
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sr_dynamic_resource.c                                  */
/*  DESCRIPTION:  Dynamic resource solving                                   */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"
#include "howard_p.h"
#include <limits.h>
#include <float.h>
#include <math.h>
#define bool_show(x) ((x) ? "true" : "false")
#define min(x, y) ((x) < (y) ? (x) : (y))
#define max(x, y) ((x) > (y) ? (x) : (y))

#define UNDEF_INT -1
#define UNDEF_FLOAT -1.0

#define DEBUG1 1		/* main functions		*/
#define DEBUG2 0		/* final solutions		*/
#define DEBUG3 0		/* checking adopted solutions	*/
#define DEBUG4 0		/* KheDrsNodeMake	*/
#define DEBUG5 0		/* KheDynamicResourceSolverDebug*/
#define DEBUG6 0		/* KheDrsFullSolnAdopt		*/
#define DEBUG7 0		/* running time (testing only)	*/
#define DEBUG8 0		/* manual tests               	*/
#define DEBUG9 0		/* hash table efficiency      	*/
#define DEBUG10 0		/* day dom tests	      	*/
#define DEBUG11 1		/* brief print of costs		*/
#define DEBUG12 0		/* trie dominance tests 	*/
#define DEBUG13 0		/* monitor sorting              */
#define DEBUG14 1		/* cluster minimum solving      */
#define DEBUG15 1		/* task classes                 */

#define DEBUG_NODE(node) (HaArrayFirst((node)->signature) == 1 &&	\
  (node)->day->day_index == 27)

#define TESTING 1

/*****************************************************************************/
/*                                                                           */
/*  Submodule "type declarations for expressions, etc."                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_KIND - the kind of dominance testing being used              */
/*                                                                           */
/*****************************************************************************/

/* *** defined in khe_solvers.h
typedef enum {
  KHE_DRS_DOM_WEAK,
  KHE_DRS_DOM_MEDIUM,
  KHE_DRS_DOM_STRONG,
  KHE_DRS_DOM_TRIE,
} KHE_DRS_DOM_KIND;
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TYPE - the type of a dominance test                          */
/*                                                                           */
/*****************************************************************************/

typedef enum {
  KHE_DRS_DOM_UNUSED,		/* "-"      should never be used      */
  KHE_DRS_DOM_GE,		/* ">="     min limit only            */
  KHE_DRS_DOM_LE,		/* "<="     max limit only            */
  KHE_DRS_DOM_EQ,		/* "==",    for uncertain cases       */
  KHE_DRS_DOM_GE_LOOSE,		/* ">=*"    min limit with loosening  */
  KHE_DRS_DOM_EQ_LOOSE		/* "==*"    max + min with loosening  */
} KHE_DRS_DOM_TYPE;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST - a dominance test                                      */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_dom_test_rec {
  KHE_DRS_DOM_TYPE	type;			/* type of test              */
  int			min_limit;		/* GE_LOOSE and EQ_LOOSE     */
} KHE_DRS_DOM_TEST;

typedef HA_ARRAY(KHE_DRS_DOM_TEST) ARRAY_KHE_DRS_DOM_TEST;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_TAG                                                         */
/*                                                                           */
/*  Tags for expression types.                                               */
/*                                                                           */
/*  Subtypes of KHE_DRS_EXPR_INT have a stored value of type int on their    */
/*  active days; subtypes of KHE_DRS_EXPR_FLOAT have a stored value of type  */
/*  float.  These are not always the same as the result type of the node.    */
/*                                                                           */
/*****************************************************************************/

typedef enum {
  KHE_DRS_EXPR_BUSY_TIME_TAG,		/* subtype of KHE_DRS_EXPR_INT	     */
  KHE_DRS_EXPR_FREE_TIME_TAG,		/* subtype of KHE_DRS_EXPR_INT	     */
  KHE_DRS_EXPR_WORK_TIME_TAG,		/* subtype of KHE_DRS_EXPR_FLOAT     */
  KHE_DRS_EXPR_BUSY_DAY_TAG,		/* subtype of KHE_DRS_EXPR_INT	     */
  KHE_DRS_EXPR_FREE_DAY_TAG,		/* subtype of KHE_DRS_EXPR_INT	     */
  KHE_DRS_EXPR_WORK_DAY_TAG,		/* subtype of KHE_DRS_EXPR_FLOAT     */
  KHE_DRS_EXPR_ASSIGNED_TASK_TAG,	/* subtype of KHE_DRS_EXPR_INT       */
  KHE_DRS_EXPR_OR_TAG,			/* subtype of KHE_DRS_EXPR_INT	     */
  KHE_DRS_EXPR_AND_TAG,			/* subtype of KHE_DRS_EXPR_INT	     */
  KHE_DRS_EXPR_INT_SUM_TAG,		/* subtype of KHE_DRS_EXPR_INT	     */
  /* KHE_DRS_EXPR_INT_SEQ_TAG, */
  KHE_DRS_EXPR_FLOAT_SUM_TAG,		/* subtype of KHE_DRS_EXPR_FLOAT     */
  KHE_DRS_EXPR_INT_DEV_TAG,
  KHE_DRS_EXPR_FLOAT_DEV_TAG,
  KHE_DRS_EXPR_COST_TAG,
  /* KHE_DRS_EXPR_COST_SUM_TAG, */
  KHE_DRS_EXPR_INT_SUM_COMB_TAG,	/* subtype of KHE_DRS_EXPR_INT	     */
  KHE_DRS_EXPR_INT_SEQ_COMB_TAG		/* subtype of KHE_DRS_EXPR_INT	     */
} KHE_DRS_EXPR_TAG;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_COMB_TYPE - types of comb nodes                                  */
/*                                                                           */
/*****************************************************************************/

typedef enum {
  KHE_DRS_COMB_ORDINARY,
  KHE_DRS_COMB_NO_MAX,
  KHE_DRS_COMB_LINEAR,
  KHE_DRS_COMB_STEP
} KHE_DRS_COMB_TYPE;


/*****************************************************************************/
/*                                                                           */
/*  INHERIT_KHE_DRS_EXPR - fields inherited by all expression types          */
/*                                                                           */
/*  For each active day di of e, except its very last active day,            */
/*  sig_index[di - e->first_day_index] is the position of e's value in       */
/*  the signatures of nodes ending on day di.                                */
/*                                                                           */
/*  For each active day di of e, child_index[di - e->first_day_index] is     */
/*  the index of the first child whose last_day_index is di or more.  There  */
/*  is also a sentinel value, so for each active day di, the iterator        */
/*                                                                           */
/*    start_child_index = child_index[di - e->first_day_index];              */
/*    stop_child_index  = child_index[di + 1 - e->first_day_index];          */
/*    for( i = start_child_index;  i < stop_child_index;  i++ )              */
/*      child_e = HaArray(e->children, i);                                   */
/*                                                                           */
/*  visits the children of e whose last_day_index is di.                     */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_rec *KHE_DRS_EXPR;
typedef HA_ARRAY(KHE_DRS_EXPR) ARRAY_KHE_DRS_EXPR;

#define INHERIT_KHE_DRS_EXPR						\
  KHE_DRS_EXPR_TAG	tag;			/* tag field      */	\
  KHE_DRS_DOM_TEST	dom_test;		/* dom test       */	\
  ARRAY_KHE_DRS_EXPR	children;		/* child exprs    */	\
  int			first_day_index;	/* first day      */	\
  int			last_day_index;		/* last day       */	\
  HA_ARRAY_INT		sig_indexes;		/* for each day   */	\
  HA_ARRAY_INT		child_indexes;		/* for each day   */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT - abstract supertype of integer-valued exprs            */
/*                                                                           */
/*****************************************************************************/

#define INHERIT_KHE_DRS_EXPR_INT					\
  INHERIT_KHE_DRS_EXPR							\
  int				value;		/* stored value */

typedef struct khe_drs_expr_int_rec {
  INHERIT_KHE_DRS_EXPR_INT
} *KHE_DRS_EXPR_INT;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FLOAT - abstract supertype of float-valued exprs            */
/*                                                                           */
/*****************************************************************************/

#define INHERIT_KHE_DRS_EXPR_FLOAT					\
  INHERIT_KHE_DRS_EXPR							\
  float				value;		/* stored value */

typedef struct khe_drs_expr_float_rec {
  INHERIT_KHE_DRS_EXPR_FLOAT
} *KHE_DRS_EXPR_FLOAT;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_BUSY_TIME                                                   */
/*                                                                           */
/*  An expression whose value is 1 when the resource is busy at a given      */
/*  time, and 0 when the resource is free then.  The expression stores the   */
/*  time, but only for debugging.                                            */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_busy_time_rec {
  INHERIT_KHE_DRS_EXPR_INT
  KHE_TIME			time;		/* the time we're busy at    */
} *KHE_DRS_EXPR_BUSY_TIME;

/* typedef HA_ARRAY(KHE_DRS_EXPR_BUSY_TIME) ARRAY_KHE_DRS_EXPR_BUSY_TIME; */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FREE_TIME                                                   */
/*                                                                           */
/*  An expression whose value is 1 when the resource is free at a given      */
/*  time, and 0 when the resource is busy then.  The expression stores the   */
/*  time, but only for debugging.                                            */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_free_time_rec {
  INHERIT_KHE_DRS_EXPR_INT
  KHE_TIME			time;		/* the time we're free at    */
} *KHE_DRS_EXPR_FREE_TIME;

/* typedef HA_ARRAY(KHE_DRS_EXPR_FREE_TIME) ARRAY_KHE_DRS_EXPR_FREE_TIME; */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_WORK_TIME                                                   */
/*                                                                           */
/*  An expression whose value is the workload incurred by the resource at    */
/*  a given time.  This will be 0.0 if the resource is free then.  The       */
/*  expression stores the time, but only for debugging.                      */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_work_time_rec {
  INHERIT_KHE_DRS_EXPR_FLOAT
  KHE_TIME			time;		/* the time of this workload */
} *KHE_DRS_EXPR_WORK_TIME;

/* typedef HA_ARRAY(KHE_DRS_EXPR_WORK_TIME) ARRAY_KHE_DRS_EXPR_WORK_TIME; */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_BUSY_DAY                                                    */
/*                                                                           */
/*  An expression whose value is 1 when the resource is busy on a given      */
/*  day, and 0 when the resource is free then.  The expression stores the    */
/*  day (as a time group), but only for debugging.                           */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_busy_day_rec {
  INHERIT_KHE_DRS_EXPR_INT
  KHE_TIME_GROUP		time_group;	/* the day we're busy at     */
} *KHE_DRS_EXPR_BUSY_DAY;

/* typedef HA_ARRAY(KHE_DRS_EXPR_BUSY_DAY) ARRAY_KHE_DRS_EXPR_BUSY_DAY; */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FREE_DAY                                                    */
/*                                                                           */
/*  An expression whose value is 1 when the resource is free on a given      */
/*  day, and 0 when the resource is busy then.  The expression stores the    */
/*  day (as a time group), but only for debugging.                           */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_free_day_rec {
  INHERIT_KHE_DRS_EXPR_INT
  KHE_TIME_GROUP		time_group;	/* the day we're free at     */
} *KHE_DRS_EXPR_FREE_DAY;

/* typedef HA_ARRAY(KHE_DRS_EXPR_FREE_DAY) ARRAY_KHE_DRS_EXPR_FREE_DAY; */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_WORK_DAY                                                    */
/*                                                                           */
/*  An expression whose value is the workload incurred by the resource on    */
/*  a given day.  This will be 0.0 if the resource is free then.  The        */
/*  expression stores the day (as a time group), but only for debugging.     */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_work_day_rec {
  INHERIT_KHE_DRS_EXPR_FLOAT
  KHE_TIME_GROUP		time_group;	/* the day we're free at     */
} *KHE_DRS_EXPR_WORK_DAY;

/* typedef HA_ARRAY(KHE_DRS_EXPR_WORK_DAY) ARRAY_KHE_DRS_EXPR_WORK_DAY; */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_ASSIGNED_TASK                                               */
/*                                                                           */
/*  An expression whose value is the duration of the given task if that      */
/*  task is assigned a resource from the given resource group, and 0         */
/*  otherwise.  The task may be a compound task.                             */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_assigned_task_rec {
  INHERIT_KHE_DRS_EXPR_INT
  int				duration;
  KHE_RESOURCE_GROUP		resource_group;
} *KHE_DRS_EXPR_ASSIGNED_TASK;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_OR                                                          */
/*                                                                           */
/*  An expression whose value is the logical `or' of its children's values.  */
/*  Here Booleans are represented by integers (0 for false, 1 for true).     */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_or_rec {
  INHERIT_KHE_DRS_EXPR_INT
} *KHE_DRS_EXPR_OR;

/* typedef HA_ARRAY(KHE_DRS_EXPR_OR) ARRAY_KHE_DRS_EXPR_OR; */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_AND                                                         */
/*                                                                           */
/*  An expression whose value is the logical `and' of its children's values. */
/*  Here Booleans are represented by integers (0 for false, 1 for true).     */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_and_rec {
  INHERIT_KHE_DRS_EXPR_INT
} *KHE_DRS_EXPR_AND;

/* typedef HA_ARRAY(KHE_DRS_EXPR_AND) ARRAY_KHE_DRS_EXPR_AND; */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SUM                                                     */
/*                                                                           */
/*  An expression whose value is the sum of its children's values.           */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_int_sum_rec {
  INHERIT_KHE_DRS_EXPR_INT
} *KHE_DRS_EXPR_INT_SUM;

/* typedef HA_ARRAY(KHE_DRS_EXPR_INT_SUM) ARRAY_KHE_DRS_EXPR_INT_SUM; */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FLOAT_SUM                                                   */
/*                                                                           */
/*  An expression whose value is the sum of its children's values.  The      */
/*  expression's value, and its children's values, have type float.          */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_float_sum_rec {
  INHERIT_KHE_DRS_EXPR_FLOAT
} *KHE_DRS_EXPR_FLOAT_SUM;

/* typedef HA_ARRAY(KHE_DRS_EXPR_FLOAT_SUM) ARRAY_KHE_DRS_EXPR_FLOAT_SUM; */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_DEV                                                     */
/*                                                                           */
/*  An expression whose value is the deviation of its integer-valued child.  */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_int_dev_rec {
  INHERIT_KHE_DRS_EXPR_INT
  int				min_limit;	/* minimum limit             */
  int				max_limit;	/* maximum limit             */
  bool				allow_zero;	/* true if 0 has deviation 0 */
} *KHE_DRS_EXPR_INT_DEV;

/* typedef HA_ARRAY(KHE_DRS_EXPR_INT_DEV) ARRAY_KHE_DRS_EXPR_INT_DEV; */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FLOAT_DEV                                                   */
/*                                                                           */
/*  An expression whose value is the integer-valued deviation of its         */
/*  float-valued child.                                                      */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_float_dev_rec {
  INHERIT_KHE_DRS_EXPR_INT
  int				min_limit;	/* minimum limit             */
  int				max_limit;	/* maximum limit             */
  bool				allow_zero;	/* true if 0 has deviation 0 */
} *KHE_DRS_EXPR_FLOAT_DEV;

/* typedef HA_ARRAY(KHE_DRS_EXPR_FLOAT_DEV) ARRAY_KHE_DRS_EXPR_FLOAT_DEV; */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_COST                                                        */
/*                                                                           */
/*  An expression whose value is a cost based on its sole child's deviation. */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_cost_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_MONITOR			monitor;	/* the monitor (debug only)  */
  KHE_COST_FUNCTION		cost_fn;	/* the cost function         */
  KHE_COST			combined_weight; /* the combined weight      */
} *KHE_DRS_EXPR_COST;

/* typedef HA_ARRAY(KHE_DRS_EXPR_COST) ARRAY_KHE_DRS_EXPR_COST; */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_COST_SUM                                                    */
/*                                                                           */
/*  An expression whose value is the sum of the values of its children.      */
/*  All these values are costs.                                              */
/*                                                                           */
/*****************************************************************************/

/* ***
typedef struct khe_drs_expr_cost_sum_rec {
  INHERIT_KHE_DRS_EXPR
} *KHE_DRS_EXPR_COST_SUM;

typedef HA_ARRAY(KHE_DRS_EXPR_COST_SUM) ARRAY_KHE_DRS_EXPR_COST_SUM;
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SUM_COMB                                                */
/*                                                                           */
/*  Combines INT_SUM, INT_DEV, and COST into a single node.  This allows     */
/*  for important optimizations in some cases, depending on comb_type.       */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_int_sum_comb_rec {
  INHERIT_KHE_DRS_EXPR_INT
  KHE_MONITOR			monitor;	/* the monitor (debug only)  */
  KHE_COST_FUNCTION		cost_fn;	/* the cost function         */
  KHE_COST			combined_weight; /* the combined weight      */
  int				min_limit;	/* minimum limit             */
  int				max_limit;	/* maximum limit             */
  bool				allow_zero;	/* true if 0 has deviation 0 */
  KHE_DRS_COMB_TYPE		comb_type;	/* type of combined value    */
  int				history_before;	/* history value before      */
  int				history_after;	/* history value after       */
} *KHE_DRS_EXPR_INT_SUM_COMB;

/* ***
typedef HA_ARRAY(KHE_DRS_EXPR_INT_SUM_COMB) ARRAY_KHE_DRS_EXPR_INT_SUM_COMB;
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SEQ_COMB                                                */
/*                                                                           */
/*  Combines INT_SEQ, INT_DEV (without allow_zero), and COST into a single   */
/*  node.  This allows the same optimizations as KHE_DRS_EXPR_INT_SUM_COMB.  */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_int_seq_comb_rec {
  INHERIT_KHE_DRS_EXPR_INT
  KHE_MONITOR			monitor;	/* the monitor (debug only)  */
  KHE_COST_FUNCTION		cost_fn;	/* the cost function         */
  KHE_COST			combined_weight; /* the combined weight      */
  int				min_limit;	/* minimum limit             */
  int				max_limit;	/* maximum limit             */
  KHE_DRS_COMB_TYPE		comb_type;	/* type of combined value    */
  int				history_before;	/* history value before      */
  int				history_after;	/* history value after       */
} *KHE_DRS_EXPR_INT_SEQ_COMB;

/* ***
typedef HA_ARRAY(KHE_DRS_EXPR_INT_SEQ_COMB) ARRAY_KHE_DRS_EXPR_INT_SEQ_COMB;
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR - abstract supertype of expressions                         */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_expr_rec {
  INHERIT_KHE_DRS_EXPR
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "type declarations for resources, tasks, etc."                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DAY_RANGE - an interval of days                                  */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_interval_rec {
  int			first_day_index;
  int			last_day_index;
} KHE_DRS_DAY_RANGE;

typedef HA_ARRAY(KHE_DRS_DAY_RANGE) ARRAY_KHE_DRS_DAY_RANGE;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE_ON_DAY - what a resource is doing on one day            */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_task_rec *KHE_DRS_TASK;

typedef struct khe_drs_resource_on_day_rec {
  KHE_DRS_TASK			assigned_task;		/* initial asst      */
  KHE_DRS_TASK			solve_task;		/* fixed if non-NULL */
  ARRAY_KHE_DRS_EXPR		active_today;		/* active today      */
  ARRAY_KHE_DRS_EXPR		leaf_today;		/* leaf today        */
  int				sig_len;		/* length of sig     */
  int				sig_offset;		/* offset on day     */
  ARRAY_KHE_DRS_DOM_TEST	dom_tests;		/* for each sig pos  */
  HA_ARRAY_INT			eq_dom_test_indexes;	/* where = tests are */
} *KHE_DRS_RESOURCE_ON_DAY;

typedef HA_ARRAY(KHE_DRS_RESOURCE_ON_DAY) ARRAY_KHE_DRS_RESOURCE_ON_DAY;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE - one resource                                          */
/*                                                                           */
/*  A resource is selected for solving when solve_day_ranges is non-empty.   */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_resource_rec {
  KHE_RESOURCE			resource;		/* the resource      */
  ARRAY_KHE_DRS_EXPR		exprs;			/* from monitors     */
  /* ARRAY_KHE_DRS_DAY_RANGE	solve_day_ranges; */	/* to be assigned    */
  ARRAY_KHE_DRS_RESOURCE_ON_DAY	days;			/* doing each day    */
} *KHE_DRS_RESOURCE;

typedef HA_ARRAY(KHE_DRS_RESOURCE) ARRAY_KHE_DRS_RESOURCE;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK - one task, or possibly a free day                          */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_task_rec {
  KHE_TASK			task;
  KHE_TASK_CLASS		orig_task_class;
  /* links to expression tree nodes still to do */
};

typedef HA_ARRAY(KHE_DRS_TASK) ARRAY_KHE_DRS_TASK;

/* *** obsolete (predates task classes)
typedef struct khe_drs_task_group_rec *KHE_DRS_TASK_GROUP;

struct khe_drs_task_rec {
  KHE_TASK			task;			** the task, or NULL **
  KHE_DRS_TASK_GROUP		task_group;		** native task group **
  KHE_DRS_DAY_RANGE		day_interval;		** the days it covers**
  KHE_DAILY_SCHEDULE		daily_schedule;		** its daily schedule**
  KHE_COST			cost_reduction;		** its cost reduction**
  int				domain_count;		** its domain size   **
  ** bool			fixed; **		** fixed to resource **
  bool				already_assigned;	** already assigned  **
};
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK_CLASS - a set of classes, a subset of one KHE_TASK_CLASS    */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_task_class_rec {
  KHE_TASK_CLASS		orig_task_class;	/* origin of this    */
  ARRAY_KHE_DRS_TASK		tasks;			/* open tasks        */
  int				tmp_used;		/* used by curr edge */
} *KHE_DRS_TASK_CLASS;

typedef HA_ARRAY(KHE_DRS_TASK_CLASS) ARRAY_KHE_DRS_TASK_CLASS;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK_GROUP - a set of equivalent tasks, sorted best first        */
/*                                                                           */
/*  Here all_tasks contains all tasks that suit the group, and solve_tasks   */
/*  is a subset of all_tasks which is available for assignment during the    */
/*  current solve.                                                           */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete (replaced by task classes)
struct khe_drs_task_group_rec {
  ARRAY_KHE_DRS_TASK		all_tasks;	** tasks that suit the group **
  ARRAY_KHE_DRS_TASK		solve_tasks;	** all_tasks avail for asst  **
};

typedef HA_ARRAY(KHE_DRS_TASK_GROUP) ARRAY_KHE_DRS_TASK_GROUP;
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EDGE - one edge, connecting one solution to the next             */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_edge_rec {
  ARRAY_KHE_DRS_TASK		tasks;		/* indexed by resources      */
} *KHE_DRS_EDGE;

typedef HA_ARRAY(KHE_DRS_EDGE) ARRAY_KHE_DRS_EDGE;


/*****************************************************************************/
/*                                                                           */
/*  Submodule "type declarations for nodes, days, and solvers"               */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TIME - one time                                                  */
/*                                                                           */
/*****************************************************************************/

/* ***
typedef struct khe_drs_time_rec {
  KHE_TIME			time;		** the time represented      **
  ARRAY_KHE_DRS_TASK		best_tasks;	** best tasks starting here  **
} *KHE_DRS_TIME;

typedef HA_ARRAY(KHE_DRS_TIME) ARRAY_KHE_DRS_TIME;
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_NODE - a node of the search tree, containing a partial solution  */
/*                                                                           */
/*  The task field is a bit tricky.  It is NULL only in the overall root     */
/*  node, the one containing an empty partial solution.  In all other nodes  */
/*  it is non-NULL, but it may have a NULL task field, in which case it      */
/*  represents a free day.                                                   */
/*                                                                           */
/*****************************************************************************/
typedef struct khe_drs_day_rec *KHE_DRS_DAY;

typedef struct khe_drs_node_rec *KHE_DRS_NODE;

struct khe_drs_node_rec {
  KHE_DRS_NODE		prev_node;		/* previous node             */
  KHE_DRS_EDGE		prev_edge;		/* edge from prev_node       */
  KHE_DRS_DAY		day;			/* day containing this node  */
  int			best_of;		/* how many this is best of  */
  KHE_COST		cost;			/* resource cost so far      */
  HA_ARRAY_INT		signature;		/* signature of this node    */
};

typedef HA_ARRAY(KHE_DRS_NODE) ARRAY_KHE_DRS_NODE;
typedef HP_TABLE(KHE_DRS_NODE) TABLE_KHE_DRS_NODE;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_NODES - a list of nodes                                       */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_node_set_rec {
  ARRAY_KHE_DRS_NODE	nodes;			/* the nodes                 */
} *KHE_DRS_NODES;

typedef HA_ARRAY(KHE_DRS_NODES) ARRAY_KHE_DRS_NODES;
typedef HP_TABLE(KHE_DRS_NODES) TABLE_KHE_DRS_NODES;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_NODE_TRIE - a trie of nodes                                      */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_node_trie_rec *KHE_DRS_NODE_TRIE;

typedef HA_ARRAY(KHE_DRS_NODE_TRIE) ARRAY_KHE_DRS_NODE_TRIE;

struct khe_drs_node_trie_rec {
  KHE_DRS_NODE			node;
  ARRAY_KHE_DRS_NODE_TRIE	children;
};


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_NODE_SET - a set of nodes, with a choice of implementations      */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_day_node_set_rec {
  int				count;			/* number of nodes   */
  /* KHE_TIME_GROUP		time_group; */		/* for debugging only*/
  TABLE_KHE_DRS_NODE		weak_dom_set;		/* weak dominance    */
  TABLE_KHE_DRS_NODES		medium_dom_set;		/* medium dominance  */
  ARRAY_KHE_DRS_NODES		strong_dom_set;		/* strong dominance  */
  KHE_DRS_NODE_TRIE		trie_dom_set;		/* trie dominance    */
} *KHE_DRS_NODE_SET;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DAY - one day                                                    */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_day_rec {
  int				day_index;		/* index of this day */
  KHE_TIME_GROUP		time_group;		/* day's time group  */
  /* ARRAY_KHE_DRS_TIME		times; */		/* times of this day */
  /* *** moved to resources
  int				sig_len;		** length of sig     **
  ARRAY_KHE_DRS_EXPR		active_today;		** active on this day**
  ARRAY_KHE_DRS_EXPR		leaf_today;		** leaf expressions  **
  *** */
  ARRAY_KHE_DRS_DOM_TEST	dom_tests;		/* from resources    */
  HA_ARRAY_INT			eq_dom_test_indexes;	/* from resources    */
  ARRAY_KHE_DRS_TASK_CLASS	task_classes;		/* today's tasks     */
  KHE_DRS_TASK			free_task;		/* free on this day  */
  KHE_DRS_NODE_SET		node_set;		/* today's nodes     */
};

typedef HA_ARRAY(KHE_DRS_DAY) ARRAY_KHE_DRS_DAY;


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_FULL_SOLN - used when adopting and unadopting complete solutions */
/*                                                                           */
/*****************************************************************************/

/* ***
typedef struct khe_drs_full_soln_rec {
  KHE_RESOURCE			resource;
  int				asst_count;
  ARRAY_KHE_DRS_TASK		tasks;
  bool				adopted;
} *KHE_DRS_FULL_SOLN;
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DYNAMIC_RESOURCE_SOLVER - a dynamic resource solver                  */
/*                                                                           */
/*****************************************************************************/

typedef enum khe_drs_solve_state {
  KHE_DRS_SOLVE_LOADING,
  KHE_DRS_SOLVE_SOLVED
} KHE_DRS_SOLVE_STATE;

struct khe_dynamic_resource_solver_rec {

  /* fields which are constant throughout the lifetime of the solver */
  HA_ARENA			arena;
  KHE_SOLN			soln;
  KHE_RESOURCE_TYPE		resource_type;
  KHE_OPTIONS			options;
  KHE_FRAME			days_frame;
  KHE_EVENT_TIMETABLE_MONITOR	etm;
  KHE_TASK_FINDER		task_finder;
  KHE_TASK_CLASS_SOLVER		task_class_solver;

  /* free list fields */
  /* *** no longer deleting expressiopns
  ARRAY_KHE_DRS_EXPR_BUSY_TIME	free_expr_busy_time;
  ARRAY_KHE_DRS_EXPR_FREE_TIME	free_expr_free_time;
  ARRAY_KHE_DRS_EXPR_WORK_TIME	free_expr_work_time;
  ARRAY_KHE_DRS_EXPR_BUSY_DAY	free_expr_busy_day;
  ARRAY_KHE_DRS_EXPR_FREE_DAY	free_expr_free_day;
  ARRAY_KHE_DRS_EXPR_WORK_DAY	free_expr_work_day;
  ARRAY_KHE_DRS_EXPR_OR		free_expr_or;
  ARRAY_KHE_DRS_EXPR_AND	free_expr_and;
  ARRAY_KHE_DRS_EXPR_INT_SUM	free_expr_int_sum;
  ARRAY_KHE_DRS_EXPR_FLOAT_SUM	free_expr_float_sum;
  ARRAY_KHE_DRS_EXPR_INT_DEV	free_expr_int_dev;
  ARRAY_KHE_DRS_EXPR_FLOAT_DEV	free_expr_float_dev;
  ARRAY_KHE_DRS_EXPR_COST	free_expr_cost;
  ARRAY_KHE_DRS_EXPR_INT_SUM_COMB free_expr_int_sum_comb;
  ARRAY_KHE_DRS_EXPR_INT_SEQ_COMB free_expr_int_seq_comb;
  *** */
  ARRAY_KHE_DRS_TASK_CLASS	free_task_classes;
  ARRAY_KHE_DRS_TASK		free_tasks;
  ARRAY_KHE_DRS_EDGE		free_edges;
  ARRAY_KHE_DRS_NODE		free_nodes;
  ARRAY_KHE_DRS_NODES	free_node_sets;
  ARRAY_KHE_DRS_NODE_TRIE	free_node_tries;

  /* fields which are constant themselves but contain varying elements */
  ARRAY_KHE_DRS_RESOURCE	all_resources;
  ARRAY_KHE_DRS_DAY		days;

  /* fields that vary with the solve */
  KHE_DRS_SOLVE_STATE		solve_state;
  /* KHE_CLUSTER_MINIMUM_SOLVER	cluster_solver; */
  int				diversifier;
  ARRAY_KHE_DRS_RESOURCE	solve_resources;  /* selected for solving */
  ARRAY_KHE_DRS_DAY_RANGE	solve_day_ranges;  /* selected for solving */
  KHE_DRS_DOM_KIND		dom_kind;
  /* int			min_assts; */
  /* int			max_assts; */
  /* KHE_COST			cost_limit; */
  /* ARRAY_KHE_DRS_EXPR		exprs; */
  ARRAY_KHE_DRS_NODE	final_solns;
  /* KHE_DRS_FULL_SOLN		full_soln; */
  bool				rerun;

#if TESTING
  KHE_TIMER			timer;
  HA_ARRAY_FLOAT		running_time_per_day;
#endif
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_DOM_TYPE and KHE_DRS_DOM_TEST"                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsDomTestMake(KHE_DRS_DOM_TYPE type)                */
/*                                                                           */
/*  Make and return a new dominance test object with no min limit attribute. */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DOM_TEST KheDrsDomTestMake(KHE_DRS_DOM_TYPE type)
{
  KHE_DRS_DOM_TEST res;
  res.type = type;
  res.min_limit = -1;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsDomTestMakeLoose(KHE_DRS_DOM_TYPE type, int val)  */
/*                                                                           */
/*  Make and return a new dominance test object with a min limit attribute.  */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DOM_TEST KheDrsDomTestMakeLoose(KHE_DRS_DOM_TYPE type, int val)
{
  KHE_DRS_DOM_TEST res;
  res.type = type;
  res.min_limit = val;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDomTestEqual(KHE_DRS_DOM_TEST dt1, KHE_DRS_DOM_TEST dt2)      */
/*                                                                           */
/*  Return true if dt1 and dt2 are equal.  This function assumes, as it      */
/*  may, that when the min_limit field is unused its value is always -1.     */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsDomTestEqual(KHE_DRS_DOM_TEST dt1, KHE_DRS_DOM_TEST dt2)
{
  return dt1.type == dt2.type && dt1.min_limit == dt2.min_limit;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDomTestDominates(KHE_DRS_DOM_TEST dt, int val1, int val2)     */
/*                                                                           */
/*  Return true if val1 dominates val2, according to dt.                     */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsDomTestDominates(KHE_DRS_DOM_TEST dt, int val1, int val2)
{
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      /* should never be called */
      HnAbort("KheDrsDomTestDominates internal error (KHE_DRS_DOM_UNUSED)");
      return false;  /* keep compiler happy */

    case KHE_DRS_DOM_GE:

      /* tight min limit */
      return val1 >= val2;

    case KHE_DRS_DOM_LE:

      /* max limit */
      return val1 <= val2;

    case KHE_DRS_DOM_EQ:

      /* values must be equal */
      return val1 == val2;

    case KHE_DRS_DOM_GE_LOOSE:

      /* loose min limit */
      return val1 >= val2 || val1 >= dt.min_limit;

    case KHE_DRS_DOM_EQ_LOOSE:

      /* max and loose min; both conditions must hold (basically equality) */
      return (val1 <= val2) && (val1 >= val2 || val1 >= dt.min_limit);

    default:

      HnAbort("KheDrsDomTestDominates internal error (%d)", dt.type);
      return false;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDomTestDominatingRange(KHE_DRS_DOM_TEST dt, int val,          */
/*    int *from_val, int *to_val)                                            */
/*                                                                           */
/*  Set [*from_val, *to_val] to the range of values that dominate val        */
/*  under dt.  That is, set it to the largest range of values x such that    */
/*                                                                           */
/*    KheDrsDomTestDominates(dt, x, val)                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDomTestDominatingRange(KHE_DRS_DOM_TEST dt, int val,
  int *from_val, int *to_val)
{
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      /* should never be called */
      HnAbort("KheDrsDomTestDominatingRange internal error (DOM_UNUSED)");
      *from_val = *to_val = 0;  /* keep compiler happy */
      break;

    case KHE_DRS_DOM_GE:

      /* largest range of values x such that x >= val */
      *from_val = val, *to_val = INT_MAX;
      break;

    case KHE_DRS_DOM_LE:

      /* largest range of values x such that x <= val */
      *from_val = 0, *to_val = val;
      break;

    case KHE_DRS_DOM_EQ:

      /* largest range of values x such that x == val */
      *from_val = val, *to_val = val;
      break;

    case KHE_DRS_DOM_GE_LOOSE:

      /* largest range of values x such that x >= val || x >= dt.min_limit */
      /* i.e. such that x >= min(val, dt.min_limit) */
      *from_val = min(val, dt.min_limit), *to_val = INT_MAX;
      break;

    case KHE_DRS_DOM_EQ_LOOSE:

      /* largest range of values x st x <= val && x >= min(val, dt.min_limit) */
      *from_val = min(val, dt.min_limit), *to_val = val;
      break;

    default:

      HnAbort("KheDrsDomTestDominatingRange internal error (%d)", dt.type);
      *from_val = *to_val = 0;  /* keep compiler happy */
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDomTestDominatedRange(KHE_DRS_DOM_TEST dt, int val,           */
/*    int *from_val, int *to_val)                                            */
/*                                                                           */
/*  Set [*from_val, *to_val] to the range of values that are dominated by    */
/*  val under dt.  That is, set it to the largest range of values x s.t.     */
/*                                                                           */
/*    KheDrsDomTestDominates(dt, val, x)                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDomTestDominatedRange(KHE_DRS_DOM_TEST dt, int val,
  int *from_val, int *to_val)
{
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      /* should never be called */
      HnAbort("KheDrsDomTestDominatedRange internal error (DOM_UNUSED)");
      *from_val = *to_val = 0;  /* keep compiler happy */
      break;

    case KHE_DRS_DOM_GE:

      /* largest range of values x such that val >= x */
      *from_val = 0, *to_val = val;
      break;

    case KHE_DRS_DOM_LE:

      /* largest range of values x such that val <= x */
      *from_val = val, *to_val = INT_MAX;
      break;

    case KHE_DRS_DOM_EQ:

      /* largest range of values x such that val == x */
      *from_val = val, *to_val = val;
      break;

    case KHE_DRS_DOM_GE_LOOSE:

      /* largest range of values x such that val >= x || val >= dt.min_limit */
      if( val >= dt.min_limit )
        *from_val = 0, *to_val = INT_MAX;
      else
        *from_val = 0, *to_val = val;
      break;

    case KHE_DRS_DOM_EQ_LOOSE:

      /* largest range of x st val <= x && (val >= x || val >= dt.min_limit) */
      if( val >= dt.min_limit )
	*from_val = val, *to_val = INT_MAX;
      else
        *from_val = val, *to_val = val;
      break;

    default:

      HnAbort("KheDrsDomTestDominatedRange internal error (%d)", dt.type);
      *from_val = *to_val = 0;  /* keep compiler happy */
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheDrsDomTestShow(KHE_DRS_DOM_TEST dt, char buff[100])             */
/*                                                                           */
/*  Show dt, using buff for scratch storage.                                 */
/*                                                                           */
/*****************************************************************************/

static char *KheDrsDomTestShow(KHE_DRS_DOM_TEST dt, char buff[100])
{
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      sprintf(buff, "-");
      break;

    case KHE_DRS_DOM_GE:

      sprintf(buff, ">=");
      break;

    case KHE_DRS_DOM_LE:

      sprintf(buff, "<=");
      break;

    case KHE_DRS_DOM_EQ:

      sprintf(buff, "==");
      break;

    case KHE_DRS_DOM_GE_LOOSE:

      sprintf(buff, ">=(%d)", dt.min_limit);
      break;

    case KHE_DRS_DOM_EQ_LOOSE:

      sprintf(buff, "=(%d)", dt.min_limit);
      break;

    default:

      sprintf(buff, "?");
      break;
  }
  return buff;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "helper code for expressions, including deviations and costs"  */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsResourceHasTimeExpr(KHE_DRS_RESOURCE dr,
  KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME time,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e);
static bool KheDrsResourceHasDayExpr(KHE_DRS_RESOURCE dr,
  KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME_GROUP tg,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e);
static void KheDrsResourceAddTimeExpr(KHE_DRS_RESOURCE dr,
  KHE_TIME time, KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR e);
static void KheDrsResourceAddDayExpr(KHE_DRS_RESOURCE dr,
  KHE_TIME_GROUP tg, KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR e);
static bool KheDrsExprInSignature(KHE_DRS_EXPR e, int day_index, int *si);
static char *KheDrsExprDayRangeShow(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, char buff[100]);
static void KheDrsExprDebugChildren(KHE_DRS_EXPR e, KHE_DRS_NODE node,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp);
static void KheDrsExprDebug(KHE_DRS_EXPR e, KHE_DRS_NODE node,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp);


/*****************************************************************************/
/*                                                                           */
/*  KheDrsExprInit(res, tg, dt, drs, free_list)                              */
/*                                                                           */
/*  Shared generic code for obtaining an expression object from a free list  */
/*  in drs and initializing those fields common to all expressions.          */
/*                                                                           */
/*****************************************************************************/

#define KheDrsExprInit(res, tg, dt, drs)				\
  HaMake(res, drs->arena);						\
  HaArrayInit(res->children, drs->arena);				\
  HaArrayInit(res->sig_indexes, drs->arena);				\
  HaArrayInit(res->child_indexes, drs->arena);				\
  res->tag = tg;							\
  res->first_day_index = res->last_day_index = UNDEF_INT;		\
  res->dom_test = dt


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSumDev(int min_limit, int max_limit, bool allow_zero,          */
/*    int history_after, int val)                                            */
/*                                                                           */
/*  Deviation function to use with sum nodes; it has allow_zero.             */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSumDev(int min_limit, int max_limit, bool allow_zero,
  int history_after, int val)
{
  if( val == 0 && allow_zero )
    return 0;
  else if( val > max_limit )
    return val - max_limit;
  else if( val + history_after < min_limit )
    return min_limit - (val + history_after);
  else
    return 0;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSeqDev(int min_limit, int max_limit,                           */
/*    int history_after, int val)                                            */
/*                                                                           */
/*  Deviation function to use with sequence nodes; it has no allow_zero.     */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSeqDev(int min_limit, int max_limit,
  int history_after, int val)
{
  if( val > max_limit )
    return val - max_limit;
  else if( val + history_after < min_limit )
    return min_limit - (val + history_after);
  else
    return 0;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_COST KheDrsCost(KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight, */
/*    int dev)                                                               */
/*                                                                           */
/*****************************************************************************/

static KHE_COST KheDrsCost(KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,
  int dev)
{
  switch( cost_fn )
  {
    case KHE_STEP_COST_FUNCTION:

        return dev > 0 ? combined_weight : 0;

    case KHE_LINEAR_COST_FUNCTION:

        return dev * combined_weight;

    case KHE_QUADRATIC_COST_FUNCTION:

	return dev * dev * combined_weight;

    default:

	HnAbort("KheDrsCost internal error");
	return KheCost(0, 0);  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_BUSY_TIME"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_BUSY_TIME KheDrsExprBusyTimeMake(KHE_TIME time,             */
/*    KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)            */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_BUSY_TIME object with these attributes.          */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_BUSY_TIME KheDrsExprBusyTimeMake(KHE_DRS_RESOURCE dr,
  KHE_TIME time, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_BUSY_TIME res;  KHE_DRS_EXPR e;
  if( KheDrsResourceHasTimeExpr(dr, KHE_DRS_EXPR_BUSY_TIME_TAG, dom_test,
	time, drs, &e) )
    return (KHE_DRS_EXPR_BUSY_TIME) e;
  KheDrsExprInit(res, KHE_DRS_EXPR_BUSY_TIME_TAG, dom_test, drs);
    /* , free_expr_busy_time); */
  res->value = UNDEF_INT;
  res->time = time;
  /* KheDrsAddExprToTime(drs, (KHE_DRS_EXPR) res, time); */
  KheDrsResourceAddTimeExpr(dr, time, drs, (KHE_DRS_EXPR) res);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_BUSY_TIME KheDrsExprBusyTimeCopy(KHE_DRS_EXPR_BUSY_TIME ebt,*/
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a fresh and unshared copy of ebt.  Do not copy the sig_indexes    */
/*  and child_indexes arrays; they should be empty anyway.                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_BUSY_TIME KheDrsExprBusyTimeCopy(KHE_DRS_EXPR_BUSY_TIME ebt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_BUSY_TIME res;
  KheDrsExprInit(res, KHE_DRS_EXPR_BUSY_TIME_TAG, ebt->dom_test, drs);
    /* , free_expr_busy_time); */
  res->first_day_index = ebt->first_day_index;
  res->last_day_index = ebt->last_day_index;
  res->value = ebt->value;
  res->time = ebt->time;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeDelete(KHE_DRS_EXPR_BUSY_TIME ebt,                */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete ebt and its descendants (there are no descendants here).          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprBusyTimeDelete(KHE_DRS_EXPR_BUSY_TIME ebt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_expr_busy_time, ebt);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeDoDay(KHE_DRS_EXPR_BUSY_TIME ebt, int day_index,  */
/*    KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active node ebt on day_index.                                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyTimeDoDay(KHE_DRS_EXPR_BUSY_TIME ebt, int day_index,
  KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;
  if( day_index == ebt->first_day_index )
  {
    /* this is ebt's first day */
    ebt->value = (time == ebt->time ? 1 : 0);
  }
  else
  {
    /* this is not ebt's first day, so we need to retrieve a previous value */
    si = HaArray(ebt->sig_indexes, day_index - 1 - ebt->first_day_index);
    ebt->value = HaArray(prev_node->signature, prev_so + si);
  }
  if( day_index < ebt->last_day_index )
  {
    /* this not ebt's last day; store value in next_node */
    si = HaArray(ebt->sig_indexes, day_index - ebt->first_day_index);
    HnAssert(HaArrayCount(next_node->signature) == next_so + si,
      "KheDrsExprBusyTimeDoDay internal error");
    HaArrayAddLast(next_node->signature, ebt->value);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeDebug(KHE_DRS_EXPR_BUSY_TIME ebt,                 */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for busy time expression ebt.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyTimeDebug(KHE_DRS_EXPR_BUSY_TIME ebt,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff1[100], buff2[100];  int si;
  if( node != NULL )
  {
    /* unindented version with a node value */
    fprintf(fp, "BUSY_TIME(%s)", KheTimeId(ebt->time));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) ebt, node->day->day_index, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(node->signature, si));
  }
  else
  {
    /* indented version without a node value */
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "[ BUSY_TIME(%s) %s %s ]", KheTimeId(ebt->time),
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) ebt, drs, buff1),
      KheDrsDomTestShow(ebt->dom_test, buff2));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_FREE_TIME"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FREE_TIME KheDrsExprFreeTimeMake(KHE_DRS_RESOURCE dr,       */
/*    KHE_TIME time, KHE_DRS_DOM_TEST dom_test,                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_FREE_TIME object with these attributes.          */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FREE_TIME KheDrsExprFreeTimeMake(KHE_DRS_RESOURCE dr,
  KHE_TIME time, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FREE_TIME res;  KHE_DRS_EXPR e;
  if( KheDrsResourceHasTimeExpr(dr, KHE_DRS_EXPR_FREE_TIME_TAG, dom_test,
      time, drs, &e) )
    return (KHE_DRS_EXPR_FREE_TIME) e;
  KheDrsExprInit(res, KHE_DRS_EXPR_FREE_TIME_TAG, dom_test, drs);
    /* free_expr_free_time); */
  res->value = UNDEF_INT;
  res->time = time;
  KheDrsResourceAddTimeExpr(dr, time, drs, (KHE_DRS_EXPR) res);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FREE_TIME KheDrsExprFreeTimeCopy(KHE_DRS_EXPR_FREE_TIME eft,*/
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a fresh and unshared copy of eft.  Do not copy the sig_indexes    */
/*  and child_indexes arrays; they should be empty anyway.                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FREE_TIME KheDrsExprFreeTimeCopy(KHE_DRS_EXPR_FREE_TIME eft,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FREE_TIME res;
  KheDrsExprInit(res, KHE_DRS_EXPR_FREE_TIME_TAG, eft->dom_test, drs);
    /* , free_expr_free_time); */
  res->first_day_index = eft->first_day_index;
  res->last_day_index = eft->last_day_index;
  res->value = eft->value;
  res->time = eft->time;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeDelete(KHE_DRS_EXPR_FREE_TIME ef,                 */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete ef and its descendants (there are no descendants here).           */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFreeTimeDelete(KHE_DRS_EXPR_FREE_TIME ef,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_expr_free_time, ef);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeDoDay(KHE_DRS_EXPR_FREE_TIME eft, int day_index,  */
/*    KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active node eft on day_index.                                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeTimeDoDay(KHE_DRS_EXPR_FREE_TIME eft, int day_index,
  KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;
  if( day_index == eft->first_day_index )
  {
    /* this is eft's first day */
    eft->value = (time == eft->time ? 0 : 1);
  }
  else
  {
    /* this is not eft's first day, so we need to retrieve a previous value */
    si = HaArray(eft->sig_indexes, day_index - 1 - eft->first_day_index);
    eft->value = HaArray(prev_node->signature, prev_so + si);
  }
  if( day_index < eft->last_day_index )
  {
    /* this not eft's last day; store value in next_node */
    si = HaArray(eft->sig_indexes, day_index - eft->first_day_index);
    HnAssert(HaArrayCount(next_node->signature) == next_so + si,
      "KheDrsExprFreeTimeDoDay internal error");
    HaArrayAddLast(next_node->signature, eft->value);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeDebug(KHE_DRS_EXPR_FREE_TIME eft,                 */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for free time expression eft.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeTimeDebug(KHE_DRS_EXPR_FREE_TIME eft,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff1[100], buff2[100];  int si;
  if( node != NULL )
  {
    /* unindented version with a node value */
    fprintf(fp, "FREE_TIME(%s)", KheTimeId(eft->time));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eft, node->day->day_index, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(node->signature, si));
  }
  else
  {
    /* indented version without a node value */
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "[ FREE_TIME(%s) %s %s ]", KheTimeId(eft->time),
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) eft, drs, buff1),
      KheDrsDomTestShow(eft->dom_test, buff2));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_WORK_TIME"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_WORK_TIME KheDrsExprWorkTimeMake(KHE_DRS_RESOURCE dr,       */
/*    KHE_TIME time, KHE_DRS_DOM_TEST dom_test,                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_WORK_TIME object with these attributes.          */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_WORK_TIME KheDrsExprWorkTimeMake(KHE_DRS_RESOURCE dr,
  KHE_TIME time, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_WORK_TIME res;  KHE_DRS_EXPR e;
  if( KheDrsResourceHasTimeExpr(dr, KHE_DRS_EXPR_WORK_TIME_TAG, dom_test,
      time, drs, &e) )
    return (KHE_DRS_EXPR_WORK_TIME) e;
  KheDrsExprInit(res, KHE_DRS_EXPR_WORK_TIME_TAG, dom_test, drs);
    /* , free_expr_work_time); */
  res->value = UNDEF_FLOAT;
  res->time = time;
  KheDrsResourceAddTimeExpr(dr, time, drs, (KHE_DRS_EXPR) res);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_WORK_TIME KheDrsExprWorkTimeCopy(KHE_DRS_EXPR_WORK_TIME ewt,*/
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a fresh and unshared copy of ewt.  Do not copy the sig_indexes    */
/*  and child_indexes arrays; they should be empty anyway.                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_WORK_TIME KheDrsExprWorkTimeCopy(KHE_DRS_EXPR_WORK_TIME ewt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_WORK_TIME res;
  KheDrsExprInit(res, KHE_DRS_EXPR_WORK_TIME_TAG, ewt->dom_test, drs);
    /* , free_expr_work_time); */
  res->first_day_index = ewt->first_day_index;
  res->last_day_index = ewt->last_day_index;
  res->value = ewt->value;
  res->time = ewt->time;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeDelete(KHE_DRS_EXPR_WORK_TIME ewt,                */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete ewt and its descendants (there are no descendants here).          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprWorkTimeDelete(KHE_DRS_EXPR_WORK_TIME ewt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_expr_work_time, ewt);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeDoDay(KHE_DRS_EXPR_WORK_TIME ewt, int day_index,  */
/*    KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active node ewt on day_index.                                    */
/*                                                                           */
/*  Implementation note.  If time == ewt->time, then since ewt->time is      */
/*  always non-NULL, time is non-NULL, and this implies task is non-NULL.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkTimeDoDay(KHE_DRS_EXPR_WORK_TIME ewt, int day_index,
  KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;
  if( day_index == ewt->first_day_index )
  {
    /* this is ewt's first day */
    ewt->value = (time == ewt->time ? KheTaskWorkloadPerTime(task) : 0.0);
  }
  else
  {
    /* this is not ewt's first day, so we need to retrieve a previous value */
    si = HaArray(ewt->sig_indexes, day_index - 1 - ewt->first_day_index);
    ewt->value = HaArray(prev_node->signature, prev_so + si);
  }
  if( day_index < ewt->last_day_index )
  {
    /* this not ewt's last day; store value in next_node */
    si = HaArray(ewt->sig_indexes, day_index - ewt->first_day_index);
    HnAssert(HaArrayCount(next_node->signature) == next_so + si,
      "KheDrsExprWorkTimeDoDay internal error");
    HaArrayAddLast(next_node->signature, ewt->value);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeDebug(KHE_DRS_EXPR_WORK_TIME ewt,                 */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for work time expression ewt.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkTimeDebug(KHE_DRS_EXPR_WORK_TIME ewt,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff1[100], buff2[100];  int si;
  if( node != NULL )
  {
    /* unindented version with a node value */
    fprintf(fp, "WORK_TIME(%s)", KheTimeId(ewt->time));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) ewt, node->day->day_index, &si) )
      fprintf(fp, " $%d:%.2f$", si, (float) HaArray(node->signature, si) / 100.0);
  }
  else
  {
    /* indented version without a node value */
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "[ WORK_TIME(%s) %s %s ]", KheTimeId(ewt->time),
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) ewt, drs, buff1),
      KheDrsDomTestShow(ewt->dom_test, buff2));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_BUSY_DAY"                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_BUSY_DAY KheDrsExprBusyDayMake(KHE_DRS_RESOURCE dr,         */
/*    KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_BUSY_DAY object with these attributes.           */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_BUSY_DAY KheDrsExprBusyDayMake(KHE_DRS_RESOURCE dr,
  KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_BUSY_DAY res;  KHE_DRS_EXPR e;
  if( KheDrsResourceHasDayExpr(dr, KHE_DRS_EXPR_BUSY_DAY_TAG, dom_test,
      tg, drs, &e) )
    return (KHE_DRS_EXPR_BUSY_DAY) e;
  KheDrsExprInit(res, KHE_DRS_EXPR_BUSY_DAY_TAG, dom_test, drs);
    /* , free_expr_busy_day); */
  res->value = UNDEF_INT;
  res->time_group = tg;
  KheDrsResourceAddDayExpr(dr, tg, drs, (KHE_DRS_EXPR) res);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_BUSY_DAY KheDrsExprBusyDayCopy(KHE_DRS_EXPR_BUSY_DAY ebd,   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a fresh and unshared copy of ebd.  Do not copy the sig_indexes    */
/*  and child_indexes arrays; they should be empty anyway.                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_BUSY_DAY KheDrsExprBusyDayCopy(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_BUSY_DAY res;
  KheDrsExprInit(res, KHE_DRS_EXPR_BUSY_DAY_TAG, ebd->dom_test, drs);
    /* , free_expr_busy_day); */
  res->first_day_index = ebd->first_day_index;
  res->last_day_index = ebd->last_day_index;
  res->value = ebd->value;
  res->time_group = ebd->time_group;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDayDelete(KHE_DRS_EXPR_BUSY_DAY ebd,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete ebd and its descendants (there are no descendants here).          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprBusyDayDelete(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_expr_busy_day, ebd);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDayDoDay(KHE_DRS_EXPR_BUSY_DAY ebd, int day_index,    */
/*    KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active node ebd on day_index.                                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyDayDoDay(KHE_DRS_EXPR_BUSY_DAY ebd, int day_index,
  KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;
  if( day_index == ebd->first_day_index )
  {
    /* this is ebd's first day */
    ebd->value = (time != NULL ? 1 : 0);
  }
  else
  {
    /* this is not ebd's first day, so we need to retrieve a previous value */
    si = HaArray(ebd->sig_indexes, day_index - 1 - ebd->first_day_index);
    ebd->value = HaArray(prev_node->signature, prev_so + si);
  }
  if( day_index < ebd->last_day_index )
  {
    /* this not ebd's last day; store value in next_node */
    si = HaArray(ebd->sig_indexes, day_index - ebd->first_day_index);
    HnAssert(HaArrayCount(next_node->signature) == next_so + si,
      "KheDrsExprBusyDayDoDay internal error");
    HaArrayAddLast(next_node->signature, ebd->value);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDayDebug(KHE_DRS_EXPR_BUSY_DAY ebd,                   */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for busy day expression ebd.                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyDayDebug(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff1[100], buff2[100];  int si;
  if( node != NULL )
  {
    /* unindented version with a node value */
    fprintf(fp, "BUSY_DAY(%s)", KheTimeGroupId(ebd->time_group));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) ebd, node->day->day_index, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(node->signature, si));
  }
  else
  {
    /* indented version without a node value */
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "[ BUSY_DAY(%s) %s %s ]", KheTimeGroupId(ebd->time_group),
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) ebd, drs, buff1),
      KheDrsDomTestShow(ebd->dom_test, buff2));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_FREE_DAY"                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FREE_DAY KheDrsExprFreeDayMake(KHE_DRS_RESOURCE dr,         */
/*    KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_FREE_DAY object with these attributes.           */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FREE_DAY KheDrsExprFreeDayMake(KHE_DRS_RESOURCE dr,
  KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FREE_DAY res;  KHE_DRS_EXPR e;
  if( KheDrsResourceHasDayExpr(dr, KHE_DRS_EXPR_FREE_DAY_TAG, dom_test,
      tg, drs, &e) )
    return (KHE_DRS_EXPR_FREE_DAY) e;
  KheDrsExprInit(res, KHE_DRS_EXPR_FREE_DAY_TAG, dom_test, drs);
    /* , free_expr_free_day); */
  res->value = UNDEF_INT;
  res->time_group = tg;
  KheDrsResourceAddDayExpr(dr, tg, drs, (KHE_DRS_EXPR) res);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FREE_DAY KheDrsExprFreeDayCopy(KHE_DRS_EXPR_FREE_DAY efd,   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a fresh and unshared copy of efd.  Do not copy the sig_indexes    */
/*  and child_indexes arrays; they should be empty anyway.                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FREE_DAY KheDrsExprFreeDayCopy(KHE_DRS_EXPR_FREE_DAY efd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FREE_DAY res;
  KheDrsExprInit(res, KHE_DRS_EXPR_FREE_DAY_TAG, efd->dom_test, drs);
    /* , free_expr_free_day); */
  res->first_day_index = efd->first_day_index;
  res->last_day_index = efd->last_day_index;
  res->value = efd->value;
  res->time_group = efd->time_group;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDayDelete(KHE_DRS_EXPR_FREE_DAY efd,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete efd and its descendants (there are no descendants here).          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFreeDayDelete(KHE_DRS_EXPR_FREE_DAY efd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_expr_free_day, efd);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDayDoDay(KHE_DRS_EXPR_FREE_DAY efd, int day_index,    */
/*    KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active node efd on day_index.                                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeDayDoDay(KHE_DRS_EXPR_FREE_DAY efd, int day_index,
  KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;
  if( day_index == efd->first_day_index )
  {
    /* this is efd's first day */
    efd->value = (time == NULL ? 1 : 0);
  }
  else
  {
    /* this is not efd's first day, so we need to retrieve a previous value */
    si = HaArray(efd->sig_indexes, day_index - 1 - efd->first_day_index);
    efd->value = HaArray(prev_node->signature, prev_so + si);
  }
  if( day_index < efd->last_day_index )
  {
    /* this not efd's last day; store value in next_node */
    si = HaArray(efd->sig_indexes, day_index - efd->first_day_index);
    HnAssert(HaArrayCount(next_node->signature) == next_so + si,
      "KheDrsExprFreeDayDoDay internal error");
    HaArrayAddLast(next_node->signature, efd->value);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDayDebug(KHE_DRS_EXPR_FREE_DAY efd,                   */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for free time expression efd.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeDayDebug(KHE_DRS_EXPR_FREE_DAY efd,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff1[100], buff2[100];  int si;
  if( node != NULL )
  {
    /* unindented version with a node value */
    fprintf(fp, "FREE_DAY(%s)", KheTimeGroupId(efd->time_group));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) efd, node->day->day_index, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(node->signature, si));
  }
  else
  {
    /* indented version without a node value */
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "[ FREE_DAY(%s) %s %s ]", KheTimeGroupId(efd->time_group),
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) efd, drs, buff1),
      KheDrsDomTestShow(efd->dom_test, buff2));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_WORK_DAY"                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_WORK_DAY KheDrsExprWorkDayMake(KHE_TIME_GROUP tg,           */
/*    KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)            */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_WORK_DAY object with these attributes.           */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused, although I really should be using it
static KHE_DRS_EXPR_WORK_DAY KheDrsExprWorkDayMake(KHE_TIME_GROUP tg,
  KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_WORK_DAY res;  KHE_DRS_EXPR e;
  if( KheDrsHasDayExpr(drs, KHE_DRS_EXPR_WORK_DAY_TAG, dom_test, tg, &e) )
    return (KHE_DRS_EXPR_WORK_DAY) e;
  KheDrsExprInit(res, KHE_DRS_EXPR_WORK_DAY_TAG, dom_test, drs,
    free_expr_work_day);
  res->value = UNDEF_INT;
  res->time_group = tg;
  KheDrsAddDayExpr(drs, tg, (KHE_DRS_EXPR) res);
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_WORK_DAY KheDrsExprWorkDayCopy(KHE_DRS_EXPR_WORK_DAY ewd,   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a fresh and unshared copy of ewd.  Do not copy the sig_indexes    */
/*  and child_indexes arrays; they should be empty anyway.                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_WORK_DAY KheDrsExprWorkDayCopy(KHE_DRS_EXPR_WORK_DAY ewd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_WORK_DAY res;
  KheDrsExprInit(res, KHE_DRS_EXPR_WORK_DAY_TAG, ewd->dom_test, drs);
    /* , free_expr_work_day); */
  res->first_day_index = ewd->first_day_index;
  res->last_day_index = ewd->last_day_index;
  res->value = ewd->value;
  res->time_group = ewd->time_group;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDayDelete(KHE_DRS_EXPR_WORK_DAY ewd,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete ewd and its descendants (there are no descendants here).          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprWorkDayDelete(KHE_DRS_EXPR_WORK_DAY ewd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_expr_work_day, ewd);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDayDoDay(KHE_DRS_EXPR_WORK_DAY ewd, int day_index,    */
/*    KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active node ewd on day_index.                                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkDayDoDay(KHE_DRS_EXPR_WORK_DAY ewd, int day_index,
  KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;
  if( day_index == ewd->first_day_index )
  {
    /* this is ewd's first day */
    ewd->value = (task != NULL ? KheTaskWorkloadPerTime(task) : 0.0);
  }
  else
  {
    /* this is not ewd's first day, so we need to retrieve a previous value */
    si = HaArray(ewd->sig_indexes, day_index - 1 - ewd->first_day_index);
    ewd->value = HaArray(prev_node->signature, prev_so + si);
  }
  if( day_index < ewd->last_day_index )
  {
    /* this not ewd's last day; store value in next_node */
    si = HaArray(ewd->sig_indexes, day_index - ewd->first_day_index);
    HnAssert(HaArrayCount(next_node->signature) == next_so + si,
      "KheDrsExprWorkDayDoDay internal error");
    HaArrayAddLast(next_node->signature, ewd->value);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDayDebug(KHE_DRS_EXPR_WORK_DAY ewd,                   */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for work day expression ewd.                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkDayDebug(KHE_DRS_EXPR_WORK_DAY ewd,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff1[100], buff2[100];  int si;
  if( node != NULL )
  {
    /* unindented version with a node value */
    fprintf(fp, "WORK_DAY(%s)", KheTimeGroupId(ewd->time_group));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) ewd, node->day->day_index, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(node->signature, si));
  }
  else
  {
    /* indented version without a node value */
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "[ WORK_DAY(%s) %s %s ]", KheTimeGroupId(ewd->time_group),
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) ewd, drs, buff1),
      KheDrsDomTestShow(ewd->dom_test, buff2));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_ASSIGNED_TASK"                                   */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_ASSIGNED_TASK KheDrsExprAssignedTaskMake(int durn,          */
/*    KHE_RESOURCE_GROUP rg, KHE_DYNAMIC_RESOURCE_SOLVER drs)                */
/*                                                                           */
/*  Make an assigned task expression object with these attributes.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_ASSIGNED_TASK KheDrsExprAssignedTaskMake(int durn,
  KHE_RESOURCE_GROUP rg, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_ASSIGNED_TASK res;  KHE_DRS_DOM_TEST dom_test;
  dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  KheDrsExprInit(res, KHE_DRS_EXPR_ASSIGNED_TASK_TAG, dom_test, drs);
  res->value = UNDEF_INT;
  res->duration = durn;
  res->resource_group = rg;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAssignedTaskDoDay(KHE_DRS_EXPR_ASSIGNED_TASK eat,         */
/*    int day_index, KHE_DRS_NODE prev_node, int prev_so, KHE_RESOURCE r,    */
/*    KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active node eat on day_index.                                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAssignedTaskDoDay(KHE_DRS_EXPR_ASSIGNED_TASK eat,
  int day_index, KHE_DRS_NODE prev_node, int prev_so, KHE_RESOURCE r,
  KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;

  /* initialize eo->value depending on whether day_index is its first day */
  if( day_index == eat->first_day_index )
  {
    if( eat->resource_group == NULL ||
	KheResourceGroupContains(eat->resource_group, r) )
      eat->value = eat->duration;
    else
      eat->value = 0;
  }
  else
  {
    /* this is not eo's first day, so we need to retrieve a previous value */
    si = HaArray(eat->sig_indexes, day_index - 1 - eat->first_day_index);
    eat->value = HaArray(prev_node->signature, prev_so + si);
  }

  if( day_index < eat->last_day_index )
  {
    /* this not eat's last day; store value in next_node */
    si = HaArray(eat->sig_indexes, day_index - eat->first_day_index);
    HnAssert(HaArrayCount(next_node->signature) == next_so + si,
      "KheDrsExprAssignedTaskDoDay internal error");
    HaArrayAddLast(next_node->signature, eat->value);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAssignedTaskDebug(KHE_DRS_EXPR_ASSIGNED_TASK eat,         */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for assigned task expression eat.              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAssignedTaskDebug(KHE_DRS_EXPR_ASSIGNED_TASK eat,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff1[100], buff2[100];  int si;
  if( node != NULL )
  {
    /* unindented version with a node value */
    fprintf(fp, "ASSIGNED_TASK");
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eat, node->day->day_index, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(node->signature, si));
  }
  else if( indent >= 0 )
  {
    /* indented version without a node value */
    fprintf(fp, "%*s[ ASSIGNED_TASK %s %s %s %d ]\n", indent, "",
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) eat, drs, buff1),
      KheDrsDomTestShow(eat->dom_test, buff2),
      KheResourceGroupId(eat->resource_group), eat->duration);
  }
  else
  {
    /* unindented version without a node value */
    fprintf(fp, "[ ASSIGNED_TASK %s %s %s %d ]",
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) eat, drs, buff1),
      KheDrsDomTestShow(eat->dom_test, buff2),
      KheResourceGroupId(eat->resource_group), eat->duration);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_OR"                                              */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_OR KheDrsExprOrMake(KHE_DRS_DOM_TEST dom_test,              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_OR object with these attributes.                 */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_OR KheDrsExprOrMake(KHE_DRS_DOM_TEST dom_test,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_OR res;
  KheDrsExprInit(res, KHE_DRS_EXPR_OR_TAG, dom_test, drs /* , free_expr_or */);
  res->value = UNDEF_INT;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrDelete(KHE_DRS_EXPR_OR eo,                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete eo and its descendants.                                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprOrDelete(KHE_DRS_EXPR_OR eo,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(eo->children, e, i)
    KheDrsExprDelete(e, drs);
  HaArrayAddLast(drs->free_expr_or, eo);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrDoDay(KHE_DRS_EXPR_OR eo, int day_index,                */
/*    KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active node eo on day_index.                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOrDoDay(KHE_DRS_EXPR_OR eo, int day_index,
  KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, start_index, stop_index;  KHE_DRS_EXPR_INT child_e;

  /* initialize eo->value depending on whether day_index is its first day */
  if( day_index == eo->first_day_index )
    eo->value = 0;
  else
  {
    /* this is not eo's first day, so we need to retrieve a previous value */
    si = HaArray(eo->sig_indexes, day_index - 1 - eo->first_day_index);
    eo->value = HaArray(prev_node->signature, prev_so + si);
  }

  /* accumulate the values of the children of eo that finalized today */
  start_index = HaArray(eo->child_indexes, day_index - eo->first_day_index);
  stop_index =  HaArray(eo->child_indexes, day_index + 1 - eo->first_day_index);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = (KHE_DRS_EXPR_INT) HaArray(eo->children, i);
    if( child_e->value == 1 )
      eo->value = 1;
  }

  if( day_index < eo->last_day_index )
  {
    /* this not eo's last day; store value in next_node */
    si = HaArray(eo->sig_indexes, day_index - eo->first_day_index);
    HnAssert(HaArrayCount(next_node->signature) == next_so + si,
      "KheDrsExprOrDoDay internal error");
    HaArrayAddLast(next_node->signature, eo->value);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrDebug(KHE_DRS_EXPR_OR eo,                               */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for or expression eo.                          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOrDebug(KHE_DRS_EXPR_OR eo,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff1[100], buff2[100];  int si;
  if( node != NULL )
  {
    /* unindented version with a node value */
    fprintf(fp, "OR");
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eo, node->day->day_index, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(node->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eo, node, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    /* indented version without a node value */
    fprintf(fp, "%*s[ OR %s %s\n", indent, "",
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) eo, drs, buff1),
      KheDrsDomTestShow(eo->dom_test, buff2));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eo, node, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    /* unindented version without a node value */
    fprintf(fp, "[ OR %s %s (%d children) ]",
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) eo, drs, buff1),
      KheDrsDomTestShow(eo->dom_test, buff2), HaArrayCount(eo->children));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_AND"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_AND KheDrsExprAndMake(KHE_DRS_DOM_TEST dom_test,            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_AND object with these attributes.                */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_AND KheDrsExprAndMake(KHE_DRS_DOM_TEST dom_test,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_AND res;
  KheDrsExprInit(res, KHE_DRS_EXPR_AND_TAG, dom_test, drs /*, free_expr_and */);
  res->value = UNDEF_INT;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndDelete(KHE_DRS_EXPR_AND ea,                            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete ea and its descendants.                                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprAndDelete(KHE_DRS_EXPR_AND ea,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(ea->children, e, i)
    KheDrsExprDelete(e, drs);
  HaArrayAddLast(drs->free_expr_and, ea);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndDoDay(KHE_DRS_EXPR_AND ea, int day_index,              */
/*    KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active node ea on day_index.                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAndDoDay(KHE_DRS_EXPR_AND ea, int day_index,
  KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, start_index, stop_index;  KHE_DRS_EXPR_INT child_e;

  /* initialize ea->value depending on whether day_index is its first day */
  if( day_index == ea->first_day_index )
    ea->value = 1;
  else
  {
    /* this is not ea's first day, so we need to retrieve a previous value */
    si = HaArray(ea->sig_indexes, day_index - 1 - ea->first_day_index);
    ea->value = HaArray(prev_node->signature, prev_so + si);
  }

  /* accumulate the values of the children of ea that finalized today */
  start_index = HaArray(ea->child_indexes, day_index - ea->first_day_index);
  stop_index =  HaArray(ea->child_indexes, day_index + 1 - ea->first_day_index);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = (KHE_DRS_EXPR_INT) HaArray(ea->children, i);
    if( child_e->value == 0 )
      ea->value = 0;
  }

  if( day_index < ea->last_day_index )
  {
    /* this not ea's last day; store value in next_node */
    si = HaArray(ea->sig_indexes, day_index - ea->first_day_index);
    HnAssert(HaArrayCount(next_node->signature) == next_so + si,
      "KheDrsExprAndDoDay internal error");
    HaArrayAddLast(next_node->signature, ea->value);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndDebug(KHE_DRS_EXPR_AND ea,                             */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for and expression ea.                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAndDebug(KHE_DRS_EXPR_AND ea,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff1[100], buff2[100];  int si;
  if( node != NULL )
  {
    /* unindented version with a node value */
    fprintf(fp, "AND");
    if( KheDrsExprInSignature((KHE_DRS_EXPR) ea, node->day->day_index, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(node->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) ea, node, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    /* indented version without a node value */
    fprintf(fp, "%*s[ AND %s %s\n", indent, "",
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) ea, drs, buff1),
      KheDrsDomTestShow(ea->dom_test, buff2));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) ea, node, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    /* unindented version without a node value */
    fprintf(fp, "[ AND %s %s (%d children) ]",
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) ea, drs, buff1),
      KheDrsDomTestShow(ea->dom_test, buff2), HaArrayCount(ea->children));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_SUM"                                         */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SUM KheDrsExprIntSumMake(KHE_DRS_DOM_TEST dom_test,     */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_SUM object with these attributes.            */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_INT_SUM KheDrsExprIntSumMake(KHE_DRS_DOM_TEST dom_test,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_SUM res;
  KheDrsExprInit(res, KHE_DRS_EXPR_INT_SUM_TAG, dom_test, drs);
    /* , free_expr_int_sum); */
  res->value = UNDEF_INT;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumDelete(KHE_DRS_EXPR_INT_SUM eis,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete eis and its descendants.                                          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumDelete(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(eis->children, e, i)
    KheDrsExprDelete(e, drs);
  HaArrayAddLast(drs->free_expr_int_sum, eis);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumDoDay(KHE_DRS_EXPR_INT_SUM eis, int day_index,      */
/*    KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active node eis on day_index.                                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumDoDay(KHE_DRS_EXPR_INT_SUM eis, int day_index,
  KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, start_index, stop_index;  KHE_DRS_EXPR_INT child_e;

  /* initialize eis->value depending on whether day_index is its first day */
  if( day_index == eis->first_day_index )
    eis->value = 0;
  else
  {
    /* this is not eis's first day, so we need to retrieve a previous value */
    si = HaArray(eis->sig_indexes, day_index - 1 - eis->first_day_index);
    eis->value = HaArray(prev_node->signature, prev_so + si);
  }

  /* accumulate the values of the children of eis that finalized today */
  start_index = HaArray(eis->child_indexes, day_index - eis->first_day_index);
  stop_index =  HaArray(eis->child_indexes, day_index+1 - eis->first_day_index);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = (KHE_DRS_EXPR_INT) HaArray(eis->children, i);
    eis->value += child_e->value;
  }

  if( day_index < eis->last_day_index )
  {
    /* this not eis's last day; store value in next_node */
    si = HaArray(eis->sig_indexes, day_index - eis->first_day_index);
    HnAssert(HaArrayCount(next_node->signature) == next_so + si,
      "KheDrsExprIntSumDoDay internal error");
    HaArrayAddLast(next_node->signature, eis->value);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumDebug(KHE_DRS_EXPR_INT_SUM eis,                     */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for busy time expression eis.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumDebug(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff1[100], buff2[100];  int si;
  if( node != NULL )
  {
    /* unindented version with a node value */
    fprintf(fp, "INT_SUM");
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eis, node->day->day_index, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(node->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eis, node, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    /* indented version without a node value */
    fprintf(fp, "%*s[ INT_SUM %s %s\n", indent, "",
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) eis, drs, buff1),
      KheDrsDomTestShow(eis->dom_test, buff2));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eis, node, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    /* unindented version without a node value */
    fprintf(fp, "[ INT_SUM %s %s (%d children) ]",
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) eis, drs, buff1),
      KheDrsDomTestShow(eis->dom_test, buff2), HaArrayCount(eis->children));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_SEQ"                                         */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SEQ KheDrsExprIntSeqMake(int history_before,            */
/*    int history_after, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_SEQ object with these attributes.            */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static KHE_DRS_EXPR_INT_SEQ KheDrsExprIntSeqMake(int history_before,
  int history_after, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_SEQ res;
  KheDrsExprInit(res, KHE_DRS_EXPR_INT_SEQ_TAG, dom_test, drs,
    free_expr_int_seq);
  res->history_before = history_before;
  res->history_after = history_after;
  res->tmp_value = UNDEF_INT;
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqDelete(KHE_DRS_EXPR_INT_SEQ eis,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete eis and its descendants.                                          */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static void KheDrsExprIntSeqDelete(KHE_DRS_EXPR_INT_SEQ eis,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(eis->children, e, i)
    KheDrsExprDelete(e, drs);
  HaArrayAddLast(drs->free_expr_int_seq, eis);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsExprIntSeqDomTest(KHE_DRS_EXPR_INT_SEQ eis)       */
/*                                                                           */
/*  Return a suitable dom test for eis.                                      */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static KHE_DRS_DOM_TEST KheDrsExprIntSeqDomTest(KHE_DRS_EXPR_INT_SEQ eis)
{
  return KheExprMonotoneDomTest((KHE_DRS_EXPR) eis);  ** is this right? **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqBeginDay(KHE_DRS_EXPR_INT_SEQ eis, int day_index,   */
/*    KHE_DRS_NODE prev_node, KHE_DYNAMIC_RESOURCE_SOLVER drs)               */
/*                                                                           */
/*  Inform eis that day day_index is beginning.                              */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static void KheDrsExprIntSeqBeginDay(KHE_DRS_EXPR_INT_SEQ eis, int day_index,
  KHE_DRS_NODE prev_node, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eis->tmp_value = KheDrsExprIntBeginDay((KHE_DRS_EXPR) eis, day_index,
    eis->history_before, prev_node, drs, "IntSeq");
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqReceiveValueInt(KHE_DRS_EXPR_INT_SEQ eis,           */
/*    int value, KHE_DYNAMIC_RESOURCE_SOLVER drs)                            */
/*                                                                           */
/*  Inform eis that one of its children has just finalized its value.        */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static void KheDrsExprIntSeqReceiveValueInt(KHE_DRS_EXPR_INT_SEQ eis,
  int value, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HnAssert(value == 0 || value == 1,
    "KheDrsExprIntSeqReceiveValueInt internal error");
  if( value == 1 )
  {
    ** a sequence is beginning or continuing **
    eis->tmp_value++;
  }
  else if( eis->tmp_value > 0 )
  {
    ** a non-empty sequence has just ended **
    KheDrsExprReceiveValueInt(eis->parent, eis->tmp_value, drs);
    eis->tmp_value = 0;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqEndDay(KHE_DRS_EXPR_INT_SEQ eis, int day_index,     */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs)                    */
/*                                                                           */
/*  Inform eis that day day_index is ending.                                 */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static void KheDrsExprIntSeqEndDay(KHE_DRS_EXPR_INT_SEQ eis, int day_index,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eis->tmp_value = KheDrsExprIntEndDay((KHE_DRS_EXPR) eis, day_index,
    ** eis->tmp_value > 0, ** eis->tmp_value, node, drs);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqDebug(KHE_DRS_EXPR_INT_SEQ eis,                     */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of eis onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static void KheDrsExprIntSeqDebug(KHE_DRS_EXPR_INT_SEQ eis,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KheDrsExprIntDebug((KHE_DRS_EXPR) eis, "INT_SEQ", eis->tmp_value,
    drs, verbosity, indent, fp);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_FLOAT_SUM"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FLOAT_SUM KheDrsExprFloatSumMake(                           */
/*    KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)            */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_FLOAT_SUM object with these attributes.          */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FLOAT_SUM KheDrsExprFloatSumMake(
  KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FLOAT_SUM res;
  KheDrsExprInit(res, KHE_DRS_EXPR_FLOAT_SUM_TAG, dom_test, drs);
    /* , free_expr_float_sum); */
  res->value = UNDEF_FLOAT;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumDelete(KHE_DRS_EXPR_FLOAT_SUM efs,                */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete efs and its descendants.                                          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFloatSumDelete(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(efs->children, e, i)
    KheDrsExprDelete(e, drs);
  HaArrayAddLast(drs->free_expr_float_sum, efs);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumDoDay(KHE_DRS_EXPR_FLOAT_SUM efs, int day_index,  */
/*    KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active node efs on day_index.                                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatSumDoDay(KHE_DRS_EXPR_FLOAT_SUM efs, int day_index,
  KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, start_index, stop_index;  KHE_DRS_EXPR_FLOAT child_e;

  /* initialize efs->value depending on whether day_index is its first day */
  if( day_index == efs->first_day_index )
    efs->value = 0.0;
  else
  {
    /* this is not efs's first day, so we need to retrieve a previous value */
    si = HaArray(efs->sig_indexes, day_index - 1 - efs->first_day_index);
    efs->value = (float) HaArray(prev_node->signature, prev_so + si) / 100.0;
  }

  /* accumulate the values of the children of efs that finalized today */
  start_index = HaArray(efs->child_indexes, day_index - efs->first_day_index);
  stop_index =  HaArray(efs->child_indexes, day_index+1 - efs->first_day_index);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = (KHE_DRS_EXPR_FLOAT) HaArray(efs->children, i);
    efs->value += child_e->value;
  }

  if( day_index < efs->last_day_index )
  {
    /* this not efs's last day; store value in next_node */
    si = HaArray(efs->sig_indexes, day_index - efs->first_day_index);
    HnAssert(HaArrayCount(next_node->signature) == next_so + si,
      "KheDrsExprFloatSumDoDay internal error");
    HaArrayAddLast(next_node->signature, (int) (efs->value * 100.0));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumDebug(KHE_DRS_EXPR_FLOAT_SUM efs,                 */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of the header part of efs in node onto fp.                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatSumDebug(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff1[100], buff2[100];  int si;
  if( node != NULL )
  {
    /* unindented version with a node value */
    fprintf(fp, "FLOAT_SUM");
    if( KheDrsExprInSignature((KHE_DRS_EXPR) efs, node->day->day_index, &si) )
      fprintf(fp, " $%d:%.2f$", si, (float) HaArray(node->signature, si) /100.0);
    KheDrsExprDebugChildren((KHE_DRS_EXPR) efs, node, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    /* indented version without a node value */
    fprintf(fp, "%*s[ FLOAT_SUM %s %s\n", indent, "",
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) efs, drs, buff1),
      KheDrsDomTestShow(efs->dom_test, buff2));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) efs, node, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    /* unindented version without a node value */
    fprintf(fp, "[ FLOAT_SUM %s %s (%d children) ]",
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) efs, drs, buff1),
      KheDrsDomTestShow(efs->dom_test, buff2), HaArrayCount(efs->children));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_DEV"                                         */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_DEV KheDrsExprIntDevMake(int min_limit,                 */
/*    int max_limit, bool allow_zero, KHE_DYNAMIC_RESOURCE_SOLVER drs)       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_DEV object with these attributes.            */
/*                                                                           */
/*  There is no dom_test because its type is always KHE_DRS_DOM_UNUSED.      */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_INT_DEV KheDrsExprIntDevMake(int min_limit,
  int max_limit, bool allow_zero, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_DEV res;  KHE_DRS_DOM_TEST dom_test;
  dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  KheDrsExprInit(res, KHE_DRS_EXPR_INT_DEV_TAG, dom_test, drs);
    /* , free_expr_int_dev); */
  res->value = UNDEF_INT;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  res->allow_zero = allow_zero;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevDelete(KHE_DRS_EXPR_INT_DEV eid,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete eid and its descendants.                                          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntDevDelete(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(eid->children, e, i)
    KheDrsExprDelete(e, drs);
  HaArrayAddLast(drs->free_expr_int_dev, eid);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevDoDay(KHE_DRS_EXPR_INT_DEV eid, int day_index,      */
/*    KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active node eid on day_index.                                    */
/*                                                                           */
/*  Implementation note.  This code assumes (correctly) that there is a      */
/*  single child, which becomes final on eid'd first day.                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntDevDoDay(KHE_DRS_EXPR_INT_DEV eid, int day_index,
  KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;  KHE_DRS_EXPR_INT e;

  /* initialize eid->value depending on whether day_index is its first day */
  if( day_index == eid->first_day_index )
  {
    e = (KHE_DRS_EXPR_INT) HaArrayFirst(eid->children);
    eid->value = KheDrsSumDev(eid->min_limit, eid->max_limit,
      eid->allow_zero, 0, e->value);
  }
  else
  {
    /* this is not eid's first day, so we need to retrieve a previous value */
    si = HaArray(eid->sig_indexes, day_index - 1 - eid->first_day_index);
    eid->value = HaArray(prev_node->signature, prev_so + si);
  }

  if( day_index < eid->last_day_index )
  {
    /* this not eid's last day; store value in next_node */
    si = HaArray(eid->sig_indexes, day_index - eid->first_day_index);
    HnAssert(HaArrayCount(next_node->signature) == next_so + si,
      "KheDrsExprIntDevDoDay internal error");
    HaArrayAddLast(next_node->signature, eid->value);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevDebug(KHE_DRS_EXPR_INT_DEV eid,                     */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of the header part of eid in node onto fp.                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntDevDebug(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff1[100], buff2[100];  int si;
  if( node != NULL )
  {
    /* unindented version with a node value (actually no node value here) */
    fprintf(fp, "INT_DEV");
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eid, node->day->day_index, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(node->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eid, node, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    /* indented version without a node value */
    fprintf(fp, "%*s[ INT_DEV %s %s\n", indent, "",
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) eid, drs, buff1),
      KheDrsDomTestShow(eid->dom_test, buff2));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eid, node, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    /* unindented version without a node value */
    fprintf(fp, "[ INT_DEV %s %s (%d children) ]",
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) eid, drs, buff1),
      KheDrsDomTestShow(eid->dom_test, buff2), HaArrayCount(eid->children));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_FLOAT_DEV"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FLOAT_DEV KheDrsExprFloatDevMake(int min_limit,             */
/*    int max_limit, bool allow_zero, KHE_DYNAMIC_RESOURCE_SOLVER drs)       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_FLOAT_DEV object with these attributes.          */
/*                                                                           */
/*  There is no dom_test because its type is always KHE_DRS_DOM_UNUSED.      */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FLOAT_DEV KheDrsExprFloatDevMake(int min_limit,
  int max_limit, bool allow_zero, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FLOAT_DEV res;  KHE_DRS_DOM_TEST dom_test;
  dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  KheDrsExprInit(res, KHE_DRS_EXPR_FLOAT_DEV_TAG, dom_test, drs);
    /* , free_expr_float_dev); */
  res->value = UNDEF_INT;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  res->allow_zero = allow_zero;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevDelete(KHE_DRS_EXPR_FLOAT_DEV efd,                */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete efd and its descendants.                                          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFloatDevDelete(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(efd->children, e, i)
    KheDrsExprDelete(e, drs);
  HaArrayAddLast(drs->free_expr_float_dev, efd);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheLimitWorkloadMonitorDev(KHE_DRS_EXPR_FLOAT_DEV efd,               */
/*    float workload)                                                        */
/*                                                                           */
/*  Return the deviation of workload wrt efd.                                */
/*                                                                           */
/*****************************************************************************/

static int KheLimitWorkloadMonitorDev(KHE_DRS_EXPR_FLOAT_DEV efd,
  float workload)
{
  if( efd->allow_zero && workload < 0.001 )
    return 0;
  else if( workload < efd->min_limit - 0.001 )
    return (int) ceil(efd->min_limit - 0.001 - workload);
  else if( workload > efd->max_limit + 0.001 )
    return (int) ceil(workload - efd->max_limit - 0.001);
  else
    return 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevDoDay(KHE_DRS_EXPR_FLOAT_DEV efd, int day_index,  */
/*    KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active node efd on day_index.                                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatDevDoDay(KHE_DRS_EXPR_FLOAT_DEV efd, int day_index,
  KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;  KHE_DRS_EXPR_FLOAT e;

  /* initialize efd->value depending on whether day_index is its first day */
  if( day_index == efd->first_day_index )
  {
    e = (KHE_DRS_EXPR_FLOAT) HaArrayFirst(efd->children);
    efd->value = KheLimitWorkloadMonitorDev(efd, e->value);
  }
  else
  {
    /* this is not efd's first day, so we need to retrieve a previous value */
    si = HaArray(efd->sig_indexes, day_index - 1 - efd->first_day_index);
    efd->value = HaArray(prev_node->signature, prev_so + si);
  }

  if( day_index < efd->last_day_index )
  {
    /* this not efd's last day; store value in next_node */
    si = HaArray(efd->sig_indexes, day_index - efd->first_day_index);
    HnAssert(HaArrayCount(next_node->signature) == next_so + si,
      "KheDrsExprFloatDevDoDay internal error");
    HaArrayAddLast(next_node->signature, efd->value);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevDebug(KHE_DRS_EXPR_FLOAT_DEV efd,                 */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for float dev expression efd.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatDevDebug(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff1[100], buff2[100];  int si;
  if( node != NULL )
  {
    /* unindented version with a node value (actually no node value here) */
    fprintf(fp, "FLOAT_DEV");
    if( KheDrsExprInSignature((KHE_DRS_EXPR) efd, node->day->day_index, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(node->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) efd, node, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    /* indented version without a node value */
    fprintf(fp, "%*s[ FLOAT_DEV %s %s\n", indent, "",
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) efd, drs, buff1),
      KheDrsDomTestShow(efd->dom_test, buff2));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) efd, node, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    /* unindented version without a node value */
    fprintf(fp, "[ FLOAT_DEV %s %s (%d children) ]",
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) efd, drs, buff1),
      KheDrsDomTestShow(efd->dom_test, buff2), HaArrayCount(efd->children));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_COST"                                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_COST KheDrsExprCostMake(KHE_MONITOR m,                      */
/*    KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,                   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_COST object with these attributes.               */
/*                                                                           */
/*  This function has no dom_test parameter, because there is never any      */
/*  dominance testing at this node.                                          */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_COST KheDrsExprCostMake(KHE_MONITOR m,
  KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_COST res;  KHE_DRS_DOM_TEST dom_test;
  dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  KheDrsExprInit(res, KHE_DRS_EXPR_COST_TAG, dom_test, drs);
    /* , free_expr_cost); */
  res->monitor = m;
  res->cost_fn = cost_fn;
  res->combined_weight = combined_weight;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostDelete(KHE_DRS_EXPR_COST ec,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete ec and its descendants.                                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprCostDelete(KHE_DRS_EXPR_COST ec,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(ec->children, e, i)
    KheDrsExprDelete(e, drs);
  HaArrayAddLast(drs->free_expr_cost, ec);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostDoDay(KHE_DRS_EXPR_COST ec, int day_index,            */
/*    KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active node ec on day_index.                                     */
/*                                                                           */
/*  Implementation note.  Cost expressions always report their values        */
/*  directly to the new node as soon as they are available.  They do         */
/*  not store values for retrieval by a parent; they have no parents.        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprCostDoDay(KHE_DRS_EXPR_COST ec, int day_index,
  KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_COST cost;  KHE_DRS_EXPR_INT e;
  e = (KHE_DRS_EXPR_INT) HaArrayFirst(ec->children);
  cost = KheDrsCost(ec->cost_fn, ec->combined_weight, e->value);
  if( cost > 0 )
  {
    if( drs->rerun )
      fprintf(stderr, "  %s reporting cost %.5f\n",
	KheMonitorId(ec->monitor), KheCostShow(cost));
    next_node->cost += cost;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostDebug(KHE_DRS_EXPR_COST ec,                           */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for cost expression ec.                        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprCostDebug(KHE_DRS_EXPR_COST ec,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff1[100], buff2[100];  int si;
  if( node != NULL )
  {
    /* unindented version with a node value */
    fprintf(fp, "COST %s", KheMonitorId(ec->monitor));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) ec, node->day->day_index, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(node->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) ec, node, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    /* indented version without a node value */
    fprintf(fp, "%*s[ COST %s %s %s\n", indent, "", KheMonitorId(ec->monitor),
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) ec, drs, buff1),
      KheDrsDomTestShow(ec->dom_test, buff2));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) ec, node, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    /* unindented version without a node value */
    fprintf(fp, "[ COST %s %s %s (%d children) ]", KheMonitorId(ec->monitor),
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) ec, drs, buff1),
      KheDrsDomTestShow(ec->dom_test, buff2), HaArrayCount(ec->children));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_SUM_COMB"                                    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_COMB_TYPE KheDrsCombType(KHE_COST_FUNCTION cost_fn,              */
/*    int max_limit)                                                         */
/*                                                                           */
/*  Return the comb type, as determined by the cost function and max limit.  */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_COMB_TYPE KheDrsCombType(KHE_COST_FUNCTION cost_fn,
  int max_limit)
{
  if( max_limit == INT_MAX )
    return KHE_DRS_COMB_NO_MAX;
  else if( cost_fn == KHE_LINEAR_COST_FUNCTION )
    return KHE_DRS_COMB_LINEAR;
  else if( cost_fn == KHE_STEP_COST_FUNCTION )
    return KHE_DRS_COMB_STEP;
  else
    return KHE_DRS_COMB_ORDINARY;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SUM_COMB KheDrsExprIntSumCombMake(KHE_MONITOR m,        */
/*    KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,                   */
/*    int min_limit, int max_limit, bool allow_zero, int history_before,     */
/*    int history_after, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_SUM_COMB object with these attributes.       */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_INT_SUM_COMB KheDrsExprIntSumCombMake(KHE_MONITOR m,
  KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,
  int min_limit, int max_limit, bool allow_zero, int history_before,
  int history_after, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_SUM_COMB res;
  KheDrsExprInit(res, KHE_DRS_EXPR_INT_SUM_COMB_TAG, dom_test, drs);
    /* , free_expr_int_sum_comb); */
  res->value = UNDEF_INT;
  res->monitor = m;
  res->cost_fn = cost_fn;
  res->combined_weight = combined_weight;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  res->allow_zero = allow_zero;
  res->history_before = history_before;
  res->history_after = history_after;
  res->comb_type = KheDrsCombType(cost_fn, max_limit);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombDelete(KHE_DRS_EXPR_INT_SUM_COMB eisc,          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete eisc and its descendants.                                         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumCombDelete(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(eisc->children, e, i)
    KheDrsExprDelete(e, drs);
  HaArrayAddLast(drs->free_expr_int_sum_comb, eisc);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombDoDay(KHE_DRS_EXPR_INT_SUM_COMB eisc,           */
/*    int day_index,                                                         */
/*    KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active node eisc on day_index.                                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumCombDoDay(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  int day_index,
  KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, start_index, stop_index, dev;
  KHE_DRS_EXPR_INT child_e;  KHE_COST cost;

  /* initialize eisc->value depending on whether day_index is its first day */
  if( day_index == eisc->first_day_index )
    eisc->value = eisc->history_before;
  else
  {
    /* this is not eisc's first day, so we need to retrieve a previous value */
    si = HaArray(eisc->sig_indexes, day_index - 1 - eisc->first_day_index);
    eisc->value = HaArray(prev_node->signature, prev_so + si);
  }

  /* accumulate the values of the children of eisc that finalized today */
  start_index = HaArray(eisc->child_indexes, day_index - eisc->first_day_index);
  stop_index =  HaArray(eisc->child_indexes, day_index+1-eisc->first_day_index);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = (KHE_DRS_EXPR_INT) HaArray(eisc->children, i);
    eisc->value += child_e->value;
  }

  /* adjust in various special cases */
  switch( eisc->comb_type )
  {
    case KHE_DRS_COMB_ORDINARY:

      /* no adjustment here */
      break;

    case KHE_DRS_COMB_NO_MAX:

      /* no maximum limit; store minimum limit at most */
      if( eisc->value > eisc->min_limit )
	eisc->value = eisc->min_limit;
      break;

    case KHE_DRS_COMB_LINEAR:

      /* linear cost; if over maximum, report excess immediately */
      if( eisc->value > eisc->max_limit )
      {
	cost = (eisc->value - eisc->max_limit) * eisc->combined_weight;
	if( cost > 0 )
	{
	  if( drs->rerun )
	    fprintf(stderr, "  %s reporting linear cost %.5f\n",
	      KheMonitorId(eisc->monitor), KheCostShow(cost));
	  next_node->cost += cost;
	}
	eisc->value = eisc->max_limit;
      }
      break;

    case KHE_DRS_COMB_STEP:

      /* step cost; store maximum plus one at most */
      if( eisc->value > eisc->max_limit + 1 )
	eisc->value = eisc->max_limit + 1;
      break;

    default:

      HnAbort("KheDrsExprIntSumCombDoDay internal error");
      break;
  }

  if( day_index < eisc->last_day_index )
  {
    /* this not eisc's last day; store value in next_node */
    si = HaArray(eisc->sig_indexes, day_index - eisc->first_day_index);
    HnAssert(HaArrayCount(next_node->signature) == next_so + si,
      "KheDrsExprIntSumCombDoDay internal error");
    HaArrayAddLast(next_node->signature, eisc->value);
  }
  else
  {
    /* this eisc's last day; report any cost immediately to next_node */
    dev = KheDrsSumDev(eisc->min_limit, eisc->max_limit, eisc->allow_zero,
      eisc->history_after, eisc->value);
    cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
    if( cost > 0 )
    {
      if( drs->rerun )
	fprintf(stderr, "  %s reporting cost %.5f\n",
	  KheMonitorId(eisc->monitor), KheCostShow(cost));
      next_node->cost += cost;
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombDebug(KHE_DRS_EXPR_INT_SUM_COMB eisc,           */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for int sum comb expression eisc.              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumCombDebug(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff1[100], buff2[100];  int si;
  if( node != NULL )
  {
    /* unindented version with a node value */
    fprintf(fp, "INT_SUM_COMB %s", KheMonitorId(eisc->monitor));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eisc, node->day->day_index, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(node->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eisc, node, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    /* indented version without a node value */
    fprintf(fp, "%*s[ INT_SUM_COMB %s %s %s\n", indent, "",
      KheMonitorId(eisc->monitor),
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) eisc, drs, buff1),
      KheDrsDomTestShow(eisc->dom_test, buff2));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eisc, node, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    /* unindented version without a node value */
    fprintf(fp, "[ INT_SUM_COMB %s %s %s (%d children) ]",
      KheMonitorId(eisc->monitor),
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) eisc, drs, buff1),
      KheDrsDomTestShow(eisc->dom_test, buff2), HaArrayCount(eisc->children));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_SEQ_COMB"                                    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SEQ_COMB KheDrsExprIntSeqCombMake(KHE_MONITOR m,        */
/*    KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,                   */
/*    KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight, int min_limit,    */
/*    int max_limit, int history_before, int history_after,                  */
/*    KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)            */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_SEQ_COMB object with these attributes.       */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_INT_SEQ_COMB KheDrsExprIntSeqCombMake(KHE_MONITOR m,
  KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight, int min_limit,
  int max_limit, int history_before, int history_after,
  KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_SEQ_COMB res;
  KheDrsExprInit(res, KHE_DRS_EXPR_INT_SEQ_COMB_TAG, dom_test, drs);
    /* , free_expr_int_seq_comb); */
  res->value = UNDEF_INT;
  res->monitor = m;
  res->cost_fn = cost_fn;
  res->combined_weight = combined_weight;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  res->history_before = history_before;
  res->history_after = history_after;
  res->comb_type = KheDrsCombType(cost_fn, max_limit);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombDelete(KHE_DRS_EXPR_INT_SEQ_COMB eisc,          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete eisc and its descendants.                                         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSeqCombDelete(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(eisc->children, e, i)
    KheDrsExprDelete(e, drs);
  HaArrayAddLast(drs->free_expr_int_seq_comb, eisc);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR KheDrsExprOptionalCopy(KHE_DRS_EXPR e,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  If e is a BUSY_TIME, FREE_TIME, WORK_TIME, BUSY_DAY, FREE_DAY, or        */
/*  WORK_DAY expression, return a fresh copy of it.  Otherwise return e.     */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR KheDrsExprOptionalCopy(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      return (KHE_DRS_EXPR) KheDrsExprBusyTimeCopy((KHE_DRS_EXPR_BUSY_TIME) e,
	drs);

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      return (KHE_DRS_EXPR) KheDrsExprFreeTimeCopy((KHE_DRS_EXPR_FREE_TIME) e,
	drs);

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      return (KHE_DRS_EXPR) KheDrsExprWorkTimeCopy((KHE_DRS_EXPR_WORK_TIME) e,
	drs);

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      return (KHE_DRS_EXPR) KheDrsExprBusyDayCopy((KHE_DRS_EXPR_BUSY_DAY) e,
	drs);

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      return (KHE_DRS_EXPR) KheDrsExprFreeDayCopy((KHE_DRS_EXPR_FREE_DAY) e,
	drs);

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      return (KHE_DRS_EXPR) KheDrsExprWorkDayCopy((KHE_DRS_EXPR_WORK_DAY) e,
	drs);

    default:

      /* other node types are not shared, just return e */
      return e;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombDoDay(KHE_DRS_EXPR_INT_SEQ_COMB eisc,           */
/*    int day_index,                                                         */
/*    KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active node eisc on day_index.                                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCombDoDay(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  int day_index,
  KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, dev, start_index, stop_index;
  KHE_DRS_EXPR_INT child_e;  KHE_COST cost;

  /* initialize eisc->value depending on whether day_index is its first day */
  if( day_index == eisc->first_day_index )
    eisc->value = eisc->history_before;
  else
  {
    /* this is not eisc's first day, so we need to retrieve a previous value */
    si = HaArray(eisc->sig_indexes, day_index-1 - eisc->first_day_index);
    eisc->value = HaArray(prev_node->signature, prev_so + si);
  }

  /* accumulate the values of the children of eisc that finalized today */
  start_index = HaArray(eisc->child_indexes, day_index - eisc->first_day_index);
  stop_index =  HaArray(eisc->child_indexes, day_index+1-eisc->first_day_index);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = (KHE_DRS_EXPR_INT) HaArray(eisc->children, i);
    if( child_e->value == 1 )
    {
      /* a sequence is beginning or continuing */
      eisc->value++;

      /* adjust in various special cases */
      switch( eisc->comb_type )
      {
	case KHE_DRS_COMB_ORDINARY:

	  /* no adjustment here */
	  break;

	case KHE_DRS_COMB_NO_MAX:

	  /* no maximum limit; store minimum limit at most */
	  if( eisc->value > eisc->min_limit )
	    eisc->value = eisc->min_limit;
	  break;

	case KHE_DRS_COMB_LINEAR:

	  /* linear cost; if over maximum, report excess immediately */
	  if( eisc->value > eisc->max_limit )
	  {
	    dev = KheDrsSeqDev(eisc->min_limit, eisc->max_limit, 0,
	      eisc->value);
	    cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
	    if( cost > 0 )
	    {
	      if( drs->rerun )
		fprintf(stderr, "  %s reporting linear cost %.5f\n",
		  KheMonitorId(eisc->monitor), KheCostShow(cost));
	      next_node->cost += cost;
	    }
	    eisc->value = eisc->max_limit;
	  }
	  break;

	case KHE_DRS_COMB_STEP:

	  /* step cost; store maximum plus one at most */
	  if( eisc->value > eisc->max_limit + 1 )
	    eisc->value = eisc->max_limit + 1;
	  break;

	default:

	  HnAbort("KheDrsExprIntSeqCombDoDay internal error");
	  break;
      }
    }
    else if( eisc->value > 0 )
    {
      /* a non-empty sequence not at the end of the cycle has just ended */
      dev = KheDrsSeqDev(eisc->min_limit, eisc->max_limit, 0, eisc->value);
      if( dev > 0 )
      {
	cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
	if( cost > 0 )
	{
	  if( drs->rerun )
	    fprintf(stderr, "  %s reporting cost %.5f\n",
	      KheMonitorId(eisc->monitor), KheCostShow(cost));
	  next_node->cost += cost;
	}
      }
      eisc->value = 0;
    }
  }

  if( day_index < eisc->last_day_index )
  {
    /* this not eisc's last day; store value in next_node */
    si = HaArray(eisc->sig_indexes, day_index - eisc->first_day_index);
    HnAssert(HaArrayCount(next_node->signature) == next_so + si,
      "KheDrsExprIntSeqDoDay internal error 3");
    HaArrayAddLast(next_node->signature, eisc->value);
  }
  else if( eisc->value > 0 )
  {
    /* this is the last day and we have a sequence to report */
    dev = KheDrsSeqDev(eisc->min_limit, eisc->max_limit,
      eisc->history_after, eisc->value);
    if( dev > 0 )
    {
      cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
      if( cost > 0 )
      {
	if( drs->rerun )
	  fprintf(stderr, "  %s reporting cost %.5f\n",
	    KheMonitorId(eisc->monitor), KheCostShow(cost));
	next_node->cost += cost;
      }
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombDebug(KHE_DRS_EXPR_INT_SEQ_COMB eisc,           */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of the header part of eisc in node onto fp.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCombDebug(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char buff1[100], buff2[100];  int si;
  if( node != NULL )
  {
    /* unindented version with a node value */
    fprintf(fp, "INT_SEQ_COMB %s", KheMonitorId(eisc->monitor));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eisc, node->day->day_index, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(node->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eisc, node, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    /* indented version without a node value */
    fprintf(fp, "%*s[ INT_SEQ_COMB %s %s %s\n", indent, "",
      KheMonitorId(eisc->monitor),
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) eisc, drs, buff1),
      KheDrsDomTestShow(eisc->dom_test, buff2));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eisc, node, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    /* unindented version without a node value */
    fprintf(fp, "[ INT_SEQ_COMB %s %s %s (%d children) ]",
      KheMonitorId(eisc->monitor),
      KheDrsExprDayRangeShow((KHE_DRS_EXPR) eisc, drs, buff1),
      KheDrsDomTestShow(eisc->dom_test, buff2), HaArrayCount(eisc->children));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAddChild(KHE_DRS_EXPR parent, KHE_DRS_EXPR child)         */
/*                                                                           */
/*  Make child a child of parent.                                            */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAddChild(KHE_DRS_EXPR parent, KHE_DRS_EXPR child)
{
  HaArrayAddLast(parent->children, child);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprEnsureDayRangeCoversDayIndex(KHE_DRS_EXPR e,              */
/*    int last_day_index)                                                    */
/*                                                                           */
/*  Ensure that e's day range covers last_day_index.                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprEnsureDayRangeCoversDayIndex(KHE_DRS_EXPR e,
  int day_index)
{
  if( day_index < e->first_day_index )
    e->first_day_index = day_index;
  if( day_index > e->last_day_index )
    e->last_day_index = day_index;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprIncreasingLastDayIndexCmp(const void *t1, const void *t2)  */
/*                                                                           */
/*  Comparison function for sorting an array of expressions by increasing    */
/*  last day index.                                                          */
/*                                                                           */
/*****************************************************************************/

static int KheDrsExprIncreasingLastDayIndexCmp(const void *t1, const void *t2)
{
  KHE_DRS_EXPR e1 = * (KHE_DRS_EXPR *) t1;
  KHE_DRS_EXPR e2 = * (KHE_DRS_EXPR *) t2;
  return e1->last_day_index - e2->last_day_index;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprSetChildIndexes(KHE_DRS_EXPR e)                           */
/*                                                                           */
/*  Assuming that e's children's day indexes are set, set e's child indexes. */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprSetChildIndexes(KHE_DRS_EXPR e)
{
  int i, curr_last_day_index;  KHE_DRS_EXPR child_e;
  HaArrayClear(e->child_indexes);
  curr_last_day_index = e->first_day_index - 1;
  HaArrayForEach(e->children, child_e, i)
  {
    while( child_e->last_day_index > curr_last_day_index )
    {
      HaArrayAddLast(e->child_indexes, i);
      curr_last_day_index++;
    }
  }
  HaArrayAddLast(e->child_indexes, i);  /* sentinel */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprSetDayRange(KHE_DRS_EXPR e, int min_last_day_index,       */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Set the day range (first_day_index and last_day_index) of e.             */
/*                                                                           */
/*  If min_last_day_index != -1, ensure that e->last_day_index is at least   */
/*  min_last_day_index.  Used for the children of INT_SEQ and INT_SEQ_COMB   */
/*  nodes, this ensures that those children report in the order they appear. */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprSetDayRange(KHE_DRS_EXPR e, /* int min_last_day_index, */
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, index /* , prev_last_day_index */;  KHE_DRS_EXPR child_e, prev_child_e;
  KHE_DRS_EXPR_BUSY_TIME eb;  KHE_DRS_EXPR_FREE_TIME ef;
  KHE_DRS_EXPR_WORK_TIME ew;  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_DRS_EXPR_FREE_DAY efd;  KHE_DRS_EXPR_WORK_DAY ewd;
  KHE_DRS_EXPR_INT_SEQ_COMB eisc;

  /* initialize range to empty, or to just min_last_day_index */
  e->first_day_index = INT_MAX;
  e->last_day_index = 0;
  /* ***
  if( min_last_day_index != -1 )
    KheDrsExprEnsureDayRangeCoversDayIndex(e, min_last_day_index);
  *** */

  /* increase range depending on the expression type */
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      eb = (KHE_DRS_EXPR_BUSY_TIME) e;
      index = KheFrameTimeIndex(drs->days_frame, eb->time);
      KheDrsExprEnsureDayRangeCoversDayIndex(e, index);
      /* KheDrsExprEnsureDayRangeCovers(e, index, index); */
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      ef = (KHE_DRS_EXPR_FREE_TIME) e;
      index = KheFrameTimeIndex(drs->days_frame, ef->time);
      KheDrsExprEnsureDayRangeCoversDayIndex(e, index);
      /* KheDrsExprEnsureDayRangeCovers(e, index, index); */
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      ew = (KHE_DRS_EXPR_WORK_TIME) e;
      index = KheFrameTimeIndex(drs->days_frame, ew->time);
      KheDrsExprEnsureDayRangeCoversDayIndex(e, index);
      /* KheDrsExprEnsureDayRangeCovers(e, index, index); */
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      ebd = (KHE_DRS_EXPR_BUSY_DAY) e;
      index = KheFrameTimeIndex(drs->days_frame,
	KheTimeGroupTime(ebd->time_group, 0));
      KheDrsExprEnsureDayRangeCoversDayIndex(e, index);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      efd = (KHE_DRS_EXPR_FREE_DAY) e;
      index = KheFrameTimeIndex(drs->days_frame,
	KheTimeGroupTime(efd->time_group, 0));
      KheDrsExprEnsureDayRangeCoversDayIndex(e, index);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      ewd = (KHE_DRS_EXPR_WORK_DAY) e;
      index = KheFrameTimeIndex(drs->days_frame,
	KheTimeGroupTime(ewd->time_group, 0));
      KheDrsExprEnsureDayRangeCoversDayIndex(e, index);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      eisc = (KHE_DRS_EXPR_INT_SEQ_COMB) e;
      prev_child_e = NULL;
      HaArrayForEach(eisc->children, child_e, i)
      {
        KheDrsExprSetDayRange(child_e, /* prev_last_day_index, */ drs);
	if( prev_child_e != NULL &&
	    prev_child_e->last_day_index > child_e->last_day_index )
	{
	  child_e = KheDrsExprOptionalCopy(child_e, drs);
	  HaArrayPut(eisc->children, i, child_e);
          child_e->last_day_index = prev_child_e->last_day_index;
	}
        KheDrsExprEnsureDayRangeCoversDayIndex(e, child_e->last_day_index);
	prev_child_e = child_e;
      }
      /* children already have increasing last day index, must not sort them */
      break;

    default:

      HaArrayForEach(e->children, child_e, i)
      {
        KheDrsExprSetDayRange(child_e, /* -1, */ drs);
        KheDrsExprEnsureDayRangeCoversDayIndex(e, child_e->last_day_index);
      }
      HaArraySort(e->children, &KheDrsExprIncreasingLastDayIndexCmp);
      break;
  }

  /* set the child indexes */
  KheDrsExprSetChildIndexes(e);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprSetSigIndexes(KHE_DRS_EXPR e,                             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Set the signature indexes in e and its descendants.                      */
/*                                                                           */
/*  These are assigned in postorder, because that is the order that          */
/*  KheDrsExprEndDay visits the nodes.                                       */
/*                                                                           */
/*****************************************************************************/
/* static void KheDrsDayAddDomTest(KHE_DRS_DAYday,KHE_DRS_DOM_TESTdom_test); */
static void KheDrsResourceOnDayAddDomTest(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_DOM_TEST dom_test);

static void KheDrsExprSetSigIndexes(KHE_DRS_EXPR e, KHE_DRS_RESOURCE dr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR child_e;  int i;  /* KHE_DRS_DAY day; */
  KHE_DRS_RESOURCE_ON_DAY drd;

  /* set the signature indexes in e's descendants */
  HaArrayForEach(e->children, child_e, i)
    KheDrsExprSetSigIndexes(child_e, dr, drs);

  /* set the signature indexes in e (NB < is correct, not <= here) */
  for( i = e->first_day_index;  i < e->last_day_index;  i++ )
  {
    /* ***
    day = HaArray(drs->days, i);
    HaArrayAddLast(e->sig_indexes, day->sig_len++);
    *** */
    drd = HaArray(dr->days, i);
    HaArrayAddLast(e->sig_indexes, drd->sig_len++);
    HnAssert(e->dom_test.type != KHE_DRS_DOM_UNUSED,
      "KheDrsExprSetSigIndexes internal error");
    KheDrsResourceOnDayAddDomTest(drd, e->dom_test);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAddExpressionsToDays(KHE_DRS_EXPR e, KHE_DRS_RESOURCE dr) */
/*                                                                           */
/*  Add e and its descendants to the days of dr they are active on.  Most    */
/*  importantly, do it in postorder.                                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAddExpressionsToDays(KHE_DRS_EXPR e, KHE_DRS_RESOURCE dr)
{
  KHE_DRS_EXPR child_e;  int i;  KHE_DRS_RESOURCE_ON_DAY drd;

  /* do it for e's proper descendants (first!) */
  HaArrayForEach(e->children, child_e, i)
    KheDrsExprAddExpressionsToDays(child_e, dr);

  /* do it for e (NB <= is correct, not < here) */
  for( i = e->first_day_index;  i <= e->last_day_index;  i++ )
  {
    drd = HaArray(dr->days, i);
    HaArrayAddLast(drd->active_today, e);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAddExpressionsToDays(KHE_DRS_EXPR e,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add e and its descendants to the days they are active on.  Most          */
/*  importantly, do it in postorder.                                         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprAddExpressionsToDays(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR child_e;  int i;  KHE_DRS_DAY day;

  ** do it for e's proper descendants (first!) **
  HaArrayForEach(e->children, child_e, i)
    KheDrsExprAddExpressionsToDays(child_e, drs);

  ** do it for e (NB <= is correct, not < here) **
  for( i = e->first_day_index;  i <= e->last_day_index;  i++ )
  {
    day = HaArray(drs->days, i);
    HaArrayAddLast(day->active_today, e);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprDelete(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs)   */
/*                                                                           */
/*  Delete e and its descendants, taking care not to delete something that   */
/*  has already been deleted.                                                */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprDelete(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( !e->deleted )
  {
    e->deleted = true;
    switch( e->tag )
    {
      case KHE_DRS_EXPR_BUSY_TIME_TAG:

	KheDrsExprBusyTimeDelete((KHE_DRS_EXPR_BUSY_TIME) e, drs);
	break;

      case KHE_DRS_EXPR_FREE_TIME_TAG:

	KheDrsExprFreeTimeDelete((KHE_DRS_EXPR_FREE_TIME) e, drs);
	break;

      case KHE_DRS_EXPR_WORK_TIME_TAG:

	KheDrsExprWorkTimeDelete((KHE_DRS_EXPR_WORK_TIME) e, drs);
	break;

      case KHE_DRS_EXPR_BUSY_DAY_TAG:

	KheDrsExprBusyDayDelete((KHE_DRS_EXPR_BUSY_DAY) e, drs);
	break;

      case KHE_DRS_EXPR_FREE_DAY_TAG:

	KheDrsExprFreeDayDelete((KHE_DRS_EXPR_FREE_DAY) e, drs);
	break;

      case KHE_DRS_EXPR_WORK_DAY_TAG:

	KheDrsExprWorkDayDelete((KHE_DRS_EXPR_WORK_DAY) e, drs);
	break;

      case KHE_DRS_EXPR_OR_TAG:

	KheDrsExprOrDelete((KHE_DRS_EXPR_OR) e, drs);
	break;

      case KHE_DRS_EXPR_AND_TAG:

	KheDrsExprAndDelete((KHE_DRS_EXPR_AND) e, drs);
	break;

      case KHE_DRS_EXPR_INT_SUM_TAG:

	KheDrsExprIntSumDelete((KHE_DRS_EXPR_INT_SUM) e, drs);
	break;

      ** *** currently unused
      case KHE_DRS_EXPR_INT_SEQ_TAG:

	KheDrsExprIntSeqDelete((KHE_DRS_EXPR_INT_SEQ) e, drs);
	break;
      *** **

      case KHE_DRS_EXPR_FLOAT_SUM_TAG:

	KheDrsExprFloatSumDelete((KHE_DRS_EXPR_FLOAT_SUM) e, drs);
	break;

      case KHE_DRS_EXPR_INT_DEV_TAG:

	KheDrsExprIntDevDelete((KHE_DRS_EXPR_INT_DEV) e, drs);
	break;

      case KHE_DRS_EXPR_FLOAT_DEV_TAG:

	KheDrsExprFloatDevDelete((KHE_DRS_EXPR_FLOAT_DEV) e, drs);
	break;

      case KHE_DRS_EXPR_COST_TAG:

	KheDrsExprCostDelete((KHE_DRS_EXPR_COST) e, drs);
	break;

      case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

	KheDrsExprIntSumCombDelete((KHE_DRS_EXPR_INT_SUM_COMB) e, drs);
	break;

      case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

	KheDrsExprIntSeqCombDelete((KHE_DRS_EXPR_INT_SEQ_COMB) e, drs);
	break;

      default:

	HnAbort("KheDrsExprDelete internal error: unknown tag %d", e->tag);
	break;
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprDoDay(KHE_DRS_EXPR e, int day_index,                      */
/*    KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active expression e on day day_index.                            */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprDoDay(KHE_DRS_EXPR e, int day_index,
  KHE_DRS_NODE prev_node, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_NODE next_node, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      KheDrsExprBusyTimeDoDay((KHE_DRS_EXPR_BUSY_TIME) e, day_index,
	prev_node, prev_so, task, time, next_node, next_so, drs);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      KheDrsExprFreeTimeDoDay((KHE_DRS_EXPR_FREE_TIME) e, day_index,
	prev_node, prev_so, task, time, next_node, next_so, drs);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      KheDrsExprWorkTimeDoDay((KHE_DRS_EXPR_WORK_TIME) e, day_index,
	prev_node, prev_so, task, time, next_node, next_so, drs);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      KheDrsExprBusyDayDoDay((KHE_DRS_EXPR_BUSY_DAY) e, day_index,
	prev_node, prev_so, task, time, next_node, next_so, drs);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      KheDrsExprFreeDayDoDay((KHE_DRS_EXPR_FREE_DAY) e, day_index,
	prev_node, prev_so, task, time, next_node, next_so, drs);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      KheDrsExprWorkDayDoDay((KHE_DRS_EXPR_WORK_DAY) e, day_index,
	prev_node, prev_so, task, time, next_node, next_so, drs);
      break;

    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      KheDrsExprAssignedTaskDoDay((KHE_DRS_EXPR_ASSIGNED_TASK) e, day_index,
        prev_node, prev_so, KheTaskAsstResource(task), next_node, next_so, drs);
      break;

    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrDoDay((KHE_DRS_EXPR_OR) e, day_index,
	prev_node, prev_so, task, time, next_node, next_so, drs);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndDoDay((KHE_DRS_EXPR_AND) e, day_index,
	prev_node, prev_so, task, time, next_node, next_so, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumDoDay((KHE_DRS_EXPR_INT_SUM) e, day_index,
	prev_node, prev_so, task, time, next_node, next_so, drs);
      break;

    /* *** currently unused
    case KHE_DRS_EXPR_INT_SEQ_TAG:

      KheDrsExprIntSeqDoDay((KHE_DRS_EXPR_INT_SEQ) e, day_index,
	prev_node, prev_so, task, time, next_node, next_so, drs);
      break;
    *** */

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumDoDay((KHE_DRS_EXPR_FLOAT_SUM) e, day_index,
	prev_node, prev_so, task, time, next_node, next_so, drs);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevDoDay((KHE_DRS_EXPR_INT_DEV) e, day_index,
	prev_node, prev_so, task, time, next_node, next_so, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevDoDay((KHE_DRS_EXPR_FLOAT_DEV) e, day_index,
	prev_node, prev_so, task, time, next_node, next_so, drs);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostDoDay((KHE_DRS_EXPR_COST) e, day_index,
	prev_node, prev_so, task, time, next_node, next_so, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombDoDay((KHE_DRS_EXPR_INT_SUM_COMB) e, day_index,
	prev_node, prev_so, task, time, next_node, next_so, drs);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombDoDay((KHE_DRS_EXPR_INT_SEQ_COMB) e, day_index,
	prev_node, prev_so, task, time, next_node, next_so, drs);
      break;

    default:

      HnAbort("KheDrsExprDoDay internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsExprIsActive(KHE_DRS_EXPR e, int day_index)                   */
/*                                                                           */
/*  Return true if e is active on day_index.                                 */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; good idea though
static bool KheDrsExprIsActive(KHE_DRS_EXPR e, int day_index)
{
  return e->first_day_index <= day_index && day_index <= e->last_day_index;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsExprInSignature(KHE_DRS_EXPR e, int day_index, int *si)       */
/*                                                                           */
/*  If e is in the signature on day_index, return true and set *si to its    */
/*  position in the signature.  Otherwise return false.                      */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsExprInSignature(KHE_DRS_EXPR e, int day_index, int *si)
{
  if( e->first_day_index <= day_index && day_index < e->last_day_index )
    return *si = HaArray(e->sig_indexes, day_index - e->first_day_index), true;
  else
    return *si = -1, false;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprFirstDebugDayIndex(KHE_DRS_EXPR e)                         */
/*                                                                           */
/*  The minimum, over e and its descendants, of the first_day_index field    */
/*  of e.                                                                    */
/*                                                                           */
/*  Implementation note.  We only search down the first children.  If that   */
/*  is wrong in some obscure case, it doesn't matter, because this function  */
/*  is called only when debugging.                                           */
/*                                                                           */
/*****************************************************************************/

static int KheDrsExprFirstDebugDayIndex(KHE_DRS_EXPR e)
{
  int res;
  res = e->first_day_index;
  while( HaArrayCount(e->children) > 0 )
  {
    e = HaArrayFirst(e->children);
    if( e->first_day_index < res )
      res = e->first_day_index;
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprDebugChildren(KHE_DRS_EXPR e, KHE_DRS_NODE node,          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Carry out the part of KheDrsExprDebug which prints the children.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprDebugChildren(KHE_DRS_EXPR e, KHE_DRS_NODE node,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KHE_DRS_EXPR child_e;  int i, di, first_day_index;  bool child_printed;
  if( node == NULL )
  {
    /* print all children */
    di = e->first_day_index;
    HaArrayForEach(e->children, child_e, i)
    {
      for( ;  HaArray(e->child_indexes, di - e->first_day_index) <= i;  di++ )
	fprintf(fp, "%*sending on %s:\n", indent + 2, "",
	  KheTimeGroupId(KheFrameTimeGroup(drs->days_frame, di)));
      KheDrsExprDebug(child_e, node, drs, verbosity, indent + 2, fp);
    }
  }
  else
  {
    /* print the children with a descendant with an entry in the signature */
    if( HaArrayCount(e->children) > 0 )
    {
      fprintf(fp, "(");
      child_printed = false;
      HaArrayForEach(e->children, child_e, i)
      {
	first_day_index = KheDrsExprFirstDebugDayIndex(child_e);
	if( first_day_index <= node->day->day_index &&
	    node->day->day_index <= child_e->last_day_index )
	{
	  if( child_printed )
	    fprintf(fp, ", ");
	  KheDrsExprDebug(child_e, node, drs, verbosity, -1, fp);
	  child_printed = true;
	}
      }
      fprintf(fp, ")");
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheDrsExprDayRangeShow(KHE_DRS_EXPR e,                             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, char buff[100])                       */
/*                                                                           */
/*  Return a string showing the day range of e onto fp.  Use buff for        */
/*  scratch memory.                                                          */
/*                                                                           */
/*****************************************************************************/

static char *KheDrsExprDayRangeShow(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, char buff[100])
{
  KHE_TIME_GROUP tg1, tg2;
  if( e->first_day_index == e->last_day_index )
  {
    tg1 = KheFrameTimeGroup(drs->days_frame, e->first_day_index);
    snprintf(buff, 100, "%s", KheTimeGroupId(tg1));
  }
  else
  {
    tg1 = KheFrameTimeGroup(drs->days_frame, e->first_day_index);
    tg2 = KheFrameTimeGroup(drs->days_frame, e->last_day_index);
    snprintf(buff, 100, "%s-%s", KheTimeGroupId(tg1), KheTimeGroupId(tg2));
  }
  return buff;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprDebug(KHE_DRS_EXPR e, KHE_DRS_NODE node,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of e onto fp with the given verbosity and indent.            */
/*                                                                           */
/*  If node != NULL, only those parts of e that have a value stored in       */
/*  node's signature are printed, along with those values.  In that case     */
/*  indent is ignored and the whole print is on one line.                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprDebug(KHE_DRS_EXPR e, KHE_DRS_NODE node,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      KheDrsExprBusyTimeDebug((KHE_DRS_EXPR_BUSY_TIME) e, node, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      KheDrsExprFreeTimeDebug((KHE_DRS_EXPR_FREE_TIME) e, node, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      KheDrsExprWorkTimeDebug((KHE_DRS_EXPR_WORK_TIME) e, node, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      KheDrsExprBusyDayDebug((KHE_DRS_EXPR_BUSY_DAY) e, node, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      KheDrsExprFreeDayDebug((KHE_DRS_EXPR_FREE_DAY) e, node, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      KheDrsExprWorkDayDebug((KHE_DRS_EXPR_WORK_DAY) e, node, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      KheDrsExprAssignedTaskDebug((KHE_DRS_EXPR_ASSIGNED_TASK) e, node, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrDebug((KHE_DRS_EXPR_OR) e, node, drs, verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndDebug((KHE_DRS_EXPR_AND) e, node, drs, verbosity, indent,fp);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumDebug((KHE_DRS_EXPR_INT_SUM) e, node, drs,
	verbosity, indent, fp);
      break;

    /* *** currently unused
    case KHE_DRS_EXPR_INT_SEQ_TAG:

      KheDrsExprIntSeqDebug((KHE_DRS_EXPR_INT_SEQ) e, node, drs, verbosity,
	indent, fp);
      break;
    *** */

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumDebug((KHE_DRS_EXPR_FLOAT_SUM) e, node, drs,
	verbosity, indent,fp);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevDebug((KHE_DRS_EXPR_INT_DEV) e, node, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevDebug((KHE_DRS_EXPR_FLOAT_DEV) e, node, drs,
	verbosity, indent,fp);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostDebug((KHE_DRS_EXPR_COST) e, node, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombDebug((KHE_DRS_EXPR_INT_SUM_COMB) e, node, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombDebug((KHE_DRS_EXPR_INT_SEQ_COMB) e, node, drs,
	verbosity, indent, fp);
      break;

    default:

      HnAbort("KheDrsExprDebug internal error: unknown tag %d", e->tag);
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_DAY_RANGE"                                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DAY_RANGE KheDrsDayRangeMake(int first_day_index,                */
/*    int last_day_index)                                                    */
/*                                                                           */
/*  Return a new interval with these endpoints.                              */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DAY_RANGE KheDrsDayRangeMake(int first_day_index,
  int last_day_index)
{
  KHE_DRS_DAY_RANGE res;
  res.first_day_index = first_day_index;
  res.last_day_index = last_day_index;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayRangeMerge(ARRAY_KHE_DRS_DAY_RANGE *day_ranges,            */
/*    KHE_DRS_DAY_RANGE ddr)                                                 */
/*                                                                           */
/*  Merge ddr into *day_ranges, a sorted array of day ranges.                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayRangeMerge(ARRAY_KHE_DRS_DAY_RANGE *day_ranges,
  KHE_DRS_DAY_RANGE ddr)
{
  KHE_DRS_DAY_RANGE ddr2;  int i;

  /* if ddr is empty, it changes nothing */
  if( ddr.first_day_index > ddr.last_day_index )
    return;

  HaArrayForEach(*day_ranges, ddr2, i)
  {
    /***********************************************************************/
    /*                                                                     */
    /*  Invariant: day_ranges and ddr may have changed, but it remains      */
    /*  true that merging ddr with day_ranges gives the desired result.     */
    /*  Also, day_ranges[0..i-1] precede and do not interact with ddr.      */
    /*                                                                     */
    /***********************************************************************/

    if( ddr2.last_day_index + 1 < ddr.first_day_index )
    {
      /* ddr2 precedes ddr and does not interact with it; continue */
      continue;
    }

    if( ddr.last_day_index + 1 < ddr2.first_day_index )
    {
      /* ddr precedes ddr2 and does not interact with it, or with day_ranges */
      HaArrayAdd(*day_ranges, i, ddr);
      return;
    }

    /* ddr2 interacts with ddr, so expand ddr to include ddr2, then */
    /* delete ddr2 and continue */
    if( ddr2.first_day_index < ddr.first_day_index )
      ddr.first_day_index = ddr2.first_day_index;
    if( ddr2.last_day_index > ddr.last_day_index )
      ddr.last_day_index = ddr2.last_day_index;
    HaArrayDeleteAndShift(*day_ranges, i);
    i--;
  }

  /* ddr does not interact with any day_ranges, and it follows them all */
  HaArrayAddLast(*day_ranges, ddr);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDayRangeSubset(KHE_DRS_DAY_RANGE si1, KHE_DRS_DAY_RANGE si2)  */
/*                                                                           */
/*  Return true if si1 is a subset of si2.                                   */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsDayRangeSubset(KHE_DRS_DAY_RANGE si1, KHE_DRS_DAY_RANGE si2)
{
  return si1.first_day_index >= si2.first_day_index &&
    si1.last_day_index <= si2.last_day_index;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDayRangesDebug(ARRAY_KHE_DRS_DAY_RANGE *day_ranges,              */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of *day_ranges onto fp with the given verbosity and indent.  */
/*                                                                           */
/*****************************************************************************/

static void KheDayRangesDebug(ARRAY_KHE_DRS_DAY_RANGE *day_ranges,
  int verbosity, int indent, FILE *fp)
{
  KHE_DRS_DAY_RANGE si;  int i;
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "[");
  HaArrayForEach(*day_ranges, si, i)
  {
    if( i > 0 )
      fprintf(fp, ", ");
    if( si.first_day_index == si.last_day_index )
      fprintf(fp, "%d", si.first_day_index);
    else
      fprintf(fp, "%d-%d", si.first_day_index, si.last_day_index);
  }
  fprintf(fp, "]");
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_RESOURCE_ON_DAY"                                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDayMake(                         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new resource on day object.                                       */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDayMake(
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_RESOURCE_ON_DAY res;
  HaMake(res, drs->arena);
  res->assigned_task = NULL;
  res->solve_task = NULL;
  HaArrayInit(res->active_today, drs->arena);
  HaArrayInit(res->leaf_today, drs->arena);
  res->sig_len = 0;
  HaArrayInit(res->dom_tests, drs->arena);
  HaArrayInit(res->eq_dom_test_indexes, drs->arena);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceOnDayAddDomTest(KHE_DRS_RESOURCE_ON_DAY drd,          */
/*    KHE_DRS_DOM_TEST dom_test)                                             */
/*                                                                           */
/*  Add dom_test to drd.  This may change the eq_dom_test_indexes array as   */
/*  well as the dom_tests array.                                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceOnDayAddDomTest(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_DOM_TEST dom_test)
{
  if( dom_test.type == KHE_DRS_DOM_EQ || dom_test.type == KHE_DRS_DOM_EQ_LOOSE )
    HaArrayAddLast(drd->eq_dom_test_indexes, HaArrayCount(drd->dom_tests));
  HaArrayAddLast(drd->dom_tests, dom_test);
}



/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceOnDayDebug(KHE_DRS_RESOURCE_ON_DAY drd,               */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of drd onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/
static void KheDrsTaskDebug(KHE_DRS_TASK dt, int verbosity,
  int indent, FILE *fp);

static void KheDrsResourceOnDayDebug(KHE_DRS_RESOURCE_ON_DAY drd,
  int verbosity, int indent, FILE *fp)
{
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  if( drd->assigned_task != NULL )
    KheDrsTaskDebug(drd->assigned_task, 1, -1, fp);
  else
    fprintf(fp, "-");
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_RESOURCE - converting monitors to expressions"        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddAvoidClashesMonitor(KHE_DRS_RESOURCE dr,           */
/*    KHE_AVOID_CLASHES_MONITOR m, KHE_DYNAMIC_RESOURCE_SOLVER drs)          */
/*                                                                           */
/*  Add avoid clashes monitor m to dr.                                       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceAddAvoidClashesMonitor(KHE_DRS_RESOURCE dr,
  KHE_AVOID_CLASHES_MONITOR m, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* ignoring this; there will be no clashes */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddAvoidUnavailableTimesMonitor(KHE_DRS_RESOURCE dr,  */
/*    KHE_AVOID_UNAVAILABLE_TIMES_MONITOR m, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  Add avoid unavailable times monitor m to dr.                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceAddAvoidUnavailableTimesMonitor(KHE_DRS_RESOURCE dr,
  KHE_AVOID_UNAVAILABLE_TIMES_MONITOR m, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_AVOID_UNAVAILABLE_TIMES_CONSTRAINT c;  KHE_DRS_EXPR_BUSY_TIME eb;
  int i;  KHE_TIME t;  KHE_DRS_EXPR_COST ec;  KHE_COST_FUNCTION cost_fn;
  KHE_TIME_GROUP tg;  KHE_DRS_EXPR_INT_SUM eis;  KHE_COST combined_weight;
  KHE_DRS_DOM_TEST unused_dom_test, le_dom_test;
  c = KheAvoidUnavailableTimesMonitorConstraint(m);
  tg = KheAvoidUnavailableTimesConstraintUnavailableTimes(c);
  unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  combined_weight = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  if( combined_weight > 0 && KheTimeGroupTimeCount(tg) > 0 )
  {
    cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
    if( cost_fn == KHE_LINEAR_COST_FUNCTION || KheTimeGroupTimeCount(tg) == 1 )
    {
      /* a separate expression tree for each time */
      for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
      {
	t = KheTimeGroupTime(tg, i);
	ec = KheDrsExprCostMake((KHE_MONITOR) m, cost_fn, combined_weight, drs);
	HaArrayAddLast(dr->exprs, (KHE_DRS_EXPR) ec);
	/* ***
	KheDrsExprAddChild((KHE_DRS_EXPR) drs->root_expr, (KHE_DRS_EXPR) ec);
	*** */
	eb = KheDrsExprBusyTimeMake(dr, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eb);
      }
    }
    else
    {
      /* a single expression tree (max limit of 0, in effect) */
      ec = KheDrsExprCostMake((KHE_MONITOR) m, cost_fn, combined_weight, drs);
      HaArrayAddLast(dr->exprs, (KHE_DRS_EXPR) ec);
      /* KheDrsExprAddChild((KHE_DRS_EXPR) drs->root_expr, (KHE_DRS_EXPR) ec);*/
      le_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);
      eis = KheDrsExprIntSumMake(le_dom_test, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eis);
      for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
      {
	t = KheTimeGroupTime(tg, i);
	eb = KheDrsExprBusyTimeMake(dr, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eis, (KHE_DRS_EXPR) eb);
      }
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddLimitIdleTimesMonitor(KHE_DRS_RESOURCE dr,         */
/*    KHE_LIMIT_IDLE_TIMES_MONITOR m, KHE_DYNAMIC_RESOURCE_SOLVER drs)       */
/*                                                                           */
/*  Add limit idle times monitor m to dr.                                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceAddLimitIdleTimesMonitor(KHE_DRS_RESOURCE dr,
  KHE_LIMIT_IDLE_TIMES_MONITOR m, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* not included in the solve, and that fact is documented */
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheTimeGroupContainsWholeDay(KHE_TIME_GROUP tg, int start_index,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_TIME_GROUP *day_tg)               */
/*                                                                           */
/*  If tg contains a complete day's worth of times, starting from            */
/*  start_index, then return true and set *day_tg to the day time group,     */
/*  otherwise return false.                                                  */
/*                                                                           */
/*****************************************************************************/

static bool KheTimeGroupContainsWholeDay(KHE_TIME_GROUP tg, int start_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_TIME_GROUP *day_tg)
{
  KHE_TIME tg_t, day_t;  int i;  KHE_TIME_GROUP d_tg;
  tg_t = KheTimeGroupTime(tg, start_index);
  d_tg = KheFrameTimeTimeGroup(drs->days_frame, tg_t);
  for( i = 0;  i < KheTimeGroupTimeCount(d_tg);  i++ )
  {
    day_t = KheTimeGroupTime(d_tg, i);
    if( start_index + i >= KheTimeGroupTimeCount(tg) )
      return *day_tg = NULL, false;
    tg_t = KheTimeGroupTime(tg, start_index + i);
    if( day_t != tg_t )
      return *day_tg = NULL, false;
  }
  return *day_tg = d_tg, true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMakeOrNodeAndChildren(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,    */
/*    KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make an OR node with one BUSY child for each time of tg, and add the     */
/*  OR node as a child of prnt.  If tg has exactly one time, optimize by     */
/*  adding just the BUSY node for that time as a child of prnt.              */
/*                                                                           */
/*  Either way, add dom_test to the root of these vertices.                  */
/*                                                                           */
/*****************************************************************************/

static void KheMakeOrNodeAndChildren(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,
  KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_OR eo;  KHE_DRS_EXPR_BUSY_TIME ebt;  int i;  KHE_TIME t;
  KHE_DRS_DOM_TEST unused_dom_test;  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_TIME_GROUP day_tg;
  if( KheTimeGroupTimeCount(tg) == 1 )
  {
    /* just a single BUSY_TIME node */
    t = KheTimeGroupTime(tg, 0);
    ebt = KheDrsExprBusyTimeMake(dr, t, dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) ebt);
  }
  else if( KheTimeGroupContainsWholeDay(tg, 0, drs, &day_tg) &&
    KheTimeGroupTimeCount(tg) == KheTimeGroupTimeCount(day_tg) )
  {
    /* just a single BUSY_DAY node */
    ebd = KheDrsExprBusyDayMake(dr, day_tg, dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) ebd);
  }
  else
  {
    /* an OR node with BUSY_DAY and BUSY_TIME children */
    eo = KheDrsExprOrMake(dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) eo);
    unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
    for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
    {
      if( KheTimeGroupContainsWholeDay(tg, i, drs, &day_tg) )
      {
	/* add a BUSY_DAY node */
	ebd = KheDrsExprBusyDayMake(dr, day_tg, dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eo, (KHE_DRS_EXPR) ebd);
	i += KheTimeGroupTimeCount(day_tg) - 1;
      }
      else
      {
	/* add a BUSY_TIME node */
	t = KheTimeGroupTime(tg, i);
	ebt = KheDrsExprBusyTimeMake(dr, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eo, (KHE_DRS_EXPR) ebt);
      }
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMakeAndNodeAndChildren(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,   */
/*    KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make an AND node with one FREE child for each time of tg, and add the    */
/*  AND node as a child of prnt.  If tg has exactly one time, optimize by    */
/*  adding just the FREE node for that time as a child of prnt.              */
/*                                                                           */
/*  Either way, add dom_test to the root of these vertices.                  */
/*                                                                           */
/*****************************************************************************/

static void KheMakeAndNodeAndChildren(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,
  KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_AND ea;  KHE_DRS_EXPR_FREE_TIME eft;  int i;  KHE_TIME t;
  KHE_DRS_DOM_TEST unused_dom_test;  KHE_DRS_EXPR_FREE_DAY efd;
  KHE_TIME_GROUP day_tg;
  if( KheTimeGroupTimeCount(tg) == 1 )
  {
    /* just a single FREE node */
    t = KheTimeGroupTime(tg, 0);
    eft = KheDrsExprFreeTimeMake(dr, t, dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) eft);
  }
  else if( KheTimeGroupContainsWholeDay(tg, 0, drs, &day_tg) &&
    KheTimeGroupTimeCount(tg) == KheTimeGroupTimeCount(day_tg) )
  {
    /* just a single FREE_DAY node */
    efd = KheDrsExprFreeDayMake(dr, day_tg, dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) efd);
  }
  else
  {
    /* an AND node with FREE_DAY and FREE_TIME children */
    ea = KheDrsExprAndMake(dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) ea);
    unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
    for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
    {
      if( KheTimeGroupContainsWholeDay(tg, i, drs, &day_tg) )
      {
	/* add a FREE_DAY node */
	efd = KheDrsExprFreeDayMake(dr, day_tg, dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) ea, (KHE_DRS_EXPR) efd);
	i += KheTimeGroupTimeCount(day_tg) - 1;
      }
      else
      {
	/* add a FREE_TIME node */
	t = KheTimeGroupTime(tg, i);
	eft = KheDrsExprFreeTimeMake(dr, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) ea, (KHE_DRS_EXPR) eft);
      }
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheLimitsToTests(int min_limit, int max_limit, int max_max_limit,   */
/*    bool allow_zero, KHE_DRS_DOM_TEST *alpha_dom_test,                     */
/*    KHE_DRS_DOM_TEST *beta_dom_test)                                       */
/*                                                                           */
/*  Convert limits (min_limit, max_limit, allow_zero) into dominance tests.  */
/*  if min_limit <= 0 there is no min limit; if max_limit >= max_max_limit   */
/*  there is no maximum limit.                                               */
/*                                                                           */
/*  The values returned in *alpha_dom_test and *beta_dom_test are referred   */
/*  to as alpha and beta in the User's Guide.                                */
/*                                                                           */
/*  Note.  If there is a trivial allow_zero flag (if min_limit <= 1), an     */
/*  adjustment will have already been made by now to remove it.  So this     */
/*  call may assume that allow_zero implies min_limit > 1.                   */
/*                                                                           */
/*****************************************************************************/

static void KheLimitsToTests(int min_limit, int max_limit, int max_max_limit,
  bool allow_zero, KHE_DRS_DOM_TEST *alpha_dom_test,
  KHE_DRS_DOM_TEST *beta_dom_test)
{
  if( allow_zero /* && min_limit > 1 */ )
  {
    /* fall back on equality when there is a non-trivial allow_zero flag */
    *alpha_dom_test = *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_EQ);
  }
  else if( max_limit >= max_max_limit )
  {
    if( min_limit <= 0 )
    {
      /* no max_limit, and no min_limit (should never happen) */
      HnAbort("KheLimitsToTests: no max limit, no min limit");
      *alpha_dom_test = *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
    }
    else
    {
      /* no max_limit, and min_limit */
      *alpha_dom_test = KheDrsDomTestMakeLoose(KHE_DRS_DOM_GE_LOOSE, min_limit);
      *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_GE);
    }
  }
  else
  {
    if( min_limit <= 0 )
    {
      /* max_limit, and no min_limit */
      *alpha_dom_test = *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);
    }
    else
    {
      /* max_limit, and min_limit */
      *alpha_dom_test = KheDrsDomTestMakeLoose(KHE_DRS_DOM_EQ_LOOSE, min_limit);
      *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_EQ);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsClusterMonitorContainsWholePositiveDay(                       */
/*    KHE_CLUSTER_BUSY_TIMES_MONITOR m, int start_index,                     */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_TIME_GROUP *day_tg)               */
/*                                                                           */
/*  If m contains a sequence of singleton time groups starting at day_index  */
/*  that have postive polarity and together cover a complete day, return     */
/*  true and set *day_tg to the day time group;  otherwise return false.     */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsClusterMonitorContainsWholePositiveDay(
  KHE_CLUSTER_BUSY_TIMES_MONITOR m, int start_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_TIME_GROUP *day_tg)
{
  KHE_TIME day_t;  int i, tg_count;  KHE_TIME_GROUP m_tg, d_tg;
  KHE_POLARITY po;  KHE_TIME t;
  tg_count = KheClusterBusyTimesMonitorTimeGroupCount(m);
  m_tg = KheClusterBusyTimesMonitorTimeGroup(m, start_index, &po);
  if( po == KHE_POSITIVE && KheTimeGroupTimeCount(m_tg) == 1 )
  {
    t = KheTimeGroupTime(m_tg, 0);
    d_tg = KheFrameTimeTimeGroup(drs->days_frame, t);
    for( i = 0;  i < KheTimeGroupTimeCount(d_tg);  i++ )
    {
      day_t = KheTimeGroupTime(d_tg, i);
      if( start_index + i >= tg_count )
	return *day_tg = NULL, false;
      m_tg = KheClusterBusyTimesMonitorTimeGroup(m, start_index + i, &po);
      if( po != KHE_POSITIVE || KheTimeGroupTimeCount(m_tg) != 1 ||
	  KheTimeGroupTime(m_tg, 0) != day_t )
	return *day_tg = NULL, false;
    }
    return *day_tg = d_tg, true;
  }
  else
    return *day_tg = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddClusterBusyTimesMonitor(KHE_DRS_RESOURCE dr,       */
/*    KHE_CLUSTER_BUSY_TIMES_MONITOR m, KHE_DYNAMIC_RESOURCE_SOLVER drs)     */
/*                                                                           */
/*  Add cluster busy times monitor m to dr.                                  */
/*                                                                           */
/*  Implementation note.  If m contains a sequence of consecutive            */
/*  singleton time groups with positive polarity that add up to a            */
/*  complete day, then one BUSY_DAY node is generated for them instead       */
/*  of one BUSY_TIME node for each.  This is correct the sum reported        */
/*  in either case is 1 if the day is busy and 0 otherwise.  It cannot       */
/*  be more than 1 because the resource takes at most one task each day.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceAddClusterBusyTimesMonitor(KHE_DRS_RESOURCE dr,
  KHE_CLUSTER_BUSY_TIMES_MONITOR m, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c;  KHE_TIME_GROUP tg, day_tg;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_weight;  bool allow_zero;
  int i, min_limit, max_limit, history_before, history_after, tg_count;
  KHE_DRS_EXPR_INT_SUM_COMB eisc;  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_POLARITY po;  KHE_DRS_DOM_TEST alpha_dom_test, beta_dom_test;

  /* boilerplate, and return if constraint does nothing */
  c = KheClusterBusyTimesMonitorConstraint(m);
  combined_weight = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  min_limit = KheClusterBusyTimesConstraintMinimum(c);
  max_limit = KheClusterBusyTimesConstraintMaximum(c);
  allow_zero = KheClusterBusyTimesConstraintAllowZero(c);
  history_before = KheClusterBusyTimesMonitorHistory(m);
  history_after = KheClusterBusyTimesMonitorHistoryAfter(m);
  if( allow_zero && min_limit <= 1 )
    min_limit = 0, allow_zero = false;
  tg_count = KheClusterBusyTimesConstraintTimeGroupCount(c);
  if( (min_limit <= 0 && max_limit >= tg_count) || combined_weight == 0 )
    return;

  /* work out the various dom_tests */
  KheLimitsToTests(min_limit, max_limit, tg_count, allow_zero,
    &alpha_dom_test, &beta_dom_test);

  /* one int_cost_comb node at top of tree */
  eisc = KheDrsExprIntSumCombMake((KHE_MONITOR) m, cost_fn, combined_weight,
    min_limit, max_limit, allow_zero, history_before, history_after,
    alpha_dom_test, drs);
  HaArrayAddLast(dr->exprs, (KHE_DRS_EXPR) eisc);
  /* KheDrsExprAddChild((KHE_DRS_EXPR) drs->root_expr, (KHE_DRS_EXPR) eisc); */

  /* one OR or AND node for each time group */
  /* or a BUSY_DAY or FREE_DAY node if all singles of one day present */
  for( i = 0;  i < tg_count;  i++ )
  {
    if( KheDrsClusterMonitorContainsWholePositiveDay(m, i, drs, &day_tg) )
    {
      /* add a BUSY_DAY node */
      ebd = KheDrsExprBusyDayMake(dr, day_tg, beta_dom_test, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) ebd);
      i += KheTimeGroupTimeCount(day_tg) - 1;
    }
    else
    {
      tg = KheClusterBusyTimesMonitorTimeGroup(m, i, &po);
      if( po == KHE_POSITIVE )
      {
	/* make an OR node with one BUSY node for each time of tg */
	KheMakeOrNodeAndChildren(dr, (KHE_DRS_EXPR) eisc, tg,beta_dom_test,drs);
      }
      else
      {
	/* make an AND node with one FREE node for each time of tg */
	KheMakeAndNodeAndChildren(dr, (KHE_DRS_EXPR) eisc,tg,beta_dom_test,drs);
      }
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddLimitBusyTimesMonitor(KHE_DRS_RESOURCE dr,         */
/*    KHE_LIMIT_BUSY_TIMES_MONITOR m, KHE_DYNAMIC_RESOURCE_SOLVER drs)       */
/*                                                                           */
/*  Add limit busy times monitor m to dr.                                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceAddLimitBusyTimesMonitor(KHE_DRS_RESOURCE dr,
  KHE_LIMIT_BUSY_TIMES_MONITOR m, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_LIMIT_BUSY_TIMES_CONSTRAINT c;  KHE_TIME_GROUP tg, day_tg;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_weight;
  int i, j, junk, min_limit, max_limit;  bool allow_zero;
  KHE_DRS_EXPR_INT_SUM_COMB eisc;  KHE_TIME t;
  KHE_DRS_EXPR_INT_SUM eis, eis2;  KHE_DRS_EXPR_COST ec;
  KHE_DRS_EXPR_INT_DEV eid;
  KHE_DRS_EXPR_BUSY_TIME ebt;
  KHE_DRS_DOM_TEST unused_dom_test, le_dom_test, alpha_dom_test, beta_dom_test;

  /* boilerplate, and return if constraint does nothing */
  c = KheLimitBusyTimesMonitorConstraint(m);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  combined_weight = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  min_limit = KheLimitBusyTimesConstraintMinimum(c);
  max_limit = KheLimitBusyTimesConstraintMaximum(c);
  allow_zero = KheLimitBusyTimesConstraintAllowZero(c);
  if( allow_zero && min_limit <= 1 )
    min_limit = 0, allow_zero = false;
  unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  le_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);
  if( combined_weight == 0 )
    return;

  if( cost_fn == KHE_LINEAR_COST_FUNCTION ||
      KheLimitBusyTimesMonitorTimeGroupCount(m) == 1 )
  {
    /* one INT_SUM_COMB for each time group */
    /* unless the time group is one day, min_limit = 0, and max_limit = 1 */
    for( i = 0;  i < KheLimitBusyTimesMonitorTimeGroupCount(m);  i++ )
    {
      tg = KheLimitBusyTimesMonitorTimeGroup(m, i, &junk);
      if( min_limit == 0 && max_limit == 1 &&
	  KheTimeGroupContainsWholeDay(tg, 0, drs, &day_tg) &&
	  KheTimeGroupTimeCount(tg) == KheTimeGroupTimeCount(day_tg) )
      {
	/* constraint specifies at most one task per day; trivial, so omit */
      }
      else if( min_limit > 0 || max_limit < KheTimeGroupTimeCount(tg) )
      {
	KheLimitsToTests(min_limit, max_limit, KheTimeGroupTimeCount(tg),
	  allow_zero, &alpha_dom_test, &beta_dom_test);
	eisc = KheDrsExprIntSumCombMake((KHE_MONITOR) m, cost_fn,
	  combined_weight, min_limit, max_limit, allow_zero, 0, 0,
	  alpha_dom_test, drs);
	HaArrayAddLast(dr->exprs, (KHE_DRS_EXPR) eisc);
	/* ***
	KheDrsExprAddChild((KHE_DRS_EXPR) drs->root_expr, (KHE_DRS_EXPR) eisc);
	*** */
	for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
	{
	  t = KheTimeGroupTime(tg, j);
	  ebt = KheDrsExprBusyTimeMake(dr, t, unused_dom_test, drs);
	  KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) ebt);
	}
      }
    }
  }
  else
  {
    /* one COST and INT_SUM at root of tree */
    ec = KheDrsExprCostMake((KHE_MONITOR) m, cost_fn, combined_weight, drs);
    HaArrayAddLast(dr->exprs, (KHE_DRS_EXPR) ec);
    /* KheDrsExprAddChild((KHE_DRS_EXPR) drs->root_expr, (KHE_DRS_EXPR) ec); */
    eis = KheDrsExprIntSumMake(le_dom_test, drs);
    KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eis);

    /* one INT_DEV and INT_SUM for each time group */
    for( i = 0;  i < KheLimitBusyTimesMonitorTimeGroupCount(m);  i++ )
    {
      tg = KheLimitBusyTimesMonitorTimeGroup(m, i, &junk);
      eid = KheDrsExprIntDevMake(min_limit, max_limit, allow_zero, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) eis, (KHE_DRS_EXPR) eid);
      KheLimitsToTests(min_limit, max_limit, KheTimeGroupTimeCount(tg),
	allow_zero, &alpha_dom_test, &beta_dom_test);
      eis2 = KheDrsExprIntSumMake(alpha_dom_test, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) eid, (KHE_DRS_EXPR) eis2);
      for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
      {
	t = KheTimeGroupTime(tg, j);
	ebt = KheDrsExprBusyTimeMake(dr, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eis2, (KHE_DRS_EXPR) ebt);
      }
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddLimitWorkloadMonitor(KHE_DRS_RESOURCE dr,          */
/*    KHE_LIMIT_WORKLOAD_MONITOR m, KHE_DYNAMIC_RESOURCE_SOLVER drs)         */
/*                                                                           */
/*  Add limit workload monitor m to dr.                                      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceAddLimitWorkloadMonitor(KHE_DRS_RESOURCE dr,
  KHE_LIMIT_WORKLOAD_MONITOR m, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_LIMIT_WORKLOAD_CONSTRAINT c;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_weight;
  int i, j, min_limit, max_limit;  bool allow_zero;  float junk;
  KHE_TIME_GROUP tg;  KHE_TIME t;  KHE_DRS_EXPR_WORK_TIME ew;
  KHE_DRS_EXPR_INT_SUM eis;  KHE_DRS_EXPR_COST ec;
  KHE_DRS_EXPR_FLOAT_SUM efs;  KHE_DRS_EXPR_FLOAT_DEV efd;
  KHE_DRS_DOM_TEST unused_dom_test, le_dom_test, alpha_dom_test, beta_dom_test;

  /* boilerplate, and return if constraint does nothing */
  c = KheLimitWorkloadMonitorConstraint(m);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  combined_weight = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  min_limit = KheLimitWorkloadConstraintMinimum(c);
  max_limit = KheLimitWorkloadConstraintMaximum(c);
  allow_zero = KheLimitWorkloadConstraintAllowZero(c);
  if( allow_zero && min_limit <= 1 )
    min_limit = 0, allow_zero = false;
  unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  le_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);
  if( combined_weight == 0 )
    return;

  /* one COST and INT_SUM at the root */
  ec = KheDrsExprCostMake((KHE_MONITOR) m, cost_fn, combined_weight, drs);
  HaArrayAddLast(dr->exprs, (KHE_DRS_EXPR) ec);
  /* KheDrsExprAddChild((KHE_DRS_EXPR) drs->root_expr, (KHE_DRS_EXPR) ec); */
  eis = KheDrsExprIntSumMake(le_dom_test, drs);
  KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eis);

  /* one FLOAT_DEV and FLOAT_SUM for each time group */
  KheLimitsToTests(min_limit, max_limit, INT_MAX, allow_zero,
    &alpha_dom_test, &beta_dom_test);
  for( i = 0;  i < KheLimitWorkloadMonitorTimeGroupCount(m);  i++ )
  {
    tg = KheLimitWorkloadMonitorTimeGroup(m, i, &junk);
    efd = KheDrsExprFloatDevMake(min_limit, max_limit, allow_zero, drs);
    KheDrsExprAddChild((KHE_DRS_EXPR) eis, (KHE_DRS_EXPR) efd);
    efs = KheDrsExprFloatSumMake(alpha_dom_test, drs);
    KheDrsExprAddChild((KHE_DRS_EXPR) efd, (KHE_DRS_EXPR) efs);
    for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
    {
      t = KheTimeGroupTime(tg, j);
      ew = KheDrsExprWorkTimeMake(dr, t, unused_dom_test, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) efs, (KHE_DRS_EXPR) ew);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddLimitActiveIntervalsMonitor(KHE_DRS_RESOURCE dr,   */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR m, KHE_DYNAMIC_RESOURCE_SOLVER drs) */
/*                                                                           */
/*  Add limit active intervals monitor m to dr.                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceAddLimitActiveIntervalsMonitor(KHE_DRS_RESOURCE dr,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR m, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c;  KHE_TIME_GROUP tg;  KHE_POLARITY po;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_weight;
  int i, min_limit, max_limit, history_before, history_after, tg_count;
  KHE_DRS_EXPR_INT_SEQ_COMB eisc;
  KHE_DRS_DOM_TEST alpha_dom_test, beta_dom_test;

  /* boilerplate, and return if constraint does nothing */
  c = KheLimitActiveIntervalsMonitorConstraint(m);
  combined_weight = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  min_limit = KheLimitActiveIntervalsConstraintMinimum(c);
  max_limit = KheLimitActiveIntervalsConstraintMaximum(c);
  history_before = KheLimitActiveIntervalsMonitorHistory(m);
  history_after = KheLimitActiveIntervalsMonitorHistoryAfter(m);
  tg_count = KheLimitActiveIntervalsConstraintTimeGroupCount(c);
  if( (min_limit <= 1 && max_limit >= tg_count) || combined_weight == 0 )
    return;

  /* work out the various dom_tests */
  KheLimitsToTests(min_limit, max_limit, tg_count, false,
    &alpha_dom_test, &beta_dom_test);

  /* one int_seq_comb node at top of tree */
  eisc = KheDrsExprIntSeqCombMake((KHE_MONITOR) m, cost_fn, combined_weight,
    min_limit, max_limit, history_before, history_after, alpha_dom_test, drs);
  HaArrayAddLast(dr->exprs, (KHE_DRS_EXPR) eisc);
  /* KheDrsExprAddChild((KHE_DRS_EXPR) drs->root_expr, (KHE_DRS_EXPR) eisc); */

  /* one OR or AND node for each time group */
  for( i = 0;  i < KheLimitActiveIntervalsMonitorTimeGroupCount(m);  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(m, i, &po);
    if( po == KHE_POSITIVE )
    {
      /* make an OR node with one BUSY node for each time of tg */
      KheMakeOrNodeAndChildren(dr, (KHE_DRS_EXPR) eisc, tg, beta_dom_test, drs);
    }
    else
    {
      /* make an AND node with one FREE node for each time of tg */
      KheMakeAndNodeAndChildren(dr, (KHE_DRS_EXPR) eisc, tg, beta_dom_test,drs);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddMonitor(KHE_DRS_RESOURCE dr,                       */
/*    KHE_MONITOR m, KHE_DYNAMIC_RESOURCE_SOLVER drs)                        */
/*                                                                           */
/*  Add m to dr.                                                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceAddMonitor(KHE_DRS_RESOURCE dr,
  KHE_MONITOR m, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( KheMonitorTag(m) )
  {
    case KHE_AVOID_CLASHES_MONITOR_TAG:

      KheDrsResourceAddAvoidClashesMonitor(dr,
        (KHE_AVOID_CLASHES_MONITOR) m, drs);
      break;

    case KHE_AVOID_UNAVAILABLE_TIMES_MONITOR_TAG:

      KheDrsResourceAddAvoidUnavailableTimesMonitor(dr,
        (KHE_AVOID_UNAVAILABLE_TIMES_MONITOR) m, drs);
      break;

    case KHE_LIMIT_IDLE_TIMES_MONITOR_TAG:

      KheDrsResourceAddLimitIdleTimesMonitor(dr,
        (KHE_LIMIT_IDLE_TIMES_MONITOR) m, drs);
      break;

    case KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG:

      KheDrsResourceAddClusterBusyTimesMonitor(dr,
        (KHE_CLUSTER_BUSY_TIMES_MONITOR) m, drs);
      break;

    case KHE_LIMIT_BUSY_TIMES_MONITOR_TAG:

      KheDrsResourceAddLimitBusyTimesMonitor(dr,
        (KHE_LIMIT_BUSY_TIMES_MONITOR) m, drs);
      break;

    case KHE_LIMIT_WORKLOAD_MONITOR_TAG:

      KheDrsResourceAddLimitWorkloadMonitor(dr,
        (KHE_LIMIT_WORKLOAD_MONITOR) m, drs);
      break;

    case KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG:

      KheDrsResourceAddLimitActiveIntervalsMonitor(dr,
        (KHE_LIMIT_ACTIVE_INTERVALS_MONITOR) m, drs);
      break;

    default:

      HnAbort("KheDrsResourceAddMonitor internal error (tag %d)",
        KheMonitorTag(m));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  int KheMonitorLimit(KHE_MONITOR m)                                       */
/*                                                                           */
/*  Return the maximum limit of m if there is one, or else the minimum       */
/*  limit if there is one, or else 0.                                        */
/*                                                                           */
/*****************************************************************************/

static int KheMonitorLimit(KHE_MONITOR m)
{
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT cbtc;  int val;
  KHE_LIMIT_BUSY_TIMES_CONSTRAINT lbtc;
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim;
  switch( KheMonitorTag(m) )
  {
    case KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG:

      cbtc = KheClusterBusyTimesMonitorConstraint(
	(KHE_CLUSTER_BUSY_TIMES_MONITOR) m);
      val = KheClusterBusyTimesConstraintMaximum(cbtc);
      if( val < KheClusterBusyTimesConstraintTimeGroupCount(cbtc) )
	return val;
      val = KheClusterBusyTimesConstraintMinimum(cbtc);
      return val;

    case KHE_LIMIT_BUSY_TIMES_MONITOR_TAG:

      lbtc = KheLimitBusyTimesMonitorConstraint(
	(KHE_LIMIT_BUSY_TIMES_MONITOR) m);
      val = KheLimitBusyTimesConstraintMaximum(lbtc);
      if( val < INT_MAX )
	return val;
      val = KheLimitBusyTimesConstraintMinimum(lbtc);
      if( val < INT_MAX )
	return val;
      return 0;

    case KHE_LIMIT_WORKLOAD_MONITOR_TAG:

      /* could do better here */
      return 0;

    case KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG:

      laim = (KHE_LIMIT_ACTIVE_INTERVALS_MONITOR) m;
      val = KheLimitActiveIntervalsMonitorMaximum(laim);
      if( val < KheLimitActiveIntervalsMonitorTimeGroupCount(laim) )
	return val;
      val = KheLimitActiveIntervalsMonitorMinimum(laim);
      if( val < INT_MAX )
	return val;
      return 0;

    case KHE_AVOID_CLASHES_MONITOR_TAG:
    case KHE_AVOID_UNAVAILABLE_TIMES_MONITOR_TAG:
    case KHE_LIMIT_IDLE_TIMES_MONITOR_TAG:
    default:

      return 0;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  int KheMonitorLargestLimitFirstCmp(const void *p1, const void *p2)       */
/*                                                                           */
/*  Comparison function for sorting monitors so that those with the          */
/*  largest limits come first.                                               */
/*                                                                           */
/*****************************************************************************/

static int KheMonitorLargestLimitFirstCmp(const void *p1, const void *p2)
{
  KHE_MONITOR m1 = * (KHE_MONITOR *) p1;
  KHE_MONITOR m2 = * (KHE_MONITOR *) p2;
  return KheMonitorLimit(m2) - KheMonitorLimit(m1);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_RESOURCE"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsResourceHasTimeExpr(KHE_DRS_RESOURCE dr,                      */
/*    KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME time,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e)                  */
/*                                                                           */
/*  If dr has a BUSY_TIME, FREE_TIME, or WORK_TIME expression with the       */
/*  given tag, dom_test, and time, then return true and set *res_e to it.    */
/*  Otherwise return false with *res_e set to NULL.                          */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsResourceHasTimeExpr(KHE_DRS_RESOURCE dr,
  KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME time,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e)
{
  int i;  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_DRS_EXPR e;
  KHE_DRS_EXPR_BUSY_TIME ebt;
  KHE_DRS_EXPR_FREE_TIME eft;
  KHE_DRS_EXPR_WORK_TIME ewt;
  i = KheFrameTimeIndex(drs->days_frame, time);
  drd = HaArray(dr->days, i);
  HaArrayForEach(drd->leaf_today, e, i)
    if( e->tag == tag && KheDrsDomTestEqual(e->dom_test, dom_test) )
    {
      switch( e->tag )
      {
	case KHE_DRS_EXPR_BUSY_TIME_TAG:

	  ebt = (KHE_DRS_EXPR_BUSY_TIME) e;
	  if( ebt->time == time )
	    return *res_e = e, true;
	  break;

	case KHE_DRS_EXPR_FREE_TIME_TAG:

	  eft = (KHE_DRS_EXPR_FREE_TIME) e;
	  if( eft->time == time )
	    return *res_e = e, true;
	  break;

	case KHE_DRS_EXPR_WORK_TIME_TAG:

	  ewt = (KHE_DRS_EXPR_WORK_TIME) e;
	  if( ewt->time == time )
	    return *res_e = e, true;
	  break;

	default:

	  HnAbort("KheDrsResourceHasTimeExpr: invalid tag (%d)", e->tag);
      }
    }
  return *res_e = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsHasTimeExpr(KHE_DYNAMIC_RESOURCE_SOLVER drs,                  */
/*    KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME time,        */
/*    KHE_DRS_EXPR *res_e)                                                   */
/*                                                                           */
/*  If a BUSY_TIME, FREE_TIME, or WORK_TIME expression exists with the       */
/*  given tag, dom_test, and time, then return true and set *res_e to it.    */
/*  Otherwise return false with *res_e set to NULL.                          */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDrsHasTimeExpr(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME time,
  KHE_DRS_EXPR *res_e)
{
  int i;  KHE_DRS_DAY day;  KHE_DRS_EXPR e;
  KHE_DRS_EXPR_BUSY_TIME ebt;
  KHE_DRS_EXPR_FREE_TIME eft;
  KHE_DRS_EXPR_WORK_TIME ewt;
  i = KheFrameTimeIndex(drs->days_frame, time);
  day = HaArray(drs->days, i);
  HaArrayForEach(day->leaf_today, e, i)
    if( e->tag == tag && KheDrsDomTestEqual(e->dom_test, dom_test) )
    {
      switch( e->tag )
      {
	case KHE_DRS_EXPR_BUSY_TIME_TAG:

	  ebt = (KHE_DRS_EXPR_BUSY_TIME) e;
	  if( ebt->time == time )
	    return *res_e = e, true;
	  break;

	case KHE_DRS_EXPR_FREE_TIME_TAG:

	  eft = (KHE_DRS_EXPR_FREE_TIME) e;
	  if( eft->time == time )
	    return *res_e = e, true;
	  break;

	case KHE_DRS_EXPR_WORK_TIME_TAG:

	  ewt = (KHE_DRS_EXPR_WORK_TIME) e;
	  if( ewt->time == time )
	    return *res_e = e, true;
	  break;

	default:

	  HnAbort("KheDrsHasTimeExpr: invalid tag (%d)", e->tag);
      }
    }
  return *res_e = NULL, false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsResourceHasDayExpr(KHE_DRS_RESOURCE dr,                       */
/*    KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME_GROUP tg,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e)                  */
/*                                                                           */
/*  If dr has a BUSY_DAY, FREE_DAY, or WORK_DAY expression with the given    */
/*  tag, dom_test, and time group, then return true and set *res_e to it.    */
/*  Otherwise return false with *res_e set to NULL.                          */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsResourceHasDayExpr(KHE_DRS_RESOURCE dr,
  KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME_GROUP tg,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e)
{
  int i;  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_DRS_EXPR e;
  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_DRS_EXPR_FREE_DAY efd;
  KHE_DRS_EXPR_WORK_DAY ewd;
  if( KheTimeGroupTimeCount(tg) > 0 )
  {
    i = KheFrameTimeIndex(drs->days_frame, KheTimeGroupTime(tg, 0));
    drd = HaArray(dr->days, i);
    /* ***
    HnAssert(KheTimeGroupEqual(day->time_group, tg),
      "KheDrsHasDayExpr internal error");
    *** */
    HaArrayForEach(drd->leaf_today, e, i)
      if( e->tag == tag && KheDrsDomTestEqual(e->dom_test, dom_test) )
      {
	switch( e->tag )
	{
	  case KHE_DRS_EXPR_BUSY_DAY_TAG:

	    ebd = (KHE_DRS_EXPR_BUSY_DAY) e;
	    if( KheTimeGroupEqual(ebd->time_group, tg) )
	      return *res_e = e, true;
	    break;

	  case KHE_DRS_EXPR_FREE_DAY_TAG:

	    efd = (KHE_DRS_EXPR_FREE_DAY) e;
	    if( KheTimeGroupEqual(efd->time_group, tg) )
	      return *res_e = e, true;
	    break;

	  case KHE_DRS_EXPR_WORK_DAY_TAG:

	    ewd = (KHE_DRS_EXPR_WORK_DAY) e;
	    if( KheTimeGroupEqual(ewd->time_group, tg) )
	      return *res_e = e, true;
	    break;

	  default:

	    HnAbort("KheDrsResourceHasDayExpr: invalid tag (%d)", e->tag);
	}
      }
  }
  return *res_e = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsHasDayExpr(KHE_DYNAMIC_RESOURCE_SOLVER drs,                   */
/*    KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME_GROUP tg,    */
/*    KHE_DRS_EXPR *res_e)                                                   */
/*                                                                           */
/*  If a BUSY_DAY, FREE_DAY, or WORK_DAY expression exists with the given    */
/*  tag, dom_test, and time group, then return true and set *res_e to it.    */
/*  Otherwise return false with *res_e set to NULL.                          */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDrsHasDayExpr(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME_GROUP tg,
  KHE_DRS_EXPR *res_e)
{
  int i;  KHE_DRS_DAY day;  KHE_DRS_EXPR e;
  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_DRS_EXPR_FREE_DAY efd;
  KHE_DRS_EXPR_WORK_DAY ewd;
  if( KheTimeGroupTimeCount(tg) > 0 )
  {
    i = KheFrameTimeIndex(drs->days_frame, KheTimeGroupTime(tg, 0));
    day = HaArray(drs->days, i);
    HnAssert(KheTimeGroupEqual(day->time_group, tg),
      "KheDrsHasDayExpr internal error");
    HaArrayForEach(day->leaf_today, e, i)
      if( e->tag == tag && KheDrsDomTestEqual(e->dom_test, dom_test) )
      {
	switch( e->tag )
	{
	  case KHE_DRS_EXPR_BUSY_DAY_TAG:

	    ebd = (KHE_DRS_EXPR_BUSY_DAY) e;
	    if( KheTimeGroupEqual(ebd->time_group, tg) )
	      return *res_e = e, true;
	    break;

	  case KHE_DRS_EXPR_FREE_DAY_TAG:

	    efd = (KHE_DRS_EXPR_FREE_DAY) e;
	    if( KheTimeGroupEqual(efd->time_group, tg) )
	      return *res_e = e, true;
	    break;

	  case KHE_DRS_EXPR_WORK_DAY_TAG:

	    ewd = (KHE_DRS_EXPR_WORK_DAY) e;
	    if( KheTimeGroupEqual(ewd->time_group, tg) )
	      return *res_e = e, true;
	    break;

	  default:

	    HnAbort("KheDrsHasDayExpr: invalid tag (%d)", e->tag);
	}
      }
  }
  return *res_e = NULL, false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddTimeExpr(KHE_DRS_RESOURCE dr,                      */
/*    KHE_TIME time, KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR e)        */
/*                                                                           */
/*  Add time expression e whose time is time to drs.                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceAddTimeExpr(KHE_DRS_RESOURCE dr,
  KHE_TIME time, KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR e)
{
  int i;  KHE_DRS_RESOURCE_ON_DAY drd;
  i = KheFrameTimeIndex(drs->days_frame, time);
  drd = HaArray(dr->days, i);
  HaArrayAddLast(drd->leaf_today, e);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddDayExpr(KHE_DRS_RESOURCE dr,                       */
/*    KHE_TIME_GROUP tg, KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR e)    */
/*                                                                           */
/*  Add day expression e whose time group is tg to dr.                       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceAddDayExpr(KHE_DRS_RESOURCE dr,
  KHE_TIME_GROUP tg, KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR e)
{
  int i;  KHE_DRS_RESOURCE_ON_DAY drd;
  HnAssert(KheTimeGroupTimeCount(tg) > 0,
    "KheDrsResourceAddDayExpr internal error 1");
  i = KheFrameTimeIndex(drs->days_frame, KheTimeGroupTime(tg, 0));
  drd = HaArray(dr->days, i);
  /* ***
  HnAssert(KheTimeGroupEqual(drd->time_group, tg),
    "KheDrsAddDayExpr internal error 2");
  *** */
  HaArrayAddLast(drd->leaf_today, e);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE KheDrsResourceMake(KHE_RESOURCE r,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new resource object with these attributes.                        */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_RESOURCE KheDrsResourceMake(KHE_RESOURCE r,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_RESOURCE res;  int i, day_count;  KHE_DRS_DOM_TEST eq_dom_test;
  KHE_MONITOR m;  KHE_DRS_EXPR e;  KHE_DRS_RESOURCE_ON_DAY drd;

  /* build the object, including one resource on day object for each day */
  HaMake(res, drs->arena);
  res->resource = r;
  HaArrayInit(res->exprs, drs->arena);
  HaArrayInit(res->days, drs->arena);
  day_count = KheFrameTimeGroupCount(drs->days_frame);
  for( i = 0;  i < day_count;  i++ )
    HaArrayAddLast(res->days, KheDrsResourceOnDayMake(drs));

  /* add expressions to dr based on the resource's monitors */
  KheSolnResourceMonitorSort(drs->soln, r, &KheMonitorLargestLimitFirstCmp);
  for( i = 0;  i < KheSolnResourceMonitorCount(drs->soln, r);  i++ )
  {
    m = KheSolnResourceMonitor(drs->soln, r, i);
    if( DEBUG13 )
    {
      fprintf(stderr, "    limit %d: ", KheMonitorLimit(m));
      KheMonitorDebug(m, 2, 0, stderr);
    }
    KheDrsResourceAddMonitor(res, m, drs);
  }

  /* initialize tricky fields of the expressions */
  HaArrayForEach(res->exprs, e, i)
  {
    KheDrsExprSetDayRange(e, drs);
    KheDrsExprSetSigIndexes(e, res, drs);
    KheDrsExprAddExpressionsToDays(e, res);
  }

  /* add final dominance tests for resource end point */
  eq_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_EQ);
  HaArrayForEach(res->days, drd, i)
    KheDrsResourceOnDayAddDomTest(drd, eq_dom_test);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceMakeAndAdd(KHE_RESOURCE r,                            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a drs resource object corresponding to r (assumed to have the       */
/*  correct type) and add it to drs.                                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceMakeAndAdd(KHE_RESOURCE r,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->all_resources, KheDrsResourceMake(r, drs));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddSolveInterval(KHE_DRS_RESOURCE dr,                 */
/*    KHE_DRS_DAY_RANGE si, KHE_DYNAMIC_RESOURCE_SOLVER drs)                 */
/*                                                                           */
/*  Add si to the solve intervals of dr.  Also make sure that dr is on       */
/*  the selected list, if it isn't already.                                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsResourceAddSolveInterval(KHE_DRS_RESOURCE dr,
  KHE_DRS_DAY_RANGE si, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ** add dr to the selected list if it isn't there already **
  if( HaArrayCount(dr->solve_day_ranges) == 0 )
    HaArrayAddLast(drs->solve_resources, dr);

  ** add si to the day intervals **
  HnAssert(si.first_day_index <= si.last_day_index,
    "KheDrsResourceAddSolveInterval internal error:  empty interval");
  KheDrsDayRangeMerge(&dr->solve_day_ranges, si);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceInitFixed(KHE_DRS_RESOURCE dr, KHE_DRS_DAY_RANGE si,  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Initialize fixed interval si.                                            */
/*                                                                           */
/*  Throughout this interval, the resource must preserve its existing        */
/*  timetable.  This is indicated by placing a non-NULL drs task into        */
/*  each position.  If the resource is already assigned a task, that         */
/*  is the task placed there.  If the resource is free, a free day task      */
/*  for the day is used.                                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceInitFixed(KHE_DRS_RESOURCE dr, KHE_DRS_DAY_RANGE si,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int di;  KHE_DRS_TASK dt;  KHE_DRS_DAY day;  KHE_DRS_RESOURCE_ON_DAY drd;
  for( di = si.first_day_index;  di <= si.last_day_index;  di++ )
  {
    drd = HaArray(dr->days, di);
    dt = drd->assigned_task;
    if( dt == NULL )
    {
      day = HaArray(drs->days, di);
      dt = day->free_task;
    }
    drd->solve_task = dt;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceInitFree(KHE_DRS_RESOURCE dr, KHE_DRS_DAY_RANGE si,   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Initialize free interval si.                                             */
/*                                                                           */
/*  Throughout this interval, the resource is free to take on a different    */
/*  timetable.  This is indicated by placing a NULL drs task into each       */
/*  position.  However, if the resource is assigned a task on some day       */
/*  and that task extends beyond si, the resource must preserve that task,   */
/*  and that is indicated by placing it into solve_tasks.                    */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete
static void KheDrsTaskAddToSolveTasks(KHE_DRS_TASK dt);

static void KheDrsResourceInitFree(KHE_DRS_RESOURCE dr, KHE_DRS_DAY_RANGE si,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int di;  KHE_DRS_TASK dt;  KHE_DRS_RESOURCE_ON_DAY drd;
  for( di = si.first_day_index;  di <= si.last_day_index;  di++ )
  {
    drd = HaArray(dr->days, di);
    dt = drd->assigned_task;
    if( dt != NULL && KheDrsDayRangeSubset(dt->day_interval, si) )
    {
      ** this task is assigned but it can be reassigned **
      KheDrsTaskAddToSolveTasks(dt);
      dt = NULL;
    }
    drd->solve_task = dt;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceInitForSolving(KHE_DRS_RESOURCE dr,                   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Assuming that dr is one of the resources selected for solving,           */
/*  initialize it for solving.                                               */
/*                                                                           */
/*  There will always be at least one free interval, although as it happens  */
/*  this code works correctly even if there isn't:  in that case, it makes   */
/*  all days fixed.                                                          */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete
static void KheDrsResourceInitForSolving(KHE_DRS_RESOURCE dr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int prev_index, i;  KHE_DRS_DAY_RANGE free_si, fixed_si;

  ** initialize the solve_tasks array for solving **
  prev_index = 0;
  HaArrayForEach(dr->solve_day_ranges, free_si, i)
  {
    fixed_si = KheDrsDayRangeMake(prev_index, free_si.first_day_index - 1);
    KheDrsResourceInitFixed(dr, fixed_si, drs);
    KheDrsResourceInitFree(dr, free_si, drs);
    prev_index = free_si.last_day_index + 1;
  }
  fixed_si = KheDrsDayRangeMake(prev_index, HaArrayCount(dr->days));
  KheDrsResourceInitFixed(dr, fixed_si, drs);

  ** add expressions to days **
  ** ***
  HaArrayForEach(dr->exprs, e, i)
    KheDrsExprAddExpressionsToDays(e, dr);
  *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceEndSolving(KHE_DRS_RESOURCE dr)                       */
/*                                                                           */
/*  End the selection of dr for solving.                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceEndSolving(KHE_DRS_RESOURCE dr)
{
  /* clear out the solve intervals */
  /* HaArrayClear(dr->solve_day_ranges); */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddTask(KHE_DRS_RESOURCE dr, KHE_DRS_TASK dt)         */
/*                                                                           */
/*  Assuming that dt is bound to dr's resource, add dt to dr.                */
/*                                                                           */
/*****************************************************************************/
static int KheDrsTaskFirstDayIndex(KHE_DRS_TASK dt);
static int KheDrsTaskLastDayIndex(KHE_DRS_TASK dt);

static void KheDrsResourceAddTask(KHE_DRS_RESOURCE dr, KHE_DRS_TASK dt)
{
  int first_day_index, last_day_index, i;  KHE_DRS_RESOURCE_ON_DAY drd;
  first_day_index = KheDrsTaskFirstDayIndex(dt);
  last_day_index = KheDrsTaskLastDayIndex(dt);
  for( i = first_day_index;  i <= last_day_index;  i++ )
  {
    drd = HaArray(dr->days, i);
    drd->assigned_task = dt;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceDebug(KHE_DRS_RESOURCE dr, int verbosity,             */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of dr onto fp with the given verbosity and indent.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceDebug(KHE_DRS_RESOURCE dr, int verbosity,
  int indent, FILE *fp)
{
  KHE_DRS_RESOURCE_ON_DAY drd;  int i;
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "[ %s", KheResourceId(dr->resource));
  /* KheDayRangesDebug(&dr->solve_day_ranges, verbosity, -1, fp); */
  if( verbosity > 1 && indent >= 0 )
  {
    fprintf(fp, "\n");
    HaArrayForEach(dr->days, drd, i)
      KheDrsResourceOnDayDebug(drd, verbosity, indent + 2, fp);
  }
  fprintf(fp, "]");
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_TASK"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK KheDrsTaskGet(KHE_DYNAMIC_RESOURCE_SOLVER drs)              */
/*                                                                           */
/*  Get a new drs task object.  Only the times_by_day and tasks_by day       */
/*  fields are initialized.                                                  */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_TASK KheDrsTaskGet(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK res;
  if( HaArrayCount(drs->free_tasks) > 0 )
    res = HaArrayLastAndDelete(drs->free_tasks);
  else
    HaMake(res, drs->arena);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK KheDrsTaskMake(KHE_TASK task, KHE_TASK_CLASS orig_tc,       */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new drs task with these attributes.                               */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_TASK KheDrsTaskMake(KHE_TASK task, KHE_TASK_CLASS orig_tc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK res;
  res = KheDrsTaskGet(drs);
  res->task = task;
  res->orig_task_class = orig_tc;
  return res;

  /* ***
  KHE_DRS_TASK res;  KHE_RESOURCE r;  int first_day_index, last_day_index;
  res = KheDrsTaskGet(drs);
  res->task = task;
  res->task_group = NULL;	** will be filled in later **
  res->daily_schedule = KheTaskFinderTaskDailySchedule(drs->task_finder, task);
  first_day_index = KheDailyScheduleFirstDayIndex(res->daily_schedule);
  last_day_index = KheDailyScheduleLastDayIndex(res->daily_schedule);
  res->day_interval = KheDrsDayRangeMake(first_day_index, last_day_index);
  res->cost_reduction = KheTaskAssignmentCostReduction(task);
  res->domain_count = KheResourceGroupResourceCount(KheTaskDomain(task));
  ** res->fixed = false; **	** placeholder value **
  r = KheTaskAsstResource(task);
  res->already_assigned = (r != NULL);
  return res;
  *** */
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsTaskFirstDayIndex(KHE_DRS_TASK dt)                             */
/*                                                                           */
/*  Return the index of the first day when dt is running.                    */
/*                                                                           */
/*****************************************************************************/

static int KheDrsTaskFirstDayIndex(KHE_DRS_TASK dt)
{
  /* *** replacement for this still to do
  return KheDailyScheduleFirstDayIndex(dt->daily_schedule);
  *** */
  return -1;  /* placeholder */
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsTaskLastDayIndex(KHE_DRS_TASK dt)                              */
/*                                                                           */
/*  Return the index of the last day when dt is running.                     */
/*                                                                           */
/*****************************************************************************/

static int KheDrsTaskLastDayIndex(KHE_DRS_TASK dt)
{
  /* *** replacement for this still to do
  return KheDailyScheduleLastDayIndex(dt->daily_schedule);
  *** */
  return -1;  /* placeholder */
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_TIME KheDrsTaskFirstTime(KHE_DRS_TASK dt)                            */
/*                                                                           */
/*  Return the first time of dt.                                             */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_TIME KheDrsTaskFirstTime(KHE_DRS_TASK dt)
{
  int index;
  index = KheDailyScheduleFirstDayIndex(dt->daily_schedule);
  return KheDailyScheduleTime(dt->daily_schedule, index);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_TIME KheDrsTaskLastTime(KHE_DRS_TASK dt)                             */
/*                                                                           */
/*  Return the last time of dt.                                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_TIME KheDrsTaskLastTime(KHE_DRS_TASK dt)
{
  int index;
  index = KheDailyScheduleLastDayIndex(dt->daily_schedule);
  return KheDailyScheduleTime(dt->daily_schedule, index);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTaskIsBound(KHE_DRS_TASK dt, KHE_RESOURCE *r)                 */
/*                                                                           */
/*  If dt is bound, return true and set *r to the associated resource.       */
/*  Otherwise return false and set *r to NULL.                               */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsTaskIsBound(KHE_DRS_TASK dt, KHE_RESOURCE *r)
{
  /* return false if dt represents free time */
  if( dt->task == NULL )
    return false;

  /* return true if assigned */
  *r = KheTaskAsstResource(dt->task);
  if( *r != NULL )
    return true;

  /* return true if preassigned */
  return KheTaskIsPreassigned(dt->task, r);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE KheDrsResource(KHE_RESOURCE r,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Assuming that r has the correct resource type, return its corresponding  */
/*  drs resource.                                                            */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_RESOURCE KheDrsResource(KHE_RESOURCE r,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  return HaArray(drs->all_resources, KheResourceResourceTypeIndex(r));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskMakeAndAdd(KHE_TASK task,                                 */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new drs task object for task, which is a proper root task of the  */
/*  wanted type, and add it to the solver generally.                         */
/*                                                                           */
/*  Implementation note.  The drs task is added to one task group on its     */
/*  first day, and if assigned (or bound generally) it is added to the drs   */
/*  resource it is assigned or bound to.                                     */
/*                                                                           */
/*****************************************************************************/
/* ***
static KHE_DRS_TASK_GROUP KheDrsDayAddTask(KHE_DRS_DAY day, KHE_DRS_TASK dt);

static void KheDrsTaskMakeAndAdd(KHE_TASK task,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK dt;  KHE_RESOURCE r;  KHE_DRS_TASK_GROUP dtg;  KHE_DRS_DAY day;
  int first_day_index;

  ** make the drs task object **
  dt = KheDrsTaskMake(task, drs);

  ** add it to the day object of the first day that it is running **
  first_day_index = KheDailyScheduleFirstDayIndex(dt->daily_schedule);
  day = HaArray(drs->days, first_day_index);
  dtg = KheDrsDayAddTask(day, dt);

  ** record which task group it got added to **
  dt->task_group = dtg;

  ** if task is assigned or preassigned, add it to the associated resource **
  if( KheDrsTaskIsBound(dt, &r) )
    KheDrsResourceAddTask(KheDrsResource(r, drs), dt);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK KheDrsTaskMakeFree(int day_index,                           */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new drs task which represents doing nothing on the day with       */
/*  this day_index.                                                          */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_TASK KheDrsTaskMakeFree(int day_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK res;
  res = KheDrsTaskGet(drs);
  res->task = NULL;
  res->daily_schedule = KheTaskFinderNullDailySchedule(drs->task_finder,
    day_index, day_index);
  res->cost_reduction = 0;
  res->domain_count = 0;
  res->already_assigned = false;
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskAddToSolveTasks(KHE_DRS_TASK dt)                          */
/*                                                                           */
/*  Make dt a solve task, by adding it to its task group's solve_tasks list. */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTaskAddToSolveTasks(KHE_DRS_TASK dt)
{
  ** sorting by decreasing cost still to do **
  HaArrayAddLast(dt->task_group->solve_tasks, dt);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskMakeUnAvail(KHE_DRS_TASK dt)                              */
/*                                                                           */
/*  Make dt unavailable.                                                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTaskMakeUnAvail(KHE_DRS_TASK dt)
{
  int pos;
  if( HaArrayContains(dt->task_group->solve_tasks, dt, &pos) )
    HaArrayDeleteAndShift(dt->task_group->solve_tasks, pos);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskDelete(KHE_DRS_TASK dt, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Delete dt.                                                               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskDelete(KHE_DRS_TASK dt, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* KheDailyScheduleDelete(dt->daily_schedule); */
  HaArrayAddLast(drs->free_tasks, dt);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTaskBetterThan(KHE_DRS_TASK dt1, KHE_DRS_TASK dt2)            */
/*                                                                           */
/*  Return true if dt1 is a better choice than dt2, because it has a         */
/*  larger cost reduction, or equal cost reduction and a smaller domain.     */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDrsTaskBetterThan(KHE_DRS_TASK dt1, KHE_DRS_TASK dt2)
{
  return dt1->cost_reduction > dt2->cost_reduction ||
    (dt1->cost_reduction == dt2->cost_reduction &&
     dt1->domain_count < dt2->domain_count);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskDebug(KHE_DRS_TASK dt, int verbosity,                     */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of dt onto fp with the given verbosity and indent.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskDebug(KHE_DRS_TASK dt, int verbosity,
  int indent, FILE *fp)
{
  /* ***
  KHE_TIME first_time, last_time;  int first_day_index, last_day_index;
  KHE_TASK_FINDER tf;  KHE_TIME_GROUP tg1, tg2;  KHE_FRAME frame;
  *** */
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  /* ***
  if( dt->task != NULL )
  {
    first_time = KheDrsTaskFirstTime(dt);
    last_time = KheDrsTaskLastTime(dt);
    KheTaskDebug(dt->task, verbosity, -1, fp);
    if( first_time != last_time )
      fprintf(fp, " (%s-%s", KheTimeId(first_time), KheTimeId(last_time));
    else
      fprintf(fp, " (%s", KheTimeId(first_time));
    fprintf(fp, ", cost %.5f, domain %d)",
      KheCostShow(dt->cost_reduction), dt->domain_count);
  }
  else
  {
    tf = KheDailyScheduleTaskFinder(dt->daily_schedule);
    frame = KheTaskFinderFrame(tf);
    first_day_index = KheDailyScheduleFirstDayIndex(dt->daily_schedule);
    last_day_index = KheDailyScheduleLastDayIndex(dt->daily_schedule);
    tg1 = KheFrameTimeGroup(frame, first_day_index);
    tg2 = KheFrameTimeGroup(frame, last_day_index);
    if( tg1 != tg2 )
      fprintf(fp, "Free (%s-%s)", KheTimeGroupId(tg1), KheTimeGroupId(tg2));
    else
      fprintf(fp, "Free (%s)", KheTimeGroupId(tg1));
  }
  *** */
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_TASK_CLASS"                                           */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK_CLASS KheDrsTaskClassMake(KHE_TASK_CLASS orig_tc,           */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new task class object based on orig_tc.                           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_TASK_CLASS KheDrsTaskClassMake(KHE_TASK_CLASS orig_tc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_CLASS res;

  /* get or make a drs task class object */
  if( HaArrayCount(drs->free_task_classes) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_task_classes);
    HaArrayClear(res->tasks);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->tasks, drs->arena);
  }
  res->orig_task_class = orig_tc;
  /* adding open tasks from orig_tc still to do */
  res->tmp_used = 0;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassDelete(KHE_DRS_TASK_CLASS drc,                       */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete drc by adding it to drs's free list.                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassDelete(KHE_DRS_TASK_CLASS drc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK dt;  int i;
  HaArrayForEach(drc->tasks, dt, i)
    KheDrsTaskDelete(dt, drs);
  HaArrayAddLast(drs->free_task_classes, drc);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTaskGroupAccepts(KHE_DRS_TASK_GROUP sg, KHE_DRS_TASK dt)      */
/*                                                                           */
/*  Return true if sg accepts dt.  The start times, end times, and domains   */
/*  must be equal.                                                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDrsTaskGroupAccepts(KHE_DRS_TASK_GROUP sg, KHE_DRS_TASK dt)
{
  KHE_DRS_TASK sg_dt;  int di1, di2;
  KHE_DAILY_SCHEDULE sched1, sched2;

  ** get the daily schedules of the two tasks, and their first and last days **
  sg_dt = HaArrayFirst(sg->all_tasks);
  sched1 = sg_dt->daily_schedule;
  sched2 = dt->daily_schedule;

  ** check the first times **
  di1 = KheDailyScheduleFirstDayIndex(sched1);
  di2 = KheDailyScheduleFirstDayIndex(sched2);
  if( KheDailyScheduleTime(sched1, di1) != KheDailyScheduleTime(sched2, di2) )
    return false;

  ** check the last times **
  di1 = KheDailyScheduleLastDayIndex(sched1);
  di2 = KheDailyScheduleLastDayIndex(sched2);
  if( KheDailyScheduleTime(sched1, di1) != KheDailyScheduleTime(sched2, di2) )
    return false;

  ** check the domains **
  if( !KheResourceGroupEqual(KheTaskDomain(dt->task), 
	KheTaskDomain(sg_dt->task)) )
      return false;

  ** all good **
  return true;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskGroupAddTask(KHE_DRS_TASK_GROUP sg, KHE_DRS_TASK dt)      */
/*                                                                           */
/*  Add dt to sg, largest cost reduction first.                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTaskGroupAddTask(KHE_DRS_TASK_GROUP sg, KHE_DRS_TASK dt)
{
  KHE_DRS_TASK dt2;  int i;
  HaArrayAddLast(sg->all_tasks, dt);
  for( i = HaArrayCount(sg->all_tasks) - 2;  i >= 0;  i-- )
  {
    dt2 = HaArray(sg->all_tasks, i);
    if( dt2->cost_reduction <= dt->cost_reduction )
      HaArrayPut(sg->all_tasks, i + 1, dt2);
  }
  HaArrayPut(sg->all_tasks, i + 1, dt);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskGroupClearAvail(KHE_DRS_TASK_GROUP dtg)                   */
/*                                                                           */
/*  Clear the available tasks of dtg.                                        */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTaskGroupClearAvail(KHE_DRS_TASK_GROUP dtg)
{
  HaArrayClear(dtg->solve_tasks);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskGroupMakeUnboundTasksAvail(KHE_DRS_TASK_GROUP dtg)        */
/*                                                                           */
/*  Make the unbound tasks of dtg available.                                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTaskGroupMakeUnboundTasksAvail(KHE_DRS_TASK_GROUP dtg)
{
  KHE_DRS_TASK dt;  int i;  KHE_RESOURCE r;
  HaArrayForEach(dtg->all_tasks, dt, i)
    if( !KheDrsTaskIsBound(dt, &r) )
      HaArrayAddLast(dtg->solve_tasks, dt);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassDebug(KHE_DRS_TASK_CLASS dtc, int verbosity,         */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of dtc onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassDebug(KHE_DRS_TASK_CLASS dtc, int verbosity,
  int indent, FILE *fp)
{
  /* ***
  KHE_DRS_TASK dt;  KHE_DAILY_SCHEDULE sched;  KHE_TIME ft, lt;
  KHE_RESOURCE_GROUP rg;
  dt = HaArrayFirst(sg->all_tasks);
  sched = dt->daily_schedule;
  ft = KheDailyScheduleTime(sched, KheDailyScheduleFirstDayIndex(sched));
  lt = KheDailyScheduleTime(sched, KheDailyScheduleLastDayIndex(sched));
  rg = KheTaskDomain(dt->task);
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "[ %d tasks: ", HaArrayCount(sg->all_tasks));
  if( ft != lt )
    fprintf(fp, "%s-%s", KheTimeId(ft), KheTimeId(lt));
  else
    fprintf(fp, "%s", KheTimeId(ft));
  fprintf(fp, " %s ]", KheResourceGroupId(rg));
  if( indent >= 0 )
    fprintf(fp, "\n");
  *** */
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EDGE"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EDGE KheDrsEdgeMake(KHE_DYNAMIC_RESOURCE_SOLVER drs)             */
/*                                                                           */
/*  Make a new, empty edge object.                                           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EDGE KheDrsEdgeMake(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EDGE res;
  if( HaArrayCount(drs->free_edges) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_edges);
    HaArrayClear(res->tasks);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->tasks, drs->arena);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EDGE KheDrsEdgeCopy(KHE_DRS_EDGE de,                             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a copy of edge.                                                     */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EDGE KheDrsEdgeCopy(KHE_DRS_EDGE de,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EDGE res;  int i;
  res = KheDrsEdgeMake(drs);
  HaArrayAppend(res->tasks, de->tasks, i);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsEdgeDelete(KHE_DRS_EDGE de, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Delete de; add it to the free list in drs.                               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsEdgeDelete(KHE_DRS_EDGE de, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_edges, de);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsEdgeDebug(KHE_DRS_EDGE de, KHE_DYNAMIC_RESOURCE_SOLVER drs,   */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of de onto fp with the given verbosity and indent.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsEdgeDebug(KHE_DRS_EDGE de, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  KHE_DRS_TASK dt;  int i;  KHE_DRS_RESOURCE dr;
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "[ edge ");
  HaArrayForEach(de->tasks, dt, i)
  {
    if( i > 0 )
      fprintf(fp, ", ");
    dr = HaArray(drs->solve_resources, i);
    fprintf(fp, "%s:", KheResourceId(dr->resource));
    KheDrsTaskDebug(dt, 1, -1, fp);
  }
  fprintf(fp, " ]");
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_TIME"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TIME KheDrsTimeMake(KHE_TIME t, HA_ARENA a)                      */
/*                                                                           */
/*  Make and return a new drs time object representing t.                    */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_TIME KheDrsTimeMake(KHE_TIME t, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TIME res;
  HaMake(res, drs->arena);
  res->time = t;
  ** HaArrayInit(res->leaf_exprs, drs->arena); **
  HaArrayInit(res->best_tasks, drs->arena);
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTimeClearTasks(KHE_DRS_TIME st,                               */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Clear the tasks of st.                                                   */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTimeClearTasks(KHE_DRS_TIME st,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i;  KHE_DRS_TASK dt;
  HaArrayForEach(st->best_tasks, dt, i)
    KheDrsTaskDelete(dt, drs);
  HaArrayClear(st->best_tasks);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTimeContainsLastTime(KHE_DRS_TIME stime,                      */
/*    KHE_TIME last_time, KHE_DRS_TASK *dt, int *pos)                        */
/*                                                                           */
/*  If stime contains a task with the given last_time, return true with      */
/*  *dt set to that task and *pos set to its position.  Otherwise return     */
/*  false with *dt set to NULL and *pos set to -1.                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDrsTimeContainsLastTime(KHE_DRS_TIME stime,
  KHE_TIME last_time, KHE_DRS_TASK *dt, int *pos)
{
  KHE_DRS_TASK dt2;  int i;
  HaArrayForEach(stime->best_tasks, dt2, i)
    if( KheDrsTaskLastTime(dt2) == last_time )
      return *dt = dt2, *pos = i, true;
  return *dt = NULL, *pos = -1, false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTimeAddTask(KHE_DRS_TIME st, KHE_DRS_TASK dt,                 */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add dt to st.  More precisely, if it's better than what st already       */
/*  has, then add it, otherwise delete it.                                   */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTimeAddTask(KHE_DRS_TIME st, KHE_DRS_TASK dt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK dt2;  int pos;
  if( !KheDrsTimeContainsLastTime(st, KheDrsTaskLastTime(dt), &dt2, &pos) )
  {
    ** st has no task with the same last time as dt, so add dt to st **
    HaArrayAddLast(st->best_tasks, dt);
  }
  else if( KheDrsTaskBetterThan(dt, dt2) )
  {
    ** dt is better than the similar task dt2, so replace dt2 by dt **
    KheDrsTaskDelete(dt2, drs);
    HaArrayPut(st->best_tasks, pos, dt);
  }
  else
  {
    ** dt is worse than the similar dt2, so delete dt **
    KheDrsTaskDelete(dt, drs);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTimeDebug(KHE_DRS_TIME st, KHE_DYNAMIC_RESOURCE_SOLVER drs,   */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of st onto fp with the given verbosity and indent.           */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTimeDebug(KHE_DRS_TIME st, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int i;  ** KHE_DRS_EXPR e; **  KHE_DRS_TASK dt;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ Time %s:\n", indent, "", KheTimeId(st->time));
    ** ***
    HaArrayForEach(st->leaf_exprs, e, i)
      KheDrsExprDebug(e, drs, verbosity, indent + 2, fp);
    *** **
    HaArrayForEach(st->best_tasks, dt, i)
      KheDrsTaskDebug(dt, verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
    fprintf(fp, "[ Time %s (%d best_tasks) ]", KheTimeId(st->time),
      HaArrayCount(st->best_tasks));
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_NODE"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_NODE KheDrsNodeGet(KHE_DYNAMIC_RESOURCE_SOLVER drs)              */
/*                                                                           */
/*  Get a new node object.  Only the signature field is initialized.         */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_NODE KheDrsNodeGet(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_NODE res;
  if( HaArrayCount(drs->free_nodes) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_nodes);
    HaArrayClear(res->signature);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->signature, drs->arena);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_NODE KheDrsNodeMakeRoot(KHE_DYNAMIC_RESOURCE_SOLVER drs)         */
/*                                                                           */
/*  Make the root node that the search grows from.  It does not end on       */
/*  any day.                                                                 */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_NODE KheDrsNodeMakeRoot(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_NODE res;
  res = KheDrsNodeGet(drs);
  res->prev_node = NULL;
  res->prev_edge = NULL;
  res->day = NULL;
  res->best_of = 1;
  res->cost = 0;
  HaArrayAddLast(res->signature, 0);  /* number of assignments */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_NODE KheDrsNodeMakeNonRoot(KHE_DRS_NODE prev_node,               */
/*    KHE_DRS_EDGE prev_edge, KHE_DRS_DAY day,                               */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new non-root node with these attributes.                          */
/*                                                                           */
/*****************************************************************************/
static void KheDrsNodeDebug(KHE_DRS_NODE node,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp);

static KHE_DRS_NODE KheDrsNodeMakeNonRoot(KHE_DRS_NODE prev_node,
  KHE_DRS_EDGE prev_edge, KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_NODE res;  int asst_count, i, j;  KHE_TIME time;
  KHE_DRS_EXPR e;  KHE_TASK task;  /* KHE_DRS_TASK dt; */
  KHE_DRS_RESOURCE dr;  KHE_DRS_RESOURCE_ON_DAY drd, prev_drd;

  /* make the basic object */
  HnAssert(prev_edge != NULL, "KheDrsNodeMake internal error (prev_edge)");
  if( DEBUG4 )
    fprintf(stderr, "[ KheDrsNodeMake(prev_node, prev_edge, %s, drs)\n",
      KheTimeGroupId(day->time_group));
  res = KheDrsNodeGet(drs);
  res->prev_node = prev_node;
  res->prev_edge = prev_edge;
  res->day = day;
  res->best_of = 1;

  /* set the cost and signature */
  res->cost = prev_node->cost;
  asst_count = HaArrayFirst(prev_node->signature);
  HaArrayAddLast(res->signature, asst_count);
  HaArrayForEach(drs->solve_resources, dr, i)
  {
    /* *** replacement for this still to do
    dt = HaArray(prev_edge->tasks, i);
    task = KheDailyScheduleTask(dt->daily_schedule, day->day_index);
    time = KheDailyScheduleTime(dt->daily_schedule, day->day_index);
    *** */
    task = NULL;
    time = NULL;
    if( task != NULL )
      asst_count++;
    drd = HaArray(dr->days, day->day_index);
    prev_drd = HaArray(dr->days, day->day_index - 1);
    HaArrayForEach(drd->active_today, e, j)
      KheDrsExprDoDay(e, day->day_index, prev_node, prev_drd->sig_offset,
	task, time, res, drd->sig_offset, drs);
    /* *** replacement for this still to do
    HaArrayAddLast(res->signature,
      KheDailyScheduleLastDayIndex(dt->daily_schedule) - day->day_index);
    *** */
  }
  HaArrayPut(res->signature, 0, asst_count);
  /* *** stil l to do (must iterate over selected resources and get tasks 
  HaArrayForEach(day->active_today, e, i)
    KheDrsExprDoDay(e, day->day_index, prev_node, task, time, res, drs);
  *** */

  if( DEBUG4 )
  {
    KheDrsNodeDebug(res, drs, 1, 2, stderr);
    fprintf(stderr, "] KheDrsNodeMake returning\n");
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeDelete(KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  Delete node.                                                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodeDelete(KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_nodes, node);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsNodeSignatureFullHash(KHE_DRS_NODE node)                       */
/*                                                                           */
/*  Hash function for hashing the full signature of node.                    */
/*                                                                           */
/*****************************************************************************/

static int KheDrsNodeSignatureFullHash(KHE_DRS_NODE node)
{
  int res, val, i;
  res = 0;
  HaArrayForEach(node->signature, val, i)
    res = (res + val) * 3;
  return (res < 0 ? - res : res);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsNodeSignatureFullHashUntyped(void *p)                          */
/*                                                                           */
/*  Untyped version of KheDrsNodeSignatureHash.                              */
/*                                                                           */
/*****************************************************************************/

static int KheDrsNodeSignatureFullHashUntyped(void *p)
{
  return KheDrsNodeSignatureFullHash((KHE_DRS_NODE) p);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsNodeSignaturePartialHash(KHE_DRS_NODE node)                    */
/*                                                                           */
/*  Hash function for hashing the partial signature of node, that is, only   */
/*  the parts of it whose dominance tests are equality (or loose equality).  */
/*                                                                           */
/*****************************************************************************/

static int KheDrsNodeSignaturePartialHash(KHE_DRS_NODE node)
{
  int res, index, val, i;  KHE_DRS_DAY day;
  res = 0;
  day = node->day;
  HaArrayForEach(day->eq_dom_test_indexes, index, i)
  {
    val = HaArray(node->signature, index);
    res = (res + val) * 3;
  }
  if( res < 0 )
    res = - res;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsNodeSignatureFullEqual(KHE_DRS_NODE node1, KHE_DRS_NODE node2)*/
/*                                                                           */
/*  Return true if the signatures of node1 and node2 are equal.  The lengths */
/*  of their signatures must be equal.                                       */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsNodeSignatureFullEqual(KHE_DRS_NODE node1, KHE_DRS_NODE node2)
{
  int i;
  HnAssert(HaArrayCount(node1->signature) == HaArrayCount(node2->signature),
    "KheDrsNodeSignaturesEqual: signatures have different lengths");
  for( i = 0;  i < HaArrayCount(node1->signature);  i++ )
    if( HaArray(node1->signature, i) != HaArray(node2->signature, i) )
      return false;
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsNodeSignatureFullEqualUntyped(void *p1, void *p2)             */
/*                                                                           */
/*  Untyped version of KheDrsNodeSignaturesEqual.                            */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsNodeSignatureFullEqualUntyped(void *p1, void *p2)
{
  return KheDrsNodeSignatureFullEqual((KHE_DRS_NODE) p1,
    (KHE_DRS_NODE) p2);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsNodeSignaturePartialEqual(KHE_DRS_NODE node1,                 */
/*    KHE_DRS_NODE node2)                                                    */
/*                                                                           */
/*  Return true if the signatures of node1 and node2 are partially equal,    */
/*  that is, the entries whose dominance tests are equality (or loose        */
/*  equality) are equal.                                                     */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsNodeSignaturePartialEqual(KHE_DRS_NODE node1,
  KHE_DRS_NODE node2)
{
  int i, index;  KHE_DRS_DAY day;
  HnAssert(HaArrayCount(node1->signature) == HaArrayCount(node2->signature),
   "KheDrsNodeSignaturePartialEqual: signatures have different lengths");
  day = node1->day;
  HaArrayForEach(day->eq_dom_test_indexes, index, i)
    if( HaArray(node1->signature, index) != HaArray(node2->signature, index) )
      return false;
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeSignatureDebug(KHE_DRS_NODE node, FILE *fp)               */
/*                                                                           */
/*  Debug print of the signature of node onto fp (no verbosity or indent).   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodeSignatureDebug(KHE_DRS_NODE node, FILE *fp)
{
  int i, val;
  fprintf(fp, "[");
  HaArrayForEach(node->signature, val, i)
  {
    if( i > 0 )
      fprintf(fp, " ");
    fprintf(fp, "%d", val);
  }
  fprintf(fp, "]");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeSignatureDebugUntyped(void *p, FILE *fp)                  */
/*                                                                           */
/*  Untyped version of KheDrsNodeSignatureDebug.                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodeSignatureDebugUntyped(void *p, FILE *fp)
{
  KheDrsNodeSignatureDebug((KHE_DRS_NODE) p, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsNodeDominates(KHE_DRS_NODE node1, KHE_DRS_NODE node2,         */
/*    int start_depth)                                                       */
/*                                                                           */
/*  Return true if node1 dominates node2.  They come from the same day.      */
/*  Parameter start_depth says how far along the signature to start.         */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsNodeDominates(KHE_DRS_NODE node1, KHE_DRS_NODE node2,
  int start_depth)
{
  KHE_DRS_DOM_TEST dt;  int i, val1, val2, sig_len;  KHE_DRS_DAY day;
  HnAssert(node1->day == node2->day,
    "KheDrsNodeDominates internal error: node1->day = %s, node2->day = %s",
    node1->day != NULL ? KheTimeGroupId(node1->day->time_group) : "-",
    node2->day != NULL ? KheTimeGroupId(node2->day->time_group) : "-");
  if( node2->cost < node1->cost )
    return false;
  day = node1->day;
  sig_len = HaArrayCount(node1->signature);
  for( i = start_depth;  i < sig_len;  i++ )
  {
    dt = HaArray(day->dom_tests, i);
    val1 = HaArray(node1->signature, i);
    val2 = HaArray(node2->signature, i);
    if( !KheDrsDomTestDominates(dt, val1, val2) )
      return false;
  }
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeOverWrite(KHE_DRS_NODE dest_node, KHE_DRS_NODE src_node)  */
/*                                                                           */
/*  Overwrite src_node onto dest_node.  Their signatures are known to be     */
/*  equal, so those don't need to be touched.  Also the best_of field        */
/*  is not touched here; it gets updated separately.                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodeOverWrite(KHE_DRS_NODE dest_node, KHE_DRS_NODE src_node)
{
  dest_node->prev_node = src_node->prev_node;
  dest_node->prev_edge = src_node->prev_edge;
  dest_node->cost = src_node->cost;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsNodeIncreasingAsstCountCmp(const void *p1, void *p2)           */
/*                                                                           */
/*  Comparison function for sorting an array of nodes by increasing          */
/*  assignment count.                                                        */
/*                                                                           */
/*****************************************************************************/

static int KheDrsNodeIncreasingAsstCountCmp(const void *p1,
  const void *p2)
{
  KHE_DRS_NODE node1 = * (KHE_DRS_NODE *) p1;
  KHE_DRS_NODE node2 = * (KHE_DRS_NODE *) p2;
  return HaArrayFirst(node1->signature) - HaArrayFirst(node2->signature);
}


/*****************************************************************************/
/*                                                                           */
/*  void DoExtend(KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_NODE node,        */
/*    KHE_DRS_EDGE de)                                                       */
/*                                                                           */
/*  Extend node using de.                                                    */
/*                                                                           */
/*****************************************************************************/

static void DoExtend(KHE_DRS_NODE node, KHE_DRS_EDGE de,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* still to do ***
  KHE_DRS_NODE curr_node, last_node;
  int di, first_day_index, last_day_index;  ** KHE_DRS_DAY last_day; **
  ** KHE_TIME time;  KHE_DRS_TIME stime;  KHE_TASK task; **

  ** make one node for each day of dt, ending at last_node **
  last_node = node;
  first_day_index = KheDailyScheduleFirstDayIndex(dt->daily_schedule);
  last_day_index = KheDailyScheduleLastDayIndex(dt->daily_schedule);
  for( di = first_day_index;  di <= last_day_index;  di++ )
    last_node = KheDrsNodeMakeNonRoot(last_node, dt, HaArray(drs->days, di), drs);
  HnAssert(last_node != node, "DoExtend internal error");

  ** delete all nodes between node and last_node exclusive **
  ** NB field curr_node->prev is usable after calling KheNodeDelete **
  for( curr_node = last_node->prev_node;  curr_node != node;  curr_node = curr_node->prev_node )
    KheDrsNodeDelete(curr_node, drs);
  last_node->prev_node = node;

  ** add last_node to last_day, and increment the diversifier **
  ** last_day = HaArray(drs->days, last_day_index); **
  KheDrsDayAddNode(last_node->day, last_node, drs);
  drs->diversifier++;
  *** */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeDoExtendOpen(KHE_DRS_NODE node, int day_index,            */
/*    KHE_DRS_EDGE de, KHE_DYNAMIC_RESOURCE_SOLVER drs)                      */
/*                                                                           */
/*  Extend node using de, only build de up to full length first.             */
/*                                                                           */
/*****************************************************************************/
static bool KheFindContinuingTask(KHE_DRS_NODE node, int i, KHE_DRS_TASK *dt);

static void KheDrsNodeDoExtendOpen(KHE_DRS_NODE node, int day_index,
  KHE_DRS_EDGE de, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_RESOURCE dr;  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_DRS_DAY day;
  /* KHE_DRS_TASK_GROUP dtg; */
  int i;  KHE_DRS_TASK dt;
  if( HaArrayCount(de->tasks) == HaArrayCount(drs->solve_resources) )
  {
    /* de is ready to use; but copy it first */
    de = KheDrsEdgeCopy(de, drs);
    DoExtend(node, de, drs);
  }
  else
  {
    /* find alternative tasks for the next resource */ 
    i = HaArrayCount(de->tasks);
    dr = HaArray(drs->solve_resources, i);
    drd = HaArray(dr->days, day_index);
    if( KheFindContinuingTask(node, i, &dt) )
    {
      /* dr's current task, dt, is continuing */
      HaArrayAddLast(de->tasks, dt);
      KheDrsNodeDoExtendOpen(node, day_index, de, drs);
      HaArrayDeleteLast(de->tasks);
    }
    else if( drd->solve_task != NULL )
    {
      /* dr has a particular task on this day that must be used */
      HaArrayAddLast(de->tasks, drd->solve_task);
      KheDrsNodeDoExtendOpen(node, day_index, de, drs);
      HaArrayDeleteLast(de->tasks);
    }
    else
    {
      /* try the first available task from each task group of day */
      /* *** replacing this still to do
      HaArrayForEach(day->task_groups, dtg, i)
        if( KheTaskGroupContainsSuitableTask(dtg, dr, de, &dt) )
	{
	  HaArrayAddLast(de->tasks, dt);
	  KheDrsNodeDoExtendOpen(node, day_index, de, drs);
	  HaArrayDeleteLast(de->tasks);
	}
      *** */

      /* try a free day */
      day = HaArray(drs->days, day_index);
      HaArrayAddLast(de->tasks, day->free_task);
      KheDrsNodeDoExtendOpen(node, day_index, de, drs);
      HaArrayDeleteLast(de->tasks);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeExtendOpen(KHE_DRS_NODE node, int day_index,              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Extend the dynamic programming search by searching out of node.  Each    */
/*  new task starts on the day with index day_index, which is open (part     */
/*  of a selected interval).                                                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodeExtendOpen(KHE_DRS_NODE node, int day_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* int i, j;  KHE_DRS_TASK dt; */  /* KHE_DRS_TIME stime; */
  KHE_DRS_EDGE de;
  de = KheDrsEdgeMake(drs);
  KheDrsNodeDoExtendOpen(node, day_index, de, drs);
  KheDrsEdgeDelete(de, drs);

  /* extend for each task beginning on day */
  /* *** 
  HaArrayForEach(day->times, stime, i)
    HaArrayForEach(stime->best_tasks, dt, j)
      DoExtend(node, dt, drs);
  ** extend by taking day off **
  DoExtend(node, day->free_task, drs);
  *** */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeDebugBasics(KHE_DRS_NODE node, FILE *fp)                  */
/*                                                                           */
/*  Produce an unindented one-line debug print of the basics of node.        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodeDebugBasics(KHE_DRS_NODE node, FILE *fp)
{
  /* task */
  fprintf(fp, "NODE(");
  /* ***
  if( node->task != NULL && node->task->task != NULL )
    KheTaskDebug(node->task->task, 1, -1, fp);
  else
    fprintf(fp, "-");
  *** */

  /* day_index */
  if( node->day == NULL )
    fprintf(fp, "to -");
  else
    fprintf(fp, "to %s", KheTimeGroupId(node->day->time_group));

  /* assts, best_of, and cost */
  fprintf(fp, ", assts %d, best of %d, cost %.5f)",
    HaArray(node->signature, 0), node->best_of, KheCostShow(node->cost));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeDebug(KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs, */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of node onto fp with the given verbosity and indent.         */
/*                                                                           */
/*  NB If verbosity == 1 or indent < 0, it is safe to pass NULL for drs.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodeDebug(KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  KHE_DRS_EXPR e;  KHE_DRS_RESOURCE dr;  int i, j;
  if( indent < 0 )
    KheDrsNodeDebugBasics(node, fp);
  else if( verbosity == 1 )
  {
    fprintf(fp, "%*s", indent, "");
    KheDrsNodeDebugBasics(node, fp);
    fprintf(fp, "\n");
  }
  else
  {
    /* basics */
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsNodeDebugBasics(node, fp);
    fprintf(fp, "\n");

    /* incoming edge */
    if( node->prev_edge != NULL )
      KheDrsEdgeDebug(node->prev_edge, drs, verbosity, indent + 2, fp);

    /* signature */
    HaArrayForEach(drs->solve_resources, dr, j)
      HaArrayForEach(dr->exprs, e, i)
      {
	if( KheDrsExprFirstDebugDayIndex(e) <= node->day->day_index &&
	    node->day->day_index <= e->last_day_index )
	{
	  fprintf(fp, "%*s", indent + 2, "");
	  KheDrsExprDebug(e, node, drs, verbosity, -1, fp);
	  fprintf(fp, "\n");
	}
      }
    fprintf(fp, "%*s]\n", indent, "");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_NODES"                                                */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_NODES KheDrsNodesMake(KHE_DYNAMIC_RESOURCE_SOLVER drs)           */
/*                                                                           */
/*  Return a new, empty KHE_DRS_NODES object.                                */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_NODES KheDrsNodesMake(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_NODES res;
  if( HaArrayCount(drs->free_node_sets) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_node_sets);
    HaArrayClear(res->nodes);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->nodes, drs->arena);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodesDelete(KHE_DRS_NODES ns, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  Delete node_set and the nodes in it.                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodesDelete(KHE_DRS_NODES ns, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_NODE node;  int i;
  HaArrayForEach(ns->nodes, node, i)
    KheDrsNodeDelete(node, drs);
  HaArrayAddLast(drs->free_node_sets, ns);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsNodesSignaturePartialHashUntyped(void *p)                      */
/*                                                                           */
/*  Hash function for nodess, being the partial hash of the set's first      */
/*  element.                                                                 */
/*                                                                           */
/*****************************************************************************/

static int KheDrsNodesSignaturePartialHashUntyped(void *p)
{
  KHE_DRS_NODES ns;  KHE_DRS_NODE node;
  ns = (KHE_DRS_NODES) p;
  node = HaArrayFirst(ns->nodes);
  return KheDrsNodeSignaturePartialHash(node);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsNodesSignaturePartialEqualUntyped(void *p1, void *p2)         */
/*                                                                           */
/*  Equality function for nodes, being the partial equality of the sets'     */
/*  first elements.                                                          */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsNodesSignaturePartialEqualUntyped(void *p1, void *p2)
{
  KHE_DRS_NODES ns1, ns2;  KHE_DRS_NODE node1, node2;
  ns1 = (KHE_DRS_NODES) p1;
  ns2 = (KHE_DRS_NODES) p2;
  node1 = HaArrayFirst(ns1->nodes);
  node2 = HaArrayFirst(ns2->nodes);
  return KheDrsNodeSignaturePartialEqual(node1, node2);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodesDoStrongDominanceTest(KHE_DRS_NODES ns,                  */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs, int *total_count)  */
/*                                                                           */
/*  Carry out a strong dominance test between ns and node.                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodesDoStrongDominanceTest(KHE_DRS_NODES ns,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs, int *total_count)
{
  KHE_DRS_NODE other_node;  int i;

  /* if node is dominated by anything else, delete node and return */
  HaArrayForEach(ns->nodes, other_node, i)
    if( KheDrsNodeDominates(other_node, node, 0) )
    {
      KheDrsNodeDelete(node, drs);
      return;
    }

  /* remove other nodes that node dominates */
  HaArrayForEach(ns->nodes, other_node, i)
    if( KheDrsNodeDominates(node, other_node, 0) )
    {
      KheDrsNodeDelete(other_node, drs);
      HaArrayDeleteAndPlug(ns->nodes, i);
      (*total_count)--;
      i--;
    }

  /* add node to ns */
  HaArrayAddLast(ns->nodes, node);
  (*total_count)++;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodesSignatureDebugUntyped(void *p, FILE *fp)                 */
/*                                                                           */
/*  Debug print of the signature of node set p.                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodesSignatureDebugUntyped(void *p, FILE *fp)
{
  KHE_DRS_NODES ns;  KHE_DRS_NODE node;
  ns = (KHE_DRS_NODES) p;
  node = HaArrayFirst(ns->nodes);
  KheDrsNodeSignatureDebug(node, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_NODE_TRIE"                                            */
/*                                                                           */
/*  Abstraction function                                                     */
/*  --------------------                                                     */
/*                                                                           */
/*  A node_trie represents a set of nodes.  If node_trie == NULL, the set    */
/*  is empty.  Otherwise the set consists of node_trie->node (if non-NULL)   */
/*  plus the sets stored in each of the children.                            */
/*                                                                           */
/*  Invariant                                                                */
/*  ---------                                                                */
/*                                                                           */
/*  Exactly one of these three cases applies:                                */
/*                                                                           */
/*    (1)  node_trie == NULL                                                 */
/*    (2)  node_trie->node != NULL                                           */
/*    (3)  HaArrayCount(node_trie->children) > 0                             */
/*                                                                           */
/*  and in the last case, the last child of node_trie->children is non-NULL. */
/*  Each node appears in the trie in a place determined by the sequence      */
/*  of values in its signature, in the usual trie manner.                    */
/*                                                                           */
/*  There is no requirement that two or more nodes be present in case (3).   */
/*  If there are no nodes, case (1) always applies; if there are two or      */
/*  more nodes, case (3) always applies; but if there is exactly one node,   */
/*  then either of cases (2) or (3) applies.  This is because it would be    */
/*  too slow and tedious to detect situations where a case (3) node could    */
/*  be simplified to a case (2) node.                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_NODE_TRIE KheDrsNodeTrieMake(KHE_DRS_NODE node,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new trie object containing node.  It satisfies case (2) above.    */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_NODE_TRIE KheDrsNodeTrieMake(KHE_DRS_NODE node,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_NODE_TRIE res;
  HnAssert(node != NULL, "KheDrsNodeTrieMake internal error (node == NULL)");
  if( HaArrayCount(drs->free_node_tries) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_node_tries);
    HaArrayClear(res->children);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->children, drs->arena);
  }
  res->node = node;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeTrieDelete(KHE_DRS_NODE_TRIE *node_trie,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Delete *node_trie and all its descendants and nodes.                     */
/*  Set *node_trie to NULL.                                                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodeTrieDelete(KHE_DRS_NODE_TRIE *node_trie,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i;
  if( *node_trie == NULL )
  {
    /* case (1), do nothing */
  }
  else if( (*node_trie)->node != NULL )
  {
    /* case (2), delete (*node_trie)->node and *node_trie */
    KheDrsNodeDelete((*node_trie)->node, drs);
    HaArrayAddLast(drs->free_node_tries, *node_trie);
    *node_trie = NULL;
  }
  else
  {
    /* case (3), delete children and *node_trie */
    for( i = 0;  i < HaArrayCount((*node_trie)->children);  i++ )
      KheDrsNodeTrieDelete(&HaArray((*node_trie)->children, i), drs);
    HaArrayAddLast(drs->free_node_tries, *node_trie);
    *node_trie = NULL;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeTrieInsertChild(KHE_DRS_NODE_TRIE node_trie,              */
/*    KHE_DRS_NODE node, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)         */
/*                                                                           */
/*  Insert node in its appropriate place among the children of node_trie.    */
/*                                                                           */
/*****************************************************************************/
static void KheDrsNodeTrieInsert(KHE_DRS_NODE_TRIE *node_trie,
  KHE_DRS_NODE node, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsNodeTrieInsertChild(KHE_DRS_NODE_TRIE node_trie,
  KHE_DRS_NODE node, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int index;
  index = HaArray(node->signature, depth);
  HaArrayFill(node_trie->children, index + 1, NULL);
  KheDrsNodeTrieInsert(&HaArray(node_trie->children, index), node,
    depth + 1, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeTrieInsert(KHE_DRS_NODE_TRIE *node_trie,                  */
/*    KHE_DRS_NODE node, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)         */
/*                                                                           */
/*  Insert node into *node_trie, possibly changing *node_trie as a result.   */
/*  The depth parameter indicates how far along the signature we are.        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodeTrieInsert(KHE_DRS_NODE_TRIE *node_trie,
  KHE_DRS_NODE node, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( *node_trie == NULL )
  {
    /* case (1), make a new node containing just node, -> case (2) */
    *node_trie = KheDrsNodeTrieMake(node, drs);
  }
  else if( (*node_trie)->node != NULL )
  {
    /* case (2), add node and (*node_trie)->node to children, -> case (3) */
    KheDrsNodeTrieInsertChild(*node_trie, node, depth, drs);
    node = (*node_trie)->node;
    (*node_trie)->node = NULL;
    KheDrsNodeTrieInsertChild(*node_trie, node, depth, drs);
  }
  else
  {
    /* case (3), add node to children, -> case (3) */
    KheDrsNodeTrieInsertChild(*node_trie, node, depth, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsNodeTrieDominates(KHE_DRS_NODE_TRIE node_trie,                */
/*    KHE_DRS_NODE node, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)         */
/*                                                                           */
/*  Return true if there is a node in node_trie that dominates node.         */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsNodeTrieDominates(KHE_DRS_NODE_TRIE node_trie,
  KHE_DRS_NODE node, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int val, from_val, to_val, i;  KHE_DRS_DOM_TEST dom_test;
  if( node_trie == NULL )
  {
    /* case (1); no nodes in node_trie, so none dominate */
    return false;
  }
  else if( node_trie->node != NULL )
  {
    /* case (2); the only possible dominator is node_trie->node */
    return KheDrsNodeDominates(node_trie->node, node, depth);
  }
  else
  {
    /* case (3); find range of dominators and try each of them */
    dom_test = HaArray(node->day->dom_tests, depth);
    val = HaArray(node->signature, depth);
    KheDrsDomTestDominatingRange(dom_test, val, &from_val, &to_val);
    if( to_val >= HaArrayCount(node_trie->children) )
      to_val = HaArrayCount(node_trie->children) - 1;
    for( i = from_val;  i <= to_val;  i++ )
      if( KheDrsNodeTrieDominates(HaArray(node_trie->children,i), node, depth+1, drs))
	return true;
    return false;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeTrieRemoveDominated(KHE_DRS_NODE_TRIE *node_trie,         */
/*    KHE_DRS_NODE node, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs,         */
/*    int *total_count)                                                      */
/*                                                                           */
/*  Remove those elements of *node_trie that are dominated by node.  This    */
/*  can change *node_trie.                                                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodeTrieRemoveDominated(KHE_DRS_NODE_TRIE *node_trie,
  KHE_DRS_NODE node, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int *total_count)
{
  int val, from_val, to_val, i;  KHE_DRS_DOM_TEST dom_test;
  if( *node_trie == NULL )
  {
    /* case (1); there is nothing to remove */
  }
  else if( (*node_trie)->node != NULL )
  {
    /* case (2); if (*node_trie)->node is dominated, remove whole trie */
    if( KheDrsNodeDominates(node, (*node_trie)->node, depth) )
    {
      KheDrsNodeTrieDelete(node_trie, drs);
      (*total_count)--;
    }
  }
  else
  {
    /* case (3); removed dominated nodes among the children */
    dom_test = HaArray(node->day->dom_tests, depth);
    val = HaArray(node->signature, depth);
    KheDrsDomTestDominatedRange(dom_test, val, &from_val, &to_val);
    if( to_val >= HaArrayCount((*node_trie)->children) )
      to_val = HaArrayCount((*node_trie)->children) - 1;
    for( i = from_val;  i <= to_val;  i++ )
      KheDrsNodeTrieRemoveDominated(&HaArray((*node_trie)->children, i), node,
	depth + 1, drs, total_count);

    /* remove trailing NULL children */
    while( HaArrayCount((*node_trie)->children) > 0 &&
	HaArrayLast((*node_trie)->children) == NULL )
      HaArrayDeleteLast((*node_trie)->children);

    /* if children array is empty then the whole trie is empty, so delete it */
    if( HaArrayCount((*node_trie)->children) == 0 )
      KheDrsNodeTrieDelete(node_trie, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeTrieDoStrongDominanceTest(KHE_DRS_NODE_TRIE *node_trie,   */
/*    KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int *total_count)                                                      */
/*                                                                           */
/*  Carry out a strong dominance test between *node_trie and node.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodeTrieDoStrongDominanceTest(KHE_DRS_NODE_TRIE *node_trie,
  KHE_DRS_NODE node, KHE_DYNAMIC_RESOURCE_SOLVER drs, int *total_count)
{
  if( DEBUG12 )
    fprintf(stderr, "[ KheDrsNodeTrieDoStrongDominanceTest(node for %s)\n",
      KheTimeGroupId(node->day->time_group));
  if( KheDrsNodeTrieDominates(*node_trie, node, 0, drs) )
  {
    /* node is dominated by some other node, so delete node */
    if( DEBUG12 )
      fprintf(stderr, "  trie deleting node for %s\n",
	KheTimeGroupId(node->day->time_group));
    KheDrsNodeDelete(node, drs);
  }
  else
  {
    /* remove nodes that node dominates, and insert node */
    if( DEBUG12 )
      fprintf(stderr, "  trie inserting node for %s\n",
	KheTimeGroupId(node->day->time_group));
    KheDrsNodeTrieRemoveDominated(node_trie, node, 0, drs, total_count);
    KheDrsNodeTrieInsert(node_trie, node, 0, drs);
    (*total_count)++;
  }
  if( DEBUG12 )
    fprintf(stderr, "] KheDrsNodeTrieDoStrongDominanceTest returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeTrieExtend(KHE_DRS_NODE_TRIE node_trie,                   */
/*    int day_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)                        */
/*                                                                           */
/*  Extend the solutions in node_trie to next_day.                           */
/*                                                                           */
/*****************************************************************************/
static void KheDrsNodeExtendOpen(KHE_DRS_NODE node, int day_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsNodeTrieExtendOpen(KHE_DRS_NODE_TRIE node_trie,
  int day_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_NODE_TRIE child_trie;  int i;
  if( node_trie == NULL )
  {
    /* case (1); do nothing */
  }
  else if( node_trie->node != NULL )
  {
    /* case (2); extend node_trie->node */
    KheDrsNodeExtendOpen(node_trie->node, day_index, drs);
  }
  else
  {
    /* case (3); handle children */
    HaArrayForEach(node_trie->children, child_trie, i)
      KheDrsNodeTrieExtendOpen(child_trie, day_index, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeTrieExtractFinalSolns(KHE_DRS_NODE_TRIE node_trie,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Extract all the solutions from node_trie and make them final solutions.  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodeTrieExtractFinalSolns(KHE_DRS_NODE_TRIE node_trie,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_NODE_TRIE child_trie;  int i;
  if( node_trie == NULL )
  {
    /* case (1); do nothing */
  }
  else if( node_trie->node != NULL )
  {
    /* case (2); extract node_trie->node */
    HaArrayAddLast(drs->final_solns, node_trie->node);
  }
  else
  {
    /* case (3); handle children */
    HaArrayForEach(node_trie->children, child_trie, i)
      KheDrsNodeTrieExtractFinalSolns(child_trie, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_NODE_SET"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_NODE_SET KheDrsNodeSetMake(KHE_TIME_GROUP tg,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a new KHE_DRS_NODE_SET object.                                    */
/*                                                                           */
/*****************************************************************************/

KHE_DRS_NODE_SET KheDrsNodeSetMake(/* KHE_TIME_GROUP tg, */
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_NODE_SET res;
  HaMake(res, drs->arena);
  /* res->time_group = tg; */
  res->count = 0;
  HpTableInit(res->weak_dom_set,
    &KheDrsNodeSignatureFullHashUntyped,
    &KheDrsNodeSignatureFullEqualUntyped,
    &KheDrsNodeSignatureDebugUntyped, drs->arena);
  HpTableInit(res->medium_dom_set,
    &KheDrsNodesSignaturePartialHashUntyped,
    &KheDrsNodesSignaturePartialEqualUntyped,
    &KheDrsNodesSignatureDebugUntyped, drs->arena);
  HaArrayInit(res->strong_dom_set, drs->arena);
  res->trie_dom_set = NULL;  /* represents the empty set */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeSetClear(KHE_DRS_NODE_SET dns,                            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Remove and delete the nodes and node_sets held in dns.                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodeSetClear(KHE_DRS_NODE_SET dns,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int pos;  KHE_DRS_NODE node;  KHE_DRS_NODES ns;
  HpTableForEachValue(dns->weak_dom_set, node, pos)
    KheDrsNodeDelete(node, drs);
  HpTableClear(dns->weak_dom_set);
  HpTableForEachValue(dns->medium_dom_set, ns, pos)
    KheDrsNodesDelete(ns, drs);
  HpTableClear(dns->medium_dom_set);
  HaArrayForEach(dns->strong_dom_set, ns, pos)
    KheDrsNodesDelete(ns, drs);
  HaArrayClear(dns->strong_dom_set);
  KheDrsNodeTrieDelete(&dns->trie_dom_set, drs);
  HnAssert(dns->trie_dom_set == NULL,
    "KheDrsNodeSetClear internal error");
  dns->count = 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeSetAddNode(KHE_DRS_NODE_SET dns, KHE_DRS_NODE node,       */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add node to dns.                                                         */
/*                                                                           */
/*  Strong dominance                                                         */
/*  ----------------                                                         */
/*                                                                           */
/*  For each node other_node in dns, if node is dominated by other_node,     */
/*  then delete node and return.                                             */
/*                                                                           */
/*  Otherwise, delete each other_node that node dominates, then add node.    */
/*                                                                           */
/*  Weak dominance                                                           */
/*  --------------                                                           */
/*                                                                           */
/*  If dns does not already contain a node with the same signature as node,  */
/*  just add node to dns and return.                                         */
/*                                                                           */
/*  If dns already contains a node (call it other_node) with the same        */
/*  signature as node, keep whichever one of node and other_node has the     */
/*  smallest cost, by overwriting other_node's fields with node's fields,    */
/*  if required.                                                             */
/*                                                                           */
/*  If node and other_node have equal cost, keep one chosen at random, with  */
/*  weights which ensure that if there are K nodes with the same signature   */
/*  and the same cost, then each has probability 1/K of being the one kept.  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodeSetAddNode(KHE_DRS_NODE_SET dns, KHE_DRS_NODE node,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_NODE other_node;  int asst_count;
  KHE_DRS_NODES node_set, other_node_set;
  switch( drs->dom_kind )
  {
    case KHE_DRS_DOM_WEAK:

      /* do equivalence testing */
      if( HpTableAddUnique(dns->weak_dom_set, (void *) node, node, other_node) )
	dns->count++;
      else
      {
	/* other_node will be best of one more node than before */
	other_node->best_of++;

	/* ensure other_node is the better of other_node and node */
	if( node->cost < other_node->cost || (node->cost == other_node->cost &&
	      drs->diversifier % other_node->best_of == 0) )
	{
	  /* node is better than other_node, so overwrite with node */
	  KheDrsNodeOverWrite(other_node, node);
	}

	/* free node */
	KheDrsNodeDelete(node, drs);
      }
      break;

    case KHE_DRS_DOM_MEDIUM:

      /* make node_set holding just node and try adding it to medium_dom_set */
      node_set = KheDrsNodesMake(drs);
      HaArrayAddLast(node_set->nodes, node);
      if( HpTableAddUnique(dns->medium_dom_set, (void *) node_set, node_set,
	    other_node_set) )
      {
	/* node_set added, so that's all that's needed */
	dns->count++;
      }
      else
      {
	/* remove node_set and do strong dom. of node with other_node_set */
	HaArrayClear(node_set->nodes);
	KheDrsNodesDelete(node_set, drs);
        KheDrsNodesDoStrongDominanceTest(other_node_set, node, drs,
	  &dns->count);
      }
      break;

    case KHE_DRS_DOM_STRONG:
    
      /* find or make node_set, the KHE_DRS_NODES which should hold node */
      asst_count = HaArrayFirst(node->signature);
      while( HaArrayCount(dns->strong_dom_set) <= asst_count )
      {
	node_set = KheDrsNodesMake(drs);
	HaArrayAddLast(dns->strong_dom_set, node_set);
      }
      node_set = HaArray(dns->strong_dom_set, asst_count);

      /* carry out strong dominance test between node_set and node */
      KheDrsNodesDoStrongDominanceTest(node_set, node, drs, &dns->count);
      break;

    case KHE_DRS_DOM_TRIE:
    
      KheDrsNodeTrieDoStrongDominanceTest(&dns->trie_dom_set, node, drs,
	&dns->count);
      break;
    
    default:

      HnAbort("KheDrsNodeSetAddNode internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeSetExtendOpen(KHE_DRS_NODE_SET dns, int day_index,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Extend dns by adding tasks from the day with index day_index.            */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodeSetExtendOpen(KHE_DRS_NODE_SET dns, int day_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_NODE node;  int i, j;  KHE_DRS_NODES node_set;
  switch( drs->dom_kind )
  {
    case KHE_DRS_DOM_WEAK:

      if( DEBUG9 )
      {
	float probes;
	probes = HpTableProbeLength(dns->weak_dom_set);
	fprintf(stderr, "  hash table: count %d, size %d, probes %.2f\n",
	  /* KheTimeGroupId(dns->time_group), */ dns->count,
	  HpTableSize(dns->weak_dom_set), probes);
	if( probes > 14.0 )
	  HpTableDebug(dns->weak_dom_set, 2, stderr);
      }
      HpTableForEachValue(dns->weak_dom_set, node, i)
	KheDrsNodeExtendOpen(node, day_index, drs);
      break;

    case KHE_DRS_DOM_MEDIUM:

      HpTableForEachValue(dns->medium_dom_set, node_set, i)
	HaArrayForEach(node_set->nodes, node, j)
	  KheDrsNodeExtendOpen(node, day_index, drs);
      break;

    case KHE_DRS_DOM_STRONG:

      HaArrayForEach(dns->strong_dom_set, node_set, i)
	HaArrayForEach(node_set->nodes, node, j)
	  KheDrsNodeExtendOpen(node, day_index, drs);
      break;

    case KHE_DRS_DOM_TRIE:
    
      KheDrsNodeTrieExtendOpen(dns->trie_dom_set, day_index, drs);
      break;
    
    default:

      HnAbort("KheDrsNodeSetExtend internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsNodeSetExtractFinalSolns(KHE_DRS_NODE_SET dns,                */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Extract final solutions from dns.                                        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsNodeSetExtractFinalSolns(KHE_DRS_NODE_SET dns,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_NODE node;  int i, j;  KHE_DRS_NODES node_set;
  switch( drs->dom_kind )
  {
    case KHE_DRS_DOM_WEAK:

      HpTableForEachValue(dns->weak_dom_set, node, i)
	HaArrayAddLast(drs->final_solns, node);
      break;

    case KHE_DRS_DOM_MEDIUM:

      HpTableForEachValue(dns->medium_dom_set, node_set, i)
	HaArrayForEach(node_set->nodes, node, j)
	  HaArrayAddLast(drs->final_solns, node);
      break;

    case KHE_DRS_DOM_STRONG:

      HaArrayForEach(dns->strong_dom_set, node_set, i)
	HaArrayForEach(node_set->nodes, node, j)
	  HaArrayAddLast(drs->final_solns, node);
      break;

    case KHE_DRS_DOM_TRIE:
    
      KheDrsNodeTrieExtractFinalSolns(dns->trie_dom_set, drs);
      break;
    
    default:

      HnAbort("KheDrsNodeSetExtractFinalSolns internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_DAY"                                                  */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DAY KheDrsDayMake(int day_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)*/
/*                                                                           */
/*  Make a new day object for the day with this day index.                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DAY KheDrsDayMake(int day_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_DAY res;  /* KHE_TIME t;  int i;  KHE_DRS_TIME stime; */
  HaMake(res, drs->arena);
  res->day_index = day_index;
  res->time_group = KheFrameTimeGroup(drs->days_frame, day_index);
  /* ***
  HaArrayInit(res->times, drs->arena);
  for( i = 0;  i < KheTimeGroupTimeCount(res->time_group);  i++ )
  {
    t = KheTimeGroupTime(res->time_group, i);
    stime = KheDrsTimeMake(t, drs);
    HaArrayAddLast(res->times, stime);
  }
  *** */
  /* res->nodes_count = 0; */
  /* ***
  res->sig_len = 1;  ** asst_count is always at start of signature **
  HaArrayInit(res->dom_tests, drs->arena);
  HaArrayInit(res->eq_dom_test_indexes, drs->arena);
  KheDrsDayAddDomTest(res, KheDrsDomTestMake(KHE_DRS_DOM_EQ));
  *** */
  /* ***
  HaArrayInit(res->active_today, drs->arena);
  HaArrayInit(res->leaf_today, drs->arena);
  *** */
  HaArrayInit(res->dom_tests, drs->arena);
  HaArrayInit(res->eq_dom_test_indexes, drs->arena);
  HaArrayInit(res->task_classes, drs->arena);
  /* adding classes (based on orig classes) still to do */
  res->free_task = NULL;  /* still to do */
  /* res->free_task = KheDrsTaskMakeFree(day_index, drs); */
  res->node_set = KheDrsNodeSetMake(/* res->time_group, */ drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK_GROUP KheDrsDayAddTask(KHE_DRS_DAY day, KHE_DRS_TASK dt)    */
/*                                                                           */
/*  Add dt to day, and return the task group it was added to.                */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_TASK_GROUP KheDrsDayAddTask(KHE_DRS_DAY day, KHE_DRS_TASK dt)
{
  KHE_DRS_TASK_GROUP sg;  int i;
  HaArrayForEach(day->task_groups, sg, i)
    if( KheDrsTaskGroupAccepts(sg, dt) )
    {
      KheDrsTaskGroupAddTask(sg, dt);
      return sg;
    }
  sg = KheDrsTaskGroupMake(dt, HaArrayArena(day->task_groups));
  HaArrayAddLast(day->task_groups, sg);
  return sg;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayClearAvail(KHE_DRS_DAY day)                                */
/*                                                                           */
/*  Clear away all the avail tasks on day.                                   */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsDayClearAvail(KHE_DRS_DAY day)
{
  KHE_DRS_TASK_GROUP dtg;  int i;
  HaArrayForEach(day->task_groups, dtg, i)
    KheDrsTaskGroupClearAvail(dtg);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayMakeUnboundTasksAvail(KHE_DRS_DAY day)                     */
/*                                                                           */
/*  Make the unbound tasks in the task groups of day available.              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsDayMakeUnboundTasksAvail(KHE_DRS_DAY day)
{
  KHE_DRS_TASK_GROUP dtg;  int i;
  HaArrayForEach(day->task_groups, dtg, i)
    KheDrsTaskGroupMakeUnboundTasksAvail(dtg);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayClearTasks(KHE_DRS_DAY day,                                */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Clear out all the tasks running on day, except the free task.            */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsDayClearTasks(KHE_DRS_DAY day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i;  KHE_DRS_TIME t;
  HaArrayForEach(day->times, t, i)
    KheDrsTimeClearTasks(t, drs);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayInitializeForSolving(KHE_DRS_DAY day,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Initialize day for solving.                                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayInitializeForSolving(KHE_DRS_DAY day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_CLASS dtc;  KHE_TASK_CLASS tc;  int i, j;
  KHE_TASK_CLASS_SOLVER tcs;  KHE_TIME time;

  /* delete old task classes */
  HaArrayForEach(day->task_classes, dtc, i)
    KheDrsTaskClassDelete(dtc, drs);
  HaArrayClear(day->task_classes);

  /* make one drs task class for each task class on that day */
  tcs = drs->task_class_solver;
  for( i = 0;  i < KheTimeGroupTimeCount(day->time_group);  i++ )
  {
    time = KheTimeGroupTime(day->time_group, i);
    for( j = 0;  j < KheTaskClassSolverClassAtTimeCount(tcs, time);  j++ )
    {
      tc = KheTaskClassSolverClassAtTime(tcs, time, i);
      dtc = KheDrsTaskClassMake(tc, drs);
      HaArrayAddLast(day->task_classes, dtc);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayClear(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Clear day, ready for a fresh solve.                                      */
/*                                                                           */
/*  Implementation note.  The expressions in leaf_today will be deleted      */
/*  separately, as part of KheDynamicResourceSolverClear.                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayClear(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* KheDrsDayClearTasks(day, drs); */
  /* ***
  day->sig_len = 1;
  HaArrayClear(day->dom_tests);
  HaArrayClear(day->eq_dom_test_indexes);
  KheDrsDayAddDomTest(day, KheDrsDomTestMake(KHE_DRS_DOM_EQ));
  HaArrayClear(day->active_today);
  HaArrayClear(day->leaf_today);
  *** */
  KheDrsNodeSetClear(day->node_set, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDayAddTasks(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs, */
/*    KHE_DRS_TASK *preassigned_dt)                                          */
/*                                                                           */
/*  Add suitable tasks to day.                                               */
/*                                                                           */
/*  If true is returned, it means that a preassigned task was found; a task  */
/*  for it will be the only one added.  It is returned in *preassigned_dt.   */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete
static bool KheDrsDayAddTasks(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_DRS_TASK *preassigned_dt)
{
  int i, j, k;  KHE_TIME t;  KHE_TASK task;  KHE_DRS_TIME stime;
  KHE_DRS_TASK dt;  KHE_MEET meet;
  HaArrayForEach(day->times, stime, i)
  {
    ** for each time of day **
    t = stime->time;
    for( j = 0;  j < KheEventTimetableMonitorTimeMeetCount(drs->etm, t);  j++ )
    {
      ** for each meet running at time **
      meet = KheEventTimetableMonitorTimeMeet(drs->etm, stime->time, j);
      for( k = 0;  k < KheMeetTaskCount(meet);  k++ )
      {
	** for the proper root task of each task of meet **
	task = KheTaskProperRoot(KheMeetTask(meet, k));
	if( KheTaskAsstResource(task) == drs->resource )
	{
	  ** task is already assigned, so it is the only legal choice **
	  KheDrsDayClearTasks(day, drs);
	  dt = KheDrsTaskMake(task, drs);
	  HaArrayAddLast(stime->best_tasks, dt);
	  *preassigned_dt = dt;
	  return true;
	}
	else if( KheTaskAssignResourceCheck(task, drs->resource) )
	{
	  dt = KheDrsTaskMake(task, drs);
	  if( KheDrsTaskFirstTime(dt) == t )
	  {
	    ** add dt to stime, possibly replacing an existing task **
	    KheDrsTimeAddTask(stime, dt, drs);
	  }
	  else
	  {
	    ** dt starts at the wrong time, so ignore it **
	    KheDrsTaskDelete(dt, drs);
	  }
	}
      }
    }
  }
  *preassigned_dt = NULL;
  return false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayAddDomTest(KHE_DRS_DAY day, KHE_DRS_DOM_TEST dom_test)     */
/*                                                                           */
/*  Add dom_test to day.  This may change the eq_dom_test_indexes array as   */
/*  well as the dom_tests array.                                             */
/*                                                                           */
/*****************************************************************************/

/* *** replacement for this is KheDrsResourceOnDayAddDomTest
static void KheDrsDayAddDomTest(KHE_DRS_DAY day, KHE_DRS_DOM_TEST dom_test)
{
  if( dom_test.type == KHE_DRS_DOM_EQ || dom_test.type == KHE_DRS_DOM_EQ_LOOSE )
    HaArrayAddLast(day->eq_dom_test_indexes, HaArrayCount(day->dom_tests));
  HaArrayAddLast(day->dom_tests, dom_test);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayDomTestsDebug(KHE_DRS_DAY day, int verbosity,              */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of the dom tests of day.                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayDomTestsDebug(KHE_DRS_DAY day, int verbosity,
  int indent, FILE *fp)
{
  int dom_test_freq[KHE_DRS_DOM_EQ_LOOSE + 1];  int i;
  KHE_DRS_DOM_TEST dom_test;  char buff[100];

  /* initialize dom_test_freq */
  for( i = KHE_DRS_DOM_UNUSED;  i <= KHE_DRS_DOM_EQ_LOOSE;  i++ )
    dom_test_freq[i] = 0;

  /* accumulate dom_test_freq */
  HaArrayForEach(day->dom_tests, dom_test, i)
    dom_test_freq[dom_test.type]++;

  /* display frequencies */
  if( indent >= 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "[ Day %s dom tests: ", KheTimeGroupId(day->time_group));
  for( i = 0;  i <= KHE_DRS_DOM_EQ_LOOSE;  i++ )
  {
    if( i > 0 )
      fprintf(fp, ", ");
    fprintf(fp, "%d %s", dom_test_freq[i],
      KheDrsDomTestShow(KheDrsDomTestMake(i), buff));
  }
  fprintf(fp, " ]");
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayDebug(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs,    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of day onto fp.                                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayDebug(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  /* int i, pos */;  /* KHE_DRS_TIME stime; */
  /* KHE_DRS_DOM_TEST dom_test; */  /* char buff[100]; */
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ Day ", indent, "");
    KheTimeGroupDebug(day->time_group, 1, 0, stderr);
    /* ***
    fprintf(fp, " (sig_len %d)\n", day->sig_len);
    fprintf(fp, " (times %d, sig_len %d)\n", HaArrayCount(day->times),
      day->sig_len);
    fprintf(fp, " (times %d, sig_len %d, active_today %d)\n",
      HaArrayCount(day->times), day->sig_len, HaArrayCount(day->active_today));
    HaArrayForEach(day->times, stime, i)
      KheDrsTimeDebug(stime, drs, verbosity, indent + 2, fp);
    fprintf(fp, "%*s  dom_tests [", indent, "");
    HaArrayForEach(day->dom_tests, dom_test, i)
      fprintf(fp, "%s%s%s", i > 0 ? ", " : "",
      HaArrayContains(day->eq_dom_test_indexes, i, &pos) ? "*" : "",
      KheDrsDomTestShow(dom_test, buff));
    *** */
    fprintf(fp, "]\n");
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
    KheTimeGroupDebug(day->time_group, 1, -1, stderr);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_FULL_SOLN"                                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_FULL_SOLN KheDrsFullSolnMake(KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Make a new, empty full soln object.                                      */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_FULL_SOLN KheDrsFullSolnMake(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_FULL_SOLN res;
  HaMake(res, drs->arena);
  res->asst_count = 0;
  HaArrayInit(res->tasks, drs->arena);
  res->resource = NULL;
  res->adopted = false;
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsFullSolnLoad(KHE_DRS_FULL_SOLN fs, KHE_DRS_NODE node)         */
/*                                                                           */
/*  Load node into fs.                                                       */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsFullSolnLoad(KHE_DRS_FULL_SOLN fs, KHE_DRS_NODE node,
  KHE_RESOURCE r)
{
  HaArrayClear(fs->tasks);
  fs->resource = r;
  fs->asst_count = HaArrayFirst(node->signature);
  while( node->task != NULL )
  {
    HaArrayAddLast(fs->tasks, node->task);
    node = node->prev_node;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsFullSolnAdopt(KHE_DRS_FULL_SOLN fs)                           */
/*                                                                           */
/*  Adopt fs.                                                                */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsFullSolnAdopt(KHE_DRS_FULL_SOLN fs)
{
  int i;  KHE_DRS_TASK dt;
  HnAssert(!fs->adopted, "KheDrsFullSolnAdopt: already adopted");
  HaArrayForEach(fs->tasks, dt, i)
    if( dt->task != NULL && !dt->already_assigned )
    {
      if( !KheTaskAssignResource(dt->task, fs->resource) )
      {
	KheTaskDebug(dt->task, 1, 0, stderr);
	HnAbort("KheDrsFullSolnAdopt error (cannot assign task %d to %s)",
	  i, KheResourceId(fs->resource));
      }
      if( DEBUG6 )
      {
	fprintf(stderr, "  KheDrsFullSolnAdopt assigning %s to ",
	  KheResourceId(fs->resource));
	KheTaskDebug(dt->task, 1, 0, stderr);
      }
    }
  fs->adopted = true;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsFullSolnUnAdopt(KHE_DRS_FULL_SOLN fs)                         */
/*                                                                           */
/*  Undo the matching call to KheDrsFullSolnAdopt.                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsFullSolnUnAdopt(KHE_DRS_FULL_SOLN fs)
{
  int i;  KHE_DRS_TASK dt;
  HnAssert(fs->adopted, "KheDrsFullSolnUnAdopt: not adopted");
  HaArrayForEach(fs->tasks, dt, i)
    if( dt->task != NULL && !dt->already_assigned )
    {
      if( !KheTaskUnAssignResource(dt->task) )
	HnAbort("KheDrsFullSolnAdopt internal error");
    }
  fs->adopted = false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_NODE KheFullSolnReRun(KHE_DRS_FULL_SOLN fs,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Rerun fs, returning the resulting node.                                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_NODE KheFullSolnReRun(KHE_DRS_FULL_SOLN fs,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, di, first_day_index, last_day_index;
  KHE_DRS_NODE node;  KHE_TIME_GROUP tg;  KHE_DRS_TASK dt;
  HnAssert(!fs->adopted, "KheFullSolnReRun: already adopted");
  drs->rerun = true;
  node = KheDrsNodeMakeRoot(drs);
  HaArrayForEachReverse(fs->tasks, dt, i)
  {
    first_day_index = KheDailyScheduleFirstDayIndex(dt->daily_schedule);
    last_day_index = KheDailyScheduleLastDayIndex(dt->daily_schedule);
    for( di = first_day_index;  di <= last_day_index;  di++ )
    {
      tg = KheFrameTimeGroup(drs->days_frame, di);
      fprintf(stderr, "  starting day %s\n", KheTimeGroupId(tg));
      node = KheDrsNodeMake(node, dt, HaArray(drs->days, di), drs);
    }
  }
  drs->rerun = false;
  return node;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsFullSolnDebug(KHE_DRS_FULL_SOLN fs, int verbosity,            */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of fs onto fp.                                               */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsFullSolnDebug(KHE_DRS_FULL_SOLN fs, int verbosity,
  int indent, FILE *fp)
{
  int i;  KHE_DRS_TASK dt;
  fprintf(fp, "%*s[ Full Soln (%d assts%s) for %s:\n", indent, "",
    fs->asst_count, fs->adopted ? ", adopted" : "",KheResourceId(fs->resource));
  HaArrayForEachReverse(fs->tasks, dt, i)
    KheDrsTaskDebug(dt, verbosity, indent + 2, fp);
  fprintf(fp, "%*s]\n", indent, "");
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "converting tasks into task groups"                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverAddTasks(KHE_DYNAMIC_RESOURCE_SOLVER drs)   */
/*                                                                           */
/*  Add the tasks, in the form of tasks grouped by equivalent, to drs.       */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDynamicResourceSolverAddTasks(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_TASK task;  KHE_DRS_TASK dt;  int i, first_day_index;  KHE_DRS_DAY day;
  KHE_DRS_TASK_GROUP dtg;
  for( i = 0;  i < KheSolnTaskCount(drs->soln);  i++ )
  {
    task = KheSolnTask(drs->soln, i);
    if( KheTaskResourceType(task) == drs->resource_type &&
	!KheTaskIsCycleTask(task) && KheTaskAsst(task) == NULL )
    {
      dt = KheDrsTaskMake(task, drs);
      first_day_index = KheDailyScheduleFirstDayIndex(dt->daily_schedule);
      day = HaArray(drs->days, first_day_index);
      dtg = KheDrsDayAddTask(day, dt);
      dt->task_group = dtg;
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "public functions"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DYNAMIC_RESOURCE_SOLVER KheDynamicResourceSolverMake(KHE_SOLN soln,  */
/*    KHE_RESOURCE_TYPE rt, KHE_OPTIONS options)                             */
/*                                                                           */
/*  Make and return a new dynamic resource solver with these attributes.     */
/*                                                                           */
/*****************************************************************************/

KHE_DYNAMIC_RESOURCE_SOLVER KheDynamicResourceSolverMake(KHE_SOLN soln,
  KHE_RESOURCE_TYPE rt, KHE_OPTIONS options)
{
  KHE_DYNAMIC_RESOURCE_SOLVER res;  HA_ARENA a;  /* KHE_TASK task; */  int i;
  KHE_RESOURCE r;  KHE_FRAME days_frame;  KHE_EVENT_TIMETABLE_MONITOR etm;

  /* get the common frame and event timetable monitor; return NULL if can't */
  days_frame = KheOptionsFrame(options, "gs_common_frame", soln);
  etm = (KHE_EVENT_TIMETABLE_MONITOR)
    KheOptionsGetObject(options, "gs_event_timetable_monitor", NULL);
  if( days_frame == NULL || etm == NULL )
    return NULL;

  /* make the basic object */
  a = KheSolnArenaBegin(soln, false);
  HaMake(res, a);

  /* fields which are constant throughout the lifetime of the solver */
  res->arena = a;
  res->soln = soln;
  res->resource_type = rt;
  res->options = options;
  res->days_frame = days_frame;
  res->etm = etm;
  res->task_finder = KheTaskFinderMake(soln, options, a);
  res->task_class_solver = KheTaskClassSolverMake(soln, rt, days_frame);
  KheTaskClassSolverAddAllTasks(res->task_class_solver);
  if( DEBUG15 )
    KheTaskClassSolverDebug(res->task_class_solver, 2, 2, stderr);

  /* free list fields */
  /* ***
  HaArrayInit(res->free_expr_busy_time, a);
  HaArrayInit(res->free_expr_free_time, a);
  HaArrayInit(res->free_expr_work_time, a);
  HaArrayInit(res->free_expr_busy_day, a);
  HaArrayInit(res->free_expr_free_day, a);
  HaArrayInit(res->free_expr_work_day, a);
  HaArrayInit(res->free_expr_or, a);
  HaArrayInit(res->free_expr_and, a);
  HaArrayInit(res->free_expr_int_sum, a);
  HaArrayInit(res->free_expr_float_sum, a);
  HaArrayInit(res->free_expr_int_dev, a);
  HaArrayInit(res->free_expr_float_dev, a);
  HaArrayInit(res->free_expr_cost, a);
  HaArrayInit(res->free_expr_int_sum_comb, a);
  HaArrayInit(res->free_expr_int_seq_comb, a);
  *** */
  HaArrayInit(res->free_task_classes, a);
  HaArrayInit(res->free_tasks, a);
  HaArrayInit(res->free_edges, a);
  HaArrayInit(res->free_nodes, a);
  HaArrayInit(res->free_node_sets, a);
  HaArrayInit(res->free_node_tries, a);

  /* fields which are constant themselves but contain varying elements */
  /* one drs resource object for each resource of type rt */
  HaArrayInit(res->all_resources, a);
  for( i = 0;  i < KheResourceTypeResourceCount(rt);  i++ )
  {
    r = KheResourceTypeResource(rt, i);
    KheDrsResourceMakeAndAdd(r, res);
  }

  /* one day object for each day of the frame */
  HaArrayInit(res->days, a);
  for( i = 0;  i < KheFrameTimeGroupCount(days_frame);  i++ )
    HaArrayAddLast(res->days, KheDrsDayMake(i, res));

  /* one drs task object for each proper root task of type rt */
  /* *** obsolete
  for( i = 0;  i < KheSolnTaskCount(soln);  i++ )
  {
    task = KheSolnTask(soln, i);
    if( KheTaskResourceType(task) == rt && KheTaskIsProperRoot(task) )
      KheDrsTaskMakeAndAdd(task, res);
  }
  *** */

  /* fields that vary with the solve */
  res->solve_state = KHE_DRS_SOLVE_LOADING;
  /* res->cluster_solver = KheClusterMinimumSolverMake(a); */
  res->diversifier = 0;
  HaArrayInit(res->solve_resources, a);
  HaArrayInit(res->solve_day_ranges, a);
  res->dom_kind = KHE_DRS_DOM_STRONG;
  /* res->min_assts = res->max_assts = 0; */
  /* res->cost_limit = KheCost(0, 0); */
  /* HaArrayInit(res->exprs, a); */
  HaArrayInit(res->final_solns, a);
  /* res->full_soln = KheDrsFullSolnMake(res); */
  res->rerun = false;

#if TESTING
  res->timer = KheTimerMake("dynamic", KHE_NO_TIME, a);
  HaArrayInit(res->running_time_per_day, a);
#endif

  /* all done */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverDelete(KHE_DYNAMIC_RESOURCE_SOLVER drs)     */
/*                                                                           */
/*  Delete drs, reclaiming its memory.                                       */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverDelete(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KheSolnArenaEnd(drs->soln, drs->arena);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverAddResource(KHE_DYNAMIC_RESOURCE_SOLVER drs,*/
/*    KHE_RESOURCE r)                                                        */
/*                                                                           */
/*  Add resource r to the solve.                                             */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverAddResource(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_RESOURCE r)
{
  KHE_DRS_RESOURCE dr;
  HnAssert(KheResourceResourceType(r) == drs->resource_type,
    "KheDynamicResourceSolverAddResource: r has wrong resource type for drs");
  drs->solve_state = KHE_DRS_SOLVE_LOADING;
  dr = KheDrsResource(r, drs);
  HaArrayAddLast(drs->solve_resources, dr);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverAddDayRange(KHE_DYNAMIC_RESOURCE_SOLVER drs,*/
/*    int first_day_index, int last_day_index)                               */
/*                                                                           */
/*  Add day range first_day_index .. last_day_index to the solve.            */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverAddDayRange(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int first_day_index, int last_day_index)
{
  KHE_DRS_DAY_RANGE si;
  HnAssert(first_day_index <= last_day_index,
    "KheDynamicResourceSolverAddResource: empty day interval");
  si = KheDrsDayRangeMake(first_day_index, last_day_index);
  KheDrsDayRangeMerge(&drs->solve_day_ranges, si);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayAddNode(KHE_DRS_DAY day, KHE_DRS_NODE node,                */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add node to day, or do nothing if node has a hard constraint violation   */
/*  or contains too few or too many assignments.  Otherwise what to do       */
/*  depends on whether we are using strong dominance or weak dominance.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayAddNode(KHE_DRS_DAY day, KHE_DRS_NODE node,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* ***
  ** KHE_DRS_NODE other_node; **  int asst_count ** , i **;
  asst_count = HaArrayFirst(node->signature);
  if( asst_count + HaArrayCount(drs->days) - day->day_index < drs->min_assts ||
      asst_count > drs->max_assts || node->cost > drs->cost_limit )
  {
    ** do nothing except free node; it is unacceptable **
    KheDrsNodeDelete(node, drs);
  }
  else
  *** */

  KheDrsNodeSetAddNode(day->node_set, node, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheFindContinuingTask(KHE_DRS_NODE node, int i, KHE_DRS_TASK *dt)   */
/*                                                                           */
/*  If there is a task for the resource at index i running beyond the last   */
/*  day of node, return true and set *dt to that task.                       */
/*                                                                           */
/*****************************************************************************/

static bool KheFindContinuingTask(KHE_DRS_NODE node, int i, KHE_DRS_TASK *dt)
{
  return *dt = NULL, false;  /* placeholder */
  /* *** the idea of this function is still good, but new version still to do
  KHE_DRS_TASK dt2;
  if( node->prev_edge == NULL )
    return *dt = NULL, false;
  dt2 = HaArray(node->prev_edge->tasks, i);
  if( KheDailyScheduleLastDayIndex(dt2->daily_schedule) > node->day->day_index )
    return *dt = dt2, true;
  else
    return *dt = NULL, false;
  *** */
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheTaskGroupContainsSuitableTask(KHE_DRS_TASK_GROUP dtg,            */
/*    KHE_DRS_RESOURCE dr, KHE_DRS_EDGE de, KHE_DRS_TASK *dt)                */
/*                                                                           */
/*  If dtg contains a task to which dr can be assigned and which is not      */
/*  already assigned to a resource in de, return true and set *dt to the     */
/*  first such task.  Otherwise return false with *dt set to NULL.           */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete (except for domain check)
static bool KheTaskGroupContainsSuitableTask(KHE_DRS_TASK_GROUP dtg,
  KHE_DRS_RESOURCE dr, KHE_DRS_EDGE de, KHE_DRS_TASK *dt)
{
  int i, pos;  KHE_DRS_TASK dt2;
  if( HaArrayCount(dtg->solve_tasks) > 0 )
  {
    dt2 = HaArrayFirst(dtg->solve_tasks);
    if( KheResourceGroupContains(KheTaskDomain(dt2->task), dr->resource) )
    {
      ** dtg's tasks have a suitable domain; now find one not in de **
      HaArrayForEach(dtg->solve_tasks, dt2, i)
	if( !HaArrayContains(de->tasks, dt2, &pos) )
	  return *dt = dt2, true;
    }
  }

  ** no luck **
  return *dt = NULL, false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverClear(KHE_DYNAMIC_RESOURCE_SOLVER drs)      */
/*                                                                           */
/*  Clear drs, ready for a fresh solve.                                      */
/*                                                                           */
/*****************************************************************************/

static void KheDynamicResourceSolverClear(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i;  KHE_DRS_DAY day;  /* KHE_DRS_EXPR e; */
  drs->diversifier = 0;
  HaArrayForEach(drs->days, day, i)
    KheDrsDayClear(day, drs);
  HaArrayClear(drs->final_solns);
  /* ***
  HaArrayForEach(drs->exprs, e, i)
    KheDrsExprDelete(e, drs);
  HaArrayClear(drs->exprs);
  *** */
  /* KheDrsExprCostSumDelete(drs->root_expr, drs); */
  /* drs->root_expr = KheDrsExprCostSumMake(drs); */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsInitializeForSolving(KHE_DYNAMIC_RESOURCE_SOLVER drs,         */
/*    KHE_RESOURCE r, KHE_DRS_DOM_KIND dom_kind, int min_assts,              */
/*    int max_assts, KHE_COST cost_limit)                                    */
/*                                                                           */
/*  Initialize drs for a solve with these attributes.                        */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete
static void KheDrsInitializeForSolving(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_RESOURCE r, KHE_DRS_DOM_KIND dom_kind, int min_assts,
  int max_assts, KHE_COST cost_limit)
{
  int di, i;  KHE_MONITOR m;  KHE_DRS_DAY day;
  KHE_DRS_TASK preassigned_dt;  KHE_DRS_EXPR e;

  ** if there was already a resource, clear out old stuff **
  if( drs->resource != NULL )
    KheDynamicResourceSolverClear(drs);

  ** add the resource, and the expression tree representing the monitors **
  drs->resource = r;
  drs->dom_kind = dom_kind;
  drs->min_assts = min_assts;
  drs->max_assts = max_assts;
  drs->cost_limit = cost_limit;
  KheSolnResourceMonitorSort(drs->soln, r, &KheMonitorLargestLimitFirstCmp);
  if( DEBUG13 )
    fprintf(stderr, "  [ monitor limits:\n");
  for( i = 0;  i < KheSolnResourceMonitorCount(drs->soln, r);  i++ )
  {
    m = KheSolnResourceMonitor(drs->soln, r, i);
    if( DEBUG13 )
    {
      fprintf(stderr, "    limit %d: ", KheMonitorLimit(m));
      KheMonitorDebug(m, 2, 0, stderr);
    }
    KheDrsResourceAddMonitor(drs, m);
  }
  if( DEBUG13 )
    fprintf(stderr, "  ]\n");

  ** sort out the day info for the expression tree **
  HaArrayForEach(drs->exprs, e, i)
  {
    KheDrsExprSetDayRange(e, drs);
    KheDrsExprSetSigIndexes(e, drs);
    KheDrsExprAddExpressionsToDays(e, drs);
  }
  ** ***
  KheDrsExprSetDayRange((KHE_DRS_EXPR) drs->root_expr, 0, drs);
  KheDrsExprSetSigIndexes((KHE_DRS_EXPR) drs->root_expr, drs);
  KheDrsExprAddExpressionsToDays((KHE_DRS_EXPR) drs->root_expr, drs);
  *** **

  ** add suitable tasks for each day, skipping days with preassigned tasks **
  HaArrayForEach(drs->days, day, di)
    if( KheDrsDayAddTasks(day, drs, &preassigned_dt) )
      di = KheDailyScheduleLastDayIndex(preassigned_dt->daily_schedule);
           ** continue from after here **

  ** work out the minimum cost reduction over all selected tasks **
  ** *** sti ll to do
  drs->min_cost_reduction = KheCost(INT_MAX, INT_MAX);
  HaArrayForEach(drs->days, day, di)
    HaArrayForEach(day->times, stime, i)
      HaArrayForEach(stime->best_tasks, dt, j)
        if( !dt->already_assigned &&
	    dt->cost_reduction < drs->min_cost_reduction )
	  drs->min_cost_reduction = dt->cost_reduction;
  *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDynamicResourceSolverSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs,      */
/*    KHE_DRS_DOM_KIND dom_kind, int min_assts, int max_assts,               */
/*    KHE_COST cost_limit, bool preserve_existing, KHE_COST *cost)           */
/*                                                                           */
/*  Solve drs.  If a solution is found, return true and set *cost to its     */
/*  cost.                                                                    */
/*                                                                           */
/*****************************************************************************/

bool KheDynamicResourceSolverSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_DRS_DOM_KIND dom_kind, KHE_COST *cost)
{
  int i, j, k, sig_offset, val, di, prev_di;  KHE_DRS_DAY day;
  KHE_DRS_NODE node;  KHE_DRS_RESOURCE dr;  KHE_DRS_RESOURCE_ON_DAY drd;
  KHE_DRS_NODE_SET dns;  KHE_DRS_DAY_RANGE open_ddr, closed_ddr;

  if( DEBUG1 )
    fprintf(stderr, "[ KheDynamicResourceSolverSolve(drs)\n");

  /* initialize drs and its solve_resources for solving */
  drs->dom_kind = dom_kind;
  /* drs->min_assts = min_assts; */
  /* drs->max_assts = max_assts; */
  /* drs->cost_limit = cost_limit; */
  /* *** replacing this is still to do
  HaArrayForEach(drs->solve_resources, dr, i)
    KheDrsResourceInitForSolving(dr, drs);
  *** */

  /* initialize the day and resource_on_day objects for solving */
  HaArrayForEach(drs->days, day, i)
  {
    KheDrsDayInitializeForSolving(day, drs);

    /* initialize day->dom_tests and day->eq_dom_test_indexes to a single EQ */
    HaArrayClear(day->dom_tests);
    HaArrayAddLast(day->dom_tests, KheDrsDomTestMake(KHE_DRS_DOM_EQ));
    HaArrayClear(day->eq_dom_test_indexes);
    HaArrayAddLast(day->eq_dom_test_indexes, 0);

    /* extend day->dom_tests and day->eq_dom_test_indexes from the resources */
    sig_offset = 1;  /* to leave room for the test just added */
    HaArrayForEach(drs->solve_resources, dr, j)
    {
      drd = HaArray(dr->days, i);
      drd->sig_offset = sig_offset;
      HaArrayAppend(day->dom_tests, drd->dom_tests, k);
      HaArrayForEach(drd->eq_dom_test_indexes, val, k)
	HaArrayAddLast(day->eq_dom_test_indexes, val + sig_offset);
      sig_offset += drd->sig_len;
    }
  }

  if( DEBUG5 )
    KheDynamicResourceSolverDebug(drs, 1, 2, stderr);
  if( DEBUG10 )
    HaArrayForEach(drs->days, day, i)
      KheDrsDayDomTestsDebug(day, 1, 2, stderr);

  /* build up the nodes for each day */
  dns = KheDrsNodeSetMake(drs);
  node = KheDrsNodeMakeRoot(drs);
  KheDrsNodeSetAddNode(dns, node, drs);
  prev_di = -1;
  HaArrayForEach(drs->solve_day_ranges, open_ddr, i)
  {
    /* closed day range [prev_di + 1, ddr.first_day_index - 1] */
    if( prev_di < open_ddr.first_day_index - 1 )
    {
      closed_ddr = KheDrsDayRangeMake(prev_di + 1, open_ddr.first_day_index-1);
      KheDrsNodeSetExtendClosed(dns, closed_ddr, drs);
      day = HaArray(drs->days, open_ddr.first_day_index - 1);
      dns = day->node_set;
    }

    /* open day range [ddr.first_day_index, ddr.last_day_index] */
    for( di = open_ddr.first_day_index;  di <= open_ddr.last_day_index;  di++ )
    {
      KheDrsNodeSetExtendOpen(dns, di, drs);
      day = HaArray(drs->days, di);
      dns = day->node_set;
    }
    prev_di = open_ddr.last_day_index;
  }

  /* final closed day range [prev_di + 1, last_day] */
  if( prev_di < HaArrayCount(drs->days) - 1 )
  {
    closed_ddr = KheDrsDayRangeMake(prev_di + 1, HaArrayCount(drs->days) - 1);
    KheDrsNodeSetExtendClosed(dns, closed_ddr, drs);
    day = HaArray(drs->days, HaArrayCount(drs->days) - 1);
    dns = day->node_set;
  }

  /* *** old version
  HaArrayForEach(drs->days, day, day_index)
  {
    if( prev_day == NULL )
    {
      ** virtual prev_day contains a single node (an empty one) **
      if( DEBUG2 )
	fprintf(stderr, "  extending from day 0 (1 node)\n");
      KheDrsNodeExtendOpen(node, day_index, drs);
    }
    else
    {
      ** extend each node from prev_day with tasks from day **
      KheDrsNodeSetExtendOpen(prev_day->node_set, day_index, drs);
    }
    prev_day = day;
  }
  *** */

/* *** moved out of loop just above to make things clearer for now
#if TESTING
      if( DEBUG7 )
	fprintf(stderr, "  running_time_per_day = %d, days = %d, di = %d\n",
	  HaArrayCount(drs->running_time_per_day), HaArrayCount(drs->days),
	  day_index);
      HaArray(drs->running_time_per_day, day_index) =
	KheTimerElapsedTime(drs->timer);
#endif
      if( DEBUG2 )
	fprintf(stderr, "  extending from day %s (%d nodes)\n",
	  KheTimeGroupId(prev_day->time_group),
	  prev_day->node_set->count);
*** */

  /* extract final solutions whose assignment count is in the desired range */
  day = HaArrayLast(drs->days);
  KheDrsNodeSetExtractFinalSolns(day->node_set, drs);
  HaArraySort(drs->final_solns, &KheDrsNodeIncreasingAsstCountCmp);
  if( DEBUG2 )
  {
    fprintf(stderr, "  final solutions:\n");
    HaArrayForEach(drs->final_solns, node, i)
      KheDrsNodeDebug(node, drs, 1, 2, stderr);
  }

  /* clear out the resources and intervals */
  HaArrayForEach(drs->solve_resources, dr, i)
    KheDrsResourceEndSolving(dr);
  HaArrayClear(drs->solve_resources);
  HaArrayClear(drs->solve_day_ranges);
  drs->solve_state = KHE_DRS_SOLVE_SOLVED;
  if( DEBUG1 )
    fprintf(stderr, "] KheDynamicResourceSolverSolve returning\n");
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDynamicResourceSolverTimetableCount(                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return the number of timetables found.                                   */
/*                                                                           */
/*****************************************************************************/

int KheDynamicResourceSolverTimetableCount(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HnAssert(drs->solve_state == KHE_DRS_SOLVE_SOLVED,
    "KheDynamicResourceSolverTimetableCount called out of order");
  return HaArrayCount(drs->final_solns);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverTimetable(KHE_DYNAMIC_RESOURCE_SOLVER drs,  */
/*    int i, int *asst_count, KHE_COST *r_cost, KHE_COST *soln_cost)         */
/*                                                                           */
/*  Report on the i'th timetable found by the solver.                        */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverTimetable(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int i, int *asst_count, KHE_COST *cost)
{
  /* make sure there is an i'th solution */
  KHE_DRS_NODE node;
  HnAssert(drs->solve_state == KHE_DRS_SOLVE_SOLVED,
    "KheDynamicResourceSolverTimetable called out of order");
  HnAssert(0 <= i && i < HaArrayCount(drs->final_solns),
    "KheDynamicResourceSolverTimetable:  i (%d) out of range (0 .. %d)",
    i, HaArrayCount(drs->final_solns) - 1);

  /* return the relevant fields of the i'th solution */
  node = HaArray(drs->final_solns, i);
  *asst_count = HaArrayFirst(node->signature);
  *cost = node->cost;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAdoptedSolnDebug(KHE_DYNAMIC_RESOURCE_SOLVER drs,             */
/*    KHE_COST *cost)                                                        */
/*                                                                           */
/*  Debug print of r's timetable and costs.                                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAdoptedSolnDebug(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_COST *cost)
{
  int i, ri;  KHE_COST m_cost;  KHE_MONITOR m;  KHE_DRS_RESOURCE dr;
  KHE_RESOURCE_TIMETABLE_MONITOR rtm;  KHE_RESOURCE r;
  *cost = KheCost(0, 0);
  HaArrayForEach(drs->solve_resources, dr, ri)
  {
    rtm = KheResourceTimetableMonitor(drs->soln, dr->resource);
    KheResourceTimetableMonitorPrintTimetable(rtm, 6, 2, stderr);
    for( i = 0;  i < KheSolnResourceMonitorCount(drs->soln, r);  i++ )
    {
      m = KheSolnResourceMonitor(drs->soln, r, i);
      switch( KheMonitorTag(m) )
      {
	case KHE_AVOID_UNAVAILABLE_TIMES_MONITOR_TAG:
	case KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG:
	case KHE_LIMIT_BUSY_TIMES_MONITOR_TAG:
	case KHE_LIMIT_WORKLOAD_MONITOR_TAG:
	case KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG:

	  m_cost = KheMonitorCost(m);
	  if( m_cost > 0 )
	  {
	    fprintf(stderr, "  true cost of %s is %.5f\n", KheMonitorId(m),
	      KheCostShow(m_cost));
	    *cost += m_cost;
	  }
	  break;

	default:

	  /* ignore monitors of other types */
	  break;
      }
    }
  }
  fprintf(stderr, "  total cost is %.5f\n", KheCostShow(*cost));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverAdopt(KHE_DYNAMIC_RESOURCE_SOLVER drs, int i)*/
/*                                                                           */
/*  Change drs's solution to include the i'th timetable.                     */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverAdopt(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* make sure there is an i'th solution */
  KHE_DRS_NODE final_node, rerun_node;  KHE_COST cost;
  HnAssert(drs->solve_state == KHE_DRS_SOLVE_SOLVED,
    "KheDynamicResourceSolverAdopt called out of order");
  /* ***
  HnAssert(0 <= i && i < HaArrayCount(drs->final_solns),
    "KheDynamicResourceSolverAdopt:  i (%d) out of range (0 .. %d)",
    i, HaArrayCount(drs->final_solns) - 1);
  *** */

  /* load the i'th solution, and adopt it */
  final_node = HaArray(drs->final_solns, 0);
  /* *** still to do
  KheDrsFullSolnLoad(drs->full_soln, final_node, drs->resource);
  KheDrsFullSolnAdopt(drs->full_soln);
  *** */

  /* check that the resulting resource costs agree with the final soln */
  if( DEBUG3 )
  {
    fprintf(stderr, "[ KheDynamicResourceSolverAdopt(drs, %d)\n", 0);
    KheDrsAdoptedSolnDebug(drs, &cost);
    if( cost != final_node->cost )
    {
      fprintf(stderr, "KheDynamicResourceSolverAdopt: cost discrepancy "
	"(true %.5f != dynamic %.5f):\n", KheCostShow(cost),
	KheCostShow(final_node->cost));
      /* *** still to do
      KheDrsFullSolnDebug(drs->full_soln, 1, 2, stderr);
      KheDrsFullSolnUnAdopt(drs->full_soln);
      rerun_node = KheFullSolnReRun(drs->full_soln, drs);
      *** */
      rerun_node = NULL;  /* still to do */
      fprintf(stderr, "  rerun_node cost is %.5f\n", KheCostShow(rerun_node->cost));
      HnAbort("aborting now (cost discrepancy)");
    }
    fprintf(stderr, "  ?? assts %d, cost %.5f\n",
      HaArrayFirst(final_node->signature), KheCostShow(cost));
    /* ***
    fprintf(stderr, "  %s assts %d, cost %.5f\n", KheResourceId(drs->resource),
      HaArrayFirst(final_node->signature), KheCostShow(cost));
    *** */
    fprintf(stderr, "] KheDynamicResourceSolverAdopt returning\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverUnAdopt(KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Unadopt the most recently adopted full solution.                         */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverUnAdopt(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* *** still to do
  KheDrsFullSolnUnAdopt(drs->full_soln);
  *** */
}


/*****************************************************************************/
/*                                                                           */
/*  int KheNodeAsstCount(KHE_DRS_NODE node)                                  */
/*                                                                           */
/*  Return the assignment count of node.  (First element of the signature.)  */
/*                                                                           */
/*****************************************************************************/

static int KheNodeAsstCount(KHE_DRS_NODE node)
{
  return HaArrayFirst(node->signature);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_COST KheNodeAdjustedCost(KHE_DRS_NODE node, KHE_COST c)              */
/*                                                                           */
/*  Return the adjusted cost of node.                                        */
/*                                                                           */
/*****************************************************************************/

static KHE_COST KheNodeAdjustedCost(KHE_DRS_NODE node, KHE_COST c)
{
  return node->cost - KheNodeAsstCount(node) * c;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDynamicResourceSolverBest(KHE_DYNAMIC_RESOURCE_SOLVER drs,        */
/*    KHE_COST cost_reduction)                                               */
/*                                                                           */
/*  Return the index of the final solution which has a mminimal value of     */
/*                                                                           */
/*    r_cost - asst_count * cost_reduction                                   */
/*                                                                           */
/*  Break ties in favour of final solutions with larger asst_count values.   */
/*                                                                           */
/*****************************************************************************/

int KheDynamicResourceSolverBest(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_COST cost_reduction)
{
  int best_i, i, asst_count, best_asst_count;  KHE_DRS_NODE node, best_node;
  KHE_COST adjusted_cost, best_adjusted_cost;  
  HnAssert(drs->solve_state == KHE_DRS_SOLVE_SOLVED,
    "KheDynamicResourceSolverBest called out of order");
  HnAssert(HaArrayCount(drs->final_solns) > 0,
    "KheDynamicResourceSolverBest:  there are no solutions");
  best_node = NULL;
  best_i = -1;
  best_adjusted_cost = 0;
  best_asst_count = 0;
  HaArrayForEach(drs->final_solns, node, i)
  {
    asst_count = KheNodeAsstCount(node);
    adjusted_cost = KheNodeAdjustedCost(node, cost_reduction);
    if( best_node == NULL || adjusted_cost < best_adjusted_cost ||
	(adjusted_cost == best_adjusted_cost && asst_count > best_asst_count) )
    {
      best_node = node;
      best_i = i;
      best_adjusted_cost = adjusted_cost;
      best_asst_count = asst_count;
    }
  }
  return best_i;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverDebug(KHE_DYNAMIC_RESOURCE_SOLVER drs,      */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug printf of drs onto fp with the given verbosity and indent.         */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverDebug(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  KHE_DRS_DAY day;  KHE_DRS_RESOURCE dr;  int i;
  HnAssert(indent >= 0, "KheDynamicResourceSolverDebug: indent < 0");
  fprintf(fp, "%*s[ DynamicResourceSolver(", indent, "");
  HaArrayForEach(drs->solve_resources, dr, i)
  {
    if( i > 0 )
      fprintf(fp, ", ");
    fprintf(fp, "%s", KheResourceId(dr->resource));
  }
  fprintf(fp, ")\n");
  HaArrayForEach(drs->days, day, i)
    KheDrsDayDebug(day, drs, verbosity, indent + 2, fp);
  fprintf(fp, "%*s]\n", indent, "");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "manual testing"                                               */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_FULL_SOLN KheDrsManualSolnBegin(KHE_DYNAMIC_RESOURCE_SOLVER drs, */
/*    char *resource_id)                                                     */
/*                                                                           */
/*  Begin loading a manual solution for this resource.                       */
/*                                                                           */
/*****************************************************************************/

#if TESTING
/* ***
static KHE_DRS_FULL_SOLN KheDrsManualSolnBegin(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  char *resource_id)
{
  KHE_INSTANCE ins;  KHE_RESOURCE r;  KHE_DRS_FULL_SOLN res;
  ins = KheSolnInstance(drs->soln);
  if( !KheInstanceRetrieveResource(ins, resource_id, &r) )
    return NULL;
  res = KheDrsFullSolnMake(drs);
  res->resource = r;
  return res;
}
*** */
#endif


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsManualSolnAddTask(KHE_DRS_FULL_SOLN fs, char *time_id,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add the best task for time_id to fs.                                     */
/*                                                                           */
/*****************************************************************************/

#if TESTING
/* ***
static void KheDrsManualSolnAddTask(KHE_DRS_FULL_SOLN fs, char *time_id,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_INSTANCE ins;  KHE_TIME time;  int day_index;  KHE_DRS_DAY day;
  KHE_DRS_TIME t;  int i;  KHE_DRS_TASK best_dt;
  ins = KheSolnInstance(drs->soln);
  if( !KheInstanceRetrieveTime(ins, time_id, &time) )
    HnAbort("KheDrsManualSolnAddTask:  unknown time %s", time_id);
  day_index = KheFrameTimeIndex(drs->days_frame, time);
  day = HaArray(drs->days, day_index);
  HaArrayForEach(day->times, t, i)
    if( t->time == time && HaArrayCount(t->best_tasks) > 0 )
    {
      best_dt = HaArrayFirst(t->best_tasks);
      if( KheDailyScheduleFirstDayIndex(best_dt->daily_schedule) ==
	  KheDailyScheduleLastDayIndex(best_dt->daily_schedule) )
      {
	HaArrayAddLast(fs->tasks, best_dt);
	return;
      }
    }
  HnAbort("KheDrsManualSolnAddTask:  no best task at time %s", time_id);
}
*** */
#endif


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsManualSolnTest1(KHE_DYNAMIC_RESOURCE_SOLVER drs)              */
/*                                                                           */
/*  Carry out a specific manual test, if the instance is INRC2-4-030-1-6291. */
/*                                                                           */
/*****************************************************************************/

#if TESTING
/* ***
static void KheDrsManualSolnTest1(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_INSTANCE ins;  KHE_DRS_FULL_SOLN fs;  KHE_COST cost;
  ins = KheSolnInstance(drs->soln);
  if( strcmp(KheInstanceId(ins), "INRC2-4-030-1-6291") == 0 )
  {
    fprintf(stderr, "[ KheDrsManualSolnTest1(%s, %s)\n",
      KheInstanceId(ins), "HN_0");
    fs = KheDrsManualSolnBegin(drs, "HN_0");
    KheDrsInitializeForSolving(drs, fs->resource, false, 0, INT_MAX,
      KheCost(1, 0));
    KheDrsManualSolnAddTask(fs, "1Wed1", drs);
    KheDrsManualSolnAddTask(fs, "1Thu1", drs);
    KheDrsManualSolnAddTask(fs, "1Fri1", drs);

    KheDrsManualSolnAddTask(fs, "2Mon2", drs);
    KheDrsManualSolnAddTask(fs, "2Tue2", drs);
    KheDrsManualSolnAddTask(fs, "2Wed2", drs);
    KheDrsManualSolnAddTask(fs, "2Thu2", drs);
    KheDrsManualSolnAddTask(fs, "2Fri2", drs);

    KheDrsManualSolnAddTask(fs, "3Mon1", drs);
    KheDrsManualSolnAddTask(fs, "3Tue1", drs);
    KheDrsManualSolnAddTask(fs, "3Wed1", drs);
    KheDrsManualSolnAddTask(fs, "3Thu3", drs);
    KheDrsManualSolnAddTask(fs, "3Fri3", drs);

    KheDrsManualSolnAddTask(fs, "4Mon2", drs);
    KheDrsManualSolnAddTask(fs, "4Tue2", drs);
    KheDrsManualSolnAddTask(fs, "4Wed2", drs);
    KheDrsManualSolnAddTask(fs, "4Thu2", drs);
    KheDrsManualSolnAddTask(fs, "4Fri2", drs);
    KheDrsFullSolnAdopt(fs);
    KheDrsAdoptedSolnDebug(drs, &cost);
    KheDrsFullSolnUnAdopt(fs);
    fprintf(stderr, "] KheDrsManualSolnTest1 returning\n");
  }
}
*** */
#endif


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsManualSolnTest2(KHE_DYNAMIC_RESOURCE_SOLVER drs)              */
/*                                                                           */
/*  Carry out a specific manual test, if the instance is INRC2-4-030-1-6291. */
/*                                                                           */
/*****************************************************************************/

#if TESTING
/* ***
static void KheDrsManualSolnTest2(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_INSTANCE ins;  KHE_DRS_FULL_SOLN fs;  KHE_COST cost;
  ins = KheSolnInstance(drs->soln);
  if( strcmp(KheInstanceId(ins), "INRC2-4-030-1-6291") == 0 )
  {
    fprintf(stderr, "[ KheDrsManualSolnTest2(%s, %s)\n",
      KheInstanceId(ins), "HN_0");
    fs = KheDrsManualSolnBegin(drs, "HN_0");
    KheDrsInitializeForSolving(drs, fs->resource, false, 0, INT_MAX,
      KheCost(1, 0));
    KheDrsManualSolnAddTask(fs, "1Wed1", drs);
    KheDrsManualSolnAddTask(fs, "1Thu1", drs);
    KheDrsManualSolnAddTask(fs, "1Fri1", drs);

    KheDrsManualSolnAddTask(fs, "2Mon2", drs);
    KheDrsManualSolnAddTask(fs, "2Tue2", drs);
    KheDrsManualSolnAddTask(fs, "2Wed2", drs);
    KheDrsManualSolnAddTask(fs, "2Thu2", drs);
    KheDrsManualSolnAddTask(fs, "2Fri2", drs);

    KheDrsManualSolnAddTask(fs, "3Mon1", drs);
    KheDrsManualSolnAddTask(fs, "3Tue1", drs);
    KheDrsManualSolnAddTask(fs, "3Wed1", drs);
    KheDrsManualSolnAddTask(fs, "3Thu1", drs);
    KheDrsManualSolnAddTask(fs, "3Fri1", drs);

    KheDrsManualSolnAddTask(fs, "4Mon2", drs);
    KheDrsManualSolnAddTask(fs, "4Tue2", drs);
    KheDrsManualSolnAddTask(fs, "4Wed2", drs);
    KheDrsManualSolnAddTask(fs, "4Thu2", drs);
    KheDrsManualSolnAddTask(fs, "4Fri2", drs);
    KheDrsFullSolnAdopt(fs);
    KheDrsAdoptedSolnDebug(drs, &cost);
    KheDrsFullSolnUnAdopt(fs);
    fprintf(stderr, "] KheDrsManualSolnTest2 returning\n");
  }
}
*** */
#endif




/*****************************************************************************/
/*                                                                           */
/*  Submodule "testing"                                                      */
/*                                                                           */
/*****************************************************************************/

#if TESTING

typedef HA_ARRAY(KHE_COST) ARRAY_KHE_COST;
#endif


/*****************************************************************************/
/*                                                                           */
/*  char *KheDrsTestTypeShow(KHE_DRS_TEST_TYPE type)                         */
/*                                                                           */
/*  Show a test type.                                                        */
/*                                                                           */
/*****************************************************************************/

#if TESTING
/* ***
static char *KheDrsTestTypeShow(KHE_DRS_TEST_TYPE type)
{
  switch( type )
  {
    case KHE_DRS_TEST_WEAK:		return "w";
    case KHE_DRS_TEST_STRONG:		return "s";
    case KHE_DRS_TEST_WEAK_AND_STRONG:	return "ws";
    default:				return "?";
  }
}
*** */
#endif


/*****************************************************************************/
/*                                                                           */
/*  void TestName(KHE_INSTANCE ins, KHE_RESOURCE r, KHE_DRS_TEST_TYPE type,  */
/*    int min_assts, int max_assts, ARRAY_KHE_COST *cost_limits,char *suffix,*/
/*    char buff[200])                                                        */
/*  void TestName(KHE_INSTANCE ins, KHE_RESOURCE r, KHE_DRS_TEST_TYPE type,  */
/*    int min_assts, int max_assts, char *suffix, char buff[200])            */
/*                                                                           */
/*  Set buff to the name of a test with these attributes.                    */
/*                                                                           */
/*****************************************************************************/

#if TESTING
static void TestName(KHE_INSTANCE ins, char *suffix, char buff[200])
{
  int len;

  /* add standard prefix to buff */
  len = snprintf(buff, 200, "%s", KheInstanceId(ins));

  /* add suffix */
  if( strcmp(suffix, "") != 0 )
    len += snprintf(&buff[len], 200 - len, "-%s", suffix);
}
#endif


/*****************************************************************************/
/*                                                                           */
/*  KHE_DATASET_POINTS_TYPE KheDomKindToPointsType(KHE_DRS_DOM_KIND dom_kind)*/
/*                                                                           */
/*  Return the dataset points type corresponding to dom_kind.                */
/*                                                                           */
/*****************************************************************************/

static KHE_DATASET_POINTS_TYPE KheDomKindToPointsType(KHE_DRS_DOM_KIND dom_kind)
{
  switch( dom_kind )
  {
    case KHE_DRS_DOM_WEAK:	return KHE_DATASET_POINTS_SQUARE;
    case KHE_DRS_DOM_MEDIUM:	return KHE_DATASET_POINTS_CIRCLE;
    case KHE_DRS_DOM_STRONG:	return KHE_DATASET_POINTS_PLUS;
    case KHE_DRS_DOM_TRIE:	return KHE_DATASET_POINTS_CROSS;

    default:

      HnAbort("KheDomKindToPointsType: unknown dom_kind (%d)", dom_kind);
      return KHE_DATASET_POINTS_CROSS;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DATASET_PAIRS_TYPE KheDomKindToPairsType(KHE_DRS_DOM_KIND dom_kind)  */
/*                                                                           */
/*  Return the dataset pairs type corresponding to dom_kind.                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DATASET_PAIRS_TYPE KheDomKindToPairsType(KHE_DRS_DOM_KIND dom_kind)
{
  switch( dom_kind )
  {
    case KHE_DRS_DOM_WEAK:	return KHE_DATASET_PAIRS_DOTTED;
    case KHE_DRS_DOM_MEDIUM:	return KHE_DATASET_PAIRS_DASHED;
    case KHE_DRS_DOM_STRONG:	return KHE_DATASET_PAIRS_SOLID;
    case KHE_DRS_DOM_TRIE:	return KHE_DATASET_PAIRS_DOT_DASHED;

    default:

      HnAbort("KheDomKindToPairsType: unknown dom_kind (%d)", dom_kind);
      return KHE_DATASET_PAIRS_SOLID;  ** keep compiler happy **
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  char *KheDomKindLabel(KHE_DRS_DOM_KIND dom_kind)                         */
/*                                                                           */
/*  Return a suitable label for dom_kind.                                    */
/*                                                                           */
/*****************************************************************************/

static char *KheDomKindLabel(KHE_DRS_DOM_KIND dom_kind)
{
  switch( dom_kind )
  {
    case KHE_DRS_DOM_WEAK:	return "Weak";
    case KHE_DRS_DOM_MEDIUM:	return "Medium";
    case KHE_DRS_DOM_STRONG:	return "Strong";
    case KHE_DRS_DOM_TRIE:	return "Trie";

    default:

      HnAbort("KheDomKindLabel: unknown dom_kind (%d)", dom_kind);
      return NULL;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsPrintPoint(KHE_DYNAMIC_RESOURCE_SOLVER drs, int i)            */
/*                                                                           */
/*  Return true if we want to print a point with index i.                    */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsPrintPoint(KHE_DYNAMIC_RESOURCE_SOLVER drs, int i)
{
  return HaArrayCount(drs->days) <= 28 || i % 2 == 0 ||
    i == HaArrayCount(drs->days) - 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolveTest(KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_RESOURCE r,    */
/*    KHE_DRS_DOM_KIND dom_kind, int min_assts, int max_assts,               */
/*    ARRAY_KHE_COST *cost_limits, KHE_GRAPH time_graph,                     */
/*    KHE_GRAPH size_graph, KHE_GRAPH cost_graph)                            */
/*                                                                           */
/*  Carry out a test with these attributes, and add a dataset to time_graph, */
/*  size_graph, and cost_graph recording the results.                        */
/*                                                                           */
/*****************************************************************************/

#if TESTING
static void KheDrsSolveTest(KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_RESOURCE r,
  KHE_DRS_DOM_KIND dom_kind, KHE_GRAPH time_graph,
  KHE_GRAPH size_graph, KHE_GRAPH cost_graph)
{
  int i, best_i, asst_count;  float val;  KHE_DATASET kd;  KHE_DRS_DAY day;
  KHE_COST r_cost, marginal_cost;  KHE_BALANCE_SOLVER bs;
  KHE_DATASET_POINTS_TYPE points_type; KHE_DATASET_PAIRS_TYPE pairs_type;
  char *label;  KHE_DRS_NODE node;  char buff[200];  KHE_COST cost;
  if( DEBUG11 )
  {
    TestName(KheSolnInstance(drs->soln), KheDomKindLabel(dom_kind), buff);
    fprintf(stderr, "[ KheDrsSolveTest(%s)\n", buff);
  }

  /* solve, measuring running time */
  HaArrayClear(drs->running_time_per_day);
  HaArrayFill(drs->running_time_per_day, HaArrayCount(drs->days), 0.0);
  KheTimerResetStartTime(drs->timer);
  KheDynamicResourceSolverSolve(drs, dom_kind, &cost);

  /* verify costs */
  /* *** still to do
  for( i = 0;  i < KheDynamicResourceSolverTimetableCount(drs);  i++ )
  {
    KheDynamicResourceSolverAdopt(drs, i);
    KheDynamicResourceSolverUnAdopt(drs);
  }
  *** */

  if( DEBUG11 )
  {
    bs = KheBalanceSolverMake(drs->soln, drs->options,
      KheResourceResourceType(r), drs->arena);
    KheBalanceSolverDebug(bs, 2, 2, stderr);
    marginal_cost = KheBalanceSolverMarginalCost(bs);
    fprintf(stderr, "  marginal cost = %.5f\n", KheCostShow(marginal_cost));
    if( KheDynamicResourceSolverTimetableCount(drs) > 0 )
    {
      best_i = KheDynamicResourceSolverBest(drs, marginal_cost);
      fprintf(stderr, "  best of best has index %d\n", best_i);
    }
    HaArrayForEach(drs->final_solns, node, i)
      fprintf(stderr, "  best soln (%d assts) has cost %.5f%s\n",
	HaArrayFirst(node->signature), KheCostShow(node->cost),
	i == best_i ? " (overall best)" : "");
  }

  /* sort out the points type and pairs type */
  points_type = KheDomKindToPointsType(dom_kind);
  pairs_type = KHE_DATASET_PAIRS_DASHED;
  label = KheDomKindLabel(dom_kind);

  /* add one data set to the running time graph */
  kd = KheDataSetAdd(time_graph, points_type, pairs_type, label);
  HaArrayForEach(drs->running_time_per_day, val, i)
    if( KheDrsPrintPoint(drs, i) )
      KhePointAdd(kd, (float) i + 1, val);

  /* add one data set to the table size graph */
  kd = KheDataSetAdd(size_graph, points_type, pairs_type, label);
  KhePointAdd(kd, 0.0, 0.0);
  HaArrayForEach(drs->days, day, i)
    if( KheDrsPrintPoint(drs, i) )
      KhePointAdd(kd, (float) i + 1, day->node_set->count);

  /* add one data set to the cost graph */
  kd = KheDataSetAdd(cost_graph, points_type, pairs_type, label);
  for( i = 0;  i < KheDynamicResourceSolverTimetableCount(drs);  i++ )
  {
    KheDynamicResourceSolverTimetable(drs, i, &asst_count, &r_cost);
    KhePointAdd(kd, (float) asst_count, (float) KheSoftCost(r_cost));
  }

  if( DEBUG11 )
    fprintf(stderr, "] KheDrsSolveTest returning\n");
}
#endif


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverDoTest(KHE_DYNAMIC_RESOURCE_SOLVER drs,     */
/*    KHE_RESOURCE r, KHE_DRS_TEST_TYPE test_type, int min_assts,            */
/*    int max_assts, ARRAY_KHE_COST *cost_limits)                            */
/*                                                                           */
/*  Do a test with these values for min_assts, max_assts, and cost_limits.   */
/*                                                                           */
/*****************************************************************************/

#if TESTING
static void KheDynamicResourceSolverDoTest(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_RESOURCE r)
{
  char time_file_name[200], size_file_name[200], cost_file_name[200];
  KHE_FILE time_file, size_file, cost_file;
  KHE_GRAPH time_graph, size_graph, cost_graph;

  /* initialize the time file and graph */
  TestName(KheSolnInstance(drs->soln), "time", time_file_name);
  time_file = KheFileBegin(time_file_name, KHE_FILE_LOUT_STANDALONE);
  time_graph = KheGraphBegin(time_file);
  KheGraphSetBelowCaption(time_graph, "Days");
  KheGraphSetLeftCaptionAndGap(time_graph, "Secs", "1.0c");

  /* initialize the table size file and graph */
  TestName(KheSolnInstance(drs->soln), "size", size_file_name);
  size_file = KheFileBegin(size_file_name, KHE_FILE_LOUT_STANDALONE);
  size_graph = KheGraphBegin(size_file);
  KheGraphSetBelowCaption(size_graph, "Days");
  KheGraphSetLeftCaptionAndGap(size_graph, "Size", "1.0c");

  /* initialize the cost file and graph */
  TestName(KheSolnInstance(drs->soln), "cost", cost_file_name);
  cost_file = KheFileBegin(cost_file_name, KHE_FILE_LOUT_STANDALONE);
  cost_graph = KheGraphBegin(cost_file);
  KheGraphSetBelowCaption(cost_graph, "Assignments");
  KheGraphSetLeftCaptionAndGap(cost_graph, "Cost", "1.0c");

  /* test with strong, medium, and weak dominance */
  KheDrsSolveTest(drs, r, KHE_DRS_DOM_WEAK,
    time_graph, size_graph, cost_graph);
  KheDrsSolveTest(drs, r, KHE_DRS_DOM_MEDIUM,
    time_graph, size_graph, cost_graph);
  KheDrsSolveTest(drs, r, KHE_DRS_DOM_STRONG,
    time_graph, size_graph, cost_graph);
  KheDrsSolveTest(drs, r, KHE_DRS_DOM_TRIE,
    time_graph, size_graph, cost_graph);

  /* tidy up graphs and stats files */
  KheGraphEnd(time_graph);
  KheFileEnd(time_file);
  KheGraphEnd(size_graph);
  KheFileEnd(size_file);
  KheGraphEnd(cost_graph);
  KheFileEnd(cost_file);
}
#endif


/*****************************************************************************/
/*                                                                           */
/*  bool KheMinConstraintWeight(KHE_SOLN soln, KHE_RESOURCE r,               */
/*    KHE_COST *min_weight)                                                  */
/*                                                                           */
/*  If r has at least one constraint whose cost will be optimized and        */
/*  whose weight is non-negative, return the minimum of those weights.       */
/*                                                                           */
/*****************************************************************************/

static bool KheMinConstraintWeight(KHE_SOLN soln, KHE_RESOURCE r,
  KHE_COST *min_weight)
{
  int i;  KHE_CONSTRAINT c;  KHE_COST weight;
  *min_weight = KheCost(INT_MAX, INT_MAX);
  for( i = 0;  i < KheResourceConstraintCount(r);  i++ )
  {
    c = KheResourceConstraint(r, i);
    switch( KheConstraintTag(c) )
    {
      case KHE_AVOID_UNAVAILABLE_TIMES_CONSTRAINT_TAG:
      case KHE_CLUSTER_BUSY_TIMES_CONSTRAINT_TAG:
      case KHE_LIMIT_BUSY_TIMES_CONSTRAINT_TAG:
      case KHE_LIMIT_WORKLOAD_CONSTRAINT_TAG:
      case KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT_TAG:

	weight = KheConstraintCombinedWeight(c);
	if( weight > 0 && weight < *min_weight )
	  *min_weight = weight;
	break;

      default:

	/* ignore other constraints */
	break;
    }
  }
  return *min_weight < KheCost(INT_MAX, INT_MAX);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverTest(KHE_SOLN soln,                         */
/*    KHE_RESOURCE_TYPE rt, KHE_OPTIONS options, KHE_RESOURCE r1,            */
/*    KHE_RESOURCE r2, KHE_RESOURCE r3, KHE_RESOURCE r4,                     */
/*    int first_day_index1, int last_day_index1,                             */
/*    int first_day_index2, int last_day_index2,                             */
/*    int first_day_index3, int last_day_index3,                             */
/*    int first_day_index4, int last_day_index4)                             */
/*                                                                           */
/*  Test dynamic resource solving on r1, r2, r3, and r4.  Some may be NULL,  */
/*  allowing testing on fewer than 4 resources.                              */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverTest(KHE_SOLN soln,
  KHE_RESOURCE_TYPE rt, KHE_OPTIONS options, KHE_RESOURCE r1,
  KHE_RESOURCE r2, KHE_RESOURCE r3, KHE_RESOURCE r4,
  int first_day_index1, int last_day_index1,
  int first_day_index2, int last_day_index2,
  int first_day_index3, int last_day_index3,
  int first_day_index4, int last_day_index4)
{
#if TESTING
  KHE_DYNAMIC_RESOURCE_SOLVER drs;  int max_busy_times;  KHE_RESOURCE r;

  /* make a solver */
  r = r1;  /* temporary hack, still to do */
  drs = KheDynamicResourceSolverMake(soln, rt, options);

  /* do the test */
  KheDynamicResourceSolverDoTest(drs, r);
  if( KheResourceMaxBusyTimes(drs->soln, r, &max_busy_times) )
  {
    /* unlimited cost, limited busy days range */
    KheDynamicResourceSolverDoTest(drs, r);
  }

  /* manual tests and return */
  /* ***
  if( DEBUG8 )
  {
    KheDrsManualSolnTest1(drs);
    KheDrsManualSolnTest2(drs);
  }
  *** */
  KheDynamicResourceSolverDelete(drs);
#else
  fprintf(stderr, "to test, set TESTING to 1 in khe_sr_dynamic_resource.c"\n);
#endif
}
