
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sr_dynamic_resource.c                                  */
/*  DESCRIPTION:  Dynamic resource solving                                   */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"
#include "howard_p.h"
#include <limits.h>
#include <float.h>
#include <math.h>
#define bool_show(x) ((x) ? "true" : "false")
#define min(x, y) ((x) < (y) ? (x) : (y))
#define max(x, y) ((x) > (y) ? (x) : (y))

#define UNDEF_INT -1
#define UNDEF_FLOAT -1.0

#define DEBUG1 1		/* main functions		*/
#define DEBUG2 0		/* final solutions		*/
#define DEBUG3 0		/* checking adopted solutions	*/
#define DEBUG4 0		/* KheDrsSolnMake		*/
#define DEBUG5 0		/* KheDynamicResourceSolverDebug*/
#define DEBUG6 0		/* KheDrsFullSolnAdopt		*/
#define DEBUG7 0		/* running time (testing only)	*/
#define DEBUG8 0		/* manual tests               	*/
#define DEBUG9 0		/* hash table efficiency      	*/
#define DEBUG10 0		/* day dom tests	      	*/
#define DEBUG11 1		/* brief print of costs		*/
#define DEBUG12 0		/* trie dominance tests 	*/
#define DEBUG13 0		/* monitor sorting              */
#define DEBUG14 1		/* cluster minimum solving      */
#define DEBUG15 1		/* task classes                 */
#define DEBUG16 1		/* monitor framer               */

/* ***
#define DEBUG_SOLN(soln) (HaArrayFirst((soln)->signature) == 1 &&	\
  (soln)->day->day_index == 27)
*** */

#define TESTING 1

/*****************************************************************************/
/*                                                                           */
/*  Submodule "type declarations for expressions, etc."                      */
/*                                                                           */
/*****************************************************************************/

/* some forward declarations */
typedef struct khe_drs_resource_on_day_rec *KHE_DRS_RESOURCE_ON_DAY;
typedef struct khe_drs_resource_rec *KHE_DRS_RESOURCE;
typedef struct khe_drs_task_on_day_rec *KHE_DRS_TASK_ON_DAY;
typedef struct khe_drs_task_rec *KHE_DRS_TASK;
typedef struct khe_drs_task_class_rec *KHE_DRS_TASK_CLASS;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_DAY_RANGE - an interval of days                             */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_day_range_rec {
  int			first;
  int			last;
} KHE_DRS_DAY_RANGE;

typedef HA_ARRAY(KHE_DRS_DAY_RANGE) ARRAY_KHE_DRS_DAY_RANGE;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_DOM_KIND - the kind of dominance testing being used         */
/*                                                                           */
/*****************************************************************************/

/* *** defined in khe_solvers.h
typedef enum {
  KHE_DRS_DOM_NONE,
  KHE_DRS_DOM_WEAK,
  KHE_DRS_DOM_MEDIUM,
  KHE_DRS_DOM_STRONG,
  KHE_DRS_DOM_TRIE,
} KHE_DRS_DOM_KIND;
*** */


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_DOM_TYPE - the type of a dominance test                     */
/*                                                                           */
/*****************************************************************************/

typedef enum {
  KHE_DRS_DOM_UNUSED,		/* "-"      should never be used      */
  KHE_DRS_DOM_GE,		/* ">="     min limit only            */
  KHE_DRS_DOM_LE,		/* "<="     max limit only            */
  KHE_DRS_DOM_EQ,		/* "==",    for uncertain cases       */
  KHE_DRS_DOM_GE_LOOSE,		/* ">=*"    min limit with loosening  */
  KHE_DRS_DOM_EQ_LOOSE		/* "==*"    max + min with loosening  */
} KHE_DRS_DOM_TYPE;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_DOM_TEST - a dominance test                                 */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_dom_test_rec {
  KHE_DRS_DOM_TYPE	type;			/* type of test              */
  int			min_limit;		/* GE_LOOSE and EQ_LOOSE     */
} KHE_DRS_DOM_TEST;

typedef HA_ARRAY(KHE_DRS_DOM_TEST) ARRAY_KHE_DRS_DOM_TEST;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_TAG                                                    */
/*                                                                           */
/*  Tags for expression types.                                               */
/*                                                                           */
/*****************************************************************************/

typedef enum {

  /* subtypes of KHE_DRS_EXPR_EXTERNAL */
  KHE_DRS_EXPR_BUSY_TIME_TAG,		/* int-valued	     */
  KHE_DRS_EXPR_FREE_TIME_TAG,		/* int-valued	     */
  KHE_DRS_EXPR_WORK_TIME_TAG,		/* float-valued      */
  KHE_DRS_EXPR_BUSY_DAY_TAG,		/* int-valued	     */
  KHE_DRS_EXPR_FREE_DAY_TAG,		/* int-valued	     */
  KHE_DRS_EXPR_WORK_DAY_TAG,		/* float-valued      */
  KHE_DRS_EXPR_ASSIGNED_TASK_TAG,	/* int-valued        */

  /* subtypes of KHE_DRS_EXPR_INTERNAL */
  KHE_DRS_EXPR_OR_TAG,			/* int-valued	     */
  KHE_DRS_EXPR_AND_TAG,			/* int-valued	     */
  KHE_DRS_EXPR_INT_SUM_TAG,		/* int-valued	     */
  /* KHE_DRS_EXPR_INT_SEQ_TAG, */
  KHE_DRS_EXPR_FLOAT_SUM_TAG,		/* float-valued      */
  KHE_DRS_EXPR_INT_DEV_TAG,
  KHE_DRS_EXPR_FLOAT_DEV_TAG,
  KHE_DRS_EXPR_COST_TAG,
  /* KHE_DRS_EXPR_COST_SUM_TAG, */
  KHE_DRS_EXPR_INT_SUM_COMB_TAG,	/* int-valued	     */
  KHE_DRS_EXPR_INT_SEQ_COMB_TAG		/* int-valued	     */
} KHE_DRS_EXPR_TAG;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_COMB_TYPE - types of comb expressions                       */
/*                                                                           */
/*****************************************************************************/

typedef enum {
  KHE_DRS_COMB_ORDINARY,
  KHE_DRS_COMB_NO_MAX,
  KHE_DRS_COMB_LINEAR,
  KHE_DRS_COMB_STEP
} KHE_DRS_COMB_TYPE;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_VALUE - the value of an expression; could be int or float   */
/*                                                                           */
/*****************************************************************************/

typedef union {
  int			int_val;		/* value, if int   */
  float			float_val;		/* value, if float */
} KHE_DRS_VALUE;


/*****************************************************************************/
/*                                                                           */
/*  INHERIT_KHE_DRS_EXPR - fields inherited by all expression types          */
/*                                                                           */
/*  For each active day di of e, except its very last active day,            */
/*  sig_index[di - e->first_day_index] is the position of e's value in       */
/*  the signatures of solutions ending on day di.                            */
/*                                                                           */
/*  For each active day di of e, child_index[di - e->first_day_index] is     */
/*  the index of the first child whose last_day_index is di or more.  There  */
/*  is also a sentinel value, so for each active day di, the iterator        */
/*                                                                           */
/*    start_child_index = child_index[di - e->first_day_index];              */
/*    stop_child_index  = child_index[di + 1 - e->first_day_index];          */
/*    for( i = start_child_index;  i < stop_child_index;  i++ )              */
/*      child_e = HaArray(e->children, i);                                   */
/*                                                                           */
/*  visits the children of e whose last_day_index is di.                     */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_rec *KHE_DRS_EXPR;
typedef HA_ARRAY(KHE_DRS_EXPR) ARRAY_KHE_DRS_EXPR;

typedef struct khe_drs_parent_rec {
  KHE_DRS_EXPR		expr;			/* the parent expression */
  int			index;			/* index in parent       */
} KHE_DRS_PARENT;

typedef HA_ARRAY(KHE_DRS_PARENT) ARRAY_KHE_DRS_PARENT;

#define INHERIT_KHE_DRS_EXPR						\
  KHE_DRS_EXPR_TAG	tag;			/* tag field      */	\
  bool			gathered;		/* for opening    */	\
  KHE_DRS_DOM_TEST	dom_test;		/* dom test       */	\
  int			postorder_index;	/* expr index     */	\
  KHE_DRS_VALUE		u;			/* value          */	\
  ARRAY_KHE_DRS_PARENT	parents;		/* parent exprs   */	\
  ARRAY_KHE_DRS_EXPR	children;		/* child exprs    */	\
  /* KHE_DRS_DAY_RANGE	day_range; */		/* day range      */	\
  /* bool		open; */		/* open for solve */	\
  KHE_DRS_DAY_RANGE	open_day_range;		/* open day range */	\
  ARRAY_KHE_DRS_EXPR	open_children;		/* open children  */	\
  HA_ARRAY_INT		open_child_indexes;	/* for each day   */	\
  /* int		first_day_index; */	/* first day      */	\
  /* int		last_day_index; */	/* last day       */	\
  HA_ARRAY_INT		sig_indexes;		/* for each day   */


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_INT - abstract supertype of integer-valued exprs       */
/*                                                                           */
/*****************************************************************************/

/* ***
#define INHERIT_KHE_DRS_EXPR_INT					\
  INHERIT_KHE_DRS_EXPR							\
  int				value;		** stored value **

typedef struct khe_drs_expr_int_rec {
  INHERIT_KHE_DRS_EXPR_INT
} *KHE_DRS_EXPR_INT;
*** */


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_FLOAT - abstract supertype of float-valued exprs       */
/*                                                                           */
/*****************************************************************************/

/* ***
#define INHERIT_KHE_DRS_EXPR_FLOAT					\
  INHERIT_KHE_DRS_EXPR							\
  float				value;		** stored value **

typedef struct khe_drs_expr_float_rec {
  INHERIT_KHE_DRS_EXPR_FLOAT
} *KHE_DRS_EXPR_FLOAT;
*** */


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_BUSY_TIME                                              */
/*                                                                           */
/*  An expression whose value is 1 when the resource is busy at a given      */
/*  time, and 0 when the resource is free then.  The expression stores the   */
/*  time, but only for debugging.                                            */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_busy_time_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_RESOURCE_ON_DAY	resource_on_day;  /* the resource and day    */
  KHE_TIME			time;		/* the time we're busy at    */
} *KHE_DRS_EXPR_BUSY_TIME;

typedef HA_ARRAY(KHE_DRS_EXPR_BUSY_TIME) ARRAY_KHE_DRS_EXPR_BUSY_TIME;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_FREE_TIME                                              */
/*                                                                           */
/*  An expression whose value is 1 when the resource is free at a given      */
/*  time, and 0 when the resource is busy then.  The expression stores the   */
/*  time, but only for debugging.                                            */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_free_time_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_RESOURCE_ON_DAY	resource_on_day;  /* the resource and day    */
  KHE_TIME			time;		/* the time we're free at    */
} *KHE_DRS_EXPR_FREE_TIME;

typedef HA_ARRAY(KHE_DRS_EXPR_FREE_TIME) ARRAY_KHE_DRS_EXPR_FREE_TIME;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_WORK_TIME                                              */
/*                                                                           */
/*  An expression whose value is the workload incurred by the resource at    */
/*  a given time.  This will be 0.0 if the resource is free then.  The       */
/*  expression stores the time, but only for debugging.                      */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_work_time_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_RESOURCE_ON_DAY	resource_on_day;  /* the resource and day    */
  KHE_TIME			time;		/* the time of this workload */
} *KHE_DRS_EXPR_WORK_TIME;

typedef HA_ARRAY(KHE_DRS_EXPR_WORK_TIME) ARRAY_KHE_DRS_EXPR_WORK_TIME;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_BUSY_DAY                                               */
/*                                                                           */
/*  An expression whose value is 1 when the resource is busy on a given      */
/*  day, and 0 when the resource is free then.  The expression stores the    */
/*  day (as a time group), but only for debugging.                           */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_busy_day_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_RESOURCE_ON_DAY	resource_on_day;  /* the resource and day    */
  KHE_TIME_GROUP		time_group;	/* the day we're busy at     */
} *KHE_DRS_EXPR_BUSY_DAY;

typedef HA_ARRAY(KHE_DRS_EXPR_BUSY_DAY) ARRAY_KHE_DRS_EXPR_BUSY_DAY;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_FREE_DAY                                               */
/*                                                                           */
/*  An expression whose value is 1 when the resource is free on a given      */
/*  day, and 0 when the resource is busy then.  The expression stores the    */
/*  day (as a time group), but only for debugging.                           */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_free_day_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_RESOURCE_ON_DAY	resource_on_day;  /* the resource and day    */
  KHE_TIME_GROUP		time_group;	/* the day we're free at     */
} *KHE_DRS_EXPR_FREE_DAY;

typedef HA_ARRAY(KHE_DRS_EXPR_FREE_DAY) ARRAY_KHE_DRS_EXPR_FREE_DAY;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_WORK_DAY                                               */
/*                                                                           */
/*  An expression whose value is the workload incurred by the resource on    */
/*  a given day.  This will be 0.0 if the resource is free then.  The        */
/*  expression stores the day (as a time group), but only for debugging.     */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_work_day_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_RESOURCE_ON_DAY	resource_on_day;  /* the resource and day    */
  KHE_TIME_GROUP		time_group;	/* the day we're free at     */
} *KHE_DRS_EXPR_WORK_DAY;

typedef HA_ARRAY(KHE_DRS_EXPR_WORK_DAY) ARRAY_KHE_DRS_EXPR_WORK_DAY;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_ASSIGNED_TASK                                          */
/*                                                                           */
/*  An expression derived from an atomic task, whose value is 1 if the task  */
/*  is assigned a resource from the given resource group, and 0 otherwise.   */
/*                                                                           */
/*  We include the task because we are going to merge equal expressions of   */
/*  this kind, and that is only possible when the task is in the expression. */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_assigned_task_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_DRS_TASK_ON_DAY		task_on_day;
  KHE_RESOURCE_GROUP		resource_group;
} *KHE_DRS_EXPR_ASSIGNED_TASK;

typedef HA_ARRAY(KHE_DRS_EXPR_ASSIGNED_TASK) ARRAY_KHE_DRS_EXPR_ASSIGNED_TASK;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_OR                                                     */
/*                                                                           */
/*  An expression whose value is the logical `or' of its children's values.  */
/*  Here Booleans are represented by integers (0 for false, 1 for true).     */
/*                                                                           */
/*  The closed state here is the number of closed children with value 1.     */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_or_rec {
  INHERIT_KHE_DRS_EXPR
  int				closed_state;
} *KHE_DRS_EXPR_OR;

typedef HA_ARRAY(KHE_DRS_EXPR_OR) ARRAY_KHE_DRS_EXPR_OR;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_AND                                                    */
/*                                                                           */
/*  An expression whose value is the logical `and' of its children's values. */
/*  Here Booleans are represented by integers (0 for false, 1 for true).     */
/*                                                                           */
/*  The closed state here is the number of closed children with value 0.     */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_and_rec {
  INHERIT_KHE_DRS_EXPR
  int				closed_state;
} *KHE_DRS_EXPR_AND;

typedef HA_ARRAY(KHE_DRS_EXPR_AND) ARRAY_KHE_DRS_EXPR_AND;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_INT_SUM                                                */
/*                                                                           */
/*  An expression whose value is the sum of its children's values.           */
/*                                                                           */
/*  The closed state here is the sum of the values of the closed children.   */
/*  We may be able to use just the value field to hold this, but if so that  */
/*  is still to do.                                                          */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_int_sum_rec {
  INHERIT_KHE_DRS_EXPR
  int				closed_state;
} *KHE_DRS_EXPR_INT_SUM;

typedef HA_ARRAY(KHE_DRS_EXPR_INT_SUM) ARRAY_KHE_DRS_EXPR_INT_SUM;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_FLOAT_SUM                                              */
/*                                                                           */
/*  An expression whose value is the sum of its children's values.  The      */
/*  expression's value, and its children's values, have type float.          */
/*                                                                           */
/*  The closed state here is the sum of the values of the closed children.   */
/*  We may be able to use just the value field to hold this, but if so that  */
/*  is still to do.                                                          */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_float_sum_rec {
  INHERIT_KHE_DRS_EXPR
  float				closed_state;
} *KHE_DRS_EXPR_FLOAT_SUM;

typedef HA_ARRAY(KHE_DRS_EXPR_FLOAT_SUM) ARRAY_KHE_DRS_EXPR_FLOAT_SUM;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_INT_DEV                                                */
/*                                                                           */
/*  An expression whose value is the deviation of its integer-valued child.  */
/*                                                                           */
/*  The closed state here is the value of the closed child.  There is no     */
/*  need to store it explicitly.                                             */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_int_dev_rec {
  INHERIT_KHE_DRS_EXPR
  int				min_limit;	/* minimum limit             */
  int				max_limit;	/* maximum limit             */
  bool				allow_zero;	/* true if 0 has deviation 0 */
  /* int			closed_state; */
} *KHE_DRS_EXPR_INT_DEV;

typedef HA_ARRAY(KHE_DRS_EXPR_INT_DEV) ARRAY_KHE_DRS_EXPR_INT_DEV;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_FLOAT_DEV                                              */
/*                                                                           */
/*  An expression whose value is the integer-valued deviation of its         */
/*  float-valued child.                                                      */
/*                                                                           */
/*  The closed state here is the value of the closed child.  There is no     */
/*  need to store it explicitly.                                             */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_float_dev_rec {
  INHERIT_KHE_DRS_EXPR
  int				min_limit;	/* minimum limit             */
  int				max_limit;	/* maximum limit             */
  bool				allow_zero;	/* true if 0 has deviation 0 */
  /* float			closed_state; */
} *KHE_DRS_EXPR_FLOAT_DEV;

typedef HA_ARRAY(KHE_DRS_EXPR_FLOAT_DEV) ARRAY_KHE_DRS_EXPR_FLOAT_DEV;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_COST                                                   */
/*                                                                           */
/*  An expression whose value is a cost based on its sole child's deviation. */
/*                                                                           */
/*  The closed state here is the value of the closed child.                  */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_cost_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_MONITOR			monitor;	/* the monitor (debug only)  */
  KHE_COST_FUNCTION		cost_fn;	/* the cost function         */
  KHE_COST			combined_weight; /* the combined weight      */
  /* int			closed_state; */
} *KHE_DRS_EXPR_COST;

typedef HA_ARRAY(KHE_DRS_EXPR_COST) ARRAY_KHE_DRS_EXPR_COST;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_COST_SUM                                               */
/*                                                                           */
/*  An expression whose value is the sum of the values of its children.      */
/*  All these values are costs.                                              */
/*                                                                           */
/*****************************************************************************/

/* ***
typedef struct khe_drs_expr_cost_sum_rec {
  INHERIT_KHE_DRS_EXPR
} *KHE_DRS_EXPR_COST_SUM;

typedef HA_ARRAY(KHE_DRS_EXPR_COST_SUM) ARRAY_KHE_DRS_EXPR_COST_SUM;
*** */


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_INT_SUM_COMB                                           */
/*                                                                           */
/*  Combines INT_SUM, INT_DEV, and COST into a single expression.  This      */
/*  allows important optimizations in some cases, depending on comb_type.    */
/*                                                                           */
/*  The closed state here is the sum of the values of the closed children.   */
/*  We may be able to use just the value field to hold this, but if so that  */
/*  is still to do.                                                          */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_int_sum_comb_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_MONITOR			monitor;	/* the monitor (debug only)  */
  KHE_COST_FUNCTION		cost_fn;	/* the cost function         */
  KHE_COST			combined_weight; /* the combined weight      */
  int				min_limit;	/* minimum limit             */
  int				max_limit;	/* maximum limit             */
  bool				allow_zero;	/* true if 0 has deviation 0 */
  KHE_DRS_COMB_TYPE		comb_type;	/* type of combined value    */
  int				history_before;	/* history value before      */
  int				history_after;	/* history value after       */
  int				closed_state;
} *KHE_DRS_EXPR_INT_SUM_COMB;

typedef HA_ARRAY(KHE_DRS_EXPR_INT_SUM_COMB) ARRAY_KHE_DRS_EXPR_INT_SUM_COMB;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR_INT_SEQ_COMB                                           */
/*                                                                           */
/*  Combines INT_SEQ, INT_DEV (without allow_zero), and COST into a single   */
/*  expression, for the same optimizations as KHE_DRS_EXPR_INT_SUM_COMB.     */
/*                                                                           */
/*  The closed state here will be a complex record of the state of the       */
/*  closed intervals.  This is very much still to do.                        */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_expr_int_seq_comb_rec {
  INHERIT_KHE_DRS_EXPR
  KHE_MONITOR			monitor;	/* the monitor (debug only)  */
  KHE_COST_FUNCTION		cost_fn;	/* the cost function         */
  KHE_COST			combined_weight; /* the combined weight      */
  int				min_limit;	/* minimum limit             */
  int				max_limit;	/* maximum limit             */
  KHE_DRS_COMB_TYPE		comb_type;	/* type of combined value    */
  int				history_before;	/* history value before      */
  int				history_after;	/* history value after       */
} *KHE_DRS_EXPR_INT_SEQ_COMB;

typedef HA_ARRAY(KHE_DRS_EXPR_INT_SEQ_COMB) ARRAY_KHE_DRS_EXPR_INT_SEQ_COMB;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EXPR - abstract supertype of expressions                    */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_expr_rec {
  INHERIT_KHE_DRS_EXPR
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "type declarations for resources, tasks, etc."                 */
/*                                                                           */
/*****************************************************************************/

typedef HA_ARRAY(KHE_DRS_RESOURCE) ARRAY_KHE_DRS_RESOURCE;
typedef HA_ARRAY(KHE_DRS_RESOURCE_ON_DAY) ARRAY_KHE_DRS_RESOURCE_ON_DAY;
typedef HA_ARRAY(KHE_DRS_TASK_ON_DAY) ARRAY_KHE_DRS_TASK_ON_DAY;
typedef HA_ARRAY(KHE_DRS_TASK) ARRAY_KHE_DRS_TASK;
typedef HA_ARRAY(KHE_DRS_TASK_CLASS) ARRAY_KHE_DRS_TASK_CLASS;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_RESOURCE_ON_DAY - what one resource is doing on one day     */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_day_rec *KHE_DRS_DAY;

struct khe_drs_resource_on_day_rec {
  KHE_DRS_RESOURCE		encl_dr;		/* the resource      */
  KHE_DRS_DAY			day;			/* the day           */
  /* bool			open; */		/* open for solving  */
  KHE_DRS_TASK_ON_DAY		closed_asst;		/* asst when closed  */
  ARRAY_KHE_DRS_EXPR		external_today;		/* leaf today        */
};


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_RESOURCE - one resource                                     */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_resource_rec {
  KHE_RESOURCE			resource;		/* the resource      */
  /* int			open_resource_index; */	/* index when open   */
  ARRAY_KHE_DRS_RESOURCE_ON_DAY	days;			/* doing each day    */
};


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_TASK_ON_DAY - what one proper root task is doing on one day */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_task_on_day_rec {
  KHE_DRS_TASK			encl_dt;		/* enclosing task    */
  KHE_DRS_DAY			day;			/* the day           */
  KHE_TASK			task;			/* atomic task today */
  KHE_TIME			time;			/* time on day       */
  KHE_DRS_RESOURCE_ON_DAY	closed_asst;		/* asst when closed  */
  ARRAY_KHE_DRS_EXPR		external_today;		/* leaf today        */
};


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_TASK - one proper root task                                 */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_task_rec {
  KHE_DRS_TASK_CLASS		encl_dtc;
  int				index_in_encl_dtc;
  bool				open;			/* true when open    */
  KHE_TASK			task;
  KHE_DRS_RESOURCE		closed_asst;		/* asst when closed  */
  ARRAY_KHE_DRS_TASK_ON_DAY	days;
};


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_TASK_CLASS - a set of tasks, taken from one KHE_TASK_CLASS  */
/*                                                                           */
/*  For efficiency, open_tasks is managed lazily.  Every unassigned task is  */
/*  in it, but not int order, and there may also be assigned tasks in it.    */
/*  These defects are corrected when the task class is opened for solving.   */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_task_class_rec {
  KHE_TASK_CLASS		orig_task_class;	/* origin of this    */
  KHE_DRS_DAY_RANGE		day_range;		/* days running      */
  ARRAY_KHE_DRS_TASK		all_tasks;		/* all tasks in class*/
  ARRAY_KHE_DRS_TASK		open_tasks;		/* unassigned tasks  */
  int				tmp_used;		/* used by curr edge */
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "type declarations for edges, solutions, days, and solvers"    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_EDGE - one edge, connecting one solution to the next        */
/*                                                                           */
/*****************************************************************************/

/* ***
typedef struct khe_drs_soln_rec *KHE_DRS_SOLN;

typedef struct khe_drs_edge_rec {
  KHE_DRS_SOLN			prev_soln;	** previous soln             **
  ** ARRAY_KHE_DRS_TASK		tasks; **	** indexed by resources      **
  ARRAY_KHE_DRS_TASK_ON_DAY	tasks;		** indexed by resources      **
} *KHE_DRS_EDGE;

typedef HA_ARRAY(KHE_DRS_EDGE) ARRAY_KHE_DRS_EDGE;
*** */


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_SOLN - a partial solution, part of the search tree          */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_soln_rec *KHE_DRS_SOLN;

struct khe_drs_soln_rec {
  KHE_DRS_SOLN		prev_soln;		/* edge from prev_soln       */
  KHE_DRS_DAY		day;			/* day containing this soln  */
  ARRAY_KHE_DRS_TASK_ON_DAY prev_tasks;		/* indexed by resources      */
  int			best_of;		/* how many this is best of  */
  KHE_COST		cost;			/* soln cost so far          */
  HA_ARRAY_INT		signature;		/* signature of this soln    */
};

typedef HA_ARRAY(KHE_DRS_SOLN) ARRAY_KHE_DRS_SOLN;
typedef HP_TABLE(KHE_DRS_SOLN) TABLE_KHE_DRS_SOLN;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_SOLN_LIST - a list of solns; has several uses               */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_soln_list_rec {
  ARRAY_KHE_DRS_SOLN	solns;			/* the solns                 */
} *KHE_DRS_SOLN_LIST;

typedef HA_ARRAY(KHE_DRS_SOLN_LIST) ARRAY_KHE_DRS_SOLN_LIST;
typedef HP_TABLE(KHE_DRS_SOLN_LIST) TABLE_KHE_DRS_SOLN_LIST;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_SOLN_TRIE - a trie of solns                                 */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_soln_trie_rec *KHE_DRS_SOLN_TRIE;

typedef HA_ARRAY(KHE_DRS_SOLN_TRIE) ARRAY_KHE_DRS_SOLN_TRIE;

struct khe_drs_soln_trie_rec {
  KHE_DRS_SOLN			soln;
  ARRAY_KHE_DRS_SOLN_TRIE	children;
};


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_SOLN_SET - a set of solns, with a choice of implementations */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_soln_set_rec {
  int				count;			/* number of solns   */
  KHE_DRS_DOM_KIND		dom_kind;		/* kind of dominance */
  union {
    KHE_DRS_SOLN_LIST		none_dom_set;		/* none dominance    */
    TABLE_KHE_DRS_SOLN		weak_dom_set;		/* weak dominance    */
    TABLE_KHE_DRS_SOLN_LIST	medium_dom_set;		/* medium dominance  */
    KHE_DRS_SOLN_LIST		strong_dom_set;		/* strong dominance  */
    KHE_DRS_SOLN_TRIE		trie_dom_set;		/* trie dominance    */
  } u;
} *KHE_DRS_SOLN_SET;

typedef HA_ARRAY(KHE_DRS_SOLN_SET) ARRAY_KHE_DRS_SOLN_SET;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DRS_DAY - one day, open for solving                             */
/*                                                                           */
/*****************************************************************************/

struct khe_drs_day_rec {
  int				frame_index;		/* index in frame    */
  int				open_day_index;		/* index when open   */
  KHE_TIME_GROUP		time_group;		/* day's time group  */
  ARRAY_KHE_DRS_TASK_CLASS	task_classes;		/* today's tasks     */
  ARRAY_KHE_DRS_EXPR		internal_today;		/* active today      */
  ARRAY_KHE_DRS_DOM_TEST	dom_tests;		/* for each sig pos  */
  HA_ARRAY_INT			eq_dom_test_indexes;	/* where = tests are */
  KHE_DRS_SOLN_SET		soln_set;		/* today's solns     */
  /* ***
  int				sig_len;		** length of sig     **
  KHE_DRS_TASK			free_task;		** free on this day  **
  ARRAY_KHE_DRS_TIME		times;			** times of this day **
  ARRAY_KHE_DRS_EXPR		internal_today;		** active on this day**
  ARRAY_KHE_DRS_EXPR		external_today;		** leaf expressions  **
  *** */
};

typedef HA_ARRAY(KHE_DRS_DAY) ARRAY_KHE_DRS_DAY;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DYNAMIC_RESOURCE_SOLVER - a dynamic resource solver             */
/*                                                                           */
/*****************************************************************************/

typedef HA_ARRAY(KHE_MONITOR) ARRAY_KHE_MONITOR;

/* ***
typedef enum khe_drs_solve_state {
  KHE_DRS_SOLVE_LOADING,
  KHE_DRS_SOLVE_SOLVED
} KHE_DRS_SOLVE_STATE;
*** */

struct khe_dynamic_resource_solver_rec {

  /* fields which are constant throughout the lifetime of the solver */
  HA_ARENA			arena;
  KHE_SOLN			soln;
  KHE_RESOURCE_TYPE		resource_type;
  KHE_OPTIONS			options;
  KHE_FRAME			days_frame;
  KHE_EVENT_TIMETABLE_MONITOR	etm;
  KHE_TASK_CLASS_SOLVER		task_class_solver;
  ARRAY_KHE_DRS_RESOURCE	all_resources;
  ARRAY_KHE_DRS_DAY		all_days;
  ARRAY_KHE_DRS_TASK		all_root_tasks;
  ARRAY_KHE_DRS_EXPR		all_root_exprs;	/* roots of expr trees */
  int				postorder_count;

  /* free list fields (edges and solns) */
  /* ARRAY_KHE_DRS_EDGE		free_edge; */
  ARRAY_KHE_DRS_SOLN		free_soln;
  ARRAY_KHE_DRS_SOLN_LIST	free_soln_list;
  ARRAY_KHE_DRS_SOLN_TRIE	free_soln_trie;
  ARRAY_KHE_DRS_SOLN_SET	free_soln_set;

  /* fields that vary with the solve */
  /* KHE_DRS_SOLVE_STATE	solve_state; */
  KHE_RESOURCE_SET		selected_resource_set; /* selected for solve */
  ARRAY_KHE_DRS_DAY_RANGE	selected_day_ranges;   /* selected for solve */
  ARRAY_KHE_DRS_RESOURCE	open_resources;
  ARRAY_KHE_DRS_DAY		open_days;
  ARRAY_KHE_DRS_EXPR		open_exprs;
  KHE_COST			solve_init_cost;
  int				solve_soln_limit;
  int				solve_diversifier;
  bool				solve_rerun;
  /* ARRAY_KHE_DRS_SOLN		solve_final_solns; */
  /* KHE_DRS_SOLN		solve_final_soln; */

  /* fields for recording running time */
#if TESTING
  KHE_TIMER			timer;
  HA_ARRAY_FLOAT		running_time_per_day;
#endif
};

  /* *** obsolete
  KHE_CLUSTER_MINIMUM_SOLVER	cluster_solver;
  int				min_assts;
  int				max_assts;
  KHE_COST			cost_limit;
  ARRAY_KHE_DRS_EXPR		exprs;
  KHE_DRS_FULL_SOLN		full_soln;
  *** */
  /* fields holding temporary information */
  /* ARRAY_KHE_MONITOR		er_monitors; */
  /* KHE_TASK_FINDER		task_finder; */
  /* KHE_MONITOR_FRAMER		monitor_framer; */
  /* KHE_DRS_DOM_KIND		solve_dom_kind;  moved to days */
  /* ARRAY_KHE_DRS_MONITOR_CONVERTER er_monitor_converters; */
  /* free list fields - expressions */
  /* ***
  ARRAY_KHE_DRS_EXPR_BUSY_TIME	free_expr_busy_time;
  ARRAY_KHE_DRS_EXPR_FREE_TIME	free_expr_free_time;
  ARRAY_KHE_DRS_EXPR_WORK_TIME	free_expr_work_time;
  ARRAY_KHE_DRS_EXPR_BUSY_DAY	free_expr_busy_day;
  ARRAY_KHE_DRS_EXPR_FREE_DAY	free_expr_free_day;
  ARRAY_KHE_DRS_EXPR_WORK_DAY	free_expr_work_day;
  ARRAY_KHE_DRS_EXPR_ASSIGNED_TASK free_expr_assigned_task;
  ARRAY_KHE_DRS_EXPR_OR		free_expr_or;
  ARRAY_KHE_DRS_EXPR_AND	free_expr_and;
  ARRAY_KHE_DRS_EXPR_INT_SUM	free_expr_int_sum;
  ARRAY_KHE_DRS_EXPR_FLOAT_SUM	free_expr_float_sum;
  ARRAY_KHE_DRS_EXPR_INT_DEV	free_expr_int_dev;
  ARRAY_KHE_DRS_EXPR_FLOAT_DEV	free_expr_float_dev;
  ARRAY_KHE_DRS_EXPR_COST	free_expr_cost;
  ARRAY_KHE_DRS_EXPR_INT_SUM_COMB free_expr_int_sum_comb;
  ARRAY_KHE_DRS_EXPR_INT_SEQ_COMB free_expr_int_seq_comb;
  *** */

  /* free list fields - resources, tasks, and task classes */
  /* ***
  ARRAY_KHE_DRS_RESOURCE_ON_DAY	free_resource_on_day;
  ARRAY_KHE_DRS_RESOURCE	free_resource;
  ARRAY_KHE_DRS_TASK_ON_DAY	free_task_on_day;
  ARRAY_KHE_DRS_TASK		free_task;
  ARRAY_KHE_DRS_TASK_CLASS	free_task_class;
  *** */
  /* ARRAY_KHE_DRS_DAY		free_day; */
  /* ARRAY_KHE_DRS_MONITOR_CONVERTER free_monitor_converter; */



/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_DAY_RANGE"                                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DAY_RANGE KheDrsDayRangeMake(int first, int last)                */
/*                                                                           */
/*  Return a new interval with these endpoints.                              */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DAY_RANGE KheDrsDayRangeMake(int first, int last)
{
  KHE_DRS_DAY_RANGE res;
  res.first = first;
  res.last = last;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDayRangeIsEmpty(KHE_DRS_DAY_RANGE ddr)                        */
/*                                                                           */
/*  Return true if ddr is empty.                                             */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsDayRangeIsEmpty(KHE_DRS_DAY_RANGE ddr)
{
  return ddr.first > ddr.last;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayRangeAdd(KHE_DRS_DAY_RANGE *ddr, int index)                */
/*                                                                           */
/*  Update *ddr to include index.                                            */
/*                                                                           */
/*  Implementation note.  Both cases are needed here, as proved by a trace   */
/*  of Add([1, 0], 4).  The second case wrongly produces [1, 4].             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayRangeAdd(KHE_DRS_DAY_RANGE *ddr, int index)
{
  if( ddr->first > ddr->last )
  {
    /* ddr is empty, so make it cover just index */
    ddr->first = ddr->last = index;
  }
  else
  {
    /* ddr is non-empty, so expand it to include index */
    if( index < ddr->first )
      ddr->first = index;
    if( index > ddr->last )
      ddr->last = index;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayRangeMerge(ARRAY_KHE_DRS_DAY_RANGE *day_ranges,            */
/*    KHE_DRS_DAY_RANGE ddr)                                                 */
/*                                                                           */
/*  Merge ddr into *day_ranges, a sorted array of day ranges.                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayRangeMerge(ARRAY_KHE_DRS_DAY_RANGE *day_ranges,
  KHE_DRS_DAY_RANGE ddr)
{
  KHE_DRS_DAY_RANGE ddr2;  int i;

  /* if ddr is empty, it changes nothing (this can't happen, actually) */
  if( ddr.first > ddr.last )
    return;

  HaArrayForEach(*day_ranges, ddr2, i)
  {
    /***********************************************************************/
    /*                                                                     */
    /*  Invariant: day_ranges and ddr may have changed, but it remains      */
    /*  true that merging ddr with day_ranges gives the desired result.     */
    /*  Also, day_ranges[0..i-1] precede and do not interact with ddr.      */
    /*                                                                     */
    /***********************************************************************/

    if( ddr2.last + 1 < ddr.first )
    {
      /* ddr2 precedes ddr and does not interact with it; continue */
      continue;
    }

    if( ddr.last + 1 < ddr2.first )
    {
      /* ddr precedes ddr2 and does not interact with it, or with day_ranges */
      HaArrayAdd(*day_ranges, i, ddr);
      return;
    }

    /* ddr2 interacts with ddr, so expand ddr to include ddr2, then */
    /* delete ddr2 and continue */
    if( ddr2.first < ddr.first )
      ddr.first = ddr2.first;
    if( ddr2.last > ddr.last )
      ddr.last = ddr2.last;
    HaArrayDeleteAndShift(*day_ranges, i);
    i--;
  }

  /* ddr does not interact with any day_ranges, and it follows them all */
  HaArrayAddLast(*day_ranges, ddr);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDayRangeContains(KHE_DRS_DAY_RANGE si, int i)                 */
/*                                                                           */
/*  Return true if si contains i.                                            */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDrsDayRangeContains(KHE_DRS_DAY_RANGE si, int i)
{
  return si.first <= i && i <= si.last;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDayRangeSubset(KHE_DRS_DAY_RANGE si1, KHE_DRS_DAY_RANGE si2)  */
/*                                                                           */
/*  Return true if si1 is a subset of si2.                                   */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsDayRangeSubset(KHE_DRS_DAY_RANGE si1, KHE_DRS_DAY_RANGE si2)
{
  return si1.first >= si2.first && si1.last <= si2.last;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDayRangesDebug(ARRAY_KHE_DRS_DAY_RANGE *day_ranges,              */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of *day_ranges onto fp with the given verbosity and indent.  */
/*                                                                           */
/*****************************************************************************/

static void KheDayRangesDebug(ARRAY_KHE_DRS_DAY_RANGE *day_ranges,
  int verbosity, int indent, FILE *fp)
{
  KHE_DRS_DAY_RANGE si;  int i;
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "[");
  HaArrayForEach(*day_ranges, si, i)
  {
    if( i > 0 )
      fprintf(fp, ", ");
    if( si.first == si.last )
      fprintf(fp, "%d", si.first);
    else
      fprintf(fp, "%d-%d", si.first, si.last);
  }
  fprintf(fp, "]");
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_DOM_TYPE and KHE_DRS_DOM_TEST"                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsDomTestMake(KHE_DRS_DOM_TYPE type)                */
/*                                                                           */
/*  Make and return a new dominance test object with no min limit attribute. */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DOM_TEST KheDrsDomTestMake(KHE_DRS_DOM_TYPE type)
{
  KHE_DRS_DOM_TEST res;
  res.type = type;
  res.min_limit = -1;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsDomTestMakeLoose(KHE_DRS_DOM_TYPE type, int val)  */
/*                                                                           */
/*  Make and return a new dominance test object with a min limit attribute.  */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DOM_TEST KheDrsDomTestMakeLoose(KHE_DRS_DOM_TYPE type, int val)
{
  KHE_DRS_DOM_TEST res;
  res.type = type;
  res.min_limit = val;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDomTestEqual(KHE_DRS_DOM_TEST dt1, KHE_DRS_DOM_TEST dt2)      */
/*                                                                           */
/*  Return true if dt1 and dt2 are equal.  This function assumes, as it      */
/*  may, that when the min_limit field is unused its value is always -1.     */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsDomTestEqual(KHE_DRS_DOM_TEST dt1, KHE_DRS_DOM_TEST dt2)
{
  return dt1.type == dt2.type && dt1.min_limit == dt2.min_limit;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDomTestDominates(KHE_DRS_DOM_TEST dt, int val1, int val2)     */
/*                                                                           */
/*  Return true if val1 dominates val2, according to dt.                     */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsDomTestDominates(KHE_DRS_DOM_TEST dt, int val1, int val2)
{
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      /* should never be called */
      HnAbort("KheDrsDomTestDominates internal error (KHE_DRS_DOM_UNUSED)");
      return false;  /* keep compiler happy */

    case KHE_DRS_DOM_GE:

      /* tight min limit */
      return val1 >= val2;

    case KHE_DRS_DOM_LE:

      /* max limit */
      return val1 <= val2;

    case KHE_DRS_DOM_EQ:

      /* values must be equal */
      return val1 == val2;

    case KHE_DRS_DOM_GE_LOOSE:

      /* loose min limit */
      return val1 >= val2 || val1 >= dt.min_limit;

    case KHE_DRS_DOM_EQ_LOOSE:

      /* max and loose min; both conditions must hold (basically equality) */
      return (val1 <= val2) && (val1 >= val2 || val1 >= dt.min_limit);

    default:

      HnAbort("KheDrsDomTestDominates internal error (%d)", dt.type);
      return false;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDomTestDominatingRange(KHE_DRS_DOM_TEST dt, int val,          */
/*    int *from_val, int *to_val)                                            */
/*                                                                           */
/*  Set [*from_val, *to_val] to the range of values that dominate val        */
/*  under dt.  That is, set it to the largest range of values x such that    */
/*                                                                           */
/*    KheDrsDomTestDominates(dt, x, val)                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDomTestDominatingRange(KHE_DRS_DOM_TEST dt, int val,
  int *from_val, int *to_val)
{
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      /* should never be called */
      HnAbort("KheDrsDomTestDominatingRange internal error (DOM_UNUSED)");
      *from_val = *to_val = 0;  /* keep compiler happy */
      break;

    case KHE_DRS_DOM_GE:

      /* largest range of values x such that x >= val */
      *from_val = val, *to_val = INT_MAX;
      break;

    case KHE_DRS_DOM_LE:

      /* largest range of values x such that x <= val */
      *from_val = 0, *to_val = val;
      break;

    case KHE_DRS_DOM_EQ:

      /* largest range of values x such that x == val */
      *from_val = val, *to_val = val;
      break;

    case KHE_DRS_DOM_GE_LOOSE:

      /* largest range of values x such that x >= val || x >= dt.min_limit */
      /* i.e. such that x >= min(val, dt.min_limit) */
      *from_val = min(val, dt.min_limit), *to_val = INT_MAX;
      break;

    case KHE_DRS_DOM_EQ_LOOSE:

      /* largest range of values x st x <= val && x >= min(val, dt.min_limit) */
      *from_val = min(val, dt.min_limit), *to_val = val;
      break;

    default:

      HnAbort("KheDrsDomTestDominatingRange internal error (%d)", dt.type);
      *from_val = *to_val = 0;  /* keep compiler happy */
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDomTestDominatedRange(KHE_DRS_DOM_TEST dt, int val,           */
/*    int *from_val, int *to_val)                                            */
/*                                                                           */
/*  Set [*from_val, *to_val] to the range of values that are dominated by    */
/*  val under dt.  That is, set it to the largest range of values x s.t.     */
/*                                                                           */
/*    KheDrsDomTestDominates(dt, val, x)                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDomTestDominatedRange(KHE_DRS_DOM_TEST dt, int val,
  int *from_val, int *to_val)
{
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      /* should never be called */
      HnAbort("KheDrsDomTestDominatedRange internal error (DOM_UNUSED)");
      *from_val = *to_val = 0;  /* keep compiler happy */
      break;

    case KHE_DRS_DOM_GE:

      /* largest range of values x such that val >= x */
      *from_val = 0, *to_val = val;
      break;

    case KHE_DRS_DOM_LE:

      /* largest range of values x such that val <= x */
      *from_val = val, *to_val = INT_MAX;
      break;

    case KHE_DRS_DOM_EQ:

      /* largest range of values x such that val == x */
      *from_val = val, *to_val = val;
      break;

    case KHE_DRS_DOM_GE_LOOSE:

      /* largest range of values x such that val >= x || val >= dt.min_limit */
      if( val >= dt.min_limit )
        *from_val = 0, *to_val = INT_MAX;
      else
        *from_val = 0, *to_val = val;
      break;

    case KHE_DRS_DOM_EQ_LOOSE:

      /* largest range of x st val <= x && (val >= x || val >= dt.min_limit) */
      if( val >= dt.min_limit )
	*from_val = val, *to_val = INT_MAX;
      else
        *from_val = val, *to_val = val;
      break;

    default:

      HnAbort("KheDrsDomTestDominatedRange internal error (%d)", dt.type);
      *from_val = *to_val = 0;  /* keep compiler happy */
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheDrsDomTestShow(KHE_DRS_DOM_TEST dt, char buff[100])             */
/*                                                                           */
/*  Show dt, using buff for scratch storage.                                 */
/*                                                                           */
/*****************************************************************************/

static char *KheDrsDomTestShow(KHE_DRS_DOM_TEST dt, char buff[100])
{
  switch( dt.type )
  {
    case KHE_DRS_DOM_UNUSED:

      sprintf(buff, "-");
      break;

    case KHE_DRS_DOM_GE:

      sprintf(buff, ">=");
      break;

    case KHE_DRS_DOM_LE:

      sprintf(buff, "<=");
      break;

    case KHE_DRS_DOM_EQ:

      sprintf(buff, "==");
      break;

    case KHE_DRS_DOM_GE_LOOSE:

      sprintf(buff, ">=(%d)", dt.min_limit);
      break;

    case KHE_DRS_DOM_EQ_LOOSE:

      sprintf(buff, "=(%d)", dt.min_limit);
      break;

    default:

      sprintf(buff, "?");
      break;
  }
  return buff;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "helper code for expressions, including deviations and costs"  */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsResourceOnDayHasTimeExpr(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME time,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e);
static bool KheDrsResourceOnDayHasDayExpr(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME_GROUP tg,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e);
static void KheDrsResourceOnDayAddExpr(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_EXPR e);
/* ***
static void KheDrsResourceAddTimeExpr(KHE_DRS_RESOURCE dr,
  KHE_TIME time, KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR e);
static void KheDrsExprFreeInternal(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);
static void KheDrsExprFreeExternal(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);
static void KheDrsResourceAddDayExpr(KHE_DRS_RESOURCE dr,
  KHE_TIME_GROUP tg, KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR e);
*** */
static bool KheDrsExprInSignature(KHE_DRS_EXPR e, KHE_DRS_DAY day, int *si);
/* ***
static char *KheDrsExprDayRangeShow(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, char buff[100]);
*** */
static void KheDrsExprDebugChildren(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp);
static void KheDrsExprDebug(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp);


/*****************************************************************************/
/*                                                                           */
/*  KheDrsExprInitBegin(res, tg, dt, drs, free_list)                         */
/*                                                                           */
/*  Shared generic code for creating an expression object and                */
/*  initializing those fields common to all expressions.                     */
/*                                                                           */
/*****************************************************************************/

#define KheDrsExprInitBegin(res, tg, dt, drs)				\
  HaMake(res, drs->arena);						\
  HaArrayInit(res->parents, drs->arena);				\
  HaArrayInit(res->children, drs->arena);				\
  HaArrayInit(res->open_children, drs->arena);				\
  HaArrayInit(res->open_child_indexes, drs->arena);			\
  HaArrayInit(res->sig_indexes, drs->arena);				\
  res->tag = tg;							\
  res->gathered = false;						\
  res->postorder_index = -1;						\
  res->open_day_range = KheDrsDayRangeMake(1, 0);			\
  res->dom_test = dt

#define OldKheDrsExprInit(res, tg, dt, drs, free_list)			\
  if( HaArrayCount(drs->free_list) > 0 )				\
  {									\
    res = HaArrayLastAndDelete(drs->free_list);				\
    HaArrayClear(res->parents);						\
    HaArrayClear(res->children);					\
    HaArrayClear(res->open_children);					\
    HaArrayClear(res->open_child_indexes);				\
    HaArrayClear(res->sig_indexes);					\
  }									\
  else									\
  {									\
    HaMake(res, drs->arena);						\
    HaArrayInit(res->parents, drs->arena);				\
    HaArrayInit(res->children, drs->arena);				\
    HaArrayInit(res->open_children, drs->arena);			\
    HaArrayInit(res->open_child_indexes, drs->arena);			\
    HaArrayInit(res->sig_indexes, drs->arena);				\
  }									\
  res->tag = tg;							\
  res->postorder_index = -1;						\
  /* res->day_range = KheDrsDayRangeMake(UNDEF_INT, UNDEF_INT); */	\
  /* res->open = false; */						\
  res->open_day_range = KheDrsDayRangeMake(1, 0);			\
  /* res->first_day_index = res->last_day_index = UNDEF_INT; */		\
  res->dom_test = dt


/*****************************************************************************/
/*                                                                           */
/*  KheDrsExprInitEnd(res, drs)                                              */
/*                                                                           */
/*  Shared generic code for ending the intialization of res by setting       */
/*  its postorder index and (closed) value.                                  */
/*                                                                           */
/*****************************************************************************/

#define KheDrsExprInitEnd(res, drs)					\
  (res)->postorder_index = (drs)->postorder_count++,			\
  KheDrsExprSetClosedValue((KHE_DRS_EXPR) (res), (drs))


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsExprOpenDayIsFirst(KHE_DRS_EXPR e, int open_day_index)        */
/*                                                                           */
/*  Return true if open_day_index is the index of the first open day of e.   */
/*                                                                           */
/*****************************************************************************/

#define KheDrsExprOpenDayIsFirst(e, open_day_index)			\
  ((open_day_index) == (e)->open_day_range.first)


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprSigVal(KHE_DRS_EXPR e, int open_day_index,                 */
/*    KHE_DRS_SOLN prev_soln)                                                */
/*                                                                           */
/*  Retrieve from prev_soln the signature of e on open day open_day_index.   */
/*                                                                           */
/*****************************************************************************/

#define KheDrsExprSigVal(e, open_day_index, prev_soln)			\
  HaArray((prev_soln)->signature,					\
    HaArray((e)->sig_indexes, (open_day_index) - (e)->open_day_range.first))


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprForEachOpenDayChild(KHE_DRS_EXPR e, int open_day_index,   */
/*    KHE_DRS_EXPR x, int i)                                                 */
/*                                                                           */
/*  Iterator which visits every open child of e whose last open day is       */
/*  open_day_index, setting x to each child in turn and i to its index       */
/*  in e->open_children.                                                     */
/*                                                                           */
/*****************************************************************************/

#define KheDrsExprForEachOpenDayChild(e, odi, x, i)			      \
  i1 = HaArray((e)->open_child_indexes, (odi)   - (e)->open_day_range.first); \
  i2 = HaArray((e)->open_child_indexes, (odi)+1 - (e)->open_day_range.first); \
  for( (i) = i1;							      \
       (i) < i2 ? ((x) = HaArray((e)->open_children, i), true) : false;	      \
       (i)++ )


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsExprOpenDayIsLast(KHE_DRS_EXPR e, int open_day_index)         */
/*                                                                           */
/*  Return true if open_day_index is the index of the last open day of e.    */
/*                                                                           */
/*****************************************************************************/

#define KheDrsExprOpenDayIsLast(e, open_day_index)			      \
  ((open_day_index) == (e)->open_day_range.last)


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprPutSigVal(KHE_DRS_EXPR e, int open_day_index,             */
/*    KHE_DRS_SOLN next_soln, int sig_val)                                   */
/*                                                                           */
/*  Add sig_val, the signature of e on open_day_index, to next_soln.         */
/*                                                                           */
/*****************************************************************************/

#define KheDrsExprPutSigVal(e, open_day_index, next_soln, val)		      \
  HnAssert(HaArrayCount((next_soln)->signature) ==			      \
    HaArray((e)->sig_indexes, (open_day_index) - (e)->open_day_range.first),  \
    "KheDrsExprPutSigVal internal error"),			      \
  HaArrayAddLast((next_soln)->signature, (val));


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSumDev(int min_limit, int max_limit, bool allow_zero,          */
/*    int history_after, int val)                                            */
/*                                                                           */
/*  Deviation function to use with sum expressions; it has allow_zero.       */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSumDev(int min_limit, int max_limit, bool allow_zero,
  int history_after, int val)
{
  if( val == 0 && allow_zero )
    return 0;
  else if( val > max_limit )
    return val - max_limit;
  else if( val + history_after < min_limit )
    return min_limit - (val + history_after);
  else
    return 0;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSeqDev(int min_limit, int max_limit,                           */
/*    int history_after, int val)                                            */
/*                                                                           */
/*  Deviation function to use with sequence expressions; no allow_zero.      */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSeqDev(int min_limit, int max_limit,
  int history_after, int val)
{
  if( val > max_limit )
    return val - max_limit;
  else if( val + history_after < min_limit )
    return min_limit - (val + history_after);
  else
    return 0;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_COST KheDrsCost(KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight, */
/*    int dev)                                                               */
/*                                                                           */
/*****************************************************************************/

static KHE_COST KheDrsCost(KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,
  int dev)
{
  switch( cost_fn )
  {
    case KHE_STEP_COST_FUNCTION:

        return dev > 0 ? combined_weight : 0;

    case KHE_LINEAR_COST_FUNCTION:

        return dev * combined_weight;

    case KHE_QUADRATIC_COST_FUNCTION:

	return dev * dev * combined_weight;

    default:

	HnAbort("KheDrsCost internal error");
	return KheCost(0, 0);  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_BUSY_TIME"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeSetClosedValue(KHE_DRS_EXPR_BUSY_TIME ebt,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing ebt, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyTimeSetClosedValue(KHE_DRS_EXPR_BUSY_TIME ebt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;
  dtd = ebt->resource_on_day->closed_asst;
  ebt->u.int_val = (dtd != NULL && dtd->time == ebt->time ? 1 : 0);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_BUSY_TIME KheDrsExprBusyTimeMake(                           */
/*    KHE_DRS_RESOURCE_ON_DAY drd, KHE_TIME time,                            */
/*    KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)            */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_BUSY_TIME object with these attributes.          */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/
static void KheDrsExprSetClosedValue(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);

static KHE_DRS_EXPR_BUSY_TIME KheDrsExprBusyTimeMake(
  KHE_DRS_RESOURCE_ON_DAY drd, KHE_TIME time,
  KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_BUSY_TIME res;  KHE_DRS_EXPR e;
  if( KheDrsResourceOnDayHasTimeExpr(drd, KHE_DRS_EXPR_BUSY_TIME_TAG,
	dom_test, time, drs, &e) )
    return (KHE_DRS_EXPR_BUSY_TIME) e;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_BUSY_TIME_TAG, dom_test, drs);
  /* res->u.int_val = 0;  set below, by KheDrsExprBusyTimeSetClosedValue */
  res->resource_on_day = drd;
  res->time = time;
  /* KheDrsAddExprToTime(drs, (KHE_DRS_EXPR) res, time); */
  KheDrsResourceOnDayAddExpr(drd, (KHE_DRS_EXPR) res);
  KheDrsExprInitEnd(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_BUSY_TIME KheDrsExprBusyTimeCopy(KHE_DRS_EXPR_BUSY_TIME ebt,*/
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a fresh and unshared copy of ebt.  Do not copy the sig_indexes    */
/*  and child_indexes arrays; they should be empty anyway.                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_BUSY_TIME KheDrsExprBusyTimeCopy(KHE_DRS_EXPR_BUSY_TIME ebt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_BUSY_TIME res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_BUSY_TIME_TAG, ebt->dom_test, drs);
  /* ***
  res->day_range = ebt->day_range;
  res->first_day_index = ebt->first_day_index;
  res->last_day_index = ebt->last_day_index;
  *** */
  res->u.int_val = ebt->u.int_val;
  res->time = ebt->time;
  KheDrsExprInitEnd(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeFreeExternal(KHE_DRS_EXPR_BUSY_TIME ebt,          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free ebt and its descendants (there are no descendants here).            */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprBusyTimeFreeExternal(KHE_DRS_EXPR_BUSY_TIME ebt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_expr_busy_time, ebt);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeSetValue(KHE_DRS_EXPR_BUSY_TIME ebt,              */
/*    int day_index, KHE_TASK task, KHE_TIME time, KHE_RESOURCE r)           */
/*                                                                           */
/*  Set the value of ebt, given that r (which must be non-NULL and which     */
/*  ebt is concerned with) is assigned to task at time on day day_index.     */
/*                                                                           */
/*  Implementation note.  The condition "time == ebt->time" is true when     */
/*  r is assigned a task which is running at ebt's time.  If r is not        */
/*  assigned a task, time is NULL so the condition is false.                 */
/*                                                                           */
/*****************************************************************************/
/* ***
static void KheDrsExprChildHasClosed(KHE_DRS_EXPR e, KHE_DRS_EXPR child_e,
  int child_index);
*** */

/* ***
static void KheDrsExprBusyTimeSetValue(KHE_DRS_EXPR_BUSY_TIME ebt,
  ** int day_index, ** KHE_TASK task, KHE_TIME time, KHE_RESOURCE r)
{
  ** KHE_DRS_PARENT p;  int i; **
  ebt->u.int_val = (time == ebt->time ? 1 : 0);
  ** ***
  HaArrayForEach(ebt->parents, p, i)
    KheDrsExprChildHasClosed(p.expr, (KHE_DRS_EXPR) ebt, p.index);
  *** **
    ** ***
    KheDrsExprChildHasClosed(p.expr, (KHE_DRS_EXPR) ebt,
      p.index, day_index, ebt->u);
    *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeLeafSet(KHE_DRS_EXPR_BUSY_TIME ebt,               */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Set the value of ebt, given that dtd has just been assigned drd.         */
/*  Here dtd is known to be an actual task, not a non-assignment, and drd    */
/*  is known to be concerned with the same resource as ebt is.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyTimeLeafSet(KHE_DRS_EXPR_BUSY_TIME ebt,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  if( dtd->time == ebt->time )
    ebt->u.int_val = 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeLeafClear(KHE_DRS_EXPR_BUSY_TIME ebt)             */
/*                                                                           */
/*  Inform leaf expression ebt that the assignment of drd to dtd carried     */
/*  out by KheDrsExprBusyTimeLeafSet above has been removed.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyTimeLeafClear(KHE_DRS_EXPR_BUSY_TIME ebt)
{
  ebt->u.int_val = 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeDoDay(KHE_DRS_EXPR_BUSY_TIME ebt, int day_index,  */
/*    KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active expression ebt on day_index.                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprBusyTimeDoDay(KHE_DRS_EXPR_BUSY_TIME ebt, int day_index,
  KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;
  if( day_index == ebt->day_range.first )
  {
    ** this is ebt's first day **
    ebt->u.int_val = (time == ebt->time ? 1 : 0);
  }
  else
  {
    ** this is not ebt's first day, so we need to retrieve a previous value **
    si = HaArray(ebt->sig_indexes, day_index - 1 - ebt->day_range.first);
    ebt->u.int_val = HaArray(prev_soln->signature, prev_so + si);
  }
  if( day_index < ebt->day_range.last )
  {
    ** this not ebt's last day; store value in next_soln **
    si = HaArray(ebt->sig_indexes, day_index - ebt->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == next_so + si,
      "KheDrsExprBusyTimeDoDay internal error");
    HaArrayAddLast(next_soln->signature, ebt->u.int_val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyTimeDebug(KHE_DRS_EXPR_BUSY_TIME ebt,                 */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for busy time expression ebt.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyTimeDebug(KHE_DRS_EXPR_BUSY_TIME ebt,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char /* buff1[100], */ buff2[100];  int si;
  if( soln != NULL )
  {
    /* unindented version with a soln value */
    fprintf(fp, "BUSY_TIME(%s)", KheTimeId(ebt->time));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) ebt, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
  }
  else
  {
    /* indented version without a soln value */
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "[ BUSY_TIME(%s) %s ]", KheTimeId(ebt->time),
      /* KheDrsExprDayRangeShow((KHE_DRS_EXPR) ebt, drs, buff1), */
      KheDrsDomTestShow(ebt->dom_test, buff2));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_FREE_TIME"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FREE_TIME KheDrsExprFreeTimeMake(KHE_DRS_RESOURCE dr,       */
/*    KHE_TIME time, KHE_DRS_DOM_TEST dom_test,                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_FREE_TIME object with these attributes.          */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FREE_TIME KheDrsExprFreeTimeMake(
  KHE_DRS_RESOURCE_ON_DAY drd, KHE_TIME time,
  KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FREE_TIME res;  KHE_DRS_EXPR e;
  if( KheDrsResourceOnDayHasTimeExpr(drd, KHE_DRS_EXPR_FREE_TIME_TAG,
      dom_test, time, drs, &e) )
    return (KHE_DRS_EXPR_FREE_TIME) e;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_FREE_TIME_TAG, dom_test, drs);
  res->u.int_val = 0;    /* default value, will be used sometimes */
  res->resource_on_day = drd;
  res->time = time;
  KheDrsResourceOnDayAddExpr(drd, (KHE_DRS_EXPR) res);
  KheDrsExprInitEnd(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FREE_TIME KheDrsExprFreeTimeCopy(KHE_DRS_EXPR_FREE_TIME eft,*/
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a fresh and unshared copy of eft.  Do not copy the sig_indexes    */
/*  and child_indexes arrays; they should be empty anyway.                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FREE_TIME KheDrsExprFreeTimeCopy(KHE_DRS_EXPR_FREE_TIME eft,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FREE_TIME res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_FREE_TIME_TAG, eft->dom_test, drs);
  /* res->day_range = eft->day_range; */
  res->u.int_val = eft->u.int_val;
  res->time = eft->time;
  KheDrsExprInitEnd(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeFreeExternal(KHE_DRS_EXPR_FREE_TIME ef,           */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free ef and its descendants (there are no descendants here).             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFreeTimeFreeExternal(KHE_DRS_EXPR_FREE_TIME ef,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_expr_free_time, ef);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeSetValue(KHE_DRS_EXPR_FREE_TIME eft,              */
/*    int day_index, KHE_TASK task, KHE_TIME time, KHE_RESOURCE r)           */
/*                                                                           */
/*  Set the value of eft, given that r (which must be non-NULL and which     */
/*  ebt is concerned with) is assigned to task at time on day day_index.     */
/*                                                                           */
/*  Implementation note.  The condition "time == eft->time" is true when     */
/*  r is assigned a task which is running at eft's time.  If r is not        */
/*  assigned a task, time is NULL so the condition is false.                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFreeTimeSetValue(KHE_DRS_EXPR_FREE_TIME eft,
  ** int day_index, ** KHE_TASK task, KHE_TIME time, KHE_RESOURCE r)
{
  ** KHE_DRS_PARENT p;  int i; **
  eft->u.int_val = (time == eft->time ? 0 : 1);
  ** ***
  HaArrayForEach(eft->parents, p, i)
    KheDrsExprChildHasClosed(p.expr, (KHE_DRS_EXPR) eft, p.index);
  *** **
    ** ***
    KheDrsExprChildHasClosed(p.expr, (KHE_DRS_EXPR) eft,
      p.index, day_index, eft->u);
    *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeSetClosedValue(KHE_DRS_EXPR_FREE_TIME eft,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eft, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeTimeSetClosedValue(KHE_DRS_EXPR_FREE_TIME eft,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;
  dtd = eft->resource_on_day->closed_asst;
  eft->u.int_val = (dtd != NULL && dtd->time == eft->time ? 0 : 1);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeLeafSet(KHE_DRS_EXPR_FREE_TIME eft,               */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Set the value of eft, given that dtd has just been assigned drd.         */
/*  Here dtd is known to be an actual task, not a non-assignment, and drd    */
/*  is known to be concerned with the same resource as eft is.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeTimeLeafSet(KHE_DRS_EXPR_FREE_TIME eft,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  if( dtd->time == eft->time )
    eft->u.int_val = 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeLeafClear(KHE_DRS_EXPR_FREE_TIME eft)             */
/*                                                                           */
/*  Inform leaf expression eft that the assignment of drd to dtd carried     */
/*  out by KheDrsExprFreeTimeLeafSet above has been removed.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeTimeLeafClear(KHE_DRS_EXPR_FREE_TIME eft)
{
  eft->u.int_val = 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeDoDay(KHE_DRS_EXPR_FREE_TIME eft, int day_index,  */
/*    KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active expression eft on day_index.                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFreeTimeDoDay(KHE_DRS_EXPR_FREE_TIME eft, int day_index,
  KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;
  if( day_index == eft->day_range.first )
  {
    ** this is eft's first day **
    eft->u.int_val = (time == eft->time ? 0 : 1);
  }
  else
  {
    ** this is not eft's first day, so we need to retrieve a previous value **
    si = HaArray(eft->sig_indexes, day_index - 1 - eft->day_range.first);
    eft->u.int_val = HaArray(prev_soln->signature, prev_so + si);
  }
  if( day_index < eft->day_range.last )
  {
    ** this not eft's last day; store value in next_soln **
    si = HaArray(eft->sig_indexes, day_index - eft->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == next_so + si,
      "KheDrsExprFreeTimeDoDay internal error");
    HaArrayAddLast(next_soln->signature, eft->u.int_val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeTimeDebug(KHE_DRS_EXPR_FREE_TIME eft,                 */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for free time expression eft.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeTimeDebug(KHE_DRS_EXPR_FREE_TIME eft,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char /* buff1[100], */ buff2[100];  int si;
  if( soln != NULL )
  {
    /* unindented version with a soln value */
    fprintf(fp, "FREE_TIME(%s)", KheTimeId(eft->time));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eft, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
  }
  else
  {
    /* indented version without a soln value */
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "[ FREE_TIME(%s) %s ]", KheTimeId(eft->time),
      /* KheDrsExprDayRangeShow((KHE_DRS_EXPR) eft, drs, buff1), */
      KheDrsDomTestShow(eft->dom_test, buff2));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_WORK_TIME"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_WORK_TIME KheDrsExprWorkTimeMake(KHE_DRS_RESOURCE dr,       */
/*    KHE_TIME time, KHE_DRS_DOM_TEST dom_test,                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_WORK_TIME object with these attributes.          */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_WORK_TIME KheDrsExprWorkTimeMake(
  KHE_DRS_RESOURCE_ON_DAY drd, KHE_TIME time,
  KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_WORK_TIME res;  KHE_DRS_EXPR e;
  if( KheDrsResourceOnDayHasTimeExpr(drd, KHE_DRS_EXPR_WORK_TIME_TAG,
	dom_test, time, drs, &e) )
    return (KHE_DRS_EXPR_WORK_TIME) e;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_WORK_TIME_TAG, dom_test, drs);
  res->u.float_val = 0.0;    /* default value, will be used sometimes */
  res->resource_on_day = drd;
  res->time = time;
  KheDrsResourceOnDayAddExpr(drd, (KHE_DRS_EXPR) res);
  KheDrsExprInitEnd(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_WORK_TIME KheDrsExprWorkTimeCopy(KHE_DRS_EXPR_WORK_TIME ewt,*/
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a fresh and unshared copy of ewt.  Do not copy the sig_indexes    */
/*  and child_indexes arrays; they should be empty anyway.                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_WORK_TIME KheDrsExprWorkTimeCopy(KHE_DRS_EXPR_WORK_TIME ewt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_WORK_TIME res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_WORK_TIME_TAG, ewt->dom_test, drs);
  /* res->day_range = ewt->day_range; */
  res->u.float_val = ewt->u.float_val;
  res->time = ewt->time;
  KheDrsExprInitEnd(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeFreeExternal(KHE_DRS_EXPR_WORK_TIME ewt,          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free ewt and its descendants (there are no descendants here).            */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprWorkTimeFreeExternal(KHE_DRS_EXPR_WORK_TIME ewt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_expr_work_time, ewt);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeSetValue(KHE_DRS_EXPR_WORK_TIME ewt,              */
/*    int day_index, KHE_TASK task, KHE_TIME time, KHE_RESOURCE r)           */
/*                                                                           */
/*  Set the value of ewt, given that r (which must be non-NULL and which     */
/*  ewt is concerned with) is assigned to task at time on day day_index.     */
/*                                                                           */
/*  Implementation note.  The condition "time == ewt->time" is true when     */
/*  r is assigned a task which is running at ewt's time.  If r is not        */
/*  assigned a task, time is NULL so the condition is false.                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprWorkTimeSetValue(KHE_DRS_EXPR_WORK_TIME ewt,
  ** int day_index, ** KHE_TASK task, KHE_TIME time, KHE_RESOURCE r)
{
  ** KHE_DRS_PARENT p;  int i; **
  ewt->u.float_val = (time == ewt->time ? KheTaskWorkloadPerTime(task) : 0.0);
  ** ***
  HaArrayForEach(ewt->parents, p, i)
    KheDrsExprChildHasClosed(p.expr, (KHE_DRS_EXPR) ewt, p.index);
  *** **
    ** ***
    KheDrsExprChildHasClosed(p.expr, (KHE_DRS_EXPR) ewt,
      p.index, day_index, ewt->u);
    *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeSetClosedValue(KHE_DRS_EXPR_WORK_TIME ewt,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing ewt, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkTimeSetClosedValue(KHE_DRS_EXPR_WORK_TIME ewt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;
  dtd = ewt->resource_on_day->closed_asst;
  ewt->u.float_val = (dtd != NULL && dtd->time == ewt->time ?
    KheTaskWorkloadPerTime(dtd->task) : 0.0);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeLeafSet(KHE_DRS_EXPR_WORK_TIME ewt,               */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Set the value of ewt, given that dtd has just been assigned drd.         */
/*  Here dtd is known to be an actual task, not a non-assignment, and drd    */
/*  is known to be concerned with the same resource as ewt is.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkTimeLeafSet(KHE_DRS_EXPR_WORK_TIME ewt,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  if( dtd->time == ewt->time )
    ewt->u.float_val = KheTaskWorkloadPerTime(dtd->task);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeLeafClear(KHE_DRS_EXPR_WORK_TIME ewt)             */
/*                                                                           */
/*  Inform leaf expression ebt that the assignment of drd to dtd carried     */
/*  out by KheDrsExprWorkTimeLeafSet above has been removed.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkTimeLeafClear(KHE_DRS_EXPR_WORK_TIME ewt)
{
  ewt->u.float_val = 0.0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeDoDay(KHE_DRS_EXPR_WORK_TIME ewt, int day_index,  */
/*    KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active expression ewt on day_index.                              */
/*                                                                           */
/*  Implementation note.  If time == ewt->time, then since ewt->time is      */
/*  always non-NULL, time is non-NULL, and this implies task is non-NULL.    */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprWorkTimeDoDay(KHE_DRS_EXPR_WORK_TIME ewt, int day_index,
  KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;
  if( day_index == ewt->day_range.first )
  {
    ** this is ewt's first day **
    ewt->u.float_val = (time == ewt->time ? KheTaskWorkloadPerTime(task) : 0.0);
  }
  else
  {
    ** this is not ewt's first day, so we need to retrieve a previous value **
    si = HaArray(ewt->sig_indexes, day_index - 1 - ewt->day_range.first);
    ewt->u.float_val = HaArray(prev_soln->signature, prev_so + si);
  }
  if( day_index < ewt->day_range.last )
  {
    ** this not ewt's last day; store value in next_soln **
    si = HaArray(ewt->sig_indexes, day_index - ewt->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == next_so + si,
      "KheDrsExprWorkTimeDoDay internal error");
    HaArrayAddLast(next_soln->signature, ewt->u.float_val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkTimeDebug(KHE_DRS_EXPR_WORK_TIME ewt,                 */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for work time expression ewt.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkTimeDebug(KHE_DRS_EXPR_WORK_TIME ewt,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char /* buff1[100], */ buff2[100];  int si;
  if( soln != NULL )
  {
    /* unindented version with a soln value */
    fprintf(fp, "WORK_TIME(%s)", KheTimeId(ewt->time));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) ewt, soln->day, &si) )
      fprintf(fp, " $%d:%.2f$", si, (float) HaArray(soln->signature, si) / 100.0);
  }
  else
  {
    /* indented version without a soln value */
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "[ WORK_TIME(%s) %s ]", KheTimeId(ewt->time),
      /* KheDrsExprDayRangeShow((KHE_DRS_EXPR) ewt, drs, buff1), */
      KheDrsDomTestShow(ewt->dom_test, buff2));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_BUSY_DAY"                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_BUSY_DAY KheDrsExprBusyDayMake(KHE_DRS_RESOURCE dr,         */
/*    KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_BUSY_DAY object with these attributes.           */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_BUSY_DAY KheDrsExprBusyDayMake(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_BUSY_DAY res;  KHE_DRS_EXPR e;
  if( KheDrsResourceOnDayHasDayExpr(drd, KHE_DRS_EXPR_BUSY_DAY_TAG, dom_test,
      tg, drs, &e) )
    return (KHE_DRS_EXPR_BUSY_DAY) e;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_BUSY_DAY_TAG, dom_test, drs);
  res->u.int_val = 0;    /* default value, will be used sometimes */
  res->resource_on_day = drd;
  res->time_group = tg;
  KheDrsResourceOnDayAddExpr(drd, (KHE_DRS_EXPR) res);
  KheDrsExprInitEnd(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_BUSY_DAY KheDrsExprBusyDayCopy(KHE_DRS_EXPR_BUSY_DAY ebd,   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a fresh and unshared copy of ebd.  Do not copy the sig_indexes    */
/*  and child_indexes arrays; they should be empty anyway.                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_BUSY_DAY KheDrsExprBusyDayCopy(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_BUSY_DAY res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_BUSY_DAY_TAG, ebd->dom_test, drs);
  /* res->day_range = ebd->day_range; */
  res->u.int_val = ebd->u.int_val;
  res->time_group = ebd->time_group;
  KheDrsExprInitEnd(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDayFreeExternal(KHE_DRS_EXPR_BUSY_DAY ebd,            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free ebd and its descendants (there are no descendants here).            */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprBusyDayFreeExternal(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_expr_busy_day, ebd);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDaySetValue(KHE_DRS_EXPR_BUSY_DAY ebd,                */
/*    int day_index, KHE_TASK task, KHE_TIME time, KHE_RESOURCE r)           */
/*                                                                           */
/*  Set the value of ebd, given that r (which must be non-NULL and which     */
/*  ebd is concerned with) is assigned to task at time on day day_index.     */
/*                                                                           */
/*  Implementation note.  The condition "time != NULL" is true when r is     */
/*  assigned a task which is running at some time on ebd's day.  If r is     */
/*  not assigned a task, time is NULL so the condition is false.             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprBusyDaySetValue(KHE_DRS_EXPR_BUSY_DAY ebd,
  ** int day_index, ** KHE_TASK task, KHE_TIME time, KHE_RESOURCE r)
{
  ** KHE_DRS_PARENT p;  int i; **
  ebd->u.int_val = (time != NULL ? 1 : 0);
  ** ***
  HaArrayForEach(ebd->parents, p, i)
    KheDrsExprChildHasClosed(p.expr, (KHE_DRS_EXPR) ebd, p.index);
  *** **
    ** ***
    KheDrsExprChildHasClosed(p.expr, (KHE_DRS_EXPR) ebd,
      p.index, day_index, ebd->u);
    *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDaySetClosedValue(KHE_DRS_EXPR_BUSY_DAY ebd,          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing ebd, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyDaySetClosedValue(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;
  dtd = ebd->resource_on_day->closed_asst;
  ebd->u.int_val = (dtd != NULL ? 1 : 0);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDayLeafSet(KHE_DRS_EXPR_BUSY_DAY ebd,                 */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Set the value of ebd, given that dtd has just been assigned drd.         */
/*  Here dtd is known to be an actual task, not a non-assignment, and drd    */
/*  is known to be concerned with the same resource as ebt is.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyDayLeafSet(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  ebd->u.int_val = 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDayLeafClear(KHE_DRS_EXPR_BUSY_DAY ebd)               */
/*                                                                           */
/*  Inform leaf soln ebd that the assignment of drd to dtd carried out by    */
/*  KheDrsExprBusyDayLeafSet above has been removed.                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyDayLeafClear(KHE_DRS_EXPR_BUSY_DAY ebd)
{
  ebd->u.int_val = 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDayDoDay(KHE_DRS_EXPR_BUSY_DAY ebd, int day_index,    */
/*    KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active soln ebd on day_index.                                    */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprBusyDayDoDay(KHE_DRS_EXPR_BUSY_DAY ebd, int day_index,
  KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;
  if( day_index == ebd->day_range.first )
  {
    ** this is ebd's first day **
    ebd->u.int_val = (time != NULL ? 1 : 0);
  }
  else
  {
    ** this is not ebd's first day, so we need to retrieve a previous value **
    si = HaArray(ebd->sig_indexes, day_index - 1 - ebd->day_range.first);
    ebd->u.int_val = HaArray(prev_soln->signature, prev_so + si);
  }
  if( day_index < ebd->day_range.last )
  {
    ** this not ebd's last day; store value in next_soln **
    si = HaArray(ebd->sig_indexes, day_index - ebd->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == next_so + si,
      "KheDrsExprBusyDayDoDay internal error");
    HaArrayAddLast(next_soln->signature, ebd->u.int_val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprBusyDayDebug(KHE_DRS_EXPR_BUSY_DAY ebd,                   */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for busy day expression ebd.                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprBusyDayDebug(KHE_DRS_EXPR_BUSY_DAY ebd,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char /* buff1[100], */ buff2[100];  int si;
  if( soln != NULL )
  {
    /* unindented version with a soln value */
    fprintf(fp, "BUSY_DAY(%s)", KheTimeGroupId(ebd->time_group));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) ebd, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
  }
  else
  {
    /* indented version without a soln value */
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "[ BUSY_DAY(%s) %s ]", KheTimeGroupId(ebd->time_group),
      /* KheDrsExprDayRangeShow((KHE_DRS_EXPR) ebd, drs, buff1), */
      KheDrsDomTestShow(ebd->dom_test, buff2));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_FREE_DAY"                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FREE_DAY KheDrsExprFreeDayMake(KHE_DRS_RESOURCE dr,         */
/*    KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_FREE_DAY object with these attributes.           */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FREE_DAY KheDrsExprFreeDayMake(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FREE_DAY res;  KHE_DRS_EXPR e;
  if( KheDrsResourceOnDayHasDayExpr(drd, KHE_DRS_EXPR_FREE_DAY_TAG,
	dom_test, tg, drs, &e) )
    return (KHE_DRS_EXPR_FREE_DAY) e;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_FREE_DAY_TAG, dom_test, drs);
  res->u.int_val = 1;    /* default value, will be used sometimes */
  res->resource_on_day = drd;
  res->time_group = tg;
  KheDrsResourceOnDayAddExpr(drd, (KHE_DRS_EXPR) res);
  KheDrsExprInitEnd(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FREE_DAY KheDrsExprFreeDayCopy(KHE_DRS_EXPR_FREE_DAY efd,   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a fresh and unshared copy of efd.  Do not copy the sig_indexes    */
/*  and child_indexes arrays; they should be empty anyway.                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FREE_DAY KheDrsExprFreeDayCopy(KHE_DRS_EXPR_FREE_DAY efd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FREE_DAY res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_FREE_DAY_TAG, efd->dom_test, drs);
  /* res->day_range = efd->day_range; */
  res->u.int_val = efd->u.int_val;
  res->time_group = efd->time_group;
  KheDrsExprInitEnd(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDayFreeExternal(KHE_DRS_EXPR_FREE_DAY efd,            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free efd and its descendants (there are no descendants here).            */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFreeDayFreeExternal(KHE_DRS_EXPR_FREE_DAY efd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_expr_free_day, efd);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDaySetValue(KHE_DRS_EXPR_FREE_DAY efd,                */
/*    int day_index, KHE_TASK task, KHE_TIME time, KHE_RESOURCE r)           */
/*                                                                           */
/*  Set the value of efd, given that r (which must be non-NULL and which     */
/*  efd is concerned with) is assigned to task at time on day day_index.     */
/*                                                                           */
/*  Implementation note.  The condition "time != NULL" is true when r is     */
/*  assigned a task which is running at some time on efd's day.  If r is     */
/*  not assigned a task, time is NULL so the condition is false.             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFreeDaySetValue(KHE_DRS_EXPR_FREE_DAY efd,
  ** int day_index, ** KHE_TASK task, KHE_TIME time, KHE_RESOURCE r)
{
  ** KHE_DRS_PARENT p;  int i; **
  efd->u.int_val = (time != NULL ? 0 : 1);
  ** ***
  HaArrayForEach(efd->parents, p, i)
    KheDrsExprChildHasClosed(p.expr, (KHE_DRS_EXPR) efd, p.index);
  *** **
    ** ***
    KheDrsExprChildHasClosed(p.expr, (KHE_DRS_EXPR) efd,
      p.index, day_index, efd->u);
    *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDaySetClosedValue(KHE_DRS_EXPR_FREE_DAY efd,          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing efd, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeDaySetClosedValue(KHE_DRS_EXPR_FREE_DAY efd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;
  dtd = efd->resource_on_day->closed_asst;
  efd->u.int_val = (dtd != NULL ? 0 : 1);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDayLeafSet(KHE_DRS_EXPR_FREE_DAY efd,                 */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Set the value of efd, given that dtd has just been assigned drd.         */
/*  Here dtd is known to be an actual task, not a non-assignment, and drd    */
/*  is known to be concerned with the same resource as efd is.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeDayLeafSet(KHE_DRS_EXPR_FREE_DAY efd,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  efd->u.int_val = 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDayLeafClear(KHE_DRS_EXPR_FREE_DAY efd)               */
/*                                                                           */
/*  Inform leaf expression eft that the assignment of drd to dtd carried     */
/*  out by KheDrsExprFreeDayLeafSet above has been removed.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeDayLeafClear(KHE_DRS_EXPR_FREE_DAY efd)
{
  efd->u.int_val = 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDayDoDay(KHE_DRS_EXPR_FREE_DAY efd, int day_index,    */
/*    KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active expression efd on day_index.                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFreeDayDoDay(KHE_DRS_EXPR_FREE_DAY efd, int day_index,
  KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;
  if( day_index == efd->day_range.first )
  {
    ** this is efd's first day **
    efd->u.int_val = (time == NULL ? 1 : 0);
  }
  else
  {
    ** this is not efd's first day, so we need to retrieve a previous value **
    si = HaArray(efd->sig_indexes, day_index - 1 - efd->day_range.first);
    efd->u.int_val = HaArray(prev_soln->signature, prev_so + si);
  }
  if( day_index < efd->day_range.last )
  {
    ** this not efd's last day; store value in next_soln **
    si = HaArray(efd->sig_indexes, day_index - efd->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == next_so + si,
      "KheDrsExprFreeDayDoDay internal error");
    HaArrayAddLast(next_soln->signature, efd->u.int_val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeDayDebug(KHE_DRS_EXPR_FREE_DAY efd,                   */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for free time expression efd.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFreeDayDebug(KHE_DRS_EXPR_FREE_DAY efd,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char /* buff1[100], */ buff2[100];  int si;
  if( soln != NULL )
  {
    /* unindented version with a soln value */
    fprintf(fp, "FREE_DAY(%s)", KheTimeGroupId(efd->time_group));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) efd, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
  }
  else
  {
    /* indented version without a soln value */
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "[ FREE_DAY(%s) %s ]", KheTimeGroupId(efd->time_group),
      /* KheDrsExprDayRangeShow((KHE_DRS_EXPR) efd, drs, buff1), */
      KheDrsDomTestShow(efd->dom_test, buff2));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_WORK_DAY"                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_WORK_DAY KheDrsExprWorkDayMake(KHE_TIME_GROUP tg,           */
/*    KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)            */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_WORK_DAY object with these attributes.           */
/*  Or retrieve an equivalent existing expression if there is one.           */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused, although I really should be using it
static KHE_DRS_EXPR_WORK_DAY KheDrsExprWorkDayMake(KHE_TIME_GROUP tg,
  KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_WORK_DAY res;  KHE_DRS_EXPR e;
  if( KheDrsHasDayExpr(drs, KHE_DRS_EXPR_WORK_DAY_TAG, dom_test, tg, &e) )
    return (KHE_DRS_EXPR_WORK_DAY) e;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_WORK_DAY_TAG, dom_test, drs);
  res->u.float_val = 0.0;  ** default value, used sometimes **
  res->time_group = tg;
  KheDrsAddDayExpr(drs, tg, (KHE_DRS_EXPR) res);
  KheDrsExprInitEnd(res, drs);
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_WORK_DAY KheDrsExprWorkDayCopy(KHE_DRS_EXPR_WORK_DAY ewd,   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a fresh and unshared copy of ewd.  Do not copy the sig_indexes    */
/*  and child_indexes arrays; they should be empty anyway.                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_WORK_DAY KheDrsExprWorkDayCopy(KHE_DRS_EXPR_WORK_DAY ewd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_WORK_DAY res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_WORK_DAY_TAG, ewd->dom_test, drs);
  /* res->day_range = ewd->day_range; */
  res->u.float_val = ewd->u.float_val;
  res->time_group = ewd->time_group;
  KheDrsExprInitEnd(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDayFreeExternal(KHE_DRS_EXPR_WORK_DAY ewd,            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free ewd and its descendants (there are no descendants here).            */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprWorkDayFreeExternal(KHE_DRS_EXPR_WORK_DAY ewd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_expr_work_day, ewd);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDaySetValue(KHE_DRS_EXPR_WORK_DAY ewd,                */
/*    int day_index, KHE_TASK task, KHE_TIME time, KHE_RESOURCE r)           */
/*                                                                           */
/*  Set the value of ewd, given that r (which must be non-NULL and which     */
/*  ewd is concerned with) is assigned to task at time on day day_index.     */
/*                                                                           */
/*  Implementation note.  The condition "time != NULL" is true when r is     */
/*  assigned a task which is running at some time on ewd's day.  If r is     */
/*  not assigned a task, time is NULL so the condition is false.             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprWorkDaySetValue(KHE_DRS_EXPR_WORK_DAY ewd,
  ** int day_index, ** KHE_TASK task, KHE_TIME time, KHE_RESOURCE r)
{
  ** KHE_DRS_PARENT p;  int i; **
  ewd->u.float_val = (time != NULL ? KheTaskWorkloadPerTime(task) : 0.0);
  ** ***
  HaArrayForEach(ewd->parents, p, i)
    KheDrsExprChildHasClosed(p.expr, (KHE_DRS_EXPR) ewd, p.index);
  *** **
    ** ***
    KheDrsExprChildHasClosed(p.expr, (KHE_DRS_EXPR) ewd,
      p.index, day_index, ewd->u);
    *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDaySetClosedValue(KHE_DRS_EXPR_WORK_DAY ewd,          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing ewd, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkDaySetClosedValue(KHE_DRS_EXPR_WORK_DAY ewd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;
  dtd = ewd->resource_on_day->closed_asst;
  ewd->u.int_val = (dtd != NULL ? KheTaskWorkloadPerTime(dtd->task) : 0.0);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDayLeafSet(KHE_DRS_EXPR_WORK_DAY ewd,                 */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Set the value of ewd, given that dtd has just been assigned drd.         */
/*  Here dtd is known to be an actual task, not a non-assignment, and drd    */
/*  is known to be concerned with the same resource as ewd is.               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkDayLeafSet(KHE_DRS_EXPR_WORK_DAY ewd,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  ewd->u.float_val = KheTaskWorkloadPerTime(dtd->task);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDayLeafClear(KHE_DRS_EXPR_WORK_DAY ewd)               */
/*                                                                           */
/*  Inform leaf expression ewt that the assignment of drd to dtd carried     */
/*  out by KheDrsExprWorkDayLeafSet above has been removed.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkDayLeafClear(KHE_DRS_EXPR_WORK_DAY ewd)
{
  ewd->u.float_val = 0.0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDayDoDay(KHE_DRS_EXPR_WORK_DAY ewd, int day_index,    */
/*    KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active expression ewd on day_index.                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprWorkDayDoDay(KHE_DRS_EXPR_WORK_DAY ewd, int day_index,
  KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;
  if( day_index == ewd->day_range.first )
  {
    ** this is ewd's first day **
    ewd->u.float_val = (task != NULL ? KheTaskWorkloadPerTime(task) : 0.0);
  }
  else
  {
    ** this is not ewd's first day, so we need to retrieve a previous value **
    si = HaArray(ewd->sig_indexes, day_index - 1 - ewd->day_range.first);
    ewd->u.float_val = HaArray(prev_soln->signature, prev_so + si);
  }
  if( day_index < ewd->day_range.last )
  {
    ** this not ewd's last day; store value in next_soln **
    si = HaArray(ewd->sig_indexes, day_index - ewd->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == next_so + si,
      "KheDrsExprWorkDayDoDay internal error");
    HaArrayAddLast(next_soln->signature, ewd->u.float_val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprWorkDayDebug(KHE_DRS_EXPR_WORK_DAY ewd,                   */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for work day expression ewd.                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprWorkDayDebug(KHE_DRS_EXPR_WORK_DAY ewd,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char /* buff1[100], */ buff2[100];  int si;
  if( soln != NULL )
  {
    /* unindented version with a soln value */
    fprintf(fp, "WORK_DAY(%s)", KheTimeGroupId(ewd->time_group));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) ewd, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
  }
  else
  {
    /* indented version without a soln value */
    if( indent > 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "[ WORK_DAY(%s) %s ]", KheTimeGroupId(ewd->time_group),
      /* KheDrsExprDayRangeShow((KHE_DRS_EXPR) ewd, drs, buff1), */
      KheDrsDomTestShow(ewd->dom_test, buff2));
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_ASSIGNED_TASK"                                   */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_ASSIGNED_TASK KheDrsExprAssignedTaskMake(KHE_TASK task,     */
/*    KHE_RESOURCE_GROUP rg, KHE_DYNAMIC_RESOURCE_SOLVER drs)                */
/*                                                                           */
/*  Make an assigned task expression object with these attributes.           */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_EXPR_ASSIGNED_TASK KheDrsExprAssignedTaskMake(KHE_DRS_TASK dt,
  KHE_TASK task, KHE_RESOURCE_GROUP rg, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_ASSIGNED_TASK res;  KHE_DRS_DOM_TEST dom_test;  KHE_DRS_EXPR e;
  dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  if( KheDrsTaskHasAssignedTaskExpr(dt, task, rg, &res) )
    return res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_ASSIGNED_TASK_TAG, dom_test, drs);
  res->value = 0;    ** default value, will be used sometimes **
  res->task = task;
  res->resource_group = rg;
  KheDrsExprInitEnd(res, drs);
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_ASSIGNED_TASK KheDrsExprAssignedTaskMake(                   */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_RESOURCE_GROUP rg,                        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make an assigned task expression object with these attributes.           */
/*  Or if there is already such an expression in dtd's list, use that.       */
/*                                                                           */
/*****************************************************************************/
static bool KheDrsTaskOnDayHasAssignedTaskExpr(KHE_DRS_TASK_ON_DAY dtd,
  KHE_RESOURCE_GROUP rg, KHE_DRS_EXPR_ASSIGNED_TASK *res);
static void KheDrsTaskOnDayAddAssignedTaskExpr(KHE_DRS_TASK_ON_DAY dtd,
  KHE_DRS_EXPR_ASSIGNED_TASK eat);

static KHE_DRS_EXPR_ASSIGNED_TASK KheDrsExprAssignedTaskMake(
  KHE_DRS_TASK_ON_DAY dtd, KHE_RESOURCE_GROUP rg,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_ASSIGNED_TASK res;  KHE_DRS_DOM_TEST dom_test;
  dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  HnAssert(dtd->task != NULL, "KheDrsExprAssignedTaskMake internal error");
  if( KheDrsTaskOnDayHasAssignedTaskExpr(dtd, rg, &res) )
    return res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_ASSIGNED_TASK_TAG, dom_test, drs);
  res->u.int_val = 0;    /* default value, will be used sometimes */
  res->task_on_day = dtd;
  res->resource_group = rg;
  KheDrsTaskOnDayAddAssignedTaskExpr(dtd, res);
  KheDrsExprInitEnd(res, drs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAssignedTaskFreeExternal(KHE_DRS_EXPR_ASSIGNED_TASK eat,  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free eat.                                                                */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprAssignedTaskFreeExternal(KHE_DRS_EXPR_ASSIGNED_TASK eat,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_expr_assigned_task, eat);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprAssignedTaskValue(KHE_DRS_EXPR_ASSIGNED_TASK eat,          */
/*    KHE_RESOURCE r)                                                        */
/*                                                                           */
/*  Return the value of eat when its task is assigned r (possibly NULL).     */
/*  This will be 1 when r is non-NULL and lies in eat's resource group.      */
/*  For efficienty, eat's resource group may be NULL, meaning all resources. */
/*                                                                           */
/*****************************************************************************/

static int KheDrsExprAssignedTaskValue(KHE_DRS_EXPR_ASSIGNED_TASK eat,
  KHE_RESOURCE r)
{
  KHE_RESOURCE_GROUP rg;
  if( r == NULL )
    return 0;
  rg = eat->resource_group;
  return (rg == NULL || KheResourceGroupContains(rg, r)) ? 1 : 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAssignedTaskSetValue(KHE_DRS_EXPR_ASSIGNED_TASK eat,      */
/*    int day_index, KHE_TASK task, KHE_TIME time, KHE_RESOURCE r)           */
/*                                                                           */
/*  Optionally set the value of eat, given that task has just been           */
/*  assigned r (possibly NULL) at time.                                      */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprAssignedTaskSetValue(KHE_DRS_EXPR_ASSIGNED_TASK eat,
  ** int day_index, ** KHE_TASK task, KHE_TIME time, KHE_RESOURCE r)
{
  ** KHE_DRS_PARENT p;  int i; **
  eat->u.int_val = KheDrsExprAssignedTaskValue(eat, r);
  ** ***
  HaArrayForEach(eat->parents, p, i)
    KheDrsExprChildHasClosed(p.expr, (KHE_DRS_EXPR) eat, p.index);
  *** **
    ** ***
    KheDrsExprChildHasClosed(p.expr, (KHE_DRS_EXPR) eat,
      p.index, day_index, eat->u);
    *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAssignedTaskSetClosedValue(KHE_DRS_EXPR_ASSIGNED_TASK eat,*/
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eat, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAssignedTaskSetClosedValue(KHE_DRS_EXPR_ASSIGNED_TASK eat,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_RESOURCE r;
  drd = eat->task_on_day->closed_asst;
  r = (drd == NULL ? NULL : drd->encl_dr->resource);
  eat->u.int_val = KheDrsExprAssignedTaskValue(eat, r);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAssignedTaskLeafSet(KHE_DRS_EXPR_ASSIGNED_TASK eat,       */
/*    KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)                  */
/*                                                                           */
/*  Update the value of eat to reflect the assignment of drd to dtd.         */
/*  Here dtd is known to be the relevant task, and drd is known to           */
/*  represent a non-NULL resource, not a non-assignment.  Also the           */
/*  value of eat must currently be 0.                                        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAssignedTaskLeafSet(KHE_DRS_EXPR_ASSIGNED_TASK eat,
  KHE_DRS_TASK_ON_DAY dtd, KHE_DRS_RESOURCE_ON_DAY drd)
{
  if( KheDrsExprAssignedTaskValue(eat, drd->encl_dr->resource) == 1 )
    eat->u.int_val = 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAssignedTaskLeafClear(KHE_DRS_EXPR_ASSIGNED_TASK eat)     */
/*                                                                           */
/*  Inform leaf expression ebt that the assignment of drd to dtd carried     */
/*  out by KheDrsExprAssignedTaskLeafSet above has been removed.             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAssignedTaskLeafClear(KHE_DRS_EXPR_ASSIGNED_TASK eat)
{
  eat->u.int_val = 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAssignedTaskDoDay(KHE_DRS_EXPR_ASSIGNED_TASK eat,         */
/*    int day_index, KHE_DRS_SOLN prev_soln, int prev_so, KHE_RESOURCE r,    */
/*    KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active expression eat on day_index.                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprAssignedTaskDoDay(KHE_DRS_EXPR_ASSIGNED_TASK eat,
  int day_index, KHE_DRS_SOLN prev_soln, int prev_so, KHE_RESOURCE r,
  KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;

  ** initialize eo->value depending on whether day_index is its first day **
  if( day_index == eat->day_range.first )
  {
    if( eat->resource_group == NULL ||
	KheResourceGroupContains(eat->resource_group, r) )
      eat->u.int_val = 1;
      ** eat->value = eat->duration; **
    else
      eat->u.int_val = 0;
  }
  else
  {
    ** this is not eo's first day, so we need to retrieve a previous value **
    si = HaArray(eat->sig_indexes, day_index - 1 - eat->day_range.first);
    eat->u.int_val = HaArray(prev_soln->signature, prev_so + si);
  }

  if( day_index < eat->day_range.last )
  {
    ** this not eat's last day; store value in next_soln **
    si = HaArray(eat->sig_indexes, day_index - eat->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == next_so + si,
      "KheDrsExprAssignedTaskDoDay internal error");
    HaArrayAddLast(next_soln->signature, eat->u.int_val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAssignedTaskDebug(KHE_DRS_EXPR_ASSIGNED_TASK eat,         */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for assigned task expression eat.              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAssignedTaskDebug(KHE_DRS_EXPR_ASSIGNED_TASK eat,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char /* buff1[100], */ buff2[100];  int si;
  if( soln != NULL )
  {
    /* unindented version with a soln value */
    fprintf(fp, "ASSIGNED_TASK");
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eat, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
  }
  else if( indent >= 0 )
  {
    /* indented version without a soln value */
    fprintf(fp, "%*s[ ASSIGNED_TASK %s %s ]\n", indent, "",
      /* KheDrsExprDayRangeShow((KHE_DRS_EXPR) eat, drs, buff1), */
      KheDrsDomTestShow(eat->dom_test, buff2),
      KheResourceGroupId(eat->resource_group));
  }
  else
  {
    /* unindented version without a soln value */
    fprintf(fp, "[ ASSIGNED_TASK %s %s ]",
      /* KheDrsExprDayRangeShow((KHE_DRS_EXPR) eat, drs, buff1), */
      KheDrsDomTestShow(eat->dom_test, buff2),
      KheResourceGroupId(eat->resource_group));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_OR"                                              */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_OR KheDrsExprOrMake(KHE_DRS_DOM_TEST dom_test,              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_OR object with these attributes.                 */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_OR KheDrsExprOrMake(KHE_DRS_DOM_TEST dom_test,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_OR res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_OR_TAG, dom_test, drs);
  res->u.int_val = UNDEF_INT;
  res->closed_state = 0;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrFreeInternal(KHE_DRS_EXPR_OR eo,                        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free eo and its internal descendants.                                    */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprOrFreeInternal(KHE_DRS_EXPR_OR eo,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(eo->children, e, i)
    KheDrsExprFreeInternal(e, drs);
  HaArrayAddLast(drs->free_expr_or, eo);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrChildHasOpened(KHE_DRS_EXPR_OR eo,                      */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform eo that its child_index'th child, child_e, has opened on          */
/*  day_index, and that its previous closed value was v.                     */
/*                                                                           */
/*****************************************************************************/
/* ***
static void KheDrsInternalExprChildHasOpened(KHE_DRS_EXPR e,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsExprOrChildHasOpened(KHE_DRS_EXPR_OR eo,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;  int i;
  if( v.int_val == 1 )
    eo->closed_state -= 1;
  eo->open_children++;
  if( eo->open_children == 1 )
  {
    ** this expression has just opened **
    HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) eo);
    HaArrayForEach(eo->parents, prnt, i)
      KheDrsInternalExprChildHasOpened(prnt.expr, (KHE_DRS_EXPR) eo,
	prnt.index, day_index, eo->u, drs);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrChildHasClosed(KHE_DRS_EXPR_OR eo,                      */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform eo that its child_index'th child, child_e, has                    */
/*  closed on day_index, and that child_e's closed value is v.               */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprOrChildHasClosed(KHE_DRS_EXPR_OR eo,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v)
{
  KHE_DRS_PARENT prnt;  int i;
  if( v.int_val == 1 )
    eo->closed_state += 1;
  eo->open_children--;
  if( eo->open_children == 0 )
  {
    ** this expression has just closed **
    eo->u.int_val = (eo->closed_state > 0 ? 1 : 0);
    HaArrayForEach(eo->parents, prnt, i)
      KheDrsExprChildHasClosed(prnt.expr, (KHE_DRS_EXPR) eo,
	prnt.index, day_index, eo->u);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrChildHasOpened(KHE_DRS_EXPR_OR eo,                      */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  As part of opening eo for solving, eo's child_index'th child, child_e,   */
/*  has been opened.  Update eo's closed state to reflect this.              */
/*                                                                           */
/*  The closed state of eo is the number of eo's closed children whose       */
/*  value is 1, so this update subtracts 1 from the closed state if          */
/*  child_e's value (which preserves its value before opening) is 1.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOrChildHasOpened(KHE_DRS_EXPR_OR eo,
  KHE_DRS_EXPR child_e, int child_index)
{
  if( child_e->u.int_val == 1 )
    eo->closed_state -= 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrChildHasClosed(KHE_DRS_EXPR_OR eo,                      */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  As part of closing eo, eo's child_index'th child, child_e, has been      */
/*  closed.  Update eo's closed state to reflect this.                       */
/*                                                                           */
/*  The closed state of eo is the number of eo's closed children whose       */
/*  value is 1, so this update adds 1 to the closed state if child_e's       */
/*  value (which is up to date because it is closed first) is 1.             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOrChildHasClosed(KHE_DRS_EXPR_OR eo,
  KHE_DRS_EXPR child_e, int child_index)
{
  if( child_e->u.int_val == 1 )
    eo->closed_state += 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrSetValueFromClosedState(KHE_DRS_EXPR_OR eo,             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Here eo is closing.  Set its value, based on its closed state.           */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprOrSetValueFromClosedState(KHE_DRS_EXPR_OR eo,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eo->u.int_val = (eo->closed_state > 0 ? 1 : 0);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrSetClosedValue(KHE_DRS_EXPR_OR eo,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eo, set its value suitably for its closed state.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOrSetClosedValue(KHE_DRS_EXPR_OR eo,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eo->u.int_val = (eo->closed_state > 0 ? 1 : 0);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrEvalDay(KHE_DRS_EXPR_OR eo,                             */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eo on day open_day_index.                            */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOrEvalDay(KHE_DRS_EXPR_OR eo, int open_day_index,
  KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2;  KHE_DRS_EXPR child_e;

  if( KheDrsExprOpenDayIsFirst(eo, open_day_index) )
  {
    /* first day, so (ignoring closed_state) we have a 0 (false) value here */
    eo->u.int_val = 0;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    eo->u.int_val = KheDrsExprSigVal(eo, open_day_index - 1, prev_soln);
  }

  /* accumulate the open values of the children of eo that finalized today */
  KheDrsExprForEachOpenDayChild(eo, open_day_index, child_e, i)
    if( child_e->u.int_val == 1 )
      eo->u.int_val = 1;

  if( KheDrsExprOpenDayIsLast(eo, open_day_index) )
  {
    /* last day; add closed_state to the value */
    if( eo->closed_state > 0 )
      eo->u.int_val = 1;
  }
  else
  {
    /* this not eo's last day; store value in next_soln */
    KheDrsExprPutSigVal(eo, open_day_index, next_soln, eo->u.int_val);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrInternalDoDay(KHE_DRS_EXPR_OR eo,                       */
/*    int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Process internal expression eo on day_index.                             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprOrInternalDoDay(KHE_DRS_EXPR_OR eo,
  int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, start_index, stop_index;  KHE_DRS_EXPR child_e;

  ** initialize eo->value depending on whether day_index is its first day **
  if( day_index == eo->day_range.first )
    eo->u.int_val = 0;
  else
  {
    ** this is not eo's first day, so we need to retrieve a previous value **
    si = HaArray(eo->sig_indexes, day_index - 1 - eo->day_range.first);
    eo->u.int_val = HaArray(prev_soln->signature, si);
  }

  ** accumulate the values of the children of eo that finalized today **
  start_index = HaArray(eo->open_child_indexes,
    day_index - eo->day_range.first);
  stop_index =  HaArray(eo->open_child_indexes,
    day_index + 1 - eo->day_range.first);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = HaArray(eo->children, i);
    if( child_e->u.int_val == 1 )
      eo->u.int_val = 1;
  }

  if( day_index < eo->day_range.last )
  {
    ** this not eo's last day; store value in next_soln **
    si = HaArray(eo->sig_indexes, day_index - eo->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == si,
      "KheDrsExprOrInternalDoDay internal error");
    HaArrayAddLast(next_soln->signature, eo->u.int_val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrDoDay(KHE_DRS_EXPR_OR eo, int day_index,                */
/*    KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active expression eo on day_index.                               */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprOrDoDay(KHE_DRS_EXPR_OR eo, int day_index,
  KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, start_index, stop_index;  KHE_DRS_EXPR child_e;

  ** initialize eo->value depending on whether day_index is its first day **
  if( day_index == eo->day_range.first )
    eo->u.int_val = 0;
  else
  {
    ** this is not eo's first day, so we need to retrieve a previous value **
    si = HaArray(eo->sig_indexes, day_index - 1 - eo->day_range.first);
    eo->u.int_val = HaArray(prev_soln->signature, prev_so + si);
  }

  ** accumulate the values of the children of eo that finalized today **
  start_index = HaArray(eo->open_child_indexes,
    day_index - eo->day_range.first);
  stop_index =  HaArray(eo->open_child_indexes,
    day_index + 1 - eo->day_range.first);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = HaArray(eo->children, i);
    if( child_e->u.int_val == 1 )
      eo->u.int_val = 1;
  }

  if( day_index < eo->day_range.last )
  {
    ** this not eo's last day; store value in next_soln **
    si = HaArray(eo->sig_indexes, day_index - eo->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == next_so + si,
      "KheDrsExprOrDoDay internal error");
    HaArrayAddLast(next_soln->signature, eo->u.int_val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprHeaderDebug(KHE_DRS_EXPR e, char *type_tag, FILE *fp)     */
/*                                                                           */
/*  Print the header of e (with the given type tag) onto fp.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprHeaderDebug(KHE_DRS_EXPR e, char *type_tag,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, FILE *fp)
{
  char /* buff1[100], */ buff2[100];
  fprintf(fp, "%s p%d o%d r%s", type_tag, e->postorder_index,
    HaArrayCount(e->open_children), /* KheDrsExprDayRangeShow(e, drs, buff1), */
    KheDrsDomTestShow(e->dom_test, buff2));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOrDebug(KHE_DRS_EXPR_OR eo,                               */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for or expression eo.                          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprOrDebug(KHE_DRS_EXPR_OR eo,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int si;
  if( soln != NULL )
  {
    /* unindented version with a soln value */
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eo, "OR", drs, fp);
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eo, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eo, soln, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    /* indented version without a soln value */
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eo, "OR", drs, fp);
    fprintf(fp, "\n");
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eo, soln, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    /* unindented version without a soln value */
    fprintf(fp, "[ ");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eo, "OR", drs, fp);
    fprintf(fp, " (%d children) ]", HaArrayCount(eo->children));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_AND"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_AND KheDrsExprAndMake(KHE_DRS_DOM_TEST dom_test,            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_AND object with these attributes.                */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_AND KheDrsExprAndMake(KHE_DRS_DOM_TEST dom_test,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_AND res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_AND_TAG, dom_test, drs);
  res->u.int_val = UNDEF_INT;
  res->closed_state = 0;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndFreeInternal(KHE_DRS_EXPR_AND ea,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free ea and its internal descendants.                                    */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprAndFreeInternal(KHE_DRS_EXPR_AND ea,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(ea->children, e, i)
    KheDrsExprFreeInternal(e, drs);
  HaArrayAddLast(drs->free_expr_and, ea);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndChildHasOpened(KHE_DRS_EXPR_AND ea,                    */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform ea that its child_index'th child, child_e, has opened on          */
/*  day_index, and that its previous closed value was v.                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprAndChildHasOpened(KHE_DRS_EXPR_AND ea,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;  int i;
  if( v.int_val == 0 )
    ea->closed_state -= 1;
  ea->open_children++;
  if( ea->open_children == 1 )
  {
    ** this expression has just opened **
    HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) ea);
    HaArrayForEach(ea->parents, prnt, i)
      KheDrsInternalExprChildHasOpened(prnt.expr, (KHE_DRS_EXPR) ea,
	prnt.index, day_index, ea->u, drs);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndChildHasClosed(KHE_DRS_EXPR_AND ea,                    */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform ea that its child_index'th child, child_e, has closed on          */
/*  day_index, and that child_e's closed value is v.                         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprAndChildHasClosed(KHE_DRS_EXPR_AND ea,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v)
{
  KHE_DRS_PARENT prnt;  int i;
  if( v.int_val == 0 )
    ea->closed_state += 1;
  ea->open_children--;
  if( ea->open_children == 0 )
  {
    ** this expression has just closed **
    ea->u.int_val = (ea->closed_state > 0 ? 0 : 1);
    HaArrayForEach(ea->parents, prnt, i)
      KheDrsExprChildHasClosed(prnt.expr, (KHE_DRS_EXPR) ea,
	prnt.index, day_index, ea->u);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndChildHasOpened(KHE_DRS_EXPR_AND ea,                    */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  As part of opening ea for solving, ea's child_index'th child, child_e,   */
/*  has been opened.  Update ea's closed state to reflect this.              */
/*                                                                           */
/*  The closed state of ea is the number of ea's closed children whose       */
/*  value is 0, so this update subtracts 1 from the closed state if          */
/*  child_e's value (which preserves its value before opening) is 0.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAndChildHasOpened(KHE_DRS_EXPR_AND ea,
  KHE_DRS_EXPR child_e, int child_index)
{
  if( child_e->u.int_val == 0 )
    ea->closed_state -= 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndChildHasClosed(KHE_DRS_EXPR_AND ea,                    */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  As part of closing ea, ea's child_index'th child, child_e, has been      */
/*  closed.  Update ea's closed state to reflect this.                       */
/*                                                                           */
/*  The closed state of ea is the number of ea's closed children whose       */
/*  value is 0, so this update adds 1 to the closed state if child_e's       */
/*  value (which is up to date because it is closed first) is 0.             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAndChildHasClosed(KHE_DRS_EXPR_AND ea,
  KHE_DRS_EXPR child_e, int child_index)
{
  if( child_e->u.int_val == 0 )
    ea->closed_state += 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndSetValueFromClosedState(KHE_DRS_EXPR_AND ea,           */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Here ea is closing.  Set its value, based on its closed state.           */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprAndSetValueFromClosedState(KHE_DRS_EXPR_AND ea,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ea->u.int_val = (ea->closed_state > 0 ? 0 : 1);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndSetClosedValue(KHE_DRS_EXPR_AND ea,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing ea, set its value suitably for its closed state.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAndSetClosedValue(KHE_DRS_EXPR_AND ea,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ea->u.int_val = (ea->closed_state > 0 ? 0 : 1);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndEvalDay(KHE_DRS_EXPR_AND ea,                           */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression ea on day open_day_index.                            */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAndEvalDay(KHE_DRS_EXPR_AND ea,
  int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2;  KHE_DRS_EXPR child_e;

  if( KheDrsExprOpenDayIsFirst(ea, open_day_index) )
  {
    /* first day, so (ignoring closed_state) we have a 1 (true) value here */
    ea->u.int_val = 1;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    ea->u.int_val = KheDrsExprSigVal(ea, open_day_index - 1, prev_soln);
  }

  /* accumulate the open values of the children of ea that finalized today */
  KheDrsExprForEachOpenDayChild(ea, open_day_index, child_e, i)
    if( child_e->u.int_val == 0 )
      ea->u.int_val = 0;

  if( KheDrsExprOpenDayIsLast(ea, open_day_index) )
  {
    /* last day; add closed_state to the value */
    if( ea->closed_state > 0 )
      ea->u.int_val = 0;
  }
  else
  {
    /* this not ea's last day; store value in next_soln */
    KheDrsExprPutSigVal(ea, open_day_index, next_soln, ea->u.int_val);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndInternalDoDay(KHE_DRS_EXPR_AND ea,                     */
/*    int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Process internal expression ea on day_index.                             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprAndInternalDoDay(KHE_DRS_EXPR_AND ea,
  int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, start_index, stop_index;  KHE_DRS_EXPR child_e;

  ** initialize ea->value depending on whether day_index is its first day **
  if( day_index == ea->day_range.first )
    ea->u.int_val = 1;
  else
  {
    ** this is not ea's first day, so we need to retrieve a previous value **
    si = HaArray(ea->sig_indexes, day_index - 1 - ea->day_range.first);
    ea->u.int_val = HaArray(prev_soln->signature, si);
  }

  ** accumulate the values of the children of ea that finalized today **
  start_index = HaArray(ea->open_child_indexes,
    day_index - ea->day_range.first);
  stop_index =  HaArray(ea->open_child_indexes,
    day_index + 1 - ea->day_range.first);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = HaArray(ea->children, i);
    if( child_e->u.int_val == 0 )
      ea->u.int_val = 0;
  }

  if( day_index < ea->day_range.last )
  {
    ** this not ea's last day; store value in next_soln **
    si = HaArray(ea->sig_indexes, day_index - ea->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == si,
      "KheDrsExprAndInternalDoDay internal error");
    HaArrayAddLast(next_soln->signature, ea->u.int_val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndDoDay(KHE_DRS_EXPR_AND ea, int day_index,              */
/*    KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active expression ea on day_index.                               */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprAndDoDay(KHE_DRS_EXPR_AND ea, int day_index,
  KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, start_index, stop_index;  KHE_DRS_EXPR child_e;

  ** initialize ea->value depending on whether day_index is its first day **
  if( day_index == ea->day_range.first )
    ea->u.int_val = 1;
  else
  {
    ** this is not ea's first day, so we need to retrieve a previous value **
    si = HaArray(ea->sig_indexes, day_index - 1 - ea->day_range.first);
    ea->u.int_val = HaArray(prev_soln->signature, prev_so + si);
  }

  ** accumulate the values of the children of ea that finalized today **
  start_index = HaArray(ea->open_child_indexes,
    day_index - ea->day_range.first);
  stop_index =  HaArray(ea->open_child_indexes,
    day_index + 1 - ea->day_range.first);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = HaArray(ea->children, i);
    if( child_e->u.int_val == 0 )
      ea->u.int_val = 0;
  }

  if( day_index < ea->day_range.last )
  {
    ** this not ea's last day; store value in next_soln **
    si = HaArray(ea->sig_indexes, day_index - ea->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == next_so + si,
      "KheDrsExprAndDoDay internal error");
    HaArrayAddLast(next_soln->signature, ea->u.int_val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAndDebug(KHE_DRS_EXPR_AND ea,                             */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for and expression ea.                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprAndDebug(KHE_DRS_EXPR_AND ea,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int si;
  if( soln != NULL )
  {
    /* unindented version with a soln value */
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) ea, "AND", drs, fp);
    if( KheDrsExprInSignature((KHE_DRS_EXPR) ea, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) ea, soln, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    /* indented version without a soln value */
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) ea, "AND", drs, fp);
    fprintf(fp, "\n");
    KheDrsExprDebugChildren((KHE_DRS_EXPR) ea, soln, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    /* unindented version without a soln value */
    fprintf(fp, "[ ");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) ea, "AND", drs, fp);
    fprintf(fp, " (%d children) ]", HaArrayCount(ea->children));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_SUM"                                         */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SUM KheDrsExprIntSumMake(KHE_DRS_DOM_TEST dom_test,     */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_SUM object with these attributes.            */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_INT_SUM KheDrsExprIntSumMake(KHE_DRS_DOM_TEST dom_test,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_SUM res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_INT_SUM_TAG, dom_test, drs);
  res->u.int_val = UNDEF_INT;
  res->closed_state = 0;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumFreeInternal(KHE_DRS_EXPR_INT_SUM eis,              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free eis and its internal descendants.                                   */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumFreeInternal(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(eis->children, e, i)
    KheDrsExprFreeInternal(e, drs);
  HaArrayAddLast(drs->free_expr_int_sum, eis);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumChildHasOpened(KHE_DRS_EXPR_INT_SUM eis,            */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform eis that its child_index'th child, child_e, has opened on         */
/*  day_index, and that its previous closed value was v.                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumChildHasOpened(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;  int i;
  eis->closed_state -= v.int_val;
  eis->open_children++;
  if( eis->open_children == 1 )
  {
    ** this expression has just opened **
    HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) eis);
    HaArrayForEach(eis->parents, prnt, i)
      KheDrsInternalExprChildHasOpened(prnt.expr, (KHE_DRS_EXPR) eis,
	prnt.index, day_index, eis->u, drs);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumChildHasClosed(KHE_DRS_EXPR_INT_SUM eis,            */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform eis that its child_index'th child, child_e, has closed on         */
/*  day_index, and that child_e's closed value is v.                         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumChildHasClosed(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v)
{
  KHE_DRS_PARENT prnt;  int i;
  eis->closed_state += v.int_val;
  eis->open_children--;
  if( eis->open_children == 0 )
  {
    ** this expression has just closed **
    eis->u.int_val = eis->closed_state;
    HaArrayForEach(eis->parents, prnt, i)
      KheDrsExprChildHasClosed(prnt.expr, (KHE_DRS_EXPR) eis,
	prnt.index, day_index, eis->u);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumChildHasOpened(KHE_DRS_EXPR_INT_SUM eis,            */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  As part of opening eis for solving, eis's child_index'th child, child_e, */
/*  has been opened.  Update eis's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of eis is the sum of the values of eis's s closed       */
/*  children, so this update subtracts child_e's value (which preserves      */
/*  its value before opening) from the closed state.                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumChildHasOpened(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DRS_EXPR child_e, int child_index)
{
  eis->closed_state -= child_e->u.int_val;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumChildHasClosed(KHE_DRS_EXPR_INT_SUM eis,            */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  As part of closing eis, eis's child_index'th child, child_e, has been    */
/*  closed.  Update eis's closed state to reflect this.                      */
/*                                                                           */
/*  The closed state of eis is the sum of the values of eis's s closed       */
/*  children, so this update adds child_e's value (which preserves           */
/*  its value before opening) to the closed state.                           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumChildHasClosed(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DRS_EXPR child_e, int child_index)
{
  eis->closed_state += child_e->u.int_val;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumSetValueFromClosedState(KHE_DRS_EXPR_INT_SUM eis,   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Here eis is closing.  Set its value, based on its closed state.          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumSetValueFromClosedState(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eis->u.int_val = eis->closed_state;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumSetClosedValue(KHE_DRS_EXPR_INT_SUM eis,            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eis, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumSetClosedValue(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eis->u.int_val = eis->closed_state;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumEvalDay(KHE_DRS_EXPR_INT_SUM eis,                   */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eis on day open_day_index.                           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumEvalDay(KHE_DRS_EXPR_INT_SUM eis,
  int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2;  KHE_DRS_EXPR child_e;

  if( KheDrsExprOpenDayIsFirst(eis, open_day_index) )
  {
    /* first day, so (ignoring closed_state) we have a 0 value here */
    eis->u.int_val = 0;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    eis->u.int_val = KheDrsExprSigVal(eis, open_day_index - 1, prev_soln);
  }

  /* accumulate the open values of the children of eis that finalized today */
  KheDrsExprForEachOpenDayChild(eis, open_day_index, child_e, i)
    eis->u.int_val += child_e->u.int_val;

  if( KheDrsExprOpenDayIsLast(eis, open_day_index) )
  {
    /* last day; add closed_state to the value */
    eis->u.int_val += eis->closed_state;
  }
  else
  {
    /* this not eis's last day; store value in next_soln */
    KheDrsExprPutSigVal(eis, open_day_index, next_soln, eis->u.int_val);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumInternalDoDay(KHE_DRS_EXPR_INT_SUM eis,             */
/*    int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Process internal expression eis on day_index.                            */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumInternalDoDay(KHE_DRS_EXPR_INT_SUM eis,
  int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, start_index, stop_index;  KHE_DRS_EXPR child_e;

  ** initialize eis->value depending on whether day_index is its first day **
  if( day_index == eis->day_range.first )
    eis->u.int_val = 0;
  else
  {
    ** this is not eis's first day, so we need to retrieve a previous value **
    si = HaArray(eis->sig_indexes, day_index - 1 - eis->day_range.first);
    eis->u.int_val = HaArray(prev_soln->signature, si);
  }

  ** accumulate the values of the children of eis that finalized today **
  start_index = HaArray(eis->open_child_indexes,
    day_index - eis->day_range.first);
  stop_index =  HaArray(eis->open_child_indexes,
    day_index+1 - eis->day_range.first);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = HaArray(eis->children, i);
    eis->u.int_val += child_e->u.int_val;
  }

  if( day_index < eis->day_range.last )
  {
    ** this not eis's last day; store value in next_soln **
    si = HaArray(eis->sig_indexes, day_index - eis->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == si,
      "KheDrsExprIntSumInternalDoDay internal error");
    HaArrayAddLast(next_soln->signature, eis->u.int_val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumDoDay(KHE_DRS_EXPR_INT_SUM eis, int day_index,      */
/*    KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active expression eis on day_index.                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumDoDay(KHE_DRS_EXPR_INT_SUM eis, int day_index,
  KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, start_index, stop_index;  KHE_DRS_EXPR child_e;

  ** initialize eis->value depending on whether day_index is its first day **
  if( day_index == eis->day_range.first )
    eis->u.int_val = 0;
  else
  {
    ** this is not eis's first day, so we need to retrieve a previous value **
    si = HaArray(eis->sig_indexes, day_index - 1 - eis->day_range.first);
    eis->u.int_val = HaArray(prev_soln->signature, prev_so + si);
  }

  ** accumulate the values of the children of eis that finalized today **
  start_index = HaArray(eis->open_child_indexes,
    day_index - eis->day_range.first);
  stop_index =  HaArray(eis->open_child_indexes,
    day_index+1 - eis->day_range.first);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = HaArray(eis->children, i);
    eis->u.int_val += child_e->u.int_val;
  }

  if( day_index < eis->day_range.last )
  {
    ** this not eis's last day; store value in next_soln **
    si = HaArray(eis->sig_indexes, day_index - eis->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == next_so + si,
      "KheDrsExprIntSumDoDay internal error");
    HaArrayAddLast(next_soln->signature, eis->u.int_val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumDebug(KHE_DRS_EXPR_INT_SUM eis,                     */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for busy time expression eis.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumDebug(KHE_DRS_EXPR_INT_SUM eis,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int si;
  if( soln != NULL )
  {
    /* unindented version with a soln value */
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eis, "INT_SUM", drs, fp);
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eis, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eis, soln, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    /* indented version without a soln value */
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eis, "INT_SUM", drs, fp);
    fprintf(fp, "\n");
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eis, soln, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    /* unindented version without a soln value */
    fprintf(fp, "[ ");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eis, "INT_SUM", drs, fp);
    fprintf(fp, " (%d children) ]", HaArrayCount(eis->children));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_SEQ"                                         */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SEQ KheDrsExprIntSeqMake(int history_before,            */
/*    int history_after, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_SEQ object with these attributes.            */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static KHE_DRS_EXPR_INT_SEQ KheDrsExprIntSeqMake(int history_before,
  int history_after, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_SEQ res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_INT_SEQ_TAG, dom_test, drs);
  res->history_before = history_before;
  res->history_after = history_after;
  res->tmp_value = UNDEF_INT;
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqFree(KHE_DRS_EXPR_INT_SEQ eis,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free eis and its descendants.                                            */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static void KheDrsExprIntSeqFree(KHE_DRS_EXPR_INT_SEQ eis,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(eis->children, e, i)
    KheDrsExprFree(e, drs);
  HaArrayAddLast(drs->free_expr_int_seq, eis);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_TEST KheDrsExprIntSeqDomTest(KHE_DRS_EXPR_INT_SEQ eis)       */
/*                                                                           */
/*  Return a suitable dom test for eis.                                      */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static KHE_DRS_DOM_TEST KheDrsExprIntSeqDomTest(KHE_DRS_EXPR_INT_SEQ eis)
{
  return KheExprMonotoneDomTest((KHE_DRS_EXPR) eis);  ** is this right? **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqBeginDay(KHE_DRS_EXPR_INT_SEQ eis, int day_index,   */
/*    KHE_DRS_SOLN prev_soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)               */
/*                                                                           */
/*  Inform eis that day day_index is beginning.                              */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static void KheDrsExprIntSeqBeginDay(KHE_DRS_EXPR_INT_SEQ eis, int day_index,
  KHE_DRS_SOLN prev_soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eis->tmp_value = KheDrsExprIntBeginDay((KHE_DRS_EXPR) eis, day_index,
    eis->history_before, prev_soln, drs, "IntSeq");
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqReceiveValueInt(KHE_DRS_EXPR_INT_SEQ eis,           */
/*    int value, KHE_DYNAMIC_RESOURCE_SOLVER drs)                            */
/*                                                                           */
/*  Inform eis that one of its children has just finalized its value.        */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static void KheDrsExprIntSeqReceiveValueInt(KHE_DRS_EXPR_INT_SEQ eis,
  int value, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HnAssert(value == 0 || value == 1,
    "KheDrsExprIntSeqReceiveValueInt internal error");
  if( value == 1 )
  {
    ** a sequence is beginning or continuing **
    eis->tmp_value++;
  }
  else if( eis->tmp_value > 0 )
  {
    ** a non-empty sequence has just ended **
    KheDrsExprReceiveValueInt(eis->parent, eis->tmp_value, drs);
    eis->tmp_value = 0;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqEndDay(KHE_DRS_EXPR_INT_SEQ eis, int day_index,     */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)                    */
/*                                                                           */
/*  Inform eis that day day_index is ending.                                 */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static void KheDrsExprIntSeqEndDay(KHE_DRS_EXPR_INT_SEQ eis, int day_index,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  eis->tmp_value = KheDrsExprIntEndDay((KHE_DRS_EXPR) eis, day_index,
    ** eis->tmp_value > 0, ** eis->tmp_value, soln, drs);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqDebug(KHE_DRS_EXPR_INT_SEQ eis,                     */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of eis onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; only KHE_DRS_EXPR_INT_SEQ_COMB_TAG is used
static void KheDrsExprIntSeqDebug(KHE_DRS_EXPR_INT_SEQ eis,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  KheDrsExprIntDebug((KHE_DRS_EXPR) eis, "INT_SEQ", eis->tmp_value,
    drs, verbosity, indent, fp);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_FLOAT_SUM"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FLOAT_SUM KheDrsExprFloatSumMake(                           */
/*    KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)            */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_FLOAT_SUM object with these attributes.          */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FLOAT_SUM KheDrsExprFloatSumMake(
  KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FLOAT_SUM res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_FLOAT_SUM_TAG, dom_test, drs);
  res->u.float_val = UNDEF_FLOAT;
  res->closed_state = 0.0;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumFreeInternal(KHE_DRS_EXPR_FLOAT_SUM efs,          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free efs and its internal descendants.                                   */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFloatSumFreeInternal(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(efs->children, e, i)
    KheDrsExprFreeInternal(e, drs);
  HaArrayAddLast(drs->free_expr_float_sum, efs);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumChildHasOpened(KHE_DRS_EXPR_FLOAT_SUM efs,        */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform efs that its child_index'th child, child_e, has opened on         */
/*  day_index, and that its previous closed value was v.                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFloatSumChildHasOpened(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;  int i;
  efs->closed_state -= v.float_val;
  efs->open_children++;
  if( efs->open_children == 1 )
  {
    ** this expression has just opened **
    HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) efs);
    HaArrayForEach(efs->parents, prnt, i)
      KheDrsInternalExprChildHasOpened(prnt.expr, (KHE_DRS_EXPR) efs,
	prnt.index, day_index, efs->u, drs);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumChildHasClosed(KHE_DRS_EXPR_FLOAT_SUM efs,        */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform efs that its child_index'th child, child_e, has closed on         */
/*  day_index, and that child_e's closed value is v.                         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFloatSumChildHasClosed(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v)
{
  KHE_DRS_PARENT prnt;  int i;
  efs->closed_state += v.float_val;
  efs->open_children--;
  if( efs->open_children == 0 )
  {
    ** this expression has just closed **
    efs->u.float_val = efs->closed_state;
    HaArrayForEach(efs->parents, prnt, i)
      KheDrsExprChildHasClosed(prnt.expr, (KHE_DRS_EXPR) efs,
	prnt.index, day_index, efs->u);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumChildHasOpened(KHE_DRS_EXPR_FLOAT_SUM efs,        */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  As part of opening efs for solving, efs's child_index'th child, child_e, */
/*  has been opened.  Update efs's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of efs is the sum of the values of efs's closed         */
/*  children, so this update subtracts child_e's value (which preserves      */
/*  its value before opening) from the closed state.                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatSumChildHasOpened(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DRS_EXPR child_e, int child_index)
{
  efs->closed_state -= child_e->u.float_val;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumChildHasClosed(KHE_DRS_EXPR_FLOAT_SUM efs,        */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  As part of opening efs for solving, efs's child_index'th child, child_e, */
/*  has been closed.  Update efs's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of efs is the sum of the values of efs's closed         */
/*  children, so this update adds child_e's value (which preserves           */
/*  its value before opening) to the closed state.                           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatSumChildHasClosed(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DRS_EXPR child_e, int child_index)
{
  efs->closed_state += child_e->u.float_val;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumSetValueFromClosedState(                          */
/*    KHE_DRS_EXPR_FLOAT_SUM efs, KHE_DYNAMIC_RESOURCE_SOLVER drs)           */
/*                                                                           */
/*  Here efs is closing.  Set its value, based on its closed state.          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFloatSumSetValueFromClosedState(
  KHE_DRS_EXPR_FLOAT_SUM efs, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  efs->u.float_val = efs->closed_state;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumSetClosedValue(KHE_DRS_EXPR_FLOAT_SUM efs,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing efs, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatSumSetClosedValue(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  efs->u.float_val = efs->closed_state;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumEvalDay(KHE_DRS_EXPR_FLOAT_SUM efs,               */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression efs on day open_day_index.                           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatSumEvalDay(KHE_DRS_EXPR_FLOAT_SUM efs,
  int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2;  KHE_DRS_EXPR child_e;

  if( KheDrsExprOpenDayIsFirst(efs, open_day_index) )
  {
    /* first day, so (ignoring closed_state) we have a 0 value here */
    efs->u.float_val = 0.0;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    efs->u.float_val =
      (float) KheDrsExprSigVal(efs, open_day_index - 1, prev_soln) / 100.0;
  }

  /* accumulate the open values of the children of efs that finalized today */
  KheDrsExprForEachOpenDayChild(efs, open_day_index, child_e, i)
    efs->u.float_val += child_e->u.float_val;

  if( KheDrsExprOpenDayIsLast(efs, open_day_index) )
  {
    /* last day; add closed_state to the value */
    efs->u.float_val += efs->closed_state;
  }
  else
  {
    /* this not efs's last day; store value in next_soln */
    KheDrsExprPutSigVal(efs, open_day_index, next_soln,
      (int) (efs->u.float_val * 100.0));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumInternalDoDay(KHE_DRS_EXPR_FLOAT_SUM efs,         */
/*    int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Process active expression efs on day_index.                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFloatSumInternalDoDay(KHE_DRS_EXPR_FLOAT_SUM efs,
  int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, start_index, stop_index;  KHE_DRS_EXPR child_e;

  ** initialize efs->value depending on whether day_index is its first day **
  if( day_index == efs->day_range.first )
    efs->u.float_val = 0.0;
  else
  {
    ** this is not efs's first day, so we need to retrieve a previous value **
    si = HaArray(efs->sig_indexes, day_index - 1 - efs->day_range.first);
    efs->u.float_val = (float) HaArray(prev_soln->signature, si) / 100.0;
  }

  ** accumulate the values of the children of efs that finalized today **
  start_index = HaArray(efs->open_child_indexes,
    day_index - efs->day_range.first);
  stop_index =  HaArray(efs->open_child_indexes,
    day_index+1 - efs->day_range.first);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = HaArray(efs->children, i);
    efs->u.float_val += child_e->u.float_val;
  }

  if( day_index < efs->day_range.last )
  {
    ** this not efs's last day; store value in next_soln **
    si = HaArray(efs->sig_indexes, day_index - efs->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == si,
      "KheDrsExprFloatSumInternalDoDay internal error");
    HaArrayAddLast(next_soln->signature, (int) (efs->u.float_val * 100.0));
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumDoDay(KHE_DRS_EXPR_FLOAT_SUM efs, int day_index,  */
/*    KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active expression efs on day_index.                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFloatSumDoDay(KHE_DRS_EXPR_FLOAT_SUM efs, int day_index,
  KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, start_index, stop_index;  KHE_DRS_EXPR child_e;

  ** initialize efs->value depending on whether day_index is its first day **
  if( day_index == efs->day_range.first )
    efs->u.float_val = 0.0;
  else
  {
    ** this is not efs's first day, so we need to retrieve a previous value **
    si = HaArray(efs->sig_indexes, day_index - 1 - efs->day_range.first);
    efs->u.float_val = (float) HaArray(prev_soln->signature, prev_so+si) /100.0;
  }

  ** accumulate the values of the children of efs that finalized today **
  start_index = HaArray(efs->open_child_indexes,
    day_index - efs->day_range.first);
  stop_index =  HaArray(efs->open_child_indexes,
    day_index+1 - efs->day_range.first);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = HaArray(efs->children, i);
    efs->u.float_val += child_e->u.float_val;
  }

  if( day_index < efs->day_range.last )
  {
    ** this not efs's last day; store value in next_soln **
    si = HaArray(efs->sig_indexes, day_index - efs->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == next_so + si,
      "KheDrsExprFloatSumDoDay internal error");
    HaArrayAddLast(next_soln->signature, (int) (efs->u.float_val * 100.0));
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatSumDebug(KHE_DRS_EXPR_FLOAT_SUM efs,                 */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of the header part of efs in soln onto fp.                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatSumDebug(KHE_DRS_EXPR_FLOAT_SUM efs,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int si;
  if( soln != NULL )
  {
    /* unindented version with a soln value */
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) efs, "FLOAT_SUM", drs, fp);
    if( KheDrsExprInSignature((KHE_DRS_EXPR) efs, soln->day, &si) )
      fprintf(fp, " $%d:%.2f$", si, (float) HaArray(soln->signature, si)/100.0);
    KheDrsExprDebugChildren((KHE_DRS_EXPR) efs, soln, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    /* indented version without a soln value */
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) efs, "FLOAT_SUM", drs, fp);
    fprintf(fp, "\n");
    KheDrsExprDebugChildren((KHE_DRS_EXPR) efs, soln, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    /* unindented version without a soln value */
    fprintf(fp, "[ ");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) efs, "FLOAT_SUM", drs, fp);
    fprintf(fp, " (%d children) ]", HaArrayCount(efs->children));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_DEV"                                         */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_DEV KheDrsExprIntDevMake(int min_limit,                 */
/*    int max_limit, bool allow_zero, KHE_DYNAMIC_RESOURCE_SOLVER drs)       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_DEV object with these attributes.            */
/*                                                                           */
/*  There is no dom_test because its type is always KHE_DRS_DOM_UNUSED.      */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_INT_DEV KheDrsExprIntDevMake(int min_limit,
  int max_limit, bool allow_zero, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_DEV res;  KHE_DRS_DOM_TEST dom_test;
  dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_INT_DEV_TAG, dom_test, drs);
  res->u.int_val = UNDEF_INT;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  res->allow_zero = allow_zero;
  /* res->closed_state = 0; */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevFreeInternal(KHE_DRS_EXPR_INT_DEV eid,              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free eid and its internal descendants.                                   */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntDevFreeInternal(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(eid->children, e, i)
    KheDrsExprFreeInternal(e, drs);
  HaArrayAddLast(drs->free_expr_int_dev, eid);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevChildHasOpened(KHE_DRS_EXPR_INT_DEV eid,            */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform eid that its child_index'th child, child_e, has opened on         */
/*  day_index, and that its previous closed value was v.                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntDevChildHasOpened(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;  int i;
  eid->closed_state = 0;
  eid->open_children++;
  if( eid->open_children == 1 )
  {
    ** this expression has just opened **
    HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) eid);
    HaArrayForEach(eid->parents, prnt, i)
      KheDrsInternalExprChildHasOpened(prnt.expr, (KHE_DRS_EXPR) eid,
	prnt.index, day_index, eid->u, drs);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevChildHasClosed(KHE_DRS_EXPR_INT_DEV eid,            */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform eid that its child_index'th child, child_e, has closed on         */
/*  day_index, and that child_e's closed value is v.                         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntDevChildHasClosed(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v)
{
  KHE_DRS_PARENT prnt;  int i;
  eid->closed_state = v.int_val;
  eid->open_children--;
  if( eid->open_children == 0 )
  {
    ** this expression has just closed **
    eid->u.int_val = KheDrsSumDev(eid->min_limit, eid->max_limit,
      eid->allow_zero, 0, eid->closed_state);
    HaArrayForEach(eid->parents, prnt, i)
      KheDrsExprChildHasClosed(prnt.expr, (KHE_DRS_EXPR) eid,
	prnt.index, day_index, eid->u);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevChildHasOpened(KHE_DRS_EXPR_INT_DEV eid,            */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  As part of opening eid for solving, eid's child_index'th child, child_e, */
/*  has been opened.  Update eid's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of eid is the value of eid's single closed child.       */
/*  This is not stored explicitly, so this update does nothing.              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntDevChildHasOpened(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DRS_EXPR child_e, int child_index)
{
  /* eid->closed_state = 0; */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevChildHasClosed(KHE_DRS_EXPR_INT_DEV eid,            */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  As part of opening eid for solving, eid's child_index'th child, child_e, */
/*  has been closed.  Update eid's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of eid is the value of eid's single closed child.       */
/*  This is not stored explicitly, so this update does nothing.              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntDevChildHasClosed(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DRS_EXPR child_e, int child_index)
{
  /* eid->closed_state = child_e->u.int_val; */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevSetValueFromClosedState(                            */
/*    KHE_DRS_EXPR_INT_DEV eid, KHE_DYNAMIC_RESOURCE_SOLVER drs)             */
/*                                                                           */
/*  Here eid is closing.  Set its value, based on its closed state.          */
/*  The closed state of eid is the value of eid's single closed child.       */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntDevSetValueFromClosedState(
  KHE_DRS_EXPR_INT_DEV eid, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR child_e;
  child_e = HaArrayFirst(eid->children);
  eid->u.int_val = KheDrsSumDev(eid->min_limit, eid->max_limit,
    eid->allow_zero, 0, child_e->u.int_val);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevSetClosedValue(KHE_DRS_EXPR_INT_DEV eid,            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eid, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntDevSetClosedValue(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR child_e;
  child_e = HaArrayFirst(eid->children);
  eid->u.int_val = KheDrsSumDev(eid->min_limit, eid->max_limit,
    eid->allow_zero, 0, child_e->u.int_val);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevEvalDay(KHE_DRS_EXPR_INT_DEV eid,                   */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eid on day open_day_index.                           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntDevEvalDay(KHE_DRS_EXPR_INT_DEV eid,
  int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2;  KHE_DRS_EXPR child_e;

  if( KheDrsExprOpenDayIsFirst(eid, open_day_index) )
  {
    /* first day */
    eid->u.int_val = 0;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    HnAbort("KheDrsExprIntDevEvalDay internal error (non-first day)");
  }

  /* accumulate the open values of the children of eid that finalized today */
  KheDrsExprForEachOpenDayChild(eid, open_day_index, child_e, i)
    eid->u.int_val += KheDrsSumDev(eid->min_limit, eid->max_limit,
      eid->allow_zero, 0, child_e->u.int_val);

  if( KheDrsExprOpenDayIsLast(eid, open_day_index) )
  {
    /* last day; add closed_state to the value */
    /* don't add anything because the sole child must be open so there */
    /* is no closed state to add in */
  }
  else
  {
    /* this not eid's last day; store value in next_soln */
    HnAbort("KheDrsExprIntDevEvalDay internal error (non-last day)");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevInternalDoDay(KHE_DRS_EXPR_INT_DEV eid,             */
/*    int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Process internal expression eid on day_index.                            */
/*                                                                           */
/*  Implementation note.  This code assumes (correctly) that there is a      */
/*  single child, which becomes final on eid'd first day.                    */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntDevInternalDoDay(KHE_DRS_EXPR_INT_DEV eid,
  int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;  KHE_DRS_EXPR e;

  ** initialize eid->value depending on whether day_index is its first day **
  if( day_index == eid->day_range.first )
  {
    e = HaArrayFirst(eid->children);
    eid->u.int_val = KheDrsSumDev(eid->min_limit, eid->max_limit,
      eid->allow_zero, 0, e->u.int_val);
  }
  else
  {
    ** this is not eid's first day, so we need to retrieve a previous value **
    si = HaArray(eid->sig_indexes, day_index - 1 - eid->day_range.first);
    eid->u.int_val = HaArray(prev_soln->signature, si);
  }

  if( day_index < eid->day_range.last )
  {
    ** this not eid's last day; store value in next_soln **
    si = HaArray(eid->sig_indexes, day_index - eid->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == si,
      "KheDrsExprIntDevInternalDoDay internal error");
    HaArrayAddLast(next_soln->signature, eid->u.int_val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevDoDay(KHE_DRS_EXPR_INT_DEV eid, int day_index,      */
/*    KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active expression eid on day_index.                              */
/*                                                                           */
/*  Implementation note.  This code assumes (correctly) that there is a      */
/*  single child, which becomes final on eid'd first day.                    */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntDevDoDay(KHE_DRS_EXPR_INT_DEV eid, int day_index,
  KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;  KHE_DRS_EXPR e;

  ** initialize eid->value depending on whether day_index is its first day **
  if( day_index == eid->day_range.first )
  {
    e = HaArrayFirst(eid->children);
    eid->u.int_val = KheDrsSumDev(eid->min_limit, eid->max_limit,
      eid->allow_zero, 0, e->u.int_val);
  }
  else
  {
    ** this is not eid's first day, so we need to retrieve a previous value **
    si = HaArray(eid->sig_indexes, day_index - 1 - eid->day_range.first);
    eid->u.int_val = HaArray(prev_soln->signature, prev_so + si);
  }

  if( day_index < eid->day_range.last )
  {
    ** this not eid's last day; store value in next_soln **
    si = HaArray(eid->sig_indexes, day_index - eid->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == next_so + si,
      "KheDrsExprIntDevDoDay internal error");
    HaArrayAddLast(next_soln->signature, eid->u.int_val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntDevDebug(KHE_DRS_EXPR_INT_DEV eid,                     */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of the header part of eid in soln onto fp.                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntDevDebug(KHE_DRS_EXPR_INT_DEV eid,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int si;
  if( soln != NULL )
  {
    /* unindented version with a soln value (actually no soln value here) */
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eid, "INT_DEV", drs, fp);
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eid, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eid, soln, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    /* indented version without a soln value */
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eid, "INT_DEV", drs, fp);
    fprintf(fp, "\n");
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eid, soln, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    /* unindented version without a soln value */
    fprintf(fp, "[ ");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eid, "INT_DEV", drs, fp);
    fprintf(fp, " (%d children) ]", HaArrayCount(eid->children));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_FLOAT_DEV"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_FLOAT_DEV KheDrsExprFloatDevMake(int min_limit,             */
/*    int max_limit, bool allow_zero, KHE_DYNAMIC_RESOURCE_SOLVER drs)       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_FLOAT_DEV object with these attributes.          */
/*                                                                           */
/*  There is no dom_test because its type is always KHE_DRS_DOM_UNUSED.      */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_FLOAT_DEV KheDrsExprFloatDevMake(int min_limit,
  int max_limit, bool allow_zero, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_FLOAT_DEV res;  KHE_DRS_DOM_TEST dom_test;
  dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_FLOAT_DEV_TAG, dom_test, drs);
  res->u.int_val = UNDEF_INT;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  res->allow_zero = allow_zero;
  /* res->closed_state = 0.0; */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevFreeInternal(KHE_DRS_EXPR_FLOAT_DEV efd,          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free efd and its internal descendants.                                   */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFloatDevFreeInternal(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(efd->children, e, i)
    KheDrsExprFreeInternal(e, drs);
  HaArrayAddLast(drs->free_expr_float_dev, efd);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheLimitWorkloadMonitorDev(KHE_DRS_EXPR_FLOAT_DEV efd,               */
/*    float workload)                                                        */
/*                                                                           */
/*  Return the deviation of workload wrt efd.                                */
/*                                                                           */
/*****************************************************************************/

static int KheLimitWorkloadMonitorDev(KHE_DRS_EXPR_FLOAT_DEV efd,
  float workload)
{
  if( efd->allow_zero && workload < 0.001 )
    return 0;
  else if( workload < efd->min_limit - 0.001 )
    return (int) ceil(efd->min_limit - 0.001 - workload);
  else if( workload > efd->max_limit + 0.001 )
    return (int) ceil(workload - efd->max_limit - 0.001);
  else
    return 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevChildHasOpened(KHE_DRS_EXPR_FLOAT_DEV efd,        */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform efd that its child_index'th child, child_e, has opened on         */
/*  day_index, and that its previous closed value was v.                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFloatDevChildHasOpened(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;  int i;
  efd->closed_state = 0.0;
  efd->open_children++;
  if( efd->open_children == 1 )
  {
    ** this expression has just opened **
    HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) efd);
    HaArrayForEach(efd->parents, prnt, i)
      KheDrsInternalExprChildHasOpened(prnt.expr, (KHE_DRS_EXPR) efd,
	prnt.index, day_index, efd->u, drs);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevChildHasClosed(KHE_DRS_EXPR_FLOAT_DEV efd,        */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform efd that its child_index'th child, child_e, has closed on         */
/*  day_index, and that child_e's closed value is v.                         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFloatDevChildHasClosed(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v)
{
  KHE_DRS_PARENT prnt;  int i;
  efd->closed_state = v.float_val;
  efd->open_children--;
  if( efd->open_children == 0 )
  {
    ** this expression has just closed **
    efd->u.int_val = KheLimitWorkloadMonitorDev(efd, efd->closed_state);
    HaArrayForEach(efd->parents, prnt, i)
      KheDrsExprChildHasClosed(prnt.expr, (KHE_DRS_EXPR) efd,
	prnt.index, day_index, efd->u);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevChildHasOpened(KHE_DRS_EXPR_FLOAT_DEV efd,        */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  As part of opening efd for solving, efd's child_index'th child, child_e, */
/*  has been opened.  Update efd's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of efd is the value of efd's single closed child,       */
/*  so this update needs to do nothing.                                      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatDevChildHasOpened(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DRS_EXPR child_e, int child_index)
{
  /* efd->closed_state = 0.0; */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevChildHasClosed(KHE_DRS_EXPR_FLOAT_DEV efd,        */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  As part of closing efd, efd's child_index'th child, child_e,             */
/*  has been closed.  Update efd's closed state to reflect this.             */
/*                                                                           */
/*  The closed state of efd is the value of efd's single closed child,       */
/*  so this update needs to do nothing.                                      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatDevChildHasClosed(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DRS_EXPR child_e, int child_index)
{
  /* efd->closed_state = child_e->u.float_val; */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevSetValueFromClosedState(                          */
/*    KHE_DRS_EXPR_FLOAT_DEV efd, KHE_DYNAMIC_RESOURCE_SOLVER drs)           */
/*                                                                           */
/*  Here efd is closing.  Set its value, based on its closed state.          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFloatDevSetValueFromClosedState(
  KHE_DRS_EXPR_FLOAT_DEV efd, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR child_e;
  child_e = HaArrayFirst(efd->children);
  efd->u.int_val = KheLimitWorkloadMonitorDev(efd, child_e->u.float_val);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevSetClosedValue(KHE_DRS_EXPR_FLOAT_DEV efd,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing efd, set its value suitably for its closed state.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatDevSetClosedValue(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR child_e;
  child_e = HaArrayFirst(efd->children);
  efd->u.int_val = KheLimitWorkloadMonitorDev(efd, child_e->u.float_val);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevEvalDay(KHE_DRS_EXPR_FLOAT_DEV efd,               */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression efd on day open_day_index.                           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatDevEvalDay(KHE_DRS_EXPR_FLOAT_DEV efd,
  int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2;  KHE_DRS_EXPR child_e;

  if( KheDrsExprOpenDayIsFirst(efd, open_day_index) )
  {
    /* first day */
    efd->u.int_val = 0;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    HnAbort("KheDrsExprFloatDevEvalDay internal error (non-first day)");
  }

  /* accumulate the open values of the children of efd that finalized today */
  KheDrsExprForEachOpenDayChild(efd, open_day_index, child_e, i)
    efd->u.int_val += KheLimitWorkloadMonitorDev(efd, child_e->u.float_val);

  if( KheDrsExprOpenDayIsLast(efd, open_day_index) )
  {
    /* last day; add closed_state to the value */
    /* don't add anything because the sole child must be open so there */
    /* is no closed state to add in */
  }
  else
  {
    /* this not efd's last day; store value in next_soln */
    HnAbort("KheDrsExprFloatDevEvalDay internal error (non-last day)");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevInternalDoDay(KHE_DRS_EXPR_FLOAT_DEV efd,         */
/*    int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Process internal expression efd on day_index.                            */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFloatDevInternalDoDay(KHE_DRS_EXPR_FLOAT_DEV efd,
  int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;  KHE_DRS_EXPR e;  ** NB child e has a float value **

  ** initialize efd->value depending on whether day_index is its first day **
  if( day_index == efd->day_range.first )
  {
    e = HaArrayFirst(efd->children);
    efd->u.int_val = KheLimitWorkloadMonitorDev(efd, e->u.float_val);
  }
  else
  {
    ** this is not efd's first day, so we need to retrieve a previous value **
    si = HaArray(efd->sig_indexes, day_index - 1 - efd->day_range.first);
    efd->u.int_val = HaArray(prev_soln->signature, si);
  }

  if( day_index < efd->day_range.last )
  {
    ** this not efd's last day; store value in next_soln **
    si = HaArray(efd->sig_indexes, day_index - efd->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == si,
      "KheDrsExprFloatDevInternalDoDay internal error");
    HaArrayAddLast(next_soln->signature, efd->u.int_val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevDoDay(KHE_DRS_EXPR_FLOAT_DEV efd, int day_index,  */
/*    KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active expression efd on day_index.                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFloatDevDoDay(KHE_DRS_EXPR_FLOAT_DEV efd, int day_index,
  KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si;  KHE_DRS_EXPR e;

  ** initialize efd->value depending on whether day_index is its first day **
  if( day_index == efd->day_range.first )
  {
    e = HaArrayFirst(efd->children);
    efd->u.int_val = KheLimitWorkloadMonitorDev(efd, e->u.float_val);
  }
  else
  {
    ** this is not efd's first day, so we need to retrieve a previous value **
    si = HaArray(efd->sig_indexes, day_index - 1 - efd->day_range.first);
    efd->u.int_val = HaArray(prev_soln->signature, prev_so + si);
  }

  if( day_index < efd->day_range.last )
  {
    ** this not efd's last day; store value in next_soln **
    si = HaArray(efd->sig_indexes, day_index - efd->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == next_so + si,
      "KheDrsExprFloatDevDoDay internal error");
    HaArrayAddLast(next_soln->signature, efd->u.int_val);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFloatDevDebug(KHE_DRS_EXPR_FLOAT_DEV efd,                 */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for float dev expression efd.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprFloatDevDebug(KHE_DRS_EXPR_FLOAT_DEV efd,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int si;
  if( soln != NULL )
  {
    /* unindented version with a soln value (actually no soln value here) */
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) efd, "FLOAT_DEV", drs, fp);
    if( KheDrsExprInSignature((KHE_DRS_EXPR) efd, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) efd, soln, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    /* indented version without a soln value */
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) efd, "FLOAT_DEV", drs, fp);
    fprintf(fp, "\n");
    KheDrsExprDebugChildren((KHE_DRS_EXPR) efd, soln, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    /* unindented version without a soln value */
    fprintf(fp, "[ ");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) efd, "FLOAT_DEV", drs, fp);
    fprintf(fp, " (%d children) ]", HaArrayCount(efd->children));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_COST"                                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_COST KheDrsExprCostMake(KHE_MONITOR m,                      */
/*    KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,                   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_COST object with these attributes.               */
/*                                                                           */
/*  This function has no dom_test parameter, because there is never any      */
/*  dominance testing at this expression.                                    */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_COST KheDrsExprCostMake(KHE_MONITOR m,
  KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_COST res;  KHE_DRS_DOM_TEST dom_test;
  dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_COST_TAG, dom_test, drs);
  res->monitor = m;
  res->cost_fn = cost_fn;
  res->combined_weight = combined_weight;
  res->u.int_val = -1;  /* no value at this expression */
  /* res->closed_state = 0; */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostFreeInternal(KHE_DRS_EXPR_COST ec,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free ec and its internal descendants.                                    */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprCostFreeInternal(KHE_DRS_EXPR_COST ec,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(ec->children, e, i)
    KheDrsExprFreeInternal(e, drs);
  HaArrayAddLast(drs->free_expr_cost, ec);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostChildHasOpened(KHE_DRS_EXPR_COST ec,                  */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform ec that its child_index'th child, child_e, has opened on          */
/*  day_index, and that its previous closed value was v.                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprCostChildHasOpened(KHE_DRS_EXPR_COST ec,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ** KHE_DRS_PARENT prnt;  int i; **
  ec->closed_state = 0;
  ec->open_children++;
  if( ec->open_children == 1 )
  {
    ** this expression has just opened **
    ** *** it has no parents and no value, at least, not at present
    HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) ec);
    HaArrayForEach(ec->parents, prnt, i)
      KheDrsInternalExprChildHasOpened(prnt.expr, (KHE_DRS_EXPR) ec,
	prnt.index, day_index, ec->u, drs);
    *** **
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostChildHasClosed(KHE_DRS_EXPR_COST ec,                  */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform ec that its child_index'th child, child_e, has closed on          */
/*  day_index, and that child_e's closed value is v.                         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprCostChildHasClosed(KHE_DRS_EXPR_COST ec,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v)
{
  ** KHE_DRS_PARENT prnt;  int i; **
  ec->closed_state = v.int_val;
  ec->open_children--;
  if( ec->open_children == 0 )
  {
    ** this expression has just closed **
    ** *** it has no parents and no value, at least, not at present
    cost = KheDrsCost(ec->cost_fn, ec->combined_weight, ec->closed_state);
    HaArrayForEach(ec->parents, prnt, i)
      KheDrsExprChildHasClosed(prnt.expr, (KHE_DRS_EXPR) ec,
	prnt.index, day_index, ec->u);
    *** **
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostChildHasOpened(KHE_DRS_EXPR_COST ec,                  */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  As part of opening ec for solving, ec's child_index'th child, child_e,   */
/*  has been opened.  Update ec's closed state to reflect this.              */
/*                                                                           */
/*  The closed state of ec is the value of ec's single closed child, so      */
/*  this update does nothing.                                                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprCostChildHasOpened(KHE_DRS_EXPR_COST ec,
  KHE_DRS_EXPR child_e, int child_index)
{
  /* ec->closed_state = 0; */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostChildHasClosed(KHE_DRS_EXPR_COST ec,                  */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  As part of closing ec, ec's child_index'th child, child_e,               */
/*  has been closed.  Update ec's closed state to reflect this.              */
/*                                                                           */
/*  The closed state of ec is the value of ec's single closed child, so      */
/*  this update does nothing.                                                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprCostChildHasClosed(KHE_DRS_EXPR_COST ec,
  KHE_DRS_EXPR child_e, int child_index)
{
  /* ec->closed_state = child_e->u.int_val; */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostSetValueFromClosedState(                              */
/*    KHE_DRS_EXPR_COST ec, KHE_DYNAMIC_RESOURCE_SOLVER drs)                 */
/*                                                                           */
/*  Here ec is closing.  Set its value, based on its closed state.           */
/*                                                                           */
/*  The cost here needs to be reported directly to somewhere; I don't        */
/*  know where yet, that's still to do.                                      */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprCostSetValueFromClosedState(
  KHE_DRS_EXPR_COST ec, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ** *** lhs here still to do; rhs is correct
  KHE_DRS_EXPR child_e;
  child_e = HaArrayFirst(ec->children);
  somewhere += KheDrsCost(ec->cost_fn, ec->combined_weight, child_e->u.int_val);
  *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostSetClosedValue(KHE_DRS_EXPR_COST ec,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing ec, set its value suitably for its closed state.      */
/*                                                                           */
/*  The cost here needs to be reported directly to somewhere; I don't        */
/*  know where yet, that's still to do.                                      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprCostSetClosedValue(KHE_DRS_EXPR_COST ec,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* *** lhs here still to do; rhs is correct
  KHE_DRS_EXPR child_e;
  child_e = HaArrayFirst(ec->children);
  somewhere += KheDrsCost(ec->cost_fn, ec->combined_weight, child_e->u.int_val);
  *** */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostEvalDay(KHE_DRS_EXPR_COST ec,                         */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression ec on day open_day_index.                            */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprCostEvalDay(KHE_DRS_EXPR_COST ec,
  int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2;  KHE_DRS_EXPR child_e;  KHE_COST cost;

  if( KheDrsExprOpenDayIsFirst(ec, open_day_index) )
  {
    /* first day */
    cost = KheCost(0, 0);
  }
  else
  {
    /* not first day, so retrieve a previous value */
    HnAbort("KheDrsExprCostEvalDay internal error (non-first day)");
    cost =  KheCost(0, 0);  /* keep compiler happy */
  }

  /* accumulate the open values of the children of ec that finalized today */
  KheDrsExprForEachOpenDayChild(ec, open_day_index, child_e, i)
    cost += KheDrsCost(ec->cost_fn, ec->combined_weight, child_e->u.int_val);

  if( KheDrsExprOpenDayIsLast(ec, open_day_index) )
  {
    /* last day; add closed_state to the value */
    /* don't add anything because the sole child must be open so there */
    /* is no closed state to add in */
    if( cost > 0 )
    {
      if( drs->solve_rerun )
	fprintf(stderr, "  %s reporting cost %.5f\n",
	  KheMonitorId(ec->monitor), KheCostShow(cost));
      next_soln->cost += cost;
    }
  }
  else
  {
    /* this not ec's last day; store value in next_soln */
    HnAbort("KheDrsExprCostEvalDay internal error (non-last day)");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostInternalDoDay(KHE_DRS_EXPR_COST ec,                   */
/*    int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Process internal expression ec on day_index.                             */
/*                                                                           */
/*  Implementation note.  Cost expressions always report their values        */
/*  directly to the new soln as soon as they are available.  They do         */
/*  not store values for retrieval by a parent; they have no parents.        */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprCostInternalDoDay(KHE_DRS_EXPR_COST ec,
  int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_COST cost;  KHE_DRS_EXPR e;
  e = HaArrayFirst(ec->children);
  cost = KheDrsCost(ec->cost_fn, ec->combined_weight, e->u.int_val);
  if( cost > 0 )
  {
    if( drs->solve_rerun )
      fprintf(stderr, "  %s reporting cost %.5f\n",
	KheMonitorId(ec->monitor), KheCostShow(cost));
    next_soln->cost += cost;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostDoDay(KHE_DRS_EXPR_COST ec, int day_index,            */
/*    KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active expression ec on day_index.                               */
/*                                                                           */
/*  Implementation note.  Cost expressions always report their values        */
/*  directly to the new soln as soon as they are available.  They do         */
/*  not store values for retrieval by a parent; they have no parents.        */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprCostDoDay(KHE_DRS_EXPR_COST ec, int day_index,
  KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_COST cost;  KHE_DRS_EXPR e;
  e = HaArrayFirst(ec->children);
  cost = KheDrsCost(ec->cost_fn, ec->combined_weight, e->u.int_val);
  if( cost > 0 )
  {
    if( drs->solve_rerun )
      fprintf(stderr, "  %s reporting cost %.5f\n",
	KheMonitorId(ec->monitor), KheCostShow(cost));
    next_soln->cost += cost;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprCostDebug(KHE_DRS_EXPR_COST ec,                           */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for cost expression ec.                        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprCostDebug(KHE_DRS_EXPR_COST ec,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int si;
  if( soln != NULL )
  {
    /* unindented version with a soln value */
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) ec, "COST", drs, fp);
    if( KheDrsExprInSignature((KHE_DRS_EXPR) ec, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) ec, soln, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    /* indented version without a soln value */
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) ec, "COST", drs, fp);
    fprintf(fp, "\n");
    KheDrsExprDebugChildren((KHE_DRS_EXPR) ec, soln, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    /* unindented version without a soln value */
    fprintf(fp, "[ ");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) ec, "COST", drs, fp);
    fprintf(fp, " (%d children) ]", HaArrayCount(ec->children));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_SUM_COMB"                                    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_COMB_TYPE KheDrsCombType(KHE_COST_FUNCTION cost_fn,              */
/*    int max_limit)                                                         */
/*                                                                           */
/*  Return the comb type, as determined by the cost function and max limit.  */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_COMB_TYPE KheDrsCombType(KHE_COST_FUNCTION cost_fn,
  int max_limit)
{
  if( max_limit == INT_MAX )
    return KHE_DRS_COMB_NO_MAX;
  else if( cost_fn == KHE_LINEAR_COST_FUNCTION )
    return KHE_DRS_COMB_LINEAR;
  else if( cost_fn == KHE_STEP_COST_FUNCTION )
    return KHE_DRS_COMB_STEP;
  else
    return KHE_DRS_COMB_ORDINARY;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SUM_COMB KheDrsExprIntSumCombMake(KHE_MONITOR m,        */
/*    KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,                   */
/*    int min_limit, int max_limit, bool allow_zero, int history_before,     */
/*    int history_after, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_SUM_COMB object with these attributes.       */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_INT_SUM_COMB KheDrsExprIntSumCombMake(KHE_MONITOR m,
  KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,
  int min_limit, int max_limit, bool allow_zero, int history_before,
  int history_after, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_SUM_COMB res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_INT_SUM_COMB_TAG, dom_test, drs);
  res->u.int_val = UNDEF_INT;
  res->monitor = m;
  res->cost_fn = cost_fn;
  res->combined_weight = combined_weight;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  res->allow_zero = allow_zero;
  res->comb_type = KheDrsCombType(cost_fn, max_limit);
  res->history_before = history_before;
  res->history_after = history_after;
  res->closed_state = 0;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombFreeInternal(KHE_DRS_EXPR_INT_SUM_COMB eisc,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free eisc and its internal descendants.                                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumCombFreeInternal(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(eisc->children, e, i)
    KheDrsExprFreeInternal(e, drs);
  HaArrayAddLast(drs->free_expr_int_sum_comb, eisc);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombChildHasOpened(KHE_DRS_EXPR_INT_SUM_COMB eisc,  */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform eisc that its child_index'th child, child_e, has opened on        */
/*  day_index, and that its previous closed value was v.                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumCombChildHasOpened(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ** KHE_DRS_PARENT prnt;  int i; **
  eisc->closed_state -= v.int_val;
  eisc->open_children++;
  if( eisc->open_children == 1 )
  {
    ** this expression has just opened **
    ** *** it has no parents and no value, at least, not at present
    HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) eisc);
    HaArrayForEach(eisc->parents, prnt, i)
      KheDrsInternalExprChildHasOpened(prnt.expr, (KHE_DRS_EXPR) eisc,
	prnt.index, day_index, eisc->u, drs);
    *** **
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombChildHasClosed(KHE_DRS_EXPR_INT_SUM_COMB eisc,  */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform eisc that its child_index'th child, child_e, has closed on        */
/*  day_index, and that child_e's closed value is v.                         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumCombChildHasClosed(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v)
{
  ** KHE_DRS_PARENT prnt;  int i; **
  eisc->closed_state += v.int_val;
  eisc->open_children--;
  if( eisc->open_children == 0 )
  {
    ** this expression has just closed **
    ** *** it has no parents and no value, at least, not at present
    dev = KheDrsSumDev(eisc->min_limit, eisc->max_limit, eisc->allow_zero,
      eisc->history_after, eisc->u.int_val);
    cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
    HaArrayForEach(eisc->parents, prnt, i)
      KheDrsExprChildHasClosed(prnt.expr, (KHE_DRS_EXPR) eisc,
	prnt.index, day_index, eisc->u);
    *** **
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombChildHasOpened(KHE_DRS_EXPR_INT_SUM_COMB eisc,  */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  As part of opening eisc for solving, eisc's child_index'th child,        */
/*  child_e, has been opened.  Update eisc's closed state to reflect this.   */
/*                                                                           */
/*  The closed state of eisc is the sum of the values of eisc's closed       */
/*  children, so this update subtracts child_e's value (which preserves      */
/*  its value before opening) from the closed state.                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumCombChildHasOpened(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  KHE_DRS_EXPR child_e, int child_index)
{
  eisc->closed_state -= child_e->u.int_val;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombChildHasClosed(KHE_DRS_EXPR_INT_SUM_COMB eisc,  */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  As part of closing eisc, eisc's child_index'th child, child_e, has       */
/*  been opened.  Update eisc's closed state to reflect this.                */
/*                                                                           */
/*  The closed state of eisc is the sum of the values of eisc's closed       */
/*  children, so this update adds child_e's value to the closed state.       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumCombChildHasClosed(
  KHE_DRS_EXPR_INT_SUM_COMB eisc, KHE_DRS_EXPR child_e, int child_index)
{
  eisc->closed_state += child_e->u.int_val;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombSetValueFromClosedState(                        */
/*    KHE_DRS_EXPR_INT_SUM_COMB eisc, KHE_DYNAMIC_RESOURCE_SOLVER drs)       */
/*                                                                           */
/*  Here eisc is closing.  Set its value, based on its closed state.         */
/*                                                                           */
/*  The cost here needs to be reported directly to somewhere; I don't        */
/*  know where yet, that's still to do.                                      */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumCombSetValueFromClosedState(
  KHE_DRS_EXPR_INT_SUM_COMB eisc, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int dev;  KHE_COST cost;
  dev = KheDrsSumDev(eisc->min_limit, eisc->max_limit, eisc->allow_zero,
    eisc->history_after, eisc->closed_state);
  if( dev > 0 )
  {
    cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
    if( cost > 0 )
    {
      ** reporting cost here still to do **
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombSetClosedValue(KHE_DRS_EXPR_INT_SUM_COMB eisc,  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eisc, set its value suitably for its closed state.    */
/*                                                                           */
/*  The cost here needs to be reported directly to somewhere; I don't        */
/*  know where yet, that's still to do.                                      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumCombSetClosedValue(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int dev;  KHE_COST cost;
  dev = KheDrsSumDev(eisc->min_limit, eisc->max_limit, eisc->allow_zero,
    eisc->history_after, eisc->closed_state);
  if( dev > 0 )
  {
    cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
    if( cost > 0 )
    {
      /* reporting cost here still to do */
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombEvalDay(KHE_DRS_EXPR_INT_SUM_COMB eisc,         */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eisc on day open_day_index.                          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumCombEvalDay(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, i1, i2, dev, val;  KHE_DRS_EXPR child_e;  KHE_COST cost;

  if( KheDrsExprOpenDayIsFirst(eisc, open_day_index) )
  {
    /* first day */
    val = eisc->history_before + eisc->closed_state;
  }
  else
  {
    /* not first day, so retrieve a previous value */
    val = KheDrsExprSigVal(eisc, open_day_index - 1, prev_soln);
  }

  /* accumulate the open values of the children of eisc that finalized today */
  KheDrsExprForEachOpenDayChild(eisc, open_day_index, child_e, i)
    val += child_e->u.int_val;

  /* adjust in various special cases */
  switch( eisc->comb_type )
  {
    case KHE_DRS_COMB_ORDINARY:

      /* no adjustment here */
      break;

    case KHE_DRS_COMB_NO_MAX:

      /* no maximum limit; store minimum limit at most */
      if( val > eisc->min_limit )
	val = eisc->min_limit;
      break;

    case KHE_DRS_COMB_LINEAR:

      /* linear cost; if over maximum, report excess immediately */
      if( val > eisc->max_limit )
      {
	cost = (val - eisc->max_limit) * eisc->combined_weight;
	if( cost > 0 )
	{
	  if( drs->solve_rerun )
	    fprintf(stderr, "  %s reporting linear cost %.5f\n",
	      KheMonitorId(eisc->monitor), KheCostShow(cost));
	  next_soln->cost += cost;
	}
	val = eisc->max_limit;
      }
      break;

    case KHE_DRS_COMB_STEP:

      /* step cost; store maximum plus one at most */
      if( val > eisc->max_limit + 1 )
	val = eisc->max_limit + 1;
      break;

    default:

      HnAbort("KheDrsExprIntSumCombEvalDay internal error");
      break;
  }

  if( KheDrsExprOpenDayIsLast(eisc, open_day_index) )
  {
    /* this eisc's last day; report any cost immediately to next_soln */
    dev = KheDrsSumDev(eisc->min_limit, eisc->max_limit, eisc->allow_zero,
      eisc->history_after, val);
    if( dev > 0 )
    {
      cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
      if( cost > 0 )
      {
	if( drs->solve_rerun )
	  fprintf(stderr, "  %s reporting cost %.5f\n",
	    KheMonitorId(eisc->monitor), KheCostShow(cost));
	next_soln->cost += cost;
      }
    }
  }
  else
  {
    /* this not eisc's last day; store value in next_soln */
    KheDrsExprPutSigVal(eisc, open_day_index, next_soln, val);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombInternalDoDay(KHE_DRS_EXPR_INT_SUM_COMB eisc,   */
/*    int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Process active expression eisc on day_index.                             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumCombInternalDoDay(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, start_index, stop_index, dev;
  KHE_DRS_EXPR child_e;  KHE_COST cost;

  ** initialize eisc->value depending on whether day_index is its first day **
  if( day_index == eisc->day_range.first )
    eisc->u.int_val = eisc->history_before;
  else
  {
    ** this is not eisc's first day, so we need to retrieve a previous value **
    si = HaArray(eisc->sig_indexes, day_index - 1 - eisc->day_range.first);
    eisc->u.int_val = HaArray(prev_soln->signature, si);
  }

  ** accumulate the values of the children of eisc that finalized today **
  start_index = HaArray(eisc->open_child_indexes,
    day_index - eisc->day_range.first);
  stop_index =  HaArray(eisc->open_child_indexes,
    day_index+1-eisc->day_range.first);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = HaArray(eisc->children, i);
    eisc->u.int_val += child_e->u.int_val;
  }

  ** adjust in various special cases **
  switch( eisc->comb_type )
  {
    case KHE_DRS_COMB_ORDINARY:

      ** no adjustment here **
      break;

    case KHE_DRS_COMB_NO_MAX:

      ** no maximum limit; store minimum limit at most **
      if( eisc->u.int_val > eisc->min_limit )
	eisc->u.int_val = eisc->min_limit;
      break;

    case KHE_DRS_COMB_LINEAR:

      ** linear cost; if over maximum, report excess immediately **
      if( eisc->u.int_val > eisc->max_limit )
      {
	cost = (eisc->u.int_val - eisc->max_limit) * eisc->combined_weight;
	if( cost > 0 )
	{
	  if( drs->solve_rerun )
	    fprintf(stderr, "  %s reporting linear cost %.5f\n",
	      KheMonitorId(eisc->monitor), KheCostShow(cost));
	  next_soln->cost += cost;
	}
	eisc->u.int_val = eisc->max_limit;
      }
      break;

    case KHE_DRS_COMB_STEP:

      ** step cost; store maximum plus one at most **
      if( eisc->u.int_val > eisc->max_limit + 1 )
	eisc->u.int_val = eisc->max_limit + 1;
      break;

    default:

      HnAbort("KheDrsExprIntSumCombDoDay internal error");
      break;
  }

  if( day_index < eisc->day_range.last )
  {
    ** this not eisc's last day; store value in next_soln **
    si = HaArray(eisc->sig_indexes, day_index - eisc->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == si,
      "KheDrsExprIntSumCombInternalDoDay internal error");
    HaArrayAddLast(next_soln->signature, eisc->u.int_val);
  }
  else
  {
    ** this eisc's last day; report any cost immediately to next_soln **
    dev = KheDrsSumDev(eisc->min_limit, eisc->max_limit, eisc->allow_zero,
      eisc->history_after, eisc->u.int_val);
    cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
    if( cost > 0 )
    {
      if( drs->solve_rerun )
	fprintf(stderr, "  %s reporting cost %.5f\n",
	  KheMonitorId(eisc->monitor), KheCostShow(cost));
      next_soln->cost += cost;
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombDoDay(KHE_DRS_EXPR_INT_SUM_COMB eisc,           */
/*    int day_index,                                                         */
/*    KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active expression eisc on day_index.                             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSumCombDoDay(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  int day_index,
  KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, start_index, stop_index, dev;
  KHE_DRS_EXPR child_e;  KHE_COST cost;

  ** initialize eisc->value depending on whether day_index is its first day **
  if( day_index == eisc->day_range.first )
    eisc->u.int_val = eisc->history_before;
  else
  {
    ** this is not eisc's first day, so we need to retrieve a previous value **
    si = HaArray(eisc->sig_indexes, day_index - 1 - eisc->day_range.first);
    eisc->u.int_val = HaArray(prev_soln->signature, prev_so + si);
  }

  ** accumulate the values of the children of eisc that finalized today **
  start_index = HaArray(eisc->open_child_indexes,
    day_index - eisc->day_range.first);
  stop_index =  HaArray(eisc->open_child_indexes,
    day_index+1-eisc->day_range.first);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = HaArray(eisc->children, i);
    eisc->u.int_val += child_e->u.int_val;
  }

  ** adjust in various special cases **
  switch( eisc->comb_type )
  {
    case KHE_DRS_COMB_ORDINARY:

      ** no adjustment here **
      break;

    case KHE_DRS_COMB_NO_MAX:

      ** no maximum limit; store minimum limit at most **
      if( eisc->u.int_val > eisc->min_limit )
	eisc->u.int_val = eisc->min_limit;
      break;

    case KHE_DRS_COMB_LINEAR:

      ** linear cost; if over maximum, report excess immediately **
      if( eisc->u.int_val > eisc->max_limit )
      {
	cost = (eisc->u.int_val - eisc->max_limit) * eisc->combined_weight;
	if( cost > 0 )
	{
	  if( drs->solve_rerun )
	    fprintf(stderr, "  %s reporting linear cost %.5f\n",
	      KheMonitorId(eisc->monitor), KheCostShow(cost));
	  next_soln->cost += cost;
	}
	eisc->u.int_val = eisc->max_limit;
      }
      break;

    case KHE_DRS_COMB_STEP:

      ** step cost; store maximum plus one at most **
      if( eisc->u.int_val > eisc->max_limit + 1 )
	eisc->u.int_val = eisc->max_limit + 1;
      break;

    default:

      HnAbort("KheDrsExprIntSumCombDoDay internal error");
      break;
  }

  if( day_index < eisc->day_range.last )
  {
    ** this not eisc's last day; store value in next_soln **
    si = HaArray(eisc->sig_indexes, day_index - eisc->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == next_so + si,
      "KheDrsExprIntSumCombDoDay internal error");
    HaArrayAddLast(next_soln->signature, eisc->u.int_val);
  }
  else
  {
    ** this eisc's last day; report any cost immediately to next_soln **
    dev = KheDrsSumDev(eisc->min_limit, eisc->max_limit, eisc->allow_zero,
      eisc->history_after, eisc->u.int_val);
    cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
    if( cost > 0 )
    {
      if( drs->solve_rerun )
	fprintf(stderr, "  %s reporting cost %.5f\n",
	  KheMonitorId(eisc->monitor), KheCostShow(cost));
      next_soln->cost += cost;
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSumCombDebug(KHE_DRS_EXPR_INT_SUM_COMB eisc,           */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Implement KheDrsExprDebug for int sum comb expression eisc.              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSumCombDebug(KHE_DRS_EXPR_INT_SUM_COMB eisc,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int si;
  if( soln != NULL )
  {
    /* unindented version with a soln value */
     KheDrsExprHeaderDebug((KHE_DRS_EXPR) eisc, "INT_SUM_COMB", drs, fp);
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eisc, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eisc, soln, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    /* indented version without a soln value */
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eisc, "INT_SUM_COMB", drs, fp);
    fprintf(fp, "\n");
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eisc, soln, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    /* unindented version without a soln value */
    fprintf(fp, "[ ");
    KheDrsExprHeaderDebug((KHE_DRS_EXPR) eisc, "INT_SUM_COMB", drs, fp);
    fprintf(fp, " (%d children) ]", HaArrayCount(eisc->children));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR_INT_SEQ_COMB"                                    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR_INT_SEQ_COMB KheDrsExprIntSeqCombMake(KHE_MONITOR m,        */
/*    KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight,                   */
/*    KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight, int min_limit,    */
/*    int max_limit, int history_before, int history_after,                  */
/*    KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)            */
/*                                                                           */
/*  Make a new KHE_DRS_EXPR_INT_SEQ_COMB object with these attributes.       */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR_INT_SEQ_COMB KheDrsExprIntSeqCombMake(KHE_MONITOR m,
  KHE_COST_FUNCTION cost_fn, KHE_COST combined_weight, int min_limit,
  int max_limit, int history_before, int history_after,
  KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_SEQ_COMB res;
  KheDrsExprInitBegin(res, KHE_DRS_EXPR_INT_SEQ_COMB_TAG, dom_test, drs);
  res->u.int_val = UNDEF_INT;
  res->monitor = m;
  res->cost_fn = cost_fn;
  res->combined_weight = combined_weight;
  res->min_limit = min_limit;
  res->max_limit = max_limit;
  res->comb_type = KheDrsCombType(cost_fn, max_limit);
  res->history_before = history_before;
  res->history_after = history_after;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombFreeInternal(KHE_DRS_EXPR_INT_SEQ_COMB eisc,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free eisc and its internal descendants.                                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSeqCombFreeInternal(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(eisc->children, e, i)
    KheDrsExprFreeInternal(e, drs);
  HaArrayAddLast(drs->free_expr_int_seq_comb, eisc);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombChildHasOpened(KHE_DRS_EXPR_INT_SEQ_COMB eisc,  */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform eisc that its child_index'th child, child_e, has opened on        */
/*  day_index, and that its previous closed value was v.                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSeqCombChildHasOpened(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ** st ill to do **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombChildHasClosed(KHE_DRS_EXPR_INT_SEQ_COMB eisc,  */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform eisc that its child_index'th child, child_e, has closed on        */
/*  day_index, and that child_e's closed value is v.                         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSeqCombChildHasClosed(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v)
{
  ** sti ll to do **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombChildHasOpened(KHE_DRS_EXPR_INT_SEQ_COMB eisc,  */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  As part of opening eisc for solving, eisc's child_index'th child,        */
/*  child_e, has been opened.  Update eisc's closed state to reflect this.   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCombChildHasOpened(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  KHE_DRS_EXPR child_e, int child_index)
{
  /* still to do */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombChildHasClosed(KHE_DRS_EXPR_INT_SEQ_COMB eisc,  */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  As part of closing eisc, eisc's child_index'th child, child_e, has       */
/*  been opened.  Update eisc's closed state to reflect this.                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCombChildHasClosed(
  KHE_DRS_EXPR_INT_SEQ_COMB eisc, KHE_DRS_EXPR child_e, int child_index)
{
  /* still to do */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombSetValueFromClosedState(                        */
/*    KHE_DRS_EXPR_INT_SEQ_COMB eisc, KHE_DYNAMIC_RESOURCE_SOLVER drs)       */
/*                                                                           */
/*  Here eisc is closing.  Set its value, based on its closed state.         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSeqCombSetValueFromClosedState(
  KHE_DRS_EXPR_INT_SEQ_COMB eisc, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ** still to do **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombSetClosedValue(KHE_DRS_EXPR_INT_SEQ_COMB eisc,  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing eisc, set its value suitably for its closed state.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCombSetClosedValue(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* still to do */
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EXPR KheDrsExprOptionalCopy(KHE_DRS_EXPR e,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  If e is a BUSY_TIME, FREE_TIME, WORK_TIME, BUSY_DAY, FREE_DAY, or        */
/*  WORK_DAY expression, return a fresh copy of it.  Otherwise return e.     */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_EXPR KheDrsExprOptionalCopy(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      return (KHE_DRS_EXPR) KheDrsExprBusyTimeCopy((KHE_DRS_EXPR_BUSY_TIME) e,
	drs);

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      return (KHE_DRS_EXPR) KheDrsExprFreeTimeCopy((KHE_DRS_EXPR_FREE_TIME) e,
	drs);

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      return (KHE_DRS_EXPR) KheDrsExprWorkTimeCopy((KHE_DRS_EXPR_WORK_TIME) e,
	drs);

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      return (KHE_DRS_EXPR) KheDrsExprBusyDayCopy((KHE_DRS_EXPR_BUSY_DAY) e,
	drs);

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      return (KHE_DRS_EXPR) KheDrsExprFreeDayCopy((KHE_DRS_EXPR_FREE_DAY) e,
	drs);

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      return (KHE_DRS_EXPR) KheDrsExprWorkDayCopy((KHE_DRS_EXPR_WORK_DAY) e,
	drs);

    default:

      /* other expression types are not shared, just return e */
      return e;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombEvalDay(KHE_DRS_EXPR_INT_SEQ_COMB eisc,         */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression eisc on day open_day_index.                          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCombEvalDay(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* still to do */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombInternalDoDay(KHE_DRS_EXPR_INT_SEQ_COMB eisc,   */
/*    int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Process active expression eisc on day_index.                             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSeqCombInternalDoDay(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  int day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, dev, start_index, stop_index;
  KHE_DRS_EXPR child_e;  KHE_COST cost;

  ** initialize eisc->value depending on whether day_index is its first day **
  if( day_index == eisc->day_range.first )
    eisc->u.int_val = eisc->history_before;
  else
  {
    ** this is not eisc's first day, so we need to retrieve a previous value **
    si = HaArray(eisc->sig_indexes, day_index-1 - eisc->day_range.first);
    eisc->u.int_val = HaArray(prev_soln->signature, si);
  }

  ** accumulate the values of the children of eisc that finalized today **
  start_index = HaArray(eisc->open_child_indexes,
    day_index - eisc->day_range.first);
  stop_index =  HaArray(eisc->open_child_indexes,
    day_index+1-eisc->day_range.first);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = HaArray(eisc->children, i);
    if( child_e->u.int_val == 1 )
    {
      ** a sequence is beginning or continuing **
      eisc->u.int_val++;

      ** adjust in various special cases **
      switch( eisc->comb_type )
      {
	case KHE_DRS_COMB_ORDINARY:

	  ** no adjustment here **
	  break;

	case KHE_DRS_COMB_NO_MAX:

	  ** no maximum limit; store minimum limit at most **
	  if( eisc->u.int_val > eisc->min_limit )
	    eisc->u.int_val = eisc->min_limit;
	  break;

	case KHE_DRS_COMB_LINEAR:

	  ** linear cost; if over maximum, report excess immediately **
	  if( eisc->u.int_val > eisc->max_limit )
	  {
	    dev = KheDrsSeqDev(eisc->min_limit, eisc->max_limit, 0,
	      eisc->u.int_val);
	    cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
	    if( cost > 0 )
	    {
	      if( drs->solve_rerun )
		fprintf(stderr, "  %s reporting linear cost %.5f\n",
		  KheMonitorId(eisc->monitor), KheCostShow(cost));
	      next_soln->cost += cost;
	    }
	    eisc->u.int_val = eisc->max_limit;
	  }
	  break;

	case KHE_DRS_COMB_STEP:

	  ** step cost; store maximum plus one at most **
	  if( eisc->u.int_val > eisc->max_limit + 1 )
	    eisc->u.int_val = eisc->max_limit + 1;
	  break;

	default:

	  HnAbort("KheDrsExprIntSeqCombDoDay internal error");
	  break;
      }
    }
    else if( eisc->u.int_val > 0 )
    {
      ** a non-empty sequence not at the end of the cycle has just ended **
      dev = KheDrsSeqDev(eisc->min_limit, eisc->max_limit, 0, eisc->u.int_val);
      if( dev > 0 )
      {
	cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
	if( cost > 0 )
	{
	  if( drs->solve_rerun )
	    fprintf(stderr, "  %s reporting cost %.5f\n",
	      KheMonitorId(eisc->monitor), KheCostShow(cost));
	  next_soln->cost += cost;
	}
      }
      eisc->u.int_val = 0;
    }
  }

  if( day_index < eisc->day_range.last )
  {
    ** this not eisc's last day; store value in next_soln **
    si = HaArray(eisc->sig_indexes, day_index - eisc->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == si,
      "KheDrsExprIntSeqInternalDoDay internal error 3");
    HaArrayAddLast(next_soln->signature, eisc->u.int_val);
  }
  else if( eisc->u.int_val > 0 )
  {
    ** this is the last day and we have a sequence to report **
    dev = KheDrsSeqDev(eisc->min_limit, eisc->max_limit,
      eisc->history_after, eisc->u.int_val);
    if( dev > 0 )
    {
      cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
      if( cost > 0 )
      {
	if( drs->solve_rerun )
	  fprintf(stderr, "  %s reporting cost %.5f\n",
	    KheMonitorId(eisc->monitor), KheCostShow(cost));
	next_soln->cost += cost;
      }
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombDoDay(KHE_DRS_EXPR_INT_SEQ_COMB eisc,           */
/*    int day_index,                                                         */
/*    KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active expression eisc on day_index.                             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprIntSeqCombDoDay(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  int day_index,
  KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int si, i, dev, start_index, stop_index;
  KHE_DRS_EXPR child_e;  KHE_COST cost;

  ** initialize eisc->value depending on whether day_index is its first day **
  if( day_index == eisc->day_range.first )
    eisc->u.int_val = eisc->history_before;
  else
  {
    ** this is not eisc's first day, so we need to retrieve a previous value **
    si = HaArray(eisc->sig_indexes, day_index-1 - eisc->day_range.first);
    eisc->u.int_val = HaArray(prev_soln->signature, prev_so + si);
  }

  ** accumulate the values of the children of eisc that finalized today **
  start_index = HaArray(eisc->open_child_indexes,
    day_index - eisc->day_range.first);
  stop_index =  HaArray(eisc->open_child_indexes,
    day_index+1-eisc->day_range.first);
  for( i = start_index;  i < stop_index;  i++ )
  {
    child_e = HaArray(eisc->children, i);
    if( child_e->u.int_val == 1 )
    {
      ** a sequence is beginning or continuing **
      eisc->u.int_val++;

      ** adjust in various special cases **
      switch( eisc->comb_type )
      {
	case KHE_DRS_COMB_ORDINARY:

	  ** no adjustment here **
	  break;

	case KHE_DRS_COMB_NO_MAX:

	  ** no maximum limit; store minimum limit at most **
	  if( eisc->u.int_val > eisc->min_limit )
	    eisc->u.int_val = eisc->min_limit;
	  break;

	case KHE_DRS_COMB_LINEAR:

	  ** linear cost; if over maximum, report excess immediately **
	  if( eisc->u.int_val > eisc->max_limit )
	  {
	    dev = KheDrsSeqDev(eisc->min_limit, eisc->max_limit, 0,
	      eisc->u.int_val);
	    cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
	    if( cost > 0 )
	    {
	      if( drs->solve_rerun )
		fprintf(stderr, "  %s reporting linear cost %.5f\n",
		  KheMonitorId(eisc->monitor), KheCostShow(cost));
	      next_soln->cost += cost;
	    }
	    eisc->u.int_val = eisc->max_limit;
	  }
	  break;

	case KHE_DRS_COMB_STEP:

	  ** step cost; store maximum plus one at most **
	  if( eisc->u.int_val > eisc->max_limit + 1 )
	    eisc->u.int_val = eisc->max_limit + 1;
	  break;

	default:

	  HnAbort("KheDrsExprIntSeqCombDoDay internal error");
	  break;
      }
    }
    else if( eisc->u.int_val > 0 )
    {
      ** a non-empty sequence not at the end of the cycle has just ended **
      dev = KheDrsSeqDev(eisc->min_limit, eisc->max_limit, 0, eisc->u.int_val);
      if( dev > 0 )
      {
	cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
	if( cost > 0 )
	{
	  if( drs->solve_rerun )
	    fprintf(stderr, "  %s reporting cost %.5f\n",
	      KheMonitorId(eisc->monitor), KheCostShow(cost));
	  next_soln->cost += cost;
	}
      }
      eisc->u.int_val = 0;
    }
  }

  if( day_index < eisc->day_range.last )
  {
    ** this not eisc's last day; store value in next_soln **
    si = HaArray(eisc->sig_indexes, day_index - eisc->day_range.first);
    HnAssert(HaArrayCount(next_soln->signature) == next_so + si,
      "KheDrsExprIntSeqDoDay internal error 3");
    HaArrayAddLast(next_soln->signature, eisc->u.int_val);
  }
  else if( eisc->u.int_val > 0 )
  {
    ** this is the last day and we have a sequence to report **
    dev = KheDrsSeqDev(eisc->min_limit, eisc->max_limit,
      eisc->history_after, eisc->u.int_val);
    if( dev > 0 )
    {
      cost = KheDrsCost(eisc->cost_fn, eisc->combined_weight, dev);
      if( cost > 0 )
      {
	if( drs->solve_rerun )
	  fprintf(stderr, "  %s reporting cost %.5f\n",
	    KheMonitorId(eisc->monitor), KheCostShow(cost));
	next_soln->cost += cost;
      }
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprIntSeqCombDebug(KHE_DRS_EXPR_INT_SEQ_COMB eisc,           */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of the header part of eisc in soln onto fp.                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprIntSeqCombDebug(KHE_DRS_EXPR_INT_SEQ_COMB eisc,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  char /* buff1[100], */ buff2[100];  int si;
  if( soln != NULL )
  {
    /* unindented version with a soln value */
    fprintf(fp, "INT_SEQ_COMB %s", KheMonitorId(eisc->monitor));
    if( KheDrsExprInSignature((KHE_DRS_EXPR) eisc, soln->day, &si) )
      fprintf(fp, " $%d:%d$", si, HaArray(soln->signature, si));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eisc, soln, drs, verbosity, -1, fp);
  }
  else if( indent >= 0 )
  {
    /* indented version without a soln value */
    fprintf(fp, "%*s[ INT_SEQ_COMB %s %s\n", indent, "",
      KheMonitorId(eisc->monitor),
      /* KheDrsExprDayRangeShow((KHE_DRS_EXPR) eisc, drs, buff1), */
      KheDrsDomTestShow(eisc->dom_test, buff2));
    KheDrsExprDebugChildren((KHE_DRS_EXPR) eisc, soln, drs,
      verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
  {
    /* unindented version without a soln value */
    fprintf(fp, "[ INT_SEQ_COMB %s %s (%d children) ]",
      KheMonitorId(eisc->monitor),
      /* KheDrsExprDayRangeShow((KHE_DRS_EXPR) eisc, drs, buff1), */
      KheDrsDomTestShow(eisc->dom_test, buff2), HaArrayCount(eisc->children));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAddChild(KHE_DRS_EXPR parent, KHE_DRS_EXPR child)         */
/*                                                                           */
/*  Make child a child of parent.                                            */
/*                                                                           */
/*  Implementation note.  During expression construction, all expressions    */
/*  are closed by the time they are made children of their parents (by       */
/*  calls on this function).  Hence KheDrsExprChildHasClosed below.          */
/*                                                                           */
/*****************************************************************************/
static void KheDrsExprChildHasClosed(KHE_DRS_EXPR e,
  KHE_DRS_EXPR child_e, int child_index);

static void KheDrsExprAddChild(KHE_DRS_EXPR parent, KHE_DRS_EXPR child)
{
  KHE_DRS_PARENT prnt;
  HnAssert(HaArrayCount(parent->parents) == 0,
    "KheDrsExprAddChild internal error:  too late to add child to parent");
  prnt.expr = parent;
  prnt.index = HaArrayCount(parent->children);
  HaArrayAddLast(parent->children, child);
  HaArrayAddLast(child->parents, prnt);
  KheDrsExprChildHasClosed(parent, child,
    HaArrayCount(parent->children) - 1);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprCmp(const void *t1, const void *t2)                        */
/*                                                                           */
/*  Comparison function for sorting an array of expressions by increasing    */
/*  postorder index.                                                         */
/*                                                                           */
/*****************************************************************************/

static int KheDrsExprCmp(const void *t1, const void *t2)
{
  KHE_DRS_EXPR e1 = * (KHE_DRS_EXPR *) t1;
  KHE_DRS_EXPR e2 = * (KHE_DRS_EXPR *) t2;
  return e1->postorder_index - e2->postorder_index;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprSetPostorderIndex(KHE_DRS_EXPR e, int *val)               */
/*                                                                           */
/*  Set the postorder index field in every descendant of e, including e      */
/*  itself, unless it has already been set.                                  */
/*                                                                           */
/*****************************************************************************/

/* *** done during creation now
static void KheDrsExprSetPostorderIndex(KHE_DRS_EXPR e, int *val)
{
  KHE_DRS_EXPR child_e;  int i;
  if( e->postorder_index == -1 )
  {
    HaArrayForEach(e->children, child_e, i)
      KheDrsExprSetPostorderIndex(child_e, val);
    e->postorder_index = (*val)++;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR - opening"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprGatherForOpening(KHE_DRS_EXPR e,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Gather e into drs->open_exprs, the list of expressions that are to be    */
/*  opened.  This does not actually open e; that comes later.                */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprGatherForOpening(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;  int i;
  if( !e->gathered )
  {
    e->gathered = true;
    HaArrayAddLast(drs->open_exprs, e);
    HaArrayForEach(e->parents, prnt, i)
      KheDrsExprGatherForOpening(prnt.expr, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprChildHasOpened(KHE_DRS_EXPR e, KHE_DRS_EXPR child_e,      */
/*    int child_index)                                                       */
/*                                                                           */
/*  As part of opening e for solving, inform e that its child_index'th       */
/*  child, child_e, has opened.                                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprChildHasOpened(KHE_DRS_EXPR e, KHE_DRS_EXPR child_e,
  int child_index)
{
  int i;  KHE_DRS_EXPR e2;

  /* add child_e's last solve day to open_day_range */
  KheDrsDayRangeAdd(&e->open_day_range, child_e->open_day_range.last);

  /* add child_e to e->open_children in last day order */
  HaArrayAddLast(e->open_children, NULL);
  for( i = HaArrayCount(e->open_children) - 2;  i >= 0;  i-- )
  {
    e2 = HaArray(e->open_children, i);
    if( e2->open_day_range.last <= child_e->open_day_range.last )
      break;
    HaArrayPut(e->open_children, i + 1, e2);
  }
  HaArrayPut(e->open_children, i + 1, child_e);

  /* delete child_e's value from the closed state of e */
  switch( e->tag )
  {
    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrChildHasOpened((KHE_DRS_EXPR_OR) e,
	child_e, child_index);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndChildHasOpened((KHE_DRS_EXPR_AND) e,
	child_e, child_index);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumChildHasOpened((KHE_DRS_EXPR_INT_SUM) e,
	child_e, child_index);
      break;

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumChildHasOpened((KHE_DRS_EXPR_FLOAT_SUM) e,
	child_e, child_index);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevChildHasOpened((KHE_DRS_EXPR_INT_DEV) e,
	child_e, child_index);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevChildHasOpened((KHE_DRS_EXPR_FLOAT_DEV) e,
	child_e, child_index);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostChildHasOpened((KHE_DRS_EXPR_COST) e,
	child_e, child_index);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombChildHasOpened((KHE_DRS_EXPR_INT_SUM_COMB) e,
	child_e, child_index);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombChildHasOpened((KHE_DRS_EXPR_INT_SEQ_COMB) e,
	child_e, child_index);
      break;

    default:

      HnAbort("KheDrsExprChildHasOpened internal error (tag %d)", e->tag);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprOpen(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs)     */
/*                                                                           */
/*  Open e for solving.                                                      */
/*                                                                           */
/*  By the time this is called, all of e's open children will have           */
/*  called KheDrsExprChildHasOpened.                                         */
/*                                                                           */
/*  The value of e->open_child_indexes[i - e->open_day_range.first] is the   */
/*  index in e->open_children of the first child whose last open day is i:   */
/*                                                                           */
/*    start = HaArray(e->open_child_indexes, i - e->open_day_range.first);   */
/*    stop =  HaArray(e->open_child_indexes, i+1 - e->open_day_range.first); */
/*    for( j = start;  j < stop;  j++ )                                      */
/*      child_e = HaArray(e->open_children, j);                              */
/*                                                                           */
/*  visits all children whose last open day is i.                            */
/*                                                                           */
/*  Implementation note - how e->open_day_range is set                       */
/*  --------------------------------------------------                       */
/*  In external expressions, e->open_day_range receives its final value      */
/*  when KheDrsExprGatherForOpening is called.  In internal expressions,     */
/*  its value is initially empty and is updated by each call to              */
/*  KheDrsExprChildHasOpened until it reaches its final value.               */
/*                                                                           */
/*****************************************************************************/
static int KheDrsDayAddDomTest(KHE_DRS_DAY day, KHE_DRS_DOM_TEST dom_test);
static void KheDrsDayAddOpenExpr(KHE_DRS_DAY day, KHE_DRS_EXPR e);

static void KheDrsExprOpen(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;  int i, curr_last_day_index;
  KHE_DRS_DAY day;  KHE_DRS_EXPR child_e;

  /* inform e's parents that e is open */
  e->gathered = false;
  HaArrayForEach(e->parents, prnt, i)
    KheDrsExprChildHasOpened(prnt.expr, e, prnt.index);

  /* nothing more to do if e is an external expression */
  if( e->tag <= KHE_DRS_EXPR_ASSIGNED_TASK_TAG )
    return;

  /* build open_child_indexes array into e's open children (already sorted) */
  HaArrayClear(e->open_child_indexes);
  curr_last_day_index = e->open_day_range.first - 1;
  HaArrayForEach(e->open_children, child_e, i)
  {
    while( child_e->open_day_range.last > curr_last_day_index )
    {
      HaArrayAddLast(e->open_child_indexes, i);
      curr_last_day_index++;
    }
  }
  HaArrayAddLast(e->open_child_indexes, i);

  /* inform e's days that e is open */
  HaArrayClear(e->sig_indexes);
  for( i = e->open_day_range.first;  i <= e->open_day_range.last;  i++ )
  {
    /* add e to day's list of open internal expressions */
    day = HaArray(drs->open_days, i);
    KheDrsDayAddOpenExpr(day, e);

    /* signatures and dom tests (needed on each day except the last) */
    if( i < e->open_day_range.last )
      HaArrayAddLast(e->sig_indexes, KheDrsDayAddDomTest(day, e->dom_test));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprExternalOpen(KHE_DRS_EXPR e, int day_index,               */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Open external expression e on day_index.                                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprExternalOpen(KHE_DRS_EXPR e, int day_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;  int i;
  HaArrayForEach(e->parents, prnt, i)
    KheDrsInternalExprChildHasOpened(prnt.expr, e, prnt.index, day_index,
      e->u, drs);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsInternalExprChildHasOpened(KHE_DRS_EXPR e,                    */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v, */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Inform internal expression e that its child_index'th child, child_e,     */
/*  has opened on day_index, and that its previous closed value was v.       */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsInternalExprChildHasOpened(KHE_DRS_EXPR e,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrChildHasOpened((KHE_DRS_EXPR_OR) e,
	child_e, child_index, day_index, v, drs);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndChildHasOpened((KHE_DRS_EXPR_AND) e,
	child_e, child_index, day_index, v, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumChildHasOpened((KHE_DRS_EXPR_INT_SUM) e,
	child_e, child_index, day_index, v, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumChildHasOpened((KHE_DRS_EXPR_FLOAT_SUM) e,
	child_e, child_index, day_index, v, drs);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevChildHasOpened((KHE_DRS_EXPR_INT_DEV) e,
	child_e, child_index, day_index, v, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevChildHasOpened((KHE_DRS_EXPR_FLOAT_DEV) e,
	child_e, child_index, day_index, v, drs);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostChildHasOpened((KHE_DRS_EXPR_COST) e,
	child_e, child_index, day_index, v, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombChildHasOpened((KHE_DRS_EXPR_INT_SUM_COMB) e,
	child_e, child_index, day_index, v, drs);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombChildHasOpened((KHE_DRS_EXPR_INT_SEQ_COMB) e,
	child_e, child_index, day_index, v, drs);
      break;

    default:

      HnAbort("KheDrsInternalExprChildHasOpened internal error (%d)", e->tag);
      break;

  }
  ** *** put this into each branch separately
  KHE_DRS_PARENT prnt;  int i;
  HaArrayForEach(e->parents, prnt, i)
    KheDrsInternalExprChildHasOpened(prnt.expr, e, prnt.index, day_index);
  *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR - closing"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprChildHasClosed(KHE_DRS_EXPR e,                            */
/*    KHE_DRS_EXPR child_e, int child_index)                                 */
/*                                                                           */
/*  Inform internal expression e that its child_index's child, child_e,      */
/*  has closed.  The child will have its correct closed_state and value      */
/*  fields when this is called.                                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprChildHasClosed(KHE_DRS_EXPR e,
  KHE_DRS_EXPR child_e, int child_index)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrChildHasClosed((KHE_DRS_EXPR_OR) e,
	child_e, child_index);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndChildHasClosed((KHE_DRS_EXPR_AND) e,
	child_e, child_index);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumChildHasClosed((KHE_DRS_EXPR_INT_SUM) e,
	child_e, child_index);
      break;

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumChildHasClosed((KHE_DRS_EXPR_FLOAT_SUM) e,
	child_e, child_index);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevChildHasClosed((KHE_DRS_EXPR_INT_DEV) e,
	child_e, child_index);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevChildHasClosed((KHE_DRS_EXPR_FLOAT_DEV) e,
	child_e, child_index);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostChildHasClosed((KHE_DRS_EXPR_COST) e,
	child_e, child_index);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombChildHasClosed((KHE_DRS_EXPR_INT_SUM_COMB) e,
	child_e, child_index);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombChildHasClosed((KHE_DRS_EXPR_INT_SEQ_COMB) e,
	child_e, child_index);
      break;

    default:

      HnAbort("KheDrsExprChildHasClosed internal error (tag %d)", e->tag);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsInternalExprSetValueFromClosedState(KHE_DRS_EXPR e,           */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Here e is closing.  Set its value, based on its closed state.            */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsInternalExprSetValueFromClosedState(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrSetValueFromClosedState((KHE_DRS_EXPR_OR) e, drs);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndSetValueFromClosedState((KHE_DRS_EXPR_AND) e, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumSetValueFromClosedState((KHE_DRS_EXPR_INT_SUM) e, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumSetValueFromClosedState((KHE_DRS_EXPR_FLOAT_SUM) e,drs);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevSetValueFromClosedState((KHE_DRS_EXPR_INT_DEV) e, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevSetValueFromClosedState((KHE_DRS_EXPR_FLOAT_DEV) e,drs);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostSetValueFromClosedState((KHE_DRS_EXPR_COST) e, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombSetValueFromClosedState((KHE_DRS_EXPR_INT_SUM_COMB) e,
	drs);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombSetValueFromClosedState((KHE_DRS_EXPR_INT_SEQ_COMB) e,
	drs);
      break;

    default:

      HnAbort("KheDrsExprSetValueFromClosedState internal error (tag %d)",
	e->tag);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprSetClosedValue(KHE_DRS_EXPR e,                            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  As part of closing e, set its value suitably for its closed state.       */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprSetClosedValue(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      KheDrsExprBusyTimeSetClosedValue((KHE_DRS_EXPR_BUSY_TIME) e, drs);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      KheDrsExprFreeTimeSetClosedValue((KHE_DRS_EXPR_FREE_TIME) e, drs);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      KheDrsExprWorkTimeSetClosedValue((KHE_DRS_EXPR_WORK_TIME) e, drs);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      KheDrsExprBusyDaySetClosedValue((KHE_DRS_EXPR_BUSY_DAY) e, drs);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      KheDrsExprFreeDaySetClosedValue((KHE_DRS_EXPR_FREE_DAY) e, drs);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      KheDrsExprWorkDaySetClosedValue((KHE_DRS_EXPR_WORK_DAY) e, drs);
      break;

    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      KheDrsExprAssignedTaskSetClosedValue((KHE_DRS_EXPR_ASSIGNED_TASK) e, drs);
      break;

    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrSetClosedValue((KHE_DRS_EXPR_OR) e, drs);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndSetClosedValue((KHE_DRS_EXPR_AND) e, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumSetClosedValue((KHE_DRS_EXPR_INT_SUM) e, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumSetClosedValue((KHE_DRS_EXPR_FLOAT_SUM) e,drs);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevSetClosedValue((KHE_DRS_EXPR_INT_DEV) e, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevSetClosedValue((KHE_DRS_EXPR_FLOAT_DEV) e,drs);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostSetClosedValue((KHE_DRS_EXPR_COST) e, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombSetClosedValue((KHE_DRS_EXPR_INT_SUM_COMB) e, drs);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombSetClosedValue((KHE_DRS_EXPR_INT_SEQ_COMB) e, drs);
      break;

    default:

      HnAbort("KheDrsExprSetClosedValue internal error (tag %d)", e->tag);
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprClose(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Close e, which may be internal or external.  Whatever it needs to work   */
/*  out its closed value (an assignment, closed children, or closed state)   */
/*  is up to date.  At the end, inform e's parents that e has closed.        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprClose(KHE_DRS_EXPR e, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;  int i;

  /* set e's closed value */
  KheDrsExprSetClosedValue(e, drs);

  /* clear fields that are meaningful only when e is open */
  e->open_day_range = KheDrsDayRangeMake(1, 0);
  HaArrayClear(e->open_children);
  HaArrayClear(e->open_child_indexes);
  HaArrayClear(e->sig_indexes);

  /* inform e's parents that e has closed */
  HaArrayForEach(e->parents, prnt, i)
    KheDrsExprChildHasClosed(prnt.expr, e, prnt.index);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsInternalExprClose(KHE_DRS_EXPR e,                             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Close e.  Its children have all been closed, its closed_state is up      */
/*  to date with that, and so what we mainly need to do is to set its        */
/*  value depending on its closed state, then inform its parents.            */
/*                                                                           */
/*****************************************************************************/

/* *** replaced by KheDrsExprClose
static void KheDrsInternalExprClose(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_PARENT prnt;  int i;

  ** set e's value depending on its closed_state **
  KheDrsInternalExprSetValueFromClosedState(e, drs);

  ** clear fields that are meaningful only when e is open **
  e->open_day_range = KheDrsDayRangeMake(1, 0);
  HaArrayClear(e->open_children);
  HaArrayClear(e->open_child_indexes);
  HaArrayClear(e->sig_indexes);

  ** inform e's parents that e has closed **
  HaArrayForEach(e->parents, prnt, i)
    KheDrsExprChildHasClosed(prnt.expr, e, prnt.index);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprExternalClose(KHE_DRS_EXPR e, int day_index,              */
/*    KHE_TASK task, KHE_TIME t, KHE_RESOURCE r)                             */
/*                                                                           */
/*  Close external expression e on day_index.                                */
/*                                                                           */
/*  When e is resource-oriented, r is non-NULL and task and t are non-NULL   */
/*  when r is busy on e's day.  When e is task-oriented, task and t are      */
/*  non-NULL and task is assigned r (possibly NULL) on e's day.              */
/*                                                                           */
/*****************************************************************************/

/* *** replaced by KheDrsExprClose
static void KheDrsExprExternalClose(KHE_DRS_EXPR e, KHE_TASK task, KHE_TIME t,
  KHE_RESOURCE r)
{
  KHE_DRS_PARENT prnt;  int i;

  ** set closed value based on task, t, and r **
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      KheDrsExprBusyTimeSetValue((KHE_DRS_EXPR_BUSY_TIME) e, task, t, r);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      KheDrsExprFreeTimeSetValue((KHE_DRS_EXPR_FREE_TIME) e, task, t, r);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      KheDrsExprWorkTimeSetValue((KHE_DRS_EXPR_WORK_TIME) e, task, t, r);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      KheDrsExprBusyDaySetValue((KHE_DRS_EXPR_BUSY_DAY) e, task, t, r);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      KheDrsExprFreeDaySetValue((KHE_DRS_EXPR_FREE_DAY) e, task, t, r);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      KheDrsExprWorkDaySetValue((KHE_DRS_EXPR_WORK_DAY) e, task, t, r);
      break;

    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      KheDrsExprAssignedTaskSetValue((KHE_DRS_EXPR_ASSIGNED_TASK) e, task, t,r);
      break;

    default:

      HnAbort("KheDrsExprExternalClose internal error (%d)", e->tag);
      break;
  }

  ** inform parents that e has closed **
  HaArrayForEach(e->parents, prnt, i)
    KheDrsExprChildHasClosed(prnt.expr, e, prnt.index);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprChildHasClosed(KHE_DRS_EXPR e,                            */
/*    KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v) */
/*                                                                           */
/*  Inform internal expression e that its child_index'th child, child_e,     */
/*  has closed on day_index, and that child_e's closed value is v.           */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprChildHasClosed(KHE_DRS_EXPR e,
  KHE_DRS_EXPR child_e, int child_index, int day_index, KHE_DRS_VALUE v)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrChildHasClosed((KHE_DRS_EXPR_OR) e,
	child_e, child_index, day_index, v);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndChildHasClosed((KHE_DRS_EXPR_AND) e,
	child_e, child_index, day_index, v);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumChildHasClosed((KHE_DRS_EXPR_INT_SUM) e,
	child_e, child_index, day_index, v);
      break;

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumChildHasClosed((KHE_DRS_EXPR_FLOAT_SUM) e,
	child_e, child_index, day_index, v);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevChildHasClosed((KHE_DRS_EXPR_INT_DEV) e,
	child_e, child_index, day_index, v);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevChildHasClosed((KHE_DRS_EXPR_FLOAT_DEV) e,
	child_e, child_index, day_index, v);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostChildHasClosed((KHE_DRS_EXPR_COST) e,
	child_e, child_index, day_index, v);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombChildHasClosed((KHE_DRS_EXPR_INT_SUM_COMB) e,
	child_e, child_index, day_index, v);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombChildHasClosed((KHE_DRS_EXPR_INT_SEQ_COMB) e,
	child_e, child_index, day_index, v);
      break;

    default:

      HnAbort("KheDrsExprChildHasClosed internal error (%d)", e->tag);
      break;

  }
  ** *** put this into each branch separately
  KHE_DRS_PARENT prnt;  int i;
  HaArrayForEach(e->parents, prnt, i)
    KheDrsExprChildHasClosed(prnt.expr, e, prnt.index, day_index);
  *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprEnsureDayRangeCoversDayIndex(KHE_DRS_EXPR e,              */
/*    int last_day_index)                                                    */
/*                                                                           */
/*  Ensure that e's day range covers last_day_index.                         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprEnsureDayRangeCoversDayIndex(KHE_DRS_EXPR e,
  int day_index)
{
  if( day_index < e->day_range.first )
    e->day_range.first = day_index;
  if( day_index > e->day_range.last )
    e->day_range.last = day_index;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprIncreasingLastDayIndexCmp(const void *t1, const void *t2)  */
/*                                                                           */
/*  Comparison function for sorting an array of expressions by increasing    */
/*  last day index.                                                          */
/*                                                                           */
/*****************************************************************************/

/* ***
static int KheDrsExprIncreasingLastDayIndexCmp(const void *t1, const void *t2)
{
  KHE_DRS_EXPR e1 = * (KHE_DRS_EXPR *) t1;
  KHE_DRS_EXPR e2 = * (KHE_DRS_EXPR *) t2;
  return e1->day_range.last - e2->day_range.last;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprSetChildIndexes(KHE_DRS_EXPR e)                           */
/*                                                                           */
/*  Assuming that e's children's day indexes are set, set e's child indexes. */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprSetChildIndexes(KHE_DRS_EXPR e)
{
  int i, curr_last_day_index;  KHE_DRS_EXPR child_e;
  HaArrayClear(e->child_indexes);
  curr_last_day_index = e->day_range.first - 1;
  HaArrayForEach(e->children, child_e, i)
  {
    while( child_e->day_range.first > curr_last_day_index )
    {
      HaArrayAddLast(e->child_indexes, i);
      curr_last_day_index++;
    }
  }
  HaArrayAddLast(e->child_indexes, i);  ** sentinel **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprSetDayRange(KHE_DRS_EXPR e,                               */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Set the day range (first_day_index and last_day_index) of e.             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprSetDayRange(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, index;  KHE_DRS_EXPR child_e, prev_child_e;
  KHE_DRS_EXPR_BUSY_TIME eb;  KHE_DRS_EXPR_FREE_TIME ef;
  KHE_DRS_EXPR_WORK_TIME ew;  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_DRS_EXPR_FREE_DAY efd;  KHE_DRS_EXPR_WORK_DAY ewd;
  KHE_DRS_EXPR_INT_SEQ_COMB eisc;

  ** initialize range to empty **
  e->day_range = KheDrsDayRangeMake(INT_MAX, 0);
  ** ***
  e->first_day_index = INT_MAX;
  e->last_day_index = 0;
  *** **

  ** increase range depending on the expression type **
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      eb = (KHE_DRS_EXPR_BUSY_TIME) e;
      index = KheFrameTimeIndex(drs->days_frame, eb->time);
      KheDrsExprEnsureDayRangeCoversDayIndex(e, index);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      ef = (KHE_DRS_EXPR_FREE_TIME) e;
      index = KheFrameTimeIndex(drs->days_frame, ef->time);
      KheDrsExprEnsureDayRangeCoversDayIndex(e, index);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      ew = (KHE_DRS_EXPR_WORK_TIME) e;
      index = KheFrameTimeIndex(drs->days_frame, ew->time);
      KheDrsExprEnsureDayRangeCoversDayIndex(e, index);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      ebd = (KHE_DRS_EXPR_BUSY_DAY) e;
      index = KheFrameTimeIndex(drs->days_frame,
	KheTimeGroupTime(ebd->time_group, 0));
      KheDrsExprEnsureDayRangeCoversDayIndex(e, index);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      efd = (KHE_DRS_EXPR_FREE_DAY) e;
      index = KheFrameTimeIndex(drs->days_frame,
	KheTimeGroupTime(efd->time_group, 0));
      KheDrsExprEnsureDayRangeCoversDayIndex(e, index);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      ewd = (KHE_DRS_EXPR_WORK_DAY) e;
      index = KheFrameTimeIndex(drs->days_frame,
	KheTimeGroupTime(ewd->time_group, 0));
      KheDrsExprEnsureDayRangeCoversDayIndex(e, index);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      eisc = (KHE_DRS_EXPR_INT_SEQ_COMB) e;
      prev_child_e = NULL;
      HaArrayForEach(eisc->children, child_e, i)
      {
        KheDrsExprSetDayRange(child_e, drs);
	if( prev_child_e != NULL &&
	    prev_child_e->day_range.last > child_e->day_range.last )
	{
	  child_e = KheDrsExprOptionalCopy(child_e, drs);
	  HaArrayPut(eisc->children, i, child_e);
          child_e->day_range.last = prev_child_e->day_range.last;
	}
        KheDrsExprEnsureDayRangeCoversDayIndex(e, child_e->day_range.last);
	prev_child_e = child_e;
      }
      ** children already have increasing last day index, must not sort them **
      break;

    default:

      HaArrayForEach(e->children, child_e, i)
      {
        KheDrsExprSetDayRange(child_e, drs);
        KheDrsExprEnsureDayRangeCoversDayIndex(e, child_e->day_range.last);
      }
      HaArraySort(e->children, &KheDrsExprIncreasingLastDayIndexCmp);
      break;
  }

  ** set the child indexes **
  KheDrsExprSetChildIndexes(e);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprSetSigIndexes(KHE_DRS_EXPR e,                             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Set the signature indexes in e and its descendants.                      */
/*                                                                           */
/*  These are assigned in postorder, because that is the order that          */
/*  KheDrsExprEndDay visits the expressions.                                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsResourceOnDayAddDomTest(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_DOM_TEST dom_test);
*** */

/* *** moved to KheDrsExprOpen
static void KheDrsExprSetSigIndexes(KHE_DRS_EXPR e, ** KHE_DRS_RESOURCE dr, **
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR child_e;  int i ** , len **;  KHE_DRS_DAY day;
  ** KHE_DRS_RESOURCE_ON_DAY drd; **

  ** set the signature indexes in e's descendants **
  HaArrayForEach(e->children, child_e, i)
    KheDrsExprSetSigIndexes(child_e, ** dr, ** drs);

  ** set the signature indexes in e (NB < is correct, not <= here) **
  for( i = e->day_range.first;  i < e->day_range.last;  i++ )
  {
    ** may be wrong here; stil l to do **
    day = HaArray(drs->open_days, i);
    HaArrayAddLast(e->sig_indexes, day->sig_len++);
    ** ***
    day = HaArray(drs->days, i);
    HaArrayAddLast(e->sig_indexes, day->sig_len++);
    drd = HaArray(dr->days, i);
    len = HaArrayCount(drd->internal_today);
    HaArrayAddLast(e->sig_indexes, len);  ** may be wrong; sti ll to do **
    *** **
    ** HaArrayAddLast(e->sig_indexes, drd->sig_len++); **
    HnAssert(e->dom_test.type != KHE_DRS_DOM_UNUSED,
      "KheDrsExprSetSigIndexes internal error");
    KheDrsDayAddDomTest(day, e->dom_test);
    ** KheDrsResourceOnDayAddDomTest(drd, e->dom_test); **
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsExprIsExternal(KHE_DRS_EXPR e)                                */
/*                                                                           */
/*  Return true if e is external.                                            */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused
static bool KheDrsExprIsExternal(KHE_DRS_EXPR e)
{
  return e->tag <= KHE_DRS_EXPR_ASSIGNED_TASK_TAG;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsExprIsInternal(KHE_DRS_EXPR e)                                */
/*                                                                           */
/*  Return true if e is internal.                                            */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused
static bool KheDrsExprIsInternal(KHE_DRS_EXPR e)
{
  return !KheDrsExprIsExternal(e);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAddExpressionsToDays(KHE_DRS_EXPR e, KHE_DRS_RESOURCE dr) */
/*                                                                           */
/*  Add e and its descendants to the days of dr they are active on.  Most    */
/*  importantly, do it in postorder.                                         */
/*                                                                           */
/*****************************************************************************/

/* *** moved to KheDrsExprOpen
static void KheDrsExprAddExpressionsToDays(KHE_DRS_EXPR e, KHE_DRS_RESOURCE dr)
{
  KHE_DRS_EXPR child_e;  int i;  KHE_DRS_RESOURCE_ON_DAY drd;

  ** do it for e's proper descendants (first!) **
  HaArrayForEach(e->children, child_e, i)
    KheDrsExprAddExpressionsToDays(child_e, dr);

  ** do it for e (NB <= is correct, not < here) **
  for( i = e->first_day_index;  i <= e->last_day_index;  i++ )
  {
    drd = HaArray(dr->days, i);
    HaArrayAddLast(drd->internal_today, e);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprAddExpressionsToDays(KHE_DRS_EXPR e,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add e and its descendants to the days they are active on.  Most          */
/*  importantly, do it in postorder.                                         */
/*                                                                           */
/*****************************************************************************/

/* *** moved to KheDrsExprOpen
static void KheDrsExprAddExpressionsToDays(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR child_e;  int i;  KHE_DRS_DAY day;

  if( KheDrsExprIsInternal(e) )
  {
    ** do it for e's proper descendants (first!) **
    HaArrayForEach(e->children, child_e, i)
      KheDrsExprAddExpressionsToDays(child_e, drs);

    ** do it for e (NB <= is correct, not < here) **
    for( i = e->day_range.first;  i <= e->day_range.last;  i++ )
    {
      day = HaArray(drs->open_days, i);
      HaArrayAddLast(day->internal_today, e);
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeInternal(KHE_DRS_EXPR e,                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free e and its descendants, but only internal expressions.               */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFreeInternal(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      ** KheDrsExprBusyTimeFree((KHE_DRS_EXPR_BUSY_TIME) e, drs); **
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      ** KheDrsExprFreeTimeFree((KHE_DRS_EXPR_FREE_TIME) e, drs); **
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      ** KheDrsExprWorkTimeFree((KHE_DRS_EXPR_WORK_TIME) e, drs); **
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      ** KheDrsExprBusyDayFree((KHE_DRS_EXPR_BUSY_DAY) e, drs); **
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      ** KheDrsExprFreeDayFree((KHE_DRS_EXPR_FREE_DAY) e, drs); **
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      ** KheDrsExprWorkDayFree((KHE_DRS_EXPR_WORK_DAY) e, drs); **
      break;

    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrFreeInternal((KHE_DRS_EXPR_OR) e, drs);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndFreeInternal((KHE_DRS_EXPR_AND) e, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumFreeInternal((KHE_DRS_EXPR_INT_SUM) e, drs);
      break;

    ** *** currently unused
    case KHE_DRS_EXPR_INT_SEQ_TAG:

      KheDrsExprIntSeqFreeInternal((KHE_DRS_EXPR_INT_SEQ) e, drs);
      break;
    *** **

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumFreeInternal((KHE_DRS_EXPR_FLOAT_SUM) e, drs);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevFreeInternal((KHE_DRS_EXPR_INT_DEV) e, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevFreeInternal((KHE_DRS_EXPR_FLOAT_DEV) e, drs);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostFreeInternal((KHE_DRS_EXPR_COST) e, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombFreeInternal((KHE_DRS_EXPR_INT_SUM_COMB) e, drs);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombFreeInternal((KHE_DRS_EXPR_INT_SEQ_COMB) e, drs);
      break;

    default:

      HnAbort("KheDrsExprFree internal error: unknown tag %d", e->tag);
      break;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprFreeExternal(KHE_DRS_EXPR e,                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free external expression e.                                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprFreeExternal(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      KheDrsExprBusyTimeFreeExternal((KHE_DRS_EXPR_BUSY_TIME) e, drs);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      KheDrsExprFreeTimeFreeExternal((KHE_DRS_EXPR_FREE_TIME) e, drs);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      KheDrsExprWorkTimeFreeExternal((KHE_DRS_EXPR_WORK_TIME) e, drs);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      KheDrsExprBusyDayFreeExternal((KHE_DRS_EXPR_BUSY_DAY) e, drs);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      KheDrsExprFreeDayFreeExternal((KHE_DRS_EXPR_FREE_DAY) e, drs);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      KheDrsExprWorkDayFreeExternal((KHE_DRS_EXPR_WORK_DAY) e, drs);
      break;

    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      KheDrsExprAssignedTaskFreeExternal((KHE_DRS_EXPR_ASSIGNED_TASK) e, drs);
      break;

    default:

      HnAbort("KheDrsExprFreeExternal internal error: tag %d", e->tag);
      break;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EXPR - solving"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprLeafSet(KHE_DRS_EXPR e, KHE_DRS_TASK_ON_DAY dtd,          */
/*    KHE_DRS_RESOURCE_ON_DAY drd)                                           */
/*                                                                           */
/*  Inform leaf expression e that dtd has been assigned drd.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprLeafSet(KHE_DRS_EXPR e, KHE_DRS_TASK_ON_DAY dtd,
  KHE_DRS_RESOURCE_ON_DAY drd)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      KheDrsExprBusyTimeLeafSet((KHE_DRS_EXPR_BUSY_TIME) e, dtd, drd);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      KheDrsExprFreeTimeLeafSet((KHE_DRS_EXPR_FREE_TIME) e, dtd, drd);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      KheDrsExprWorkTimeLeafSet((KHE_DRS_EXPR_WORK_TIME) e, dtd, drd);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      KheDrsExprBusyDayLeafSet((KHE_DRS_EXPR_BUSY_DAY) e, dtd, drd);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      KheDrsExprFreeDayLeafSet((KHE_DRS_EXPR_FREE_DAY) e, dtd, drd);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      KheDrsExprWorkDayLeafSet((KHE_DRS_EXPR_WORK_DAY) e, dtd, drd);
      break;

    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      KheDrsExprAssignedTaskLeafSet((KHE_DRS_EXPR_ASSIGNED_TASK) e, dtd, drd);
      break;

    default:

      HnAbort("KheDrsExprLeafSet internal error: tag %d", e->tag);
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprLeafClear(KHE_DRS_EXPR e)                                 */
/*                                                                           */
/*  Inform leaf expression e that the assignment of drd to dtd carried out   */
/*  by KheDrsExprLeafSet above has been removed.                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprLeafClear(KHE_DRS_EXPR e)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      KheDrsExprBusyTimeLeafClear((KHE_DRS_EXPR_BUSY_TIME) e);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      KheDrsExprFreeTimeLeafClear((KHE_DRS_EXPR_FREE_TIME) e);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      KheDrsExprWorkTimeLeafClear((KHE_DRS_EXPR_WORK_TIME) e);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      KheDrsExprBusyDayLeafClear((KHE_DRS_EXPR_BUSY_DAY) e);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      KheDrsExprFreeDayLeafClear((KHE_DRS_EXPR_FREE_DAY) e);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      KheDrsExprWorkDayLeafClear((KHE_DRS_EXPR_WORK_DAY) e);
      break;

    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      KheDrsExprAssignedTaskLeafClear((KHE_DRS_EXPR_ASSIGNED_TASK) e);
      break;

    default:

      HnAbort("KheDrsExprLeafClear internal error: tag %d");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprEvalDay(KHE_DRS_EXPR e,                                   */
/*    int open_day_index, KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Evaluate expression e on day open_day_index.                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprEvalDay(KHE_DRS_EXPR e, int open_day_index,
  KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:
    case KHE_DRS_EXPR_FREE_TIME_TAG:
    case KHE_DRS_EXPR_WORK_TIME_TAG:
    case KHE_DRS_EXPR_BUSY_DAY_TAG:
    case KHE_DRS_EXPR_FREE_DAY_TAG:
    case KHE_DRS_EXPR_WORK_DAY_TAG:
    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      HnAbort("KheDrsExprEvalDay: external tag (%d)", e->tag);
      break;

    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrEvalDay((KHE_DRS_EXPR_OR) e,
	open_day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndEvalDay((KHE_DRS_EXPR_AND) e,
	open_day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumEvalDay((KHE_DRS_EXPR_INT_SUM) e,
	open_day_index, prev_soln, next_soln, drs);
      break;

    /* *** currently unused
    case KHE_DRS_EXPR_INT_SEQ_TAG:

      KheDrsExprIntSeqEvalDay((KHE_DRS_EXPR_INT_SEQ) e, open_day_index,
	prev_soln, next_soln, drs);
      break;
    *** */

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumEvalDay((KHE_DRS_EXPR_FLOAT_SUM) e,
	open_day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevEvalDay((KHE_DRS_EXPR_INT_DEV) e,
	open_day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevEvalDay((KHE_DRS_EXPR_FLOAT_DEV) e,
	open_day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostEvalDay((KHE_DRS_EXPR_COST) e,
	open_day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombEvalDay((KHE_DRS_EXPR_INT_SUM_COMB) e,
	open_day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombEvalDay((KHE_DRS_EXPR_INT_SEQ_COMB) e,
	open_day_index, prev_soln, next_soln, drs);
      break;

    default:

      HnAbort("KheDrsExprEvalDay internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsInternalExprDoDay(KHE_DRS_EXPR e, int day_index,              */
/*    KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,                        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Process internal expression e on day day_index.                          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsInternalExprDoDay(KHE_DRS_EXPR e, int day_index,
  KHE_DRS_SOLN prev_soln, KHE_DRS_SOLN next_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:
    case KHE_DRS_EXPR_FREE_TIME_TAG:
    case KHE_DRS_EXPR_WORK_TIME_TAG:
    case KHE_DRS_EXPR_BUSY_DAY_TAG:
    case KHE_DRS_EXPR_FREE_DAY_TAG:
    case KHE_DRS_EXPR_WORK_DAY_TAG:
    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      HnAbort("KheDrsInternalExprDoDay: external tag (%d)", e->tag);
      break;

    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrInternalDoDay((KHE_DRS_EXPR_OR) e,
	day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndInternalDoDay((KHE_DRS_EXPR_AND) e,
	day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumInternalDoDay((KHE_DRS_EXPR_INT_SUM) e,
	day_index, prev_soln, next_soln, drs);
      break;

    ** *** currently unused
    case KHE_DRS_EXPR_INT_SEQ_TAG:

      KheDrsExprIntSeqInternalDoDay((KHE_DRS_EXPR_INT_SEQ) e, day_index,
	prev_soln, next_soln, drs);
      break;
    *** **

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumInternalDoDay((KHE_DRS_EXPR_FLOAT_SUM) e,
	day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevInternalDoDay((KHE_DRS_EXPR_INT_DEV) e,
	day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevInternalDoDay((KHE_DRS_EXPR_FLOAT_DEV) e,
	day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostInternalDoDay((KHE_DRS_EXPR_COST) e,
	day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombInternalDoDay((KHE_DRS_EXPR_INT_SUM_COMB) e,
	day_index, prev_soln, next_soln, drs);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombInternalDoDay((KHE_DRS_EXPR_INT_SEQ_COMB) e,
	day_index, prev_soln, next_soln, drs);
      break;

    default:

      HnAbort("KheDrsInternalExprDoDay internal error");
      break;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprDoDay(KHE_DRS_EXPR e, int day_index,                      */
/*    KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,     */
/*    KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Process active expression e on day day_index.                            */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsExprDoDay(KHE_DRS_EXPR e, int day_index,
  KHE_DRS_SOLN prev_soln, int prev_so, KHE_TASK task, KHE_TIME time,
  KHE_DRS_SOLN next_soln, int next_so, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      KheDrsExprBusyTimeDoDay((KHE_DRS_EXPR_BUSY_TIME) e, day_index,
	prev_soln, prev_so, task, time, next_soln, next_so, drs);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      KheDrsExprFreeTimeDoDay((KHE_DRS_EXPR_FREE_TIME) e, day_index,
	prev_soln, prev_so, task, time, next_soln, next_so, drs);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      KheDrsExprWorkTimeDoDay((KHE_DRS_EXPR_WORK_TIME) e, day_index,
	prev_soln, prev_so, task, time, next_soln, next_so, drs);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      KheDrsExprBusyDayDoDay((KHE_DRS_EXPR_BUSY_DAY) e, day_index,
	prev_soln, prev_so, task, time, next_soln, next_so, drs);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      KheDrsExprFreeDayDoDay((KHE_DRS_EXPR_FREE_DAY) e, day_index,
	prev_soln, prev_so, task, time, next_soln, next_so, drs);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      KheDrsExprWorkDayDoDay((KHE_DRS_EXPR_WORK_DAY) e, day_index,
	prev_soln, prev_so, task, time, next_soln, next_so, drs);
      break;

    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      KheDrsExprAssignedTaskDoDay((KHE_DRS_EXPR_ASSIGNED_TASK) e, day_index,
        prev_soln, prev_so, KheTaskAsstResource(task), next_soln, next_so, drs);
      break;

    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrDoDay((KHE_DRS_EXPR_OR) e, day_index,
	prev_soln, prev_so, task, time, next_soln, next_so, drs);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndDoDay((KHE_DRS_EXPR_AND) e, day_index,
	prev_soln, prev_so, task, time, next_soln, next_so, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumDoDay((KHE_DRS_EXPR_INT_SUM) e, day_index,
	prev_soln, prev_so, task, time, next_soln, next_so, drs);
      break;

    ** *** currently unused
    case KHE_DRS_EXPR_INT_SEQ_TAG:

      KheDrsExprIntSeqDoDay((KHE_DRS_EXPR_INT_SEQ) e, day_index,
	prev_soln, prev_so, task, time, next_soln, next_so, drs);
      break;
    *** **

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumDoDay((KHE_DRS_EXPR_FLOAT_SUM) e, day_index,
	prev_soln, prev_so, task, time, next_soln, next_so, drs);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevDoDay((KHE_DRS_EXPR_INT_DEV) e, day_index,
	prev_soln, prev_so, task, time, next_soln, next_so, drs);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevDoDay((KHE_DRS_EXPR_FLOAT_DEV) e, day_index,
	prev_soln, prev_so, task, time, next_soln, next_so, drs);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostDoDay((KHE_DRS_EXPR_COST) e, day_index,
	prev_soln, prev_so, task, time, next_soln, next_so, drs);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombDoDay((KHE_DRS_EXPR_INT_SUM_COMB) e, day_index,
	prev_soln, prev_so, task, time, next_soln, next_so, drs);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombDoDay((KHE_DRS_EXPR_INT_SEQ_COMB) e, day_index,
	prev_soln, prev_so, task, time, next_soln, next_so, drs);
      break;

    default:

      HnAbort("KheDrsExprDoDay internal error");
      break;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsExprIsActive(KHE_DRS_EXPR e, int day_index)                   */
/*                                                                           */
/*  Return true if e is active on day_index.                                 */
/*                                                                           */
/*****************************************************************************/

/* *** currently unused; good idea though
static bool KheDrsExprIsActive(KHE_DRS_EXPR e, int day_index)
{
  return e->first_day_index <= day_index && day_index <= e->last_day_index;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsExprInSignature(KHE_DRS_EXPR e, KHE_DRS_DAY day, int *si)     */
/*                                                                           */
/*  If e is in the signature on day, return true and set *si to its          */
/*  position in the signature.  Otherwise return false.                      */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsExprInSignature(KHE_DRS_EXPR e, KHE_DRS_DAY day, int *si)
{
  int i = day->open_day_index;
  if( e->open_day_range.first <= i && i < e->open_day_range.last )
    return *si = HaArray(e->sig_indexes, i - e->open_day_range.first), true;
  else
    return *si = -1, false;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsExprFirstDebugDayIndex(KHE_DRS_EXPR e)                         */
/*                                                                           */
/*  The minimum, over e and its descendants, of the first_day_index field    */
/*  of e.                                                                    */
/*                                                                           */
/*  Implementation note.  We only search down the first children.  If that   */
/*  is wrong in some obscure case, it doesn't matter, because this function  */
/*  is called only when debugging.                                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static int KheDrsExprFirstDebugDayIndex(KHE_DRS_EXPR e)
{
  int res;
  res = e->day_range.first;
  while( HaArrayCount(e->children) > 0 )
  {
    e = HaArrayFirst(e->children);
    if( e->day_range.first < res )
      res = e->day_range.first;
  }
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprDebugChildren(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Carry out the part of KheDrsExprDebug which prints the children.         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprDebugChildren(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  /* KHE_DRS_EXPR child_e;  int i, di, first_day_index; */
  /* bool child_printed; */
  if( soln == NULL )
  {
    /* print all children */
    /* *** needs revision here still to do
    di = e->day_range.first;
    HaArrayForEach(e->children, child_e, i)
    {
      for(; HaArray(e->open_child_indexes, di - e->day_range.first) <= i; di++ )
	fprintf(fp, "%*sending on %s:\n", indent + 2, "",
	  KheTimeGroupId(KheFrameTimeGroup(drs->days_frame, di)));
      KheDrsExprDebug(child_e, soln, drs, verbosity, indent + 2, fp);
    }
    *** */
  }
  else
  {
    /* print the children with a descendant with an entry in the signature */
    if( HaArrayCount(e->children) > 0 )
    {
      fprintf(fp, "(");
      /* *** needs revision here still to do
      child_printed = false;
      HaArrayForEach(e->children, child_e, i)
      {
	first_day_index = KheDrsExprFirstDebugDayIndex(child_e);
	if( first_day_index <= soln->day->solve_index &&
	    soln->day->solve_index <= child_e->last_day_index )
	{
	  if( child_printed )
	    fprintf(fp, ", ");
	  KheDrsExprDebug(child_e, soln, drs, verbosity, -1, fp);
	  child_printed = true;
	}
      }
      *** */
      fprintf(fp, ")");
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheDrsExprDayRangeShow(KHE_DRS_EXPR e,                             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, char buff[100])                       */
/*                                                                           */
/*  Return a string showing the day range of e onto fp.  Use buff for        */
/*  scratch memory.                                                          */
/*                                                                           */
/*****************************************************************************/

/* ***
static char *KheDrsExprDayRangeShow(KHE_DRS_EXPR e,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, char buff[100])
{
  KHE_TIME_GROUP tg1, tg2;
  if( e->day_range.first == e->day_range.last )
  {
    tg1 = KheFrameTimeGroup(drs->days_frame, e->day_range.first);
    snprintf(buff, 100, "%s", KheTimeGroupId(tg1));
  }
  else
  {
    tg1 = KheFrameTimeGroup(drs->days_frame, e->day_range.first);
    tg2 = KheFrameTimeGroup(drs->days_frame, e->day_range.last);
    snprintf(buff, 100, "%s-%s", KheTimeGroupId(tg1), KheTimeGroupId(tg2));
  }
  return buff;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsExprDebug(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of e onto fp with the given verbosity and indent.            */
/*                                                                           */
/*  If soln != NULL, only those parts of e that have a value stored in       */
/*  soln's signature are printed, along with those values.  In that case     */
/*  indent is ignored and the whole print is on one line.                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsExprDebug(KHE_DRS_EXPR e, KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp)
{
  switch( e->tag )
  {
    case KHE_DRS_EXPR_BUSY_TIME_TAG:

      KheDrsExprBusyTimeDebug((KHE_DRS_EXPR_BUSY_TIME) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_FREE_TIME_TAG:

      KheDrsExprFreeTimeDebug((KHE_DRS_EXPR_FREE_TIME) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_WORK_TIME_TAG:

      KheDrsExprWorkTimeDebug((KHE_DRS_EXPR_WORK_TIME) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_BUSY_DAY_TAG:

      KheDrsExprBusyDayDebug((KHE_DRS_EXPR_BUSY_DAY) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_FREE_DAY_TAG:

      KheDrsExprFreeDayDebug((KHE_DRS_EXPR_FREE_DAY) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_WORK_DAY_TAG:

      KheDrsExprWorkDayDebug((KHE_DRS_EXPR_WORK_DAY) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_ASSIGNED_TASK_TAG:

      KheDrsExprAssignedTaskDebug((KHE_DRS_EXPR_ASSIGNED_TASK) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_OR_TAG:

      KheDrsExprOrDebug((KHE_DRS_EXPR_OR) e, soln, drs, verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_AND_TAG:

      KheDrsExprAndDebug((KHE_DRS_EXPR_AND) e, soln, drs, verbosity, indent,fp);
      break;

    case KHE_DRS_EXPR_INT_SUM_TAG:

      KheDrsExprIntSumDebug((KHE_DRS_EXPR_INT_SUM) e, soln, drs,
	verbosity, indent, fp);
      break;

    /* *** currently unused
    case KHE_DRS_EXPR_INT_SEQ_TAG:

      KheDrsExprIntSeqDebug((KHE_DRS_EXPR_INT_SEQ) e, soln, drs, verbosity,
	indent, fp);
      break;
    *** */

    case KHE_DRS_EXPR_FLOAT_SUM_TAG:

      KheDrsExprFloatSumDebug((KHE_DRS_EXPR_FLOAT_SUM) e, soln, drs,
	verbosity, indent,fp);
      break;

    case KHE_DRS_EXPR_INT_DEV_TAG:

      KheDrsExprIntDevDebug((KHE_DRS_EXPR_INT_DEV) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_FLOAT_DEV_TAG:

      KheDrsExprFloatDevDebug((KHE_DRS_EXPR_FLOAT_DEV) e, soln, drs,
	verbosity, indent,fp);
      break;

    case KHE_DRS_EXPR_COST_TAG:

      KheDrsExprCostDebug((KHE_DRS_EXPR_COST) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_INT_SUM_COMB_TAG:

      KheDrsExprIntSumCombDebug((KHE_DRS_EXPR_INT_SUM_COMB) e, soln, drs,
	verbosity, indent, fp);
      break;

    case KHE_DRS_EXPR_INT_SEQ_COMB_TAG:

      KheDrsExprIntSeqCombDebug((KHE_DRS_EXPR_INT_SEQ_COMB) e, soln, drs,
	verbosity, indent, fp);
      break;

    default:

      HnAbort("KheDrsExprDebug internal error: unknown tag %d", e->tag);
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_RESOURCE_ON_DAY"                                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDayMake(KHE_DRS_RESOURCE dr,     */
/*    KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)                      */
/*                                                                           */
/*  Make a new resource on day object with these attributes.                 */
/*  This function assumes that dr is not assigned any task on day.           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDayMake(KHE_DRS_RESOURCE dr,
  KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_RESOURCE_ON_DAY res;
  /* ***
  if( HaArrayCount(drs->free_resource_on_day) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_resource_on_day);
    HaArrayClear(res->external_today);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->external_today, drs->arena);
  }
  *** */
  HaMake(res, drs->arena);
  HaArrayInit(res->external_today, drs->arena);
  res->encl_dr = dr;
  /* res->open = false; */
  res->day = day;
  res->closed_asst = NULL;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceOnDayFree(KHE_DRS_RESOURCE_ON_DAY drd,                */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free drd.                                                                */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsResourceOnDayFree(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  ** ***
  HaArrayForEach(drd->internal_today, e, i)
    KheDrsExprFreeInternal(e, drs);
  *** **
  HaArrayForEach(drd->external_today, e, i)
    KheDrsExprFreeExternal(e, drs);
  HaArrayAddLast(drs->free_resource_on_day, drd);
}
*** */



/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceOnDayAddDomTest(KHE_DRS_RESOURCE_ON_DAY drd,          */
/*    KHE_DRS_DOM_TEST dom_test)                                             */
/*                                                                           */
/*  Add dom_test to drd.  This may change the eq_dom_test_indexes array as   */
/*  well as the dom_tests array.                                             */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete: apparently drd is not the right place to add a dom test now 
static void KheDrsResourceOnDayAddDomTest(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_DOM_TEST dom_test)
{
  if( dom_test.type == KHE_DRS_DOM_EQ || dom_test.type == KHE_DRS_DOM_EQ_LOOSE )
    HaArrayAddLast(drd->eq_dom_test_indexes, HaArrayCount(drd->dom_tests));
  HaArrayAddLast(drd->dom_tests, dom_test);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceOnDayDebug(KHE_DRS_RESOURCE_ON_DAY drd,               */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of drd onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/
static void KheDrsTaskDebug(KHE_DRS_TASK dt, int verbosity,
  int indent, FILE *fp);

static void KheDrsResourceOnDayDebug(KHE_DRS_RESOURCE_ON_DAY drd,
  int verbosity, int indent, FILE *fp)
{
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "drd");
  /* *** 
  if( drd->assigned_task != NULL )
    KheDrsTaskDebug(drd->assigned_task, 1, -1, fp);
  else
    fprintf(fp, "-");
  *** */
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_RESOURCE - converting monitors to expressions"        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddAvoidClashesMonitor(KHE_DRS_RESOURCE dr,           */
/*    KHE_AVOID_CLASHES_MONITOR m, int m_framer_index,                       */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add avoid clashes monitor m to dr.                                       */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsResourceAddAvoidClashesMonitor(KHE_DRS_RESOURCE dr,
  KHE_AVOID_CLASHES_MONITOR m, int m_framer_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ** ignoring this; there will be no clashes **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddAvoidUnavailableTimesMonitor(KHE_DRS_RESOURCE dr,  */
/*    KHE_AVOID_UNAVAILABLE_TIMES_MONITOR m, int m_framer_index,             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add avoid unavailable times monitor m to dr.                             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsResourceAddAvoidUnavailableTimesMonitor(KHE_DRS_RESOURCE dr,
  KHE_AVOID_UNAVAILABLE_TIMES_MONITOR m, int m_framer_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_AVOID_UNAVAILABLE_TIMES_CONSTRAINT c;  KHE_DRS_EXPR_BUSY_TIME eb;
  int i;  KHE_TIME t;  KHE_DRS_EXPR_COST ec;  KHE_COST_FUNCTION cost_fn;
  KHE_TIME_GROUP tg;  KHE_DRS_EXPR_INT_SUM eis;  KHE_COST combined_weight;
  KHE_DRS_DOM_TEST unused_dom_test, le_dom_test;
  c = KheAvoidUnavailableTimesMonitorConstraint(m);
  tg = KheAvoidUnavailableTimesConstraintUnavailableTimes(c);
  unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  combined_weight = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  if( combined_weight > 0 && KheTimeGroupTimeCount(tg) > 0 )
  {
    cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
    if( cost_fn == KHE_LINEAR_COST_FUNCTION || KheTimeGroupTimeCount(tg) == 1 )
    {
      ** a separate expression tree for each time **
      for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
      {
	t = KheTimeGroupTime(tg, i);
	ec = KheDrsExprCostMake((KHE_MONITOR) m, cost_fn, combined_weight, drs);
	HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) ec);
	** ***
	HaArrayAddLast(dr->exprs, (KHE_DRS_EXPR) ec);
	KheDrsExprAddChild((KHE_DRS_EXPR) drs->root_expr, (KHE_DRS_EXPR) ec);
	*** **
	eb = KheDrsExprBusyTimeMake(dr, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eb);
      }
    }
    else
    {
      ** a single expression tree (max limit of 0, in effect) **
      ec = KheDrsExprCostMake((KHE_MONITOR) m, cost_fn, combined_weight, drs);
      HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) ec);
      ** HaArrayAddLast(dr->exprs, (KHE_DRS_EXPR) ec); **
      ** KheDrsExprAddChild((KHE_DRS_EXPR) drs->root_expr, (KHE_DRS_EXPR) ec);**
      le_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);
      eis = KheDrsExprIn tSumMake(le_dom_test, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eis);
      for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
      {
	t = KheTimeGroupTime(tg, i);
	eb = KheDrsExprBusyTimeMake(dr, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eis, (KHE_DRS_EXPR) eb);
      }
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddLimitIdleTimesMonitor(KHE_DRS_RESOURCE dr,         */
/*    KHE_LIMIT_IDLE_TIMES_MONITOR m, int m_framer_index,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add limit idle times monitor m to dr.                                    */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsResourceAddLimitIdleTimesMonitor(KHE_DRS_RESOURCE dr,
  KHE_LIMIT_IDLE_TIMES_MONITOR m, int m_framer_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ** not included in the solve, and that fact is documented **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheTimeGroupContainsWholeDay(KHE_TIME_GROUP tg, int start_index,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_TIME_GROUP *day_tg)               */
/*                                                                           */
/*  If tg contains a complete day's worth of times, starting from            */
/*  start_index, then return true and set *day_tg to the day time group,     */
/*  otherwise return false.                                                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheTimeGroupContainsWholeDay(KHE_TIME_GROUP tg, int start_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_TIME_GROUP *day_tg)
{
  KHE_TIME tg_t, day_t;  int i;  KHE_TIME_GROUP d_tg;
  tg_t = KheTimeGroupTime(tg, start_index);
  d_tg = KheFrameTimeTimeGroup(drs->days_frame, tg_t);
  for( i = 0;  i < KheTimeGroupTimeCount(d_tg);  i++ )
  {
    day_t = KheTimeGroupTime(d_tg, i);
    if( start_index + i >= KheTimeGroupTimeCount(tg) )
      return *day_tg = NULL, false;
    tg_t = KheTimeGroupTime(tg, start_index + i);
    if( day_t != tg_t )
      return *day_tg = NULL, false;
  }
  return *day_tg = d_tg, true;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheMakeOrExprAndChildren(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,    */
/*    KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make an OR expression with one BUSY child for each time of tg, and add   */
/*  the OR expression as a child of prnt.  If tg has exactly one time,       */
/*  optimize by adding just the BUSY expression for that time as a child     */
/*  of prnt.                                                                 */
/*                                                                           */
/*  Either way, add dom_test to the root of these vertices.                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheMakeOrExprAndChildren(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,
  KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_OR eo;  KHE_DRS_EXPR_BUSY_TIME ebt;  int i;  KHE_TIME t;
  KHE_DRS_DOM_TEST unused_dom_test;  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_TIME_GROUP day_tg;
  if( KheTimeGroupTimeCount(tg) == 1 )
  {
    ** just a single BUSY_TIME expression **
    t = KheTimeGroupTime(tg, 0);
    ebt = KheDrsExprBusyTimeMake(dr, t, dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) ebt);
  }
  else if( KheTimeGroupContainsWholeDay(tg, 0, drs, &day_tg) &&
    KheTimeGroupTimeCount(tg) == KheTimeGroupTimeCount(day_tg) )
  {
    ** just a single BUSY_DAY expression **
    ebd = KheDrsExprBusyDayMake(dr, day_tg, dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) ebd);
  }
  else
  {
    ** an OR expression with BUSY_DAY and BUSY_TIME children **
    eo = KheDrsExprO rMake(dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) eo);
    unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
    for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
    {
      if( KheTimeGroupContainsWholeDay(tg, i, drs, &day_tg) )
      {
	** add a BUSY_DAY expression **
	ebd = KheDrsExprBusyDayMake(dr, day_tg, dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eo, (KHE_DRS_EXPR) ebd);
	i += KheTimeGroupTimeCount(day_tg) - 1;
      }
      else
      {
	** add a BUSY_TIME expression **
	t = KheTimeGroupTime(tg, i);
	ebt = KheDrsExprBusyTimeMake(dr, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eo, (KHE_DRS_EXPR) ebt);
      }
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheMakeAndExprAndChildren(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,   */
/*    KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make an AND expression with one FREE child for each time of tg, and add  */
/*  the AND expression as a child of prnt.  If tg has exactly one time,      */
/*  optimize by adding just the FREE expression for that time as a child     */
/*  of prnt.                                                                 */
/*                                                                           */
/*  Either way, add dom_test to the root of these vertices.                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheMakeAndExprAndChildren(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,
  KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_AND ea;  KHE_DRS_EXPR_FREE_TIME eft;  int i;  KHE_TIME t;
  KHE_DRS_DOM_TEST unused_dom_test;  KHE_DRS_EXPR_FREE_DAY efd;
  KHE_TIME_GROUP day_tg;
  if( KheTimeGroupTimeCount(tg) == 1 )
  {
    ** just a single FREE expression **
    t = KheTimeGroupTime(tg, 0);
    eft = KheDrsExprFreeTimeMake(dr, t, dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) eft);
  }
  else if( KheTimeGroupContainsWholeDay(tg, 0, drs, &day_tg) &&
    KheTimeGroupTimeCount(tg) == KheTimeGroupTimeCount(day_tg) )
  {
    ** just a single FREE_DAY expression **
    efd = KheDrsExprFreeDayMake(dr, day_tg, dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) efd);
  }
  else
  {
    ** an AND expression with FREE_DAY and FREE_TIME children **
    ea = KheDrsExprA ndMake(dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) ea);
    unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
    for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
    {
      if( KheTimeGroupContainsWholeDay(tg, i, drs, &day_tg) )
      {
	** add a FREE_DAY expression **
	efd = KheDrsExprFreeDayMake(dr, day_tg, dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) ea, (KHE_DRS_EXPR) efd);
	i += KheTimeGroupTimeCount(day_tg) - 1;
      }
      else
      {
	** add a FREE_TIME expression **
	t = KheTimeGroupTime(tg, i);
	eft = KheDrsExprFreeTimeMake(dr, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) ea, (KHE_DRS_EXPR) eft);
      }
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheLimitsToTests(int min_limit, int max_limit, int max_max_limit,   */
/*    bool allow_zero, KHE_DRS_DOM_TEST *alpha_dom_test,                     */
/*    KHE_DRS_DOM_TEST *beta_dom_test)                                       */
/*                                                                           */
/*  Convert limits (min_limit, max_limit, allow_zero) into dominance tests.  */
/*  if min_limit <= 0 there is no min limit; if max_limit >= max_max_limit   */
/*  there is no maximum limit.                                               */
/*                                                                           */
/*  The values returned in *alpha_dom_test and *beta_dom_test are referred   */
/*  to as alpha and beta in the User's Guide.                                */
/*                                                                           */
/*  Note.  If there is a trivial allow_zero flag (if min_limit <= 1), an     */
/*  adjustment will have already been made by now to remove it.  So this     */
/*  call may assume that allow_zero implies min_limit > 1.                   */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheLimitsToTests(int min_limit, int max_limit, int max_max_limit,
  bool allow_zero, KHE_DRS_DOM_TEST *alpha_dom_test,
  KHE_DRS_DOM_TEST *beta_dom_test)
{
  if( allow_zero ** && min_limit > 1 ** )
  {
    ** fall back on equality when there is a non-trivial allow_zero flag **
    *alpha_dom_test = *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_EQ);
  }
  else if( max_limit >= max_max_limit )
  {
    if( min_limit <= 0 )
    {
      ** no max_limit, and no min_limit (should never happen) **
      HnAbort("KheLimitsToTests: no max limit, no min limit");
      *alpha_dom_test = *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
    }
    else
    {
      ** no max_limit, and min_limit **
      *alpha_dom_test = KheDrsDomTestMakeLoose(KHE_DRS_DOM_GE_LOOSE, min_limit);
      *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_GE);
    }
  }
  else
  {
    if( min_limit <= 0 )
    {
      ** max_limit, and no min_limit **
      *alpha_dom_test = *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);
    }
    else
    {
      ** max_limit, and min_limit **
      *alpha_dom_test = KheDrsDomTestMakeLoose(KHE_DRS_DOM_EQ_LOOSE, min_limit);
      *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_EQ);
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsClusterMonitorContainsWholePositiveDay(                       */
/*    KHE_CLUSTER_BUSY_TIMES_MONITOR m, int start_index,                     */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_TIME_GROUP *day_tg)               */
/*                                                                           */
/*  If m contains a sequence of singleton time groups starting at day_index  */
/*  that have postive polarity and together cover a complete day, return     */
/*  true and set *day_tg to the day time group;  otherwise return false.     */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDrsClusterMonitorContainsWholePositiveDay(
  KHE_CLUSTER_BUSY_TIMES_MONITOR m, int start_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_TIME_GROUP *day_tg)
{
  KHE_TIME day_t;  int i, tg_count;  KHE_TIME_GROUP m_tg, d_tg;
  KHE_POLARITY po;  KHE_TIME t;
  tg_count = KheClusterBusyTimesMonitorTimeGroupCount(m);
  m_tg = KheClusterBusyTimesMonitorTimeGroup(m, start_index, &po);
  if( po == KHE_POSITIVE && KheTimeGroupTimeCount(m_tg) == 1 )
  {
    t = KheTimeGroupTime(m_tg, 0);
    d_tg = KheFrameTimeTimeGroup(drs->days_frame, t);
    for( i = 0;  i < KheTimeGroupTimeCount(d_tg);  i++ )
    {
      day_t = KheTimeGroupTime(d_tg, i);
      if( start_index + i >= tg_count )
	return *day_tg = NULL, false;
      m_tg = KheClusterBusyTimesMonitorTimeGroup(m, start_index + i, &po);
      if( po != KHE_POSITIVE || KheTimeGroupTimeCount(m_tg) != 1 ||
	  KheTimeGroupTime(m_tg, 0) != day_t )
	return *day_tg = NULL, false;
    }
    return *day_tg = d_tg, true;
  }
  else
    return *day_tg = NULL, false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddClusterBusyTimesMonitor(KHE_DRS_RESOURCE dr,       */
/*    KHE_CLUSTER_BUSY_TIMES_MONITOR m, int m_framer_index,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add cluster busy times monitor m to dr.                                  */
/*                                                                           */
/*  Implementation note.  If m contains a sequence of consecutive            */
/*  singleton time groups with positive polarity that add up to a complete   */
/*  day, then one BUSY_DAY expression is generated for them instead of one   */
/*  BUSY_TIME expression for each.  This is correct the sum reported         */
/*  in either case is 1 if the day is busy and 0 otherwise.  It cannot       */
/*  be more than 1 because the resource takes at most one task each day.     */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsResourceAddClusterBusyTimesMonitor(KHE_DRS_RESOURCE dr,
  KHE_CLUSTER_BUSY_TIMES_MONITOR m, int m_framer_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c;  KHE_TIME_GROUP tg, day_tg;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_weight;  bool allow_zero;
  int i, min_limit, max_limit, history_before, history_after, tg_count;
  KHE_DRS_EXPR_INT_SUM_COMB eisc;  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_POLARITY po;  KHE_DRS_DOM_TEST alpha_dom_test, beta_dom_test;

  ** boilerplate, and return if constraint does nothing **
  c = KheClusterBusyTimesMonitorConstraint(m);
  combined_weight = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  min_limit = KheClusterBusyTimesConstraintMinimum(c);
  max_limit = KheClusterBusyTimesConstraintMaximum(c);
  allow_zero = KheClusterBusyTimesConstraintAllowZero(c);
  history_before = KheClusterBusyTimesMonitorHistory(m);
  history_after = KheClusterBusyTimesMonitorHistoryAfter(m);
  if( allow_zero && min_limit <= 1 )
    min_limit = 0, allow_zero = false;
  tg_count = KheClusterBusyTimesConstraintTimeGroupCount(c);
  if( (min_limit <= 0 && max_limit >= tg_count) || combined_weight == 0 )
    return;

  ** work out the various dom_tests **
  KheLimitsToTests(min_limit, max_limit, tg_count, allow_zero,
    &alpha_dom_test, &beta_dom_test);

  ** one int_cost_comb expression at top of tree **
  eisc = KheDrsExprIntSumCombMake((KHE_MONITOR) m, cost_fn, combined_weight,
    min_limit, max_limit, allow_zero, history_before, history_after,
    alpha_dom_test, drs);
  HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) eisc);
  ** HaArrayAddLast(dr->exprs, (KHE_DRS_EXPR) eisc); **
  ** KheDrsExprAddChild((KHE_DRS_EXPR) drs->root_expr, (KHE_DRS_EXPR) eisc); **

  ** one OR or AND expression for each time group **
  ** or a BUSY_DAY or FREE_DAY expression if all singles of one day present **
  for( i = 0;  i < tg_count;  i++ )
  {
    if( KheDrsClusterMonitorContainsWholePositiveDay(m, i, drs, &day_tg) )
    {
      ** add a BUSY_DAY expression **
      ebd = KheDrsExprBusyDayMake(dr, day_tg, beta_dom_test, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) ebd);
      i += KheTimeGroupTimeCount(day_tg) - 1;
    }
    else
    {
      tg = KheClusterBusyTimesMonitorTimeGroup(m, i, &po);
      if( po == KHE_POSITIVE )
      {
	** make an OR expression with one BUSY expression for each time of tg **
	KheMakeOrExprAndChildren(dr, (KHE_DRS_EXPR) eisc, tg,beta_dom_test,drs);
      }
      else
      {
	** make an AND expression with one FREE expression for each time of tg **
	KheMakeAndExprAndChildren(dr, (KHE_DRS_EXPR) eisc,tg,beta_dom_test,drs);
      }
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddLimitBusyTimesMonitor(KHE_DRS_RESOURCE dr,         */
/*    KHE_LIMIT_BUSY_TIMES_MONITOR m, int m_framer_index,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add limit busy times monitor m to dr.                                    */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsResourceAddLimitBusyTimesMonitor(KHE_DRS_RESOURCE dr,
  KHE_LIMIT_BUSY_TIMES_MONITOR m, int m_framer_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_LIMIT_BUSY_TIMES_CONSTRAINT c;  KHE_TIME_GROUP tg, day_tg;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_weight;
  int i, j, junk, min_limit, max_limit;  bool allow_zero;
  KHE_DRS_EXPR_INT_SUM_COMB eisc;  KHE_TIME t;
  KHE_DRS_EXPR_INT_SUM eis, eis2;  KHE_DRS_EXPR_COST ec;
  KHE_DRS_EXPR_INT_DEV eid;
  KHE_DRS_EXPR_BUSY_TIME ebt;
  KHE_DRS_DOM_TEST unused_dom_test, le_dom_test, alpha_dom_test, beta_dom_test;

  ** boilerplate, and return if constraint does nothing **
  c = KheLimitBusyTimesMonitorConstraint(m);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  combined_weight = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  min_limit = KheLimitBusyTimesConstraintMinimum(c);
  max_limit = KheLimitBusyTimesConstraintMaximum(c);
  allow_zero = KheLimitBusyTimesConstraintAllowZero(c);
  if( allow_zero && min_limit <= 1 )
    min_limit = 0, allow_zero = false;
  unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  le_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);
  if( combined_weight == 0 )
    return;

  if( cost_fn == KHE_LINEAR_COST_FUNCTION ||
      KheLimitBusyTimesMonitorTimeGroupCount(m) == 1 )
  {
    ** one INT_SUM_COMB for each time group **
    ** unless the time group is one day, min_limit = 0, and max_limit = 1 **
    for( i = 0;  i < KheLimitBusyTimesMonitorTimeGroupCount(m);  i++ )
    {
      tg = KheLimitBusyTimesMonitorTimeGroup(m, i, &junk);
      if( min_limit == 0 && max_limit == 1 &&
	  KheTimeGroupContainsWholeDay(tg, 0, drs, &day_tg) &&
	  KheTimeGroupTimeCount(tg) == KheTimeGroupTimeCount(day_tg) )
      {
	** constraint specifies at most one task per day; trivial, so omit **
      }
      else if( min_limit > 0 || max_limit < KheTimeGroupTimeCount(tg) )
      {
	KheLimitsToTests(min_limit, max_limit, KheTimeGroupTimeCount(tg),
	  allow_zero, &alpha_dom_test, &beta_dom_test);
	eisc = KheDrsExprIntSumCombMake((KHE_MONITOR) m, cost_fn,
	  combined_weight, min_limit, max_limit, allow_zero, 0, 0,
	  alpha_dom_test, drs);
	HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) eisc);
	** ***
	HaArrayAddLast(dr->exprs, (KHE_DRS_EXPR) eisc);
	KheDrsExprAddChild((KHE_DRS_EXPR) drs->root_expr, (KHE_DRS_EXPR) eisc);
	*** **
	for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
	{
	  t = KheTimeGroupTime(tg, j);
	  ebt = KheDrsExprBusyTimeMake(dr, t, unused_dom_test, drs);
	  KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) ebt);
	}
      }
    }
  }
  else
  {
    ** one COST and INT_SUM at root of tree **
    ec = KheDrsExprCostMake((KHE_MONITOR) m, cost_fn, combined_weight, drs);
    HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) ec);
    ** HaArrayAddLast(dr->exprs, (KHE_DRS_EXPR) ec); **
    ** KheDrsExprAddChild((KHE_DRS_EXPR) drs->root_expr, (KHE_DRS_EXPR) ec); **
    eis = KheDrsExprIn tSumMake(le_dom_test, drs);
    KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eis);

    ** one INT_DEV and INT_SUM for each time group **
    for( i = 0;  i < KheLimitBusyTimesMonitorTimeGroupCount(m);  i++ )
    {
      tg = KheLimitBusyTimesMonitorTimeGroup(m, i, &junk);
      eid = KheDrsExprIn tDevMake(min_limit, max_limit, allow_zero, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) eis, (KHE_DRS_EXPR) eid);
      KheLimitsToTests(min_limit, max_limit, KheTimeGroupTimeCount(tg),
	allow_zero, &alpha_dom_test, &beta_dom_test);
      eis2 = KheDrsExprIntS umMake(alpha_dom_test, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) eid, (KHE_DRS_EXPR) eis2);
      for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
      {
	t = KheTimeGroupTime(tg, j);
	ebt = KheDrsExprBusyTimeMake(dr, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eis2, (KHE_DRS_EXPR) ebt);
      }
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddLimitWorkloadMonitor(KHE_DRS_RESOURCE dr,          */
/*    KHE_LIMIT_WORKLOAD_MONITOR m, int m_framer_index,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add limit workload monitor m to dr.                                      */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsResourceAddLimitWorkloadMonitor(KHE_DRS_RESOURCE dr,
  KHE_LIMIT_WORKLOAD_MONITOR m, int m_framer_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_LIMIT_WORKLOAD_CONSTRAINT c;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_weight;
  int i, j, min_limit, max_limit;  bool allow_zero;  float junk;
  KHE_TIME_GROUP tg;  KHE_TIME t;  KHE_DRS_EXPR_WORK_TIME ew;
  KHE_DRS_EXPR_INT_SUM eis;  KHE_DRS_EXPR_COST ec;
  KHE_DRS_EXPR_FLOAT_SUM efs;  KHE_DRS_EXPR_FLOAT_DEV efd;
  KHE_DRS_DOM_TEST unused_dom_test, le_dom_test, alpha_dom_test, beta_dom_test;

  ** boilerplate, and return if constraint does nothing **
  c = KheLimitWorkloadMonitorConstraint(m);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  combined_weight = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  min_limit = KheLimitWorkloadConstraintMinimum(c);
  max_limit = KheLimitWorkloadConstraintMaximum(c);
  allow_zero = KheLimitWorkloadConstraintAllowZero(c);
  if( allow_zero && min_limit <= 1 )
    min_limit = 0, allow_zero = false;
  unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  le_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);
  if( combined_weight == 0 )
    return;

  ** one COST and INT_SUM at the root **
  ec = KheDrsExprCostMake((KHE_MONITOR) m, cost_fn, combined_weight, drs);
  HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) ec);
  ** HaArrayAddLast(dr->exprs, (KHE_DRS_EXPR) ec); **
  ** KheDrsExprAddChild((KHE_DRS_EXPR) drs->root_expr, (KHE_DRS_EXPR) ec); **
  eis = KheDrsExprInt SumMake(le_dom_test, drs);
  KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eis);

  ** one FLOAT_DEV and FLOAT_SUM for each time group **
  KheLimitsToTests(min_limit, max_limit, INT_MAX, allow_zero,
    &alpha_dom_test, &beta_dom_test);
  for( i = 0;  i < KheLimitWorkloadMonitorTimeGroupCount(m);  i++ )
  {
    tg = KheLimitWorkloadMonitorTimeGroup(m, i, &junk);
    efd = KheDrsExprFlo atDevMake(min_limit, max_limit, allow_zero, drs);
    KheDrsExprAddChild((KHE_DRS_EXPR) eis, (KHE_DRS_EXPR) efd);
    efs = KheDrsExprFloa tSumMake(alpha_dom_test, drs);
    KheDrsExprAddChild((KHE_DRS_EXPR) efd, (KHE_DRS_EXPR) efs);
    for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
    {
      t = KheTimeGroupTime(tg, j);
      ew = KheDrsExprWorkTimeMake(dr, t, unused_dom_test, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) efs, (KHE_DRS_EXPR) ew);
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddLimitActiveIntervalsMonitor(KHE_DRS_RESOURCE dr,   */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR m, int m_framer_index,              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add limit active intervals monitor m to dr.                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsResourceAddLimitActiveIntervalsMonitor(KHE_DRS_RESOURCE dr,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR m, int m_framer_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c;  KHE_TIME_GROUP tg;  KHE_POLARITY po;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_weight;
  int i, min_limit, max_limit, history_before, history_after, tg_count;
  KHE_DRS_EXPR_INT_SEQ_COMB eisc;
  KHE_DRS_DOM_TEST alpha_dom_test, beta_dom_test;

  ** boilerplate, and return if constraint does nothing **
  c = KheLimitActiveIntervalsMonitorConstraint(m);
  combined_weight = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  min_limit = KheLimitActiveIntervalsConstraintMinimum(c);
  max_limit = KheLimitActiveIntervalsConstraintMaximum(c);
  history_before = KheLimitActiveIntervalsMonitorHistory(m);
  history_after = KheLimitActiveIntervalsMonitorHistoryAfter(m);
  tg_count = KheLimitActiveIntervalsConstraintTimeGroupCount(c);
  if( (min_limit <= 1 && max_limit >= tg_count) || combined_weight == 0 )
    return;

  ** work out the various dom_tests **
  KheLimitsToTests(min_limit, max_limit, tg_count, false,
    &alpha_dom_test, &beta_dom_test);

  ** one int_seq_comb expression at top of tree **
  eisc = KheDrsExprIntSeqCombMake((KHE_MONITOR) m, cost_fn, combined_weight,
    min_limit, max_limit, history_before, history_after, alpha_dom_test, drs);
  HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) eisc);
  ** HaArrayAddLast(dr->exprs, (KHE_DRS_EXPR) eisc); **
  ** KheDrsExprAddChild((KHE_DRS_EXPR) drs->root_expr, (KHE_DRS_EXPR) eisc); **

  ** one OR or AND expression for each time group **
  for( i = 0;  i < KheLimitActiveIntervalsMonitorTimeGroupCount(m);  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(m, i, &po);
    if( po == KHE_POSITIVE )
    {
      ** make an OR expression with one BUSY expression for each time of tg **
      KheMakeOrExprAndChildren(dr, (KHE_DRS_EXPR) eisc, tg, beta_dom_test, drs);
    }
    else
    {
      ** make an AND expression with one FREE expression for each time of tg **
      KheMakeAndExprAndChildren(dr, (KHE_DRS_EXPR) eisc, tg, beta_dom_test,drs);
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddMonitor(KHE_DRS_RESOURCE dr,                       */
/*    KHE_MONITOR m, int m_framer_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Add m to dr.  Here m_framer_index is m's index in drs->monitor_framer.   */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsResourceAddMonitor(KHE_DRS_RESOURCE dr,
  KHE_MONITOR m, int m_framer_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( KheMonitorTag(m) )
  {
    case KHE_AVOID_CLASHES_MONITOR_TAG:

      KheDrsResourceAddAvoidClashesMonitor(dr,
        (KHE_AVOID_CLASHES_MONITOR) m, m_framer_index, drs);
      break;

    case KHE_AVOID_UNAVAILABLE_TIMES_MONITOR_TAG:

      KheDrsResourceAddAvoidUnavailableTimesMonitor(dr,
        (KHE_AVOID_UNAVAILABLE_TIMES_MONITOR) m, m_framer_index, drs);
      break;

    case KHE_LIMIT_IDLE_TIMES_MONITOR_TAG:

      KheDrsResourceAddLimitIdleTimesMonitor(dr,
        (KHE_LIMIT_IDLE_TIMES_MONITOR) m, m_framer_index, drs);
      break;

    case KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG:

      KheDrsResourceAddClusterBusyTimesMonitor(dr,
        (KHE_CLUSTER_BUSY_TIMES_MONITOR) m, m_framer_index, drs);
      break;

    case KHE_LIMIT_BUSY_TIMES_MONITOR_TAG:

      KheDrsResourceAddLimitBusyTimesMonitor(dr,
        (KHE_LIMIT_BUSY_TIMES_MONITOR) m, m_framer_index, drs);
      break;

    case KHE_LIMIT_WORKLOAD_MONITOR_TAG:

      KheDrsResourceAddLimitWorkloadMonitor(dr,
        (KHE_LIMIT_WORKLOAD_MONITOR) m, m_framer_index, drs);
      break;

    case KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG:

      KheDrsResourceAddLimitActiveIntervalsMonitor(dr,
        (KHE_LIMIT_ACTIVE_INTERVALS_MONITOR) m, m_framer_index, drs);
      break;

    default:

      HnAbort("KheDrsResourceAddMonitor internal error (tag %d)",
        KheMonitorTag(m));
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheMonitorLimit(KHE_MONITOR m)                                       */
/*                                                                           */
/*  Return the maximum limit of m if there is one, or else the minimum       */
/*  limit if there is one, or else 0.                                        */
/*                                                                           */
/*****************************************************************************/

static int KheMonitorLimit(KHE_MONITOR m)
{
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT cbtc;  int val;
  KHE_LIMIT_BUSY_TIMES_CONSTRAINT lbtc;
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim;
  switch( KheMonitorTag(m) )
  {
    case KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG:

      cbtc = KheClusterBusyTimesMonitorConstraint(
	(KHE_CLUSTER_BUSY_TIMES_MONITOR) m);
      val = KheClusterBusyTimesConstraintMaximum(cbtc);
      if( val < KheClusterBusyTimesConstraintTimeGroupCount(cbtc) )
	return val;
      val = KheClusterBusyTimesConstraintMinimum(cbtc);
      return val;

    case KHE_LIMIT_BUSY_TIMES_MONITOR_TAG:

      lbtc = KheLimitBusyTimesMonitorConstraint(
	(KHE_LIMIT_BUSY_TIMES_MONITOR) m);
      val = KheLimitBusyTimesConstraintMaximum(lbtc);
      if( val < INT_MAX )
	return val;
      val = KheLimitBusyTimesConstraintMinimum(lbtc);
      if( val < INT_MAX )
	return val;
      return 0;

    case KHE_LIMIT_WORKLOAD_MONITOR_TAG:

      /* could do better here */
      return 0;

    case KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG:

      laim = (KHE_LIMIT_ACTIVE_INTERVALS_MONITOR) m;
      val = KheLimitActiveIntervalsMonitorMaximum(laim);
      if( val < KheLimitActiveIntervalsMonitorTimeGroupCount(laim) )
	return val;
      val = KheLimitActiveIntervalsMonitorMinimum(laim);
      if( val < INT_MAX )
	return val;
      return 0;

    case KHE_AVOID_CLASHES_MONITOR_TAG:
    case KHE_AVOID_UNAVAILABLE_TIMES_MONITOR_TAG:
    case KHE_LIMIT_IDLE_TIMES_MONITOR_TAG:
    default:

      return 0;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  int KheMonitorLargestLimitFirstCmp(const void *p1, const void *p2)       */
/*                                                                           */
/*  Comparison function for sorting monitors so that those with the          */
/*  largest limits come first.                                               */
/*                                                                           */
/*****************************************************************************/

static int KheMonitorLargestLimitFirstCmp(const void *p1, const void *p2)
{
  KHE_MONITOR m1 = * (KHE_MONITOR *) p1;
  KHE_MONITOR m2 = * (KHE_MONITOR *) p2;
  return KheMonitorLimit(m2) - KheMonitorLimit(m1);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_RESOURCE"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsResourceOnDayHasTimeExpr(KHE_DRS_RESOURCE_ON_DAY drd,         */
/*    KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME time,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e)                  */
/*                                                                           */
/*  If dr has a BUSY_TIME, FREE_TIME, or WORK_TIME expression with the       */
/*  given tag, dom_test, and time, then return true and set *res_e to it.    */
/*  Otherwise return false with *res_e set to NULL.                          */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsResourceOnDayHasTimeExpr(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME time,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e)
{
  int i;  KHE_DRS_EXPR e;
  KHE_DRS_EXPR_BUSY_TIME ebt;
  KHE_DRS_EXPR_FREE_TIME eft;
  KHE_DRS_EXPR_WORK_TIME ewt;
  HaArrayForEach(drd->external_today, e, i)
    if( e->tag == tag && KheDrsDomTestEqual(e->dom_test, dom_test) )
    {
      switch( e->tag )
      {
	case KHE_DRS_EXPR_BUSY_TIME_TAG:

	  ebt = (KHE_DRS_EXPR_BUSY_TIME) e;
	  if( ebt->time == time )
	    return *res_e = e, true;
	  break;

	case KHE_DRS_EXPR_FREE_TIME_TAG:

	  eft = (KHE_DRS_EXPR_FREE_TIME) e;
	  if( eft->time == time )
	    return *res_e = e, true;
	  break;

	case KHE_DRS_EXPR_WORK_TIME_TAG:

	  ewt = (KHE_DRS_EXPR_WORK_TIME) e;
	  if( ewt->time == time )
	    return *res_e = e, true;
	  break;

	default:

	  HnAbort("KheDrsResourceOnDayHasTimeExpr: invalid tag (%d)", e->tag);
      }
    }
  return *res_e = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsHasTimeExpr(KHE_DYNAMIC_RESOURCE_SOLVER drs,                  */
/*    KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME time,        */
/*    KHE_DRS_EXPR *res_e)                                                   */
/*                                                                           */
/*  If a BUSY_TIME, FREE_TIME, or WORK_TIME expression exists with the       */
/*  given tag, dom_test, and time, then return true and set *res_e to it.    */
/*  Otherwise return false with *res_e set to NULL.                          */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDrsHasTimeExpr(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME time,
  KHE_DRS_EXPR *res_e)
{
  int i;  KHE_DRS_DAY day;  KHE_DRS_EXPR e;
  KHE_DRS_EXPR_BUSY_TIME ebt;
  KHE_DRS_EXPR_FREE_TIME eft;
  KHE_DRS_EXPR_WORK_TIME ewt;
  i = KheFrameTimeIndex(drs->days_frame, time);
  day = HaArray(drs->days, i);
  HaArrayForEach(day->external_today, e, i)
    if( e->tag == tag && KheDrsDomTestEqual(e->dom_test, dom_test) )
    {
      switch( e->tag )
      {
	case KHE_DRS_EXPR_BUSY_TIME_TAG:

	  ebt = (KHE_DRS_EXPR_BUSY_TIME) e;
	  if( ebt->time == time )
	    return *res_e = e, true;
	  break;

	case KHE_DRS_EXPR_FREE_TIME_TAG:

	  eft = (KHE_DRS_EXPR_FREE_TIME) e;
	  if( eft->time == time )
	    return *res_e = e, true;
	  break;

	case KHE_DRS_EXPR_WORK_TIME_TAG:

	  ewt = (KHE_DRS_EXPR_WORK_TIME) e;
	  if( ewt->time == time )
	    return *res_e = e, true;
	  break;

	default:

	  HnAbort("KheDrsHasTimeExpr: invalid tag (%d)", e->tag);
      }
    }
  return *res_e = NULL, false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsResourceOnDayHasDayExpr(KHE_DRS_RESOURCE_ON_DAY drd,          */
/*    KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME_GROUP tg,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e)                  */
/*                                                                           */
/*  If dr has a BUSY_DAY, FREE_DAY, or WORK_DAY expression with the given    */
/*  tag, dom_test, and time group, then return true and set *res_e to it.    */
/*  Otherwise return false with *res_e set to NULL.                          */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsResourceOnDayHasDayExpr(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME_GROUP tg,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR *res_e)
{
  int i;  KHE_DRS_EXPR e;
  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_DRS_EXPR_FREE_DAY efd;
  KHE_DRS_EXPR_WORK_DAY ewd;
  if( KheTimeGroupTimeCount(tg) > 0 )
  {
    /* ***
    HnAssert(KheTimeGroupEqual(day->time_group, tg),
      "KheDrsHasDayExpr internal error");
    *** */
    HaArrayForEach(drd->external_today, e, i)
      if( e->tag == tag && KheDrsDomTestEqual(e->dom_test, dom_test) )
      {
	switch( e->tag )
	{
	  case KHE_DRS_EXPR_BUSY_DAY_TAG:

	    ebd = (KHE_DRS_EXPR_BUSY_DAY) e;
	    if( KheTimeGroupEqual(ebd->time_group, tg) )
	      return *res_e = e, true;
	    break;

	  case KHE_DRS_EXPR_FREE_DAY_TAG:

	    efd = (KHE_DRS_EXPR_FREE_DAY) e;
	    if( KheTimeGroupEqual(efd->time_group, tg) )
	      return *res_e = e, true;
	    break;

	  case KHE_DRS_EXPR_WORK_DAY_TAG:

	    ewd = (KHE_DRS_EXPR_WORK_DAY) e;
	    if( KheTimeGroupEqual(ewd->time_group, tg) )
	      return *res_e = e, true;
	    break;

	  default:

	    HnAbort("KheDrsResourceHasDayExpr: invalid tag (%d)", e->tag);
	}
      }
  }
  return *res_e = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsHasDayExpr(KHE_DYNAMIC_RESOURCE_SOLVER drs,                   */
/*    KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME_GROUP tg,    */
/*    KHE_DRS_EXPR *res_e)                                                   */
/*                                                                           */
/*  If a BUSY_DAY, FREE_DAY, or WORK_DAY expression exists with the given    */
/*  tag, dom_test, and time group, then return true and set *res_e to it.    */
/*  Otherwise return false with *res_e set to NULL.                          */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDrsHasDayExpr(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_DRS_EXPR_TAG tag, KHE_DRS_DOM_TEST dom_test, KHE_TIME_GROUP tg,
  KHE_DRS_EXPR *res_e)
{
  int i;  KHE_DRS_DAY day;  KHE_DRS_EXPR e;
  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_DRS_EXPR_FREE_DAY efd;
  KHE_DRS_EXPR_WORK_DAY ewd;
  if( KheTimeGroupTimeCount(tg) > 0 )
  {
    i = KheFrameTimeIndex(drs->days_frame, KheTimeGroupTime(tg, 0));
    day = HaArray(drs->days, i);
    HnAssert(KheTimeGroupEqual(day->time_group, tg),
      "KheDrsHasDayExpr internal error");
    HaArrayForEach(day->external_today, e, i)
      if( e->tag == tag && KheDrsDomTestEqual(e->dom_test, dom_test) )
      {
	switch( e->tag )
	{
	  case KHE_DRS_EXPR_BUSY_DAY_TAG:

	    ebd = (KHE_DRS_EXPR_BUSY_DAY) e;
	    if( KheTimeGroupEqual(ebd->time_group, tg) )
	      return *res_e = e, true;
	    break;

	  case KHE_DRS_EXPR_FREE_DAY_TAG:

	    efd = (KHE_DRS_EXPR_FREE_DAY) e;
	    if( KheTimeGroupEqual(efd->time_group, tg) )
	      return *res_e = e, true;
	    break;

	  case KHE_DRS_EXPR_WORK_DAY_TAG:

	    ewd = (KHE_DRS_EXPR_WORK_DAY) e;
	    if( KheTimeGroupEqual(ewd->time_group, tg) )
	      return *res_e = e, true;
	    break;

	  default:

	    HnAbort("KheDrsHasDayExpr: invalid tag (%d)", e->tag);
	}
      }
  }
  return *res_e = NULL, false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceOnDayAddExpr(KHE_DRS_RESOURCE_ON_DAY drd,             */
/*    KHE_DRS_EXPR e)                                                        */
/*                                                                           */
/*  Add e, known to be new, to drd.                                          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceOnDayAddExpr(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_EXPR e)
/* ***
static void KheDrsResourceOnDayAddTimeExpr(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_TIME time, KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR e)
*** */
{
  /* ***
  int i;  KHE_DRS_RESOURCE_ON_DAY drd;
  i = KheFrameTimeIndex(drs->days_frame, time);
  drd = HaArray(dr->days, i);
  *** */
  HaArrayAddLast(drd->external_today, e);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddDayExpr(KHE_DRS_RESOURCE dr,                       */
/*    KHE_TIME_GROUP tg, KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR e)    */
/*                                                                           */
/*  Add day expression e whose time group is tg to dr.                       */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsResourceAddDayExpr(KHE_DRS_RESOURCE dr,
  KHE_TIME_GROUP tg, KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_EXPR e)
{
  int i;  KHE_DRS_RESOURCE_ON_DAY drd;
  HnAssert(KheTimeGroupTimeCount(tg) > 0,
    "KheDrsResourceAddDayExpr internal error 1");
  i = KheFrameTimeIndex(drs->days_frame, KheTimeGroupTime(tg, 0));
  drd = HaArray(dr->days, i);
  ** ***
  HnAssert(KheTimeGroupEqual(drd->time_group, tg),
    "KheDrsAddDayExpr internal error 2");
  *** **
  HaArrayAddLast(drd->external_today, e);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceOpenOnDay(KHE_DRS_RESOURCE dr, int day_index,         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Open dr on the day whose index is day_index.                             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsResourceOpenOnDay(KHE_DRS_RESOURCE dr, int day_index,
  int open_day_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i;  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_DRS_EXPR e;
  drd = HaArray(dr->days, day_index);
  HaArrayForEach(drd->external_today, e, i)
  {
    KheDrsExprO pen(e, drs);
    e->open_day_range = KheDrsDayRangeMake(open_day_index, open_day_index);
    ** KheDrsExprExternalOpen(e, day_index, drs); **
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceOpen(KHE_DRS_RESOURCE dr,                             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Open dr.  This involves unassigning all the tasks assigned dr which      */
/*  lie in the selected day ranges of drs, and gathering for opening the     */
/*  expressions that depend on what dr is doing on those days, including     */
/*  on days where dr remains assigned.                                       */
/*                                                                           */
/*****************************************************************************/
static void KheDrsTaskUnAssign(KHE_DRS_TASK dt, bool task);

static void KheDrsResourceOpen(KHE_DRS_RESOURCE dr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_DAY_RANGE ddr, open_day_range;  int i, j, k, open_day_index;
  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_DRS_TASK dt;  KHE_DRS_EXPR e;
  open_day_index = 0;
  HaArrayForEach(drs->selected_day_ranges, ddr, i)
    for( j = ddr.first;  j <= ddr.last;  j++ )
    {
      /* unassign any task assigned on drd, if it lies entirely in ddr */
      drd = HaArray(dr->days, j);
      if( drd->closed_asst != NULL )
      {
	dt = drd->closed_asst->encl_dt;
	if( KheDrsDayRangeSubset(dt->encl_dtc->day_range, ddr) )
	  KheDrsTaskUnAssign(dt, true);
      }

      /* gather for opening expressions depending on drd (unassigned or not) */
      open_day_range = KheDrsDayRangeMake(open_day_index, open_day_index);
      HaArrayForEach(drd->external_today, e, k)
      {
	e->open_day_range = open_day_range;
	KheDrsExprGatherForOpening(e, drs);
      }

      /* increase open_day_index */
      open_day_index++;
    }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceCloseOnDay(KHE_DRS_RESOURCE dr, int day_index)        */
/*                                                                           */
/*  Close dr on the day whose index is day_index, unless already closed.     */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsResourceCloseOnDay(KHE_DRS_RESOURCE dr, int day_index,
  KHE_TASK task, KHE_TIME t)
{
  int i;  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_DRS_EXPR e;
  drd = HaArray(dr->days, day_index);
  if( drd->open )
  {
    drd->open = false;
    HaArrayForEach(drd->external_today, e, i)
      KheDrsExprExternalClose(e, day_index, task, t, dr->resource);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceClose(KHE_DRS_RESOURCE dr)                            */
/*                                                                           */
/*  Close dr.                                                                */
/*                                                                           */
/*****************************************************************************/

/* *** nothing to do here
static void KheDrsResourceClose(KHE_DRS_RESOURCE dr)
{
  ** dr->open_resource_index = -1; **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceOnDayClose(KHE_DRS_RESOURCE_ON_DAY drd,               */
/*    KHE_DRS_TASK_ON_DAY dtd)                                               */
/*                                                                           */
/*  Close drd.                                                               */
/*                                                                           */
/*****************************************************************************/

/* *** no longer used
static void KheDrsResourceOnDayClose(KHE_DRS_RESOURCE_ON_DAY drd,
  KHE_DRS_TASK_ON_DAY dtd)
{
  ** ***
  int i ** , day_index **;  KHE_DRS_EXPR e;
  KHE_TASK task;  KHE_TIME time;  KHE_RESOURCE r;
  HnAssert(drd->open, "KheDrsResourceOnDayClose internal error");
  drd->open = false;
  *** **
  drd->closed_asst = dtd;
  ** day_index = drd->day->frame_index; **
  ** *** doing this separately now
  if( dtd != NULL )
    task = dtd->task, time = dtd->time;
  else
    task = NULL, time = NULL;
  r = drd->encl_dr->resource;
  HaArrayForEach(drd->external_today, e, i)
    KheDrsExprExternalClose(e, ** day_index, ** task, time, r*;
  *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE KheDrsResourceMake(KHE_RESOURCE r,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new resource object with these attributes.                        */
/*                                                                           */
/*****************************************************************************/
static void KheDrsAddResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_MONITOR m, KHE_DRS_RESOURCE dr);

static KHE_DRS_RESOURCE KheDrsResourceMake(KHE_RESOURCE r,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_RESOURCE res;  int i;  KHE_DRS_RESOURCE_ON_DAY drd;
  KHE_DRS_DAY day;  KHE_MONITOR m;

  /* get the resource object */
  /* ***
  if( HaArrayCount(drs->free_resource) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_resource);
    HaArrayClear(res->days);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->days, drs->arena);
  }
  *** */

  /* initialize res, including one resource on day object for each day */
  HaMake(res, drs->arena);
  HaArrayInit(res->days, drs->arena);
  res->resource = r;
  /* res->open_resource_index = -1;  signals not currently open for solving */
  HaArrayForEach(drs->all_days, day, i)
  {
    drd = KheDrsResourceOnDayMake(res, day, drs);
    HaArrayAddLast(res->days, drd);
  }

  /* add expressions to dr based on the resource's monitors */
  KheSolnResourceMonitorSort(drs->soln, r, &KheMonitorLargestLimitFirstCmp);
  for( i = 0;  i < KheSolnResourceMonitorCount(drs->soln, r);  i++ )
  {
    m = KheSolnResourceMonitor(drs->soln, r, i);
    KheDrsAddResourceMonitor(drs, m, res);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceFree(KHE_DRS_RESOURCE dr,                             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Reclaim the memory used by dr.                                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsResourceFree(KHE_DRS_RESOURCE dr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ** KHE_DRS_EXPR e; **  int i;  KHE_DRS_RESOURCE_ON_DAY drd;
  ** ***
  HaArrayForEach(dr->exprs, e, i)
    KheDrsExprFreeInternal(e, drs);
  *** **
  HaArrayForEach(dr->days, drd, i)
    KheDrsResourceOnDayFree(drd, drs);
  HaArrayAddLast(drs->free_resource, dr);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDay(KHE_DRS_RESOURCE dr,         */
/*    KHE_DRS_DAY day)                                                       */
/*                                                                           */
/*  Return the resource on day object representing dr on day.                */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDay(KHE_DRS_RESOURCE dr,
  KHE_DRS_DAY day)
{
  return HaArray(dr->days, day->frame_index);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDayAtTime(KHE_DRS_RESOURCE dr,   */
/*    KHE_TIME time, KHE_DYNAMIC_RESOURCE_SOLVER drs)                        */
/*                                                                           */
/*  Return the resource on day object whose resource is dr and whose day     */
/*  is time's day.                                                           */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_RESOURCE_ON_DAY KheDrsResourceOnDayAtTime(KHE_DRS_RESOURCE dr,
  KHE_TIME time, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int day_index;
  day_index = KheFrameTimeIndex(drs->days_frame, time);
  return HaArray(dr->days, day_index);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceMakeAndAdd(KHE_RESOURCE r,                            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a drs resource object corresponding to r (assumed to have the       */
/*  correct type) and add it to drs.                                         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsResourceMakeAndAdd(KHE_RESOURCE r,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->all_resources, KheDrsResourceMake(r, drs));
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddSolveInterval(KHE_DRS_RESOURCE dr,                 */
/*    KHE_DRS_DAY_RANGE si, KHE_DYNAMIC_RESOURCE_SOLVER drs)                 */
/*                                                                           */
/*  Add si to the solve intervals of dr.  Also make sure that dr is on       */
/*  the selected list, if it isn't already.                                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsResourceAddSolveInterval(KHE_DRS_RESOURCE dr,
  KHE_DRS_DAY_RANGE si, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ** add dr to the selected list if it isn't there already **
  if( HaArrayCount(dr->selected_day_ranges) == 0 )
    HaArrayAddLast(drs->open_resources, dr);

  ** add si to the day intervals **
  HnAssert(si.first_day_index <= si.last,
    "KheDrsResourceAddSolveInterval internal error:  empty interval");
  KheDrsDayRangeMerge(&dr->selected_day_ranges, si);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceInitFixed(KHE_DRS_RESOURCE dr, KHE_DRS_DAY_RANGE si,  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Initialize fixed interval si.                                            */
/*                                                                           */
/*  Throughout this interval, the resource must preserve its existing        */
/*  timetable.  This is indicated by placing a non-NULL drs task into        */
/*  each position.  If the resource is already assigned a task, that         */
/*  is the task placed there.  If the resource is free, a free day task      */
/*  for the day is used.                                                     */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete (I think)
static void KheDrsResourceInitFixed(KHE_DRS_RESOURCE dr, KHE_DRS_DAY_RANGE si,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int di;  KHE_DRS_TASK dt;  KHE_DRS_DAY day;  KHE_DRS_RESOURCE_ON_DAY drd;
  for( di = si.first_day_index;  di <= si.last;  di++ )
  {
    drd = HaArray(dr->days, di);
    dt = drd->assigned_task;
    if( dt == NULL )
    {
      day = HaArray(drs->days, di);
      dt = day->free_task;
    }
    drd->solve_task = dt;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceInitFree(KHE_DRS_RESOURCE dr, KHE_DRS_DAY_RANGE si,   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Initialize free interval si.                                             */
/*                                                                           */
/*  Throughout this interval, the resource is free to take on a different    */
/*  timetable.  This is indicated by placing a NULL drs task into each       */
/*  position.  However, if the resource is assigned a task on some day       */
/*  and that task extends beyond si, the resource must preserve that task,   */
/*  and that is indicated by placing it into solve_tasks.                    */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete
static void KheDrsTaskAddToSolveTasks(KHE_DRS_TASK dt);

static void KheDrsResourceInitFree(KHE_DRS_RESOURCE dr, KHE_DRS_DAY_RANGE si,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int di;  KHE_DRS_TASK dt;  KHE_DRS_RESOURCE_ON_DAY drd;
  for( di = si.first_day_index;  di <= si.last;  di++ )
  {
    drd = HaArray(dr->days, di);
    dt = drd->assigned_task;
    if( dt != NULL && KheDrsDayRangeSubset(dt->day_interval, si) )
    {
      ** this task is assigned but it can be reassigned **
      KheDrsTaskAddToSolveTasks(dt);
      dt = NULL;
    }
    drd->solve_task = dt;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceInitForSolving(KHE_DRS_RESOURCE dr,                   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Assuming that dr is one of the resources selected for solving,           */
/*  initialize it for solving.                                               */
/*                                                                           */
/*  There will always be at least one free interval, although as it happens  */
/*  this code works correctly even if there isn't:  in that case, it makes   */
/*  all days fixed.                                                          */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete
static void KheDrsResourceInitForSolving(KHE_DRS_RESOURCE dr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int prev_index, i;  KHE_DRS_DAY_RANGE free_si, fixed_si;

  ** initialize the solve_tasks array for solving **
  prev_index = 0;
  HaArrayForEach(dr->selected_day_ranges, free_si, i)
  {
    fixed_si = KheDrsDayRangeMake(prev_index, free_si.first_day_index - 1);
    KheDrsResourceInitFixed(dr, fixed_si, drs);
    KheDrsResourceInitFree(dr, free_si, drs);
    prev_index = free_si.last + 1;
  }
  fixed_si = KheDrsDayRangeMake(prev_index, HaArrayCount(dr->days));
  KheDrsResourceInitFixed(dr, fixed_si, drs);

  ** add expressions to days **
  ** ***
  HaArrayForEach(dr->exprs, e, i)
    KheDrsExprAddExpressionsToDays(e, dr);
  *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceEndSolving(KHE_DRS_RESOURCE dr)                       */
/*                                                                           */
/*  End the selection of dr for solving.                                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsResourceEndSolving(KHE_DRS_RESOURCE dr)
{
  ** clear out the solve intervals **
  ** HaArrayClear(dr->selected_day_ranges); **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceAddTask(KHE_DRS_RESOURCE dr, KHE_DRS_TASK dt)         */
/*                                                                           */
/*  Assuming that dt is bound to dr's resource, add dt to dr.                */
/*                                                                           */
/*****************************************************************************/
/* static int KheDrsTaskFirstDayIndex(KHE_DRS_TASK dt); */
/* static int KheDrsTaskLastDayIndex(KHE_DRS_TASK dt); */

/* *** obsolete
static void KheDrsResourceAddTask(KHE_DRS_RESOURCE dr, KHE_DRS_TASK dt)
{
  int first_day_index, last_day_index, i;  KHE_DRS_RESOURCE_ON_DAY drd;
  first_day_index = KheDrsTaskFirstDayIndex(dt);
  last_day_index = KheDrsTaskLastDayIndex(dt);
  for( i = first_day_index;  i <= last_day_index;  i++ )
  {
    drd = HaArray(dr->days, i);
    drd->assigned_task = dt;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsResourceDebug(KHE_DRS_RESOURCE dr, int verbosity,             */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of dr onto fp with the given verbosity and indent.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsResourceDebug(KHE_DRS_RESOURCE dr, int verbosity,
  int indent, FILE *fp)
{
  KHE_DRS_RESOURCE_ON_DAY drd;  int i;
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "[ %s", KheResourceId(dr->resource));
  /* KheDayRangesDebug(&dr->selected_day_ranges, verbosity, -1, fp); */
  if( verbosity > 1 && indent >= 0 )
  {
    fprintf(fp, "\n");
    HaArrayForEach(dr->days, drd, i)
      KheDrsResourceOnDayDebug(drd, verbosity, indent + 2, fp);
  }
  fprintf(fp, "]");
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_TASK_ON_DAY"                                          */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK_ON_DAY KheDrsTaskOnDayMake(KHE_DRS_TASK encl_dt,            */
/*    KHE_DRS_DAY day, KHE_TASK task, KHE_TIME time,                         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new task on day object with these attributes.                     */
/*  This function assumes that the task is not assigned anything.            */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_TASK_ON_DAY KheDrsTaskOnDayMake(KHE_DRS_TASK encl_dt,
  KHE_DRS_DAY day, KHE_TASK task, KHE_TIME time,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY res;
  /* ***
  if( HaArrayCount(drs->free_task_on_day) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_task_on_day);
    HaArrayClear(res->external_today);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->external_today, drs->arena);
  }
  *** */
  HaMake(res, drs->arena);
  HaArrayInit(res->external_today, drs->arena);
  res->encl_dt = encl_dt;
  res->day = day;
  res->task = task;
  res->time = time;
  res->closed_asst = NULL;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskOnDayFree(KHE_DRS_TASK_ON_DAY td,                         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free td.                                                                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTaskOnDayFree(KHE_DRS_TASK_ON_DAY td,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR e;  int i;
  HaArrayForEach(td->external_today, e, i)
    KheDrsExprFreeInternal(e, drs);
  HaArrayAddLast(drs->free_task_on_day, td);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsTaskOnDayCmp(const void *t1, const void *t2)                   */
/*                                                                           */
/*  Comparison function for sorting an array of task on day objects by       */
/*  increasing day index.                                                    */
/*                                                                           */
/*****************************************************************************/

static int KheDrsTaskOnDayCmp(const void *t1, const void *t2)
{
  KHE_DRS_TASK_ON_DAY dtd1, dtd2;
  dtd1 = * (KHE_DRS_TASK_ON_DAY *) t1;
  dtd2 = * (KHE_DRS_TASK_ON_DAY *) t2;
  return dtd1->day->frame_index - dtd2->day->frame_index;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsOptionalResourceGroupEqual(KHE_RESOURCE_GROUP rg1,            */
/*    KHE_RESOURCE_GROUP rg2)                                                */
/*                                                                           */
/*  Return true if rg1 and rg2 are both NULL, or both non-NULL and equal.    */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsOptionalResourceGroupEqual(KHE_RESOURCE_GROUP rg1,
  KHE_RESOURCE_GROUP rg2)
{
  if( rg1 == NULL )
    return rg2 == NULL;
  else
    return rg2 != NULL && KheResourceGroupEqual(rg1, rg2);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTaskOnDayHasAssignedTaskExpr(KHE_DRS_TASK_ON_DAY dtd,         */
/*    KHE_RESOURCE_GROUP rg, KHE_DRS_EXPR_ASSIGNED_TASK *res)                */
/*                                                                           */
/*  If dtd has an assigned task expression with resource group attribute     */
/*  rg, return true with *res set to that expression; else return false.     */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsTaskOnDayHasAssignedTaskExpr(KHE_DRS_TASK_ON_DAY dtd,
  KHE_RESOURCE_GROUP rg, KHE_DRS_EXPR_ASSIGNED_TASK *res)
{
  KHE_DRS_EXPR e;  int i;  KHE_DRS_EXPR_ASSIGNED_TASK eat;
  HaArrayForEach(dtd->external_today, e, i)
  {
    eat = (KHE_DRS_EXPR_ASSIGNED_TASK) e;
    if( KheDrsOptionalResourceGroupEqual(rg, eat->resource_group) )
      return *res = eat, true;
  }
  return *res = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskOnDayAddAssignedTaskExpr(KHE_DRS_TASK_ON_DAY dtd,         */
/*    KHE_DRS_EXPR_ASSIGNED_TASK eat)                                        */
/*                                                                           */
/*  Add eat to dtd.                                                          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskOnDayAddAssignedTaskExpr(KHE_DRS_TASK_ON_DAY dtd,
  KHE_DRS_EXPR_ASSIGNED_TASK eat)
{
  HaArrayAddLast(dtd->external_today, (KHE_DRS_EXPR) eat);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskOnDayDebug(KHE_DRS_TASK_ON_DAY dtd,                       */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of dtd onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskOnDayDebug(KHE_DRS_TASK_ON_DAY dtd,
  int verbosity, int indent, FILE *fp)
{
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  KheTaskDebug(dtd->task, 1, -1, fp);
  fprintf(fp, " (day_index %d)", dtd->day->frame_index);
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_TASK"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskAssign(KHE_DRS_TASK dt, KHE_DRS_RESOURCE dr, bool task)   */
/*                                                                           */
/*  Assign dr (which must be non-NULL) to dt, by setting the close_asst      */
/*  fields in dt, dt->days, and dr.  If task is true, also assign dt's KHE   */
/*  task to dr's KHE resource, aborting if that does not succeed.            */
/*                                                                           */
/*  The enclosing task class has a list of unassigned tasks; but do not      */
/*  remove dt from it, because we are being lazy:  the unassigned tasks      */
/*  will be sorted out later, by KheDrsTaskClassOrganizeOpenTasks.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskAssign(KHE_DRS_TASK dt, KHE_DRS_RESOURCE dr, bool task)
{
  KHE_DRS_TASK_ON_DAY dtd;  int i;  KHE_DRS_RESOURCE_ON_DAY drd;
  HnAssert(dt->closed_asst == NULL, "KheDrsTaskAssign internal error 1");
  HnAssert(dr != NULL, "KheDrsTaskAssign internal error 2");
  if( task && !KheTaskAssignResource(dt->task, dr->resource) )
    HnAbort("KheDrsTaskAssign internal error 3");
  dt->closed_asst = dr;
  HaArrayForEach(dt->days, dtd, i)
  {
    drd = KheDrsResourceOnDay(dr, dtd->day);
    HnAssert(dtd->closed_asst == NULL, "KheDrsTaskAssign internal error 4");
    HnAssert(drd->closed_asst == NULL, "KheDrsTaskAssign internal error 5");
    dtd->closed_asst = drd;
    drd->closed_asst = dtd;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskUnAssign(KHE_DRS_TASK dt, bool task)                      */
/*                                                                           */
/*  Unassign dt (which must be assigned some drs resource dr), by clearing   */
/*  the closed_asst fields in dt, dt->days, and dr.  If task is true, also   */
/*  unassign dt's KHE task, or do nothing if the task will not unassign.     */
/*                                                                           */
/*  The enclosing task class has a list of unassigned tasks; add dt to it.   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskUnAssign(KHE_DRS_TASK dt, bool task)
{
  KHE_DRS_TASK_ON_DAY dtd;  int i;
  KHE_DRS_RESOURCE dr;  KHE_DRS_RESOURCE_ON_DAY drd;
  dr = dt->closed_asst;
  HnAssert(dr != NULL, "KheDrsTaskUnAssign internal error 1");
  if( task )
  {
    HnAssert(KheTaskAsstResource(dt->task) == dr->resource,
      "KheDrsTaskUnAssign internal error 2");
    if( !KheTaskUnAssignResource(dt->task) )
      return;
  }
  HaArrayForEach(dt->days, dtd, i)
  {
    drd = KheDrsResourceOnDay(dr, dtd->day);
    HnAssert(dtd->closed_asst == drd, "KheDrsTaskUnAssign internal error 3");
    HnAssert(drd->closed_asst == dtd, "KheDrsTaskUnAssign internal error 4");
    dtd->closed_asst = NULL;
    drd->closed_asst = NULL;
  }
  dt->closed_asst = NULL;
  HaArrayAddLast(dt->encl_dtc->open_tasks, dt);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheTimeSolveDayIndex(KHE_TIME t, int first_solve_index,              */
/*    int last_solve_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)                 */
/*                                                                           */
/*  Return the solve index of the day whose time group contains t.  It will  */
/*  be somewhere between first_solve_index and last_solve_index.             */
/*                                                                           */
/*****************************************************************************/

/* *** unused, not sure why
static int KheTimeSolveDayIndex(KHE_TIME t, int first_solve_index,
  int last_solve_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, pos;  KHE_DRS_DAY day;
  for( i = first_solve_index;  i <= last_solve_index;  i++ )
  {
    day = HaArray(drs->open_days, i);
    if( KheTimeGroupContains(day->time_group, t, &pos) )
      return i;
  }
  HnAbort("KheTimeSolveDayIndex internal error");
  return 0;  ** keep compiler happy **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheTaskAddDays(KHE_TASK task, KHE_DRS_TASK dt,                      */
/*    int first_solve_index, int last_solve_index,                           */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add the days covered by task and its descendants to dt.                  */
/*                                                                           */
/*****************************************************************************/

static void KheTaskAddDays(KHE_TASK task, KHE_DRS_TASK dt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_TASK child_task;  int i, durn, day_index;  KHE_DRS_DAY day;
  KHE_MEET meet;  KHE_TIME start_time, time;  KHE_DRS_TASK_ON_DAY dtd;

  /* do it for task */
  meet = KheTaskMeet(task);
  HnAssert(meet != NULL, "KheTaskAddDays internal error (no meet)");
  start_time = KheMeetAsstTime(meet);
  HnAssert(start_time != NULL,  "KheTaskAddDays internal error (no time)");
  durn = KheMeetDuration(meet);
  for( i = 0;  i < durn;  i++ )
  {
    time = KheTimeNeighbour(start_time, i);
    day_index = KheFrameTimeIndex(drs->days_frame, time);
    day = HaArray(drs->all_days, day_index);
    dtd = KheDrsTaskOnDayMake(dt, day, task, time, drs);
    HaArrayAddLast(dt->days, dtd);
  }

  /* do it for task's descendants */
  for( i = 0;  i < KheTaskAssignedToCount(task);  i++ )
  {
    child_task = KheTaskAssignedTo(task, i);
    KheTaskAddDays(child_task, dt, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskOrganizeDays(KHE_DRS_TASK dt)                             */
/*                                                                           */
/*  Assuming that dt->days are sorted by increasing day index, organize      */
/*  them so that there is one per day, by inserting NULL entries where       */
/*  there are gaps.  There can be no cases of two on one day, thanks to an   */
/*  earlier call to KheTaskClassNoOverlap.                                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskOrganizeDays(KHE_DRS_TASK dt)
{
  int init_len, reqd_len, i, final_pos;
  KHE_DRS_TASK_ON_DAY first_dtd, last_dtd, dtd;

  /* quit early if the number of days is fine as is */
  first_dtd = HaArrayFirst(dt->days);
  last_dtd = HaArrayLast(dt->days);
  reqd_len = last_dtd->day->frame_index - first_dtd->day->frame_index + 1;
  init_len = HaArrayCount(dt->days);
  if( reqd_len == init_len )
    return;

  /* add the required number of NULL entries, initially at the array's end */
  HaArrayFill(dt->days, reqd_len, NULL);

  /* move the entries to their final positions, working right to left */
  for( i = init_len - 1;  i >= 0;  i-- )
  {
    dtd = HaArray(dt->days, i);
    final_pos = dtd->day->frame_index - first_dtd->day->frame_index;
    if( final_pos != i )
    {
      HaArrayPut(dt->days, final_pos, dtd);
      HaArrayPut(dt->days, i, NULL);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK KheDrsTaskGet(KHE_DYNAMIC_RESOURCE_SOLVER drs)              */
/*                                                                           */
/*  Get a new drs task object.  Only the times_by_day and tasks_by day       */
/*  fields are initialized.                                                  */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_TASK KheDrsTaskGet(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK res;
  /* ***
  if( HaArrayCount(drs->free_task) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_task);
    HaArrayClear(res->days);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->days, drs->arena);
  }
  *** */
  HaMake(res, drs->arena);
  HaArrayInit(res->days, drs->arena);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK KheDrsTaskMake(KHE_DRS_TASK_CLASS encl_dtc,                 */
/*    int index_in_encl_dtc, KHE_TASK task, KHE_DYNAMIC_RESOURCE_SOLVER drs) */
/*                                                                           */
/*  Make a new drs task with these attributes, add it to drs->all_root_tasks */
/*  at the same index as it has in the solution, and return it.              */
/*                                                                           */
/*****************************************************************************/
static KHE_DRS_RESOURCE KheDrsResource(KHE_RESOURCE r,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);

static KHE_DRS_TASK KheDrsTaskMake(KHE_DRS_TASK_CLASS encl_dtc,
  int index_in_encl_dtc, KHE_TASK task, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK res;  int task_index;  KHE_RESOURCE r;

  /* get the basic object and initialize its fields */
  HnAssert(KheTaskIsProperRoot(task), "KheDrsTaskMake internal error");
  res = KheDrsTaskGet(drs);
  res->encl_dtc = encl_dtc;
  res->index_in_encl_dtc = index_in_encl_dtc;
  res->open = false;
  res->task = task;
  res->closed_asst = NULL;
  /* res->first_day_index = first_day_index; */
  /* res->last_day_index = last_day_index; */

  /* add, sort, and organize the task on day objects */
  KheTaskAddDays(task, res, drs);
  HaArraySort(res->days, &KheDrsTaskOnDayCmp);
  KheDrsTaskOrganizeDays(res);

  /* add res to drs->all_root_tasks */
  task_index = KheTaskSolnIndex(task);
  HaArrayFill(drs->all_root_tasks, task_index + 1, NULL);
  HnAssert(HaArray(drs->all_root_tasks, task_index) == NULL,
    "KheDrsTaskMake internal error");
  HaArrayPut(drs->all_root_tasks, task_index, res);

  /* if task is assigned, make that assignment in res */
  r = KheTaskAsstResource(task);
  if( r != NULL )
    KheDrsTaskAssign(res, KheDrsResource(r, drs), false);

  /* and also return res */
  return res;
}

/* ***
static KHE_DRS_TASK KheDrsTas kMake(KHE_TASK task,
   int first_solve_index, int last_solve_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK res;

  ** get the basic object and initialize its fields **
  HnAssert(KheTaskIsProperRoot(task), "KheDrsTaskMake internal error");
  res = KheDrsTaskGet(drs);
  res->task = task;
  ** res->orig_task_class = orig_tc; **
  ** res->first_solve_index = first_solve_index; **
  ** res->last_solve_index = last_solve_index; **
  ** res->adopted = false; **

  ** add, sort, and organize the task on day objects **
  KheTaskAddDays(task, res, first_solve_index, last_solve_index, drs);
  HaArraySort(res->days, &KheDrsTaskOnDayCmp);
  KheDrsTaskOrganizeDays(res);

  ** set task's back pointer to point to res **
  ** KheTaskSetBack(task, (void *) res); **
  return res;

  ** ***
  KHE_DRS_TASK res;  KHE_RESOURCE r;  int first_day_index, last_day_index;
  res = KheDrsTaskGet(drs);
  res->task = task;
  res->task_group = NULL;	** will be filled in later **
  res->daily_schedule = KheTaskFinderTaskDailySchedule(drs->task_finder, task);
  first_day_index = KheDailyScheduleFirstDayIndex(res->daily_schedule);
  last_day_index = KheDailyScheduleLastDayIndex(res->daily_schedule);
  res->day_interval = KheDrsDayRangeMake(first_day_index, last_day_index);
  res->cost_reduction = KheTaskAssignmentCostReduction(task);
  res->domain_count = KheResourceGroupResourceCount(KheTaskDomain(task));
  ** res->fixed = false; **	** placeholder value **
  r = KheTaskAsstResource(task);
  res->already_assigned = (r != NULL);
  return res;
  *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskFree(KHE_DRS_TASK dt, KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Free dt.                                                                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTaskFree(KHE_DRS_TASK dt, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;  int i;
  ** KheDailyScheduleDelete(dt->daily_schedule); **
  HaArrayForEach(dt->days, dtd, i)
    KheDrsTaskOnDayFree(dtd, drs);
  HaArrayAddLast(drs->free_task, dt);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsTaskIndexInEnclCmp(const void *t1, const void *t2)             */
/*                                                                           */
/*  Comparison function for sorting an array of drs tasks into increasing    */
/*  order of their index in their enclosing task class.                      */
/*                                                                           */
/*  In addition, it moves any assigned tasks to the end.                     */
/*                                                                           */
/*****************************************************************************/

static int KheDrsTaskIndexInEnclCmp(const void *t1, const void *t2)
{
  KHE_DRS_TASK dt1 = * (KHE_DRS_TASK *) t1;
  KHE_DRS_TASK dt2 = * (KHE_DRS_TASK *) t2;
  if( dt1->closed_asst != NULL )
  {
    if( dt2->closed_asst != NULL )
    {
      /* dt1 is assigned, dt2 is assigned */
      return 0;
    }
    else
    {
      /* dt1 is assigned, dt2 is not assigned */
      return 1;
    }
  }
  else
  {
    if( dt2->closed_asst != NULL )
    {
      /* dt1 is not assigned, dt2 is assigned */
      return -1;
    }
    else
    {
      /* dt1 is not assigned, dt2 is not assigned */
      return dt1->index_in_encl_dtc - dt2->index_in_encl_dtc;
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskOpen(KHE_DRS_TASK dt, KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Open task for solving.  We'll have checked previously that its days      */
/*  all lie in the current open interval, and it must currently be closed    */
/*  but unassigned; so opening dt basically just gathers for opening all     */
/*  expressions that depend on dt's assignment.                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskOpen(KHE_DRS_TASK dt, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_ON_DAY dtd;  KHE_DRS_EXPR e;  int i, j, di;
  KHE_DRS_DAY_RANGE open_day_range;

  /* open dt */
  HnAssert(!dt->open, "KheDrsTaskOpen internal error 1");
  HnAssert(dt->closed_asst == NULL, "KheDrsTaskOpen internal error 2");
  dt->open = true;

  /* gather external expressions for opening */
  HaArrayForEach(dt->days, dtd, i)
  {
    di = dtd->day->open_day_index;
    open_day_range = KheDrsDayRangeMake(di, di);
    HaArrayForEach(dtd->external_today, e, j)
    {
      e->open_day_range = open_day_range;
      KheDrsExprGatherForOpening(e, drs);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskOnDayClose(KHE_DRS_TASK_ON_DAY dtd,                       */
/*    KHE_DRS_RESOURCE_ON_DAY drd)                                           */
/*                                                                           */
/*  Close dtd.  It is assigned drd (possibly NULL).  Do not touch drd.       */
/*                                                                           */
/*****************************************************************************/

/* *** no longer used
static void KheDrsTaskOnDayClose(KHE_DRS_TASK_ON_DAY dtd,
  KHE_DRS_RESOURCE_ON_DAY drd)
{
  ** KHE_RESOURCE r;  KHE_DRS_EXPR e;  int i, day_index **;
  HnAssert(dtd->closed_asst == NULL, "KheDrsTaskOnDayClose internal error");
  dtd->closed_asst = drd;
  ** day_index = dtd->day->frame_index; **
  ** *** doing this separately now
  r = (drd != NULL ? drd->encl_dr->resource : NULL);
  HaArrayForEach(dtd->external_today, e, i)
    KheDrsExprExternalClose(e, ** day_index, ** dtd->task, dtd->time, r);
  *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClose(KHE_DRS_TASK dt, KHE_DRS_RESOURCE dr)               */
/*                                                                           */
/*  If dt is open, close it, with a call to KheDrsTaskAssign if dr != NULL.  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClose(KHE_DRS_TASK dt, KHE_DRS_RESOURCE dr)
{
  if( dt->open )
  {
    dt->open = false;
    if( dr != NULL )
      KheDrsTaskAssign(dt, dr, true);
  }
}

/* *** obsolete version
static void KheDrsTaskClose(KHE_DRS_TASK dt, KHE_DRS_RESOURCE dr)
{
  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_DRS_TASK_ON_DAY dtd;  int i;
  if( dt->open )
  {
    dt->open = false;
    dt->closed_asst = dr;
    if( dr != NULL )
    {
      ** assign dr to dt as well as closing dt **
      if( !KheTaskAssignResource(dt->task, dr->resource) )
	HnAbort("KheDrsTaskClose internal error 2");
      HaArrayForEach(dt->days, dtd, i)
      {
	drd = KheDrsResourceOnDay(dr, dtd->day);
        KheDrsTaskOnDayClose(dtd, drd);
      }
    }
    else
    {
      ** just close dt, there is no dr **
      HaArrayForEach(dt->days, dtd, i)
        KheDrsTaskOnDayClose(dtd, NULL);
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK_ON_DAY KheDrsTaskOnDay(KHE_DRS_TASK dt, KHE_DRS_DAY day)    */
/*                                                                           */
/*  Return the task on day object for dt on day.                             */
/*                                                                           */
/*****************************************************************************/

/* ***
KHE_DRS_TASK_ON_DAY KheDrsTaskOnDay(KHE_DRS_TASK dt, KHE_DRS_DAY day)
{
  ** *** needs revision stil l to do
  return HaArray(dt->days, day->solve_index);
  *** **
  return NULL;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsTaskFirstDayIndex(KHE_DRS_TASK dt)                             */
/*                                                                           */
/*  Return the index of the first day when dt is running.                    */
/*                                                                           */
/*****************************************************************************/

/* *** no longer used
static int KheDrsTaskFirstDayIndex(KHE_DRS_TASK dt)
{
  ** *** replacement for this st ill to do
  return KheDailyScheduleFirstDayIndex(dt->daily_schedule);
  *** **
  return -1;  ** placeholder **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsTaskLastDayIndex(KHE_DRS_TASK dt)                              */
/*                                                                           */
/*  Return the index of the last day when dt is running.                     */
/*                                                                           */
/*****************************************************************************/

/* *** no longer used
static int KheDrsTaskLastDayIndex(KHE_DRS_TASK dt)
{
  ** *** replacement for this sti ll to do
  return KheDailyScheduleLastDayIndex(dt->daily_schedule);
  *** **
  return -1;  ** placeholder **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_TIME KheDrsTaskFirstTime(KHE_DRS_TASK dt)                            */
/*                                                                           */
/*  Return the first time of dt.                                             */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_TIME KheDrsTaskFirstTime(KHE_DRS_TASK dt)
{
  int index;
  index = KheDailyScheduleFirstDayIndex(dt->daily_schedule);
  return KheDailyScheduleTime(dt->daily_schedule, index);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_TIME KheDrsTaskLastTime(KHE_DRS_TASK dt)                             */
/*                                                                           */
/*  Return the last time of dt.                                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_TIME KheDrsTaskLastTime(KHE_DRS_TASK dt)
{
  int index;
  index = KheDailyScheduleLastDayIndex(dt->daily_schedule);
  return KheDailyScheduleTime(dt->daily_schedule, index);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTaskIsBound(KHE_DRS_TASK dt, KHE_RESOURCE *r)                 */
/*                                                                           */
/*  If dt is bound, return true and set *r to the associated resource.       */
/*  Otherwise return false and set *r to NULL.                               */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDrsTaskIsBound(KHE_DRS_TASK dt, KHE_RESOURCE *r)
{
  ** return false if dt represents free time **
  if( dt->task == NULL )
    return false;

  ** return true if assigned **
  *r = KheTaskAsstResource(dt->task);
  if( *r != NULL )
    return true;

  ** return true if preassigned **
  return KheTaskIsPreassigned(dt->task, r);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_RESOURCE KheDrsResource(KHE_RESOURCE r,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Assuming that r has the correct resource type, return its corresponding  */
/*  drs resource.                                                            */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_RESOURCE KheDrsResource(KHE_RESOURCE r,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  return HaArray(drs->all_resources, KheResourceResourceTypeIndex(r));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskMakeAndAdd(KHE_TASK task,                                 */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new drs task object for task, which is a proper root task of the  */
/*  wanted type, and add it to the solver generally.                         */
/*                                                                           */
/*  Implementation note.  The drs task is added to one task group on its     */
/*  first day, and if assigned (or bound generally) it is added to the drs   */
/*  resource it is assigned or bound to.                                     */
/*                                                                           */
/*****************************************************************************/
/* ***
static KHE_DRS_TASK_GROUP KheDrsDayAddTask(KHE_DRS_DAY day, KHE_DRS_TASK dt);

static void KheDrsTaskMakeAndAdd(KHE_TASK task,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK dt;  KHE_RESOURCE r;  KHE_DRS_TASK_GROUP dtg;  KHE_DRS_DAY day;
  int first_day_index;

  ** make the drs task object **
  dt = KheDrsTas kMake(task, drs);

  ** add it to the day object of the first day that it is running **
  first_day_index = KheDailyScheduleFirstDayIndex(dt->daily_schedule);
  day = HaArray(drs->days, first_day_index);
  dtg = KheDrsDayAddTask(day, dt);

  ** record which task group it got added to **
  dt->task_group = dtg;

  ** if task is assigned or preassigned, add it to the associated resource **
  if( KheDrsTaskIsBound(dt, &r) )
    KheDrsResourceAddTask(KheDrsResource(r, drs), dt);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK KheDrsTaskMakeFree(int day_index,                           */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new drs task which represents doing nothing on the day with       */
/*  this day_index.                                                          */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_TASK KheDrsTaskMakeFree(int day_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK res;
  res = KheDrsTaskGet(drs);
  res->task = NULL;
  res->daily_schedule = KheTaskFinderNullDailySchedule(drs->task_finder,
    day_index, day_index);
  res->cost_reduction = 0;
  res->domain_count = 0;
  res->already_assigned = false;
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskAddToSolveTasks(KHE_DRS_TASK dt)                          */
/*                                                                           */
/*  Make dt a solve task, by adding it to its task group's solve_tasks list. */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTaskAddToSolveTasks(KHE_DRS_TASK dt)
{
  ** sorting by decreasing cost sti ll to do **
  HaArrayAddLast(dt->task_group->solve_tasks, dt);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskMakeUnAvail(KHE_DRS_TASK dt)                              */
/*                                                                           */
/*  Make dt unavailable.                                                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTaskMakeUnAvail(KHE_DRS_TASK dt)
{
  int pos;
  if( HaArrayContains(dt->task_group->solve_tasks, dt, &pos) )
    HaArrayDeleteAndShift(dt->task_group->solve_tasks, pos);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTaskBetterThan(KHE_DRS_TASK dt1, KHE_DRS_TASK dt2)            */
/*                                                                           */
/*  Return true if dt1 is a better choice than dt2, because it has a         */
/*  larger cost reduction, or equal cost reduction and a smaller domain.     */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDrsTaskBetterThan(KHE_DRS_TASK dt1, KHE_DRS_TASK dt2)
{
  return dt1->cost_reduction > dt2->cost_reduction ||
    (dt1->cost_reduction == dt2->cost_reduction &&
     dt1->domain_count < dt2->domain_count);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskDebug(KHE_DRS_TASK dt, int verbosity,                     */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of dt onto fp with the given verbosity and indent.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskDebug(KHE_DRS_TASK dt, int verbosity,
  int indent, FILE *fp)
{
  /* ***
  KHE_TIME first_time, last_time;  int first_day_index, last_day_index;
  KHE_TASK_FINDER tf;  KHE_TIME_GROUP tg1, tg2;  KHE_FRAME frame;
  *** */
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  /* ***
  if( dt->task != NULL )
  {
    first_time = KheDrsTaskFirstTime(dt);
    last_time = KheDrsTaskLastTime(dt);
    KheTaskDebug(dt->task, verbosity, -1, fp);
    if( first_time != last_time )
      fprintf(fp, " (%s-%s", KheTimeId(first_time), KheTimeId(last_time));
    else
      fprintf(fp, " (%s", KheTimeId(first_time));
    fprintf(fp, ", cost %.5f, domain %d)",
      KheCostShow(dt->cost_reduction), dt->domain_count);
  }
  else
  {
    tf = KheDailyScheduleTaskFinder(dt->daily_schedule);
    frame = KheTaskFinderFrame(tf);
    first_day_index = KheDailyScheduleFirstDayIndex(dt->daily_schedule);
    last_day_index = KheDailyScheduleLastDayIndex(dt->daily_schedule);
    tg1 = KheFrameTimeGroup(frame, first_day_index);
    tg2 = KheFrameTimeGroup(frame, last_day_index);
    if( tg1 != tg2 )
      fprintf(fp, "Free (%s-%s)", KheTimeGroupId(tg1), KheTimeGroupId(tg2));
    else
      fprintf(fp, "Free (%s)", KheTimeGroupId(tg1));
  }
  *** */
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_TASK_CLASS"                                           */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheTaskIsOpen(KHE_TASK task, KHE_DYNAMIC_RESOURCE_SOLVER drs)       */
/*                                                                           */
/*  Return true if task is open to the current solve.  We already know that  */
/*  its domain and times are OK, so it's just a matter of checking that it   */
/*  is either unassigned, or assigned one of drs's solve resources.          */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheTaskIsOpen(KHE_TASK task, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_RESOURCE r;
  r = KheTaskAsstResource(task);
  return r==NULL ||
    KheResourceSetContainsResource(drs->selected_resource_set, r);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheTaskAssignedSelected(KHE_TASK task,                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return true if task is assigned one of drs's selected resources.         */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheTaskAssignedSelected(KHE_TASK task,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_RESOURCE r;
  r = KheTaskAsstResource(task);
  return r!=NULL &&
    KheResourceSetContainsResource(drs->selected_resource_set, r);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassMake(KHE_TASK_CLASS tc,                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new drs task class based on tc and add it to its first day.       */
/*  Also make drs tasks corresponding to tc's tasks and add them to the      */
/*  drs class task and to drs->all_root_tasks.                               */
/*                                                                           */
/*  The following comment is obsolete; it relates to selecting tasks         */
/*  for opening during a solve, whereas here we select all tasks.            */
/*                                                                           */
/*  Here first_solve_index is the index in the solver (not in the days       */
/*  frame) of the first day that the tasks of tc are running, and similarly  */
/*  last_solve_index is the index in the solve of the last day that the      */
/*  tasks of tc are running.  Every day that the tasks of tc are running is  */
/*  a solve day, as we know thanks to an earlier call to KheTaskClassIsOpen. */
/*                                                                           */
/*  Which tasks are added to the result?  We have to add all tasks which     */
/*  are assigned selected resources, to allow them to be unassigned.  We     */
/*  also have to add at least as many tasks as there are selected resources  */
/*  (if there are that many in tc), to allow for the possibility of every    */
/*  selected resource being assigned a task of this class simultaneously.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassMake(KHE_TASK_CLASS tc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_CLASS res;  int i /* , m */;  KHE_TASK task;  KHE_DRS_TASK dt;
  KHE_DRS_DAY day;

  /* get or make a drs task class object */
  /* ***
  if( HaArrayCount(drs->free_task_class) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_task_class);
    HaArrayClear(res->all_tasks);
    HaArrayClear(res->open_tasks);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->open_tasks, drs->arena);
  }
  *** */
  HaMake(res, drs->arena);
  HaArrayInit(res->open_tasks, drs->arena);
  res->orig_task_class = tc;
  res->day_range = KheDrsDayRangeMake(KheTaskClassFirstDayIndex(tc),
    KheTaskClassLastDayIndex(tc));
  /* ***
  res->first_day_index = KheTaskClassFirstDayIndex(tc);
  res->last_day_index = KheTaskClassLastDayIndex(tc);
  *** */
  res->tmp_used = 0;

  /* make drs tasks for its khe tasks and add them */
  for( i = 0;  i < KheTaskClassTaskCount(tc);  i++ )
  {
    task = KheTaskClassTask(tc, i);
    dt = KheDrsTaskMake(res, i, task, drs);
    HaArrayAddLast(res->all_tasks, dt);
    if( dt->closed_asst == NULL )
      HaArrayAddLast(res->open_tasks, dt);
  }

  /* make res one of the task classes of its first day */
  day = HaArray(drs->all_days, res->day_range.first);
  HaArrayAddLast(day->task_classes, res);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassFree(KHE_DRS_TASK_CLASS drc,                         */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free drc by adding it to drs's free list.                                */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTaskClassFree(KHE_DRS_TASK_CLASS drc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK dt;  int i;
  HaArrayForEach(drc->all_tasks, dt, i)
    KheDrsTaskFree(dt, drs);
  HaArrayAddLast(drs->free_task_class, drc);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassOrganizeOpenTasks(KHE_DRS_TASK_CLASS dtc)            */
/*                                                                           */
/*  Organize the open tasks of dtc for solving.                              */
/*                                                                           */
/*  When this function is called, open_tasks may be out of order, and,       */
/*  because open_tasks is handled lazily, some of them may in fact be        */
/*  assigned, so not open.  This function fixes both of those problems.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassOrganizeOpenTasks(KHE_DRS_TASK_CLASS dtc)
{
  KHE_DRS_TASK dt;

  /* sort open_tasks into order with assigned tasks at the end */
  HaArraySortUnique(dtc->open_tasks, &KheDrsTaskIndexInEnclCmp);

  /* remove assigned tasks from the end */
  while( HaArrayCount(dtc->open_tasks) > 0 )
  {
    dt = HaArrayLast(dtc->open_tasks);
    if( dt->closed_asst == NULL )
      break;
    HaArrayDeleteLast(dtc->open_tasks);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassOpen(KHE_DRS_TASK_CLASS dtc, KHE_DRS_DAY_RANGE ddr,  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  If dtc is suitable (if its day range lies within ddr, and its resource   */
/*  domain includes at least one open resource), then open it.  Opening dtc  */
/*  basically means organizing its unassigned tasks into decreasing          */
/*  desirability order and opening them.                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassOpen(KHE_DRS_TASK_CLASS dtc, KHE_DRS_DAY_RANGE ddr,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK dt;  int i;
  if( KheDrsDayRangeSubset(dtc->day_range, ddr) &&
      !KheResourceSetDisjointGroup(drs->selected_resource_set,
	KheTaskClassDomain(dtc->orig_task_class)) )
  /* ***
    dtc->first_day_index >= ddr.first_day_index &&
      dtc->last_day_index <= ddr.last_day_index &&
  *** */
  {
    /* dtc can open; organize and open open_tasks; say none are used */
    KheDrsTaskClassOrganizeOpenTasks(dtc);
    HaArrayForEach(dtc->open_tasks, dt, i)
      KheDrsTaskOpen(dt, drs);
    dtc->tmp_used = 0;
  }
  else
  {
    /* dtc can't open; set dtc->tmp_used to make that clear */
    dtc->tmp_used = INT_MAX;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassClose(KHE_DRS_TASK_CLASS dtc,                        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Close dtc.  Some of its tasks may be already closed; that's OK.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassClose(KHE_DRS_TASK_CLASS dtc)
{
  KHE_DRS_TASK dt;  int i;
  HaArrayForEach(dtc->open_tasks, dt, i)
    KheDrsTaskClose(dt, NULL);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTaskGroupAccepts(KHE_DRS_TASK_GROUP sg, KHE_DRS_TASK dt)      */
/*                                                                           */
/*  Return true if sg accepts dt.  The start times, end times, and domains   */
/*  must be equal.                                                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDrsTaskGroupAccepts(KHE_DRS_TASK_GROUP sg, KHE_DRS_TASK dt)
{
  KHE_DRS_TASK sg_dt;  int di1, di2;
  KHE_DAILY_SCHEDULE sched1, sched2;

  ** get the daily schedules of the two tasks, and their first and last days **
  sg_dt = HaArrayFirst(sg->all_tasks);
  sched1 = sg_dt->daily_schedule;
  sched2 = dt->daily_schedule;

  ** check the first times **
  di1 = KheDailyScheduleFirstDayIndex(sched1);
  di2 = KheDailyScheduleFirstDayIndex(sched2);
  if( KheDailyScheduleTime(sched1, di1) != KheDailyScheduleTime(sched2, di2) )
    return false;

  ** check the last times **
  di1 = KheDailyScheduleLastDayIndex(sched1);
  di2 = KheDailyScheduleLastDayIndex(sched2);
  if( KheDailyScheduleTime(sched1, di1) != KheDailyScheduleTime(sched2, di2) )
    return false;

  ** check the domains **
  if( !KheResourceGroupEqual(KheTaskDomain(dt->task), 
	KheTaskDomain(sg_dt->task)) )
      return false;

  ** all good **
  return true;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskGroupAddTask(KHE_DRS_TASK_GROUP sg, KHE_DRS_TASK dt)      */
/*                                                                           */
/*  Add dt to sg, largest cost reduction first.                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTaskGroupAddTask(KHE_DRS_TASK_GROUP sg, KHE_DRS_TASK dt)
{
  KHE_DRS_TASK dt2;  int i;
  HaArrayAddLast(sg->all_tasks, dt);
  for( i = HaArrayCount(sg->all_tasks) - 2;  i >= 0;  i-- )
  {
    dt2 = HaArray(sg->all_tasks, i);
    if( dt2->cost_reduction <= dt->cost_reduction )
      HaArrayPut(sg->all_tasks, i + 1, dt2);
  }
  HaArrayPut(sg->all_tasks, i + 1, dt);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskGroupClearAvail(KHE_DRS_TASK_GROUP dtg)                   */
/*                                                                           */
/*  Clear the available tasks of dtg.                                        */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTaskGroupClearAvail(KHE_DRS_TASK_GROUP dtg)
{
  HaArrayClear(dtg->solve_tasks);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskGroupMakeUnboundTasksAvail(KHE_DRS_TASK_GROUP dtg)        */
/*                                                                           */
/*  Make the unbound tasks of dtg available.                                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTaskGroupMakeUnboundTasksAvail(KHE_DRS_TASK_GROUP dtg)
{
  KHE_DRS_TASK dt;  int i;  KHE_RESOURCE r;
  HaArrayForEach(dtg->all_tasks, dt, i)
    if( !KheDrsTaskIsBound(dt, &r) )
      HaArrayAddLast(dtg->solve_tasks, dt);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassDebug(KHE_DRS_TASK_CLASS dtc, int verbosity,         */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of dtc onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassDebug(KHE_DRS_TASK_CLASS dtc, int verbosity,
  int indent, FILE *fp)
{
  /* ***
  KHE_DRS_TASK dt;  KHE_DAILY_SCHEDULE sched;  KHE_TIME ft, lt;
  KHE_RESOURCE_GROUP rg;
  dt = HaArrayFirst(sg->all_tasks);
  sched = dt->daily_schedule;
  ft = KheDailyScheduleTime(sched, KheDailyScheduleFirstDayIndex(sched));
  lt = KheDailyScheduleTime(sched, KheDailyScheduleLastDayIndex(sched));
  rg = KheTaskDomain(dt->task);
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "[ %d tasks: ", HaArrayCount(sg->all_tasks));
  if( ft != lt )
    fprintf(fp, "%s-%s", KheTimeId(ft), KheTimeId(lt));
  else
    fprintf(fp, "%s", KheTimeId(ft));
  fprintf(fp, " %s ]", KheResourceGroupId(rg));
  if( indent >= 0 )
    fprintf(fp, "\n");
  *** */
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_MONITOR_CONVERTER"                                    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_MONITOR_CONVERTER KheDrsMonitorConverterMake(KHE_MONITOR m,      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new monitor converter object with these attributes.               */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_MONITOR_CONVERTER KheDrsMonitorConverterMake(KHE_MONITOR m,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_MONITOR_CONVERTER res;
  if( HaArrayCount(drs->free_monitor_converter) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_monitor_converter);
    HaArrayClear(res->task_on_days);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->task_on_days, drs->arena);
  }
  res->monitor = m;
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsMonitorConverterFree(KHE_DRS_MONITOR_CONVERTER dmc,           */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free dmc.                                                                */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsMonitorConverterFree(KHE_DRS_MONITOR_CONVERTER dmc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_monitor_converter, dmc);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsMonitorConverterAddTaskOnDay(KHE_DRS_MONITOR_CONVERTER dmc,   */
/*    KHE_DRS_TASK_ON_DAY dtd)                                               */
/*                                                                           */
/*  Add dtd to dmc.                                                          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsMonitorConverterAddTaskOnDay(KHE_DRS_MONITOR_CONVERTER dmc,
  KHE_DRS_TASK_ON_DAY dtd)
{
  HaArrayAddLast(dmc->task_on_days, dtd);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheAssignResourceMonitorTotalMonitoredDuration(                      */
/*    KHE_ASSIGN_RESOURCE_MONITOR arm)                                       */
/*                                                                           */
/*  Return the total duration of the tasks monitored by arm.                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static int KheAssignResourceMonitorTotalMonitoredDuration(
  KHE_ASSIGN_RESOURCE_MONITOR arm)
{
  KHE_EVENT_RESOURCE er;
  er = KheAssignResourceMonitorEventResource(arm);
  return KheEventDuration(KheEventResourceEvent(er));
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsMonitorConverterMakeAssignResourceExpr(                       */
/*    KHE_DRS_MONITOR_CONVERTER dmc, KHE_DYNAMIC_RESOURCE_SOLVER drs)        */
/*                                                                           */
/*  Make an assign resource expression from dmc.                             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsMonitorConverterMakeAssignResourceExpr(
  KHE_DRS_MONITOR_CONVERTER dmc, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_SUM_COMB eisc;  KHE_CONSTRAINT c;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_weight;
  int min_limit, i; KHE_ASSIGN_RESOURCE_MONITOR arm;
  KHE_DRS_TASK_ON_DAY dtd;  KHE_DRS_EXPR_ASSIGNED_TASK eat;
  KHE_DRS_EXPR_COST ec;  KHE_DRS_EXPR_INT_DEV eid;
  c = KheMonitorConstraint(dmc->monitor);
  combined_weight = KheConstraintCombinedWeight(c);
  cost_fn = KheConstraintCostFunction(c);
  arm = (KHE_ASSIGN_RESOURCE_MONITOR) dmc->monitor;
  min_limit = KheAssignResourceMonitorTotalMonitoredDuration(arm);
  if( cost_fn == KHE_LINEAR_COST_FUNCTION )
  {
    ** one tree per task on day **
    HaArrayForEach(dmc->task_on_days, dtd, i)
    {
      ec = KheDrsExprCostMake(dmc->monitor, cost_fn, combined_weight, drs);
      eid = KheDrsExprI ntDevMake(1, INT_MAX, false, drs);
      eat = KheDrsExprAssignedTaskMake(dtd, NULL, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eid);
      KheDrsExprAddChild((KHE_DRS_EXPR) eid, (KHE_DRS_EXPR) eat);
      HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) ec);
    }
  }
  else
  {
    ** a single tree, whose root has one child per task on day **
    eisc = KheDrsExprIntSumCombMake(dmc->monitor, cost_fn, combined_weight,
      min_limit, INT_MAX, false, 0, 0, KheDrsDomTestMake(KHE_DRS_DOM_GE), drs);
    HaArrayForEach(dmc->task_on_days, dtd, i)
    {
      eat = KheDrsExprAssignedTaskMake(dtd, NULL, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) eat);
    }
    HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) eisc);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsMonitorConverterMakePreferResourcesExpr(                      */
/*    KHE_DRS_MONITOR_CONVERTER dmc, KHE_DYNAMIC_RESOURCE_SOLVER drs)        */
/*                                                                           */
/*  Make a prefer resources expression from dmc.                             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsMonitorConverterMakePreferResourcesExpr(
  KHE_DRS_MONITOR_CONVERTER dmc, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_CONSTRAINT c;  KHE_RESOURCE_GROUP rg;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_weight;
  KHE_PREFER_RESOURCES_CONSTRAINT prc;  int i;  KHE_DRS_TASK_ON_DAY dtd;
  KHE_DRS_EXPR_COST ec;  KHE_DRS_EXPR_ASSIGNED_TASK eat;
  KHE_DRS_EXPR_INT_SUM eis;  KHE_DRS_DOM_TEST dom_test;
  c = KheMonitorConstraint(dmc->monitor);
  combined_weight = KheConstraintCombinedWeight(c);
  cost_fn = KheConstraintCostFunction(c);
  prc = (KHE_PREFER_RESOURCES_CONSTRAINT) c;
  rg = KhePreferResourcesConstraintDomainComplement(prc);
  if( cost_fn == KHE_LINEAR_COST_FUNCTION )
  {
    ** one tree per task on day **
    HaArrayForEach(dmc->task_on_days, dtd, i)
    {
      ec = KheDrsExprCostMake(dmc->monitor, cost_fn, combined_weight, drs);
      eat = KheDrsExprAssignedTaskMake(dtd, rg, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eat);
      HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) ec);
    }
  }
  else
  {
    ** a single tree, whose root has one grandchild per task on day **
    ec = KheDrsExprCostMake(dmc->monitor, cost_fn, combined_weight, drs);
    dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);
    eis = KheDrsExprInt SumMake(dom_test, drs);
    KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eis);
    HaArrayForEach(dmc->task_on_days, dtd, i)
    {
      eat = KheDrsExprAssignedTaskMake(dtd, rg, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) eis, (KHE_DRS_EXPR) eat);
    }
    HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) ec);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsMonitorConverterMakeLimitResourcesExpr(                       */
/*    KHE_DRS_MONITOR_CONVERTER dmc, KHE_DYNAMIC_RESOURCE_SOLVER drs)        */
/*                                                                           */
/*  Make a limit resources expression from dmc.                              */
/*                                                                           */
/*  The documentation mentions an unlikely case where the expression can     */
/*  be split up.  We haven't bothered testing for that here.                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsMonitorConverterMakeLimitResourcesExpr(
  KHE_DRS_MONITOR_CONVERTER dmc, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_INT_SUM_COMB eisc;  KHE_CONSTRAINT c;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_weight;
  int min_limit, max_limit, i;  KHE_RESOURCE_GROUP rg;
  KHE_LIMIT_RESOURCES_CONSTRAINT lrc;
  KHE_DRS_DOM_TEST alpha_dom_test, beta_dom_test;
  KHE_DRS_TASK_ON_DAY dtd;
  KHE_DRS_EXPR_ASSIGNED_TASK eat;
  c = KheMonitorConstraint(dmc->monitor);
  combined_weight = KheConstraintCombinedWeight(c);
  cost_fn = KheConstraintCostFunction(c);
  lrc = (KHE_LIMIT_RESOURCES_CONSTRAINT) c;
  rg = KheLimitResourcesConstraintDomain(lrc);
  min_limit = KheLimitResourcesConstraintMinimum(lrc);
  max_limit = KheLimitResourcesConstraintMaximum(lrc);
  KheLimitsToTests(min_limit, max_limit, INT_MAX, false,
    &alpha_dom_test, &beta_dom_test);
  eisc = KheDrsExprIntSumCombMake(dmc->monitor, cost_fn, combined_weight,
    min_limit, max_limit, false, 0, 0, alpha_dom_test, drs);
  HaArrayForEach(dmc->task_on_days, dtd, i)
  {
    eat = KheDrsExprAssignedTaskMake(dtd, rg, drs);
    KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) eat);
  }
  HaArrayAddLast(drs->open_exprs, (KHE_DRS_EXPR) eisc);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsMonitorConverterMakeExpr(KHE_DRS_MONITOR_CONVERTER dmc,       */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make an expression from dmc.  End by freeing dmc.                        */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsMonitorConverterMakeExpr(KHE_DRS_MONITOR_CONVERTER dmc,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  switch( KheMonitorTag(dmc->monitor) )
  {
    case KHE_ASSIGN_RESOURCE_MONITOR_TAG:

      KheDrsMonitorConverterMakeAssignResourceExpr(dmc, drs);
      break;

    case KHE_PREFER_RESOURCES_MONITOR_TAG:

      KheDrsMonitorConverterMakePreferResourcesExpr(dmc, drs);
      break;

    case KHE_AVOID_SPLIT_ASSIGNMENTS_MONITOR_TAG:

      ** we're skipping this one **
      break;

    case KHE_LIMIT_RESOURCES_MONITOR_TAG:

      KheDrsMonitorConverterMakeLimitResourcesExpr(dmc, drs);
      break;

    default:

      HnAbort("KheDrsMonitorConverterMakeExpr: unexpected monitor tag (%d)",
	KheMonitorTag(dmc->monitor));
      break;
  }

  ** end by freeing dmc and removing the pointer to it **
  KheMonitorSetBack(dmc->monitor, NULL);
  KheDrsMonitorConverterFree(dmc, drs);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsMonitorConverterDebug(KHE_DRS_MONITOR_CONVERTER dmc,          */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of dmc onto fp with the given verbosity and indent.          */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsMonitorConverterDebug(KHE_DRS_MONITOR_CONVERTER dmc,
  int verbosity, int indent, FILE *fp)
{
  KHE_DRS_TASK_ON_DAY dtd;  int i;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ MonitorConverter(%s %s)\n", indent, "",
      KheMonitorTagShow(KheMonitorTag(dmc->monitor)),
      KheMonitorId(dmc->monitor));
    HaArrayForEach(dmc->task_on_days, dtd, i)
      KheDrsTaskOnDayDebug(dtd, verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
    fprintf(fp, "MonitorConverter(%s %s, %d elts)\n",
      KheMonitorTagShow(KheMonitorTag(dmc->monitor)),
      KheMonitorId(dmc->monitor), HaArrayCount(dmc->task_on_days));
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_EDGE"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EDGE KheDrsEdgeMake(KHE_DRS_SOLN prev_soln,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new, empty edge object with these attributes.                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_EDGE KheDrsEdgeMake(KHE_DRS_SOLN prev_soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EDGE res;
  if( HaArrayCount(drs->free_edge) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_edge);
    HaArrayClear(res->tasks);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->tasks, drs->arena);
  }
  res->prev_soln = prev_soln;
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_EDGE KheDrsEdgeCopy(KHE_DRS_EDGE de,                             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a copy of edge.                                                     */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_EDGE KheDrsEdgeCopy(KHE_DRS_EDGE de,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EDGE res;  int i;
  res = KheDrsEdgeMake(de->prev_soln, drs);
  HaArrayAppend(res->tasks, de->tasks, i);
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsEdgeFree(KHE_DRS_EDGE de, KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Free de; add it to the free list in drs.                                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsEdgeFree(KHE_DRS_EDGE de, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ** de does not own its tasks; they will be deleted separately **
  HaArrayAddLast(drs->free_edge, de);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheEdgePushAsst(KHE_DRS_EDGE de, KHE_DRS_TASK_ON_DAY dtd,           */
/*    KHE_DRS_RESOURCE_ON_DAY drd)                                           */
/*                                                                           */
/*  Push the assignment of drd to dtd onto edge de, and update leaf exprs.   */
/*  Here drd must be non-NULL, but dtd could be NULL, in which case no leaf  */
/*  expr updates are required, because the default state is correct then.    */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheEdgePushAsst(KHE_DRS_EDGE de, KHE_DRS_TASK_ON_DAY dtd,
  KHE_DRS_RESOURCE_ON_DAY drd)
{
  KHE_DRS_EXPR e;  int i;

  ** add the task to the edge **
  HaArrayAddLast(de->tasks, dtd);

  if( dtd != NULL )
  {
    ** update leaf expressions of the resource on day **
    HaArrayForEach(drd->external_today, e, i)
      KheDrsExprLeafSet(e, dtd, drd);

    ** update leaf expressions of the task on day **
    HaArrayForEach(dtd->external_today, e, i)
      KheDrsExprLeafSet(e, dtd, drd);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheEdgePopAsst(KHE_DRS_EDGE de, KHE_DRS_TASK_ON_DAY dtd,            */
/*    KHE_DRS_RESOURCE_ON_DAY drd)                                           */
/*                                                                           */
/*  Undo the effect of the corresponding KheEdgePushAsst.                    */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheEdgePopAsst(KHE_DRS_EDGE de, KHE_DRS_TASK_ON_DAY dtd,
  KHE_DRS_RESOURCE_ON_DAY drd)
{
  KHE_DRS_EXPR e;  int i;

  ** remove the task from the edge **
  HaArrayDeleteLast(de->tasks);

  if( dtd != NULL )
  {
    ** update leaf expressions of the resource **
    HaArrayForEach(drd->external_today, e, i)
      KheDrsExprLeafClear(e);

    ** update leaf expressions of the task **
    HaArrayForEach(dtd->external_today, e, i)
      KheDrsExprLeafClear(e);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsEdgeDebug(KHE_DRS_EDGE de, KHE_DYNAMIC_RESOURCE_SOLVER drs,   */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of de onto fp with the given verbosity and indent.           */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsEdgeDebug(KHE_DRS_EDGE de, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  KHE_DRS_TASK_ON_DAY dtd;  int i;  KHE_DRS_RESOURCE dr;
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "[ edge ");
  HaArrayForEach(de->tasks, dtd, i)
  {
    if( i > 0 )
      fprintf(fp, ", ");
    dr = HaArray(drs->open_resources, i);
    fprintf(fp, "%s:", KheResourceId(dr->resource));
    KheDrsTaskOnDayDebug(dtd, 1, -1, fp);
  }
  fprintf(fp, " ]");
  if( indent >= 0 )
    fprintf(fp, "\n");
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_TIME"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TIME KheDrsTimeMake(KHE_TIME t, HA_ARENA a)                      */
/*                                                                           */
/*  Make and return a new drs time object representing t.                    */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_TIME KheDrsTimeMake(KHE_TIME t, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TIME res;
  HaMake(res, drs->arena);
  res->time = t;
  ** HaArrayInit(res->leaf_exprs, drs->arena); **
  HaArrayInit(res->best_tasks, drs->arena);
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTimeClearTasks(KHE_DRS_TIME st,                               */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Clear the tasks of st.                                                   */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTimeClearTasks(KHE_DRS_TIME st,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i;  KHE_DRS_TASK dt;
  HaArrayForEach(st->best_tasks, dt, i)
    KheDrsTaskFree(dt, drs);
  HaArrayClear(st->best_tasks);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTimeContainsLastTime(KHE_DRS_TIME stime,                      */
/*    KHE_TIME last_time, KHE_DRS_TASK *dt, int *pos)                        */
/*                                                                           */
/*  If stime contains a task with the given last_time, return true with      */
/*  *dt set to that task and *pos set to its position.  Otherwise return     */
/*  false with *dt set to NULL and *pos set to -1.                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDrsTimeContainsLastTime(KHE_DRS_TIME stime,
  KHE_TIME last_time, KHE_DRS_TASK *dt, int *pos)
{
  KHE_DRS_TASK dt2;  int i;
  HaArrayForEach(stime->best_tasks, dt2, i)
    if( KheDrsTaskLastTime(dt2) == last_time )
      return *dt = dt2, *pos = i, true;
  return *dt = NULL, *pos = -1, false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTimeAddTask(KHE_DRS_TIME st, KHE_DRS_TASK dt,                 */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add dt to st.  More precisely, if it's better than what st already       */
/*  has, then add it, otherwise delete it.                                   */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTimeAddTask(KHE_DRS_TIME st, KHE_DRS_TASK dt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK dt2;  int pos;
  if( !KheDrsTimeContainsLastTime(st, KheDrsTaskLastTime(dt), &dt2, &pos) )
  {
    ** st has no task with the same last time as dt, so add dt to st **
    HaArrayAddLast(st->best_tasks, dt);
  }
  else if( KheDrsTaskBetterThan(dt, dt2) )
  {
    ** dt is better than the similar task dt2, so replace dt2 by dt **
    KheDrsTaskFree(dt2, drs);
    HaArrayPut(st->best_tasks, pos, dt);
  }
  else
  {
    ** dt is worse than the similar dt2, so delete dt **
    KheDrsTaskFree(dt, drs);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTimeDebug(KHE_DRS_TIME st, KHE_DYNAMIC_RESOURCE_SOLVER drs,   */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of st onto fp with the given verbosity and indent.           */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTimeDebug(KHE_DRS_TIME st, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  int i;  ** KHE_DRS_EXPR e; **  KHE_DRS_TASK dt;
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ Time %s:\n", indent, "", KheTimeId(st->time));
    ** ***
    HaArrayForEach(st->leaf_exprs, e, i)
      KheDrsExprDebug(e, drs, verbosity, indent + 2, fp);
    *** **
    HaArrayForEach(st->best_tasks, dt, i)
      KheDrsTaskDebug(dt, verbosity, indent + 2, fp);
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
    fprintf(fp, "[ Time %s (%d best_tasks) ]", KheTimeId(st->time),
      HaArrayCount(st->best_tasks));
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_SOLN"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN KheDrsSolnGet(KHE_DYNAMIC_RESOURCE_SOLVER drs)              */
/*                                                                           */
/*  Get a new soln object.  Only the signature field is initialized.         */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_SOLN KheDrsSolnGet(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN res;
  if( HaArrayCount(drs->free_soln) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_soln);
    HaArrayClear(res->prev_tasks);
    HaArrayClear(res->signature);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->prev_tasks, drs->arena);
    HaArrayInit(res->signature, drs->arena);
  }
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN KheDrsSolnMakeRoot(KHE_DYNAMIC_RESOURCE_SOLVER drs)         */
/*                                                                           */
/*  Make the root soln that the search grows from.  It does not end on       */
/*  any day.                                                                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_SOLN KheDrsSolnMakeRoot(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN res;
  res = KheDrsSolnGet(drs);
  ** res->prev_soln = NULL; **
  res->prev_edge = NULL;
  res->day = NULL;
  res->best_of = 1;
  res->cost = 0;
  ** *** no longer including number of assts
  HaArrayAddLast(res->signature, 0);  ** number of assignments **
  *** **
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN KheDrsSolnMake(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY day,     */
/*    KHE_COST cost, KHE_DYNAMIC_RESOURCE_SOLVER drs)                        */
/*                                                                           */
/*  Make a new soln with these attributes.                                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN KheDrsSolnMake(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY day,
  KHE_COST cost, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN res;

  /* get the basic object */
  if( HaArrayCount(drs->free_soln) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_soln);
    HaArrayClear(res->prev_tasks);
    HaArrayClear(res->signature);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->prev_tasks, drs->arena);
    HaArrayInit(res->signature, drs->arena);
  }
  
  /* initialize the fields */
  res->prev_soln = prev_soln;
  res->day = day;
  res->best_of = 1;
  res->cost = cost;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN KheDrsSolnMakeNonRoot(KHE_DRS_SOLN prev_soln,               */
/*    KHE_DRS_EDGE prev_edge, KHE_DRS_DAY day,                               */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new non-root soln with these attributes.                          */
/*                                                                           */
/*****************************************************************************/
static void KheDrsSolnDebug(KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, int verbosity, int indent, FILE *fp);

/* ***
static KHE_DRS_SOLN KheDrsSolnMakeNonRoot(KHE_DRS_SOLN prev_soln,
  KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN res;  int ** asst_count, ** i ** , j **;  ** KHE_TIME time; **
  KHE_DRS_EXPR e;  ** KHE_TASK task; **  ** KHE_DRS_TASK dt; **
  ** KHE_DRS_RESOURCE dr; **  ** KHE_DRS_RESOURCE_ON_DAY drd, prev_drd **;

  ** make the basic object **
  HnAssert(prev_edge != NULL, "KheDrsSolnMake internal error (prev_edge)");
  if( DEBUG4 )
    fprintf(stderr, "[ KheDrsSolnMake(prev_soln, prev_edge, %s, drs)\n",
      KheTimeGroupId(day->time_group));
  res = KheDrsSolnGet(drs);
  ** res->prev_soln = prev_soln; **
  res->prev_edge = prev_edge;
  res->day = day;
  res->best_of = 1;

  ** set the cost and signature **
  res->cost = prev_edge->prev_soln->cost;
  ** *** no longer including number of assts
  asst_count = HaArrayFirst(prev_soln->signature);
  HaArrayAddLast(res->signature, asst_count);
  *** **
  HaArrayForEach(day->internal_today, e, i)
    KheDrsExprEvalDay(e, day->open_day_index, prev_edge->prev_soln, res, drs);
  ** *** no longer doing this, since it doesn't guarantee optimality anyway
  HaArrayForEach(drs->open_resources, dr, i)
  {
    ** adding each resource's excess to the signature s till to do **
  }
  *** **
  ** ***
  HaArrayForEach(drs->open_resources, dr, i)
  {
    ** *** replacement for this still to d*
    dt = HaArray(prev_edge->tasks, i);
    task = KheDailyScheduleTask(dt->daily_schedule, day->day_index);
    time = KheDailyScheduleTime(dt->daily_schedule, day->day_index);
    *** **
    task = NULL;
    time = NULL;
    if( task != NULL )
      asst_count++;
    drd = HaArray(dr->days, day->day_index);
    ** prev_drd = HaArray(dr->days, day->day_index - 1); **
    ** ***
    HaArrayForEach(drd->internal_today, e, j)
      KheDrsExprDoDay(e, day->day_index, prev_soln, ** prev_drd->sig_offset, **
	-1, task, time, res, ** drd->sig_offset, ** -1, drs);
    *** **
    ** *** replacement for this st ill to do
    HaArrayAddLast(res->signature,
      KheDailyScheduleLastDayIndex(dt->daily_schedule) - day->day_index);
    *** **
  }
  HaArrayPut(res->signature, 0, asst_count);
  ** *** stil l to do (must iterate over selected resources and get tasks 
  HaArrayForEach(day->internal_today, e, i)
    KheDrsExprDoDay(e, day->day_index, prev_soln, task, time, res, drs);
  *** **
  *** **

  if( DEBUG4 )
  {
    KheDrsSolnDebug(res, drs, 1, 2, stderr);
    fprintf(stderr, "] KheDrsSolnMake returning\n");
  }
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnFree(KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Free soln.                                                               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnFree(KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HaArrayAddLast(drs->free_soln, soln);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN KheDrsSolnEvaluate(KHE_DRS_SOLN soln,                       */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Copy soln and evaluate the copy.  Here soln has a non-NULL prev_soln.    */
/*                                                                           */
/*****************************************************************************/
static void KheDrsDayAddSoln(KHE_DRS_DAY day, KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsSolnEvaluate(KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN res;  KHE_DRS_EXPR e;  int i;  KHE_DRS_DAY day;

  /* quit early if over solve_soln_limit */
  if( --drs->solve_soln_limit < 0 )
    return;

  /* make res, a copy of soln */
  HnAssert(soln->prev_soln != NULL, "KheDrsSolnEvaluate internal error");
  if( DEBUG4 )
    fprintf(stderr, "[ KheDrsSolnMake(soln, drs)\n");
  res = KheDrsSolnMake(soln->prev_soln, soln->day, soln->prev_soln->cost, drs);
  HaArrayAppend(res->prev_tasks, soln->prev_tasks, i);

  /* set signature */
  day = soln->day;
  HaArrayForEach(day->internal_today, e, i)
    KheDrsExprEvalDay(e, day->open_day_index, soln->prev_soln, res, drs);

  /* check cost limit and either add to day or free soln */
  if( res->cost < drs->solve_init_cost )
    KheDrsDayAddSoln(res->day, res, drs);
  else
    KheDrsSolnFree(res, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnPushAsst(KHE_DRS_SOLN soln, KHE_DRS_TASK_ON_DAY dtd,      */
/*    KHE_DRS_RESOURCE_ON_DAY drd)                                           */
/*                                                                           */
/*  Push the assignment of drd to dtd onto soln, and update leaf exprs.      */
/*  Here drd must be non-NULL, but dtd could be NULL, in which case no leaf  */
/*  expr updates are required, because the default state is correct then.    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnPushAsst(KHE_DRS_SOLN soln, KHE_DRS_TASK_ON_DAY dtd,
  KHE_DRS_RESOURCE_ON_DAY drd)
{
  KHE_DRS_EXPR e;  int i;

  /* add the task to the edge */
  HaArrayAddLast(soln->prev_tasks, dtd);

  if( dtd != NULL )
  {
    /* update leaf expressions of the resource on day */
    HaArrayForEach(drd->external_today, e, i)
      KheDrsExprLeafSet(e, dtd, drd);

    /* update leaf expressions of the task on day */
    HaArrayForEach(dtd->external_today, e, i)
      KheDrsExprLeafSet(e, dtd, drd);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnPopAsst(KHE_DRS_SOLN soln, KHE_DRS_TASK_ON_DAY dtd,       */
/*    KHE_DRS_RESOURCE_ON_DAY drd)                                           */
/*                                                                           */
/*  Undo the effect of the corresponding KheDrsSolnPushAsst.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnPopAsst(KHE_DRS_SOLN soln, KHE_DRS_TASK_ON_DAY dtd,
  KHE_DRS_RESOURCE_ON_DAY drd)
{
  KHE_DRS_EXPR e;  int i;

  /* remove the task from the edge */
  HaArrayDeleteLast(soln->prev_tasks);

  if( dtd != NULL )
  {
    /* update leaf expressions of the resource */
    HaArrayForEach(drd->external_today, e, i)
      KheDrsExprLeafClear(e);

    /* update leaf expressions of the task */
    HaArrayForEach(dtd->external_today, e, i)
      KheDrsExprLeafClear(e);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSolnSignatureFullHash(KHE_DRS_SOLN soln)                       */
/*                                                                           */
/*  Hash function for hashing the full signature of soln.                    */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSolnSignatureFullHash(KHE_DRS_SOLN soln)
{
  int res, val, i;
  res = 0;
  HaArrayForEach(soln->signature, val, i)
    res = (res + val) * 3;
  return (res < 0 ? - res : res);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSolnSignatureFullHashUntyped(void *p)                          */
/*                                                                           */
/*  Untyped version of KheDrsSolnSignatureHash.                              */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSolnSignatureFullHashUntyped(void *p)
{
  return KheDrsSolnSignatureFullHash((KHE_DRS_SOLN) p);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSolnSignaturePartialHash(KHE_DRS_SOLN soln)                    */
/*                                                                           */
/*  Hash function for hashing the partial signature of soln, that is, only   */
/*  the parts of it whose dominance tests are equality (or loose equality).  */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSolnSignaturePartialHash(KHE_DRS_SOLN soln)
{
  int res, index, val, i;  KHE_DRS_DAY day;
  res = 0;
  day = soln->day;
  HaArrayForEach(day->eq_dom_test_indexes, index, i)
  {
    val = HaArray(soln->signature, index);
    res = (res + val) * 3;
  }
  if( res < 0 )
    res = - res;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnSignatureFullEqual(KHE_DRS_SOLN soln1, KHE_DRS_SOLN soln2)*/
/*                                                                           */
/*  Return true if the signatures of soln1 and soln2 are equal.  The lengths */
/*  of their signatures must be equal.                                       */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnSignatureFullEqual(KHE_DRS_SOLN soln1, KHE_DRS_SOLN soln2)
{
  int i;
  HnAssert(HaArrayCount(soln1->signature) == HaArrayCount(soln2->signature),
    "KheDrsSolnSignaturesEqual: signatures have different lengths");
  for( i = 0;  i < HaArrayCount(soln1->signature);  i++ )
    if( HaArray(soln1->signature, i) != HaArray(soln2->signature, i) )
      return false;
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnSignatureFullEqualUntyped(void *p1, void *p2)             */
/*                                                                           */
/*  Untyped version of KheDrsSolnSignaturesEqual.                            */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnSignatureFullEqualUntyped(void *p1, void *p2)
{
  return KheDrsSolnSignatureFullEqual((KHE_DRS_SOLN) p1,
    (KHE_DRS_SOLN) p2);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnSignaturePartialEqual(KHE_DRS_SOLN soln1,                 */
/*    KHE_DRS_SOLN soln2)                                                    */
/*                                                                           */
/*  Return true if the signatures of soln1 and soln2 are partially equal,    */
/*  that is, the entries whose dominance tests are equality (or loose        */
/*  equality) are equal.                                                     */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnSignaturePartialEqual(KHE_DRS_SOLN soln1,
  KHE_DRS_SOLN soln2)
{
  int i, index;  KHE_DRS_DAY day;
  HnAssert(HaArrayCount(soln1->signature) == HaArrayCount(soln2->signature),
   "KheDrsSolnSignaturePartialEqual: signatures have different lengths");
  day = soln1->day;
  HaArrayForEach(day->eq_dom_test_indexes, index, i)
    if( HaArray(soln1->signature, index) != HaArray(soln2->signature, index) )
      return false;
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSignatureDebug(KHE_DRS_SOLN soln, FILE *fp)               */
/*                                                                           */
/*  Debug print of the signature of soln onto fp (no verbosity or indent).   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSignatureDebug(KHE_DRS_SOLN soln, FILE *fp)
{
  int i, val;
  fprintf(fp, "[");
  HaArrayForEach(soln->signature, val, i)
  {
    if( i > 0 )
      fprintf(fp, " ");
    fprintf(fp, "%d", val);
  }
  fprintf(fp, "]");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSignatureDebugUntyped(void *p, FILE *fp)                  */
/*                                                                           */
/*  Untyped version of KheDrsSolnSignatureDebug.                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSignatureDebugUntyped(void *p, FILE *fp)
{
  KheDrsSolnSignatureDebug((KHE_DRS_SOLN) p, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnDominates(KHE_DRS_SOLN soln1, KHE_DRS_SOLN soln2,         */
/*    int start_depth)                                                       */
/*                                                                           */
/*  Return true if soln1 dominates soln2.  They come from the same day.      */
/*  Parameter start_depth says how far along the signature to start.         */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnDominates(KHE_DRS_SOLN soln1, KHE_DRS_SOLN soln2,
  int start_depth)
{
  KHE_DRS_DOM_TEST dt;  int i, val1, val2, sig_len;  KHE_DRS_DAY day;
  HnAssert(soln1->day == soln2->day,
    "KheDrsSolnDominates internal error: soln1->day = %s, soln2->day = %s",
    soln1->day != NULL ? KheTimeGroupId(soln1->day->time_group) : "-",
    soln2->day != NULL ? KheTimeGroupId(soln2->day->time_group) : "-");
  if( soln2->cost < soln1->cost )
    return false;
  day = soln1->day;
  sig_len = HaArrayCount(soln1->signature);
  for( i = start_depth;  i < sig_len;  i++ )
  {
    dt = HaArray(day->dom_tests, i);
    val1 = HaArray(soln1->signature, i);
    val2 = HaArray(soln2->signature, i);
    if( !KheDrsDomTestDominates(dt, val1, val2) )
      return false;
  }
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnOverWrite(KHE_DRS_SOLN dest_soln, KHE_DRS_SOLN src_soln)  */
/*                                                                           */
/*  Overwrite src_soln onto dest_soln.  Their signatures are known to be     */
/*  equal, so those don't need to be touched.  Also the best_of field        */
/*  is not touched here; it gets updated separately.                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnOverWrite(KHE_DRS_SOLN dest_soln, KHE_DRS_SOLN src_soln)
{
  int i;
  /* dest_soln->prev_soln = src_soln->prev_soln; */
  HaArrayClear(dest_soln->prev_tasks);
  HaArrayAppend(dest_soln->prev_tasks, src_soln->prev_tasks, i);
  dest_soln->prev_soln = src_soln->prev_soln;
  dest_soln->cost = src_soln->cost;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSolnIncreasingAsstCountCmp(const void *p1, void *p2)           */
/*                                                                           */
/*  Comparison function for sorting an array of solns by increasing          */
/*  assignment count.                                                        */
/*                                                                           */
/*****************************************************************************/

/* *** no longer storing asst count in the signature
static int KheDrsSolnIncreasingAsstCountCmp(const void *p1,
  const void *p2)
{
  KHE_DRS_SOLN soln1 = * (KHE_DRS_SOLN *) p1;
  KHE_DRS_SOLN soln2 = * (KHE_DRS_SOLN *) p2;
  return HaArrayFirst(soln1->signature) - HaArrayFirst(soln2->signature);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void DoExtend(KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_SOLN soln,        */
/*    KHE_DRS_EDGE de)                                                       */
/*                                                                           */
/*  Extend soln using de.                                                    */
/*                                                                           */
/*****************************************************************************/

/* ***
static void OldDoExtend(KHE_DRS_SOLN soln, KHE_DRS_EDGE de, KHE_DRS_DAY new_day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN new_soln;
  new_soln = KheDrsSolnMakeNonRoot(soln, de, new_day, drs);
  KheDrsDayAddSoln(new_day, new_soln, drs);
  drs->solve_diversifier++;

  ** ***
  KHE_DRS_SOLN curr_soln, last_soln;
  int di, first_day_index, last_day_index;  ** KHE_DRS_DAY last_day; **
  ** KHE_TIME time;  KHE_DRS_TIME stime;  KHE_TASK task; **

  ** make one soln for each day of dt, ending at last_soln **
  last_soln = soln;
  first_day_index = KheDailyScheduleFirstDayIndex(dt->daily_schedule);
  last_day_index = KheDailyScheduleLastDayIndex(dt->daily_schedule);
  for( di = first_day_index;  di <= last_day_index;  di++ )
    last_soln = KheDrsSolnMakeNonRoot(last_soln, dt, HaArray(drs->days, di), drs);
  HnAssert(last_soln != soln, "DoExtend internal error");

  ** delete all solns between soln and last_soln exclusive **
  ** NB field curr_soln->prev is usable after calling KheDrsSolnFree **
  for( curr_soln = last_soln->prev_soln;  curr_soln != soln;  curr_soln = curr_soln->prev_soln )
    KheDrsSolnFree(curr_soln, drs);
  last_soln->prev_soln = soln;

  ** add last_soln to last_day, and increment the diversifier **
  ** last_day = HaArray(drs->days, last_day_index); **
  KheDrsDayAddSoln(last_soln->day, last_soln, drs);
  drs->solve_diversifier++;
  *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnResourceIsAssigned(KHE_DRS_SOLN soln,                     */
/*    KHE_DRS_RESOURCE dr, int dr_solve_index, KHE_DRS_TASK_ON_DAY *dtd)     */
/*                                                                           */
/*  If dr is assigned in soln, return true and set *dtd to what it is        */
/*  assigned to; otherwise return false and set *dtd to NULL.                */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnResourceIsAssigned(KHE_DRS_SOLN soln,
  KHE_DRS_RESOURCE dr, int dr_solve_index, KHE_DRS_TASK_ON_DAY *dtd)
{
  if( soln == NULL )
    return *dtd = NULL, false;
  *dtd = HaArray(soln->prev_tasks, dr_solve_index);
  return *dtd != NULL;
  /* ***
  KHE_DRS_EDGE de;
  de = soln->prev_edge;
  if( de == NULL )
    return *dtd = NULL, false;
  *dtd = HaArray(de->tasks, dr_solve_index);
  return *dtd != NULL;
  *** */
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsResourceIsFixedOnDay(KHE_DRS_RESOURCE dr, KHE_DRS_DAY day,    */
/*    KHE_DRS_TASK *dt)                                                      */
/*                                                                           */
/*  If dr has a fixed assignment on day, return true and set *dt to the      */
/*  task it is fixed to.  Otherwise return false with *dt set to NULL.       */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheDrsResourceIsFixedOnDay(KHE_DRS_RESOURCE dr, KHE_DRS_DAY day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ** KHE_DRS_RESOURCE_ON_DAY drd; **
  KHE_RESOURCE_TIMETABLE_MONITOR rtm;
  ** drd = HaArray(dr->days, day->solve_index); **
  rtm = KheResourceTimetableMonitor(drs->soln, dr->resource);
  return !KheResourceTimetableMonitorTimeGroupAvailable(rtm,
    day->time_group, true);
  ** ***
  return drd->fixed_asst_today;
  *dt = drd->fixed_asst;
  return (*dt != NULL);
  *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTaskBusyOnDay(KHE_DRS_TASK dt, KHE_DRS_DAY day,               */
/*    KHE_DRS_TASK_ON_DAY *dtd)                                              */
/*                                                                           */
/*  If dt is running on day, return true with *dtd set to its task on day    */
/*  object for that day.  Otherwise return false with *dtd set to NULL.      */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsTaskBusyOnDay(KHE_DRS_TASK dt, KHE_DRS_DAY day,
  KHE_DRS_TASK_ON_DAY *dtd)
{
  KHE_DRS_TASK_ON_DAY dtd2;  int i;
  HaArrayForEach(dt->days, dtd2, i)
    if( dtd2->day == day )
      return *dtd = dtd2, true;
  return *dtd = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsTaskClassAcceptResourceBegin(KHE_DRS_TASK_CLASS dtc,          */
/*    KHE_DRS_RESOURCE dr, KHE_DRS_TASK *dt)                                 */
/*                                                                           */
/*  If dtc accepts dr (because dr lies in its tasks' domains, and not all    */
/*  the tasks are assigned), then return true and set *dt to one of dtc's    */
/*  tasks that dr can be assigned to.  Otherwise return false.               */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsTaskClassAcceptResourceBegin(KHE_DRS_TASK_CLASS dtc,
  KHE_DRS_RESOURCE dr, KHE_DRS_TASK *dt)
{
  KHE_RESOURCE_GROUP domain;
  domain = KheTaskClassDomain(dtc->orig_task_class);
  if( dtc->tmp_used < HaArrayCount(dtc->open_tasks) &&
	KheResourceGroupContains(domain, dr->resource) )
  {
    *dt = HaArray(dtc->open_tasks, dtc->tmp_used);
    dtc->tmp_used++;
    return true;
  }
  else
    return false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskClassAcceptResourceEnd(KHE_DRS_TASK_CLASS dtc)            */
/*                                                                           */
/*  Undo the effect of a corresponding KheDrsTaskClassAcceptResourceBegin    */
/*  which returned true.                                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsTaskClassAcceptResourceEnd(KHE_DRS_TASK_CLASS dtc)
{
  dtc->tmp_used--;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsResourceOnDayIsFixed(KHE_DRS_RESOURCE_ON_DAY drd,             */
/*    int dr_index, KHE_DRS_SOLN soln, KHE_DRS_TASK_ON_DAY *dtd)             */
/*                                                                           */
/*  If the assignment of drd (whose resource's open index is dr_index)       */
/*  is fixed, return in *dtd the task on day that it is fixed to.            */
/*                                                                           */
/*  There are two reasons why drd may be fixed:                              */
/*                                                                           */
/*    * drd may have a closed assignment;                                    */
/*                                                                           */
/*    * drd's resource may have been assigned to some task on the previous   */
/*      day, and that task may still be running                              */
/*                                                                           */
/*  We can find out whether drd's resource has been assigned to some task    */
/*  on the previous day by consulting the edge leading into soln.            */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsResourceOnDayIsFixed(KHE_DRS_RESOURCE_ON_DAY drd,
  int dr_index, KHE_DRS_SOLN soln, KHE_DRS_TASK_ON_DAY *dtd)
{
  KHE_DRS_TASK_ON_DAY dtd1, dtd2;

  /* if drd has a closed assignment, it's fixed to that */
  if( drd->closed_asst != NULL )
    return *dtd = drd->closed_asst, true;

  /* if drd's resource is assigned to a task in soln which is still running, */
  /* then drd is fixed to that */
  if( KheDrsSolnResourceIsAssigned(soln, drd->encl_dr, dr_index, &dtd1) &&
	KheDrsTaskBusyOnDay(dtd1->encl_dt, drd->day, &dtd2) )
    return *dtd = dtd2, true;

  /* otherwise drd has no fixed assignment */
  return *dtd = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnDoExtend(KHE_DRS_SOLN soln,                               */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Extend soln from soln->prev_soln to soln->day; build soln->prev_tasks    */
/*  up to full length first.                                                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnDoExtend(KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_RESOURCE dr;  int i, dr_index;  KHE_DRS_TASK dt;
  KHE_DRS_TASK_CLASS dtc;  KHE_DRS_RESOURCE_ON_DAY drd;
  KHE_DRS_TASK_ON_DAY dtd;  /* KHE_DRS_SOLN soln2; */
  if( HaArrayCount(soln->prev_tasks) >= HaArrayCount(drs->open_resources) )
  {
    /* soln is ready to use; add it to soln->day */
    KheDrsSolnEvaluate(soln, drs);
    drs->solve_diversifier++;
  }
  else
  {
    dr_index = HaArrayCount(soln->prev_tasks);
    dr = HaArray(drs->open_resources, dr_index);
    drd = KheDrsResourceOnDay(dr, soln->day);
    if( KheDrsResourceOnDayIsFixed(drd, dr_index, soln->prev_soln, &dtd) )
    {
      /* try dtd only */
      KheDrsSolnPushAsst(soln, dtd, drd);
      KheDrsSolnDoExtend(soln, drs);
      KheDrsSolnPopAsst(soln, dtd, drd);
    }
    else
    {
      /* try a task from any class that will accept dtd */
      HaArrayForEach(soln->day->task_classes, dtc, i)
	if( KheDrsTaskClassAcceptResourceBegin(dtc, dr, &dt) )
	{
	  if( !KheDrsTaskBusyOnDay(dt, soln->day, &dtd) )
	    HnAbort("KheDrsSolnDoExtend internal error");
	  KheDrsSolnPushAsst(soln, dtd, drd);
	  KheDrsSolnDoExtend(soln, drs);
	  KheDrsSolnPopAsst(soln, dtd, drd);
	  KheDrsTaskClassAcceptResourceEnd(dtc);
	}

      /* try a free day */
      KheDrsSolnPushAsst(soln, NULL, drd);
      KheDrsSolnDoExtend(soln, drs);
      KheDrsSolnPopAsst(soln, NULL, drd);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnExtend(KHE_DRS_SOLN soln, KHE_DRS_DAY new_day,            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Extend the dynamic programming search by searching out of soln.  Each    */
/*  new task starts on new_day.                                              */
/*                                                                           */
/*  Implementation note.  The call to KheDrsEdgeFree is safe here, because   */
/*  edge de does not go into the graph structure; only copies of it do.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnExtend(KHE_DRS_SOLN prev_soln, KHE_DRS_DAY new_day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN soln;
  soln = KheDrsSolnMake(prev_soln, new_day, 0 /* cost not used here */, drs);
  KheDrsSolnDoExtend(soln, drs);
  KheDrsSolnFree(soln, drs);

  /* extend for each task beginning on day */
  /* *** 
  HaArrayForEach(day->times, stime, i)
    HaArrayForEach(stime->best_tasks, dt, j)
      DoExtend(soln, dt, drs);
  ** extend by taking day off **
  DoExtend(soln, day->free_task, drs);
  *** */
}

/* ***
static void KheDrsSolnExtend(KHE_DRS_SOLN soln, KHE_DRS_DAY new_day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ** int i, j;  KHE_DRS_TASK dt; **  ** KHE_DRS_TIME stime; **
  KHE_DRS_EDGE de;
  de = KheDrsEdgeMake(soln, drs);
  KheDrsSolnDoExtend(soln, de, new_day, drs);
  KheDrsEdgeFree(de, drs);

  ** extend for each task beginning on day **
  ** *** 
  HaArrayForEach(day->times, stime, i)
    HaArrayForEach(stime->best_tasks, dt, j)
      DoExtend(soln, dt, drs);
  ** extend by taking day off **
  DoExtend(soln, day->free_task, drs);
  *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnDebugBasics(KHE_DRS_SOLN soln, FILE *fp)                  */
/*                                                                           */
/*  Produce an unindented one-line debug print of the basics of soln.        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnDebugBasics(KHE_DRS_SOLN soln, FILE *fp)
{
  /* task */
  fprintf(fp, "SOLN(");
  /* ***
  if( soln->task != NULL && soln->task->task != NULL )
    KheTaskDebug(soln->task->task, 1, -1, fp);
  else
    fprintf(fp, "-");
  *** */

  /* day_index */
  if( soln->day == NULL )
    fprintf(fp, "to -");
  else
    fprintf(fp, "to %s", KheTimeGroupId(soln->day->time_group));

  /* assts, best_of, and cost */
  fprintf(fp, ", best of %d, cost %.5f)", soln->best_of,
    KheCostShow(soln->cost));
  /* ***
  fprintf(fp, ", assts %d, best of %d, cost %.5f)",
    HaArray(soln->signature, 0), soln->best_of, KheCostShow(soln->cost));
  *** */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnDebug(KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs, */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of soln onto fp with the given verbosity and indent.         */
/*                                                                           */
/*  NB If verbosity == 1 or indent < 0, it is safe to pass NULL for drs.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnDebug(KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  /* KHE_DRS_EXPR e;  KHE_DRS_RESOURCE dr;  int i, j; */
  if( indent < 0 )
    KheDrsSolnDebugBasics(soln, fp);
  else if( verbosity == 1 )
  {
    fprintf(fp, "%*s", indent, "");
    KheDrsSolnDebugBasics(soln, fp);
    fprintf(fp, "\n");
  }
  else
  {
    /* basics */
    fprintf(fp, "%*s[ ", indent, "");
    KheDrsSolnDebugBasics(soln, fp);
    fprintf(fp, "\n");

    /* incoming soln */
    if( soln->prev_soln != NULL )
    {
      /* *** something to do here? still to do
      KheDrsEdgeDebug(soln->prev_edge, drs, verbosity, indent + 2, fp);
      *** */
    }

    /* signature */
    /* *** doesn't work this way now
    HaArrayForEach(drs->open_resources, dr, i)
    {
      HaArrayForEach(dr->exprs, e, j)
      {
	if( KheDrsExprFirstDebugDayIndex(e) <= soln->day->solve_index &&
	    soln->day->solve_index <= e->last_day_index )
	{
	  fprintf(fp, "%*s", indent + 2, "");
	  KheDrsExprDebug(e, soln, drs, verbosity, -1, fp);
	  fprintf(fp, "\n");
	}
      }
    }
    *** */
    fprintf(fp, "%*s]\n", indent, "");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_SOLN_LIST"                                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN_LIST KheDrsSolnListMake(KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Return a new, empty KHE_DRS_SOLN_LIST object.                            */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN_LIST KheDrsSolnListMake(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN_LIST res;
  if( HaArrayCount(drs->free_soln_list) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_soln_list);
    HaArrayClear(res->solns);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->solns, drs->arena);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnListFree(KHE_DRS_SOLN_LIST ns,                            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free ns and the solns in it.                                             */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnListFree(KHE_DRS_SOLN_LIST ns,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN soln;  int i;
  HaArrayForEach(ns->solns, soln, i)
    KheDrsSolnFree(soln, drs);
  HaArrayAddLast(drs->free_soln_list, ns);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSolnListSignaturePartialHashUntyped(void *p)                   */
/*                                                                           */
/*  Hash function for solnss, being the partial hash of the set's first      */
/*  element.                                                                 */
/*                                                                           */
/*****************************************************************************/

static int KheDrsSolnListSignaturePartialHashUntyped(void *p)
{
  KHE_DRS_SOLN_LIST ns;  KHE_DRS_SOLN soln;
  ns = (KHE_DRS_SOLN_LIST) p;
  soln = HaArrayFirst(ns->solns);
  return KheDrsSolnSignaturePartialHash(soln);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnListSignaturePartialEqualUntyped(void *p1, void *p2)      */
/*                                                                           */
/*  Equality function for solns, being the partial equality of the sets'     */
/*  first elements.                                                          */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnListSignaturePartialEqualUntyped(void *p1, void *p2)
{
  KHE_DRS_SOLN_LIST ns1, ns2;  KHE_DRS_SOLN soln1, soln2;
  ns1 = (KHE_DRS_SOLN_LIST) p1;
  ns2 = (KHE_DRS_SOLN_LIST) p2;
  soln1 = HaArrayFirst(ns1->solns);
  soln2 = HaArrayFirst(ns2->solns);
  return KheDrsSolnSignaturePartialEqual(soln1, soln2);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnListDoStrongDominanceTest(KHE_DRS_SOLN_LIST ns,           */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs, int *total_count)  */
/*                                                                           */
/*  Carry out a strong dominance test between ns and soln.                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnListDoStrongDominanceTest(KHE_DRS_SOLN_LIST ns,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs, int *total_count)
{
  KHE_DRS_SOLN other_soln;  int i;

  /* if soln is dominated by anything else, delete soln and return */
  HaArrayForEach(ns->solns, other_soln, i)
    if( KheDrsSolnDominates(other_soln, soln, 0) )
    {
      KheDrsSolnFree(soln, drs);
      return;
    }

  /* remove other solns that soln dominates */
  HaArrayForEach(ns->solns, other_soln, i)
    if( KheDrsSolnDominates(soln, other_soln, 0) )
    {
      KheDrsSolnFree(other_soln, drs);
      HaArrayDeleteAndPlug(ns->solns, i);
      (*total_count)--;
      i--;
    }

  /* add soln to ns */
  HaArrayAddLast(ns->solns, soln);
  (*total_count)++;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnListSignatureDebugUntyped(void *p, FILE *fp)              */
/*                                                                           */
/*  Debug print of the signature of soln set p.                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnListSignatureDebugUntyped(void *p, FILE *fp)
{
  KHE_DRS_SOLN_LIST ns;  KHE_DRS_SOLN soln;
  ns = (KHE_DRS_SOLN_LIST) p;
  soln = HaArrayFirst(ns->solns);
  KheDrsSolnSignatureDebug(soln, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_SOLN_TRIE"                                            */
/*                                                                           */
/*  Abstraction function                                                     */
/*  --------------------                                                     */
/*                                                                           */
/*  A soln_trie represents a set of solns.  If soln_trie == NULL, the set    */
/*  is empty.  Otherwise the set consists of soln_trie->soln (if non-NULL)   */
/*  plus the sets stored in each of the children.                            */
/*                                                                           */
/*  Invariant                                                                */
/*  ---------                                                                */
/*                                                                           */
/*  Exactly one of these three cases applies:                                */
/*                                                                           */
/*    (1)  soln_trie == NULL                                                 */
/*    (2)  soln_trie->soln != NULL                                           */
/*    (3)  HaArrayCount(soln_trie->children) > 0                             */
/*                                                                           */
/*  and in the last case, the last child of soln_trie->children is non-NULL. */
/*  Each soln appears in the trie in a place determined by the sequence      */
/*  of values in its signature, in the usual trie manner.                    */
/*                                                                           */
/*  There is no requirement that two or more solns be present in case (3).   */
/*  If there are no solns, case (1) always applies; if there are two or      */
/*  more solns, case (3) always applies; but if there is exactly one soln,   */
/*  then either of cases (2) or (3) applies.  This is because it would be    */
/*  too slow and tedious to detect situations where a case (3) soln could    */
/*  be simplified to a case (2) soln.                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN_TRIE KheDrsSolnTrieMake(KHE_DRS_SOLN soln,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make a new trie object containing soln.  It satisfies case (2) above.    */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_SOLN_TRIE KheDrsSolnTrieMake(KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN_TRIE res;
  HnAssert(soln != NULL, "KheDrsSolnTrieMake internal error (soln == NULL)");
  if( HaArrayCount(drs->free_soln_trie) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_soln_trie);
    HaArrayClear(res->children);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->children, drs->arena);
  }
  res->soln = soln;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieFree(KHE_DRS_SOLN_TRIE *soln_trie,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free *soln_trie and all its descendants and solns.                       */
/*  Set *soln_trie to NULL.                                                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieFree(KHE_DRS_SOLN_TRIE *soln_trie,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i;
  if( *soln_trie == NULL )
  {
    /* case (1), do nothing */
  }
  else if( (*soln_trie)->soln != NULL )
  {
    /* case (2), delete (*soln_trie)->soln and *soln_trie */
    KheDrsSolnFree((*soln_trie)->soln, drs);
    HaArrayAddLast(drs->free_soln_trie, *soln_trie);
    *soln_trie = NULL;
  }
  else
  {
    /* case (3), delete children and *soln_trie */
    for( i = 0;  i < HaArrayCount((*soln_trie)->children);  i++ )
      KheDrsSolnTrieFree(&HaArray((*soln_trie)->children, i), drs);
    HaArrayAddLast(drs->free_soln_trie, *soln_trie);
    *soln_trie = NULL;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieInsertChild(KHE_DRS_SOLN_TRIE soln_trie,              */
/*    KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)         */
/*                                                                           */
/*  Insert soln in its appropriate place among the children of soln_trie.    */
/*                                                                           */
/*****************************************************************************/
static void KheDrsSolnTrieInsert(KHE_DRS_SOLN_TRIE *soln_trie,
  KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsSolnTrieInsertChild(KHE_DRS_SOLN_TRIE soln_trie,
  KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int index;
  index = HaArray(soln->signature, depth);
  HaArrayFill(soln_trie->children, index + 1, NULL);
  KheDrsSolnTrieInsert(&HaArray(soln_trie->children, index), soln,
    depth + 1, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieInsert(KHE_DRS_SOLN_TRIE *soln_trie,                  */
/*    KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)         */
/*                                                                           */
/*  Insert soln into *soln_trie, possibly changing *soln_trie as a result.   */
/*  The depth parameter indicates how far along the signature we are.        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieInsert(KHE_DRS_SOLN_TRIE *soln_trie,
  KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  if( *soln_trie == NULL )
  {
    /* case (1), make a new soln containing just soln, -> case (2) */
    *soln_trie = KheDrsSolnTrieMake(soln, drs);
  }
  else if( (*soln_trie)->soln != NULL )
  {
    /* case (2), add soln and (*soln_trie)->soln to children, -> case (3) */
    KheDrsSolnTrieInsertChild(*soln_trie, soln, depth, drs);
    soln = (*soln_trie)->soln;
    (*soln_trie)->soln = NULL;
    KheDrsSolnTrieInsertChild(*soln_trie, soln, depth, drs);
  }
  else
  {
    /* case (3), add soln to children, -> case (3) */
    KheDrsSolnTrieInsertChild(*soln_trie, soln, depth, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsSolnTrieDominates(KHE_DRS_SOLN_TRIE soln_trie,                */
/*    KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)         */
/*                                                                           */
/*  Return true if there is a soln in soln_trie that dominates soln.         */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsSolnTrieDominates(KHE_DRS_SOLN_TRIE soln_trie,
  KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int val, from_val, to_val, i;  KHE_DRS_DOM_TEST dom_test;
  if( soln_trie == NULL )
  {
    /* case (1); no solns in soln_trie, so none dominate */
    return false;
  }
  else if( soln_trie->soln != NULL )
  {
    /* case (2); the only possible dominator is soln_trie->soln */
    return KheDrsSolnDominates(soln_trie->soln, soln, depth);
  }
  else
  {
    /* case (3); find range of dominators and try each of them */
    dom_test = HaArray(soln->day->dom_tests, depth);
    val = HaArray(soln->signature, depth);
    KheDrsDomTestDominatingRange(dom_test, val, &from_val, &to_val);
    if( to_val >= HaArrayCount(soln_trie->children) )
      to_val = HaArrayCount(soln_trie->children) - 1;
    for( i = from_val;  i <= to_val;  i++ )
      if( KheDrsSolnTrieDominates(HaArray(soln_trie->children,i), soln, depth+1, drs))
	return true;
    return false;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieRemoveDominated(KHE_DRS_SOLN_TRIE *soln_trie,         */
/*    KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs,         */
/*    int *total_count)                                                      */
/*                                                                           */
/*  Remove those elements of *soln_trie that are dominated by soln.  This    */
/*  can change *soln_trie.                                                   */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieRemoveDominated(KHE_DRS_SOLN_TRIE *soln_trie,
  KHE_DRS_SOLN soln, int depth, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int *total_count)
{
  int val, from_val, to_val, i;  KHE_DRS_DOM_TEST dom_test;
  if( *soln_trie == NULL )
  {
    /* case (1); there is nothing to remove */
  }
  else if( (*soln_trie)->soln != NULL )
  {
    /* case (2); if (*soln_trie)->soln is dominated, remove whole trie */
    if( KheDrsSolnDominates(soln, (*soln_trie)->soln, depth) )
    {
      KheDrsSolnTrieFree(soln_trie, drs);
      (*total_count)--;
    }
  }
  else
  {
    /* case (3); removed dominated solns among the children */
    dom_test = HaArray(soln->day->dom_tests, depth);
    val = HaArray(soln->signature, depth);
    KheDrsDomTestDominatedRange(dom_test, val, &from_val, &to_val);
    if( to_val >= HaArrayCount((*soln_trie)->children) )
      to_val = HaArrayCount((*soln_trie)->children) - 1;
    for( i = from_val;  i <= to_val;  i++ )
      KheDrsSolnTrieRemoveDominated(&HaArray((*soln_trie)->children, i), soln,
	depth + 1, drs, total_count);

    /* remove trailing NULL children */
    while( HaArrayCount((*soln_trie)->children) > 0 &&
	HaArrayLast((*soln_trie)->children) == NULL )
      HaArrayDeleteLast((*soln_trie)->children);

    /* if children array is empty then the whole trie is empty, so delete it */
    if( HaArrayCount((*soln_trie)->children) == 0 )
      KheDrsSolnTrieFree(soln_trie, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieDoStrongDominanceTest(KHE_DRS_SOLN_TRIE *soln_trie,   */
/*    KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    int *total_count)                                                      */
/*                                                                           */
/*  Carry out a strong dominance test between *soln_trie and soln.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieDoStrongDominanceTest(KHE_DRS_SOLN_TRIE *soln_trie,
  KHE_DRS_SOLN soln, KHE_DYNAMIC_RESOURCE_SOLVER drs, int *total_count)
{
  if( DEBUG12 )
    fprintf(stderr, "[ KheDrsSolnTrieDoStrongDominanceTest(soln for %s)\n",
      KheTimeGroupId(soln->day->time_group));
  if( KheDrsSolnTrieDominates(*soln_trie, soln, 0, drs) )
  {
    /* soln is dominated by some other soln, so delete soln */
    if( DEBUG12 )
      fprintf(stderr, "  trie deleting soln for %s\n",
	KheTimeGroupId(soln->day->time_group));
    KheDrsSolnFree(soln, drs);
  }
  else
  {
    /* remove solns that soln dominates, and insert soln */
    if( DEBUG12 )
      fprintf(stderr, "  trie inserting soln for %s\n",
	KheTimeGroupId(soln->day->time_group));
    KheDrsSolnTrieRemoveDominated(soln_trie, soln, 0, drs, total_count);
    KheDrsSolnTrieInsert(soln_trie, soln, 0, drs);
    (*total_count)++;
  }
  if( DEBUG12 )
    fprintf(stderr, "] KheDrsSolnTrieDoStrongDominanceTest returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieExtend(KHE_DRS_SOLN_TRIE soln_trie,                   */
/*    int day_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)                        */
/*                                                                           */
/*  Extend the solutions in soln_trie to next_day.                           */
/*                                                                           */
/*****************************************************************************/
static void KheDrsSolnExtend(KHE_DRS_SOLN soln, KHE_DRS_DAY new_day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs);

static void KheDrsSolnTrieExtend(KHE_DRS_SOLN_TRIE soln_trie,
  KHE_DRS_DAY new_day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN_TRIE child_trie;  int i;
  if( soln_trie == NULL )
  {
    /* case (1); do nothing */
  }
  else if( soln_trie->soln != NULL )
  {
    /* case (2); extend soln_trie->soln */
    KheDrsSolnExtend(soln_trie->soln, new_day, drs);
  }
  else
  {
    /* case (3); handle children */
    HaArrayForEach(soln_trie->children, child_trie, i)
      KheDrsSolnTrieExtend(child_trie, new_day, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnTrieExtractFinalSoln(KHE_DRS_SOLN_TRIE soln_trie,         */
/*    KHE_DRS_SOLN *final_soln)                                              */
/*                                                                           */
/*  Extract the unique solution (if any) from soln_trie.                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnTrieExtractFinalSoln(KHE_DRS_SOLN_TRIE soln_trie,
  KHE_DRS_SOLN *final_soln)
{
  KHE_DRS_SOLN_TRIE child_trie;  int i;
  if( soln_trie == NULL )
  {
    /* case (1); do nothing */
  }
  else if( soln_trie->soln != NULL )
  {
    /* case (2); extract soln_trie->soln */
    HnAssert(*final_soln == NULL,
      "KheDrsSolnTrieExtractFinalSoln internal error 1");
    *final_soln = soln_trie->soln;
  }
  else
  {
    /* case (3); handle children */
    HaArrayForEach(soln_trie->children, child_trie, i)
      KheDrsSolnTrieExtractFinalSoln(child_trie, final_soln);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_SOLN_SET"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN_SET KheDrsSolnSetMake(KHE_TIME_GROUP tg,                    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return a new KHE_DRS_SOLN_SET object.                                    */
/*                                                                           */
/*****************************************************************************/

KHE_DRS_SOLN_SET KheDrsSolnSetMake(/* KHE_TIME_GROUP tg, */
  KHE_DRS_DOM_KIND dom_kind, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN_SET res;
  if( HaArrayCount(drs->free_soln_set) > 0 )
    res = HaArrayLastAndDelete(drs->free_soln_set);
  else
    HaMake(res, drs->arena);
  /* res->time_group = tg; */
  res->count = 0;
  res->dom_kind = dom_kind;
  switch( dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      res->u.none_dom_set = KheDrsSolnListMake(drs);
      break;

    case KHE_DRS_DOM_WEAK:

      HpTableInit(res->u.weak_dom_set,
	&KheDrsSolnSignatureFullHashUntyped,
	&KheDrsSolnSignatureFullEqualUntyped,
	&KheDrsSolnSignatureDebugUntyped, drs->arena);
      break;

    case KHE_DRS_DOM_MEDIUM:

      HpTableInit(res->u.medium_dom_set,
	&KheDrsSolnListSignaturePartialHashUntyped,
	&KheDrsSolnListSignaturePartialEqualUntyped,
	&KheDrsSolnListSignatureDebugUntyped, drs->arena);
      break;

    case KHE_DRS_DOM_STRONG:

      res->u.strong_dom_set = KheDrsSolnListMake(drs);
      break;

    case KHE_DRS_DOM_TRIE:

      res->u.trie_dom_set = NULL;  /* represents the empty set */
      break;

    default:

      HnAbort("KheDrsSolnSetMake internal error: unknown dom_kind (%d)\n",
	dom_kind);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetFree(KHE_DRS_SOLN_SET dns,                             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Free dns and everything in it.                                           */
/*                                                                           */
/*  Most of the objects go onto free lists in drs for re-use later.  The     */
/*  exceptions are HpTable objects; they return to the arena they came from. */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetFree(KHE_DRS_SOLN_SET dns,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int pos;  KHE_DRS_SOLN soln;  KHE_DRS_SOLN_LIST ns;
  switch( dns->dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      KheDrsSolnListFree(dns->u.none_dom_set, drs);
      break;

    case KHE_DRS_DOM_WEAK:

      HpTableForEachValue(dns->u.weak_dom_set, soln, pos)
	KheDrsSolnFree(soln, drs);
      HpTableFree(dns->u.weak_dom_set);
      break;

    case KHE_DRS_DOM_MEDIUM:

      HpTableForEachValue(dns->u.medium_dom_set, ns, pos)
	KheDrsSolnListFree(ns, drs);
      HpTableFree(dns->u.medium_dom_set);
      break;

    case KHE_DRS_DOM_STRONG:

      KheDrsSolnListFree(dns->u.strong_dom_set, drs);
      /* ***
      HaArrayForEach(dns->u.strong_dom_set, ns, pos)
	KheDrsSolnListFree(ns, drs);
      HaArrayFree(dns->u.strong_dom_set);
      *** */
      break;

    case KHE_DRS_DOM_TRIE:

      KheDrsSolnTrieFree(&dns->u.trie_dom_set, drs);
      HnAssert(dns->u.trie_dom_set == NULL,
	"KheDrsSolnSetClear internal error");
      break;

    default:

      HnAbort("KheDrsSolnSetClear internal error: unknown dom_kind (%d)\n",
	dns->dom_kind);
  }
  HaArrayAddLast(drs->free_soln_set, dns);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetClear(KHE_DRS_SOLN_SET dns,                            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Remove and delete the solns and soln_sets held in dns.                   */
/*                                                                           */
/*****************************************************************************/

/* *** no longer used
static void KheDrsSolnSetClear(KHE_DRS_SOLN_SET dns,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int pos;  KHE_DRS_SOLN soln;  KHE_DRS_SOLN_LIST ns;
  switch( dns->dom_kind )
  {
    case KHE_DRS_DOM_WEAK:

      HpTableForEachValue(dns->u.weak_dom_set, soln, pos)
	KheDrsSolnFree(soln, drs);
      HpTableClear(dns->u.weak_dom_set);
      break;

    case KHE_DRS_DOM_MEDIUM:

      HpTableForEachValue(dns->u.medium_dom_set, ns, pos)
	KheDrsSolnListFree(ns, drs);
      HpTableClear(dns->u.medium_dom_set);
      break;

    case KHE_DRS_DOM_STRONG:

      KheDrsSolnListFree(dns->u.strong_dom_set, drs);
      ** ***
      HaArrayForEach(dns->u.strong_dom_set, ns, pos)
	KheDrsSolnListFree(ns, drs);
      HaArrayClear(dns->u.strong_dom_set);
      *** **
      break;

    case KHE_DRS_DOM_TRIE:

      KheDrsSolnTrieFree(&dns->u.trie_dom_set, drs);
      HnAssert(dns->u.trie_dom_set == NULL,
	"KheDrsSolnSetClear internal error");
      break;

    default:

      HnAbort("KheDrsSolnSetClear internal error: unknown dom_kind (%d)\n",
	dns->dom_kind);
  }
  dns->count = 0;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetAddSoln(KHE_DRS_SOLN_SET dns, KHE_DRS_SOLN soln,       */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add soln to dns.                                                         */
/*                                                                           */
/*  Strong dominance                                                         */
/*  ----------------                                                         */
/*                                                                           */
/*  For each soln other_soln in dns, if soln is dominated by other_soln,     */
/*  then delete soln and return.                                             */
/*                                                                           */
/*  Otherwise, delete each other_soln that soln dominates, then add soln.    */
/*                                                                           */
/*  Weak dominance                                                           */
/*  --------------                                                           */
/*                                                                           */
/*  If dns does not already contain a soln with the same signature as soln,  */
/*  just add soln to dns and return.                                         */
/*                                                                           */
/*  If dns already contains a soln (call it other_soln) with the same        */
/*  signature as soln, keep whichever one of soln and other_soln has the     */
/*  smallest cost, by overwriting other_soln's fields with soln's fields,    */
/*  if required.                                                             */
/*                                                                           */
/*  If soln and other_soln have equal cost, keep one chosen at random, with  */
/*  weights which ensure that if there are K solns with the same signature   */
/*  and the same cost, then each has probability 1/K of being the one kept.  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetAddSoln(KHE_DRS_SOLN_SET dns, KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN other_soln;  /* int asst_count; */
  KHE_DRS_SOLN_LIST soln_set, other_soln_set;
  switch( dns->dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      /* no dominance testing, just add soln */
      HaArrayAddLast(dns->u.none_dom_set->solns, soln);
      break;

    case KHE_DRS_DOM_WEAK:

      /* do equivalence testing */
      if( HpTableAddUnique(dns->u.weak_dom_set, (void *) soln, soln,other_soln))
	dns->count++;
      else
      {
	/* other_soln will be best of one more soln than before */
	other_soln->best_of++;

	/* ensure other_soln is the better of other_soln and soln */
	if( soln->cost < other_soln->cost || (soln->cost == other_soln->cost &&
	      drs->solve_diversifier % other_soln->best_of == 0) )
	{
	  /* soln is better than other_soln, so overwrite with soln */
	  KheDrsSolnOverWrite(other_soln, soln);
	}

	/* free soln */
	KheDrsSolnFree(soln, drs);
      }
      break;

    case KHE_DRS_DOM_MEDIUM:

      /* make soln_set holding just soln and try adding it to medium_dom_set */
      soln_set = KheDrsSolnListMake(drs);
      HaArrayAddLast(soln_set->solns, soln);
      if( HpTableAddUnique(dns->u.medium_dom_set, (void *) soln_set, soln_set,
	    other_soln_set) )
      {
	/* soln_set added, so that's all that's needed */
	dns->count++;
      }
      else
      {
	/* remove soln_set and do strong dom. of soln with other_soln_set */
	HaArrayClear(soln_set->solns);
	KheDrsSolnListFree(soln_set, drs);
        KheDrsSolnListDoStrongDominanceTest(other_soln_set, soln, drs,
	  &dns->count);
      }
      break;

    case KHE_DRS_DOM_STRONG:
    
      /* find or make soln_set, the KHE_DRS_SOLN_LIST which should hold soln */
      /* ***
      asst_count = HaArrayFirst(soln->signature);
      while( HaArrayCount(dns->u.strong_dom_set) <= asst_count )
      {
	soln_set = KheDrsSolnListMake(drs);
	HaArrayAddLast(dns->u.strong_dom_set, soln_set);
      }
      soln_set = HaArray(dns->u.strong_dom_set, asst_count);
      *** */

      /* carry out strong dominance test between soln_set and soln */
      KheDrsSolnListDoStrongDominanceTest(dns->u.strong_dom_set, soln,
	drs, &dns->count);
      break;

    case KHE_DRS_DOM_TRIE:
    
      KheDrsSolnTrieDoStrongDominanceTest(&dns->u.trie_dom_set, soln, drs,
	&dns->count);
      break;
    
    default:

      HnAbort("KheDrsSolnSetAddSoln internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetExtend(KHE_DRS_SOLN_SET dns, KHE_DRS_DAY new_day,      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Extend dns by adding tasks from new_day.                                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetExtend(KHE_DRS_SOLN_SET dns, KHE_DRS_DAY new_day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN soln;  int i, j;  KHE_DRS_SOLN_LIST soln_set;
  switch( dns->dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      HaArrayForEach(dns->u.none_dom_set->solns, soln, j)
	KheDrsSolnExtend(soln, new_day, drs);
      break;

    case KHE_DRS_DOM_WEAK:

      if( DEBUG9 )
      {
	float probes;
	probes = HpTableProbeLength(dns->u.weak_dom_set);
	fprintf(stderr, "  hash table: count %d, size %d, probes %.2f\n",
	  /* KheTimeGroupId(dns->time_group), */ dns->count,
	  HpTableSize(dns->u.weak_dom_set), probes);
	if( probes > 14.0 )
	  HpTableDebug(dns->u.weak_dom_set, 2, stderr);
      }
      HpTableForEachValue(dns->u.weak_dom_set, soln, i)
	KheDrsSolnExtend(soln, new_day, drs);
      break;

    case KHE_DRS_DOM_MEDIUM:

      HpTableForEachValue(dns->u.medium_dom_set, soln_set, i)
	HaArrayForEach(soln_set->solns, soln, j)
	  KheDrsSolnExtend(soln, new_day, drs);
      break;

    case KHE_DRS_DOM_STRONG:

      HaArrayForEach(dns->u.strong_dom_set->solns, soln, j)
	KheDrsSolnExtend(soln, new_day, drs);
      /* ***
      HaArrayForEach(dns->u.strong_dom_set, soln_set, i)
	HaArrayForEach(soln_set->solns, soln, j)
	  KheDrsSolnExtend(soln, day, drs);
      *** */
      break;

    case KHE_DRS_DOM_TRIE:
    
      KheDrsSolnTrieExtend(dns->u.trie_dom_set, new_day, drs);
      break;
    
    default:

      HnAbort("KheDrsSolnSetExtend internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolnSetExtractFinalSoln(KHE_DRS_SOLN_SET dns,                 */
/*    KHE_DRS_SOLN *final_soln)                                              */
/*                                                                           */
/*  Extract the unique final solution (if any) from dns.                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolnSetExtractFinalSoln(KHE_DRS_SOLN_SET dns,
  KHE_DRS_SOLN *final_soln)
{
  KHE_DRS_SOLN soln;  int i, j;  KHE_DRS_SOLN_LIST soln_set;
  switch( dns->dom_kind )
  {
    case KHE_DRS_DOM_NONE:

      HaArrayForEach(dns->u.none_dom_set->solns, soln, j)
      {
	HnAssert(*final_soln == NULL,
	  "KheDrsSolnSetExtractFinalSoln internal error 3");
	*final_soln = soln;
      }
      break;

    case KHE_DRS_DOM_WEAK:

      HpTableForEachValue(dns->u.weak_dom_set, soln, i)
      {
	HnAssert(*final_soln == NULL,
	  "KheDrsSolnSetExtractFinalSoln internal error 1");
	*final_soln = soln;
      }
      break;

    case KHE_DRS_DOM_MEDIUM:

      HpTableForEachValue(dns->u.medium_dom_set, soln_set, i)
	HaArrayForEach(soln_set->solns, soln, j)
	{
	  HnAssert(*final_soln == NULL,
	    "KheDrsSolnSetExtractFinalSoln internal error 2");
	  *final_soln = soln;
	}
      break;

    case KHE_DRS_DOM_STRONG:

      HaArrayForEach(dns->u.strong_dom_set->solns, soln, j)
      {
	HnAssert(*final_soln == NULL,
	  "KheDrsSolnSetExtractFinalSoln internal error 3");
	*final_soln = soln;
      }
      /* ***
      HaArrayForEach(dns->u.strong_dom_set, soln_set, i)
	HaArrayForEach(soln_set->solns, soln, j)
	  HaArrayAddLast(drs->solve_final_solns, soln);
      *** */
      break;

    case KHE_DRS_DOM_TRIE:
    
      KheDrsSolnTrieExtractFinalSoln(dns->u.trie_dom_set, final_soln);
      break;
    
    default:

      HnAbort("KheDrsSolnSetExtractFinalSoln internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_DAY"                                                  */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayMake(int frame_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)     */
/*                                                                           */
/*  Make a new day object and add it to drs->all_days.  Just make the basic  */
/*  object; the task classes, expressions, and dominance tests will be       */
/*  added later.                                                             */
/*                                                                           */
/*  Here frame_index is the index of the new day in the common frame.        */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayMake(int frame_index, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_DAY res;
  /* ***
  if( HaArrayCount(drs->free_day) > 0 )
  {
    res = HaArrayLastAndDelete(drs->free_day);
    HaArrayClear(res->task_classes);
    HaArrayClear(res->internal_today);
    HaArrayClear(res->dom_tests);
    HaArrayClear(res->eq_dom_test_indexes);
  }
  else
  {
    HaMake(res, drs->arena);
    HaArrayInit(res->task_classes, drs->arena);
    HaArrayInit(res->internal_today, drs->arena);
    HaArrayInit(res->dom_tests, drs->arena);
    HaArrayInit(res->eq_dom_test_indexes, drs->arena);
  }
  *** */
  HaMake(res, drs->arena);
  HaArrayInit(res->task_classes, drs->arena);
  HaArrayInit(res->internal_today, drs->arena);
  HaArrayInit(res->dom_tests, drs->arena);
  HaArrayInit(res->eq_dom_test_indexes, drs->arena);
  res->frame_index = frame_index;
  res->open_day_index = -1;  /* signals not currently open for solving */
  /* res->day_range = ddr; */
  res->time_group = KheFrameTimeGroup(drs->days_frame, frame_index);
  /* res->sig_len = 0; */
  res->soln_set = NULL;  /* will be initialized at the start of the solve */
  /* res->soln_set = KheDrsSolnSetMake(dom_kind, drs); */

  /* add the new day to drs->all_days */
  HaArrayAddLast(drs->all_days, res);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayAddTaskClasses(KHE_DRS_DAY day,                            */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add task classes to day.                                                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsDayAddTaskClasses(KHE_DRS_DAY day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_TIME t;  int i, j, tc_count ** , last_solve_index **;
  KHE_TASK_CLASS tc;  KHE_DRS_TASK_CLASS dtc;
  for( i = 0;  i < KheTimeGroupTimeCount(day->time_group);  i++ )
  {
    t = KheTimeGroupTime(day->time_group, i);
    tc_count = KheTaskClassSolverClassAtTimeCount(drs->task_class_solver, t);
    for( j = 0;  j < tc_count;  j++ )
    {
      tc = KheTaskClassSolverClassAtTime(drs->task_class_solver, t, j);
      dtc = KheDrsTaskCl assMake(tc, drs);
      HaArrayAddLast(day->task_classes, dtc);
      ** ***
      if( KheTaskClassIsOpen(tc, drs->selected_resource_set, day->day_range) )
      {
	last_solve_index = day->solve_index +
	  KheTaskClassLastDayIndex(tc) - KheTaskClassFirstDayIndex(tc);
	dtc = KheDrsTaskCl assMake(tc, day->solve_index, last_solve_index, drs);
	HaArrayAddLast(day->task_classes, dtc);
      }
      *** **
    }
  }
}
*** */

  /* ***
  HaArrayInit(res->times, drs->arena);
  for( i = 0;  i < KheTimeGroupTimeCount(res->time_group);  i++ )
  {
    t = KheTimeGroupTime(res->time_group, i);
    stime = KheDrsTimeMake(t, drs);
    HaArrayAddLast(res->times, stime);
  }
  *** */
  /* res->solns_count = 0; */
  /* ***
  res->sig_len = 1;  ** asst_count is always at start of signature **
  HaArrayInit(res->dom_tests, drs->arena);
  HaArrayInit(res->eq_dom_test_indexes, drs->arena);
  KheDrsDayAddDomTest(res, KheDrsDomTestMake(KHE_DRS_DOM_EQ));
  *** */
  /* ***
  HaArrayInit(res->internal_today, drs->arena);
  HaArrayInit(res->external_today, drs->arena);
  *** */
  /* adding classes (based on orig classes) sti ll to do */
  /* res->free_task = NULL; */
  /* res->free_task = KheDrsTaskMakeFree(day_index, drs); */
  /* return res; */
/* } */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayFree(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)     */
/*                                                                           */
/*  Free day.                                                                */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsDayFree(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KheDrsSolnSetFree(day->soln_set, drs);
  HaArrayAddLast(drs->free_day, day);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayOpen(KHE_DRS_DAY day, KHE_DRS_DAY_RANGE ddr,               */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, int open_day_index)                   */
/*                                                                           */
/*  Open day for solving.  It lies within day range ddr, and its index in    */
/*  the sequence of all open days is open_day_index.                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayOpen(KHE_DRS_DAY day, KHE_DRS_DAY_RANGE ddr,
  KHE_DRS_DOM_KIND dom_kind, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int open_day_index)
{
  KHE_DRS_TASK_CLASS dtc;  int i;
  day->open_day_index = open_day_index;
  HaArrayForEach(day->task_classes, dtc, i)
    KheDrsTaskClassOpen(dtc, ddr, drs);
  HaArrayClear(day->internal_today);
  HaArrayClear(day->dom_tests);
  HaArrayClear(day->eq_dom_test_indexes);
  day->soln_set = KheDrsSolnSetMake(dom_kind, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayAddOpenExpr(KHE_DRS_DAY day, KHE_DRS_EXPR e)               */
/*                                                                           */
/*  Add open expression e to day.                                            */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayAddOpenExpr(KHE_DRS_DAY day, KHE_DRS_EXPR e)
{
  HaArrayAddLast(day->internal_today, e);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsDayAddDomTest(KHE_DRS_DAY day, KHE_DRS_DOM_TEST dom_test)      */
/*                                                                           */
/*  Add dom_test to day and return its index in day->dom_tests.              */
/*                                                                           */
/*****************************************************************************/

static int KheDrsDayAddDomTest(KHE_DRS_DAY day, KHE_DRS_DOM_TEST dom_test)
{
  int res;

  /* add dom_test to day */
  res = HaArrayCount(day->dom_tests);
  HnAssert(dom_test.type != KHE_DRS_DOM_UNUSED,
    "KheDrsDayAddDomTest internal error");
  HaArrayAddLast(day->dom_tests, dom_test);

  /* if dom_test is equality, record it in eq_dom_test_indexes */
  if( dom_test.type == KHE_DRS_DOM_EQ || dom_test.type == KHE_DRS_DOM_EQ_LOOSE )
    HaArrayAddLast(day->eq_dom_test_indexes, res);

  /* return dom_test's index */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayClose(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Close day.                                                               */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayClose(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_CLASS dtc;  int i;
  day->open_day_index = -1;  /* signals no longer open */
  HaArrayForEach(day->task_classes, dtc, i)
    KheDrsTaskClassClose(dtc);
  /* ***
  HaArrayForEach(day->internal_today, e, i)
    KheDrsInternalExprClose(e, drs);
  *** */
  HaArrayClear(day->internal_today);
  HaArrayClear(day->dom_tests);
  HaArrayClear(day->eq_dom_test_indexes);
  KheDrsSolnSetFree(day->soln_set, drs);
  day->soln_set = NULL;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheTaskClassIsOpen(KHE_TASK_CLASS tc, KHE_RESOURCE_SET rs,          */
/*    KHE_DRS_DAY_RANGE ddr)                                                 */
/*                                                                           */
/*  Return true if tc is open, because its domain has a non-empty            */
/*  intersection with rs, and its range of days lies entirely within ddr.    */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheTaskClassIsOpen(KHE_TASK_CLASS tc, KHE_RESOURCE_SET rs,
  KHE_DRS_DAY_RANGE ddr)
{
  HnAssert(KheTaskClassFirstDayIndex(tc) <= KheTaskClassLastDayIndex(tc),
    "KheTaskClassIsOpen:  tc has empty day range");
  return KheTaskClassNoOverlap(tc) &&
    KheTaskClassFirstDayIndex(tc) >= ddr.first_day_index &&
    KheTaskClassLastDayIndex(tc) <= ddr.last_day_index &&
    !KheResourceSetDisjointGroup(rs, KheTaskClassDomain(tc));
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DAY KheDrsDayNext(KHE_DRS_DAY day,                               */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return the solve day after day, aborting if none.                        */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_DAY KheDrsDayNext(KHE_DRS_DAY day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HnAssert(day->solve_index < HaArrayCount(drs->open_days) - 1,
    "KheDrsDayNext internal error");
  return HaArray(drs->open_days, day->solve_index + 1);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayAddSoln(KHE_DRS_DAY day, KHE_DRS_SOLN soln,                */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add soln to day, or do nothing if soln has a hard constraint violation   */
/*  or contains too few or too many assignments.  Otherwise what to do       */
/*  depends on whether we are using strong dominance or weak dominance.      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayAddSoln(KHE_DRS_DAY day, KHE_DRS_SOLN soln,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  /* ***
  ** KHE_DRS_SOLN other_soln; **  int asst_count ** , i **;
  asst_count = HaArrayFirst(soln->signature);
  if( asst_count + HaArrayCount(drs->days) - day->day_index < drs->min_assts ||
      asst_count > drs->max_assts || soln->cost > drs->cost_limit )
  {
    ** do nothing except free soln; it is unacceptable **
    KheDrsSolnFree(soln, drs);
  }
  else
  *** */

  KheDrsSolnSetAddSoln(day->soln_set, soln, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK_GROUP KheDrsDayAddTask(KHE_DRS_DAY day, KHE_DRS_TASK dt)    */
/*                                                                           */
/*  Add dt to day, and return the task group it was added to.                */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_TASK_GROUP KheDrsDayAddTask(KHE_DRS_DAY day, KHE_DRS_TASK dt)
{
  KHE_DRS_TASK_GROUP sg;  int i;
  HaArrayForEach(day->task_groups, sg, i)
    if( KheDrsTaskGroupAccepts(sg, dt) )
    {
      KheDrsTaskGroupAddTask(sg, dt);
      return sg;
    }
  sg = KheDrsTaskGroupMake(dt, HaArrayArena(day->task_groups));
  HaArrayAddLast(day->task_groups, sg);
  return sg;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayClearAvail(KHE_DRS_DAY day)                                */
/*                                                                           */
/*  Clear away all the avail tasks on day.                                   */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsDayClearAvail(KHE_DRS_DAY day)
{
  KHE_DRS_TASK_GROUP dtg;  int i;
  HaArrayForEach(day->task_groups, dtg, i)
    KheDrsTaskGroupClearAvail(dtg);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayMakeUnboundTasksAvail(KHE_DRS_DAY day)                     */
/*                                                                           */
/*  Make the unbound tasks in the task groups of day available.              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsDayMakeUnboundTasksAvail(KHE_DRS_DAY day)
{
  KHE_DRS_TASK_GROUP dtg;  int i;
  HaArrayForEach(day->task_groups, dtg, i)
    KheDrsTaskGroupMakeUnboundTasksAvail(dtg);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayClearTasks(KHE_DRS_DAY day,                                */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Clear out all the tasks running on day, except the free task.            */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsDayClearTasks(KHE_DRS_DAY day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i;  KHE_DRS_TIME t;
  HaArrayForEach(day->times, t, i)
    KheDrsTimeClearTasks(t, drs);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayInitializeForSolving(KHE_DRS_DAY day,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Initialize day for solving.                                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsDayInitializeForSolving(KHE_DRS_DAY day,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_TASK_CLASS dtc;  KHE_TASK_CLASS tc;  int i, j;
  KHE_TASK_CLASS_SOLVER tcs;  KHE_TIME time;

  ** delete old task classes **
  HaArrayForEach(day->task_classes, dtc, i)
    KheDrsTaskClassFree(dtc, drs);
  HaArrayClear(day->task_classes);

  ** make one drs task class for each task class on that day **
  tcs = drs->task_class_solver;
  for( i = 0;  i < KheTimeGroupTimeCount(day->time_group);  i++ )
  {
    time = KheTimeGroupTime(day->time_group, i);
    for( j = 0;  j < KheTaskClassSolverClassAtTimeCount(tcs, time);  j++ )
    {
      tc = KheTaskClassSolverClassAtTime(tcs, time, i);
      dtc = KheDrsTaskC lassMake(tc, drs);
      HaArrayAddLast(day->task_classes, dtc);
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayClear(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Clear day, ready for a fresh solve.                                      */
/*                                                                           */
/*  Implementation note.  The expressions in external_today will be deleted  */
/*  separately, as part of KheDynamicResourceSolverClear.                    */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsDayClear(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ** KheDrsDayClearTasks(day, drs); **
  ** ***
  day->sig_len = 1;
  HaArrayClear(day->dom_tests);
  HaArrayClear(day->eq_dom_test_indexes);
  KheDrsDayAddDomTest(day, KheDrsDomTestMake(KHE_DRS_DOM_EQ));
  HaArrayClear(day->internal_today);
  HaArrayClear(day->external_today);
  *** **
  KheDrsSolnSetClear(day->soln_set, drs);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsDayAddTasks(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs, */
/*    KHE_DRS_TASK *preassigned_dt)                                          */
/*                                                                           */
/*  Add suitable tasks to day.                                               */
/*                                                                           */
/*  If true is returned, it means that a preassigned task was found; a task  */
/*  for it will be the only one added.  It is returned in *preassigned_dt.   */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete
static bool KheDrsDayAddTasks(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_DRS_TASK *preassigned_dt)
{
  int i, j, k;  KHE_TIME t;  KHE_TASK task;  KHE_DRS_TIME stime;
  KHE_DRS_TASK dt;  KHE_MEET meet;
  HaArrayForEach(day->times, stime, i)
  {
    ** for each time of day **
    t = stime->time;
    for( j = 0;  j < KheEventTimetableMonitorTimeMeetCount(drs->etm, t);  j++ )
    {
      ** for each meet running at time **
      meet = KheEventTimetableMonitorTimeMeet(drs->etm, stime->time, j);
      for( k = 0;  k < KheMeetTaskCount(meet);  k++ )
      {
	** for the proper root task of each task of meet **
	task = KheTaskProperRoot(KheMeetTask(meet, k));
	if( KheTaskAsstResource(task) == drs->resource )
	{
	  ** task is already assigned, so it is the only legal choice **
	  KheDrsDayClearTasks(day, drs);
	  dt = KheDrsTa skMake(task, drs);
	  HaArrayAddLast(stime->best_tasks, dt);
	  *preassigned_dt = dt;
	  return true;
	}
	else if( KheTaskAssignResourceCheck(task, drs->resource) )
	{
	  dt = KheDrsTas kMake(task, drs);
	  if( KheDrsTaskFirstTime(dt) == t )
	  {
	    ** add dt to stime, possibly replacing an existing task **
	    KheDrsTimeAddTask(stime, dt, drs);
	  }
	  else
	  {
	    ** dt starts at the wrong time, so ignore it **
	    KheDrsTaskFree(dt, drs);
	  }
	}
      }
    }
  }
  *preassigned_dt = NULL;
  return false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayAddDomTest(KHE_DRS_DAY day, KHE_DRS_DOM_TEST dom_test)     */
/*                                                                           */
/*  Add dom_test to day.  This may change the eq_dom_test_indexes array as   */
/*  well as the dom_tests array.                                             */
/*                                                                           */
/*****************************************************************************/

/* *** moved to KheDrsExprOpen
static void KheDrsDayAddDomTest(KHE_DRS_DAY day, KHE_DRS_DOM_TEST dom_test)
{
  if( dom_test.type == KHE_DRS_DOM_EQ || dom_test.type == KHE_DRS_DOM_EQ_LOOSE )
    HaArrayAddLast(day->eq_dom_test_indexes, HaArrayCount(day->dom_tests));
  HaArrayAddLast(day->dom_tests, dom_test);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayDomTestsDebug(KHE_DRS_DAY day, int verbosity,              */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of the dom tests of day.                                     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayDomTestsDebug(KHE_DRS_DAY day, int verbosity,
  int indent, FILE *fp)
{
  int dom_test_freq[KHE_DRS_DOM_EQ_LOOSE + 1];  int i;
  KHE_DRS_DOM_TEST dom_test;  char buff[100];

  /* initialize dom_test_freq */
  for( i = KHE_DRS_DOM_UNUSED;  i <= KHE_DRS_DOM_EQ_LOOSE;  i++ )
    dom_test_freq[i] = 0;

  /* accumulate dom_test_freq */
  HaArrayForEach(day->dom_tests, dom_test, i)
    dom_test_freq[dom_test.type]++;

  /* display frequencies */
  if( indent >= 0 )
    fprintf(fp, "%*s", indent, "");
  fprintf(fp, "[ Day %s dom tests: ", KheTimeGroupId(day->time_group));
  for( i = 0;  i <= KHE_DRS_DOM_EQ_LOOSE;  i++ )
  {
    if( i > 0 )
      fprintf(fp, ", ");
    fprintf(fp, "%d %s", dom_test_freq[i],
      KheDrsDomTestShow(KheDrsDomTestMake(i), buff));
  }
  fprintf(fp, " ]");
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsDayDebug(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs,    */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of day onto fp.                                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsDayDebug(KHE_DRS_DAY day, KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  /* int i, pos */;  /* KHE_DRS_TIME stime; */
  /* KHE_DRS_DOM_TEST dom_test; */  /* char buff[100]; */
  if( indent >= 0 )
  {
    fprintf(fp, "%*s[ Day ", indent, "");
    KheTimeGroupDebug(day->time_group, 1, 0, stderr);
    /* ***
    fprintf(fp, " (sig_len %d)\n", day->sig_len);
    fprintf(fp, " (times %d, sig_len %d)\n", HaArrayCount(day->times),
      day->sig_len);
    fprintf(fp, " (times %d, sig_len %d, internal_today %d)\n",
      HaArrayCount(day->times), day->sig_len, HaArrayCount(day->internal_today));
    HaArrayForEach(day->times, stime, i)
      KheDrsTimeDebug(stime, drs, verbosity, indent + 2, fp);
    fprintf(fp, "%*s  dom_tests [", indent, "");
    HaArrayForEach(day->dom_tests, dom_test, i)
      fprintf(fp, "%s%s%s", i > 0 ? ", " : "",
      HaArrayContains(day->eq_dom_test_indexes, i, &pos) ? "*" : "",
      KheDrsDomTestShow(dom_test, buff));
    *** */
    fprintf(fp, "]\n");
    fprintf(fp, "%*s]\n", indent, "");
  }
  else
    KheTimeGroupDebug(day->time_group, 1, -1, stderr);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DRS_FULL_SOLN"                                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_FULL_SOLN KheDrsFullSolnMake(KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Make a new, empty full soln object.                                      */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_FULL_SOLN KheDrsFullSolnMake(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_FULL_SOLN res;
  HaMake(res, drs->arena);
  res->asst_count = 0;
  HaArrayInit(res->tasks, drs->arena);
  res->resource = NULL;
  res->adopted = false;
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsFullSolnLoad(KHE_DRS_FULL_SOLN fs, KHE_DRS_SOLN soln)         */
/*                                                                           */
/*  Load soln into fs.                                                       */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsFullSolnLoad(KHE_DRS_FULL_SOLN fs, KHE_DRS_SOLN soln,
  KHE_RESOURCE r)
{
  HaArrayClear(fs->tasks);
  fs->resource = r;
  fs->asst_count = HaArrayFirst(soln->signature);
  while( soln->task != NULL )
  {
    HaArrayAddLast(fs->tasks, soln->task);
    soln = soln->prev_soln;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsFullSolnAdopt(KHE_DRS_FULL_SOLN fs)                           */
/*                                                                           */
/*  Adopt fs.                                                                */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsFullSolnAdopt(KHE_DRS_FULL_SOLN fs)
{
  int i;  KHE_DRS_TASK dt;
  HnAssert(!fs->adopted, "KheDrsFullSolnAdopt: already adopted");
  HaArrayForEach(fs->tasks, dt, i)
    if( dt->task != NULL && !dt->already_assigned )
    {
      if( !KheTaskAssignResource(dt->task, fs->resource) )
      {
	KheTaskDebug(dt->task, 1, 0, stderr);
	HnAbort("KheDrsFullSolnAdopt error (cannot assign task %d to %s)",
	  i, KheResourceId(fs->resource));
      }
      if( DEBUG6 )
      {
	fprintf(stderr, "  KheDrsFullSolnAdopt assigning %s to ",
	  KheResourceId(fs->resource));
	KheTaskDebug(dt->task, 1, 0, stderr);
      }
    }
  fs->adopted = true;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsFullSolnUnAdopt(KHE_DRS_FULL_SOLN fs)                         */
/*                                                                           */
/*  Undo the matching call to KheDrsFullSolnAdopt.                           */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsFullSolnUnAdopt(KHE_DRS_FULL_SOLN fs)
{
  int i;  KHE_DRS_TASK dt;
  HnAssert(fs->adopted, "KheDrsFullSolnUnAdopt: not adopted");
  HaArrayForEach(fs->tasks, dt, i)
    if( dt->task != NULL && !dt->already_assigned )
    {
      if( !KheTaskUnAssignResource(dt->task) )
	HnAbort("KheDrsFullSolnAdopt internal error");
    }
  fs->adopted = false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_SOLN KheFullSolnReRun(KHE_DRS_FULL_SOLN fs,                      */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Rerun fs, returning the resulting soln.                                  */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_DRS_SOLN KheFullSolnReRun(KHE_DRS_FULL_SOLN fs,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, di, first_day_index, last_day_index;
  KHE_DRS_SOLN soln;  KHE_TIME_GROUP tg;  KHE_DRS_TASK dt;
  HnAssert(!fs->adopted, "KheFullSolnReRun: already adopted");
  drs->solve_rerun = true;
  soln = KheDrsSolnMakeRoot(drs);
  HaArrayForEachReverse(fs->tasks, dt, i)
  {
    first_day_index = KheDailyScheduleFirstDayIndex(dt->daily_schedule);
    last_day_index = KheDailyScheduleLastDayIndex(dt->daily_schedule);
    for( di = first_day_index;  di <= last_day_index;  di++ )
    {
      tg = KheFrameTimeGroup(drs->days_frame, di);
      fprintf(stderr, "  starting day %s\n", KheTimeGroupId(tg));
      soln = KheDrsSolnMake(soln, dt, HaArray(drs->days, di), drs);
    }
  }
  drs->solve_rerun = false;
  return soln;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsFullSolnDebug(KHE_DRS_FULL_SOLN fs, int verbosity,            */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of fs onto fp.                                               */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsFullSolnDebug(KHE_DRS_FULL_SOLN fs, int verbosity,
  int indent, FILE *fp)
{
  int i;  KHE_DRS_TASK dt;
  fprintf(fp, "%*s[ Full Soln (%d assts%s) for %s:\n", indent, "",
    fs->asst_count, fs->adopted ? ", adopted" : "",KheResourceId(fs->resource));
  HaArrayForEachReverse(fs->tasks, dt, i)
    KheDrsTaskDebug(dt, verbosity, indent + 2, fp);
  fprintf(fp, "%*s]\n", indent, "");
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "converting tasks into task groups"                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverAddTasks(KHE_DYNAMIC_RESOURCE_SOLVER drs)   */
/*                                                                           */
/*  Add the tasks, in the form of tasks grouped by equivalent, to drs.       */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDynamicResourceSolverAddTasks(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_TASK task;  KHE_DRS_TASK dt;  int i, first_day_index;  KHE_DRS_DAY day;
  KHE_DRS_TASK_GROUP dtg;
  for( i = 0;  i < KheSolnTaskCount(drs->soln);  i++ )
  {
    task = KheSolnTask(drs->soln, i);
    if( KheTaskResourceType(task) == drs->resource_type &&
	!KheTaskIsCycleTask(task) && KheTaskAsst(task) == NULL )
    {
      dt = KheDrsTaskM ake(task, drs);
      first_day_index = KheDailyScheduleFirstDayIndex(dt->daily_schedule);
      day = HaArray(drs->days, first_day_index);
      dtg = KheDrsDayAddTask(day, dt);
      dt->task_group = dtg;
    }
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "event resource monitor handling"                              */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheTaskLoadEventResourceMonitors(KHE_TASK task,                     */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Load the event resource monitors that monitor dt.                        */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheTaskLoadEventResourceMonitors(KHE_TASK task,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_EVENT_RESOURCE er;  KHE_MONITOR m;  int i;  KHE_TASK child_task;

  ** do it for task **
  er = KheTaskEventResource(task);
  for( i = 0;  i < KheSolnEventResourceMonitorCount(drs->soln, er);  i++ )
  {
    m = KheSolnEventResourceMonitor(drs->soln, er, i);
    HaArrayAddLast(drs->er_monitors, m);
  }

  ** do it for task's proper descendants **
  for( i = 0;  i < KheTaskAssignedToCount(task);  i++ )
  {
    child_task = KheTaskAssignedTo(task, i);
    KheTaskLoadEventResourceMonitors(child_task, drs);
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsTaskAddToMonitorConverters(KHE_DRS_TASK dt,                   */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add dt to the monitor converters of the event resource monitors that     */
/*  monitor it.                                                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsTaskAddToMonitorConverters(KHE_DRS_TASK dt,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_EVENT_RESOURCE er;  KHE_MONITOR m;  int i, j;
  KHE_DRS_MONITOR_CONVERTER dmc;  KHE_DRS_TASK_ON_DAY dtd;
  HaArrayForEach(dt->days, dtd, i)
    if( dtd != NULL )
    {
      ** add dtd to the monitor converters that monitor dtd->task **
      er = KheTaskEventResource(dtd->task);
      for( j = 0;  j < KheSolnEventResourceMonitorCount(drs->soln, er);  j++ )
      {
	m = KheSolnEventResourceMonitor(drs->soln, er, j);
	dmc = (KHE_DRS_MONITOR_CONVERTER) KheMonitorBack(m);
	KheDrsMonitorConverterAddTaskOnDay(dmc, dtd);
      }
    }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheMonitorCmp(const void *t1, const void *t2)                        */
/*                                                                           */
/*  Comparison function for sorting (and uniqueifying) an array of monitors. */
/*                                                                           */
/*****************************************************************************/

static int KheMonitorCmp(const void *t1, const void *t2)
{
  KHE_MONITOR m1 = * (KHE_MONITOR *) t1;
  KHE_MONITOR m2 = * (KHE_MONITOR *) t2;
  return KheMonitorSolnIndex(m1) - KheMonitorSolnIndex(m2);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsLoadEventResourceMonitors(KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Load internal and external expressions representing those event          */
/*  monitors which monitor at least one selected task.                       */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheDrsLoadEventResourceMonitors(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_DAY day;  KHE_DRS_TASK_CLASS dtc;  KHE_DRS_TASK dt;  int i, j, k;
  KHE_MONITOR m;  KHE_DRS_MONITOR_CONVERTER dmc;
  ** ***
  KHE_DRS_EXPR_INT_SUM_COMB eisc;  KHE_CONSTRAINT c;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_weight;
  int min_limit, max_limit;
  KHE_LIMIT_RESOURCES_CONSTRAINT lrc;
  KHE_ASSIGN_RESOURCE_MONITOR arm;
  KHE_DRS_DOM_TEST alpha_dom_test, beta_dom_test;
  *** **

  ** load the event resource monitors that monitor selected tasks **
  HaArrayClear(drs->er_monitors);
  HaArrayForEach(drs->open_days, day, i)
    HaArrayForEach(day->task_classes, dtc, j)
      HaArrayForEach(dtc->tasks, dt, k)
        KheTaskLoadEventResourceMonitors(dt->task, drs);

  ** uniqueify the list of monitors **
  HaArraySortUnique(drs->er_monitors, &KheMonitorCmp);

  ** one monitor converter for each monitor **
  HaArrayClear(drs->er_monitor_converters);
  HaArrayForEach(drs->er_monitors, m, i)
  {
    dmc = KheDrsMonitorConverterMake(m, drs);
    HaArrayAddLast(drs->er_monitor_converters, dmc);
    KheMonitorSetBack(m, (void *) dmc);
  }

  ** add selected tasks to monitor converters **
  HaArrayForEach(drs->open_days, day, i)
    HaArrayForEach(day->task_classes, dtc, j)
      HaArrayForEach(dtc->tasks, dt, k)
        KheDrsTaskAddToMonitorConverters(dt, drs);

  ** use monitor converters to produce expressions, then delete them **
  HaArrayForEach(drs->er_monitors, m, i)
  {
    dmc = (KHE_DRS_MONITOR_CONVERTER) KheMonitorBack(m);
    KheDrsMonitorConverterMakeExpr(dmc, drs);
  }

  ** one cost-like expression for each monitor **
  ** *** doing this kind of stuff somewhere else now
  HaArrayForEach(drs->er_monitors, m, i)
  {
    c = KheMonitorConstraint(m);
    combined_weight = KheConstraintCombinedWeight(c);
    cost_fn = KheConstraintCostFunction(c);
    switch( KheMonitorTag(m) )
    {
      case KHE_ASSIGN_RESOURCE_MONITOR_TAG:

	arm = (KHE_ASSIGN_RESOURCE_MONITOR) m;
        min_limit = KheAssignResourceMonitorTotalMonitoredDuration(arm);
	eisc = KheDrsExprIntSumCombMake(m, cost_fn, combined_weight, min_limit,
	  INT_MAX, false, 0, 0, KheDrsDomTestMake(KHE_DRS_DOM_GE), drs);
	break;

      case KHE_PREFER_RESOURCES_MONITOR_TAG:

	eisc = KheDrsExprIntSumCombMake(m, cost_fn, combined_weight, 0, 0,
	  false, 0, 0, KheDrsDomTestMake(KHE_DRS_DOM_LE), drs);
	break;

      case KHE_AVOID_SPLIT_ASSIGNMENTS_MONITOR_TAG:

	** we're skipping this one **
	eisc = NULL;
	break;

      case KHE_LIMIT_RESOURCES_MONITOR_TAG:

	lrc = (KHE_LIMIT_RESOURCES_CONSTRAINT) c;
        min_limit = KheLimitResourcesConstraintMinimum(lrc);
	max_limit = KheLimitResourcesConstraintMaximum(lrc);
	KheLimitsToTests(min_limit, max_limit, INT_MAX, false,
	  &alpha_dom_test, &beta_dom_test);
	eisc = KheDrsExprIntSumCombMake(m, cost_fn, combined_weight,
	  min_limit, max_limit, false, 0, 0, alpha_dom_test, drs);
	break;

      default:

	HnAbort("KheDrsLoadEventResourceMonitors: unexpected monitor tag (%d)",
          KheMonitorTag(m));
	eisc = NULL;
	break;
    }
    KheMonitorSetBack(m, (void *) eisc);
  }

  ** remainder s  till to do **
  *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "conversion of event resource monitors into expression trees"  */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  int KheAssignResourceMonitorTotalMonitoredDuration(                      */
/*    KHE_ASSIGN_RESOURCE_MONITOR arm)                                       */
/*                                                                           */
/*  Return the total duration of the tasks monitored by arm.                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static int KheAssignResourceMonitorTotalMonitoredDuration(
  KHE_ASSIGN_RESOURCE_MONITOR arm)
{
  KHE_EVENT_RESOURCE er;
  er = KheAssignResourceMonitorEventResource(arm);
  return KheEventDuration(KheEventResourceEvent(er));
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_TASK KheTaskToDrsTask(KHE_TASK task,                             */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Convert task into the drs task corresponding to its proper root.         */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_TASK KheTaskToDrsTask(KHE_TASK task,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int task_index;
  task_index = KheTaskSolnIndex(KheTaskProperRoot(task));
  if( task_index >= HaArrayCount(drs->all_root_tasks) )
    return NULL;
  else
    return HaArray(drs->all_root_tasks, task_index);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddAssignResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,     */
/*    KHE_ASSIGN_RESOURCE_MONITOR arm)                                       */
/*                                                                           */
/*  Make and add in one or more expression trees representing arm.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddAssignResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_ASSIGN_RESOURCE_MONITOR arm)
{
  KHE_DRS_EXPR_INT_SUM_COMB eisc;  KHE_CONSTRAINT c;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;
  int min_limit, i, j;  KHE_DRS_TASK dt;  KHE_TASK task;
  KHE_DRS_TASK_ON_DAY dtd;  KHE_DRS_EXPR_ASSIGNED_TASK eat;
  KHE_DRS_EXPR_COST ec;  KHE_DRS_EXPR_INT_DEV eid;  KHE_EVENT_RESOURCE er;
  er = KheAssignResourceMonitorEventResource(arm);
  if( KheEventResourceResourceType(er) == drs->resource_type )
  {
    c = KheMonitorConstraint((KHE_MONITOR) arm);
    combined_wt = KheConstraintCombinedWeight(c);
    cost_fn = KheConstraintCostFunction(c);
    min_limit = KheEventDuration(KheEventResourceEvent(er));
    /* min_limit = KheAssignResourceMonitorTotalMonitoredDuration(arm); */
    if( cost_fn == KHE_LINEAR_COST_FUNCTION )
    {
      /* one tree per task on day */
      for( i = 0;  i < KheEventResourceTaskCount(drs->soln, er);  i++ )
      {
	task = KheEventResourceTask(drs->soln, er, i);
	dt = KheTaskToDrsTask(task, drs);
	if( dt != NULL )
	{
	  HaArrayForEach(dt->days, dtd, j)
	    if( dtd->task == task )
	    {
	      ec = KheDrsExprCostMake((KHE_MONITOR)arm,cost_fn,combined_wt,drs);
	      eid = KheDrsExprIntDevMake(1, INT_MAX, false, drs);
	      eat = KheDrsExprAssignedTaskMake(dtd, NULL, drs);
	      KheDrsExprAddChild((KHE_DRS_EXPR) eid, (KHE_DRS_EXPR) eat);
	      KheDrsExprInitEnd(eid, drs);
	      KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eid);
	      KheDrsExprInitEnd(ec, drs);
	      HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) ec);
	    }
	}
      }
    }
    else
    {
      /* a single tree, whose root has one child per task on day */
      eisc = KheDrsExprIntSumCombMake((KHE_MONITOR) arm, cost_fn, combined_wt,
	min_limit, INT_MAX, false, 0, 0, KheDrsDomTestMake(KHE_DRS_DOM_GE),drs);
      for( i = 0;  i < KheEventResourceTaskCount(drs->soln, er);  i++ )
      {
	task = KheEventResourceTask(drs->soln, er, i);
	dt = KheTaskToDrsTask(task, drs);
	if( dt != NULL )
	{
	  HaArrayForEach(dt->days, dtd, j)
	    if( dtd->task == task )
	    {
	      eat = KheDrsExprAssignedTaskMake(dtd, NULL, drs);
	      KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) eat);
	    }
	}
      }
      KheDrsExprInitEnd(eisc, drs);
      HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) eisc);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddPreferResourcesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,    */
/*    KHE_PREFER_RESOURCES_MONITOR prm)                                      */
/*                                                                           */
/*  Make and add in one or more expression trees representing prm.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddPreferResourcesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_PREFER_RESOURCES_MONITOR prm)
{
  KHE_CONSTRAINT c;  KHE_RESOURCE_GROUP rg;  KHE_EVENT_RESOURCE er;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;
  KHE_PREFER_RESOURCES_CONSTRAINT prc;  int i, j;  KHE_DRS_TASK_ON_DAY dtd;
  KHE_DRS_EXPR_COST ec;  KHE_DRS_EXPR_ASSIGNED_TASK eat;
  KHE_DRS_EXPR_INT_SUM eis;  KHE_DRS_DOM_TEST dom_test;
  KHE_TASK task;  KHE_DRS_TASK dt;
  er = KhePreferResourcesMonitorEventResource(prm);
  if( KheEventResourceResourceType(er) == drs->resource_type )
  {
    c = KheMonitorConstraint((KHE_MONITOR) prm);
    combined_wt = KheConstraintCombinedWeight(c);
    cost_fn = KheConstraintCostFunction(c);
    prc = (KHE_PREFER_RESOURCES_CONSTRAINT) c;
    rg = KhePreferResourcesConstraintDomainComplement(prc);
    if( cost_fn == KHE_LINEAR_COST_FUNCTION )
    {
      /* one tree per task on day */
      for( i = 0;  i < KheEventResourceTaskCount(drs->soln, er);  i++ )
      {
	task = KheEventResourceTask(drs->soln, er, i);
	dt = KheTaskToDrsTask(task, drs);
	if( dt != NULL )
	{
	  HaArrayForEach(dt->days, dtd, j)
	    if( dtd->task == task )
	    {
	      ec = KheDrsExprCostMake((KHE_MONITOR)prm,cost_fn,combined_wt,drs);
	      eat = KheDrsExprAssignedTaskMake(dtd, rg, drs);
	      KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eat);
	      KheDrsExprInitEnd(ec, drs);
	      HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) ec);
	    }
	}
      }
    }
    else
    {
      /* a single tree, whose root has one grandchild per task on day */
      ec = KheDrsExprCostMake((KHE_MONITOR) prm, cost_fn, combined_wt, drs);
      dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);
      eis = KheDrsExprIntSumMake(dom_test, drs);
      for( i = 0;  i < KheEventResourceTaskCount(drs->soln, er);  i++ )
      {
	task = KheEventResourceTask(drs->soln, er, i);
	dt = KheTaskToDrsTask(task, drs);
	if( dt != NULL )
	{
	  HaArrayForEach(dt->days, dtd, j)
	    if( dtd->task == task )
	    {
	      eat = KheDrsExprAssignedTaskMake(dtd, rg, drs);
	      KheDrsExprAddChild((KHE_DRS_EXPR) eis, (KHE_DRS_EXPR) eat);
	    }
	}
      }
      KheDrsExprInitEnd(eis, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eis);
      KheDrsExprInitEnd(ec, drs);
      HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) ec);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheLimitsToTests(int min_limit, int max_limit, int max_max_limit,   */
/*    bool allow_zero, KHE_DRS_DOM_TEST *alpha_dom_test,                     */
/*    KHE_DRS_DOM_TEST *beta_dom_test)                                       */
/*                                                                           */
/*  Convert limits (min_limit, max_limit, allow_zero) into dominance tests.  */
/*  if min_limit <= 0 there is no min limit; if max_limit >= max_max_limit   */
/*  there is no maximum limit.                                               */
/*                                                                           */
/*  The values returned in *alpha_dom_test and *beta_dom_test are referred   */
/*  to as alpha and beta in the User's Guide.                                */
/*                                                                           */
/*  Note.  If there is a trivial allow_zero flag (if min_limit <= 1), an     */
/*  adjustment will have already been made by now to remove it.  So this     */
/*  call may assume that allow_zero implies min_limit > 1.                   */
/*                                                                           */
/*****************************************************************************/

static void KheLimitsToTests(int min_limit, int max_limit, int max_max_limit,
  bool allow_zero, KHE_DRS_DOM_TEST *alpha_dom_test,
  KHE_DRS_DOM_TEST *beta_dom_test)
{
  if( allow_zero /* && min_limit > 1 */ )
  {
    /* fall back on equality when there is a non-trivial allow_zero flag */
    *alpha_dom_test = *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_EQ);
  }
  else if( max_limit >= max_max_limit )
  {
    if( min_limit <= 0 )
    {
      /* no max_limit, and no min_limit (should never happen) */
      HnAbort("KheLimitsToTests: no max limit, no min limit");
      *alpha_dom_test = *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
    }
    else
    {
      /* no max_limit, and min_limit */
      *alpha_dom_test = KheDrsDomTestMakeLoose(KHE_DRS_DOM_GE_LOOSE, min_limit);
      *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_GE);
    }
  }
  else
  {
    if( min_limit <= 0 )
    {
      /* max_limit, and no min_limit */
      *alpha_dom_test = *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);
    }
    else
    {
      /* max_limit, and min_limit */
      *alpha_dom_test = KheDrsDomTestMakeLoose(KHE_DRS_DOM_EQ_LOOSE, min_limit);
      *beta_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_EQ);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddLimitResourcesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,     */
/*    KHE_LIMIT_RESOURCES_MONITOR lrm)                                       */
/*                                                                           */
/*  Make and add in one or more expression trees representing lrm.           */
/*                                                                           */
/*  The documentation mentions an unlikely case where the expression can     */
/*  be split up.  We haven't bothered testing for that here.                 */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddLimitResourcesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_LIMIT_RESOURCES_MONITOR lrm)
{
  KHE_DRS_EXPR_INT_SUM_COMB eisc;  KHE_CONSTRAINT c;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;
  int min_limit, max_limit, i, egi, j;  KHE_RESOURCE_GROUP rg;
  KHE_LIMIT_RESOURCES_CONSTRAINT lrc;  KHE_EVENT_RESOURCE er;
  KHE_DRS_DOM_TEST alpha_dom_test, beta_dom_test;
  KHE_DRS_TASK_ON_DAY dtd;  KHE_TASK task;  KHE_DRS_TASK dt;
  KHE_DRS_EXPR_ASSIGNED_TASK eat;
  c = KheMonitorConstraint((KHE_MONITOR) lrm);
  combined_wt = KheConstraintCombinedWeight(c);
  cost_fn = KheConstraintCostFunction(c);
  lrc = (KHE_LIMIT_RESOURCES_CONSTRAINT) c;
  rg = KheLimitResourcesConstraintDomain(lrc);
  min_limit = KheLimitResourcesConstraintMinimum(lrc);
  max_limit = KheLimitResourcesConstraintMaximum(lrc);
  KheLimitsToTests(min_limit, max_limit, INT_MAX, false,
    &alpha_dom_test, &beta_dom_test);
  eisc = KheDrsExprIntSumCombMake((KHE_MONITOR) lrm, cost_fn, combined_wt,
    min_limit, max_limit, false, 0, 0, alpha_dom_test, drs);
  egi = KheLimitResourcesMonitorEventGroupIndex(lrm);
  for( i = 0; i < KheLimitResourcesConstraintEventResourceCount(lrc, egi); i++ )
  {
    er = KheLimitResourcesConstraintEventResource(lrc, egi, i);
    for( i = 0;  i < KheEventResourceTaskCount(drs->soln, er);  i++ )
    {
      task = KheEventResourceTask(drs->soln, er, i);
      dt = KheTaskToDrsTask(task, drs);
      if( dt != NULL )
      {
	HaArrayForEach(dt->days, dtd, j)
	  if( dtd->task == task )
	  {
	    eat = KheDrsExprAssignedTaskMake(dtd, rg, drs);
	    KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) eat);
	  }
      }
    }
  }
  KheDrsExprInitEnd(eisc, drs);
  HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) eisc);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "conversion of resource monitors into expression trees"        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddAvoidUnavailableTimesMonitor(                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs,                                       */
/*    KHE_AVOID_UNAVAILABLE_TIMES_MONITOR autm)                              */
/*                                                                           */
/*  Make and add in one or more expression trees representing autm.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddAvoidUnavailableTimesMonitor(
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_AVOID_UNAVAILABLE_TIMES_MONITOR autm,
  KHE_DRS_RESOURCE dr)
{
  KHE_AVOID_UNAVAILABLE_TIMES_CONSTRAINT c;  KHE_DRS_EXPR_BUSY_TIME eb;
  int i;  KHE_TIME t;  KHE_DRS_EXPR_COST ec;  KHE_COST_FUNCTION cost_fn;
  KHE_TIME_GROUP tg;  KHE_DRS_EXPR_INT_SUM eis;  KHE_COST combined_wt;
  KHE_DRS_DOM_TEST unused_dom_test, le_dom_test;  /* KHE_RESOURCE r; */
  KHE_DRS_RESOURCE_ON_DAY drd;

  /* boilerplate */
  /* if( KheResourceResourceType(r) == drs->resource_type ) always true */
  /* r = KheAvoidUnavailableTimesMonitorResource(autm); */
  c = KheAvoidUnavailableTimesMonitorConstraint(autm);
  tg = KheAvoidUnavailableTimesConstraintUnavailableTimes(c);
  combined_wt = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  if( combined_wt > 0 && KheTimeGroupTimeCount(tg) > 0 )
  {
    unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
    cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
    if( cost_fn == KHE_LINEAR_COST_FUNCTION || KheTimeGroupTimeCount(tg) == 1 )
    {
      /* a separate expression tree for each time */
      for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
      {
	t = KheTimeGroupTime(tg, i);
	ec = KheDrsExprCostMake((KHE_MONITOR) autm, cost_fn, combined_wt, drs);
	HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) ec);
	drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	eb = KheDrsExprBusyTimeMake(drd, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eb);
	KheDrsExprInitEnd(ec, drs);
      }
    }
    else
    {
      /* a single expression tree (max limit of 0, in effect) */
      ec = KheDrsExprCostMake((KHE_MONITOR) autm, cost_fn, combined_wt, drs);
      HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) ec);
      le_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);
      eis = KheDrsExprIntSumMake(le_dom_test, drs);
      for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
      {
	t = KheTimeGroupTime(tg, i);
	drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	eb = KheDrsExprBusyTimeMake(drd, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eis, (KHE_DRS_EXPR) eb);
      }
      KheDrsExprInitEnd(eis, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eis);
      KheDrsExprInitEnd(ec, drs);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheTimeGroupContainsWholeDay(KHE_TIME_GROUP tg, int start_index,    */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_TIME_GROUP *day_tg)               */
/*                                                                           */
/*  If tg contains a complete day's worth of times, starting from            */
/*  start_index, then return true and set *day_tg to the day time group,     */
/*  otherwise return false.                                                  */
/*                                                                           */
/*****************************************************************************/

static bool KheTimeGroupContainsWholeDay(KHE_TIME_GROUP tg, int start_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_DAY *day)
{
  KHE_TIME tg_t, day_t;  int i, day_index;  KHE_TIME_GROUP d_tg;
  tg_t = KheTimeGroupTime(tg, start_index);
  day_index = KheFrameTimeIndex(drs->days_frame, tg_t);
  d_tg = KheFrameTimeTimeGroup(drs->days_frame, tg_t);
  for( i = 0;  i < KheTimeGroupTimeCount(d_tg);  i++ )
  {
    day_t = KheTimeGroupTime(d_tg, i);
    if( start_index + i >= KheTimeGroupTimeCount(tg) )
      return *day = NULL, false;
    tg_t = KheTimeGroupTime(tg, start_index + i);
    if( day_t != tg_t )
      return *day = NULL, false;
  }
  return *day = HaArray(drs->all_days, day_index), true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMakeOrExprAndChildren(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,    */
/*    KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make an OR expression with one BUSY child for each time of tg, and add   */
/*  the OR expression as a child of prnt.  If tg has exactly one time,       */
/*  optimize by adding just the BUSY expression for that time as a child     */
/*  of prnt.                                                                 */
/*                                                                           */
/*  Either way, add dom_test to the root of these vertices.                  */
/*                                                                           */
/*****************************************************************************/

static void KheMakeOrExprAndChildren(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,
  KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_OR eo;  KHE_DRS_EXPR_BUSY_TIME ebt;  int i;  KHE_TIME t;
  KHE_DRS_DOM_TEST unused_dom_test;  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_DRS_DAY day;
  if( KheTimeGroupTimeCount(tg) == 1 )
  {
    /* just a single BUSY_TIME expression */
    t = KheTimeGroupTime(tg, 0);
    drd = KheDrsResourceOnDayAtTime(dr, t, drs);
    ebt = KheDrsExprBusyTimeMake(drd, t, dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) ebt);
  }
  else if( KheTimeGroupContainsWholeDay(tg, 0, drs, &day) &&
    KheTimeGroupTimeCount(tg) == KheTimeGroupTimeCount(day->time_group) )
  {
    /* just a single BUSY_DAY expression */
    drd = KheDrsResourceOnDay(dr, day);
    ebd = KheDrsExprBusyDayMake(drd, day->time_group, dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) ebd);
  }
  else
  {
    /* an OR expression with BUSY_DAY and BUSY_TIME children */
    eo = KheDrsExprOrMake(dom_test, drs);
    unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
    for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
    {
      if( KheTimeGroupContainsWholeDay(tg, i, drs, &day) )
      {
	/* add a BUSY_DAY expression */
	drd = KheDrsResourceOnDay(dr, day);
	ebd = KheDrsExprBusyDayMake(drd, day->time_group, dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eo, (KHE_DRS_EXPR) ebd);
	i += KheTimeGroupTimeCount(day->time_group) - 1;
      }
      else
      {
	/* add a BUSY_TIME expression */
	t = KheTimeGroupTime(tg, i);
	drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	ebt = KheDrsExprBusyTimeMake(drd, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eo, (KHE_DRS_EXPR) ebt);
      }
    }
    KheDrsExprInitEnd(eo, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) eo);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMakeAndExprAndChildren(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,   */
/*    KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test,                          */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Make an AND expression with one FREE child for each time of tg, and      */
/*  add the AND expression as a child of prnt.  If tg has exactly one time,  */
/*  optimize by adding just the FREE expression for that time as a child     */
/*  of prnt.                                                                 */
/*                                                                           */
/*  Either way, add dom_test to the root of these vertices.                  */
/*                                                                           */
/*****************************************************************************/

static void KheMakeAndExprAndChildren(KHE_DRS_RESOURCE dr, KHE_DRS_EXPR prnt,
  KHE_TIME_GROUP tg, KHE_DRS_DOM_TEST dom_test, KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_EXPR_AND ea;  KHE_DRS_EXPR_FREE_TIME eft;  int i;  KHE_TIME t;
  KHE_DRS_DOM_TEST unused_dom_test;  KHE_DRS_EXPR_FREE_DAY efd;
  KHE_DRS_DAY day;  KHE_DRS_RESOURCE_ON_DAY drd;
  if( KheTimeGroupTimeCount(tg) == 1 )
  {
    /* just a single FREE expression */
    t = KheTimeGroupTime(tg, 0);
    drd = KheDrsResourceOnDayAtTime(dr, t, drs);
    eft = KheDrsExprFreeTimeMake(drd, t, dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) eft);
  }
  else if( KheTimeGroupContainsWholeDay(tg, 0, drs, &day) &&
    KheTimeGroupTimeCount(tg) == KheTimeGroupTimeCount(day->time_group) )
  {
    /* just a single FREE_DAY expression */
    drd = KheDrsResourceOnDay(dr, day);
    efd = KheDrsExprFreeDayMake(drd, day->time_group, dom_test, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) efd);
  }
  else
  {
    /* an AND expression with FREE_DAY and FREE_TIME children */
    ea = KheDrsExprAndMake(dom_test, drs);
    unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
    for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
    {
      if( KheTimeGroupContainsWholeDay(tg, i, drs, &day) )
      {
	/* add a FREE_DAY expression */
	drd = KheDrsResourceOnDay(dr, day);
	efd = KheDrsExprFreeDayMake(drd, day->time_group, dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) ea, (KHE_DRS_EXPR) efd);
	i += KheTimeGroupTimeCount(day->time_group) - 1;
      }
      else
      {
	/* add a FREE_TIME expression */
	t = KheTimeGroupTime(tg, i);
	drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	eft = KheDrsExprFreeTimeMake(drd, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) ea, (KHE_DRS_EXPR) eft);
      }
    }
    KheDrsExprInitEnd(ea, drs);
    KheDrsExprAddChild(prnt, (KHE_DRS_EXPR) ea);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsClusterMonitorContainsWholePositiveDay(                       */
/*    KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm, int start_index,                  */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_DAY *day)                     */
/*                                                                           */
/*  If cbtm contains a sequence of singleton time groups starting at         */
/*  day_index that have positive polarity and cover a complete day, return   */
/*  true and set *day to the day;  otherwise return false.                   */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsClusterMonitorContainsWholePositiveDay(
  KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm, int start_index,
  KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_DRS_DAY *day)
{
  KHE_TIME day_t;  int i, tg_count, day_index;  KHE_TIME_GROUP m_tg, d_tg;
  KHE_POLARITY po;  KHE_TIME t;
  tg_count = KheClusterBusyTimesMonitorTimeGroupCount(cbtm);
  m_tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, start_index, &po);
  if( po == KHE_POSITIVE && KheTimeGroupTimeCount(m_tg) == 1 )
  {
    t = KheTimeGroupTime(m_tg, 0);
    d_tg = KheFrameTimeTimeGroup(drs->days_frame, t);
    day_index = KheFrameTimeIndex(drs->days_frame, t);
    for( i = 0;  i < KheTimeGroupTimeCount(d_tg);  i++ )
    {
      day_t = KheTimeGroupTime(d_tg, i);
      if( start_index + i >= tg_count )
	return *day = NULL, false;
      m_tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, start_index + i, &po);
      if( po != KHE_POSITIVE || KheTimeGroupTimeCount(m_tg) != 1 ||
	  KheTimeGroupTime(m_tg, 0) != day_t )
	return *day = NULL, false;
    }
    return *day = HaArray(drs->all_days, day_index), true;
  }
  else
    return *day = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddClusterBusyTimesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,   */
/*    KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm, KHE_DRS_RESOURCE dr)              */
/*                                                                           */
/*  Make and add in one or more expression trees representing cbtm.          */
/*                                                                           */
/*  Implementation note.  If cbtm contains a sequence of consecutive         */
/*  singleton time groups with positive polarity that add up to a complete   */
/*  day, then one BUSY_DAY expression is generated for them instead of one   */
/*  BUSY_TIME expression for each.  This is correct:  the sum reported       */
/*  in either case is 1 if the day is busy and 0 otherwise.  It cannot       */
/*  be more than 1 because the resource takes at most one task each day.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddClusterBusyTimesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm, KHE_DRS_RESOURCE dr)
{
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT c;  KHE_TIME_GROUP tg;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;  bool allow_zero;
  int i, min_limit, max_limit, history_before, history_after, tg_count;
  KHE_DRS_EXPR_INT_SUM_COMB eisc;  KHE_DRS_EXPR_BUSY_DAY ebd;
  KHE_POLARITY po;  KHE_DRS_DOM_TEST alpha_dom_test, beta_dom_test;
  KHE_DRS_RESOURCE_ON_DAY drd;  KHE_DRS_DAY day;
  /* KHE_RESOURCE r; */

  /* boilerplate */
  /* if( KheResourceResourceType(r) == drs->resource_type ) always true */
  /* r = KheClusterBusyTimesMonitorResource(cbtm); */
  c = KheClusterBusyTimesMonitorConstraint(cbtm);
  combined_wt = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  min_limit = KheClusterBusyTimesConstraintMinimum(c);
  max_limit = KheClusterBusyTimesConstraintMaximum(c);
  allow_zero = KheClusterBusyTimesConstraintAllowZero(c);
  history_before = KheClusterBusyTimesMonitorHistory(cbtm);
  history_after = KheClusterBusyTimesMonitorHistoryAfter(cbtm);

  /* adjust if allow_zero, and return if constraint does nothing */
  if( allow_zero && min_limit <= 1 )
    min_limit = 0, allow_zero = false;
  tg_count = KheClusterBusyTimesConstraintTimeGroupCount(c);
  if( (min_limit <= 0 && max_limit >= tg_count) || combined_wt == 0 )
    return;

  /* work out the various dom_tests */
  KheLimitsToTests(min_limit, max_limit, tg_count, allow_zero,
    &alpha_dom_test, &beta_dom_test);

  /* one int_cost_comb expression at top of tree */
  eisc = KheDrsExprIntSumCombMake((KHE_MONITOR) cbtm, cost_fn, combined_wt,
    min_limit, max_limit, allow_zero, history_before, history_after,
    alpha_dom_test, drs);

  /* one OR or AND expression for each time group */
  /* or a BUSY_DAY or FREE_DAY expression if all singles of one day present */
  for( i = 0;  i < tg_count;  i++ )
  {
    if( KheDrsClusterMonitorContainsWholePositiveDay(cbtm, i, drs, &day) )
    {
      /* add a BUSY_DAY expression */
      drd = KheDrsResourceOnDay(dr, day);
      ebd = KheDrsExprBusyDayMake(drd, day->time_group, beta_dom_test, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) ebd);
      i += KheTimeGroupTimeCount(day->time_group) - 1;
    }
    else
    {
      tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, i, &po);
      if( po == KHE_POSITIVE )
      {
	/* make an OR expression with one BUSY expression for each time of tg */
	KheMakeOrExprAndChildren(dr,(KHE_DRS_EXPR) eisc, tg,beta_dom_test,drs);
      }
      else
      {
	/* make an AND expression with one FREE expression for each time of tg */
	KheMakeAndExprAndChildren(dr,(KHE_DRS_EXPR)eisc, tg,beta_dom_test,drs);
      }
    }
  }
  KheDrsExprInitEnd(eisc, drs);
  HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) eisc);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddLimitBusyTimesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,     */
/*    KHE_LIMIT_BUSY_TIMES_MONITOR lbtm, KHE_DRS_RESOURCE dr)                */
/*                                                                           */
/*  Make and add in one or more expression trees representing lbtm.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddLimitBusyTimesMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_LIMIT_BUSY_TIMES_MONITOR lbtm, KHE_DRS_RESOURCE dr)
{
  KHE_LIMIT_BUSY_TIMES_CONSTRAINT c;  KHE_TIME_GROUP tg;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;
  int i, j, junk, min_limit, max_limit;  bool allow_zero;
  KHE_DRS_EXPR_INT_SUM_COMB eisc;  KHE_TIME t;
  KHE_DRS_EXPR_INT_SUM eis, eis2;  KHE_DRS_EXPR_COST ec;
  KHE_DRS_EXPR_INT_DEV eid;
  KHE_DRS_EXPR_BUSY_TIME ebt;
  KHE_DRS_DOM_TEST unused_dom_test, le_dom_test, alpha_dom_test, beta_dom_test;
  KHE_RESOURCE r;  KHE_DRS_DAY day;  KHE_DRS_RESOURCE_ON_DAY drd;

  /* boilerplate */
  /* if( KheResourceResourceType(r) == drs->resource_type ) always true */
  r = KheLimitBusyTimesMonitorResource(lbtm);
  c = KheLimitBusyTimesMonitorConstraint(lbtm);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  combined_wt = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  min_limit = KheLimitBusyTimesConstraintMinimum(c);
  max_limit = KheLimitBusyTimesConstraintMaximum(c);
  allow_zero = KheLimitBusyTimesConstraintAllowZero(c);
  unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  le_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);
  dr = KheDrsResource(r, drs);

  /* adjust allow_zero, and return if constraint does nothing */
  if( allow_zero && min_limit <= 1 )
    min_limit = 0, allow_zero = false;
  if( combined_wt == 0 )
    return;

  if( cost_fn == KHE_LINEAR_COST_FUNCTION ||
      KheLimitBusyTimesMonitorTimeGroupCount(lbtm) == 1 )
  {
    /* one INT_SUM_COMB for each time group */
    /* unless the time group is one day, min_limit = 0, and max_limit = 1 */
    for( i = 0;  i < KheLimitBusyTimesMonitorTimeGroupCount(lbtm);  i++ )
    {
      tg = KheLimitBusyTimesMonitorTimeGroup(lbtm, i, &junk);
      if( min_limit == 0 && max_limit == 1 &&
	  KheTimeGroupContainsWholeDay(tg, 0, drs, &day) &&
	  KheTimeGroupTimeCount(tg) == KheTimeGroupTimeCount(day->time_group) )
      {
	/* constraint specifies at most one task per day; trivial, so omit */
      }
      else if( min_limit > 0 || max_limit < KheTimeGroupTimeCount(tg) )
      {
	KheLimitsToTests(min_limit, max_limit, KheTimeGroupTimeCount(tg),
	  allow_zero, &alpha_dom_test, &beta_dom_test);
	eisc = KheDrsExprIntSumCombMake((KHE_MONITOR) lbtm, cost_fn,
	  combined_wt, min_limit, max_limit, allow_zero, 0, 0,
	  alpha_dom_test, drs);
	for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
	{
	  t = KheTimeGroupTime(tg, j);
	  drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	  ebt = KheDrsExprBusyTimeMake(drd, t, unused_dom_test, drs);
	  KheDrsExprAddChild((KHE_DRS_EXPR) eisc, (KHE_DRS_EXPR) ebt);
	}
	KheDrsExprInitEnd(eisc, drs);
	HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) eisc);
      }
    }
  }
  else
  {
    /* one COST and INT_SUM at root of tree */
    ec = KheDrsExprCostMake((KHE_MONITOR) lbtm, cost_fn, combined_wt, drs);
    HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) ec);
    eis = KheDrsExprIntSumMake(le_dom_test, drs);

    /* one INT_DEV and INT_SUM for each time group */
    for( i = 0;  i < KheLimitBusyTimesMonitorTimeGroupCount(lbtm);  i++ )
    {
      tg = KheLimitBusyTimesMonitorTimeGroup(lbtm, i, &junk);
      eid = KheDrsExprIntDevMake(min_limit, max_limit, allow_zero, drs);
      KheLimitsToTests(min_limit, max_limit, KheTimeGroupTimeCount(tg),
	allow_zero, &alpha_dom_test, &beta_dom_test);
      eis2 = KheDrsExprIntSumMake(alpha_dom_test, drs);
      for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
      {
	t = KheTimeGroupTime(tg, j);
	drd = KheDrsResourceOnDayAtTime(dr, t, drs);
	ebt = KheDrsExprBusyTimeMake(drd, t, unused_dom_test, drs);
	KheDrsExprAddChild((KHE_DRS_EXPR) eis2, (KHE_DRS_EXPR) ebt);
      }
      KheDrsExprInitEnd(eis2, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) eid, (KHE_DRS_EXPR) eis2);
      KheDrsExprInitEnd(eid, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) eis, (KHE_DRS_EXPR) eid);
    }
    KheDrsExprInitEnd(eis, drs);
    KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eis);
    KheDrsExprInitEnd(ec, drs);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddLimitWorkloadMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,      */
/*    KHE_LIMIT_WORKLOAD_MONITOR lwm, KHE_DRS_RESOURCE dr)                   */
/*                                                                           */
/*  Make and add in one or more expression trees representing lwm.           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddLimitWorkloadMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_LIMIT_WORKLOAD_MONITOR lwm, KHE_DRS_RESOURCE dr)
{
  KHE_LIMIT_WORKLOAD_CONSTRAINT c;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;
  int i, j, min_limit, max_limit;  bool allow_zero;  float junk;
  KHE_TIME_GROUP tg;  KHE_TIME t;  KHE_DRS_EXPR_WORK_TIME ew;
  KHE_DRS_EXPR_INT_SUM eis;  KHE_DRS_EXPR_COST ec;
  KHE_DRS_EXPR_FLOAT_SUM efs;  KHE_DRS_EXPR_FLOAT_DEV efd;
  KHE_DRS_DOM_TEST unused_dom_test, le_dom_test, alpha_dom_test, beta_dom_test;
  KHE_DRS_RESOURCE_ON_DAY drd;
  /* KHE_RESOURCE r; */

  /* boilerplate */
  /* if( KheResourceResourceType(r) == drs->resource_type ) always true */
  /* r = KheLimitWorkloadMonitorResource(lwm); */
  c = KheLimitWorkloadMonitorConstraint(lwm);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  combined_wt = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  min_limit = KheLimitWorkloadConstraintMinimum(c);
  max_limit = KheLimitWorkloadConstraintMaximum(c);
  allow_zero = KheLimitWorkloadConstraintAllowZero(c);
  unused_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_UNUSED);
  le_dom_test = KheDrsDomTestMake(KHE_DRS_DOM_LE);

  /* adjust allow_zero, and return if constraint does nothing */
  if( allow_zero && min_limit <= 1 )
    min_limit = 0, allow_zero = false;
  if( combined_wt == 0 )
    return;

  /* one COST and INT_SUM at the root */
  ec = KheDrsExprCostMake((KHE_MONITOR) lwm, cost_fn, combined_wt, drs);
  HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) ec);
  eis = KheDrsExprIntSumMake(le_dom_test, drs);

  /* one FLOAT_DEV and FLOAT_SUM for each time group */
  KheLimitsToTests(min_limit, max_limit, INT_MAX, allow_zero,
    &alpha_dom_test, &beta_dom_test);
  for( i = 0;  i < KheLimitWorkloadMonitorTimeGroupCount(lwm);  i++ )
  {
    tg = KheLimitWorkloadMonitorTimeGroup(lwm, i, &junk);
    efd = KheDrsExprFloatDevMake(min_limit, max_limit, allow_zero, drs);
    efs = KheDrsExprFloatSumMake(alpha_dom_test, drs);
    for( j = 0;  j < KheTimeGroupTimeCount(tg);  j++ )
    {
      t = KheTimeGroupTime(tg, j);
      drd = KheDrsResourceOnDayAtTime(dr, t, drs);
      ew = KheDrsExprWorkTimeMake(drd, t, unused_dom_test, drs);
      KheDrsExprAddChild((KHE_DRS_EXPR) efs, (KHE_DRS_EXPR) ew);
    }
    KheDrsExprInitEnd(efs, drs);
    KheDrsExprAddChild((KHE_DRS_EXPR) efd, (KHE_DRS_EXPR) efs);
    KheDrsExprInitEnd(efd, drs);
    KheDrsExprAddChild((KHE_DRS_EXPR) eis, (KHE_DRS_EXPR) efd);
  }
  KheDrsExprInitEnd(eis, drs);
  KheDrsExprAddChild((KHE_DRS_EXPR) ec, (KHE_DRS_EXPR) eis);
  KheDrsExprInitEnd(ec, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddLimitActiveIntervalsMonitor(                               */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs,                                       */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, KHE_DRS_RESOURCE dr)          */
/*                                                                           */
/*  Make and add in one or more expression trees representing laim.          */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddLimitActiveIntervalsMonitor(
  KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, KHE_DRS_RESOURCE dr)
{
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c;  KHE_TIME_GROUP tg;  KHE_POLARITY po;
  KHE_COST_FUNCTION cost_fn;  KHE_COST combined_wt;
  int i, min_limit, max_limit, history_before, history_after, tg_count;
  KHE_DRS_EXPR_INT_SEQ_COMB eisc;
  KHE_DRS_DOM_TEST alpha_dom_test, beta_dom_test;
  /* KHE_RESOURCE r; */

  /* boilerplate */
  /* if( KheResourceResourceType(r) == drs->resource_type ) always true */
  /* r = KheLimitActiveIntervalsMonitorResource(laim); */
  c = KheLimitActiveIntervalsMonitorConstraint(laim);
  combined_wt = KheConstraintCombinedWeight((KHE_CONSTRAINT) c);
  cost_fn = KheConstraintCostFunction((KHE_CONSTRAINT) c);
  min_limit = KheLimitActiveIntervalsConstraintMinimum(c);
  max_limit = KheLimitActiveIntervalsConstraintMaximum(c);
  history_before = KheLimitActiveIntervalsMonitorHistory(laim);
  history_after = KheLimitActiveIntervalsMonitorHistoryAfter(laim);
  tg_count = KheLimitActiveIntervalsConstraintTimeGroupCount(c);

  /* return if constraint does nothing */
  if( (min_limit <= 1 && max_limit >= tg_count) || combined_wt == 0 )
    return;

  /* work out the various dom_tests */
  KheLimitsToTests(min_limit, max_limit, tg_count, false,
    &alpha_dom_test, &beta_dom_test);

  /* one int_seq_comb expression at top of tree */
  eisc = KheDrsExprIntSeqCombMake((KHE_MONITOR) laim, cost_fn, combined_wt,
    min_limit, max_limit, history_before, history_after, alpha_dom_test, drs);

  /* one OR or AND expression for each time group */
  for( i = 0;  i < KheLimitActiveIntervalsMonitorTimeGroupCount(laim);  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, i, &po);
    if( po == KHE_POSITIVE )
    {
      /* make an OR expression with one BUSY expression for each time of tg */
      KheMakeOrExprAndChildren(dr, (KHE_DRS_EXPR) eisc, tg,beta_dom_test,drs);
    }
    else
    {
      /* make an AND expression with one FREE expression for each time of tg */
      KheMakeAndExprAndChildren(dr, (KHE_DRS_EXPR) eisc,tg,beta_dom_test,drs);
    }
  }
  KheDrsExprInitEnd(eisc, drs);
  HaArrayAddLast(drs->all_root_exprs, (KHE_DRS_EXPR) eisc);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,           */
/*    KHE_MONITOR m, KHE_DRS_RESOURCE dr)                                    */
/*                                                                           */
/*  Add an expression tree to drs for resource monitor m, if relevant;       */
/*  dr is the drs resource corresponding to the resource monitored by m.     */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_MONITOR m, KHE_DRS_RESOURCE dr)
{
  switch( KheMonitorTag(m) )
  {
    case KHE_AVOID_CLASHES_MONITOR_TAG:

      /* resource monitor, but no need to handle it */
      break;

    case KHE_AVOID_UNAVAILABLE_TIMES_MONITOR_TAG:

      KheDrsAddAvoidUnavailableTimesMonitor(drs,
	(KHE_AVOID_UNAVAILABLE_TIMES_MONITOR) m, dr);
      break;

    case KHE_LIMIT_IDLE_TIMES_MONITOR_TAG:

      /* resource monitor, but we are currently not handling it */
      break;

    case KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG:

      KheDrsAddClusterBusyTimesMonitor(drs,
	(KHE_CLUSTER_BUSY_TIMES_MONITOR) m, dr);
      break;

    case KHE_LIMIT_BUSY_TIMES_MONITOR_TAG:

      KheDrsAddLimitBusyTimesMonitor(drs, (KHE_LIMIT_BUSY_TIMES_MONITOR) m, dr);
      break;

    case KHE_LIMIT_WORKLOAD_MONITOR_TAG:

      KheDrsAddLimitWorkloadMonitor(drs, (KHE_LIMIT_WORKLOAD_MONITOR) m, dr);
      break;

    case KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG:

      KheDrsAddLimitActiveIntervalsMonitor(drs,
	(KHE_LIMIT_ACTIVE_INTERVALS_MONITOR) m, dr);
      break;

    default:

      HnAbort("KheDrsAddResourceMonitor internal error (tag %d)",
	KheMonitorTag(m));
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddEventResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,      */
/*    KHE_MONITOR m)                                                         */
/*                                                                           */
/*  Add an expression tree to drs for event resource monitor m, if relevant. */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddEventResourceMonitor(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_MONITOR m)
{
  switch( KheMonitorTag(m) )
  {
    case KHE_ASSIGN_RESOURCE_MONITOR_TAG:

      KheDrsAddAssignResourceMonitor(drs, (KHE_ASSIGN_RESOURCE_MONITOR) m);
      break;

    case KHE_PREFER_RESOURCES_MONITOR_TAG:

      KheDrsAddPreferResourcesMonitor(drs, (KHE_PREFER_RESOURCES_MONITOR) m);
      break;

    case KHE_AVOID_SPLIT_ASSIGNMENTS_MONITOR_TAG:

      /* event resource monitor, but we are currently not handling it */
      break;

    case KHE_LIMIT_RESOURCES_MONITOR_TAG:

      KheDrsAddLimitResourcesMonitor(drs, (KHE_LIMIT_RESOURCES_MONITOR) m);
      break;

    default:

      HnAbort("KheDrsAddEventResourceMonitor internal error (tag %d)",
	KheMonitorTag(m));
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "public functions"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DYNAMIC_RESOURCE_SOLVER KheDynamicResourceSolverMake(KHE_SOLN soln,  */
/*    KHE_RESOURCE_TYPE rt, KHE_OPTIONS options)                             */
/*                                                                           */
/*  Make and return a new dynamic resource solver with these attributes.     */
/*                                                                           */
/*****************************************************************************/

KHE_DYNAMIC_RESOURCE_SOLVER KheDynamicResourceSolverMake(KHE_SOLN soln,
  KHE_RESOURCE_TYPE rt, KHE_OPTIONS options)
{
  KHE_DYNAMIC_RESOURCE_SOLVER res;  HA_ARENA a;  int i, j /* , k, n, p, val */;
  KHE_FRAME days_frame;  KHE_EVENT_TIMETABLE_MONITOR etm;
  KHE_TASK_CLASS tc;  KHE_INSTANCE ins;  KHE_EVENT_RESOURCE er;
  KHE_MONITOR m;  KHE_RESOURCE r;  KHE_DRS_RESOURCE dr;  /* KHE_DRS_EXPR e; */
  ARRAY_KHE_MONITOR er_monitors;
  /* KHE_DRS_RESOURCE_ON_DAY drd;  KHE_TASK task;  KHE_TIME t; */
  /* KHE_DRS_TASK_CLASS dtc;  KHE_DRS_TASK dt; */
  /* KHE_DRS_TASK_ON_DAY dtd;  KHE_DRS_DAY day; */

  /* get the common frame and event timetable monitor; return NULL if can't */
  days_frame = KheOptionsFrame(options, "gs_common_frame", soln);
  etm = (KHE_EVENT_TIMETABLE_MONITOR)
    KheOptionsGetObject(options, "gs_event_timetable_monitor", NULL);
  if( days_frame == NULL || etm == NULL )
    return NULL;

  /*************************************************************************/
  /*                                                                       */
  /*  Make the basic object and initialize its fields                      */
  /*                                                                       */
  /*************************************************************************/

  /* make the basic object */
  a = KheSolnArenaBegin(soln, false);
  HaMake(res, a);

  /* fields which are constant throughout the lifetime of the solver */
  res->arena = a;
  res->soln = soln;
  res->resource_type = rt;
  res->options = options;
  res->days_frame = days_frame;
  res->etm = etm;
  res->task_class_solver = NULL;  /* initialized below */
  HaArrayInit(res->all_resources, a);
  HaArrayInit(res->all_days, a);
  HaArrayInit(res->all_root_tasks, a);
  HaArrayInit(res->all_root_exprs, a);
  res->postorder_count = 0;

  /* free list fields - solns and days */
  /* HaArrayInit(res->free_edge, a); */
  HaArrayInit(res->free_soln, a);
  HaArrayInit(res->free_soln_list, a);
  HaArrayInit(res->free_soln_trie, a);
  HaArrayInit(res->free_soln_set, a);

  /* fields that vary with the solve */
  /* res->solve_state = KHE_DRS_SOLVE_LOADING; */
  res->selected_resource_set = KheResourceSetMake(rt, a);
  HaArrayInit(res->open_resources, a);
  HaArrayInit(res->selected_day_ranges, a);
  HaArrayInit(res->open_days, a);
  HaArrayInit(res->open_exprs, a);
  res->solve_diversifier = 0;
  res->solve_rerun = false;
  /* HaArrayInit(res->solve_final_solns, a); */

  /* fields for recording running time */
#if TESTING
  res->timer = KheTimerMake("dynamic", KHE_NO_TIME, a);
  HaArrayInit(res->running_time_per_day, a);
#endif


  /*************************************************************************/
  /*                                                                       */
  /*  Add days, tasks, resources, and expression trees                     */
  /*                                                                       */
  /*  When adding each drs resource, its resource monitors' expression     */
  /*  trees are also added.  This assumes that all drs days are made.      */
  /*                                                                       */
  /*  When adding each drs task class, its drs tasks are added as well,    */
  /*  both to the task class and to drs->all_root_tasks.  If a drs task's  */
  /*  corresponding KHE task is currently assigned to resource r, the      */
  /*  drs task is also assigned to the corresponding drs resource.  This   */
  /*  assumes that all drs days and all drs resources are already made.    */
  /*                                                                       */
  /*  When adding the expression tree for each event resource monitor,     */
  /*  it is assumed that drs->all_root_tasks is initialized.               */
  /*                                                                       */
  /*************************************************************************/

  /* make one drs day object for each day in the frame */
  for( i = 0;  i < KheFrameTimeGroupCount(days_frame);  i++ )
    KheDrsDayMake(i, res);

  /* make one drs resource for each resource in rt */
  HaArrayInit(res->all_resources, a);
  for( i = 0;  i < KheResourceTypeResourceCount(rt);  i++ )
  {
    r = KheResourceTypeResource(rt, i);
    dr = KheDrsResourceMake(r, res);
    HaArrayAddLast(res->all_resources, dr);
  }

  /* make one drs task class for each no-overlap task class */
  /* this code assumes that drs resources are already made */
  res->task_class_solver = KheTaskClassSolverMake(soln, rt, days_frame, a);
  KheTaskClassSolverAddAllTasks(res->task_class_solver);
  if( DEBUG15 )
    KheTaskClassSolverDebug(res->task_class_solver, 2, 2, stderr);
  for( i = 0;  i < KheTaskClassSolverClassCount(res->task_class_solver);  i++ )
  {
    tc = KheTaskClassSolverClass(res->task_class_solver, i);
    if( KheTaskClassNoOverlap(tc) )
      KheDrsTaskClassMake(tc, res);
  }

  /* make one expression tree for each event resource monitor of type rt */
  ins = KheSolnInstance(soln);
  HaArrayInit(er_monitors, a);
  for( i = 0;  i < KheInstanceEventResourceCount(ins);  i++ )
  {
    er = KheInstanceEventResource(ins, i);
    if( KheEventResourceResourceType(er) == rt )
      for( j = 0;  j < KheSolnEventResourceMonitorCount(soln, er);  j++ )
      {
	m = KheSolnEventResourceMonitor(soln, er, j);
	HaArrayAddLast(er_monitors, m);
      }
  }
  HaArraySortUnique(er_monitors, &KheMonitorCmp);
  HaArrayForEach(er_monitors, m, i)
    KheDrsAddEventResourceMonitor(res, m);
  HaArrayFree(er_monitors);

  /* all done */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverDelete(KHE_DYNAMIC_RESOURCE_SOLVER drs)     */
/*                                                                           */
/*  Delete drs, reclaiming its memory.                                       */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverDelete(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KheSolnArenaEnd(drs->soln, drs->arena);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverClear(KHE_DYNAMIC_RESOURCE_SOLVER drs,      */
/*    bool full)                                                             */
/*                                                                           */
/*  Clear drs, making it ready for loading new resources and day intervals.  */
/*  This includes removing all trace of any previous solve.                  */
/*                                                                           */
/*  If full is true, the function is as just described.  If full is false,   */
/*  the solve resources and intervals are left intact, so that another       */
/*  solve can be carried out immediately.                                    */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete; solver is cleared back to the closed state at end of solve
static void KheDynamicResourceSolverClear(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  bool full)
{
  ** KHE_DRS_RESOURCE dr;  int i;  KHE_DRS_EXPR e; **

  ** change state to loading **
  drs->solve_state = KHE_DRS_SOLVE_LOADING;

  ** remove solve resources **
  ** ***
  HaArrayForEach(drs->open_resources, dr, i)
    KheDrsResourceFree(dr, drs);
  *** **
  HaArrayClear(drs->open_resources);

  ** remove solve expressions **
  ** ***
  HaArrayForEach(drs->open_exprs, e, i)
    KheDrsExprFreeInternal(e, drs);
  *** **
  HaArrayClear(drs->open_exprs);

  ** remove effect of adding resources and day intervals **
  if( full )
  {
    KheResourceSetClear(drs->selected_resource_set);
    HaArrayClear(drs->selected_day_ranges);
  }

  ** remainder (if any) still to do **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverAddResource(KHE_DYNAMIC_RESOURCE_SOLVER drs,*/
/*    KHE_RESOURCE r)                                                        */
/*                                                                           */
/*  Add resource r to the solve.                                             */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverAddResource(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_RESOURCE r)
{
  /* ***
  if( drs->solve_state == KHE_DRS_SOLVE_SOLVED )
    KheDynamicResourceSolverClear(drs, true);
  *** */
  HnAssert(KheResourceResourceType(r) == drs->resource_type,
    "KheDynamicResourceSolverAddResource: r has wrong resource type for drs");
  HnAssert(!KheResourceSetContainsResource(drs->selected_resource_set, r),
    "KheDynamicResourceSolverAddResource: r (%s) already added",
    KheResourceId(r));
  KheResourceSetAddResource(drs->selected_resource_set, r);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverAddDayRange(KHE_DYNAMIC_RESOURCE_SOLVER drs,*/
/*    int first_day_index, int last_day_index)                               */
/*                                                                           */
/*  Add day range first_day_index .. last_day_index to the solve.            */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverAddDayRange(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int first_day_index, int last_day_index)
{
  KHE_DRS_DAY_RANGE si;
  /* ***
  if( drs->solve_state == KHE_DRS_SOLVE_SOLVED )
    KheDynamicResourceSolverClear(drs, true);
  *** */
  HnAssert(first_day_index <= last_day_index,
    "KheDynamicResourceSolverAddResource: empty day range (%d .. %d)",
    first_day_index, last_day_index);
  HnAssert(first_day_index >= 0,
    "KheDynamicResourceSolverAddResource: first_day_index (%d) out of range",
    first_day_index);
  HnAssert(last_day_index < KheFrameTimeGroupCount(drs->days_frame),
    "KheDynamicResourceSolverAddResource: last_day_index (%d) out of range",
    last_day_index);
  si = KheDrsDayRangeMake(first_day_index, last_day_index);
  KheDrsDayRangeMerge(&drs->selected_day_ranges, si);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheFindContinuingTask(KHE_DRS_SOLN soln, int i, KHE_DRS_TASK *dt)   */
/*                                                                           */
/*  If there is a task for the resource at index i running beyond the last   */
/*  day of soln, return true and set *dt to that task.                       */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheFindContinuingTask(KHE_DRS_SOLN soln, int i, KHE_DRS_TASK *dt)
{
  return *dt = NULL, false;  ** placeholder **
  ** *** the idea of this function is still good, but new version stil l to do
  KHE_DRS_TASK dt2;
  if( soln->prev_edge == NULL )
    return *dt = NULL, false;
  dt2 = HaArray(soln->prev_edge->tasks, i);
  if( KheDailyScheduleLastDayIndex(dt2->daily_schedule) > soln->day->day_index )
    return *dt = dt2, true;
  else
    return *dt = NULL, false;
  *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheTaskGroupContainsSuitableTask(KHE_DRS_TASK_GROUP dtg,            */
/*    KHE_DRS_RESOURCE dr, KHE_DRS_EDGE de, KHE_DRS_TASK *dt)                */
/*                                                                           */
/*  If dtg contains a task to which dr can be assigned and which is not      */
/*  already assigned to a resource in de, return true and set *dt to the     */
/*  first such task.  Otherwise return false with *dt set to NULL.           */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete (except for domain check)
static bool KheTaskGroupContainsSuitableTask(KHE_DRS_TASK_GROUP dtg,
  KHE_DRS_RESOURCE dr, KHE_DRS_EDGE de, KHE_DRS_TASK *dt)
{
  int i, pos;  KHE_DRS_TASK dt2;
  if( HaArrayCount(dtg->solve_tasks) > 0 )
  {
    dt2 = HaArrayFirst(dtg->solve_tasks);
    if( KheResourceGroupContains(KheTaskDomain(dt2->task), dr->resource) )
    {
      ** dtg's tasks have a suitable domain; now find one not in de **
      HaArrayForEach(dtg->solve_tasks, dt2, i)
	if( !HaArrayContains(de->tasks, dt2, &pos) )
	  return *dt = dt2, true;
    }
  }

  ** no luck **
  return *dt = NULL, false;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverClear(KHE_DYNAMIC_RESOURCE_SOLVER drs)      */
/*                                                                           */
/*  Clear drs, ready for a fresh solve.                                      */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete
static void KheDynamicResourceSolverClear(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i;  KHE_DRS_DAY day;  ** KHE_DRS_EXPR e; **
  drs->solve_diversifier = 0;
  HaArrayForEach(drs->open_days, day, i)
    KheDrsDayClear(day, drs);
  HaArrayClear(drs->solve_final_solns);
  ** ***
  HaArrayForEach(drs->exprs, e, i)
    KheDrsExprFree(e, drs);
  HaArrayClear(drs->exprs);
  *** **
  ** KheDrsExprCostSumFree(drs->root_expr, drs); **
  ** drs->root_expr = KheDrsExprCostSumMake(drs); **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsInitializeForSolving(KHE_DYNAMIC_RESOURCE_SOLVER drs,         */
/*    KHE_RESOURCE r, KHE_DRS_DOM_KIND dom_kind, int min_assts,              */
/*    int max_assts, KHE_COST cost_limit)                                    */
/*                                                                           */
/*  Initialize drs for a solve with these attributes.                        */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete
static void KheDrsInitializeForSolving(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_RESOURCE r, KHE_DRS_DOM_KIND dom_kind, int min_assts,
  int max_assts, KHE_COST cost_limit)
{
  int di, i;  KHE_MONITOR m;  KHE_DRS_DAY day;
  KHE_DRS_TASK preassigned_dt;  KHE_DRS_EXPR e;

  ** if there was already a resource, clear out old stuff **
  if( drs->resource != NULL )
    KheDynamicResourceSolverClear(drs);

  ** add the resource, and the expression tree representing the monitors **
  drs->resource = r;
  drs->dom_kind = dom_kind;
  drs->min_assts = min_assts;
  drs->max_assts = max_assts;
  drs->cost_limit = cost_limit;
  KheSolnResource MonitorSort(drs->soln, r, &KheMonitorLargestLimitFirstCmp);
  if( DEBUG13 )
    fprintf(stderr, "  [ monitor limits:\n");
  for( i = 0;  i < KheSolnResourceMonitorCount(drs->soln, r);  i++ )
  {
    m = KheSolnResourceMonitor(drs->soln, r, i);
    if( DEBUG13 )
    {
      fprintf(stderr, "    limit %d: ", KheMonitorLimit(m));
      KheMonitorDebug(m, 2, 0, stderr);
    }
    KheDrsResourceA ddMonitor(drs, m);
  }
  if( DEBUG13 )
    fprintf(stderr, "  ]\n");

  ** sort out the day info for the expression tree **
  HaArrayForEach(drs->exprs, e, i)
  {
    KheDrsExprSetDayRange(e, drs);
    KheDrsExprSetSigIndexes(e, drs);
    KheDrsExprAddExpressionsToDays(e, drs);
  }
  ** ***
  KheDrsExprSetDayRange((KHE_DRS_EXPR) drs->root_expr, 0, drs);
  KheDrsExprSetSigIndexes((KHE_DRS_EXPR) drs->root_expr, drs);
  KheDrsExprAddExpressionsToDays((KHE_DRS_EXPR) drs->root_expr, drs);
  *** **

  ** add suitable tasks for each day, skipping days with preassigned tasks **
  HaArrayForEach(drs->days, day, di)
    if( KheDrsDayAddTasks(day, drs, &preassigned_dt) )
      di = KheDailyScheduleLastDayIndex(preassigned_dt->daily_schedule);
           ** continue from after here **

  ** work out the minimum cost reduction over all selected tasks **
  ** *** sti ll to do
  drs->min_cost_reduction = KheCost(INT_MAX, INT_MAX);
  HaArrayForEach(drs->days, day, di)
    HaArrayForEach(day->times, stime, i)
      HaArrayForEach(stime->best_tasks, dt, j)
        if( !dt->already_assigned &&
	    dt->cost_reduction < drs->min_cost_reduction )
	  drs->min_cost_reduction = dt->cost_reduction;
  *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolverOpen(KHE_DYNAMIC_RESOURCE_SOLVER drs)                   */
/*                                                                           */
/*  Open drs for solving of selected_resource_set and selected_day_ranges.   */
/*                                                                           */
/*  Implementation note.  The first step is to open resources.  This must    */
/*  be done before opening days, because it unassigns some tasks, and        */
/*  opening days includes sorting the unassigned tasks on those days.        */
/*                                                                           */
/*  Opening a resource r includes opening the expressions corresponding to   */
/*  the parts of the resource monitors of r that monitor what r is doing     */
/*  on the open days.  Opening a day includes opening all unassigned tasks   */
/*  t that begin on that day and lie entirely in one day range; opening t    */
/*  opens the expressions corresponding to the parts of the event resource   */
/*  monitors that monitor t.                                                 */
/*                                                                           */
/*  Either way, opening an expression includes adding it to drs->open_exprs. */
/*  After the desired expressions are all opened, this function sorts them   */
/*  by increasing postorder index, as required for updating them each day.   */
/*                                                                           */
/*****************************************************************************/

/* *** inlined now
static void KheDrsSolverOpen(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  int i, j;  KHE_DRS_DAY_RANGE ddr;  KHE_DRS_EXPR e;  KHE_DRS_DAY day;
  KHE_RESOURCE r;  KHE_DRS_RESOURCE dr;

  ** clear the open resources, open days, and open expressions **
  HaArrayClear(drs->open_resources);
  HaArrayClear(drs->open_days);
  HaArrayClear(drs->open_exprs);

  ** open the selected resources **
  for( i = 0; i < KheResourceSetResourceCount(drs->selected_resource_set); i++ )
  {
    r = KheResourceSetResource(drs->selected_resource_set, i);
    dr = HaArray(drs->all_resources, KheResourceResourceTypeIndex(r));
    KheDrsResourceOpen(dr, drs, HaArrayCount(drs->open_resources));
    HaArrayAddLast(drs->open_resources, dr);
  }

  ** open the selected days **
  HaArrayForEach(drs->selected_day_ranges, ddr, i)
    for( j = ddr.first;  j <= ddr.last;  j++ )
    {
      day = HaArray(drs->all_days, j);
      KheDrsDayOpen(day, ddr, drs, HaArrayCount(drs->open_days));
      HaArrayAddLast(drs->open_days, day);
    }

  ** sort the open expressions by increasing postorder index, and **
  ** open these expressions **
  HaArraySort(drs->open_exprs, &KheDrsExprCmp);
  HaArrayForEach(drs->open_exprs, e, i)
    KheDrsExprOpen(e, drs);

  ** sort out the tricky fields of the solve expressions **
  ** *** done within KheDrsExprOpen now
  HaArrayForEach(drs->open_exprs, e, i)
  {
    KheDrsExprSetSigIndexes(e, drs);
    KheDrsExprAddExpressionsToDays(e, drs);
  }
  *** **

  if( DEBUG5 )
    KheDynamicResourceSolverDebug(drs, 1, 2, stderr);
  if( DEBUG10 )
    HaArrayForEach(drs->open_days, day, i)
      KheDrsDayDomTestsDebug(day, 1, 2, stderr);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolverClose(KHE_DYNAMIC_RESOURCE_SOLVER drs)                  */
/*                                                                           */
/*  Close drs, based on the final solution.                                  */
/*                                                                           */
/*****************************************************************************/

/* *** inlined now
static void KheDrsSolverClose(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_DAY day;  KHE_DRS_SOLN soln;  int i;  KHE_DRS_EDGE edge;
  KHE_DRS_TASK_ON_DAY dtd;  KHE_DRS_RESOURCE dr;  KHE_DRS_RESOURCE_ON_DAY drd;
  KHE_DRS_EXPR e;

  ** extract final solution **
  HaArrayClear(drs->solve_final_solns);
  day = HaArrayLast(drs->open_days);
  KheDrsSolnSetExtractFinalSoln(day->soln_set, drs);
  HnAssert(HaArrayCount(drs->solve_final_solns) <= 1,
    "KheDynamicResourceSolverSolve internal error");
  if( DEBUG2 )
  {
    fprintf(stderr, "  final solutions:\n");
    HaArrayForEach(drs->solve_final_solns, soln, i)
      KheDrsSolnDebug(soln, drs, 1, 2, stderr);
  }

  ** traverse final solution, closing resources and tasks on each day **
  soln = HaArrayFirst(drs->solve_final_solns);
  while( soln != NULL )
  {
    day = soln->day;
    edge = soln->prev_edge;
    HaArrayForEach(edge->tasks, dtd, i)
    {
      dr = HaArray(drs->open_resources, i);
      drd = HaArray(dr->days, day->frame_index);
      KheDrsResourceOnDayClose(drd, dtd);
      if( dtd != NULL )
	KheDrsTaskClose(dtd->encl_dt, drd->encl_dr);
    }
    soln = soln->prev_soln;
  }

  ** close each day **
  HaArrayForEach(drs->open_days, day, i)
    KheDrsDayClose(day, drs);

  ** close the open expressions **
  ** correct order here sti ll to do **
  HaArrayForEach(drs->open_exprs, e, i)
    KheDrsExprClose(e, drs);

  ** close the open resources **
  HaArrayForEach(drs->open_resources, dr, i)
    KheDrsResourceClose(dr);

  ** close drs **
  KheResourceSetClear(drs->selected_resource_set);
  HaArrayClear(drs->open_resources);
  HaArrayClear(drs->selected_day_ranges);
  HaArrayClear(drs->open_days);
  HaArrayClear(drs->open_exprs);
  drs->solve_diversifier++;
  HaArrayClear(drs->solve_final_solns);
  drs->solve_state = KHE_DRS_SOLVE_SOLVED;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolveOpen(KHE_DYNAMIC_RESOURCE_SOLVER drs,                    */
/*    KHE_DRS_DOM_KIND dom_kind)                                             */
/*                                                                           */
/*  Open drs for solving.                                                    */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolveOpen(KHE_DYNAMIC_RESOURCE_SOLVER drs, int soln_limit,
  KHE_DRS_DOM_KIND dom_kind, KHE_MARK *mark)
{
  KHE_DRS_DAY_RANGE ddr;  KHE_DRS_DAY day;  int i, j;  KHE_DRS_EXPR e;
  KHE_DRS_RESOURCE dr;  KHE_RESOURCE r;

  /* boilerplate */
  drs->solve_init_cost = KheSolnCost(drs->soln);
  drs->solve_soln_limit = (soln_limit <= 0 ? INT_MAX : soln_limit);
  *mark = KheMarkBegin(drs->soln);

  /* clear the open resources, open days, and open expressions */
  HaArrayClear(drs->open_resources);
  HaArrayClear(drs->open_days);
  HaArrayClear(drs->open_exprs);

  /* open selected resources (adds drs->open_resources and drs->open_exprs) */
  for( i = 0; i < KheResourceSetResourceCount(drs->selected_resource_set); i++ )
  {
    r = KheResourceSetResource(drs->selected_resource_set, i);
    dr = HaArray(drs->all_resources, KheResourceResourceTypeIndex(r));
    KheDrsResourceOpen(dr, drs);
    HaArrayAddLast(drs->open_resources, dr);
  }

  /* open the selected days (adds drs->open_days and drs->open_exprs) */
  HaArrayForEach(drs->selected_day_ranges, ddr, i)
    for( j = ddr.first;  j <= ddr.last;  j++ )
    {
      day = HaArray(drs->all_days, j);
      KheDrsDayOpen(day, ddr, dom_kind, drs, HaArrayCount(drs->open_days));
      HaArrayAddLast(drs->open_days, day);
    }

  /* sort drs->open_exprs by increasing postorder index, then open them */
  HaArraySort(drs->open_exprs, &KheDrsExprCmp);
  HaArrayForEach(drs->open_exprs, e, i)
    KheDrsExprOpen(e, drs);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolveDoSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs,                 */
/*    KHE_DRS_DOM_KIND dom_kind)                                             */
/*                                                                           */
/*  Carry out the actual solve.                                              */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolveDoSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_DRS_DOM_KIND dom_kind)
{
  KHE_DRS_DAY new_day;  KHE_DRS_SOLN_SET dns;
  int i;  KHE_DRS_SOLN soln;

  dns = KheDrsSolnSetMake(dom_kind, drs);
  soln = KheDrsSolnMake(NULL, NULL, 0 /* cost of root soln still to do */, drs);
  KheDrsSolnSetAddSoln(dns, soln, drs);
  HaArrayForEach(drs->open_days, new_day, i)
  {
    KheDrsSolnSetExtend(dns, new_day, drs);
    dns = new_day->soln_set;
#if TESTING
    HaArray(drs->running_time_per_day, i) = KheTimerElapsedTime(drs->timer);
#endif
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolveClose(KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_MARK mark,    */
/*    bool testing_only)                                                     */
/*                                                                           */
/*  Close drs for solving.  If testing_only, this is for testing only, so    */
/*  do not adopt any solution that may have been found.                      */
/*                                                                           */
/*****************************************************************************/

static void KheDrsSolveClose(KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_MARK mark,
  bool testing_only)
{
  KHE_DRS_DAY day;  int i;  KHE_DRS_EXPR e;  KHE_DRS_SOLN soln, final_soln;
  KHE_DRS_TASK_ON_DAY dtd;  KHE_DRS_RESOURCE dr;  KHE_DRS_RESOURCE_ON_DAY drd;

  /* extract final_soln */
  day = HaArrayLast(drs->open_days);
  final_soln = NULL;
  KheDrsSolnSetExtractFinalSoln(day->soln_set, &final_soln);
  if( DEBUG2 )
  {
    if( final_soln != NULL )
    {
      fprintf(stderr, "  final solution:\n");
      KheDrsSolnDebug(final_soln, drs, 1, 2, stderr);
    }
    else
      fprintf(stderr, "  no final solution\n");
  }

  if( testing_only || final_soln == NULL ||
      final_soln->cost >= drs->solve_init_cost )
  {
    /* testing, or no final soln, or no improvement; return to initial soln */
    KheMarkEnd(mark, true);
    HnAssert(KheSolnCost(drs->soln) == drs->solve_init_cost,
      "KheDynamicResourceSolverSolve internal error 2");
  }
  else
  {
    /* improvement; traverse final solution, closing resources and tasks */
    for( soln = final_soln;  soln->prev_soln != NULL;  soln = soln->prev_soln )
    {
      HaArrayForEach(soln->prev_tasks, dtd, i)
      {
	if( dtd != NULL )
	{
	  dr = HaArray(drs->open_resources, i);
	  drd = KheDrsResourceOnDay(dr, soln->day);
	  KheDrsTaskClose(dtd->encl_dt, drd->encl_dr);
	}
      }
    }
    KheMarkEnd(mark, false);
    HnAssert(KheSolnCost(drs->soln) == final_soln->cost,
      "KheDynamicResourceSolverSolve internal error 3");
    HnAssert(KheSolnCost(drs->soln) < drs->solve_init_cost,
      "KheDynamicResourceSolverSolve internal error 4");
  }

  /* close the open days */
  HaArrayForEach(drs->open_days, day, i)
    KheDrsDayClose(day, drs);

  /* close the open expressions */
  HaArrayForEach(drs->open_exprs, e, i)
    KheDrsExprClose(e, drs);

  /* close drs */
  HaArrayClear(drs->open_resources);
  HaArrayClear(drs->open_days);
  HaArrayClear(drs->open_exprs);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDynamicResourceSolverSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs,      */
/*    int soln_limit, KHE_DRS_DOM_KIND dom_kind)                             */
/*                                                                           */
/*  Solve drs.  Adopt the best solution found and return true if it is an    */
/*  improvement on the original.                                             */
/*                                                                           */
/*****************************************************************************/

bool KheDynamicResourceSolverSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int soln_limit, KHE_DRS_DOM_KIND dom_kind)
{
  KHE_COST final_cost;  bool res;  KHE_MARK mark;
  if( DEBUG1 )
    fprintf(stderr, "[ KheDynamicResourceSolverSolve(drs)\n");

  /* open drs, solve, and close drs */
  KheDrsSolveOpen(drs, soln_limit, dom_kind, &mark);
  KheDrsSolveDoSolve(drs, dom_kind);
  KheDrsSolveClose(drs, mark, false);

  /* tidy up and return */
  drs->solve_diversifier++;
  KheResourceSetClear(drs->selected_resource_set);
  HaArrayClear(drs->selected_day_ranges);
  final_cost = KheSolnCost(drs->soln);
  HnAssert(final_cost <= drs->solve_init_cost, "KheDynamicResourceSolverSolve: "
    "final cost (%.5f) exceeds initial cost (%.5f)", KheCostShow(final_cost),
    KheCostShow(drs->solve_init_cost));
  res = (final_cost < drs->solve_init_cost);
  if( DEBUG1 )
    fprintf(stderr, "] KheDynamicResourceSolverSolve returning %s\n",
      bool_show(res));
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "testing"                                                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void TestName(KHE_INSTANCE ins, char *suffix, char buff[200])            */
/*                                                                           */
/*  Set buff to the name of a test with these attributes.                    */
/*                                                                           */
/*****************************************************************************/

#if TESTING
static void TestName(KHE_INSTANCE ins, char *suffix, char buff[200])
{
  int len;

  /* add standard prefix to buff */
  len = snprintf(buff, 200, "%s", KheInstanceId(ins));

  /* add suffix */
  if( strcmp(suffix, "") != 0 )
    len += snprintf(&buff[len], 200 - len, "-%s", suffix);
}
#endif


/*****************************************************************************/
/*                                                                           */
/*  KHE_DATASET_POINTS_TYPE KheDomKindToPointsType(KHE_DRS_DOM_KIND dom_kind)*/
/*                                                                           */
/*  Return the dataset points type corresponding to dom_kind.                */
/*                                                                           */
/*****************************************************************************/

static KHE_DATASET_POINTS_TYPE KheDomKindToPointsType(KHE_DRS_DOM_KIND dom_kind)
{
  switch( dom_kind )
  {
    case KHE_DRS_DOM_NONE:	return KHE_DATASET_POINTS_TRIANGLE;
    case KHE_DRS_DOM_WEAK:	return KHE_DATASET_POINTS_SQUARE;
    case KHE_DRS_DOM_MEDIUM:	return KHE_DATASET_POINTS_CIRCLE;
    case KHE_DRS_DOM_STRONG:	return KHE_DATASET_POINTS_PLUS;
    case KHE_DRS_DOM_TRIE:	return KHE_DATASET_POINTS_CROSS;

    default:

      HnAbort("KheDomKindToPointsType: unknown dom_kind (%d)", dom_kind);
      return KHE_DATASET_POINTS_CROSS;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheDomKindLabel(KHE_DRS_DOM_KIND dom_kind)                         */
/*                                                                           */
/*  Return a suitable label for dom_kind.                                    */
/*                                                                           */
/*****************************************************************************/

static char *KheDomKindLabel(KHE_DRS_DOM_KIND dom_kind)
{
  switch( dom_kind )
  {
    case KHE_DRS_DOM_NONE:	return "None";
    case KHE_DRS_DOM_WEAK:	return "Weak";
    case KHE_DRS_DOM_MEDIUM:	return "Medium";
    case KHE_DRS_DOM_STRONG:	return "Strong";
    case KHE_DRS_DOM_TRIE:	return "Trie";

    default:

      HnAbort("KheDomKindLabel: unknown dom_kind (%d)", dom_kind);
      return NULL;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDrsPrintPoint(KHE_DYNAMIC_RESOURCE_SOLVER drs, int i)            */
/*                                                                           */
/*  Return true if we want to print a point with index i.                    */
/*                                                                           */
/*****************************************************************************/

static bool KheDrsPrintPoint(KHE_DYNAMIC_RESOURCE_SOLVER drs, int i)
{
  return HaArrayCount(drs->open_days) <= 28 || i % 2 == 0 ||
    i == HaArrayCount(drs->open_days) - 1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheBelowCaption(KHE_DYNAMIC_RESOURCE_SOLVER drs, char buff[200])    */
/*                                                                           */
/*  Print a suitable value for the below caption onto buff.                  */
/*                                                                           */
/*****************************************************************************/

static void KheBelowCaption(KHE_DYNAMIC_RESOURCE_SOLVER drs, char buff[200])
{
  int pos, i;  KHE_DRS_DAY_RANGE ddr;  KHE_TIME_GROUP tg1, tg2;

  /* print day ranges */
  pos = 0;
  HaArrayForEach(drs->selected_day_ranges, ddr, i)
  {
    if( i > 0 )
    {
      sprintf(&buff[pos], ", ");
      pos = strlen(buff);
    }
    tg1 = KheFrameTimeGroup(drs->days_frame, ddr.first);
    tg2 = KheFrameTimeGroup(drs->days_frame, ddr.last);
    if( tg1 == tg2 )
      sprintf(&buff[pos], "%s", KheTimeGroupId(tg1));
    else
      sprintf(&buff[pos], "%s-%s", KheTimeGroupId(tg1), KheTimeGroupId(tg2));
    pos = strlen(buff);
    if( pos > 150 )
    {
      sprintf(&buff[pos], "...");
      pos = strlen(buff);
      break;
    }
  }

  /* print resource count */
  sprintf(&buff[pos], " (%d resources)",
    KheResourceSetResourceCount(drs->selected_resource_set));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverTest(KHE_DYNAMIC_RESOURCE_SOLVER drs,       */
/*    int soln_limit)                                                        */
/*                                                                           */
/*  Test drs.                                                                */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverTest(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int soln_limit)
{
#if TESTING
  KHE_DRS_DOM_KIND dom_kind;  int di, i;  KHE_MARK mark;
  char time_file_name[200], size_file_name[200], below_caption[200];
  KHE_FILE time_file, size_file;  KHE_GRAPH time_graph, size_graph;
  KHE_DATASET_POINTS_TYPE points_type; KHE_DATASET_PAIRS_TYPE pairs_type;
  KHE_DATASET kd;  char *label;  float val;  KHE_DRS_DAY day;

  /* initialize the time file and graph */
  KheBelowCaption(drs, below_caption);
  TestName(KheSolnInstance(drs->soln), "time", time_file_name);
  time_file = KheFileBegin(time_file_name, KHE_FILE_LOUT_STANDALONE);
  time_graph = KheGraphBegin(time_file);
  KheGraphSetBelowCaption(time_graph, below_caption);
  KheGraphSetLeftCaptionAndGap(time_graph, "Secs", "1.0c");

  /* initialize the table size file and graph */
  TestName(KheSolnInstance(drs->soln), "size", size_file_name);
  size_file = KheFileBegin(size_file_name, KHE_FILE_LOUT_STANDALONE);
  size_graph = KheGraphBegin(size_file);
  KheGraphSetBelowCaption(size_graph, below_caption);
  KheGraphSetLeftCaptionAndGap(size_graph, "Size", "1.0c");

  /* one solve for each kind of dominance testing */
  for( di = KHE_DRS_DOM_NONE;  di <= KHE_DRS_DOM_TRIE;  di++ )
  {
    dom_kind = (KHE_DRS_DOM_KIND) di;

    /* clear out running time stuff */
    HaArrayClear(drs->running_time_per_day);
    HaArrayFill(drs->running_time_per_day, HaArrayCount(drs->open_days), 0.0);
    KheTimerResetStartTime(drs->timer);

    /* solve drs for dominance test i */
    KheDrsSolveOpen(drs, soln_limit, dom_kind, &mark);
    KheDrsSolveDoSolve(drs, dom_kind);
    KheDrsSolveClose(drs, mark, true);

    /* sort out the points type and pairs type */
    points_type = KheDomKindToPointsType(dom_kind);
    pairs_type = KHE_DATASET_PAIRS_DASHED;
    label = KheDomKindLabel(dom_kind);

    /* add one data set to the running time graph */
    kd = KheDataSetAdd(time_graph, points_type, pairs_type, label);
    HaArrayForEach(drs->running_time_per_day, val, i)
      if( KheDrsPrintPoint(drs, i) )
        KhePointAdd(kd, (float) i + 1, val);

    /* add one data set to the table size graph */
    kd = KheDataSetAdd(size_graph, points_type, pairs_type, label);
    KhePointAdd(kd, 0.0, 0.0);
    HaArrayForEach(drs->open_days, day, i)
      if( KheDrsPrintPoint(drs, i) )
        KhePointAdd(kd, (float) i + 1, day->soln_set->count);
  }

  /* tidy up graphs and stats files */
  KheGraphEnd(time_graph);
  KheFileEnd(time_file);
  KheGraphEnd(size_graph);
  KheFileEnd(size_file);
#else
  HnAbort("KheDynamicResourceSolverTest:  TESTING flag in "
    "khe_sr_dynamic_resource.c not set");
#endif
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDynamicResourceSolverSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs,      */
/*    KHE_DRS_DOM_KIND dom_kind)                                             */
/*                                                                           */
/*  Solve drs.  Adopt the best solution found and return true if it is an    */
/*  improvement on the original.                                             */
/*                                                                           */
/*****************************************************************************/

/* *** old version
bool KheDynamicResourceSolverSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_DRS_DOM_KIND dom_kind)
{
  int i;  KHE_DRS_DAY new_day;  KHE_DRS_SOLN_SET dns;
  KHE_COST init_cost, final_cost;  bool res;

  ** initialize drs and its open_resources for solving **
  ** drs->dom_kind = dom_kind; **
  ** drs->min_assts = min_assts; **
  ** drs->max_assts = max_assts; **
  ** drs->cost_limit = cost_limit; **
  ** *** replacing this is sti ll to do
  HaArrayForEach(drs->open_resources, dr, i)
    KheDrsResourceInitForSolving(dr, drs);
  *** **

  ** boilerplate **
  if( DEBUG1 )
    fprintf(stderr, "[ KheDynamicResourceSolverSolve(drs)\n");
  HnAssert(drs->solve_state == KHE_DRS_SOLVE_LOADING,
    "KheDynamicResourceSolverSolve called out of order");
  init_cost = KheSolnCost(drs->soln);

  ** open drs for solving **
  KheDrsSolverOpen(drs);

  ** ***
  ** initialize the day and resource_on_day objects for solving **
  HaArrayForEach(drs->open_days, day, i)
  {
    ** KheDrsDayInitializeForSolving(day, drs); **

    ** initialize day->dom_tests and day->eq_dom_test_indexes to a single EQ **
    HaArrayClear(day->dom_tests);
    HaArrayAddLast(day->dom_tests, KheDrsDomTestMake(KHE_DRS_DOM_EQ));
    HaArrayClear(day->eq_dom_test_indexes);
    HaArrayAddLast(day->eq_dom_test_indexes, 0);

    ** extend day->dom_tests and day->eq_dom_test_indexes from the resources **
    ** sig_offset = 1; **  ** to leave room for the test just added **
    HaArrayForEach(drs->open_resources, dr, j)
    {
      ** drd = HaArray(dr->days, i); **
      ** drd->sig_offset = sig_offset; **
      ** HaArrayAppend(day->dom_tests, drd->dom_tests, k); **
      ** ***
      HaArrayForEach(drd->eq_dom_test_indexes, val, k)
	HaArrayAddLast(day->eq_dom_test_indexes, val + sig_offset);
      sig_offset += drd->sig_len;
      *** **
    }
  }
  *** **

  ** build up the solns for each day **
  ** this code is what we want; everything else needs to make it work **
  dns = KheDrsNod eSetMake(dom_kind, drs);
  KheDrsSolnSetAddSoln(dns, KheDrsSolnMakeRoot(drs), drs);
  HaArrayForEach(drs->open_days, new_day, i)
  {
    KheDrsSolnSetExtend(dns, new_day, drs);
    ** day = HaArray(drs->open_days, di); **
    dns = new_day->soln_set;
  }

  ** build up the solns for each day **
  ** ***
  dns = KheDrsN odeSetMake(dom_kind, drs);
  soln = KheDrsSolnMakeRoot(drs);
  KheDrsSolnSetAddSoln(dns, soln, drs);
  ** prev_di = -1; **
  HaArrayForEach(drs->selected_day_ranges, open_ddr, i)
  {
    ** closed day range [prev_di + 1, ddr.first_day_index - 1] **
    ** *** obsolete
    if( prev_di < open_ddr.first_day_index - 1 )
    {
      closed_ddr = KheDrsDayRangeMake(prev_di + 1, open_ddr.first_day_index-1);
      KheDrsSolnSetExtendCl osed(dns, closed_ddr, drs);
      day = HaArray(drs->days, open_ddr.first_day_index - 1);
      dns = day->soln_set;
    }
    *** **

    ** open day range [ddr.first_day_index, ddr.last_day_index] **
    for( di = open_ddr.first_day_index;  di <= open_ddr.last_day_index;  di++ )
    {
      KheDrsSolnSetExtend(dns, di, drs);
      day = HaArray(drs->open_days, di);
      dns = day->soln_set;
    }
    ** prev_di = open_ddr.last_day_index; **
  }

  ** final closed day range [prev_di + 1, last_day] **
  ** *** obsolete
  if( prev_di < HaArrayCount(drs->open_days) - 1 )
  {
    closed_ddr =
      KheDrsDayRangeMake(prev_di + 1, HaArrayCount(drs->open_days) - 1);
    KheDrsSolnSetExtendCl osed(dns, closed_ddr, drs);
    day = HaArray(drs->open_days, HaArrayCount(drs->open_days) - 1);
    dns = day->soln_set;
  }
  *** **

  ** *** old version
  HaArrayForEach(drs->days, day, day_index)
  {
    if( prev_day == NULL )
    {
      ** virtual prev_day contains a single soln (an empty one) **
      if( DEBUG2 )
	fprintf(stderr, "  extending from day 0 (1 soln)\n");
      KheDrsSolnExtend(soln, day_index, drs);
    }
    else
    {
      ** extend each soln from prev_day with tasks from day **
      KheDrsSolnSetExtend(prev_day->soln_set, day_index, drs);
    }
    prev_day = day;
  }
  *** **

** *** moved out of loop just above to make things clearer for now
#if TESTING
      if( DEBUG7 )
	fprintf(stderr, "  running_time_per_day = %d, days = %d, di = %d\n",
	  HaArrayCount(drs->running_time_per_day), HaArrayCount(drs->days),
	  day_index);
      HaArray(drs->running_time_per_day, day_index) =
	KheTimerElapsedTime(drs->timer);
#endif
      if( DEBUG2 )
	fprintf(stderr, "  extending from day %s (%d solns)\n",
	  KheTimeGroupId(prev_day->time_group),
	  prev_day->soln_set->count);
*** **

  ** close (including adopting the final solution) and wrap up **
  KheDrsSolverClose(drs);

  ** extract final solutions whose assignment count is in the desired range **
  ** set solve state to solved **

  ** clear out the solve expressions, resources, and intervals **
  ** *** all this should be left to KheDynamicResourceSolverClear
  HaArrayForEach(drs->open_exprs, e, i)
    KheDrsExprFreeInternal(e, drs);
  HaArrayClear(drs->open_exprs);
  HaArrayForEach(drs->open_resources, dr, i)
    KheDrsResourceEndSolving(dr);
  HaArrayClear(drs->open_resources);
  HaArrayClear(drs->selected_day_ranges);
  *** **

  ** adopt the unique best solution **
  ** sti ll to do **

  ** return true if found a solution, and set *cost to its cost **
  ** ***
  if( HaArrayCount(drs->solve_final_solns) == 1 )
  {
    soln = HaArrayFirst(drs->solve_final_solns);
    *cost = soln->cost;
    if( DEBUG1 )
      fprintf(stderr, "] KheDynamicResourceSolverSolve returning true\n");
    return true;
  }
  else
  {
    *cost = KheCost(INT_MAX, INT_MAX);
    if( DEBUG1 )
      fprintf(stderr, "] KheDynamicResourceSolverSolve returning false\n");
    return false;
  }
  *** **

  ** return **
  final_cost = KheSolnCost(drs->soln);
  HnAssert(final_cost <= init_cost, "KheDynamicResourceSolverSolve: final "
    "cost (%.5f) exceeds initial cost (%.5f)", KheCostShow(final_cost),
    KheCostShow(init_cost));
  res = (final_cost < init_cost);
  if( DEBUG1 )
    fprintf(stderr, "] KheDynamicResourceSolverSolve returning %s\n",
      bool_show(res));
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheDynamicResourceSolverTimetableCount(                              */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Return the number of timetables found.                                   */
/*                                                                           */
/*****************************************************************************/

/* *** no longer offered
int KheDynamicResourceSolverTimetableCount(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  HnAssert(drs->solve_state == KHE_DRS_SOLVE_SOLVED,
    "KheDynamicResourceSolverTimetableCount called out of order");
  return HaArrayCount(drs->solve_final_solns);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverTimetable(KHE_DYNAMIC_RESOURCE_SOLVER drs,  */
/*    int i, int *asst_count, KHE_COST *r_cost, KHE_COST *soln_cost)         */
/*                                                                           */
/*  Report on the i'th timetable found by the solver.                        */
/*                                                                           */
/*****************************************************************************/

/* *** no longer offered
void KheDynamicResourceSolverTimetable(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int i, int *asst_count, KHE_COST *cost)
{
  ** make sure there is an i'th solution **
  KHE_DRS_SOLN soln;
  HnAssert(drs->solve_state == KHE_DRS_SOLVE_SOLVED,
    "KheDynamicResourceSolverTimetable called out of order");
  HnAssert(0 <= i && i < HaArrayCount(drs->solve_final_solns),
    "KheDynamicResourceSolverTimetable:  i (%d) out of range (0 .. %d)",
    i, HaArrayCount(drs->solve_final_solns) - 1);

  ** return the relevant fields of the i'th solution **
  soln = HaArray(drs->solve_final_solns, i);
  *asst_count = HaArrayFirst(soln->signature);
  *cost = soln->cost;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAdoptedSolnDebug(KHE_DYNAMIC_RESOURCE_SOLVER drs,             */
/*    KHE_COST *cost)                                                        */
/*                                                                           */
/*  Debug print of r's timetable and costs.                                  */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAdoptedSolnDebug(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_COST *cost)
{
  int i, ri;  KHE_COST m_cost;  KHE_MONITOR m;  KHE_DRS_RESOURCE dr;
  KHE_RESOURCE_TIMETABLE_MONITOR rtm;  KHE_RESOURCE r;
  *cost = KheCost(0, 0);
  HaArrayForEach(drs->open_resources, dr, ri)
  {
    rtm = KheResourceTimetableMonitor(drs->soln, dr->resource);
    KheResourceTimetableMonitorPrintTimetable(rtm, 6, 2, stderr);
    for( i = 0;  i < KheSolnResourceMonitorCount(drs->soln, r);  i++ )
    {
      m = KheSolnResourceMonitor(drs->soln, r, i);
      switch( KheMonitorTag(m) )
      {
	case KHE_AVOID_UNAVAILABLE_TIMES_MONITOR_TAG:
	case KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG:
	case KHE_LIMIT_BUSY_TIMES_MONITOR_TAG:
	case KHE_LIMIT_WORKLOAD_MONITOR_TAG:
	case KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG:

	  m_cost = KheMonitorCost(m);
	  if( m_cost > 0 )
	  {
	    fprintf(stderr, "  true cost of %s is %.5f\n", KheMonitorId(m),
	      KheCostShow(m_cost));
	    *cost += m_cost;
	  }
	  break;

	default:

	  /* ignore monitors of other types */
	  break;
      }
    }
  }
  fprintf(stderr, "  total cost is %.5f\n", KheCostShow(*cost));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverAdopt(KHE_DYNAMIC_RESOURCE_SOLVER drs)      */
/*                                                                           */
/*  Change drs's solution to include the timetable found by the solver.      */
/*                                                                           */
/*****************************************************************************/

/* *** obsolete
void KheDynamicResourceSolverAdopt(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_DRS_SOLN soln, final_soln;  KHE_DRS_EDGE de;  KHE_DRS_TASK dt;
  KHE_DRS_RESOURCE dr;  int i;

  ** make sure there is a solution **
  HnAssert(drs->solve_state == KHE_DRS_SOLVE_SOLVED,
    "KheDynamicResourceSolverAdopt called out of order");
  HnAssert(HaArrayCount(drs->solve_final_solns) == 1,
    "KheDynamicResourceSolverAdopt:  no solution found");

  ** adopt the solution **
  ** *** sti ll to do here
  final_soln = HaArrayFirst(drs->solve_final_solns);
  for( soln = final_soln;  soln != NULL;  soln = soln->prev_soln )
  {
    de = soln->prev_edge;
    HaArrayForEach(de->tasks, dt, i)
      if( dt != NULL && !dt->adopted )
      {
	dr = HaArray(drs->open_resources, i);
	KheTaskAssignResource(dt->task, dr->resource);
	** dt->adopted = true; **
      }
  }
  *** **

  ** check that the resulting resource costs agree with the final soln **
  ** ***
  if( DEBUG3 )
  {
    fprintf(stderr, "[ KheDynamicResourceSolverAdopt(drs, %d)\n", 0);
    KheDrsAdoptedSolnDebug(drs, &cost);
    if( cost != final_soln->cost )
    {
      fprintf(stderr, "KheDynamicResourceSolverAdopt: cost discrepancy "
	"(true %.5f != dynamic %.5f):\n", KheCostShow(cost),
	KheCostShow(final_soln->cost));
      ** *** sti ll to do
      KheDrsFullSolnDebug(drs->full_soln, 1, 2, stderr);
      KheDrsFullSolnUnAdopt(drs->full_soln);
      rerun_soln = KheFullSolnReRun(drs->full_soln, drs);
      *** **
      rerun_soln = NULL;  ** sti ll to do **
      fprintf(stderr, "  rerun_soln cost is %.5f\n", KheCostShow(rerun_soln->cost));
      HnAbort("aborting now (cost discrepancy)");
    }
    fprintf(stderr, "  ?? cost %.5f\n", KheCostShow(cost));
    ** ***
    fprintf(stderr, "  ?? assts %d, cost %.5f\n",
      HaArrayFirst(final_soln->signature), KheCostShow(cost));
    fprintf(stderr, "  %s assts %d, cost %.5f\n", KheResourceId(drs->resource),
      HaArrayFirst(final_soln->signature), KheCostShow(cost));
    *** **
    fprintf(stderr, "] KheDynamicResourceSolverAdopt returning\n");
  }
  *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverUnAdopt(KHE_DYNAMIC_RESOURCE_SOLVER drs)    */
/*                                                                           */
/*  Unadopt the most recently adopted full solution.                         */
/*                                                                           */
/*****************************************************************************/

/* ***
void KheDynamicResourceSolverUnAdopt(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  ** *** st ill to do
  KheDrsFullSolnUnAdopt(drs->full_soln);
  *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheDrsSolnAsstCount(KHE_DRS_SOLN soln)                                  */
/*                                                                           */
/*  Return the assignment count of soln.  (First element of the signature.)  */
/*                                                                           */
/*****************************************************************************/

/* ***
static int KheDrsSolnAsstCount(KHE_DRS_SOLN soln)
{
  return HaArrayFirst(soln->signature);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_COST KheDrsSolnAdjustedCost(KHE_DRS_SOLN soln, KHE_COST c)              */
/*                                                                           */
/*  Return the adjusted cost of soln.                                        */
/*                                                                           */
/*****************************************************************************/

/* ***
static KHE_COST KheDrsSolnAdjustedCost(KHE_DRS_SOLN soln, KHE_COST c)
{
  return soln->cost - KheDrsSolnAsstCount(soln) * c;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheDynamicResourceSolverBest(KHE_DYNAMIC_RESOURCE_SOLVER drs,        */
/*    KHE_COST cost_reduction)                                               */
/*                                                                           */
/*  Return the index of the final solution which has a mminimal value of     */
/*                                                                           */
/*    r_cost - asst_count * cost_reduction                                   */
/*                                                                           */
/*  Break ties in favour of final solutions with larger asst_count values.   */
/*                                                                           */
/*****************************************************************************/

/* ***
int KheDynamicResourceSolverBest(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_COST cost_reduction)
{
  int best_i, i, asst_count, best_asst_count;  KHE_DRS_SOLN soln, best_soln;
  KHE_COST adjusted_cost, best_adjusted_cost;  
  HnAssert(drs->solve_state == KHE_DRS_SOLVE_SOLVED,
    "KheDynamicResourceSolverBest called out of order");
  HnAssert(HaArrayCount(drs->solve_final_solns) > 0,
    "KheDynamicResourceSolverBest:  there are no solutions");
  best_soln = NULL;
  best_i = -1;
  best_adjusted_cost = 0;
  best_asst_count = 0;
  HaArrayForEach(drs->solve_final_solns, soln, i)
  {
    asst_count = KheDrsSolnAsstCount(soln);
    adjusted_cost = KheDrsSolnAdjustedCost(soln, cost_reduction);
    if( best_soln == NULL || adjusted_cost < best_adjusted_cost ||
	(adjusted_cost == best_adjusted_cost && asst_count > best_asst_count) )
    {
      best_soln = soln;
      best_i = i;
      best_adjusted_cost = adjusted_cost;
      best_asst_count = asst_count;
    }
  }
  return best_i;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverDebug(KHE_DYNAMIC_RESOURCE_SOLVER drs,      */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug printf of drs onto fp with the given verbosity and indent.         */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverDebug(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  int verbosity, int indent, FILE *fp)
{
  KHE_DRS_DAY day;  KHE_DRS_RESOURCE dr;  int i;
  HnAssert(indent >= 0, "KheDynamicResourceSolverDebug: indent < 0");
  fprintf(fp, "%*s[ DynamicResourceSolver(", indent, "");
  HaArrayForEach(drs->open_resources, dr, i)
  {
    if( i > 0 )
      fprintf(fp, ", ");
    fprintf(fp, "%s", KheResourceId(dr->resource));
  }
  fprintf(fp, ")\n");
  HaArrayForEach(drs->open_days, day, i)
    KheDrsDayDebug(day, drs, verbosity, indent + 2, fp);
  fprintf(fp, "%*s]\n", indent, "");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "manual testing"                                               */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_FULL_SOLN KheDrsManualSolnBegin(KHE_DYNAMIC_RESOURCE_SOLVER drs, */
/*    char *resource_id)                                                     */
/*                                                                           */
/*  Begin loading a manual solution for this resource.                       */
/*                                                                           */
/*****************************************************************************/

#if TESTING
/* ***
static KHE_DRS_FULL_SOLN KheDrsManualSolnBegin(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  char *resource_id)
{
  KHE_INSTANCE ins;  KHE_RESOURCE r;  KHE_DRS_FULL_SOLN res;
  ins = KheSolnInstance(drs->soln);
  if( !KheInstanceRetrieveResource(ins, resource_id, &r) )
    return NULL;
  res = KheDrsFullSolnMake(drs);
  res->resource = r;
  return res;
}
*** */
#endif


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsManualSolnAddTask(KHE_DRS_FULL_SOLN fs, char *time_id,        */
/*    KHE_DYNAMIC_RESOURCE_SOLVER drs)                                       */
/*                                                                           */
/*  Add the best task for time_id to fs.                                     */
/*                                                                           */
/*****************************************************************************/

#if TESTING
/* ***
static void KheDrsManualSolnAddTask(KHE_DRS_FULL_SOLN fs, char *time_id,
  KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_INSTANCE ins;  KHE_TIME time;  int day_index;  KHE_DRS_DAY day;
  KHE_DRS_TIME t;  int i;  KHE_DRS_TASK best_dt;
  ins = KheSolnInstance(drs->soln);
  if( !KheInstanceRetrieveTime(ins, time_id, &time) )
    HnAbort("KheDrsManualSolnAddTask:  unknown time %s", time_id);
  day_index = KheFrameTimeIndex(drs->days_frame, time);
  day = HaArray(drs->days, day_index);
  HaArrayForEach(day->times, t, i)
    if( t->time == time && HaArrayCount(t->best_tasks) > 0 )
    {
      best_dt = HaArrayFirst(t->best_tasks);
      if( KheDailyScheduleFirstDayIndex(best_dt->daily_schedule) ==
	  KheDailyScheduleLastDayIndex(best_dt->daily_schedule) )
      {
	HaArrayAddLast(fs->tasks, best_dt);
	return;
      }
    }
  HnAbort("KheDrsManualSolnAddTask:  no best task at time %s", time_id);
}
*** */
#endif


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsManualSolnTest1(KHE_DYNAMIC_RESOURCE_SOLVER drs)              */
/*                                                                           */
/*  Carry out a specific manual test, if the instance is INRC2-4-030-1-6291. */
/*                                                                           */
/*****************************************************************************/

#if TESTING
/* ***
static void KheDrsManualSolnTest1(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_INSTANCE ins;  KHE_DRS_FULL_SOLN fs;  KHE_COST cost;
  ins = KheSolnInstance(drs->soln);
  if( strcmp(KheInstanceId(ins), "INRC2-4-030-1-6291") == 0 )
  {
    fprintf(stderr, "[ KheDrsManualSolnTest1(%s, %s)\n",
      KheInstanceId(ins), "HN_0");
    fs = KheDrsManualSolnBegin(drs, "HN_0");
    KheDrsInitializeForSolving(drs, fs->resource, false, 0, INT_MAX,
      KheCost(1, 0));
    KheDrsManualSolnAddTask(fs, "1Wed1", drs);
    KheDrsManualSolnAddTask(fs, "1Thu1", drs);
    KheDrsManualSolnAddTask(fs, "1Fri1", drs);

    KheDrsManualSolnAddTask(fs, "2Mon2", drs);
    KheDrsManualSolnAddTask(fs, "2Tue2", drs);
    KheDrsManualSolnAddTask(fs, "2Wed2", drs);
    KheDrsManualSolnAddTask(fs, "2Thu2", drs);
    KheDrsManualSolnAddTask(fs, "2Fri2", drs);

    KheDrsManualSolnAddTask(fs, "3Mon1", drs);
    KheDrsManualSolnAddTask(fs, "3Tue1", drs);
    KheDrsManualSolnAddTask(fs, "3Wed1", drs);
    KheDrsManualSolnAddTask(fs, "3Thu3", drs);
    KheDrsManualSolnAddTask(fs, "3Fri3", drs);

    KheDrsManualSolnAddTask(fs, "4Mon2", drs);
    KheDrsManualSolnAddTask(fs, "4Tue2", drs);
    KheDrsManualSolnAddTask(fs, "4Wed2", drs);
    KheDrsManualSolnAddTask(fs, "4Thu2", drs);
    KheDrsManualSolnAddTask(fs, "4Fri2", drs);
    KheDrsFullSolnAdopt(fs);
    KheDrsAdoptedSolnDebug(drs, &cost);
    KheDrsFullSolnUnAdopt(fs);
    fprintf(stderr, "] KheDrsManualSolnTest1 returning\n");
  }
}
*** */
#endif


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsManualSolnTest2(KHE_DYNAMIC_RESOURCE_SOLVER drs)              */
/*                                                                           */
/*  Carry out a specific manual test, if the instance is INRC2-4-030-1-6291. */
/*                                                                           */
/*****************************************************************************/

#if TESTING
/* ***
static void KheDrsManualSolnTest2(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
  KHE_INSTANCE ins;  KHE_DRS_FULL_SOLN fs;  KHE_COST cost;
  ins = KheSolnInstance(drs->soln);
  if( strcmp(KheInstanceId(ins), "INRC2-4-030-1-6291") == 0 )
  {
    fprintf(stderr, "[ KheDrsManualSolnTest2(%s, %s)\n",
      KheInstanceId(ins), "HN_0");
    fs = KheDrsManualSolnBegin(drs, "HN_0");
    KheDrsInitializeForSolving(drs, fs->resource, false, 0, INT_MAX,
      KheCost(1, 0));
    KheDrsManualSolnAddTask(fs, "1Wed1", drs);
    KheDrsManualSolnAddTask(fs, "1Thu1", drs);
    KheDrsManualSolnAddTask(fs, "1Fri1", drs);

    KheDrsManualSolnAddTask(fs, "2Mon2", drs);
    KheDrsManualSolnAddTask(fs, "2Tue2", drs);
    KheDrsManualSolnAddTask(fs, "2Wed2", drs);
    KheDrsManualSolnAddTask(fs, "2Thu2", drs);
    KheDrsManualSolnAddTask(fs, "2Fri2", drs);

    KheDrsManualSolnAddTask(fs, "3Mon1", drs);
    KheDrsManualSolnAddTask(fs, "3Tue1", drs);
    KheDrsManualSolnAddTask(fs, "3Wed1", drs);
    KheDrsManualSolnAddTask(fs, "3Thu1", drs);
    KheDrsManualSolnAddTask(fs, "3Fri1", drs);

    KheDrsManualSolnAddTask(fs, "4Mon2", drs);
    KheDrsManualSolnAddTask(fs, "4Tue2", drs);
    KheDrsManualSolnAddTask(fs, "4Wed2", drs);
    KheDrsManualSolnAddTask(fs, "4Thu2", drs);
    KheDrsManualSolnAddTask(fs, "4Fri2", drs);
    KheDrsFullSolnAdopt(fs);
    KheDrsAdoptedSolnDebug(drs, &cost);
    KheDrsFullSolnUnAdopt(fs);
    fprintf(stderr, "] KheDrsManualSolnTest2 returning\n");
  }
}
*** */
#endif




/*****************************************************************************/
/*                                                                           */
/*  Submodule "testing"                                                      */
/*                                                                           */
/*****************************************************************************/

#if TESTING

typedef HA_ARRAY(KHE_COST) ARRAY_KHE_COST;
#endif


/*****************************************************************************/
/*                                                                           */
/*  char *KheDrsTestTypeShow(KHE_DRS_TEST_TYPE type)                         */
/*                                                                           */
/*  Show a test type.                                                        */
/*                                                                           */
/*****************************************************************************/

#if TESTING
/* ***
static char *KheDrsTestTypeShow(KHE_DRS_TEST_TYPE type)
{
  switch( type )
  {
    case KHE_DRS_TEST_WEAK:		return "w";
    case KHE_DRS_TEST_STRONG:		return "s";
    case KHE_DRS_TEST_WEAK_AND_STRONG:	return "ws";
    default:				return "?";
  }
}
*** */
#endif


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverDoTest(KHE_DYNAMIC_RESOURCE_SOLVER drs,     */
/*    KHE_RESOURCE r, KHE_DRS_TEST_TYPE test_type, int min_assts,            */
/*    int max_assts, ARRAY_KHE_COST *cost_limits)                            */
/*                                                                           */
/*  Do a test with these values for min_assts, max_assts, and cost_limits.   */
/*                                                                           */
/*****************************************************************************/

/* ***
#if TESTING
static void KheDynamicResourceSolverDoTest(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_RESOURCE r)
{
  char time_file_name[200], size_file_name[200], cost_file_name[200];
  KHE_FILE time_file, size_file, cost_file;
  KHE_GRAPH time_graph, size_graph, cost_graph;

  ** initialize the time file and graph **
  TestName(KheSolnInstance(drs->soln), "time", time_file_name);
  time_file = KheFileBegin(time_file_name, KHE_FILE_LOUT_STANDALONE);
  time_graph = KheGraphBegin(time_file);
  KheGraphSetBelowCaption(time_graph, "Days");
  KheGraphSetLeftCaptionAndGap(time_graph, "Secs", "1.0c");

  ** initialize the table size file and graph **
  TestName(KheSolnInstance(drs->soln), "size", size_file_name);
  size_file = KheFileBegin(size_file_name, KHE_FILE_LOUT_STANDALONE);
  size_graph = KheGraphBegin(size_file);
  KheGraphSetBelowCaption(size_graph, "Days");
  KheGraphSetLeftCaptionAndGap(size_graph, "Size", "1.0c");

  ** initialize the cost file and graph **
  TestName(KheSolnInstance(drs->soln), "cost", cost_file_name);
  cost_file = KheFileBegin(cost_file_name, KHE_FILE_LOUT_STANDALONE);
  cost_graph = KheGraphBegin(cost_file);
  KheGraphSetBelowCaption(cost_graph, "Assignments");
  KheGraphSetLeftCaptionAndGap(cost_graph, "Cost", "1.0c");

  ** test with strong, medium, and weak dominance **
  KheDrsSolveTest(drs, r, KHE_DRS_DOM_WEAK,
    time_graph, size_graph, cost_graph);
  KheDrsSolveTest(drs, r, KHE_DRS_DOM_MEDIUM,
    time_graph, size_graph, cost_graph);
  KheDrsSolveTest(drs, r, KHE_DRS_DOM_STRONG,
    time_graph, size_graph, cost_graph);
  KheDrsSolveTest(drs, r, KHE_DRS_DOM_TRIE,
    time_graph, size_graph, cost_graph);

  ** tidy up graphs and stats files **
  KheGraphEnd(time_graph);
  KheFileEnd(time_file);
  KheGraphEnd(size_graph);
  KheFileEnd(size_file);
  KheGraphEnd(cost_graph);
  KheFileEnd(cost_file);
}
#endif
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheMinConstraintWeight(KHE_SOLN soln, KHE_RESOURCE r,               */
/*    KHE_COST *min_weight)                                                  */
/*                                                                           */
/*  If r has at least one constraint whose cost will be optimized and        */
/*  whose weight is non-negative, return the minimum of those weights.       */
/*                                                                           */
/*****************************************************************************/

/* ***
static bool KheMinConstraintWeight(KHE_SOLN soln, KHE_RESOURCE r,
  KHE_COST *min_weight)
{
  int i;  KHE_CONSTRAINT c;  KHE_COST weight;
  *min_weight = KheCost(INT_MAX, INT_MAX);
  for( i = 0;  i < KheResourceConstraintCount(r);  i++ )
  {
    c = KheResourceConstraint(r, i);
    switch( KheConstraintTag(c) )
    {
      case KHE_AVOID_UNAVAILABLE_TIMES_CONSTRAINT_TAG:
      case KHE_CLUSTER_BUSY_TIMES_CONSTRAINT_TAG:
      case KHE_LIMIT_BUSY_TIMES_CONSTRAINT_TAG:
      case KHE_LIMIT_WORKLOAD_CONSTRAINT_TAG:
      case KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT_TAG:

	weight = KheConstraintCombinedWeight(c);
	if( weight > 0 && weight < *min_weight )
	  *min_weight = weight;
	break;

      default:

	** ignore other constraints **
	break;
    }
  }
  return *min_weight < KheCost(INT_MAX, INT_MAX);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsSolveTest(KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_RESOURCE r,    */
/*    KHE_DRS_DOM_KIND dom_kind, int min_assts, int max_assts,               */
/*    ARRAY_KHE_COST *cost_limits, KHE_GRAPH time_graph,                     */
/*    KHE_GRAPH size_graph, KHE_GRAPH cost_graph)                            */
/*                                                                           */
/*  Carry out a test with these attributes, and add a dataset to time_graph, */
/*  size_graph, and cost_graph recording the results.                        */
/*                                                                           */
/*****************************************************************************/

#if TESTING
static void KheDrsSolveTest(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_DRS_DOM_KIND dom_kind, KHE_GRAPH time_graph,
  KHE_GRAPH size_graph /* , KHE_GRAPH cost_graph */)
{
  int i /* , best_i, asst_count */;  float val;  KHE_DATASET kd;
  KHE_DRS_DAY day;
  /* KHE_COST r_cost, marginal_cost; */  /* KHE_BALANCE_SOLVER bs; */
  KHE_DATASET_POINTS_TYPE points_type; KHE_DATASET_PAIRS_TYPE pairs_type;
  char *label;  /* KHE_DRS_SOLN soln; */  char buff[200];  /* KHE_COST cost; */
  if( DEBUG11 )
  {
    TestName(KheSolnInstance(drs->soln), KheDomKindLabel(dom_kind), buff);
    fprintf(stderr, "[ KheDrsSolveTest(%s)\n", buff);
  }

  /* solve, measuring running time */
  HaArrayClear(drs->running_time_per_day);
  HaArrayFill(drs->running_time_per_day, HaArrayCount(drs->open_days), 0.0);
  KheTimerResetStartTime(drs->timer);
  KheDynamicResourceSolverSolve(drs, 0, dom_kind);

  /* clear out for another solve with the same resources and day intervals */
  /* KheDynamicResourceSolverClear(drs, false); */

  /* verify costs */
  /* *** still to do
  for( i = 0;  i < KheDynamicResourceSolverTimetableCount(drs);  i++ )
  {
    KheDynamicResourceSolverAdopt(drs, i);
    KheDynamicResourceSolverUnAdopt(drs);
  }
  *** */

  /* ***
  if( DEBUG11 )
  {
    bs = KheBalanceSolverMake(drs->soln, drs->options,
      KheResourceResourceType(r), drs->arena);
    KheBalanceSolverDebug(bs, 2, 2, stderr);
    marginal_cost = KheBalanceSolverMarginalCost(bs);
    fprintf(stderr, "  marginal cost = %.5f\n", KheCostShow(marginal_cost));
    ** ***
    if( KheDynamicResourceSolverTimetableCount(drs) > 0 )
    {
      best_i = KheDynamicResourceSolverBest(drs, marginal_cost);
      fprintf(stderr, "  best of best has index %d\n", best_i);
    }
    *** **
    best_i = 0;  ** temporary, to keep compiler happy **
    HaArrayForEach(drs->solve_final_solns, soln, i)
      fprintf(stderr, "  best soln has cost %.5f%s\n",
	KheCostShow(soln->cost), i == best_i ? " (overall best)" : "");
      ** ***
      fprintf(stderr, "  best soln (%d assts) has cost %.5f%s\n",
	HaArrayFirst(soln->signature), KheCostShow(soln->cost),
	i == best_i ? " (overall best)" : "");
      *** **
  }
  *** */

  /* sort out the points type and pairs type */
  points_type = KheDomKindToPointsType(dom_kind);
  pairs_type = KHE_DATASET_PAIRS_DASHED;
  label = KheDomKindLabel(dom_kind);

  /* add one data set to the running time graph */
  kd = KheDataSetAdd(time_graph, points_type, pairs_type, label);
  HaArrayForEach(drs->running_time_per_day, val, i)
    if( KheDrsPrintPoint(drs, i) )
      KhePointAdd(kd, (float) i + 1, val);

  /* add one data set to the table size graph */
  kd = KheDataSetAdd(size_graph, points_type, pairs_type, label);
  KhePointAdd(kd, 0.0, 0.0);
  HaArrayForEach(drs->open_days, day, i)
    if( KheDrsPrintPoint(drs, i) )
      KhePointAdd(kd, (float) i + 1, day->soln_set->count);

  /* add one data set to the cost graph */
  /* ***
  kd = KheDataSetAdd(cost_graph, points_type, pairs_type, label);
  for( i = 0;  i < KheDynamicResourceSolverTimetableCount(drs);  i++ )
  {
    KheDynamicResourceSolverTimetable(drs, i, &asst_count, &r_cost);
    KhePointAdd(kd, (float) asst_count, (float) KheSoftCost(r_cost));
  }
  *** */

  if( DEBUG11 )
    fprintf(stderr, "] KheDrsSolveTest returning\n");
}
#endif


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverTestSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs)  */
/*                                                                           */
/*  Carry out some test solves.                                              */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceSolverTestSolve(KHE_DYNAMIC_RESOURCE_SOLVER drs)
{
#if TESTING
  char time_file_name[200], size_file_name[200] /* , cost_file_name[200] */;
  KHE_FILE time_file, size_file /* , cost_file */;
  KHE_GRAPH time_graph, size_graph /* , cost_graph */;

  /* ***
  HnAssert(drs->solve_state == KHE_DRS_SOLVE_LOADING,
    "KheDynamicResourceSolverTestSolve called out of order");
  *** */
  HnAssert(KheResourceSetResourceCount(drs->selected_resource_set) > 0,
    "KheDynamicResourceSolverTestSolve: no resources");
  HnAssert(HaArrayCount(drs->selected_day_ranges) > 0,
    "KheDynamicResourceSolverTestSolve: no day ranges");

  /* initialize the time file and graph */
  TestName(KheSolnInstance(drs->soln), "time", time_file_name);
  time_file = KheFileBegin(time_file_name, KHE_FILE_LOUT_STANDALONE);
  time_graph = KheGraphBegin(time_file);
  KheGraphSetBelowCaption(time_graph, "Days");
  KheGraphSetLeftCaptionAndGap(time_graph, "Secs", "1.0c");

  /* initialize the table size file and graph */
  TestName(KheSolnInstance(drs->soln), "size", size_file_name);
  size_file = KheFileBegin(size_file_name, KHE_FILE_LOUT_STANDALONE);
  size_graph = KheGraphBegin(size_file);
  KheGraphSetBelowCaption(size_graph, "Days");
  KheGraphSetLeftCaptionAndGap(size_graph, "Size", "1.0c");

  /* initialize the cost file and graph */
  /* ***
  TestName(KheSolnInstance(drs->soln), "cost", cost_file_name);
  cost_file = KheFileBegin(cost_file_name, KHE_FILE_LOUT_STANDALONE);
  cost_graph = KheGraphBegin(cost_file);
  KheGraphSetBelowCaption(cost_graph, "Assignments");
  KheGraphSetLeftCaptionAndGap(cost_graph, "Cost", "1.0c");
  *** */

  /* test with weak, medium, strong, and trie dominance */
  KheDrsSolveTest(drs, KHE_DRS_DOM_NONE, time_graph, size_graph);
  KheDrsSolveTest(drs, KHE_DRS_DOM_WEAK, time_graph, size_graph);
  KheDrsSolveTest(drs, KHE_DRS_DOM_MEDIUM, time_graph, size_graph);
  KheDrsSolveTest(drs, KHE_DRS_DOM_STRONG, time_graph, size_graph);
  KheDrsSolveTest(drs, KHE_DRS_DOM_TRIE, time_graph, size_graph);

  /* tidy up graphs and stats files */
  KheGraphEnd(time_graph);
  KheFileEnd(time_file);
  KheGraphEnd(size_graph);
  KheFileEnd(size_file);
  /* KheGraphEnd(cost_graph); */
  /* KheFileEnd(cost_file); */
  /* remainder still to do */
#else
  HnAbort("KheDynamicResourceSolverTestSolve: TESTING compile flag not set");
#endif
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceSolverTest(KHE_SOLN soln,                         */
/*    KHE_RESOURCE_TYPE rt, KHE_OPTIONS options, KHE_RESOURCE r1,            */
/*    KHE_RESOURCE r2, KHE_RESOURCE r3, KHE_RESOURCE r4,                     */
/*    int first_day_index1, int last_day_index1,                             */
/*    int first_day_index2, int last_day_index2,                             */
/*    int first_day_index3, int last_day_index3,                             */
/*    int first_day_index4, int last_day_index4)                             */
/*                                                                           */
/*  Test dynamic resource solving on r1, r2, r3, and r4.  Some may be NULL,  */
/*  allowing testing on fewer than 4 resources.                              */
/*                                                                           */
/*****************************************************************************/

/* ***
void KheDynamicResourceSolverTest(KHE_SOLN soln,
  KHE_RESOURCE_TYPE rt, KHE_OPTIONS options, KHE_RESOURCE r1,
  KHE_RESOURCE r2, KHE_RESOURCE r3, KHE_RESOURCE r4,
  int first_day_index1, int last_day_index1,
  int first_day_index2, int last_day_index2,
  int first_day_index3, int last_day_index3,
  int first_day_index4, int last_day_index4)
{
#if TESTING
  KHE_DYNAMIC_RESOURCE_SOLVER drs;  int max_busy_times;  KHE_RESOURCE r;

  ** make a solver **
  r = r1;  ** temporary hack, sti ll to do **
  drs = KheDynamicResourceSolverMake(soln, rt, options);

  ** do the test **
  KheDynamicResourceSolverDoTest(drs, r);
  if( KheResourceMaxBusyTimes(drs->soln, r, &max_busy_times) )
  {
    ** unlimited cost, limited busy days range **
    KheDynamicResourceSolverDoTest(drs, r);
  }

  ** manual tests and return **
  ** ***
  if( DEBUG8 )
  {
    KheDrsManualSolnTest1(drs);
    KheDrsManualSolnTest2(drs);
  }
  *** **
  KheDynamicResourceSolverDelete(drs);
#else
  fprintf(stderr, "to test, set TESTING to 1 in khe_sr_dynamic_resource.c"\n);
#endif
}
*** */
